/* advautomesh_tetra3d_particle_view.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>
#include <autogl.h>
#include <automesh.h>



#ifdef TIME_MEASURE
#include "time_measure.c"
#endif



static AutoMesh_Node3D *Node;
static AutoMesh_VolumeElement3D *Element;
static AutoMesh_NodeVector3D *Vector;
static AutoMesh_VolumeParticle3D *Particle;
static AutoMesh_Frame3D *Frame;

static double CenterX, CenterY, CenterZ;
static double Offset;
static int NSteps = 100;
static double DeltaT = 0.1;

static void SetView (void)
{
  double minX, minY, minZ;
  double maxX, maxY, maxZ;
  double centerX, centerY, centerZ;
  double size;
  
  AutoMesh_Node3D_getRange 
    (Node, 
     &minX, &minY, &minZ,
     &maxX, &maxY, &maxZ);
  centerX = (minX + maxX) / 2;
  centerY = (minY + maxY) / 2;
  centerZ = (minZ + maxZ) / 2;
  size = AutoMesh_Distance3D (minX, minY, minZ,
			      maxX, maxY, maxZ) / 2;

  AutoGL_SetViewSize (size);
  AutoGL_SetViewCenter (centerX, centerY, centerZ); 

  fprintf (stderr, " minX %f minY %f minZ %f \n",
	   minX, minY, minZ);
  fprintf (stderr, " maxX %f maxY %f maxZ %f \n",
	   maxX, maxY, maxZ);

  CenterX = centerX;
  CenterY = centerY;
  CenterZ = centerZ;
  Offset = size / 20;
}

static void MakeParticle (void)
{
  int ix, iy, iz;

  AutoMesh_VolumeParticle3D_allocateParticle 
    (Particle, 3 * 3 * 3, NSteps);

  for (ix = 0; ix < 3; ix++) {
    double x = CenterX + Offset * (ix - 1);

    for (iy = 0; iy < 3; iy++) {
      double y = CenterY + Offset * (iy - 1);

      for (iz = 0; iz < 3; iz++) {
	double z = CenterZ + Offset * (iz - 1);
	int particleId = ix * 3 * 3 + iy * 3 + iz;

	AutoMesh_VolumeParticle3D_setParticleStartPointAt 
	  (Particle, particleId, x, y, z);
      }
    }
  }

  AutoMesh_VolumeParticle3D_traceParticle 
    (Particle, Vector, DeltaT);
}

static void DrawFrame (void) 
{
  int nFrames = AutoMesh_Frame3D_nFrames (Frame);
  int iFrame;

  AutoGL_SetColor (0, 0, 1);

  for (iFrame = 0; iFrame < nFrames; iFrame++) {
    double x0, y0, z0;
    double x1, y1, z1;
    
    AutoMesh_Frame3D_getEndPointsAt 
      (Frame, iFrame,
       &x0, &y0, &z0,
       &x1, &y1, &z1);
    AutoGL_DrawLine (x0, y0, z0,
		     x1, y1, z1);
  }
}

static void DrawParticle (void) 
{
  int nParticles = AutoMesh_VolumeParticle3D_nParticles (Particle);
  int nSteps = AutoMesh_VolumeParticle3D_nSteps (Particle);
  int iParticle;
  int iStep;

  for (iParticle = 0; iParticle < nParticles; iParticle++) {
    switch (iParticle % 7) {
    case 0: 
      AutoGL_SetColor (1, 0, 0);   
      break;
    case 1:
      AutoGL_SetColor (0, 1, 0);
      break;
    case 2:
      AutoGL_SetColor (0, 0, 1);
      break;
    case 3:
      AutoGL_SetColor (1, 1, 0);
      break;
    case 4:
      AutoGL_SetColor (0, 1, 1);
      break;
    case 5:
      AutoGL_SetColor (1, 0, 1);
      break;
    case 6:
      AutoGL_SetColor (1, 1, 1);
      break;
    }

    for (iStep = 0; iStep < nSteps; iStep++) {
      double x, y, z;
      int elementId;
      double lx, ly, lz;
      
      AutoMesh_VolumeParticle3D_getTraceAt 
	(Particle, iParticle, iStep,
	 &x, &y, &z,
	 &elementId,
	 &lx, &ly, &lz);

      if (elementId != -1) {
	AutoGL_DrawCrossMark3D (x, y, z, 3);
      }
    }
  }
}

static void RedrawView (void) 
{
  DrawFrame ();
  DrawParticle ();
}

static void UpdateModel (void) 
{
#ifdef TIME_MEASURE
  StartMeasureTime ();
#endif

  MakeParticle ();
  AutoGL_DrawView ();

#ifdef TIME_MEASURE
  EndMeasureTime ();
#endif
}

static void SetParticle (void) 
{
  UpdateModel ();
}

static void IncX (void) 
{
  CenterX += Offset;
  UpdateModel ();
}

static void DecX (void) 
{
  CenterX -= Offset;
  UpdateModel ();
}

static void IncY (void) 
{
  CenterY += Offset;
  UpdateModel ();
}

static void DecY (void) 
{
  CenterY -= Offset;
  UpdateModel ();
}

static void IncZ (void) 
{
  CenterZ += Offset;
  UpdateModel ();
}

static void DecZ (void) 
{
  CenterZ -= Offset;
  UpdateModel ();
}

static void Quit (void) 
{
  AutoGL_Terminate ();
  exit (0);
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
#ifdef BINARY
  int binaryFlag = 1;
  int recordIdFlag = 0;
#else
  int binaryFlag = 0;
  int recordIdFlag = 1;
#endif
  int flag;

  if (argc != 6) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_tetra3d_particle_view  node_file  element_file \n"
       "     element_neighbor_file  node_vector_file  frame_file \n");
    exit (1);
  }

  AutoRes_Initialize (0);
  AutoMesh_Initialize ();

  AutoRes_TableInOut_SetBinaryFlag (binaryFlag);
  AutoRes_TableInOut_SetRecordIdFlag (recordIdFlag);



  Node = AutoMesh_Node3D_New ();
  assert(Node != NULL);

  flag = AutoMesh_Node3D_makeCoordColumn (Node);
  assert(flag);

  AutoRes_TableInOut_AssociateReal1DColumn (Node->nodes_coord);
  flag = AutoRes_TableInOut_Read (argv[1]);
  assert(flag);
  


  Element = AutoMesh_VolumeElement3D_New 
    (Node, AUTOMESH_POLYHEDRON_TETRA, 4);
  assert(Element != NULL);

  flag = AutoMesh_VolumeElement3D_makeNodeIdsColumn (Element);
  assert(flag);

  AutoRes_TableInOut_AssociateInteger1DColumn (Element->elements_nodeIds);
  flag = AutoRes_TableInOut_Read (argv[2]);
  assert(flag);
  
  flag = AutoMesh_VolumeElement3D_makeNeighborIdsColumn (Element);
  assert(flag);
  
  AutoRes_TableInOut_AssociateInteger1DColumn (Element->elements_neighborIds);
  flag = AutoRes_TableInOut_Read (argv[3]);
  assert(flag);

  flag = AutoMesh_VolumeElement3D_makeBucket (Element);
  assert(flag);



  Vector = AutoMesh_NodeVector3D_New (Node);
  
  flag = AutoMesh_NodeVector3D_makeValueColumn (Vector);
  assert(flag);
  
  AutoRes_TableInOut_AssociateReal1DColumn (Vector->nodes_value);
  flag = AutoRes_TableInOut_Read (argv[4]);
  assert(flag);



  Particle = AutoMesh_VolumeParticle3D_New (Element);

  flag = AutoMesh_VolumeParticle3D_makeColumn (Particle);
  assert(flag);



  Frame = AutoMesh_Frame3D_New ();
  assert(Frame != NULL);

  flag = AutoMesh_Frame3D_makeEndPointsColumn (Frame);
  assert(flag);

  AutoRes_TableInOut_AssociateReal1DColumn (Frame->frames_endPoint0);
  AutoRes_TableInOut_AssociateReal1DColumn (Frame->frames_endPoint1);
  flag = AutoRes_TableInOut_Read (argv[5]);
  assert(flag);



  SetView ();
  
  AutoGL_SetViewRedrawCallback (RedrawView);

  AutoGL_SetDefaultCallbackInMode3D (NULL);
  AutoGL_EnableDragInMode3D ();



  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode3D ();

  AutoGL_AddCallback (Quit, "Quit");       

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");

  AutoGL_AddReal (&CenterX, "CenterX");
  AutoGL_AddReal (&CenterY, "CenterY");
  AutoGL_AddReal (&CenterZ, "CenterZ");
  AutoGL_AddReal (&Offset, "Offset");
  AutoGL_AddInteger (&NSteps, "NSteps");
  AutoGL_AddReal (&DeltaT, "DeltaT");

  AutoGL_AddCallback (UpdateModel, "UpdateModel");      
  AutoGL_AddCallback (SetParticle, "SetParticle");      
  AutoGL_AddCallback (IncX, "IncX");      
  AutoGL_AddCallback (DecX, "DecX");      
  AutoGL_AddCallback (IncY, "IncY");      
  AutoGL_AddCallback (DecY, "DecY");      
  AutoGL_AddCallback (IncZ, "IncZ");      
  AutoGL_AddCallback (DecZ, "DecZ");      
}
