/* advautomesh_tetra3d_section_contour_view2.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>
#include <autogl.h>
#include <automesh.h>



#ifdef TIME_MEASURE
#include "time_measure.c"
#endif



static AutoMesh_Node3D *Node;
static AutoMesh_VolumeElement3D *Element;
static AutoMesh_NodeScalar3D *Scalar;
static AutoMesh_VolumeSection3D2 *Section;
static AutoMesh_Frame3D *Frame;

static int EdgeIsShown = 0;

static double MinRange = 0.0;
static double MaxRange = 1.0;

static double A = 0.0;
static double B = 0.0;
static double C = 1.0;
static double D = 0.0;
static double Offset = 0.0;
static double OffsetIncrement = 1.0;

static void SetScalarRange (void)
{
  AutoMesh_NodeScalar3D_getRange (Scalar, 
				  &MinRange, &MaxRange);

  fprintf (stderr, " MinRange %f MaxRange %f \n",
	   MinRange, MaxRange);
}

static void SetView (void)
{
  double minX, minY, minZ;
  double maxX, maxY, maxZ;
  double centerX, centerY, centerZ;
  double size;
  
  AutoMesh_Node3D_getRange 
    (Node, 
     &minX, &minY, &minZ,
     &maxX, &maxY, &maxZ);
  centerX = (minX + maxX) / 2;
  centerY = (minY + maxY) / 2;
  centerZ = (minZ + maxZ) / 2;
  size = AutoMesh_Distance3D (minX, minY, minZ,
			      maxX, maxY, maxZ) / 2;

  AutoGL_SetViewSize (size);
  AutoGL_SetViewCenter (centerX, centerY, centerZ); 

  fprintf (stderr, " minX %f minY %f minZ %f \n",
	   minX, minY, minZ);
  fprintf (stderr, " maxX %f maxY %f maxZ %f \n",
	   maxX, maxY, maxZ);

  D = -(minZ + maxZ) / 2;
  OffsetIncrement = (maxZ - minZ) / 10;
}

static void DrawFrame (void) 
{
  int nFrames = AutoMesh_Frame3D_nFrames (Frame);
  int iFrame;

  AutoGL_SetColor (0, 0, 1);

  for (iFrame = 0; iFrame < nFrames; iFrame++) {
    double x0, y0, z0;
    double x1, y1, z1;
    
    AutoMesh_Frame3D_getEndPointsAt 
      (Frame, iFrame,
       &x0, &y0, &z0,
       &x1, &y1, &z1);
    AutoGL_DrawLine (x0, y0, z0,
		     x1, y1, z1);
  }
}

static void DrawSectionContour (void) 
{
  int nSections;
  int iSection;

  nSections = AutoMesh_VolumeSection3D2_nSections (Section);

  for (iSection = 0; iSection < nSections; iSection++) {
    double x0, y0, z0;
    double x1, y1, z1;
    double x2, y2, z2;
    double s0, s1, s2;

    AutoMesh_VolumeSection3D2_getSectionPointAt 
      (Section, iSection,
       &x0, &y0, &z0,
       &x1, &y1, &z1,
       &x2, &y2, &z2);
    AutoMesh_VolumeSection3D2_getSectionScalarAt 
      (Section, iSection, Scalar,
       &s0, &s1, &s2);

    AutoGL_DrawSmoothContourTriangle 
      (MinRange, MaxRange, 
       x0, y0, z0, s0,
       x1, y1, z1, s1, 
       x2, y2, z2, s2);
  }

  if (EdgeIsShown) {
    AutoGL_SetColor (1, 1, 1);

    for (iSection = 0; iSection < nSections; iSection++) {
      double x0, y0, z0;
      double x1, y1, z1;
      double x2, y2, z2;
      
      AutoMesh_VolumeSection3D2_getSectionPointAt 
	(Section, iSection,
	 &x0, &y0, &z0,
	 &x1, &y1, &z1,
	 &x2, &y2, &z2);
      
      AutoGL_DrawLine (x0, y0, z0,
		       x1, y1, z1);
      AutoGL_DrawLine (x1, y1, z1,
		       x2, y2, z2);
      AutoGL_DrawLine (x2, y2, z2,
		       x0, y0, z0);
    }
  }
}

static void RedrawView (void)
{
  AutoGL_DrawContourMap (80, MinRange, MaxRange);

  AutoGL_DrawDisplayList ();
}

static void UpdateModel (void) 
{
  int flag;

#ifdef TIME_MEASURE
  StartMeasureTime ();
#endif

  AutoMesh_VolumeSection3D2_setEquation (Section, A, B, C, D);
  AutoMesh_VolumeSection3D2_makeBucket (Section);

  flag = AutoMesh_VolumeSection3D2_extract (Section, Offset);
  assert(flag);

  AutoGL_OpenDisplayList ();
  DrawFrame ();
  DrawSectionContour ();
  AutoGL_CloseDisplayList ();

  AutoGL_DrawView ();

#ifdef TIME_MEASURE
  EndMeasureTime ();
#endif
}

static void Shift (void) 
{
  int flag;

#ifdef TIME_MEASURE
  StartMeasureTime ();
#endif

  flag = AutoMesh_VolumeSection3D2_extract (Section, Offset);
  assert(flag);

  AutoGL_OpenDisplayList ();
  DrawFrame ();
  DrawSectionContour ();
  AutoGL_CloseDisplayList ();

  AutoGL_DrawView ();

#ifdef TIME_MEASURE
  EndMeasureTime ();
#endif
}

static void Inc (void) 
{
  Offset += OffsetIncrement;
  Shift ();
}

static void Dec (void) 
{
  Offset -= OffsetIncrement;
  Shift ();
}

static void Quit (void) 
{
  AutoGL_Terminate ();
  exit (0);
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
#ifdef BINARY
  int binaryFlag = 1;
  int recordIdFlag = 0;
#else
  int binaryFlag = 0;
  int recordIdFlag = 1;
#endif
  int flag;

  if (argc != 5) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_tetra3d_section_contour_view2  node_file  element_file \n"
       "        node_scalar_file  frame_file  \n");
    exit (1);
  }

  AutoRes_Initialize (0);
  AutoMesh_Initialize ();

  AutoRes_TableInOut_SetBinaryFlag (binaryFlag);
  AutoRes_TableInOut_SetRecordIdFlag (recordIdFlag);



  Node = AutoMesh_Node3D_New ();
  assert(Node != NULL);

  flag = AutoMesh_Node3D_makeCoordColumn (Node);
  assert(flag);

  AutoRes_TableInOut_AssociateReal1DColumn (Node->nodes_coord);
  flag = AutoRes_TableInOut_Read (argv[1]);
  assert(flag);
  


  Element = AutoMesh_VolumeElement3D_New 
    (Node, AUTOMESH_POLYHEDRON_TETRA, 4);
  assert(Element != NULL);

  flag = AutoMesh_VolumeElement3D_makeNodeIdsColumn (Element);
  assert(flag);

  AutoRes_TableInOut_AssociateInteger1DColumn (Element->elements_nodeIds);
  flag = AutoRes_TableInOut_Read (argv[2]);
  assert(flag);
  


  Scalar = AutoMesh_NodeScalar3D_New (Node);
  
  flag = AutoMesh_NodeScalar3D_makeValueColumn (Scalar);
  assert(flag);
  
  AutoRes_TableInOut_AssociateRealColumn (Scalar->nodes_value);
  flag = AutoRes_TableInOut_Read (argv[3]);
  assert(flag);



  Section = AutoMesh_VolumeSection3D2_New (Element);

  flag = AutoMesh_VolumeSection3D2_makeColumn (Section);
  assert(flag);



  Frame = AutoMesh_Frame3D_New ();
  assert(Frame != NULL);

  flag = AutoMesh_Frame3D_makeEndPointsColumn (Frame);
  assert(flag);

  AutoRes_TableInOut_AssociateReal1DColumn (Frame->frames_endPoint0);
  AutoRes_TableInOut_AssociateReal1DColumn (Frame->frames_endPoint1);
  flag = AutoRes_TableInOut_Read (argv[4]);
  assert(flag);



  SetScalarRange ();
  SetView ();
  
  AutoGL_SetViewRedrawCallback (RedrawView);

  AutoGL_SetDefaultCallbackInMode3D (NULL);
  AutoGL_EnableDragInMode3D ();



  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode3D ();

  AutoGL_AddCallback (Quit, "Quit");       

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");

  AutoGL_AddBoolean (&EdgeIsShown, "EdgeIsShown");
  AutoGL_AddReal (&MinRange, "MinRange");
  AutoGL_AddReal (&MaxRange, "MaxRange");
  AutoGL_AddReal (&A, "A");
  AutoGL_AddReal (&B, "B");
  AutoGL_AddReal (&C, "C");
  AutoGL_AddReal (&D, "D");
  AutoGL_AddReal (&Offset, "Offset");
  AutoGL_AddReal (&OffsetIncrement, "OffsetIncrement");
  AutoGL_SetLabel ("delta");

  AutoGL_AddCallback (UpdateModel, "UpdateModel");      
  AutoGL_AddCallback (Inc, "Inc");      
  AutoGL_AddCallback (Dec, "Dec");      
}
