/* advautomesh_tetra3d_section_grid_view.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>
#include <autogl.h>
#include <automesh.h>



#ifdef TIME_MEASURE
#include "time_measure.c"
#endif



static AutoMesh_Node3D *Node;
static AutoMesh_VolumeElement3D *Element;
static AutoMesh_NodeVector3D *Vector;
static AutoMesh_VolumeSample3D *Sample;
static AutoMesh_Frame3D *Frame;

static double Interval = 1.0;
static double Scale = 1.0;

static double A = 0.0;
static double B = 0.0;
static double C = 1.0;
static double D = 0.0;
static double Offset = 0.0;
static double OffsetIncrement = 1.0;

static void SetArrowRange (void)
{
  double maxLength;
  double minX, minY, minZ;
  double maxX, maxY, maxZ;
  double size;
  
  maxLength = AutoMesh_NodeVector3D_maxLength (Vector);
  
  AutoMesh_Node3D_getRange 
    (Node, 
     &minX, &minY, &minZ,
     &maxX, &maxY, &maxZ);
  size = AutoMesh_Distance3D (minX, minY, minZ,
			      maxX, maxY, maxZ) / 2;
  
  if (maxLength == 0.0) {
    Scale = 1.0;
  } else {
    Scale = size * 0.1 / maxLength;
  }
}

static void SetView (void)
{
  double minX, minY, minZ;
  double maxX, maxY, maxZ;
  double centerX, centerY, centerZ;
  double size;
  
  AutoMesh_Node3D_getRange 
    (Node, 
     &minX, &minY, &minZ,
     &maxX, &maxY, &maxZ);
  centerX = (minX + maxX) / 2;
  centerY = (minY + maxY) / 2;
  centerZ = (minZ + maxZ) / 2;
  size = AutoMesh_Distance3D (minX, minY, minZ,
			      maxX, maxY, maxZ) / 2;

  AutoGL_SetViewSize (size);
  AutoGL_SetViewCenter (centerX, centerY, centerZ); 

  fprintf (stderr, " minX %f minY %f minZ %f \n",
	   minX, minY, minZ);
  fprintf (stderr, " maxX %f maxY %f maxZ %f \n",
	   maxX, maxY, maxZ);

  D = -(minZ + maxZ) / 2;
  OffsetIncrement = (maxZ - minZ) / 10;

  Interval = size / 20;
}

void MakeSectionGrid (void)
{
  double length = sqrt (A * A + B * B + C * C);
  double planeA = A / length;
  double planeB = B / length;
  double planeC = C / length;
  double planeD = (D + Offset) / length;
  double minX, minY, minZ;
  double maxX, maxY, maxZ;
  double centerX, centerY, centerZ;
  double size;
  double sectionCenterX, sectionCenterY, sectionCenterZ;
  double uAxisX, uAxisY, uAxisZ;
  double vAxisX, vAxisY, vAxisZ;
  int nDivisions;
  int iu, iv;

  AutoMesh_Node3D_getRange 
    (Node, 
     &minX, &minY, &minZ,
     &maxX, &maxY, &maxZ);
  centerX = (minX + maxX) / 2;
  centerY = (minY + maxY) / 2;
  centerZ = (minZ + maxZ) / 2;
  size = AutoMesh_Distance3D (minX, minY, minZ,
			      maxX, maxY, maxZ) / 2;

  AutoMesh_GetProjectedPointOnPlane3D
    (planeA, planeB, planeC, planeD,
     centerX, centerY, centerZ, 
     &sectionCenterX, &sectionCenterY, &sectionCenterZ);
  AutoMesh_GetAnyPerpendicularDirection3D
    (planeA, planeB, planeC,
     &uAxisX, &uAxisY, &uAxisZ,
     &vAxisX, &vAxisY, &vAxisZ);
  nDivisions = size / Interval;
  
  AutoRes_Table_reserveRecords 
    (Sample->samples, 
     (nDivisions * 2 + 1) * (nDivisions * 2 + 1));

  for (iu = -nDivisions; iu <= nDivisions; iu++) {
    for (iv = -nDivisions; iv <= nDivisions; iv++) {
      double x = sectionCenterX 
	+ uAxisX * Interval * iu + vAxisX * Interval * iv;
      double y = sectionCenterY 
	+ uAxisY * Interval * iu + vAxisY * Interval * iv;
      double z = sectionCenterZ 
	+ uAxisZ * Interval * iu + vAxisZ * Interval * iv;

      AutoMesh_VolumeSample3D_addSample (Sample, x, y, z);
    }
  }

  AutoMesh_VolumeSample3D_setElementOnSample (Sample);
}

static void DrawFrame (void) 
{
  int nFrames = AutoMesh_Frame3D_nFrames (Frame);
  int iFrame;

  AutoGL_SetColor (0, 0, 1);

  for (iFrame = 0; iFrame < nFrames; iFrame++) {
    double x0, y0, z0;
    double x1, y1, z1;
    
    AutoMesh_Frame3D_getEndPointsAt 
      (Frame, iFrame,
       &x0, &y0, &z0,
       &x1, &y1, &z1);
    AutoGL_DrawLine (x0, y0, z0,
		     x1, y1, z1);
  }
}

static void DrawSampleArrow (void) 
{
  int nSamples = AutoMesh_VolumeSample3D_nSamples (Sample);
  int iSample;

  AutoGL_SetColor (1, 1, 0);

  for (iSample = 0; iSample < nSamples; iSample++) {
    int elementId;
    double lx, ly, lz;
    double x, y, z;
    double vx, vy, vz;

    AutoMesh_VolumeSample3D_getSampleAt 
      (Sample, iSample,
       &elementId,
       &lx, &ly, &lz);
    if (elementId == -1) continue;

    AutoMesh_VolumeSample3D_getSamplePointAt 
      (Sample, iSample,
       &x, &y, &z);
    AutoMesh_VolumeSample3D_getVectorAt 
      (Sample, iSample, Vector,
       &vx, &vy, &vz);

    AutoGL_DrawConeArrow3D
      (x, y, z,
       x + vx * Scale, y + vy * Scale, z + vz * Scale,
       0.25);
  }
}

static void RedrawView (void)
{
  DrawFrame ();
  DrawSampleArrow ();
}

static void UpdateModel (void) 
{
#ifdef TIME_MEASURE
  StartMeasureTime ();
#endif

  MakeSectionGrid ();
  AutoGL_DrawView ();

#ifdef TIME_MEASURE
  EndMeasureTime ();
#endif
}

static void Inc (void) 
{
  Offset += OffsetIncrement;
  UpdateModel ();
}

static void Dec (void) 
{
  Offset -= OffsetIncrement;
  UpdateModel ();
}

static void Quit (void) 
{
  AutoGL_Terminate ();
  exit (0);
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
#ifdef BINARY
  int binaryFlag = 1;
  int recordIdFlag = 0;
#else
  int binaryFlag = 0;
  int recordIdFlag = 1;
#endif
  int flag;

  if (argc != 5) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_tetra3d_section_grid_view  node_file  element_file \n"
       "        node_vector_file  frame_file  \n");
    exit (1);
  }

  AutoRes_Initialize (0);
  AutoMesh_Initialize ();

  AutoRes_TableInOut_SetBinaryFlag (binaryFlag);
  AutoRes_TableInOut_SetRecordIdFlag (recordIdFlag);



  Node = AutoMesh_Node3D_New ();
  assert(Node != NULL);

  flag = AutoMesh_Node3D_makeCoordColumn (Node);
  assert(flag);

  AutoRes_TableInOut_AssociateReal1DColumn (Node->nodes_coord);
  flag = AutoRes_TableInOut_Read (argv[1]);
  assert(flag);
  


  Element = AutoMesh_VolumeElement3D_New 
    (Node, AUTOMESH_POLYHEDRON_TETRA, 4);
  assert(Element != NULL);

  flag = AutoMesh_VolumeElement3D_makeNodeIdsColumn (Element);
  assert(flag);

  AutoRes_TableInOut_AssociateInteger1DColumn (Element->elements_nodeIds);
  flag = AutoRes_TableInOut_Read (argv[2]);
  assert(flag);
  
  flag = AutoMesh_VolumeElement3D_makeBucket (Element);
  assert(flag);



  Vector = AutoMesh_NodeVector3D_New (Node);
  
  flag = AutoMesh_NodeVector3D_makeValueColumn (Vector);
  assert(flag);
  
  AutoRes_TableInOut_AssociateReal1DColumn (Vector->nodes_value);
  flag = AutoRes_TableInOut_Read (argv[3]);
  assert(flag);



  Sample = AutoMesh_VolumeSample3D_New (Element);

  flag = AutoMesh_VolumeSample3D_makeColumn (Sample);
  assert(flag);



  Frame = AutoMesh_Frame3D_New ();
  assert(Frame != NULL);

  flag = AutoMesh_Frame3D_makeEndPointsColumn (Frame);
  assert(flag);

  AutoRes_TableInOut_AssociateReal1DColumn (Frame->frames_endPoint0);
  AutoRes_TableInOut_AssociateReal1DColumn (Frame->frames_endPoint1);
  flag = AutoRes_TableInOut_Read (argv[4]);
  assert(flag);



  SetArrowRange ();
  SetView ();
  
  AutoGL_SetViewRedrawCallback (RedrawView);

  AutoGL_SetDefaultCallbackInMode3D (NULL);
  AutoGL_EnableDragInMode3D ();



  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode3D ();

  AutoGL_AddCallback (Quit, "Quit");       

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");

  AutoGL_AddReal (&Interval, "Interval");
  AutoGL_AddReal (&Scale, "Scale");
  AutoGL_AddReal (&A, "A");
  AutoGL_AddReal (&B, "B");
  AutoGL_AddReal (&C, "C");
  AutoGL_AddReal (&D, "D");
  AutoGL_AddReal (&Offset, "Offset");
  AutoGL_AddReal (&OffsetIncrement, "OffsetIncrement");
  AutoGL_SetLabel ("delta");

  AutoGL_AddCallback (UpdateModel, "UpdateModel");      
  AutoGL_AddCallback (Inc, "Inc");      
  AutoGL_AddCallback (Dec, "Dec");      
}



