/* advautomesh_tri2d_contour_view.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>
#include <autogl.h>
#include <automesh.h>



static AutoMesh_Node2D *Node;
static AutoMesh_VolumeElement2D *Element;
static AutoMesh_NodeScalar2D *Scalar;

static int EdgeIsShown = 0;

static double MinRange = 0.0;
static double MaxRange = 1.0;

static void SetScalarRange (void)
{
  AutoMesh_NodeScalar2D_getRange (Scalar, 
				  &MinRange, &MaxRange);

  fprintf (stderr, " MinRange %f MaxRange %f \n",
	   MinRange, MaxRange);
}

static void SetView (void)
{
  double minX, minY;
  double maxX, maxY;
  double centerX, centerY;
  double size;
  
  AutoMesh_Node2D_getRange 
    (Node, 
     &minX, &minY,
     &maxX, &maxY);
  centerX = (minX + maxX) / 2;
  centerY = (minY + maxY) / 2;
  size = AutoMesh_Distance2D (minX, minY,
			      maxX, maxY) / 2;

  AutoGL_SetViewSize (size);
  AutoGL_SetViewCenter (centerX, centerY, 0.0); 

  fprintf (stderr, " minX %f minY %f \n",
	   minX, minY);
  fprintf (stderr, " maxX %f maxY %f \n",
	   maxX, maxY);
}

static void DrawContour (void) 
{
  int nElements = AutoMesh_VolumeElement2D_nElements (Element);
  int iElement;

  for (iElement = 0; iElement < nElements; iElement++) {
    int nodeId0 = AutoMesh_VolumeElement2D_nodeIdAt (Element, iElement, 0);
    int nodeId1 = AutoMesh_VolumeElement2D_nodeIdAt (Element, iElement, 1);
    int nodeId2 = AutoMesh_VolumeElement2D_nodeIdAt (Element, iElement, 2);

    double x0, y0;
    double x1, y1;
    double x2, y2;
    double s0, s1, s2;
    
    AutoMesh_Node2D_getCoordAt (Node, nodeId0,
				&x0, &y0);
    AutoMesh_Node2D_getCoordAt (Node, nodeId1,
				&x1, &y1);
    AutoMesh_Node2D_getCoordAt (Node, nodeId2,
				&x2, &y2);
    s0 = AutoMesh_NodeScalar2D_valueAt (Scalar, nodeId0);
    s1 = AutoMesh_NodeScalar2D_valueAt (Scalar, nodeId1);
    s2 = AutoMesh_NodeScalar2D_valueAt (Scalar, nodeId2);

    AutoGL_DrawSmoothContourTriangle 
      (MinRange, MaxRange, 
       x0, y0, 0.0, s0,
       x1, y1, 0.0, s1, 
       x2, y2, 0.0, s2);
  }

  if (EdgeIsShown) {
    AutoGL_SetColor (1, 1, 1);

    for (iElement = 0; iElement < nElements; iElement++) {
      int nodeId0 = AutoMesh_VolumeElement2D_nodeIdAt (Element, iElement, 0);
      int nodeId1 = AutoMesh_VolumeElement2D_nodeIdAt (Element, iElement, 1);
      int nodeId2 = AutoMesh_VolumeElement2D_nodeIdAt (Element, iElement, 2);

      double x0, y0;
      double x1, y1;
      double x2, y2;
      
      AutoMesh_Node2D_getCoordAt (Node, nodeId0,
				  &x0, &y0);
      AutoMesh_Node2D_getCoordAt (Node, nodeId1,
				  &x1, &y1);
      AutoMesh_Node2D_getCoordAt (Node, nodeId2,
				  &x2, &y2);
      
      AutoGL_DrawLine (x0, y0, 0.0,
		       x1, y1, 0.0);
      AutoGL_DrawLine (x1, y1, 0.0,
		       x2, y2, 0.0);
      AutoGL_DrawLine (x2, y2, 0.0,
		       x0, y0, 0.0);
    }
  }
}

static void RedrawView (void)
{
  AutoGL_DrawContourMap (80, MinRange, MaxRange);

  AutoGL_DrawDisplayList ();
}

static void UpdateModel (void) 
{
  AutoGL_OpenDisplayList ();
  DrawContour ();
  AutoGL_CloseDisplayList ();

  AutoGL_DrawView ();
}

static void Quit (void) 
{
  AutoGL_Terminate ();
  exit (0);
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
#ifdef BINARY
  int binaryFlag = 1;
  int recordIdFlag = 0;
#else
  int binaryFlag = 0;
  int recordIdFlag = 1;
#endif
  int flag;

  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_tri2d_contour_view  node_file  element_file \n"
       "        node_scalar_file  \n");
    exit (1);
  }

  AutoRes_Initialize (0);
  AutoMesh_Initialize ();

  AutoRes_TableInOut_SetBinaryFlag (binaryFlag);
  AutoRes_TableInOut_SetRecordIdFlag (recordIdFlag);



  Node = AutoMesh_Node2D_New ();
  assert(Node != NULL);

  flag = AutoMesh_Node2D_makeCoordColumn (Node);
  assert(flag);

  AutoRes_TableInOut_AssociateReal1DColumn (Node->nodes_coord);
  flag = AutoRes_TableInOut_Read (argv[1]);
  assert(flag);
  


  Element = AutoMesh_VolumeElement2D_New 
    (Node, AUTOMESH_POLYGON_TRI, 3);
  assert(Element != NULL);

  flag = AutoMesh_VolumeElement2D_makeNodeIdsColumn (Element);
  assert(flag);

  AutoRes_TableInOut_AssociateInteger1DColumn (Element->elements_nodeIds);
  flag = AutoRes_TableInOut_Read (argv[2]);
  assert(flag);
  


  Scalar = AutoMesh_NodeScalar2D_New (Node);
  
  flag = AutoMesh_NodeScalar2D_makeValueColumn (Scalar);
  assert(flag);
  
  AutoRes_TableInOut_AssociateRealColumn (Scalar->nodes_value);
  flag = AutoRes_TableInOut_Read (argv[3]);
  assert(flag);



  SetScalarRange ();
  SetView ();
  
  AutoGL_SetViewRedrawCallback (RedrawView);

  AutoGL_SetDefaultCallbackInMode2D (NULL);
  AutoGL_EnableDragInMode2D ();



  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode2D ();

  AutoGL_AddCallback (Quit, "Quit");       

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");

  AutoGL_AddBoolean (&EdgeIsShown, "EdgeIsShown");
  AutoGL_AddReal (&MinRange, "MinRange");
  AutoGL_AddReal (&MaxRange, "MaxRange");

  AutoGL_AddCallback (UpdateModel, "UpdateModel");      
}
