/* advautomesh_tri2d_deform_view.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>
#include <autogl.h>
#include <automesh.h>



static AutoMesh_Node2D *Node;
static AutoMesh_VolumeElement2D *Element;
static AutoMesh_NodeVector2D *Vector;

static int EdgeIsShown = 0;

static double Scale;

static void SetDeformationRange (void)
{
  double maxLength;
  double minX, minY;
  double maxX, maxY;
  double size;
  
  maxLength = AutoMesh_NodeVector2D_maxLength (Vector);
  
  AutoMesh_Node2D_getRange 
    (Node, 
     &minX, &minY,
     &maxX, &maxY);
  size = AutoMesh_Distance2D (minX, minY,
			      maxX, maxY) / 2;
  
  if (maxLength == 0.0) {
    Scale = 1.0;
  } else {
    Scale = size * 0.1 / maxLength;
  }
}

static void SetView (void)
{
  double minX, minY;
  double maxX, maxY;
  double centerX, centerY;
  double size;
  
  AutoMesh_Node2D_getRange 
    (Node, 
     &minX, &minY,
     &maxX, &maxY);
  centerX = (minX + maxX) / 2;
  centerY = (minY + maxY) / 2;
  size = AutoMesh_Distance2D (minX, minY,
			      maxX, maxY) / 2;

  AutoGL_SetViewSize (size);
  AutoGL_SetViewCenter (centerX, centerY, 0.0); 

  fprintf (stderr, " minX %f minY %f \n",
	   minX, minY);
  fprintf (stderr, " maxX %f maxY %f \n",
	   maxX, maxY);
}

static void DrawDeformation (void) 
{
  int nElements = AutoMesh_VolumeElement2D_nElements (Element);
  int iElement;

  AutoGL_SetColor (1, 0, 1);
  for (iElement = 0; iElement < nElements; iElement++) {
    int nodeId0 = AutoMesh_VolumeElement2D_nodeIdAt (Element, iElement, 0);
    int nodeId1 = AutoMesh_VolumeElement2D_nodeIdAt (Element, iElement, 1);
    int nodeId2 = AutoMesh_VolumeElement2D_nodeIdAt (Element, iElement, 2);

    double x0, y0;
    double x1, y1;
    double x2, y2;
    double u0, v0;
    double u1, v1;
    double u2, v2;
    double vx0, vy0;
    double vx1, vy1;
    double vx2, vy2;
    
    AutoMesh_Node2D_getCoordAt (Node, nodeId0,
				&x0, &y0);
    AutoMesh_Node2D_getCoordAt (Node, nodeId1,
				&x1, &y1);
    AutoMesh_Node2D_getCoordAt (Node, nodeId2,
				&x2, &y2);
    AutoMesh_NodeVector2D_getValueAt (Vector, nodeId0,
				      &u0, &v0);
    AutoMesh_NodeVector2D_getValueAt (Vector, nodeId1,
				      &u1, &v1);
    AutoMesh_NodeVector2D_getValueAt (Vector, nodeId2,
				      &u2, &v2);

    vx0 = x0 + u0 * Scale;
    vy0 = y0 + v0 * Scale;

    vx1 = x1 + u1 * Scale;
    vy1 = y1 + v1 * Scale;

    vx2 = x2 + u2 * Scale;
    vy2 = y2 + v2 * Scale;
    
    AutoGL_DrawTriangle (vx0, vy0, 0.0,
			 vx1, vy1, 0.0,
			 vx2, vy2, 0.0);
  }

  if (EdgeIsShown) {
    AutoGL_SetColor (0, 1, 0);

    for (iElement = 0; iElement < nElements; iElement++) {
      int nodeId0 = AutoMesh_VolumeElement2D_nodeIdAt (Element, iElement, 0);
      int nodeId1 = AutoMesh_VolumeElement2D_nodeIdAt (Element, iElement, 1);
      int nodeId2 = AutoMesh_VolumeElement2D_nodeIdAt (Element, iElement, 2);

      double x0, y0;
      double x1, y1;
      double x2, y2;
      double u0, v0;
      double u1, v1;
      double u2, v2;
      double vx0, vy0;
      double vx1, vy1;
      double vx2, vy2;
      
      AutoMesh_Node2D_getCoordAt (Node, nodeId0,
				  &x0, &y0);
      AutoMesh_Node2D_getCoordAt (Node, nodeId1,
				  &x1, &y1);
      AutoMesh_Node2D_getCoordAt (Node, nodeId2,
				  &x2, &y2);
      AutoMesh_NodeVector2D_getValueAt (Vector, nodeId0,
					&u0, &v0);
      AutoMesh_NodeVector2D_getValueAt (Vector, nodeId1,
					&u1, &v1);
      AutoMesh_NodeVector2D_getValueAt (Vector, nodeId2,
					&u2, &v2);
      
      vx0 = x0 + u0 * Scale;
      vy0 = y0 + v0 * Scale;
      
      vx1 = x1 + u1 * Scale;
      vy1 = y1 + v1 * Scale;
      
      vx2 = x2 + u2 * Scale;
      vy2 = y2 + v2 * Scale;
      
      AutoGL_DrawLine (vx0, vy0, 0.0,
		       vx1, vy1, 0.0);
      AutoGL_DrawLine (vx1, vy1, 0.0,
		       vx2, vy2, 0.0);
      AutoGL_DrawLine (vx2, vy2, 0.0,
		       vx0, vy0, 0.0);
    }
  }
}

static void RedrawView (void)
{
  AutoGL_DrawDisplayList ();
}

static void UpdateModel (void) 
{
  AutoGL_OpenDisplayList ();
  DrawDeformation ();
  AutoGL_CloseDisplayList ();

  AutoGL_DrawView ();
}

static void Quit (void) 
{
  AutoGL_Terminate ();
  exit (0);
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
#ifdef BINARY
  int binaryFlag = 1;
  int recordIdFlag = 0;
#else
  int binaryFlag = 0;
  int recordIdFlag = 1;
#endif
  int flag;

  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_tri2d_deform_view  node_file  element_file \n"
       "        node_vector_file  \n");
    exit (1);
  }

  AutoRes_Initialize (0);
  AutoMesh_Initialize ();

  AutoRes_TableInOut_SetBinaryFlag (binaryFlag);
  AutoRes_TableInOut_SetRecordIdFlag (recordIdFlag);



  Node = AutoMesh_Node2D_New ();
  assert(Node != NULL);

  flag = AutoMesh_Node2D_makeCoordColumn (Node);
  assert(flag);

  AutoRes_TableInOut_AssociateReal1DColumn (Node->nodes_coord);
  flag = AutoRes_TableInOut_Read (argv[1]);
  assert(flag);
  


  Element = AutoMesh_VolumeElement2D_New 
    (Node, AUTOMESH_POLYGON_TRI, 3);
  assert(Element != NULL);

  flag = AutoMesh_VolumeElement2D_makeNodeIdsColumn (Element);
  assert(flag);

  AutoRes_TableInOut_AssociateInteger1DColumn (Element->elements_nodeIds);
  flag = AutoRes_TableInOut_Read (argv[2]);
  assert(flag);
  


  Vector = AutoMesh_NodeVector2D_New (Node);
  
  flag = AutoMesh_NodeVector2D_makeValueColumn (Vector);
  assert(flag);
  
  AutoRes_TableInOut_AssociateReal1DColumn (Vector->nodes_value);
  flag = AutoRes_TableInOut_Read (argv[3]);
  assert(flag);



  SetDeformationRange ();
  SetView ();
  
  AutoGL_SetViewRedrawCallback (RedrawView);

  AutoGL_SetDefaultCallbackInMode2D (NULL);
  AutoGL_EnableDragInMode2D ();



  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode2D ();

  AutoGL_AddCallback (Quit, "Quit");       

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");

  AutoGL_AddBoolean (&EdgeIsShown, "EdgeIsShown");
  AutoGL_AddReal (&Scale, "Scale");

  AutoGL_AddCallback (UpdateModel, "UpdateModel");      
}
