/* advautomesh_tri3d_group_multi_select.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>
#include <autogl.h>
#include <automesh.h>



static AutoMesh_Node3D *Node;
static AutoMesh_SurfaceElement3D *Element;
static AutoMesh_SurfaceGrouper3D *Grouper;
static AutoMesh_Frame3D *Frame;

static int NGroups;
static AutoRes_Table *Groups;
static AutoRes_IntegerColumn *Groups_isSelected;

static int Filled = 0;
static int FrameIsShown = 0;
static int GroupIdIsShown = 0;

static void InitializeGroupSelection (void)
{
  int iGroup;
  int flag;

  Groups = AutoRes_Table_New ("Groups");
  
  Groups_isSelected = AutoRes_IntegerColumn_New 
    (Groups, "isSelected");
  flag = AutoRes_Table_addIntegerColumn 
    (Groups, Groups_isSelected);
  assert(flag);

  flag = AutoRes_Table_allocateRecords (Groups, NGroups);
  assert(flag);
  
  for (iGroup = 0; iGroup < NGroups; iGroup++) {
    AUTORES_I_SET(Groups_isSelected, iGroup, 0);
  }
}

static void SetView (void)
{
  double minX, minY, minZ;
  double maxX, maxY, maxZ;
  double centerX, centerY, centerZ;
  double size;
  
  AutoMesh_Node3D_getRange 
    (Node, 
     &minX, &minY, &minZ,
     &maxX, &maxY, &maxZ);
  centerX = (minX + maxX) / 2;
  centerY = (minY + maxY) / 2;
  centerZ = (minZ + maxZ) / 2;
  size = AutoMesh_Distance3D (minX, minY, minZ,
			      maxX, maxY, maxZ) / 2;

  AutoGL_SetViewSize (size);
  AutoGL_SetViewCenter (centerX, centerY, centerZ); 

  fprintf (stderr, " minX %f minY %f minZ %f \n",
	   minX, minY, minZ);
  fprintf (stderr, " maxX %f maxY %f maxZ %f \n",
	   maxX, maxY, maxZ);
}

static void DrawGroup (void) 
{
  int nElements = AutoMesh_SurfaceElement3D_nElements (Element);
  int nFrames = AutoMesh_Frame3D_nFrames (Frame);
  int iElement;
  int iFrame;
  char buf[256];

  if (Filled) {
    for (iElement = 0; iElement < nElements; iElement++) {
      int nodeId0 = AutoMesh_SurfaceElement3D_nodeIdAt (Element, iElement, 0);
      int nodeId1 = AutoMesh_SurfaceElement3D_nodeIdAt (Element, iElement, 1);
      int nodeId2 = AutoMesh_SurfaceElement3D_nodeIdAt (Element, iElement, 2);
      int groupId = AutoMesh_SurfaceGrouper3D_groupIdAtElement
	(Grouper, iElement);

      double x0, y0, z0;
      double x1, y1, z1;
      double x2, y2, z2;
      
      AutoMesh_Node3D_getCoordAt (Node, nodeId0,
				  &x0, &y0, &z0);
      AutoMesh_Node3D_getCoordAt (Node, nodeId1,
				  &x1, &y1, &z1);
      AutoMesh_Node3D_getCoordAt (Node, nodeId2,
				  &x2, &y2, &z2);
      
      if (AUTORES_I(Groups_isSelected, groupId)) {
	AutoGL_SetColor (1, 0, 0);
      } else {
	AutoGL_SetColor (1, 1, 0);
      }

      AutoGL_DrawTriangle (x0, y0, z0,
			   x1, y1, z1,
			   x2, y2, z2);
    }
  }

  if (FrameIsShown) {
    AutoGL_SetColor (0, 0, 1);

    for (iFrame = 0; iFrame < nFrames; iFrame++) {
      double x0, y0, z0;
      double x1, y1, z1;
	
      AutoMesh_Frame3D_getEndPointsAt 
	(Frame, iFrame,
	 &x0, &y0, &z0,
	 &x1, &y1, &z1);
      AutoGL_DrawLine (x0, y0, z0,
		       x1, y1, z1);
    }
  }

  if (GroupIdIsShown) {
    AutoGL_SetColor (1, 0, 0);

    for (iElement = 0; iElement < nElements; iElement++) {
      int nodeId0 = AutoMesh_SurfaceElement3D_nodeIdAt (Element, iElement, 0);
      int nodeId1 = AutoMesh_SurfaceElement3D_nodeIdAt (Element, iElement, 1);
      int nodeId2 = AutoMesh_SurfaceElement3D_nodeIdAt (Element, iElement, 2);
      int groupId = AutoMesh_SurfaceGrouper3D_groupIdAtElement
	(Grouper, iElement);

      double x0, y0, z0;
      double x1, y1, z1;
      double x2, y2, z2;
      double cx, cy, cz;
      
      AutoMesh_Node3D_getCoordAt (Node, nodeId0,
				  &x0, &y0, &z0);
      AutoMesh_Node3D_getCoordAt (Node, nodeId1,
				  &x1, &y1, &z1);
      AutoMesh_Node3D_getCoordAt (Node, nodeId2,
				  &x2, &y2, &z2);
      cx = (x0 + x1 + x2) / 3;
      cy = (y0 + y1 + y2) / 3;
      cz = (z0 + z1 + z2) / 3;
      
      sprintf(buf, "%d", groupId);
      AutoGL_DrawString (cx, cy, cz, buf);
    }
  }
}

static void SelectGroup (void) 
{
  int selectedGroupId = -1;
  int nElements = AutoMesh_SurfaceElement3D_nElements (Element);
  int dcX, dcY, dcZ;
  int toleranceDc = 1;
  int iElement;

  AutoGL_GetPointingDevicePositionDc 
    (&dcX, &dcY, &dcZ);
  
  AutoGL_StartSelection (dcX, dcY, toleranceDc);
  
  for (iElement = 0; iElement < nElements; iElement++) {
    int nodeId0 = AutoMesh_SurfaceElement3D_nodeIdAt (Element, iElement, 0);
    int nodeId1 = AutoMesh_SurfaceElement3D_nodeIdAt (Element, iElement, 1);
    int nodeId2 = AutoMesh_SurfaceElement3D_nodeIdAt (Element, iElement, 2);
    int groupId = AutoMesh_SurfaceGrouper3D_groupIdAtElement
      (Grouper, iElement);

    double x0, y0, z0;
    double x1, y1, z1;
    double x2, y2, z2;
    
    AutoMesh_Node3D_getCoordAt (Node, nodeId0,
				&x0, &y0, &z0);
    AutoMesh_Node3D_getCoordAt (Node, nodeId1,
				&x1, &y1, &z1);
    AutoMesh_Node3D_getCoordAt (Node, nodeId2,
				&x2, &y2, &z2);

    AutoGL_SetSelectionId (groupId);
    AutoGL_DrawTriangle (x0, y0, z0,
			 x1, y1, z1,
			 x2, y2, z2);
  }

  AutoGL_EndSelection ();
  selectedGroupId = AutoGL_GetSelectedId ();
  assert(-1 <= selectedGroupId);
  assert(selectedGroupId < NGroups);

  if (0 <= selectedGroupId) {
    if (AUTORES_I(Groups_isSelected, selectedGroupId)) {
      AUTORES_I_SET(Groups_isSelected, selectedGroupId, 0);
    } else {
      AUTORES_I_SET(Groups_isSelected, selectedGroupId, 1);
    }
  }
}

static void RedrawView (void)
{
  DrawGroup ();
}

static void HandleEvent (void)
{
  int event = AutoGL_GetViewEvent ();

  if (event == AUTOGL_EVENT_POINTING_DEVICE_RELEASE) {
    SelectGroup ();
    AutoGL_DrawView ();
  } else if (event == AUTOGL_EVENT_KEY) {
    AutoGL_HandleDefaultKeyEventInMode3D ();
  }
}

static void Quit (void) 
{
  AutoGL_Terminate ();
  exit (0);
}

static void Print (void) 
{
  int iGroup;
  int count;
  
  count = 0;
  for (iGroup = 0; iGroup < NGroups; iGroup++) {
    if (AUTORES_I(Groups_isSelected, iGroup)) {
      count++;
    }
  }
  printf ("%d\n", count);

  count = 0;
  for (iGroup = 0; iGroup < NGroups; iGroup++) {
    if (AUTORES_I(Groups_isSelected, iGroup)) {
      printf ("%d   %d\n", count, iGroup);
      count++;
    }
  }
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
#ifdef BINARY
  int binaryFlag = 1;
  int recordIdFlag = 0;
#else
  int binaryFlag = 0;
  int recordIdFlag = 1;
#endif
  int flag;

  if (argc != 5) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_tri3d_group_multi_select  node_file  element_file \n"
       "        element_group_file  frame_file  \n");
    exit (1);
  }

  AutoRes_Initialize (0);
  AutoMesh_Initialize ();

  AutoRes_TableInOut_SetBinaryFlag (binaryFlag);
  AutoRes_TableInOut_SetRecordIdFlag (recordIdFlag);



  Node = AutoMesh_Node3D_New ();
  assert(Node != NULL);

  flag = AutoMesh_Node3D_makeCoordColumn (Node);
  assert(flag);

  AutoRes_TableInOut_AssociateReal1DColumn (Node->nodes_coord);
  flag = AutoRes_TableInOut_Read (argv[1]);
  assert(flag);
  


  Element = AutoMesh_SurfaceElement3D_New 
    (Node, AUTOMESH_POLYGON_TRI, 3);
  assert(Element != NULL);

  flag = AutoMesh_SurfaceElement3D_makeNodeIdsColumn (Element);
  assert(flag);

  AutoRes_TableInOut_AssociateInteger1DColumn (Element->elements_nodeIds);
  flag = AutoRes_TableInOut_Read (argv[2]);
  assert(flag);
  


  Grouper = AutoMesh_SurfaceGrouper3D_NewWithoutEdge (Element);
  assert(Grouper != NULL);

  flag = AutoMesh_SurfaceGrouper3D_makeGroupIdAtElementColumn (Grouper);
  assert(flag);

  AutoRes_TableInOut_AssociateIntegerColumn (Grouper->elements_groupId);
  flag = AutoRes_TableInOut_Read (argv[3]);
  assert(flag);

  AutoMesh_SurfaceGrouper3D_setNGroups (Grouper);
  NGroups = AutoMesh_SurfaceGrouper3D_nGroups (Grouper);

  fprintf (stderr, " NGroups %d \n",
	   NGroups);
  
  InitializeGroupSelection ();



  Frame = AutoMesh_Frame3D_New ();
  assert(Frame != NULL);

  flag = AutoMesh_Frame3D_makeEndPointsColumn (Frame);
  assert(flag);

  AutoRes_TableInOut_AssociateReal1DColumn (Frame->frames_endPoint0);
  AutoRes_TableInOut_AssociateReal1DColumn (Frame->frames_endPoint1);
  flag = AutoRes_TableInOut_Read (argv[4]);
  assert(flag);



  SetView ();
  
  AutoGL_SetViewRedrawCallback (RedrawView);

  AutoGL_SetDefaultCallbackInMode3D (HandleEvent);
  AutoGL_EnableDragInMode3D ();



  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode3D ();

  AutoGL_AddCallback (Quit, "Quit");       

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");

  AutoGL_AddBoolean (&Filled, "Filled");
  AutoGL_AddBoolean (&FrameIsShown, "FrameIsShown");
  AutoGL_AddBoolean (&GroupIdIsShown, "GroupIdIsShown");

  AutoGL_AddCallback (Print, "Print");       
}
