/* advautomesh_tetra3d_face2_stageA.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>



/************************
	 Basic
*************************/

void SortId3 (int id0, int id1, int id2,
	      int *minId_OUT, int *midId_OUT, int *maxId_OUT)
{
  assert(id0 != id1);
  assert(id0 != id2);
  assert(id1 != id2);

  if (id0 < id1 && id0 < id2) {
    *minId_OUT = id0;
  }
  if (id1 < id0 && id1 < id2) {
    *minId_OUT = id1;
  }
  if (id2 < id0 && id2 < id1) {
    *minId_OUT = id2;
  }

  if ((id1 < id0 && id0 < id2)
      || (id2 < id0 && id0 < id1)) {
    *midId_OUT = id0;
  }
  if ((id0 < id1 && id1 < id2)
      || (id2 < id1 && id1 < id0)) {
    *midId_OUT = id1;
  }
  if ((id0 < id2 && id2 < id1)
      || (id1 < id2 && id2 < id0)) {
    *midId_OUT = id2;
  }

  if (id1 < id0 && id2 < id0) {
    *maxId_OUT = id0;
  }
  if (id0 < id1 && id2 < id1) {
    *maxId_OUT = id1;
  }
  if (id0 < id2 && id1 < id2) {
    *maxId_OUT = id2;
  }
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 2500000

int NNodes;



/* Element */

#define N_ELEMENT_NODES 4
#define N_ELEMENT_FACES 4
#define MAX_ELEMENTS 12000000

const int FaceNodeMap[N_ELEMENT_FACES][3] = {
  { 1, 2, 3 },
  { 0, 3, 2 },
  { 0, 1, 3 },
  { 0, 2, 1 }
};

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



/* Face */

#define N_FACE_NODES 3
#define MAX_FACES 30000000

int NFaces;
int Face_elementIds[MAX_FACES][2];
int Face_elementFaceIds[MAX_FACES][2];

AutoRes_ListTable FaceTable;



void Face_setRearElement 
(int faceId, 
 int elementId, int elementFaceId)
{
  assert(0 <= faceId);
  assert(faceId < NFaces);
  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(0 <= elementFaceId);
  assert(elementFaceId < N_ELEMENT_FACES);

  Face_elementIds[faceId][1] = elementId;
  Face_elementFaceIds[faceId][1] = elementFaceId;
}

void Face_getNodes 
(int faceId,
 int *nodeId0_OUT, int *nodeId1_OUT, int *nodeId2_OUT)
{
  int elementId;
  int elementFaceId;
  
  assert(0 <= faceId);
  assert(faceId < NFaces);
  
  elementId = Face_elementIds[faceId][0];
  elementFaceId = Face_elementFaceIds[faceId][0];
  *nodeId0_OUT = Element_nodeIds[elementId][FaceNodeMap[elementFaceId][0]];
  *nodeId1_OUT = Element_nodeIds[elementId][FaceNodeMap[elementFaceId][1]];
  *nodeId2_OUT = Element_nodeIds[elementId][FaceNodeMap[elementFaceId][2]];
}

int Mesh_addNewFace (int elementId, int elementFaceId)
{
  int faceId = NFaces;
  int hashTableSize;
  int hashValue;
  int nodeId0, nodeId1, nodeId2;

  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(0 <= elementFaceId);
  assert(elementFaceId < N_ELEMENT_FACES);

  assert(NFaces < MAX_FACES);
  NFaces++;

  Face_elementIds[faceId][0] = elementId;
  Face_elementIds[faceId][1] = -1;
  Face_elementFaceIds[faceId][0] = elementFaceId;
  Face_elementFaceIds[faceId][1] = -1;
  
  Face_getNodes (faceId,
		 &nodeId0, &nodeId1, &nodeId2);

  hashTableSize = AutoRes_ListTable_nLists 
    (&FaceTable);
  hashValue = (nodeId0 + nodeId1 + nodeId2) % hashTableSize;
  AutoRes_ListTable_addItemIn 
    (&FaceTable,
     hashValue, faceId);

  return faceId;
}

int Mesh_findFace (int nodeId0, int nodeId1, int nodeId2)
{
  int minNodeId, midNodeId, maxNodeId;
  int hashTableSize;
  int hashValue;
  int linkIndex;

  assert(0 <= nodeId0);
  assert(nodeId0 < NNodes);
  assert(0 <= nodeId1);
  assert(nodeId1 < NNodes);
  assert(0 <= nodeId2);
  assert(nodeId2 < NNodes);

  SortId3 (nodeId0, nodeId1, nodeId2,
	   &minNodeId, &midNodeId, &maxNodeId);

  hashTableSize = AutoRes_ListTable_nLists 
    (&FaceTable);
  hashValue = (nodeId0 + nodeId1 + nodeId2) % hashTableSize;
  for (linkIndex = AutoRes_ListTable_startLinkIndexIn 
	 (&FaceTable, 
	  hashValue);
       linkIndex != -1;
       linkIndex = AutoRes_ListTable_nextLinkIndex 
	 (&FaceTable, 
	  linkIndex)) {
    int faceId = AutoRes_ListTable_itemId 
      (&FaceTable, 
       linkIndex);
    int faceNodeId0, faceNodeId1, faceNodeId2;
    int minFaceNodeId, midFaceNodeId, maxFaceNodeId; 
    
    Face_getNodes (faceId,
		   &faceNodeId0, &faceNodeId1, &faceNodeId2);
    SortId3 (faceNodeId0, faceNodeId1, faceNodeId2,
	     &minFaceNodeId, &midFaceNodeId, &maxFaceNodeId);

    if (minFaceNodeId == minNodeId
	&& midFaceNodeId == midNodeId
	&& maxFaceNodeId == maxNodeId) {
      return faceId;
    }
  }

  return -1;
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  /* skip coordinate */

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void WriteFaceFile (const char *fileName)
{
  FILE *fp;
  int iFace;
  
  assert(0 < NFaces);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  fprintf (fp, "%d\n", 
	   NFaces);
  
  for (iFace = 0; iFace < NFaces; iFace++) {
    int nodeId0, nodeId1, nodeId2;
    
    Face_getNodes (iFace,
		   &nodeId0, &nodeId1, &nodeId2);
    
    fprintf (fp, "%d   %d %d %d   %d %d   %d %d\n", 
	     iFace,
	     nodeId0, nodeId1, nodeId2,
	     Face_elementIds[iFace][0], 
	     Face_elementFaceIds[iFace][0],
	     Face_elementIds[iFace][1], 
	     Face_elementFaceIds[iFace][1]);
  }

  fclose (fp);
}



/************************
	 Command
*************************/

void PerformCommand (void)
{
  int hashTableSize = NNodes * 10;
  int nAverageLinksPerList = 2;
  int iElement;
  int iElementFace;
  int iFace;

  AutoRes_ListTable_initialize 
    (&FaceTable,
     hashTableSize, nAverageLinksPerList);
  
  assert(NFaces == 0);

  for (iElement = 0; iElement < NElements; iElement++) {
    for (iElementFace = 0; 
	 iElementFace < N_ELEMENT_FACES; 
	 iElementFace++) {
      int faceNodeId0 
	= Element_nodeIds[iElement][FaceNodeMap[iElementFace][0]];
      int faceNodeId1 
	= Element_nodeIds[iElement][FaceNodeMap[iElementFace][1]];
      int faceNodeId2 
	= Element_nodeIds[iElement][FaceNodeMap[iElementFace][2]];
      int faceId = Mesh_findFace (faceNodeId0, faceNodeId1, faceNodeId2);

      if (faceId == -1) {
	faceId = Mesh_addNewFace (iElement, iElementFace);
      } else {
	Face_setRearElement (faceId, iElement, iElementFace);
      }
    }
  }

  for (iFace = 0; iFace < NFaces; iFace++) {
    assert(Face_elementIds[iFace][0] != -1);
    assert(Face_elementIds[iFace][1] != -1
	   || Face_elementIds[iFace][1] == -1);
  }

  AutoRes_ListTable_terminate 
    (&FaceTable);
}



#if 1
#include "../../util/time_measure.c"
#define TIME_MEASURE
#endif

int main (int argc, char *argv[])
{
  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_tetra3d_face2_stageA  node_file  element_file \n"
       "        face_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);

#ifdef TIME_MEASURE
  StartMeasureTime ();
#endif

  PerformCommand ();

#ifdef TIME_MEASURE
  EndMeasureTime ();
#endif

  WriteFaceFile (argv[3]);

  return 0;
}
