/* advautomesh_tetra3d_isosurface_view2.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>
#include <autogl.h>

#include <GL/gl.h>
#include <GL/glu.h>



#if 1
#include "../../util/time_measure.c"
#define TIME_MEASURE
#endif



/************************
	 Basic
*************************/

double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}

void GetNormalizedVector 
(double x, double y, double z,
 double *nx_OUT, double *ny_OUT, double *nz_OUT)
{
  double length = sqrt (x * x + y * y + z * z);

  *nx_OUT = x / length;
  *ny_OUT = y / length;
  *nz_OUT = z / length;
}

void GetVectorProduct 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  *x_OUT = y0 * z1 - z0 * y1;
  *y_OUT = z0 * x1 - x0 * z1;
  *z_OUT = x0 * y1 - y0 * x1;
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 2500000

int NNodes;
double Node_coord[MAX_NODES][3];
double Node_scalar[MAX_NODES];



/* Element */

#define N_ELEMENT_NODES 4
#define N_ELEMENT_EDGES 6
#define MAX_ELEMENTS 12000000

const int EdgeNodeMap[N_ELEMENT_EDGES][2] = {
  { 0, 1 }, 
  { 0, 2 }, 
  { 0, 3 }, 
  { 1, 2 }, 
  { 2, 3 }, 
  { 3, 1 }  
};

const int NIsosurfaceEdges[8] = {
  0, 3, 3, 4, 3, 4, 4, 3
};

const int IsosurfaceEdgeIds[8][4] = {
  { -1, -1, -1, -1 },
  { 0, 1, 2, -1 },
  { 0, 3, 5, -1 },
  { 3, 1, 2, 5 },
  { 3, 1, 4, -1 },
  { 0, 3, 4, 2 },
  { 0, 5, 4, 1 },
  { 2, 5, 4, -1 }
};

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



void Element_getEdgePoint 
(int elementId, 
 int elementEdgeId, double ratio,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  int nodeId0, nodeId1;
  double x0, y0, z0;
  double x1, y1, z1;

  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(0 <= elementEdgeId);
  assert(elementEdgeId < N_ELEMENT_EDGES);

  nodeId0 = Element_nodeIds
    [elementId][EdgeNodeMap[elementEdgeId][0]];
  nodeId1 = Element_nodeIds
    [elementId][EdgeNodeMap[elementEdgeId][1]];

  x0 = Node_coord[nodeId0][0];
  y0 = Node_coord[nodeId0][1];
  z0 = Node_coord[nodeId0][2];

  x1 = Node_coord[nodeId1][0];
  y1 = Node_coord[nodeId1][1];
  z1 = Node_coord[nodeId1][2];
  
  *x_OUT = x0 + (x1 - x0) * ratio;
  *y_OUT = y0 + (y1 - y0) * ratio;
  *z_OUT = z0 + (z1 - z0) * ratio;
}

double Element_edgeScalar 
(int elementId, 
 int elementEdgeId, double ratio)
{
  int nodeId0, nodeId1;
  double scalar0, scalar1;

  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(0 <= elementEdgeId);
  assert(elementEdgeId < N_ELEMENT_EDGES);

  nodeId0 = Element_nodeIds
    [elementId][EdgeNodeMap[elementEdgeId][0]];
  nodeId1 = Element_nodeIds
    [elementId][EdgeNodeMap[elementEdgeId][1]];

  scalar0 = Node_scalar[nodeId0];
  scalar1 = Node_scalar[nodeId1];
  
  return scalar0 + (scalar1 - scalar0) * ratio;
}

void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  *minRangeZ_OUT = *maxRangeZ_OUT = Node_coord[0][2];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    double z = Node_coord[iNode][2];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
    if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
  }
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void ReadNodeScalarFile (const char *fileName)
{
  FILE *fp;
  int count;
  int iNode;
  
  assert(0 < NNodes);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &count);
  assert(count == NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double scalar;

    fscanf (fp, "%d %lf", 
	    &nodeId,
	    &scalar);
    assert(nodeId == iNode);
    Node_scalar[iNode] = scalar;
  }

  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Isosurface */

#define MAX_LEVELS 10000
#define MAX_ISOSURFACES 1000000

int NLevels;
int Level_startIsosurfaceId[MAX_LEVELS];
int Level_nIsosurfaces[MAX_LEVELS];
int NIsosurfaces;
int Isosurface_elementId[MAX_ISOSURFACES];
int Isosurface_elementEdgeIds[MAX_ISOSURFACES][3];
double Isosurface_ratios[MAX_ISOSURFACES][3];

float Isosurface_vertexs[MAX_ISOSURFACES][3][3];
float Isosurface_normal[MAX_ISOSURFACES][3];



/* Wireframe */

#define MAX_WIREFRAMES 1000000

int NWireframes;
double Wireframe_coord0[MAX_WIREFRAMES][3];
double Wireframe_coord1[MAX_WIREFRAMES][3];

float Wireframe_vertexs[MAX_WIREFRAMES][2][3];



void IsosurfaceGroup_clear (void)
{
  NLevels = 0;
  NIsosurfaces = 0;
}

void IsosurfaceGroup_addLevel (void)
{
  assert(NLevels < MAX_LEVELS);

  Level_startIsosurfaceId[NLevels] = NIsosurfaces;
  Level_nIsosurfaces[NLevels] = 0;
  NLevels++;
}

void IsosurfaceGroup_addIsosurface 
(int elementId, 
 int elementEdgeId0, double ratio0,
 int elementEdgeId1, double ratio1,
 int elementEdgeId2, double ratio2)
{
  assert(0 < NLevels);
  assert(NIsosurfaces < MAX_ISOSURFACES);
  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(0 <= elementEdgeId0);
  assert(elementEdgeId0 < N_ELEMENT_EDGES);
  assert(0 <= elementEdgeId1);
  assert(elementEdgeId1 < N_ELEMENT_EDGES);
  assert(0 <= elementEdgeId2);
  assert(elementEdgeId2 < N_ELEMENT_EDGES);

  Level_nIsosurfaces[NLevels - 1]++;

  Isosurface_elementId[NIsosurfaces] = elementId;
  Isosurface_elementEdgeIds[NIsosurfaces][0] = elementEdgeId0;
  Isosurface_elementEdgeIds[NIsosurfaces][1] = elementEdgeId1;
  Isosurface_elementEdgeIds[NIsosurfaces][2] = elementEdgeId2;
  Isosurface_ratios[NIsosurfaces][0] = ratio0;
  Isosurface_ratios[NIsosurfaces][1] = ratio1;
  Isosurface_ratios[NIsosurfaces][2] = ratio2;

  NIsosurfaces++;
}

void ReadWireframeFile (const char *fileName)
{
  FILE *fp;
  int iWireframe;
  
  assert(NWireframes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NWireframes);
  assert(0 <= NWireframes);
  assert(NWireframes <= MAX_WIREFRAMES);
  
  for (iWireframe = 0; iWireframe < NWireframes; iWireframe++) {
    int wireframeId;
    double x0, y0, z0;
    double x1, y1, z1;

    fscanf (fp, "%d %lf %lf %lf %lf %lf %lf", 
	    &wireframeId,
	    &x0, &y0, &z0, 
	    &x1, &y1, &z1);
    assert(wireframeId == iWireframe);
    Wireframe_coord0[iWireframe][0] = x0;
    Wireframe_coord0[iWireframe][1] = y0;
    Wireframe_coord0[iWireframe][2] = z0;

    Wireframe_coord1[iWireframe][0] = x1;
    Wireframe_coord1[iWireframe][1] = y1;
    Wireframe_coord1[iWireframe][2] = z1;
  }

  fclose (fp);
}



/************************
	 Command
*************************/

AutoRes_Bucket1D IsosurfaceBucket;

void CreateIsosurfaceBucket (void)
{
  int iNode;
  double minRange, maxRange;
  int nDivisions;
  double offset;
  int nAverageLinksPerCell;
  int nOffsetDivisions = 23;
  
  minRange = maxRange = Node_scalar[0];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double value = Node_scalar[iNode];
    
    if (value < minRange) minRange = value;
    if (maxRange < value) maxRange = value;
  }

  nDivisions = (int)(pow (NNodes, 0.333333));
  if (nDivisions < 1) nDivisions = 1;
  nAverageLinksPerCell = (int)((double)NElements / nDivisions * 3.0);

  offset = (maxRange - minRange) / nOffsetDivisions;
  
  AutoRes_Bucket1D_initialize 
    (&IsosurfaceBucket,
     minRange - offset,
     maxRange + offset,
     nDivisions, 
     nAverageLinksPerCell);
}

void AddElementsIntoIsosurfaceBucket (void)
{
  int iElement;
  int iElementNode;

  for (iElement = 0; iElement < NElements; iElement++) {
    int nodeId0 = Element_nodeIds[iElement][0];
    double minRange, maxRange;
    
    minRange = maxRange = Node_scalar[nodeId0];
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId = Element_nodeIds[iElement][iElementNode]; 
      double value = Node_scalar[nodeId];
      
      if (value < minRange) minRange = value;
      if (maxRange < value) maxRange = value;
    }
    
    AutoRes_Bucket1D_addItemWithin 
      (&IsosurfaceBucket,
       minRange,
       maxRange,
       iElement);
  }
}

void DeleteIsosurfaceBucket (void)
{
  AutoRes_Bucket1D_terminate 
    (&IsosurfaceBucket);
}

int IsosurfaceMask (int elementId, double levelValue)
{
  int nodeId0 = Element_nodeIds[elementId][0];
  int nodeId1 = Element_nodeIds[elementId][1];
  int nodeId2 = Element_nodeIds[elementId][2];
  int nodeId3 = Element_nodeIds[elementId][3];
  unsigned int mask = 0;
  
  if (levelValue < Node_scalar[nodeId0]) mask |= 1;
  if (levelValue < Node_scalar[nodeId1]) mask |= 2;
  if (levelValue < Node_scalar[nodeId2]) mask |= 4;
  if (levelValue < Node_scalar[nodeId3]) mask |= 8;
  if ((mask & 8) != 0) mask = (~mask) & 15;

  assert(0 <= mask);
  assert(mask < 8);

  return mask;
}

int GetIsosurface 
(int elementId, double levelValue,
 int *nEdges_OUT,
 int elementEdgeIds_OUT[4], double ratios_OUT[4])
{
  unsigned int mask = IsosurfaceMask (elementId, levelValue);
  int iEdge;

  *nEdges_OUT = NIsosurfaceEdges[mask];
  if (*nEdges_OUT == 0) return 0;
  
  assert(*nEdges_OUT == 3 || *nEdges_OUT == 4);

  for (iEdge = 0; iEdge < *nEdges_OUT; iEdge++) {
    int elementEdgeId = IsosurfaceEdgeIds[mask][iEdge];

    int elementNodeId0 = EdgeNodeMap[elementEdgeId][0];
    int elementNodeId1 = EdgeNodeMap[elementEdgeId][1];

    int nodeId0 = Element_nodeIds[elementId][elementNodeId0];
    int nodeId1 = Element_nodeIds[elementId][elementNodeId1];

    double ratio = (levelValue - Node_scalar[nodeId0]) 
      / (Node_scalar[nodeId1] - Node_scalar[nodeId0]);
    
    elementEdgeIds_OUT[iEdge] = elementEdgeId; 
    ratios_OUT[iEdge] = ratio;
  }    

  return 1;
}

void CheckElementIsosurface (int elementId, double levelValue)
{
  int nEdges;
  int elementEdgeIds[4];
  double ratios[4];
  
  if (GetIsosurface
      (elementId, levelValue,
       &nEdges, elementEdgeIds, ratios)) {
    assert(nEdges == 3 || nEdges == 4);
    IsosurfaceGroup_addIsosurface 
      (elementId, 
       elementEdgeIds[0], ratios[0],
       elementEdgeIds[1], ratios[1],
       elementEdgeIds[2], ratios[2]);
    if (nEdges == 4) {
      IsosurfaceGroup_addIsosurface 
	(elementId, 
	 elementEdgeIds[2], ratios[2],
	 elementEdgeIds[3], ratios[3],
	 elementEdgeIds[0], ratios[0]);
    }
  }
}

void MakeIsosurface (double levelValue)
{
  int cx;
  int linkIndex;
  
  AutoRes_Bucket1D_getIndexAt 
    (&IsosurfaceBucket, 
     levelValue,
     &cx);
  
  if (AutoRes_Bucket1D_isOutOfRange 
      (&IsosurfaceBucket, 
       cx)) {
    return;
  } else {
    for (linkIndex = AutoRes_Bucket1D_startLinkIndexIn 
	   (&IsosurfaceBucket, 
	    cx);
	 linkIndex != -1;
	 linkIndex = AutoRes_Bucket1D_nextLinkIndex 
	   (&IsosurfaceBucket, 
	    linkIndex)) {
      int elementId = AutoRes_Bucket1D_itemId 
	(&IsosurfaceBucket, 
	 linkIndex);
      
      CheckElementIsosurface (elementId, levelValue);
    }
  }
}



/********************************
	 Visualization
********************************/

double MinRange = 0.0;
double MaxRange = 1.0;

double MinLevel = 0.0;
double MaxLevel = 1.0;
int NDivisions = 1;

double Offset = 0.0;
double Shift = 1.0;

void SetView (void)
{
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  double centerX, centerY, centerZ;
  double size;
  
  Mesh_getRange 
    (&minRangeX, &minRangeY, &minRangeZ,
     &maxRangeX, &maxRangeY, &maxRangeZ);
  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;
  centerZ = (minRangeZ + maxRangeZ) / 2;
  size = Distance (minRangeX, minRangeY, minRangeZ,
		   maxRangeX, maxRangeY, maxRangeZ) / 2;

  AutoGL_SetViewSize (size);
  AutoGL_SetViewCenter (centerX, centerY, centerZ); 
}

void SetScalarRange (void)
{
  int iNode;

  assert(0 < NNodes);

  MinRange = MaxRange = Node_scalar[0];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double scalar = Node_scalar[iNode];
    
    if (scalar < MinRange) MinRange = scalar;
    if (MaxRange < scalar) MaxRange = scalar;
  }
}

void PrepareIsosurfaceVertexAndNormal (void)
{
  int iIsosurface;

  for (iIsosurface = 0; iIsosurface < NIsosurfaces; iIsosurface++) {
    double ex0, ey0, ez0;
    double ex1, ey1, ez1;
    double ex2, ey2, ez2;
    double nx, ny, nz;

    Element_getEdgePoint 
      (Isosurface_elementId[iIsosurface], 
       Isosurface_elementEdgeIds[iIsosurface][0],
       Isosurface_ratios[iIsosurface][0],
       &ex0, &ey0, &ez0);

    Element_getEdgePoint 
      (Isosurface_elementId[iIsosurface], 
       Isosurface_elementEdgeIds[iIsosurface][1],
       Isosurface_ratios[iIsosurface][1],
       &ex1, &ey1, &ez1);

    Element_getEdgePoint 
      (Isosurface_elementId[iIsosurface], 
       Isosurface_elementEdgeIds[iIsosurface][2],
       Isosurface_ratios[iIsosurface][2],
       &ex2, &ey2, &ez2);

    Isosurface_vertexs[iIsosurface][0][0] = ex0;
    Isosurface_vertexs[iIsosurface][0][1] = ey0;
    Isosurface_vertexs[iIsosurface][0][2] = ez0;
    
    Isosurface_vertexs[iIsosurface][1][0] = ex1;
    Isosurface_vertexs[iIsosurface][1][1] = ey1;
    Isosurface_vertexs[iIsosurface][1][2] = ez1;
    
    Isosurface_vertexs[iIsosurface][2][0] = ex2;
    Isosurface_vertexs[iIsosurface][2][1] = ey2;
    Isosurface_vertexs[iIsosurface][2][2] = ez2;

    GetVectorProduct (ex0 - ex1, ey0 - ey1, ez0 - ez1,
		      ex0 - ex2, ey0 - ey2, ez0 - ez2,
		      &nx, &ny, &nz);
    GetNormalizedVector (nx, ny, nz,
			 &nx, &ny, &nz);
    Isosurface_normal[iIsosurface][0] = nx;
    Isosurface_normal[iIsosurface][1] = ny;
    Isosurface_normal[iIsosurface][2] = nz;
  }
}

void PrepareWireframe (void)
{
  int iWireframe;

  for (iWireframe = 0; iWireframe < NWireframes; iWireframe++) {
    double nodeX0 = Wireframe_coord0[iWireframe][0];
    double nodeY0 = Wireframe_coord0[iWireframe][1];
    double nodeZ0 = Wireframe_coord0[iWireframe][2];
    
    double nodeX1 = Wireframe_coord1[iWireframe][0];
    double nodeY1 = Wireframe_coord1[iWireframe][1];
    double nodeZ1 = Wireframe_coord1[iWireframe][2];
    
    Wireframe_vertexs[iWireframe][0][0] = nodeX0;
    Wireframe_vertexs[iWireframe][0][1] = nodeY0;
    Wireframe_vertexs[iWireframe][0][2] = nodeZ0;

    Wireframe_vertexs[iWireframe][1][0] = nodeX1;
    Wireframe_vertexs[iWireframe][1][1] = nodeY1;
    Wireframe_vertexs[iWireframe][1][2] = nodeZ1;
  }
}

void DrawWireframe (void)
{
  int iWireframe;

  AutoGL_BeginNativeCall ();

  glDisable (GL_LIGHTING);
  
  glColor4f (1.0f, 1.0f, 1.0f, 1.0f);
  
  glBegin (GL_LINES);
  for (iWireframe = 0; iWireframe < NWireframes; iWireframe++) {
    glVertex3f (Wireframe_vertexs[iWireframe][0][0], 
		Wireframe_vertexs[iWireframe][0][1], 
		Wireframe_vertexs[iWireframe][0][2]);
    glVertex3f (Wireframe_vertexs[iWireframe][1][0], 
		Wireframe_vertexs[iWireframe][1][1], 
		Wireframe_vertexs[iWireframe][1][2]);
  }
  glEnd ();

  AutoGL_EndNativeCall ();
}

void RedrawView (void) 
{
  int iLevel;
  int iIsosurface;

#ifdef TIME_MEASURE
  StartMeasureTime ();
#endif

  DrawWireframe ();

  AutoGL_BeginNativeCall ();

  glEnable (GL_LIGHTING);

  glBegin (GL_TRIANGLES);
  for (iLevel = 0; iLevel < NLevels; iLevel++) {
    int startId = Level_startIsosurfaceId[iLevel];
    int nIsosurfaces = Level_nIsosurfaces[iLevel];
    double levelValue = MinLevel 
      + (MaxLevel - MinLevel) / NDivisions * iLevel + Offset;
    double grade = (levelValue - MinRange) / (MaxRange - MinRange);
    double red, green, blue;

    AutoGL_GetContourColor (&red, &green, &blue,
			    grade);
    glColor4f ((float)red, (float)green, (float)blue, 1.0f);
    
    for (iIsosurface = startId; 
	 iIsosurface < startId + nIsosurfaces; 
	 iIsosurface++) {
      glNormal3f (Isosurface_normal[iIsosurface][0], 
		  Isosurface_normal[iIsosurface][1], 
		  Isosurface_normal[iIsosurface][2]);
      
      glVertex3f (Isosurface_vertexs[iIsosurface][0][0], 
		  Isosurface_vertexs[iIsosurface][0][1], 
		  Isosurface_vertexs[iIsosurface][0][2]);
      glVertex3f (Isosurface_vertexs[iIsosurface][1][0], 
		  Isosurface_vertexs[iIsosurface][1][1], 
		  Isosurface_vertexs[iIsosurface][1][2]);
      glVertex3f (Isosurface_vertexs[iIsosurface][2][0], 
		  Isosurface_vertexs[iIsosurface][2][1], 
		  Isosurface_vertexs[iIsosurface][2][2]);
      
      glNormal3f (Isosurface_normal[iIsosurface][0] * -1.0f, 
		  Isosurface_normal[iIsosurface][1] * -1.0f, 
		  Isosurface_normal[iIsosurface][2] * -1.0f);
      
      glVertex3f (Isosurface_vertexs[iIsosurface][2][0], 
		  Isosurface_vertexs[iIsosurface][2][1], 
		  Isosurface_vertexs[iIsosurface][2][2]);
      glVertex3f (Isosurface_vertexs[iIsosurface][1][0], 
		  Isosurface_vertexs[iIsosurface][1][1], 
		  Isosurface_vertexs[iIsosurface][1][2]);
      glVertex3f (Isosurface_vertexs[iIsosurface][0][0], 
		  Isosurface_vertexs[iIsosurface][0][1], 
		  Isosurface_vertexs[iIsosurface][0][2]);
    }
  }
  glEnd ();

  AutoGL_EndNativeCall ();

#ifdef TIME_MEASURE
  EndMeasureTime ();
#endif
}

void PrepareCommand (void) 
{
#ifdef TIME_MEASURE
  StartMeasureTime ();
#endif

  CreateIsosurfaceBucket ();
  AddElementsIntoIsosurfaceBucket ();
  PrepareWireframe ();

#ifdef TIME_MEASURE
  EndMeasureTime ();
#endif
}

void MakeIsosurfaceGroup (void) 
{
  int iLevel;

  IsosurfaceGroup_clear ();
  for (iLevel = 0; iLevel <= NDivisions; iLevel++) {
    double levelValue = MinLevel 
      + (MaxLevel - MinLevel) / NDivisions * iLevel;
    
    IsosurfaceGroup_addLevel ();
    MakeIsosurface (levelValue + Offset);
  }

  PrepareIsosurfaceVertexAndNormal ();
}

void SetLevel (void) 
{
#ifdef TIME_MEASURE
  StartMeasureTime ();
#endif

  Offset = 0.0;
  MakeIsosurfaceGroup ();

#ifdef TIME_MEASURE
  EndMeasureTime ();
#endif

  AutoGL_DrawView ();
}

void Increment (void) 
{
#ifdef TIME_MEASURE
  StartMeasureTime ();
#endif

  Offset += Shift;
  MakeIsosurfaceGroup ();

#ifdef TIME_MEASURE
  EndMeasureTime ();
#endif

  AutoGL_DrawView ();
}

void Decrement (void) 
{
#ifdef TIME_MEASURE
  StartMeasureTime ();
#endif

  Offset -= Shift;
  MakeIsosurfaceGroup ();

#ifdef TIME_MEASURE
  EndMeasureTime ();
#endif

  AutoGL_DrawView ();
}

void UpdateModel (void) 
{
  AutoGL_DrawView ();
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
  if (argc != 5) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_tetra3d_isosurface_view2  node_file  element_file \n"
       "     node_scalar_file  wireframe_file \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadNodeScalarFile (argv[3]);
  ReadWireframeFile (argv[4]);

  SetView ();
  SetScalarRange ();
  PrepareCommand ();
  
  AutoGL_SetViewRedrawCallback (RedrawView);

  AutoGL_SetDefaultCallbackInMode3D (NULL);
  AutoGL_EnableDragInMode3D ();



  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode3D ();

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");

  AutoGL_AddReal (&MinRange, "MinRange");
  AutoGL_AddReal (&MaxRange, "MaxRange");

  AutoGL_AddReal (&MinLevel, "MinLevel");
  AutoGL_AddReal (&MaxLevel, "MaxLevel");
  AutoGL_AddInteger (&NDivisions, "NDivisions");
  AutoGL_AddCallback (SetLevel, "SetLevel");      

  AutoGL_AddReal (&Shift, "Shift");
  AutoGL_AddCallback (Increment, "Increment");      
  AutoGL_AddCallback (Decrement, "Decrement");      

  AutoGL_AddCallback (UpdateModel, "UpdateModel");      
}
