/* advautomesh_tetra3d_sample3.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>



/************************
	 Basic
*************************/

#define VOLUME_COORD_EPS 0.00001



double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}

void GetNormalizedVector 
(double x, double y, double z,
 double *nx_OUT, double *ny_OUT, double *nz_OUT)
{
  double length = sqrt (x * x + y * y + z * z);

  *nx_OUT = x / length;
  *ny_OUT = y / length;
  *nz_OUT = z / length;
}

void GetVectorProduct 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  *x_OUT = y0 * z1 - z0 * y1;
  *y_OUT = z0 * x1 - x0 * z1;
  *z_OUT = x0 * y1 - y0 * x1;
}

void GetPlaneEquation 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double x2, double y2, double z2,
 double *a_OUT, double *b_OUT, double *c_OUT, double *d_OUT)
{
  double nx, ny, nz;

  GetVectorProduct (x1 - x0, y1 - y0, z1 - z0,
		    x2 - x0, y2 - y0, z2 - z0,
		    &nx, &ny, &nz);
  GetNormalizedVector (nx, ny, nz,
		       &nx, &ny, &nz);
  *a_OUT = nx;
  *b_OUT = ny;
  *c_OUT = nz;
  *d_OUT = -(nx * x0 + ny * y0 + nz * z0);
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 2500000

int NNodes;
double Node_coord[MAX_NODES][3];



/* Element */

#define N_ELEMENT_NODES 4
#define N_ELEMENT_EDGES 6
#define N_ELEMENT_FACES 4
#if 0
#define MAX_ELEMENTS 12000000
#else
#define MAX_ELEMENTS 3000000
#endif

const int EdgeNodeMap[N_ELEMENT_EDGES][2] = {
  { 0, 1 }, 
  { 0, 2 }, 
  { 0, 3 }, 
  { 1, 2 }, 
  { 2, 3 }, 
  { 3, 1 }  
};

const int FaceNodeMap[N_ELEMENT_FACES][3] = {
  { 1, 2, 3 },
  { 0, 3, 2 },
  { 0, 1, 3 },
  { 0, 2, 1 }
};

const int FaceOppositeNodeMap[N_ELEMENT_FACES] = {
  0, 1, 2, 3
};

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];

#if 0
#define MALLOC
#endif

#ifdef MALLOC
float *Element_a;
float *Element_b;
float *Element_c;
float *Element_d;
#else
double Element_a[MAX_ELEMENTS][N_ELEMENT_FACES];
double Element_b[MAX_ELEMENTS][N_ELEMENT_FACES];
double Element_c[MAX_ELEMENTS][N_ELEMENT_FACES];
double Element_d[MAX_ELEMENTS][N_ELEMENT_FACES];
#endif

AutoRes_Bucket3D ElementBucket;



void Element_getRange
(int elementId, 
 double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int nodeId0;
  int iElementNode;

  assert(0 <= elementId);
  assert(elementId < NElements);

  nodeId0 = Element_nodeIds[elementId][0];
  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[nodeId0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[nodeId0][1];
  *minRangeZ_OUT = *maxRangeZ_OUT = Node_coord[nodeId0][2];
  for (iElementNode = 0; iElementNode < N_ELEMENT_NODES; iElementNode++) {
    int nodeId = Element_nodeIds[elementId][iElementNode];
    double x = Node_coord[nodeId][0];
    double y = Node_coord[nodeId][1];
    double z = Node_coord[nodeId][2];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
    if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
  }
}

double Element_averageEdgeLength
(int elementId)
{
  int iElementEdge;
  double average;

  assert(0 <= elementId);
  assert(elementId < NElements);

  average = 0.0;
  for (iElementEdge = 0; 
       iElementEdge < N_ELEMENT_EDGES; 
       iElementEdge++) {
    int edgeNodeId0 
      = Element_nodeIds[elementId][EdgeNodeMap[iElementEdge][0]];
    int edgeNodeId1 
      = Element_nodeIds[elementId][EdgeNodeMap[iElementEdge][1]];
    
    double x0 = Node_coord[edgeNodeId0][0];
    double y0 = Node_coord[edgeNodeId0][1];
    double z0 = Node_coord[edgeNodeId0][2];
    
    double x1 = Node_coord[edgeNodeId1][0];
    double y1 = Node_coord[edgeNodeId1][1];
    double z1 = Node_coord[edgeNodeId1][2];
    
    average += Distance (x0, y0, z0, 
			 x1, y1, z1);
  }
  average /= N_ELEMENT_EDGES;
  return average;
}
 
int Element_containsPoint 
(int elementId, 
 double x, double y, double z,
 double *L0_OUT, double *L1_OUT, double *L2_OUT, double *L3_OUT)
{
  assert(0 <= elementId);
  assert(elementId < NElements);

#ifdef MALLOC
  *L0_OUT 
    = Element_a[elementId * N_ELEMENT_FACES + 0] * x
    + Element_b[elementId * N_ELEMENT_FACES + 0] * y
    + Element_c[elementId * N_ELEMENT_FACES + 0] * z
    + Element_d[elementId * N_ELEMENT_FACES + 0];
  *L1_OUT
    = Element_a[elementId * N_ELEMENT_FACES + 1] * x
    + Element_b[elementId * N_ELEMENT_FACES + 1] * y
    + Element_c[elementId * N_ELEMENT_FACES + 1] * z
    + Element_d[elementId * N_ELEMENT_FACES + 1];
  *L2_OUT
    = Element_a[elementId * N_ELEMENT_FACES + 2] * x
    + Element_b[elementId * N_ELEMENT_FACES + 2] * y
    + Element_c[elementId * N_ELEMENT_FACES + 2] * z
    + Element_d[elementId * N_ELEMENT_FACES + 2];
  *L3_OUT
    = Element_a[elementId * N_ELEMENT_FACES + 3] * x
    + Element_b[elementId * N_ELEMENT_FACES + 3] * y
    + Element_c[elementId * N_ELEMENT_FACES + 3] * z
    + Element_d[elementId * N_ELEMENT_FACES + 3];
#else
  *L0_OUT 
    = Element_a[elementId][0] * x
    + Element_b[elementId][0] * y
    + Element_c[elementId][0] * z
    + Element_d[elementId][0];
  *L1_OUT
    = Element_a[elementId][1] * x
    + Element_b[elementId][1] * y
    + Element_c[elementId][1] * z
    + Element_d[elementId][1];
  *L2_OUT
    = Element_a[elementId][2] * x
    + Element_b[elementId][2] * y
    + Element_c[elementId][2] * z
    + Element_d[elementId][2];
  *L3_OUT
    = Element_a[elementId][3] * x
    + Element_b[elementId][3] * y
    + Element_c[elementId][3] * z
    + Element_d[elementId][3];
#endif

  if (0.0 - VOLUME_COORD_EPS < *L0_OUT 
      && *L0_OUT < 1.0 + VOLUME_COORD_EPS
      && 0.0 - VOLUME_COORD_EPS < *L1_OUT 
      && *L1_OUT < 1.0 + VOLUME_COORD_EPS
      && 0.0 - VOLUME_COORD_EPS < *L2_OUT 
      && *L2_OUT < 1.0 + VOLUME_COORD_EPS
      && 0.0 - VOLUME_COORD_EPS < *L3_OUT 
      && *L3_OUT < 1.0 + VOLUME_COORD_EPS) {
    return 1;
  }

  return 0;
}

void Mesh_prepareInOutCheck (void)
{
  int iElement;
  int iElementFace;

#ifdef MALLOC
  Element_a = (float*)malloc (sizeof (float) * NElements * N_ELEMENT_FACES);
  Element_b = (float*)malloc (sizeof (float) * NElements * N_ELEMENT_FACES);
  Element_c = (float*)malloc (sizeof (float) * NElements * N_ELEMENT_FACES);
  Element_d = (float*)malloc (sizeof (float) * NElements * N_ELEMENT_FACES);
  assert(Element_a != NULL);
  assert(Element_b != NULL);
  assert(Element_c != NULL);
  assert(Element_d != NULL);
#endif

  for (iElement = 0; iElement < NElements; iElement++) {
    for (iElementFace = 0;
	 iElementFace < N_ELEMENT_FACES;
	 iElementFace++) {
      int nodeId0 = Element_nodeIds
	[iElement][FaceNodeMap[iElementFace][0]];
      int nodeId1 = Element_nodeIds
	[iElement][FaceNodeMap[iElementFace][1]];
      int nodeId2 = Element_nodeIds
	[iElement][FaceNodeMap[iElementFace][2]];
      int oppositeNodeId = Element_nodeIds
	[iElement][FaceOppositeNodeMap[iElementFace]];
    
      double x0 = Node_coord[nodeId0][0];
      double y0 = Node_coord[nodeId0][1];
      double z0 = Node_coord[nodeId0][2];
      
      double x1 = Node_coord[nodeId1][0];
      double y1 = Node_coord[nodeId1][1];
      double z1 = Node_coord[nodeId1][2];
      
      double x2 = Node_coord[nodeId2][0];
      double y2 = Node_coord[nodeId2][1];
      double z2 = Node_coord[nodeId2][2];
      
      double oX = Node_coord[oppositeNodeId][0];
      double oY = Node_coord[oppositeNodeId][1];
      double oZ = Node_coord[oppositeNodeId][2];
      
      double a, b, c, d;
      double height;
      
      GetPlaneEquation 
	(x0, y0, z0,
	 x1, y1, z1,
	 x2, y2, z2,
	 &a, &b, &c, &d);
      height = a * oX + b * oY + c * oZ + d;

#ifdef MALLOC
      Element_a[iElement * N_ELEMENT_FACES + iElementFace] = a / height;
      Element_b[iElement * N_ELEMENT_FACES + iElementFace] = b / height;
      Element_c[iElement * N_ELEMENT_FACES + iElementFace] = c / height;
      Element_d[iElement * N_ELEMENT_FACES + iElementFace] = d / height;
#else
      Element_a[iElement][iElementFace] = a / height;
      Element_b[iElement][iElementFace] = b / height;
      Element_c[iElement][iElementFace] = c / height;
      Element_d[iElement][iElementFace] = d / height;
#endif
    }
  }
}

void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  *minRangeZ_OUT = *maxRangeZ_OUT = Node_coord[0][2];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    double z = Node_coord[iNode][2];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
    if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
  }
}

void Mesh_createElementBucket (void)
{
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  double elementSize;
  int iElement;
  int xDivisions, yDivisions, zDivisions;
  int nOffsetDivisions = 23;
  double offsetX, offsetY, offsetZ;
#if 1
  int nAverageLinksPerCell = 50;
#else
  int nAverageLinksPerCell = 100;
#endif

  assert(0 < NNodes);

  Mesh_getRange 
    (&minRangeX, &minRangeY, &minRangeZ,
     &maxRangeX, &maxRangeY, &maxRangeZ);

  elementSize = 0.0;
  for (iElement = 0; iElement < NElements; iElement++) {
    elementSize += Element_averageEdgeLength (iElement);
  }
  elementSize /= NElements;

#if 1
  elementSize *= 0.7;
#else
  elementSize *= 0.7 * 2;
#endif

  xDivisions = (maxRangeX - minRangeX) / elementSize;
  if (xDivisions < 1) xDivisions = 1;
  yDivisions = (maxRangeY - minRangeY) / elementSize;
  if (yDivisions < 1) yDivisions = 1;
  zDivisions = (maxRangeZ - minRangeZ) / elementSize;
  if (zDivisions < 1) zDivisions = 1;

  offsetX = (maxRangeX - minRangeX) / nOffsetDivisions;
  offsetY = (maxRangeY - minRangeY) / nOffsetDivisions;
  offsetZ = (maxRangeZ - minRangeZ) / nOffsetDivisions;

  AutoRes_Bucket3D_initialize 
    (&ElementBucket,
     minRangeX - offsetX, minRangeY - offsetY, minRangeZ - offsetZ,
     maxRangeX + offsetX, maxRangeY + offsetY, maxRangeZ + offsetZ,
     xDivisions, yDivisions, zDivisions, 
     nAverageLinksPerCell);
}

void Mesh_addElementsIntoBucket (void)
{
  int iElement;

  for (iElement = 0; iElement < NElements; iElement++) {
    double minRangeX, minRangeY, minRangeZ;
    double maxRangeX, maxRangeY, maxRangeZ;
    
    Element_getRange (iElement,
		      &minRangeX, &minRangeY, &minRangeZ,
		      &maxRangeX, &maxRangeY, &maxRangeZ);
    
    AutoRes_Bucket3D_addItemWithin 
      (&ElementBucket,
       minRangeX, minRangeY, minRangeZ,
       maxRangeX, maxRangeY, maxRangeZ,
       iElement);
  }
}

int Mesh_findElementAt 
(double x, double y, double z,
 double *L0_OUT, double *L1_OUT, double *L2_OUT, double *L3_OUT)
{
  int cx, cy, cz;

  AutoRes_Bucket3D_getIndexAt 
    (&ElementBucket, 
     x, y, z,
     &cx, &cy, &cz);
  
  if (AutoRes_Bucket3D_isOutOfRange 
      (&ElementBucket, 
       cx, cy, cz)) {
    return -1;
  } else {
    int linkIndex;

    for (linkIndex = AutoRes_Bucket3D_startLinkIndexIn 
	   (&ElementBucket, 
	    cx, cy, cz);
	 linkIndex != -1;
	 linkIndex = AutoRes_Bucket3D_nextLinkIndex 
	   (&ElementBucket, 
	    linkIndex)) {
      int elementId = AutoRes_Bucket3D_itemId 
	(&ElementBucket, 
	 linkIndex);
      
      if (Element_containsPoint (elementId, x, y, z,
				 L0_OUT, L1_OUT, L2_OUT, L3_OUT)) {
	return elementId;
      }
    }
  }
  
  return -1;
}

void Mesh_deleteElementBucket (void)
{
  AutoRes_Bucket3D_terminate 
    (&ElementBucket);
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Sample */

#define MAX_SAMPLES 1000000

int NSamples;
double Sample_coord[MAX_SAMPLES][3];
int Sample_elementId[MAX_SAMPLES];
double Sample_L0[MAX_SAMPLES];
double Sample_L1[MAX_SAMPLES];
double Sample_L2[MAX_SAMPLES];
double Sample_L3[MAX_SAMPLES];



void ReadPointFile (const char *fileName)
{
  FILE *fp;
  int iSample;
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NSamples);
  assert(0 < NSamples);
  assert(NSamples <= MAX_SAMPLES);

  for (iSample = 0; iSample < NSamples; iSample++) {
    int sampleId;
    double x, y, z;
    
    fscanf (fp, "%d %lf %lf %lf", 
	    &sampleId,
	    &x, &y, &z);
    assert(sampleId == iSample);
    Sample_coord[iSample][0] = x;
    Sample_coord[iSample][1] = y;
    Sample_coord[iSample][2] = z;
  }

  fclose (fp);
}

void WriteSampleFile (const char *fileName)
{
  FILE *fp;
  int iSample;
  
  assert(0 < NSamples);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NSamples);
  
  for (iSample = 0; iSample < NSamples; iSample++) {
    fprintf (fp, "%d   %f %f %f   %d   %f %f %f %f\n", 
	     iSample,
	     Sample_coord[iSample][0],
	     Sample_coord[iSample][1],
	     Sample_coord[iSample][2],
	     Sample_elementId[iSample],
	     Sample_L0[iSample],
	     Sample_L1[iSample],
	     Sample_L2[iSample],
	     Sample_L3[iSample]);
  }

  fclose (fp);
}



/************************
	 Command
*************************/

#if 1
#include "../../util/time_measure.c"
#define TIME_MEASURE
#endif

void PerformCommand (void)
{
  int iSample;

#ifdef TIME_MEASURE
  StartMeasureTime ();
#endif

  Mesh_createElementBucket ();
  Mesh_addElementsIntoBucket ();
  Mesh_prepareInOutCheck ();

#ifdef TIME_MEASURE
  EndMeasureTime ();
#endif

#ifdef TIME_MEASURE
  StartMeasureTime ();
#endif

  for (iSample = 0; iSample < NSamples; iSample++) {
    double x = Sample_coord[iSample][0];
    double y = Sample_coord[iSample][1];
    double z = Sample_coord[iSample][2];
    int elementId;
    double L0, L1, L2, L3;
    
    elementId = Mesh_findElementAt (x, y, z,
				    &L0, &L1, &L2, &L3);

    if (elementId != -1) {
      Sample_elementId[iSample] = elementId;
      Sample_L0[iSample] = L0;
      Sample_L1[iSample] = L1;
      Sample_L2[iSample] = L2;
      Sample_L3[iSample] = L3;
    } else {
      Sample_elementId[iSample] = -1;
      Sample_L0[iSample] = 0.0;
      Sample_L1[iSample] = 0.0;
      Sample_L2[iSample] = 0.0;
      Sample_L3[iSample] = 0.0;
    }
  }

#ifdef TIME_MEASURE
  EndMeasureTime ();
#endif

  Mesh_deleteElementBucket ();
}



int main (int argc, char *argv[])
{
  if (argc != 5) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_tetra3d_sample3  node_file  element_file \n"
       "        point_file \n"
       "        sample_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadPointFile (argv[3]);

  PerformCommand ();

  WriteSampleFile (argv[4]);

  return 0;
}
