/* advautomesh_tetra3d_section_grid_view.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>
#include <autogl.h>



#if 1
#include "../../util/time_measure.c"
#define TIME_MEASURE
#endif



/************************
	 Basic
*************************/

#define VOLUME_COORD_EPS 0.00001



double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}

void GetNormalizedVector 
(double x, double y, double z,
 double *nx_OUT, double *ny_OUT, double *nz_OUT)
{
  double length = sqrt (x * x + y * y + z * z);

  *nx_OUT = x / length;
  *ny_OUT = y / length;
  *nz_OUT = z / length;
}

void GetVectorProduct 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  *x_OUT = y0 * z1 - z0 * y1;
  *y_OUT = z0 * x1 - x0 * z1;
  *z_OUT = x0 * y1 - y0 * x1;
}

void GetAnyPerpendicularDirection
(double x, double y, double z,
 double *nx0_OUT, double *ny0_OUT, double *nz0_OUT,
 double *nx1_OUT, double *ny1_OUT, double *nz1_OUT)
{
  double length;

  GetVectorProduct (x, y, z,
		    1, 0, 0,
		    nx0_OUT, ny0_OUT, nz0_OUT);
  length = sqrt (*nx0_OUT * *nx0_OUT
		 + *ny0_OUT * *ny0_OUT
		 + *nz0_OUT * *nz0_OUT);
  if (length == 0.0) {
    GetVectorProduct (x, y, z,
		      0, 1, 0,
		      nx0_OUT, ny0_OUT, nz0_OUT);
  }
  GetNormalizedVector (*nx0_OUT, *ny0_OUT, *nz0_OUT,
		       nx0_OUT, ny0_OUT, nz0_OUT);
  GetVectorProduct (x, y, z,
		    *nx0_OUT, *ny0_OUT, *nz0_OUT,
		    nx1_OUT, ny1_OUT, nz1_OUT);
  GetNormalizedVector (*nx1_OUT, *ny1_OUT, *nz1_OUT,
		       nx1_OUT, ny1_OUT, nz1_OUT);
}

void GetProjectedPointOnPlane
(double a, double b, double c, double d, 
 double x, double y, double z,
 double *px_OUT, double *py_OUT, double *pz_OUT)
{
  double value = a * x + b * y + c * z + d;
  
  *px_OUT = x - a * value;
  *py_OUT = y - b * value;
  *pz_OUT = z - c * value;
}

double SignedTetraVolume
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double x2, double y2, double z2,
 double x3, double y3, double z3)
{
  double x10 = x1 - x0;
  double y10 = y1 - y0;
  double z10 = z1 - z0;

  double x20 = x2 - x0;
  double y20 = y2 - y0;
  double z20 = z2 - z0;
  
  double x30 = x3 - x0;
  double y30 = y3 - y0;
  double z30 = z3 - z0;
  
  double d23 = y20 * z30 - z20 * y30;
  double d31 = y30 * z10 - z30 * y10;
  double d12 = y10 * z20 - z10 * y20;
  
  return (x10 * d23 + x20 * d31 + x30 * d12) / 6;
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 2500000

int NNodes;
double Node_coord[MAX_NODES][3];
double Node_vector[MAX_NODES][3];



/* Element */

#define N_ELEMENT_NODES 4
#define N_ELEMENT_EDGES 6
#define MAX_ELEMENTS 12000000

const int EdgeNodeMap[N_ELEMENT_EDGES][2] = {
  { 0, 1 }, 
  { 0, 2 }, 
  { 0, 3 }, 
  { 1, 2 }, 
  { 2, 3 }, 
  { 3, 1 }  
};

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];

AutoRes_Bucket3D ElementBucket;



void Element_getRange
(int elementId, 
 double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int nodeId0;
  int iElementNode;

  assert(0 <= elementId);
  assert(elementId < NElements);

  nodeId0 = Element_nodeIds[elementId][0];
  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[nodeId0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[nodeId0][1];
  *minRangeZ_OUT = *maxRangeZ_OUT = Node_coord[nodeId0][2];
  for (iElementNode = 0; iElementNode < N_ELEMENT_NODES; iElementNode++) {
    int nodeId = Element_nodeIds[elementId][iElementNode];
    double x = Node_coord[nodeId][0];
    double y = Node_coord[nodeId][1];
    double z = Node_coord[nodeId][2];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
    if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
  }
}

double Element_averageEdgeLength
(int elementId)
{
  int iElementEdge;
  double average;

  assert(0 <= elementId);
  assert(elementId < NElements);

  average = 0.0;
  for (iElementEdge = 0; 
       iElementEdge < N_ELEMENT_EDGES; 
       iElementEdge++) {
    int edgeNodeId0 
      = Element_nodeIds[elementId][EdgeNodeMap[iElementEdge][0]];
    int edgeNodeId1 
      = Element_nodeIds[elementId][EdgeNodeMap[iElementEdge][1]];
    
    double x0 = Node_coord[edgeNodeId0][0];
    double y0 = Node_coord[edgeNodeId0][1];
    double z0 = Node_coord[edgeNodeId0][2];
    
    double x1 = Node_coord[edgeNodeId1][0];
    double y1 = Node_coord[edgeNodeId1][1];
    double z1 = Node_coord[edgeNodeId1][2];
    
    average += Distance (x0, y0, z0, 
			 x1, y1, z1);
  }
  average /= N_ELEMENT_EDGES;
  return average;
}
 
int Element_containsPoint 
(int elementId, 
 double x, double y, double z,
 double *L0_OUT, double *L1_OUT, double *L2_OUT, double *L3_OUT)
{
  int nodeId0, nodeId1, nodeId2, nodeId3;
  double x0, y0, z0;
  double x1, y1, z1;
  double x2, y2, z2;
  double x3, y3, z3;
  double volume;
  double volume0, volume1, volume2, volume3;

  assert(0 <= elementId);
  assert(elementId < NElements);

  nodeId0 = Element_nodeIds[elementId][0];
  nodeId1 = Element_nodeIds[elementId][1];
  nodeId2 = Element_nodeIds[elementId][2];
  nodeId3 = Element_nodeIds[elementId][3];

  x0 = Node_coord[nodeId0][0];
  y0 = Node_coord[nodeId0][1];
  z0 = Node_coord[nodeId0][2];

  x1 = Node_coord[nodeId1][0];
  y1 = Node_coord[nodeId1][1];
  z1 = Node_coord[nodeId1][2];

  x2 = Node_coord[nodeId2][0];
  y2 = Node_coord[nodeId2][1];
  z2 = Node_coord[nodeId2][2];
  
  x3 = Node_coord[nodeId3][0];
  y3 = Node_coord[nodeId3][1];
  z3 = Node_coord[nodeId3][2];
  
  volume = SignedTetraVolume (x0, y0, z0, 
			      x1, y1, z1, 
			      x2, y2, z2, 
			      x3, y3, z3);
  volume0 = SignedTetraVolume (x, y, z, 
			       x1, y1, z1, 
			       x2, y2, z2, 
			       x3, y3, z3);
  volume1 = SignedTetraVolume (x0, y0, z0, 
			       x, y, z, 
			       x2, y2, z2, 
			       x3, y3, z3);
  volume2 = SignedTetraVolume (x0, y0, z0, 
			       x1, y1, z1, 
			       x, y, z, 
			       x3, y3, z3);
  volume3 = SignedTetraVolume (x0, y0, z0, 
			       x1, y1, z1, 
			       x2, y2, z2, 
			       x, y, z);

  *L0_OUT = volume0 / volume;
  *L1_OUT = volume1 / volume;
  *L2_OUT = volume2 / volume;
  *L3_OUT = volume3 / volume;

  if (0.0 - VOLUME_COORD_EPS < *L0_OUT 
      && *L0_OUT < 1.0 + VOLUME_COORD_EPS
      && 0.0 - VOLUME_COORD_EPS < *L1_OUT 
      && *L1_OUT < 1.0 + VOLUME_COORD_EPS
      && 0.0 - VOLUME_COORD_EPS < *L2_OUT 
      && *L2_OUT < 1.0 + VOLUME_COORD_EPS
      && 0.0 - VOLUME_COORD_EPS < *L3_OUT 
      && *L3_OUT < 1.0 + VOLUME_COORD_EPS) {
    return 1;
  }
  return 0;
}

void Element_getVector
(int elementId,
 double L0, double L1, double L2, double L3,
 double *vectorX_OUT, double *vectorY_OUT, double *vectorZ_OUT)
{
  int nodeId0, nodeId1, nodeId2, nodeId3;

  assert(0 <= elementId);
  assert(elementId < NElements);

  nodeId0 = Element_nodeIds[elementId][0];
  nodeId1 = Element_nodeIds[elementId][1];
  nodeId2 = Element_nodeIds[elementId][2];
  nodeId3 = Element_nodeIds[elementId][3];
  
  *vectorX_OUT 
    = Node_vector[nodeId0][0] * L0
    + Node_vector[nodeId1][0] * L1
    + Node_vector[nodeId2][0] * L2
    + Node_vector[nodeId3][0] * L3;
  *vectorY_OUT
    = Node_vector[nodeId0][1] * L0
    + Node_vector[nodeId1][1] * L1
    + Node_vector[nodeId2][1] * L2
    + Node_vector[nodeId3][1] * L3;
  *vectorZ_OUT
    = Node_vector[nodeId0][2] * L0
    + Node_vector[nodeId1][2] * L1
    + Node_vector[nodeId2][2] * L2
    + Node_vector[nodeId3][2] * L3;
}

void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  *minRangeZ_OUT = *maxRangeZ_OUT = Node_coord[0][2];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    double z = Node_coord[iNode][2];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
    if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
  }
}

void Mesh_createElementBucket (void)
{
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  double elementSize;
  int iElement;
  int xDivisions, yDivisions, zDivisions;
  int nOffsetDivisions = 23;
  double offsetX, offsetY, offsetZ;
#if 1
  int nAverageLinksPerCell = 50;
#else
  int nAverageLinksPerCell = 100;
#endif

  assert(0 < NNodes);

  Mesh_getRange 
    (&minRangeX, &minRangeY, &minRangeZ,
     &maxRangeX, &maxRangeY, &maxRangeZ);

  elementSize = 0.0;
  for (iElement = 0; iElement < NElements; iElement++) {
    elementSize += Element_averageEdgeLength (iElement);
  }
  elementSize /= NElements;

#if 1
  elementSize *= 0.7;
#else
  elementSize *= 0.7 * 2;
#endif

  xDivisions = (maxRangeX - minRangeX) / elementSize;
  if (xDivisions < 1) xDivisions = 1;
  yDivisions = (maxRangeY - minRangeY) / elementSize;
  if (yDivisions < 1) yDivisions = 1;
  zDivisions = (maxRangeZ - minRangeZ) / elementSize;
  if (zDivisions < 1) zDivisions = 1;

  offsetX = (maxRangeX - minRangeX) / nOffsetDivisions;
  offsetY = (maxRangeY - minRangeY) / nOffsetDivisions;
  offsetZ = (maxRangeZ - minRangeZ) / nOffsetDivisions;

  AutoRes_Bucket3D_initialize 
    (&ElementBucket,
     minRangeX - offsetX, minRangeY - offsetY, minRangeZ - offsetZ,
     maxRangeX + offsetX, maxRangeY + offsetY, maxRangeZ + offsetZ,
     xDivisions, yDivisions, zDivisions, 
     nAverageLinksPerCell);
}

void Mesh_addElementsIntoBucket (void)
{
  int iElement;

  for (iElement = 0; iElement < NElements; iElement++) {
    double minRangeX, minRangeY, minRangeZ;
    double maxRangeX, maxRangeY, maxRangeZ;
    
    Element_getRange (iElement,
		      &minRangeX, &minRangeY, &minRangeZ,
		      &maxRangeX, &maxRangeY, &maxRangeZ);
    
    AutoRes_Bucket3D_addItemWithin 
      (&ElementBucket,
       minRangeX, minRangeY, minRangeZ,
       maxRangeX, maxRangeY, maxRangeZ,
       iElement);
  }
}

int Mesh_findElementAt 
(double x, double y, double z,
 double *L0_OUT, double *L1_OUT, double *L2_OUT, double *L3_OUT)
{
  int cx, cy, cz;

  AutoRes_Bucket3D_getIndexAt 
    (&ElementBucket, 
     x, y, z,
     &cx, &cy, &cz);
  
  if (AutoRes_Bucket3D_isOutOfRange 
      (&ElementBucket, 
       cx, cy, cz)) {
    return -1;
  } else {
    int linkIndex;

    for (linkIndex = AutoRes_Bucket3D_startLinkIndexIn 
	   (&ElementBucket, 
	    cx, cy, cz);
	 linkIndex != -1;
	 linkIndex = AutoRes_Bucket3D_nextLinkIndex 
	   (&ElementBucket, 
	    linkIndex)) {
      int elementId = AutoRes_Bucket3D_itemId 
	(&ElementBucket, 
	 linkIndex);
      
      if (Element_containsPoint (elementId, x, y, z,
				 L0_OUT, L1_OUT, L2_OUT, L3_OUT)) {
	return elementId;
      }
    }
  }
  
  return -1;
}

void Mesh_deleteElementBucket (void)
{
  AutoRes_Bucket3D_terminate 
    (&ElementBucket);
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void ReadNodeVectorFile (const char *fileName)
{
  FILE *fp;
  int count;
  int iNode;
  
  assert(0 < NNodes);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &count);
  assert(count == NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double vectorX, vectorY, vectorZ;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId,
	    &vectorX, &vectorY, &vectorZ);
    assert(nodeId == iNode);
    Node_vector[iNode][0] = vectorX;
    Node_vector[iNode][1] = vectorY;
    Node_vector[iNode][2] = vectorZ;
  }

  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Section Grid */

double PlaneA, PlaneB, PlaneC, PlaneD;
double Interval;



/* Sample */

#define MAX_SAMPLES 1000000

int NSamples;
double Sample_coord[MAX_SAMPLES][3];
int Sample_elementId[MAX_SAMPLES];
double Sample_L0[MAX_SAMPLES];
double Sample_L1[MAX_SAMPLES];
double Sample_L2[MAX_SAMPLES];
double Sample_L3[MAX_SAMPLES];



/* Wireframe */

#define MAX_WIREFRAMES 1000000

int NWireframes;
double Wireframe_coord0[MAX_WIREFRAMES][3];
double Wireframe_coord1[MAX_WIREFRAMES][3];



void SectionGrid_setPlane (double a, double b, double c, double d)
{
  double length = sqrt (a * a + b * b + c * c);
  
  PlaneA = a / length;
  PlaneB = b / length;
  PlaneC = c / length;
  PlaneD = d / length;
}

void SampleGroup_clear (void)
{
  NSamples = 0;
} 

void SampleGroup_addPoint 
(double x, double y, double z, 
 int elementId, 
 double L0, double L1, double L2, double L3)
{
  assert(NSamples < MAX_SAMPLES);
  assert(0 <= elementId);
  assert(elementId < NElements);
  
  Sample_coord[NSamples][0] = x;
  Sample_coord[NSamples][1] = y;
  Sample_coord[NSamples][2] = z;
  Sample_elementId[NSamples] = elementId;
  Sample_L0[NSamples] = L0;
  Sample_L1[NSamples] = L1;
  Sample_L2[NSamples] = L2;
  Sample_L3[NSamples] = L3;
  
  NSamples++;
}

void ReadWireframeFile (const char *fileName)
{
  FILE *fp;
  int iWireframe;
  
  assert(NWireframes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NWireframes);
  assert(0 <= NWireframes);
  assert(NWireframes <= MAX_WIREFRAMES);
  
  for (iWireframe = 0; iWireframe < NWireframes; iWireframe++) {
    int wireframeId;
    double x0, y0, z0;
    double x1, y1, z1;

    fscanf (fp, "%d %lf %lf %lf %lf %lf %lf", 
	    &wireframeId,
	    &x0, &y0, &z0, 
	    &x1, &y1, &z1);
    assert(wireframeId == iWireframe);
    Wireframe_coord0[iWireframe][0] = x0;
    Wireframe_coord0[iWireframe][1] = y0;
    Wireframe_coord0[iWireframe][2] = z0;

    Wireframe_coord1[iWireframe][0] = x1;
    Wireframe_coord1[iWireframe][1] = y1;
    Wireframe_coord1[iWireframe][2] = z1;
  }

  fclose (fp);
}



/************************
	 Command
*************************/

void PrepareCommand (void)
{
#ifdef TIME_MEASURE
  StartMeasureTime ();
#endif

  Mesh_createElementBucket ();
  Mesh_addElementsIntoBucket ();

#ifdef TIME_MEASURE
  EndMeasureTime ();
#endif
}

void MakeSectionGrid (void)
{
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  double centerX, centerY, centerZ;
  double size;
  double sectionCenterX, sectionCenterY, sectionCenterZ;
  double uAxisX, uAxisY, uAxisZ;
  double vAxisX, vAxisY, vAxisZ;
  int nDivisions;
  int iu, iv;

#ifdef TIME_MEASURE
  StartMeasureTime ();
#endif

  Mesh_getRange
    (&minRangeX, &minRangeY, &minRangeZ,
     &maxRangeX, &maxRangeY, &maxRangeZ);
  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;
  centerZ = (minRangeZ + maxRangeZ) / 2;
  size = Distance (minRangeX, minRangeY, minRangeZ,
		   maxRangeX, maxRangeY, maxRangeZ) / 2;
  GetProjectedPointOnPlane
    (PlaneA, PlaneB, PlaneC, PlaneD,
     centerX, centerY, centerZ, 
     &sectionCenterX, &sectionCenterY, &sectionCenterZ);
  GetAnyPerpendicularDirection
    (PlaneA, PlaneB, PlaneC,
     &uAxisX, &uAxisY, &uAxisZ,
     &vAxisX, &vAxisY, &vAxisZ);
  nDivisions = size / Interval;
  
  for (iu = -nDivisions; iu <= nDivisions; iu++) {
    for (iv = -nDivisions; iv <= nDivisions; iv++) {
      double x = sectionCenterX 
	+ uAxisX * Interval * iu + vAxisX * Interval * iv;
      double y = sectionCenterY 
	+ uAxisY * Interval * iu + vAxisY * Interval * iv;
      double z = sectionCenterZ 
	+ uAxisZ * Interval * iu + vAxisZ * Interval * iv;
      int elementId;
      double L0, L1, L2, L3;
      
      elementId = Mesh_findElementAt (x, y, z,
				      &L0, &L1, &L2, &L3);
      
      if (elementId != -1) {
	SampleGroup_addPoint (x, y, z, 
			      elementId, L0, L1, L2, L3);
      }
    }
  }

#ifdef TIME_MEASURE
  EndMeasureTime ();
#endif
}



/********************************
	 Visualization
********************************/

double A, B, C = 1.0, D;

double Offset = 0.0;
double Shift = 1.0;

double Scale = 1.0;

void SetView (void)
{
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  double centerX, centerY, centerZ;
  double size;
  
  Mesh_getRange 
    (&minRangeX, &minRangeY, &minRangeZ,
     &maxRangeX, &maxRangeY, &maxRangeZ);
  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;
  centerZ = (minRangeZ + maxRangeZ) / 2;
  size = Distance (minRangeX, minRangeY, minRangeZ,
		   maxRangeX, maxRangeY, maxRangeZ) / 2;

  AutoGL_SetViewSize (size);
  AutoGL_SetViewCenter (centerX, centerY, centerZ); 
}

void SetVectorScale (void)
{
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  double centerX, centerY, centerZ;
  double size;
  double maxLength;
  int iNode;

  assert(0 < NNodes);

  Mesh_getRange 
    (&minRangeX, &minRangeY, &minRangeZ,
     &maxRangeX, &maxRangeY, &maxRangeZ);
  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;
  centerZ = (minRangeZ + maxRangeZ) / 2;
  size = Distance (minRangeX, minRangeY, minRangeZ,
		   maxRangeX, maxRangeY, maxRangeZ) / 2;

  maxLength = 0.0;
  for (iNode = 0; iNode < NNodes; iNode++) {
    double vx = Node_vector[iNode][0];
    double vy = Node_vector[iNode][1];
    double vz = Node_vector[iNode][2];
    double length = sqrt (vx * vx + vy * vy + vz * vz);
    
    if (maxLength < length) maxLength = length;
  }

  Scale = size / maxLength / 5;
}

void DrawWireframe (void)
{
  int iWireframe;

  AutoGL_SetColor (1, 1, 1);
  for (iWireframe = 0; iWireframe < NWireframes; iWireframe++) {
    AutoGL_DrawLine (Wireframe_coord0[iWireframe][0],
		     Wireframe_coord0[iWireframe][1],
		     Wireframe_coord0[iWireframe][2],
		     Wireframe_coord1[iWireframe][0],
		     Wireframe_coord1[iWireframe][1],
		     Wireframe_coord1[iWireframe][2]);
  }
}

void RedrawView (void) 
{
  int iSample;

  DrawWireframe ();

  for (iSample = 0; iSample < NSamples; iSample++) {
    double x = Sample_coord[iSample][0];
    double y = Sample_coord[iSample][1];
    double z = Sample_coord[iSample][2];
    int elementId = Sample_elementId[iSample];
    double L0 = Sample_L0[iSample];
    double L1 = Sample_L1[iSample];
    double L2 = Sample_L2[iSample];
    double L3 = Sample_L3[iSample];
    double vx, vy, vz;

    assert(elementId != -1);

    Element_getVector (elementId, L0, L1, L2, L3,
		       &vx, &vy, &vz);
    
    AutoGL_SetColor (0, 0, 1);
    AutoGL_DrawCrossMark3D (x, y, z, 3);
    AutoGL_SetColor (1, 1, 0);
    AutoGL_DrawLine (x, y, z,
		     x + vx * Scale, 
		     y + vy * Scale, 
		     z + vz * Scale);
  }
}

void SetSectionPlane (void)
{
  Offset = 0.0;
  SampleGroup_clear ();
  SectionGrid_setPlane (A, B, C, D - Offset);
  MakeSectionGrid ();

  AutoGL_DrawView ();
}

void Increment (void) 
{
  Offset += Shift;
  SampleGroup_clear ();
  SectionGrid_setPlane (A, B, C, D - Offset);
  MakeSectionGrid ();

  AutoGL_DrawView ();
}

void Decrement (void) 
{
  Offset -= Shift;
  SampleGroup_clear ();
  SectionGrid_setPlane (A, B, C, D - Offset);
  MakeSectionGrid ();

  AutoGL_DrawView ();
}

void UpdateModel (void) 
{
  AutoGL_DrawView ();
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
  if (argc != 5) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_tetra3d_section_grid_view  node_file  element_file \n"
       "     node_vector_file  wireframe_file \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadNodeVectorFile (argv[3]);
  ReadWireframeFile (argv[4]);

  SetView ();
  SetVectorScale ();
  PrepareCommand ();
  
  AutoGL_SetViewRedrawCallback (RedrawView);

  AutoGL_SetDefaultCallbackInMode3D (NULL);
  AutoGL_EnableDragInMode3D ();



  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode3D ();

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");

  AutoGL_AddReal (&A, "A");
  AutoGL_AddReal (&B, "B");
  AutoGL_AddReal (&C, "C");
  AutoGL_AddReal (&D, "D");
  AutoGL_AddReal (&Interval, "Interval");
  AutoGL_AddCallback (SetSectionPlane, "SetSectionPlane");      

  AutoGL_AddReal (&Shift, "Shift");
  AutoGL_AddCallback (Increment, "Increment");      
  AutoGL_AddCallback (Decrement, "Decrement");      

  AutoGL_AddReal (&Scale, "Scale");
  AutoGL_AddCallback (UpdateModel, "UpdateModel");      
}
