/* advautomesh_tetra3d_voxel2.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

#define VOLUME_COORD_EPS 0.00001
#define GEOMETRY_RELATIVE_EPS 0.00001



double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}

double SignedTetraVolume
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double x2, double y2, double z2,
 double x3, double y3, double z3)
{
  double x10 = x1 - x0;
  double y10 = y1 - y0;
  double z10 = z1 - z0;

  double x20 = x2 - x0;
  double y20 = y2 - y0;
  double z20 = z2 - z0;
  
  double x30 = x3 - x0;
  double y30 = y3 - y0;
  double z30 = z3 - z0;
  
  double d23 = y20 * z30 - z20 * y30;
  double d31 = y30 * z10 - z30 * y10;
  double d12 = y10 * z20 - z10 * y20;
  
  return (x10 * d23 + x20 * d31 + x30 * d12) / 6;
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 2500000

int NNodes;
double Node_coord[MAX_NODES][3];
double Node_scalar[MAX_NODES];



/* Element */

#define N_ELEMENT_NODES 4
#define N_ELEMENT_EDGES 6
#define MAX_ELEMENTS 12000000

const int EdgeNodeMap[N_ELEMENT_EDGES][2] = {
  { 0, 1 }, 
  { 0, 2 }, 
  { 0, 3 }, 
  { 1, 2 }, 
  { 2, 3 }, 
  { 3, 1 }  
};

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



void Element_getRange
(int elementId, 
 double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int nodeId0;
  int iElementNode;

  assert(0 <= elementId);
  assert(elementId < NElements);

  nodeId0 = Element_nodeIds[elementId][0];
  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[nodeId0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[nodeId0][1];
  *minRangeZ_OUT = *maxRangeZ_OUT = Node_coord[nodeId0][2];
  for (iElementNode = 0; iElementNode < N_ELEMENT_NODES; iElementNode++) {
    int nodeId = Element_nodeIds[elementId][iElementNode];
    double x = Node_coord[nodeId][0];
    double y = Node_coord[nodeId][1];
    double z = Node_coord[nodeId][2];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
    if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
  }
}

double Element_averageEdgeLength
(int elementId)
{
  int iElementEdge;
  double average;

  assert(0 <= elementId);
  assert(elementId < NElements);

  average = 0.0;
  for (iElementEdge = 0; 
       iElementEdge < N_ELEMENT_EDGES; 
       iElementEdge++) {
    int edgeNodeId0 
      = Element_nodeIds[elementId][EdgeNodeMap[iElementEdge][0]];
    int edgeNodeId1 
      = Element_nodeIds[elementId][EdgeNodeMap[iElementEdge][1]];
    
    double x0 = Node_coord[edgeNodeId0][0];
    double y0 = Node_coord[edgeNodeId0][1];
    double z0 = Node_coord[edgeNodeId0][2];
    
    double x1 = Node_coord[edgeNodeId1][0];
    double y1 = Node_coord[edgeNodeId1][1];
    double z1 = Node_coord[edgeNodeId1][2];
    
    average += Distance (x0, y0, z0, 
			 x1, y1, z1);
  }
  average /= N_ELEMENT_EDGES;
  return average;
}
 
double Element_scalar
(int elementId,
 double L0, double L1, double L2, double L3)
{
  int nodeId0, nodeId1, nodeId2, nodeId3;

  assert(0 <= elementId);
  assert(elementId < NElements);

  nodeId0 = Element_nodeIds[elementId][0];
  nodeId1 = Element_nodeIds[elementId][1];
  nodeId2 = Element_nodeIds[elementId][2];
  nodeId3 = Element_nodeIds[elementId][3];
  
  return Node_scalar[nodeId0] * L0
    + Node_scalar[nodeId1] * L1
    + Node_scalar[nodeId2] * L2
    + Node_scalar[nodeId3] * L3;
}

int Element_containsPoint 
(int elementId, 
 double x, double y, double z,
 double *L0_OUT, double *L1_OUT, double *L2_OUT, double *L3_OUT)
{
  int nodeId0, nodeId1, nodeId2, nodeId3;
  double x0, y0, z0;
  double x1, y1, z1;
  double x2, y2, z2;
  double x3, y3, z3;
  double volume;
  double volume0, volume1, volume2, volume3;

  assert(0 <= elementId);
  assert(elementId < NElements);

  nodeId0 = Element_nodeIds[elementId][0];
  nodeId1 = Element_nodeIds[elementId][1];
  nodeId2 = Element_nodeIds[elementId][2];
  nodeId3 = Element_nodeIds[elementId][3];

  x0 = Node_coord[nodeId0][0];
  y0 = Node_coord[nodeId0][1];
  z0 = Node_coord[nodeId0][2];

  x1 = Node_coord[nodeId1][0];
  y1 = Node_coord[nodeId1][1];
  z1 = Node_coord[nodeId1][2];

  x2 = Node_coord[nodeId2][0];
  y2 = Node_coord[nodeId2][1];
  z2 = Node_coord[nodeId2][2];
  
  x3 = Node_coord[nodeId3][0];
  y3 = Node_coord[nodeId3][1];
  z3 = Node_coord[nodeId3][2];
  
  volume = SignedTetraVolume (x0, y0, z0, 
			      x1, y1, z1, 
			      x2, y2, z2, 
			      x3, y3, z3);
  volume0 = SignedTetraVolume (x, y, z, 
			       x1, y1, z1, 
			       x2, y2, z2, 
			       x3, y3, z3);
  volume1 = SignedTetraVolume (x0, y0, z0, 
			       x, y, z, 
			       x2, y2, z2, 
			       x3, y3, z3);
  volume2 = SignedTetraVolume (x0, y0, z0, 
			       x1, y1, z1, 
			       x, y, z, 
			       x3, y3, z3);
  volume3 = SignedTetraVolume (x0, y0, z0, 
			       x1, y1, z1, 
			       x2, y2, z2, 
			       x, y, z);

  *L0_OUT = volume0 / volume;
  *L1_OUT = volume1 / volume;
  *L2_OUT = volume2 / volume;
  *L3_OUT = volume3 / volume;

  if (0.0 - VOLUME_COORD_EPS < *L0_OUT 
      && *L0_OUT < 1.0 + VOLUME_COORD_EPS
      && 0.0 - VOLUME_COORD_EPS < *L1_OUT 
      && *L1_OUT < 1.0 + VOLUME_COORD_EPS
      && 0.0 - VOLUME_COORD_EPS < *L2_OUT 
      && *L2_OUT < 1.0 + VOLUME_COORD_EPS
      && 0.0 - VOLUME_COORD_EPS < *L3_OUT 
      && *L3_OUT < 1.0 + VOLUME_COORD_EPS) {
    return 1;
  }
  return 0;
}

double Mesh_eps (void)
{
  int iElement;
  double averageEdgeLength;

  assert(0 < NElements);

  averageEdgeLength = 0.0;
  for (iElement = 0; iElement < NElements; iElement++) {
    averageEdgeLength += Element_averageEdgeLength (iElement);
  }
  averageEdgeLength /= NElements;

  return averageEdgeLength * GEOMETRY_RELATIVE_EPS;
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void ReadNodeScalarFile (const char *fileName)
{
  FILE *fp;
  int count;
  int iNode;
  
  assert(0 < NNodes);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &count);
  assert(count == NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double scalar;

    fscanf (fp, "%d %lf", 
	    &nodeId,
	    &scalar);
    assert(nodeId == iNode);
    Node_scalar[iNode] = scalar;
  }

  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Voxel */

#define MAX_GRIDS 512

double Voxel_minX, Voxel_minY, Voxel_minZ;
double Voxel_maxX, Voxel_maxY, Voxel_maxZ;
int Voxel_xGrids, Voxel_yGrids, Voxel_zGrids;
float Voxel_values[MAX_GRIDS][MAX_GRIDS][MAX_GRIDS];



void Voxel_setRange
(double minX, double minY, double minZ,
 double maxX, double maxY, double maxZ)
{
  assert(minX < maxX);
  assert(minY < maxY);
  assert(minZ < maxZ);

  Voxel_minX = minX;
  Voxel_minY = minY;
  Voxel_minZ = minZ;

  Voxel_maxX = maxX;
  Voxel_maxY = maxY;
  Voxel_maxZ = maxZ;
}

void Voxel_setGrid
(int xGrids, int yGrids, int zGrids)
{
  assert(1 < xGrids);
  assert(xGrids <= MAX_GRIDS);
  assert(1 < yGrids);
  assert(yGrids <= MAX_GRIDS);
  assert(1 < zGrids);
  assert(zGrids <= MAX_GRIDS);

  Voxel_xGrids = xGrids;
  Voxel_yGrids = yGrids;
  Voxel_zGrids = zGrids;
}

void WriteVoxelFile (const char *fileName)
{
  FILE *fp;
  int iGrid, jGrid, kGrid;
  
  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  fprintf (fp, "%f %f %f\n", 
	   Voxel_minX, Voxel_minY, Voxel_minZ);
  fprintf (fp, "%f %f %f\n", 
	   Voxel_maxX, Voxel_maxY, Voxel_maxZ);
  fprintf (fp, "%d %d %d\n", 
	   Voxel_xGrids, Voxel_yGrids, Voxel_zGrids);
  
  for (iGrid = 0; iGrid < Voxel_xGrids; iGrid++) {
    for (jGrid = 0; jGrid < Voxel_yGrids; jGrid++) {
      for (kGrid = 0; kGrid < Voxel_zGrids; kGrid++) {
	fprintf (fp, "%f\n", 
		 Voxel_values[iGrid][jGrid][kGrid]);
      }
    }
  }

  fclose (fp);
}



/************************
	 Command
*************************/

void PerformCommand (void)
{
  double eps = Mesh_eps ();
  double intervalX = (Voxel_maxX - Voxel_minX) / (Voxel_xGrids - 1);
  double intervalY = (Voxel_maxY - Voxel_minY) / (Voxel_yGrids - 1);
  double intervalZ = (Voxel_maxZ - Voxel_minZ) / (Voxel_zGrids - 1);
  double minRange, maxRange;
  int iNode;
  int iGrid, jGrid, kGrid;
  int iElement;

  assert(0 < NNodes);

  minRange = maxRange = Node_scalar[0];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double scalar = Node_scalar[iNode];
    
    if (scalar < minRange) minRange = scalar;
    if (maxRange < scalar) maxRange = scalar;
  }

  for (iGrid = 0; iGrid < Voxel_xGrids; iGrid++) {
    for (jGrid = 0; jGrid < Voxel_yGrids; jGrid++) {
      for (kGrid = 0; kGrid < Voxel_zGrids; kGrid++) {
	Voxel_values[iGrid][jGrid][kGrid] = 0.0;
      }
    }
  }

  for (iElement = 0; iElement < NElements; iElement++) {
    double minRangeX, minRangeY, minRangeZ;
    double maxRangeX, maxRangeY, maxRangeZ;
    double x, y, z;
    
    Element_getRange (iElement,
		      &minRangeX, &minRangeY, &minRangeZ,
		      &maxRangeX, &maxRangeY, &maxRangeZ);
    
    if (Voxel_minX + eps < minRangeX) {
      iGrid = (int)ceil ((minRangeX - eps - Voxel_minX) / intervalX);
    } else {
      iGrid = 0;
    }
    x = Voxel_minX + intervalX * iGrid;

    while (x < maxRangeX + eps) {

      if (Voxel_minY + eps < minRangeY) {
	jGrid = (int)ceil ((minRangeY - eps - Voxel_minY) / intervalY);
      } else {
	jGrid = 0;
      }
      y = Voxel_minY + intervalY * jGrid;

      while (y < maxRangeY + eps) {

	if (Voxel_minZ + eps < minRangeZ) {
	  kGrid = (int)ceil ((minRangeZ - eps - Voxel_minZ) / intervalZ);
	} else {
	  kGrid = 0;
	}
	z = Voxel_minZ + intervalZ * kGrid;

	while (z < maxRangeZ + eps) {
	  double L0, L1, L2, L3;
	  int flag = Element_containsPoint (iElement, x, y, z,
					    &L0, &L1, &L2, &L3);
	  if (flag) {
	    double scalar = Element_scalar (iElement, L0, L1, L2, L3);
	    
	    Voxel_values[iGrid][jGrid][kGrid] 
	      = (scalar - minRange) / (maxRange - minRange);
	  }

	  z += intervalZ;
	  kGrid++;
	}

	y += intervalY;
	jGrid++;
      }

      x += intervalX;
      iGrid++;
    }
  }
}



int main (int argc, char *argv[])
{
  double minX, minY, minZ;
  double maxX, maxY, maxZ;
  int xGrids, yGrids, zGrids;

  if (argc != 14) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_tetra3d_voxel2  node_file  element_file \n"
       "        node_scalar_file \n"
       "        minX minY minZ  maxX maxY maxZ  xGrids yGrids zGrids \n"
       "        voxel_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadNodeScalarFile (argv[3]);

  minX = atof (argv[4]);
  minY = atof (argv[5]);
  minZ = atof (argv[6]);

  maxX = atof (argv[7]);
  maxY = atof (argv[8]);
  maxZ = atof (argv[9]);

  Voxel_setRange (minX, minY, minZ,
		  maxX, maxY, maxZ);

  xGrids = atoi (argv[10]);
  yGrids = atoi (argv[11]);
  zGrids = atoi (argv[12]);

  Voxel_setGrid (xGrids, yGrids, zGrids);

  PerformCommand ();

  WriteVoxelFile (argv[13]);

  return 0;
}
