/* advautomesh_tri3d_edge2.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>



/************************
	 Mesh
*************************/

/* Node */

typedef enum {
  NODE_TOPOLOGY_INTERNAL,
  NODE_TOPOLOGY_EDGE,
  NODE_TOPOLOGY_VERTEX
} NodeTopologyType;

#define MAX_NODES 3000000

int NNodes;
NodeTopologyType Node_topologyType[MAX_NODES];



/* Element */

#define N_ELEMENT_NODES 3
#define N_ELEMENT_EDGES 3
#define MAX_ELEMENTS 5000000

const int EdgeNodeMap[N_ELEMENT_EDGES][2] = {
  { 1, 2 },
  { 2, 0 },
  { 0, 1 }
};

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];
int Element_edgeIds[MAX_ELEMENTS][N_ELEMENT_EDGES];
int Element_neighborIds[MAX_ELEMENTS][N_ELEMENT_EDGES];



/* Edge */

#define MAX_EDGES 7000000

int NEdges;
int Edge_nodeIds[MAX_EDGES][2];
int Edge_elementIds[MAX_EDGES][2];
int Edge_elementEdgeIds[MAX_EDGES][2];

AutoRes_ListTable EdgeTable;



void Edge_addElement 
(int edgeId, int elementId, int elementEdgeId)
{
  assert(0 <= edgeId);
  assert(edgeId < NEdges);
  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(0 <= elementEdgeId);
  assert(elementEdgeId < N_ELEMENT_EDGES);

  assert(Edge_elementIds[edgeId][1] == -1);
  if (Edge_elementIds[edgeId][0] == -1) {
    Edge_elementIds[edgeId][0] = elementId;
    Edge_elementEdgeIds[edgeId][0] = elementEdgeId;
  } else {
    Edge_elementIds[edgeId][1] = elementId;
    Edge_elementEdgeIds[edgeId][1] = elementEdgeId;
  }
}

int Mesh_addNewEdge (int nodeId0, int nodeId1)
{
  int edgeId = NEdges;
  int hashTableSize;
  int hashValue;

  assert(0 <= nodeId0);
  assert(nodeId0 < NNodes);
  assert(0 <= nodeId1);
  assert(nodeId1 < NNodes);

  assert(NEdges < MAX_EDGES);
  NEdges++;

  Edge_nodeIds[edgeId][0] = nodeId0;
  Edge_nodeIds[edgeId][1] = nodeId1;
  Edge_elementIds[edgeId][0] = -1;
  Edge_elementIds[edgeId][1] = -1;
  Edge_elementEdgeIds[edgeId][0] = -1;
  Edge_elementEdgeIds[edgeId][1] = -1;
  
  hashTableSize = AutoRes_ListTable_nLists 
    (&EdgeTable);
  hashValue = (nodeId0 + nodeId1) % hashTableSize;
  AutoRes_ListTable_addItemIn 
    (&EdgeTable,
     hashValue, edgeId);

  return edgeId;
}

int Mesh_findEdge (int nodeId0, int nodeId1)
{
  int hashTableSize;
  int hashValue;
  int linkIndex;

  assert(0 <= nodeId0);
  assert(nodeId0 < NNodes);
  assert(0 <= nodeId1);
  assert(nodeId1 < NNodes);

  hashTableSize = AutoRes_ListTable_nLists 
    (&EdgeTable);
  hashValue = (nodeId0 + nodeId1) % hashTableSize;
  for (linkIndex = AutoRes_ListTable_startLinkIndexIn 
	 (&EdgeTable, 
	  hashValue);
       linkIndex != -1;
       linkIndex = AutoRes_ListTable_nextLinkIndex 
	 (&EdgeTable, 
	  linkIndex)) {
    int edgeId = AutoRes_ListTable_itemId 
      (&EdgeTable, 
       linkIndex);
    int edgeNodeId0 = Edge_nodeIds[edgeId][0];
    int edgeNodeId1 = Edge_nodeIds[edgeId][1];
    
    if ((edgeNodeId0 == nodeId0 && edgeNodeId1 == nodeId1)
	|| (edgeNodeId1 == nodeId0 && edgeNodeId0 == nodeId1)) {
      return edgeId;
    }
  }

  return -1;
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  /* skip coordinate */

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void WriteEdgeFile (const char *fileName)
{
  FILE *fp;
  int iEdge;
  
  assert(0 < NEdges);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  fprintf (fp, "%d\n", 
	   NEdges);
  
  for (iEdge = 0; iEdge < NEdges; iEdge++) {
    fprintf (fp, "%d   %d %d   %d %d   %d %d\n", 
	     iEdge,
	     Edge_nodeIds[iEdge][0],
	     Edge_nodeIds[iEdge][1],
	     Edge_elementIds[iEdge][0],
	     Edge_elementEdgeIds[iEdge][0],
	     Edge_elementIds[iEdge][1],
	     Edge_elementEdgeIds[iEdge][1]);
  }

  fclose (fp);
}

void WriteNodeTopologyFile (const char *fileName)
{
  FILE *fp;
  int iNode;
  
  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    fprintf (fp, "%d   %d\n", 
	     iNode, Node_topologyType[iNode]);
  }

  fclose (fp);
}

void WriteElementEdgeFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementEdge;
  
  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  fprintf (fp, "%d\n", 
	   NElements);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    fprintf (fp, "%d   ", 
	     iElement);

    for (iElementEdge = 0; 
	 iElementEdge < N_ELEMENT_EDGES; 
	 iElementEdge++) {
      fprintf (fp, "%d ", 
	       Element_edgeIds[iElement][iElementEdge]);
    }

    fprintf (fp, "\n");
  }

  fclose (fp);
}

void WriteElementNeighborFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementEdge;
  
  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  fprintf (fp, "%d\n", 
	   NElements);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    fprintf (fp, "%d   ", 
	     iElement);

    for (iElementEdge = 0; 
	 iElementEdge < N_ELEMENT_EDGES; 
	 iElementEdge++) {
      fprintf (fp, "%d ", 
	       Element_neighborIds[iElement][iElementEdge]);
    }

    fprintf (fp, "\n");
  }

  fclose (fp);
}



/************************
	 Command
*************************/

void MakeEdge (void)
{
  int iElement;
  int iElementEdge;
  int iEdge;
  
  assert(NEdges == 0);

  for (iElement = 0; iElement < NElements; iElement++) {
    for (iElementEdge = 0; 
	 iElementEdge < N_ELEMENT_EDGES; 
	 iElementEdge++) {
      int edgeNodeId0 
	= Element_nodeIds[iElement][EdgeNodeMap[iElementEdge][0]];
      int edgeNodeId1 
	= Element_nodeIds[iElement][EdgeNodeMap[iElementEdge][1]];
      int edgeId = Mesh_findEdge (edgeNodeId0, edgeNodeId1);

      if (edgeId == -1) {
	edgeId = Mesh_addNewEdge (edgeNodeId0, edgeNodeId1);
      }
      Element_edgeIds[iElement][iElementEdge] = edgeId;
      Edge_addElement (edgeId, iElement, iElementEdge);
    }
  }

  for (iEdge = 0; iEdge < NEdges; iEdge++) {
    assert(Edge_elementIds[iEdge][0] != -1);
    assert(Edge_elementIds[iEdge][1] != -1
	   || Edge_elementIds[iEdge][1] == -1);
  }
}

void SetElementNeighbor (void)
{
  int iElement;
  int iElementEdge;
  
  assert(0 < NEdges);

  for (iElement = 0; iElement < NElements; iElement++) {
    for (iElementEdge = 0; 
	 iElementEdge < N_ELEMENT_EDGES; 
	 iElementEdge++) {
      int edgeId = Element_edgeIds[iElement][iElementEdge];
      int neighborId;
      
      if (Edge_elementIds[edgeId][0] == iElement) {
	neighborId = Edge_elementIds[edgeId][1];
      } else {
	assert(Edge_elementIds[edgeId][1] == iElement);
	neighborId = Edge_elementIds[edgeId][0];
      }
      Element_neighborIds[iElement][iElementEdge] = neighborId;
    }
  }
}

void SetNodeTopology (void)
{
  int iNode;
  int iEdge;
  
  assert(0 < NEdges);

  for (iNode = 0; iNode < NNodes; iNode++) {
    Node_topologyType[iNode] = NODE_TOPOLOGY_INTERNAL;
  }

  for (iEdge = 0; iEdge < NEdges; iEdge++) {
    if (Edge_elementIds[iEdge][1] == -1) {
      int nodeId0 = Edge_nodeIds[iEdge][0];
      int nodeId1 = Edge_nodeIds[iEdge][1];
      
      Node_topologyType[nodeId0] = NODE_TOPOLOGY_EDGE;
      Node_topologyType[nodeId1] = NODE_TOPOLOGY_EDGE;
    }
  }
}

void PerformCommand (void)
{
  int hashTableSize = NNodes * 10;
  int nAverageLinksPerList = 2;

  AutoRes_ListTable_initialize 
    (&EdgeTable,
     hashTableSize, nAverageLinksPerList);

  MakeEdge ();
  SetElementNeighbor ();
  SetNodeTopology ();

  AutoRes_ListTable_terminate 
    (&EdgeTable);
}



#if 1
#include "../../util/time_measure.c"
#define TIME_MEASURE
#endif

int main (int argc, char *argv[])
{
  if (argc != 7) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_tri3d_edge2  node_file  element_file \n"
       "        edge_file_OUT \n"
       "        node_topology_file_OUT \n"
       "        element_edge_file_OUT  element_neighbor_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);

#ifdef TIME_MEASURE
  StartMeasureTime ();
#endif

  PerformCommand ();

#ifdef TIME_MEASURE
  EndMeasureTime ();
#endif

  WriteEdgeFile (argv[3]);
  WriteNodeTopologyFile (argv[4]);
  WriteElementEdgeFile (argv[5]);
  WriteElementNeighborFile (argv[6]);

  return 0;
}
