/* advautomesh_tri3d_strip_view.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autogl.h>



/************************
	 Basic
*************************/

double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][3];



/* Element */

#define N_ELEMENT_NODES 3
#define MAX_ELEMENTS 2000000

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  *minRangeZ_OUT = *maxRangeZ_OUT = Node_coord[0][2];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    double z = Node_coord[iNode][2];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
    if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
  }
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Strip */

#define MAX_STRIPS 1000000
#define MAX_VERTEXS 2000000

int NStrips;
int Strip_startVertexId[MAX_STRIPS];
int Strip_nVertexs[MAX_STRIPS];
int NVertexs;
int Vertex_nodeId[MAX_VERTEXS];
int Vertex_elementId[MAX_VERTEXS];



void ReadStripFile (const char *fileName)
{
  FILE *fp;
  int iStrip;
  int iVertex;
  int count;
  
  assert(NStrips == 0);
  assert(NVertexs == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NStrips);
  assert(0 < NStrips);
  assert(NStrips <= MAX_STRIPS);
  
  count = 0;
  for (iStrip = 0; iStrip < NStrips; iStrip++) {
    int stripId;
    int nVertexs;

    fscanf (fp, "%d %d", 
	    &stripId, &nVertexs);
    assert(stripId == iStrip);
    assert(0 < nVertexs);
    Strip_startVertexId[iStrip] = count;
    Strip_nVertexs[iStrip] = nVertexs;
    count += nVertexs;
    assert(count <= MAX_VERTEXS);

    for (iVertex = Strip_startVertexId[iStrip]; 
	 iVertex < Strip_startVertexId[iStrip] + Strip_nVertexs[iStrip]; 
	 iVertex++) {
      int vertexId;
      int nodeId;
      int elementId;

      fscanf (fp, "%d %d %d", 
	      &vertexId,
	      &nodeId, 
	      &elementId);
      assert(vertexId == iVertex - Strip_startVertexId[iStrip]);

      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Vertex_nodeId[iVertex] = nodeId;

      assert(0 <= elementId || elementId == -1);
      assert(elementId < NElements);
      Vertex_elementId[iVertex] = elementId;
    }
  }
  NVertexs = count;

  fclose (fp);
}



/********************************
	 Visualization
********************************/

int Mesh = 0;
int NodeId = 0;
int StripId = 0;

void SetView (void)
{
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  double centerX, centerY, centerZ;
  double size;
  
  Mesh_getRange 
    (&minRangeX, &minRangeY, &minRangeZ,
     &maxRangeX, &maxRangeY, &maxRangeZ);
  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;
  centerZ = (minRangeZ + maxRangeZ) / 2;
  size = Distance (minRangeX, minRangeY, minRangeZ,
		   maxRangeX, maxRangeY, maxRangeZ) / 2;

  AutoGL_SetViewSize (size);
  AutoGL_SetViewCenter (centerX, centerY, centerZ); 
}

void RedrawView (void) 
{
  int iNode;
  int iElement;
  int iStrip;
  int iVertex;
  char buf[256];

  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1]; 
    double z = Node_coord[iNode][2]; 

    if (NodeId) {
      sprintf(buf, "%d", iNode);
      AutoGL_SetColor (1, 1, 0);
      AutoGL_DrawString (x, y, z,
			 buf);
    }
  }

  if (Mesh) {
    AutoGL_SetColor (1, 1, 1);

    for (iElement = 0; iElement < NElements; iElement++) {
      int nodeId0 = Element_nodeIds[iElement][0];
      int nodeId1 = Element_nodeIds[iElement][1];
      int nodeId2 = Element_nodeIds[iElement][2];
      
      double nodeX0 = Node_coord[nodeId0][0];
      double nodeY0 = Node_coord[nodeId0][1];
      double nodeZ0 = Node_coord[nodeId0][2];
      
      double nodeX1 = Node_coord[nodeId1][0];
      double nodeY1 = Node_coord[nodeId1][1];
      double nodeZ1 = Node_coord[nodeId1][2];
      
      double nodeX2 = Node_coord[nodeId2][0];
      double nodeY2 = Node_coord[nodeId2][1];
      double nodeZ2 = Node_coord[nodeId2][2];
      
      AutoGL_DrawTriangle (nodeX0, nodeY0, nodeZ0,
			   nodeX1, nodeY1, nodeZ1,
			   nodeX2, nodeY2, nodeZ2);
    }
  }

  for (iStrip = 0; iStrip < NStrips; iStrip++) {
    int startId = Strip_startVertexId[iStrip];
    int nodeId0, nodeId1, nodeId2;
    double center012X, center012Y, center012Z;

    assert(3 <= Strip_nVertexs[iStrip]);

    nodeId0 = Vertex_nodeId[startId];
    nodeId1 = Vertex_nodeId[startId + 1];
    nodeId2 = Vertex_nodeId[startId + 2];

    center012X = (Node_coord[nodeId0][0]
		  + Node_coord[nodeId1][0]
		  + Node_coord[nodeId2][0]) / 3;
    center012Y = (Node_coord[nodeId0][1]
		  + Node_coord[nodeId1][1]
		  + Node_coord[nodeId2][1]) / 3;
    center012Z = (Node_coord[nodeId0][2]
		  + Node_coord[nodeId1][2]
		  + Node_coord[nodeId2][2]) / 3;
    
    if (StripId) {
      sprintf(buf, "%d", iStrip);
      AutoGL_SetColor (0, 1, 1);
      AutoGL_DrawString (center012X, center012Y, center012Z,
			 buf);
    }
    
    AutoGL_SetColor (1, 0, 0);
    AutoGL_DrawCrossMark2D (center012X, center012Y, center012Z, 
			    5);      

    for (iVertex = Strip_startVertexId[iStrip] + 3; 
	 iVertex < Strip_startVertexId[iStrip] + Strip_nVertexs[iStrip]; 
	 iVertex++) {
      int nodeId3 = Vertex_nodeId[iVertex];
      double center123X, center123Y, center123Z;

      center012X = (Node_coord[nodeId0][0]
		    + Node_coord[nodeId1][0]
		    + Node_coord[nodeId2][0]) / 3;
      center012Y = (Node_coord[nodeId0][1]
		    + Node_coord[nodeId1][1]
		    + Node_coord[nodeId2][1]) / 3;
      center012Z = (Node_coord[nodeId0][2]
		    + Node_coord[nodeId1][2]
		    + Node_coord[nodeId2][2]) / 3;

      center123X = (Node_coord[nodeId1][0]
		    + Node_coord[nodeId2][0]
		    + Node_coord[nodeId3][0]) / 3;
      center123Y = (Node_coord[nodeId1][1]
		    + Node_coord[nodeId2][1]
		    + Node_coord[nodeId3][1]) / 3;
      center123Z = (Node_coord[nodeId1][2]
		    + Node_coord[nodeId2][2]
		    + Node_coord[nodeId3][2]) / 3;
      
      AutoGL_SetColor (1, 0, 0);
      AutoGL_DrawLine (center012X, center012Y, center012Z,
		       center123X, center123Y, center123Z);

      nodeId0 = nodeId1;
      nodeId1 = nodeId2;
      nodeId2 = nodeId3;
    }
  }
}

void UpdateModel (void) 
{
  AutoGL_DrawView ();
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_tri3d_strip_view  node_file  element_file \n"
       "     strip_file \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadStripFile (argv[3]);

  SetView ();
  
  AutoGL_SetViewRedrawCallback (RedrawView);

  AutoGL_SetDefaultCallbackInMode3D (NULL);
  AutoGL_EnableDragInMode3D ();



  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode3D ();

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");

  AutoGL_AddBoolean (&Mesh, "Mesh");
  AutoGL_AddBoolean (&NodeId, "NodeId");
  AutoGL_AddBoolean (&StripId, "StripId");

  AutoGL_AddCallback (UpdateModel, "UpdateModel");      
}
