/* advautomesh_tri3d_strip_view2.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autogl.h>

#include <GL/gl.h>
#include <GL/glu.h>



/************************
	 Basic
*************************/

double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}

void GetNormalizedVector 
(double x, double y, double z,
 double *nx_OUT, double *ny_OUT, double *nz_OUT)
{
  double length = sqrt (x * x + y * y + z * z);

  *nx_OUT = x / length;
  *ny_OUT = y / length;
  *nz_OUT = z / length;
}

void GetVectorProduct 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  *x_OUT = y0 * z1 - z0 * y1;
  *y_OUT = z0 * x1 - x0 * z1;
  *z_OUT = x0 * y1 - y0 * x1;
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][3];



/* Element */

#define N_ELEMENT_NODES 3
#define MAX_ELEMENTS 2000000

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



void Element_getNormal 
(int elementId,
 double *nx_OUT, double *ny_OUT, double *nz_OUT)
{
  int nodeId0 = Element_nodeIds[elementId][0];
  int nodeId1 = Element_nodeIds[elementId][1];
  int nodeId2 = Element_nodeIds[elementId][2];

  double x0 = Node_coord[nodeId0][0];
  double y0 = Node_coord[nodeId0][1];
  double z0 = Node_coord[nodeId0][2];

  double x1 = Node_coord[nodeId1][0];
  double y1 = Node_coord[nodeId1][1];
  double z1 = Node_coord[nodeId1][2];

  double x2 = Node_coord[nodeId2][0];
  double y2 = Node_coord[nodeId2][1];
  double z2 = Node_coord[nodeId2][2];

  GetVectorProduct (x0 - x1, y0 - y1, z0 - z1,
		    x0 - x2, y0 - y2, z0 - z2,
		    nx_OUT, ny_OUT, nz_OUT);
  GetNormalizedVector (*nx_OUT, *ny_OUT, *nz_OUT,
		       nx_OUT, ny_OUT, nz_OUT);
}

void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  *minRangeZ_OUT = *maxRangeZ_OUT = Node_coord[0][2];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    double z = Node_coord[iNode][2];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
    if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
  }
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Strip */

#define MAX_STRIPS 1000000
#define MAX_VERTEXS 2000000

int NStrips;
int Strip_startVertexId[MAX_STRIPS];
int Strip_nVertexs[MAX_STRIPS];
int NVertexs;
int Vertex_nodeId[MAX_VERTEXS];
int Vertex_elementId[MAX_VERTEXS];

int Strip_normalIsReverse[MAX_STRIPS];
float Vertex_normal[MAX_VERTEXS][3];
float Vertex_coord[MAX_VERTEXS][3];



void ReadStripFile (const char *fileName)
{
  FILE *fp;
  int iStrip;
  int iVertex;
  int count;
  
  assert(NStrips == 0);
  assert(NVertexs == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NStrips);
  assert(0 < NStrips);
  assert(NStrips <= MAX_STRIPS);
  
  count = 0;
  for (iStrip = 0; iStrip < NStrips; iStrip++) {
    int stripId;
    int nVertexs;

    fscanf (fp, "%d %d", 
	    &stripId, &nVertexs);
    assert(stripId == iStrip);
    assert(0 < nVertexs);
    Strip_startVertexId[iStrip] = count;
    Strip_nVertexs[iStrip] = nVertexs;
    count += nVertexs;
    assert(count <= MAX_VERTEXS);

    for (iVertex = Strip_startVertexId[iStrip]; 
	 iVertex < Strip_startVertexId[iStrip] + Strip_nVertexs[iStrip]; 
	 iVertex++) {
      int vertexId;
      int nodeId;
      int elementId;

      fscanf (fp, "%d %d %d", 
	      &vertexId,
	      &nodeId, 
	      &elementId);
      assert(vertexId == iVertex - Strip_startVertexId[iStrip]);

      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Vertex_nodeId[iVertex] = nodeId;

      assert(0 <= elementId || elementId == -1);
      assert(elementId < NElements);
      Vertex_elementId[iVertex] = elementId;
    }
  }
  NVertexs = count;

  fclose (fp);
}



/********************************
	 Visualization
********************************/

int Fill = 0;

void SetView (void)
{
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  double centerX, centerY, centerZ;
  double size;
  
  Mesh_getRange 
    (&minRangeX, &minRangeY, &minRangeZ,
     &maxRangeX, &maxRangeY, &maxRangeZ);
  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;
  centerZ = (minRangeZ + maxRangeZ) / 2;
  size = Distance (minRangeX, minRangeY, minRangeZ,
		   maxRangeX, maxRangeY, maxRangeZ) / 2;

  AutoGL_SetViewSize (size);
  AutoGL_SetViewCenter (centerX, centerY, centerZ); 
}

void PrepareStripVertex (void)
{
  int iVertex;
  
  for (iVertex = 0; iVertex < NVertexs; iVertex++) {
    int nodeId = Vertex_nodeId[iVertex];
    double x = Node_coord[nodeId][0];
    double y = Node_coord[nodeId][1];
    double z = Node_coord[nodeId][2];

    Vertex_coord[iVertex][0] = x;
    Vertex_coord[iVertex][1] = y;
    Vertex_coord[iVertex][2] = z;
  }
}

void PrepareStripNormal (void)
{
  int iStrip;
  int iVertex;

  for (iStrip = 0; iStrip < NStrips; iStrip++) {
    int startId = Strip_startVertexId[iStrip];
    int isReverse;
    int nodeId0, nodeId1, nodeId2;

    assert(3 <= Strip_nVertexs[iStrip]);

    nodeId0 = Vertex_nodeId[startId + 0];
    nodeId1 = Vertex_nodeId[startId + 1];
    nodeId2 = Vertex_nodeId[startId + 2];

    {
      double x0 = Node_coord[nodeId0][0];
      double y0 = Node_coord[nodeId0][1];
      double z0 = Node_coord[nodeId0][2];
      
      double x1 = Node_coord[nodeId1][0];
      double y1 = Node_coord[nodeId1][1];
      double z1 = Node_coord[nodeId1][2];
    
      double x2 = Node_coord[nodeId2][0];
      double y2 = Node_coord[nodeId2][1];
      double z2 = Node_coord[nodeId2][2];

      int firstElementId = Vertex_elementId[startId + 2];
      double enx, eny, enz;
      double nx, ny, nz;

      assert(firstElementId != -1);
      Element_getNormal (firstElementId,
			 &enx, &eny, &enz);

      GetVectorProduct (x0 - x1, y0 - y1, z0 - z1,
			x0 - x2, y0 - y2, z0 - z2,
			&nx, &ny, &nz);
      GetNormalizedVector (nx, ny, nz,
			   &nx, &ny, &nz);

      if (0.0 < nx * enx + ny * eny + nz * enz) {
	Strip_normalIsReverse[iStrip] = 0;
	isReverse = 0;
      } else {
	Strip_normalIsReverse[iStrip] = 1;
	isReverse = 1;
      }

      if (isReverse) {
	nx = -nx;
	ny = -ny;
	nz = -nz;
      }

      Vertex_normal[startId][0] = nx;
      Vertex_normal[startId][1] = ny;
      Vertex_normal[startId][2] = nz;

      Vertex_normal[startId + 1][0] = nx;
      Vertex_normal[startId + 1][1] = ny;
      Vertex_normal[startId + 1][2] = nz;

      Vertex_normal[startId + 2][0] = nx;
      Vertex_normal[startId + 2][1] = ny;
      Vertex_normal[startId + 2][2] = nz;
    }

    for (iVertex = Strip_startVertexId[iStrip] + 3; 
	 iVertex < Strip_startVertexId[iStrip] + Strip_nVertexs[iStrip]; 
	 iVertex++) {
      double x0, y0, z0;
      double x1, y1, z1;
      double x2, y2, z2;
      double nx, ny, nz;

      if (isReverse) {
	isReverse = 0;
      } else {
	isReverse = 1;
      }

      nodeId0 = nodeId1;
      nodeId1 = nodeId2;
      nodeId2 = Vertex_nodeId[iVertex];

      x0 = Node_coord[nodeId0][0];
      y0 = Node_coord[nodeId0][1];
      z0 = Node_coord[nodeId0][2];
      
      x1 = Node_coord[nodeId1][0];
      y1 = Node_coord[nodeId1][1];
      z1 = Node_coord[nodeId1][2];
    
      x2 = Node_coord[nodeId2][0];
      y2 = Node_coord[nodeId2][1];
      z2 = Node_coord[nodeId2][2];

      GetVectorProduct (x0 - x1, y0 - y1, z0 - z1,
			x0 - x2, y0 - y2, z0 - z2,
			&nx, &ny, &nz);
      GetNormalizedVector (nx, ny, nz,
			   &nx, &ny, &nz);

      if (isReverse) {
	nx = -nx;
	ny = -ny;
	nz = -nz;
      }

      Vertex_normal[iVertex][0] = nx;
      Vertex_normal[iVertex][1] = ny;
      Vertex_normal[iVertex][2] = nz;
    }
  }
}

void PrepareRendering (void)
{
  PrepareStripVertex ();
  PrepareStripNormal ();
}

#if 1
#include "../../util/time_measure.c"
#define TIME_MEASURE
#endif

void RedrawView (void) 
{
  int iStrip;
  int iVertex;

#ifdef TIME_MEASURE
  StartMeasureTime ();
#endif

  if (Fill) {
    AutoGL_BeginNativeCall ();

    glShadeModel (GL_FLAT);
    glEnable (GL_LIGHTING);
    glEnable (GL_CULL_FACE);

    glColor4f (0.0f, 1.0f, 1.0f, 1.0f);

    for (iStrip = 0; iStrip < NStrips; iStrip++) {
      int startId = Strip_startVertexId[iStrip];
      int normalIsReverse = Strip_normalIsReverse[iStrip];
      
      assert(3 <= Strip_nVertexs[iStrip]);

      if (normalIsReverse) {
	glFrontFace (GL_CW);
      } else {
	glFrontFace (GL_CCW);
      }

      glBegin (GL_TRIANGLE_STRIP);

      glNormal3f (Vertex_normal[startId][0],
		  Vertex_normal[startId][1],
		  Vertex_normal[startId][2]); 

      glVertex3f (Vertex_coord[startId][0], 
		  Vertex_coord[startId][1], 
		  Vertex_coord[startId][2]);
      glVertex3f (Vertex_coord[startId + 1][0], 
		  Vertex_coord[startId + 1][1], 
		  Vertex_coord[startId + 1][2]);
      glVertex3f (Vertex_coord[startId + 2][0], 
		  Vertex_coord[startId + 2][1], 
		  Vertex_coord[startId + 2][2]);

      for (iVertex = Strip_startVertexId[iStrip] + 3; 
	   iVertex < Strip_startVertexId[iStrip] + Strip_nVertexs[iStrip]; 
	   iVertex++) {
	glNormal3f (Vertex_normal[iVertex][0],
		    Vertex_normal[iVertex][1],
		    Vertex_normal[iVertex][2]); 
	glVertex3f (Vertex_coord[iVertex][0], 
		    Vertex_coord[iVertex][1], 
		    Vertex_coord[iVertex][2]);
      }

      glEnd ();
    }

    glShadeModel (GL_SMOOTH);
    glFrontFace (GL_CCW);
    glEnable (GL_CULL_FACE);
    glEnable (GL_LIGHTING);

    AutoGL_EndNativeCall ();
  }

#ifdef TIME_MEASURE
  EndMeasureTime ();
#endif
}

#if 0
#define USE_DISPLAY_LIST
#endif

#ifdef USE_DISPLAY_LIST
void RedrawView2 (void)
{
#ifdef TIME_MEASURE
  StartMeasureTime ();
#endif

  AutoGL_DrawDisplayList ();

#ifdef TIME_MEASURE
  EndMeasureTime ();
#endif
}
#endif

void UpdateModel (void) 
{
#ifdef USE_DISPLAY_LIST
  AutoGL_OpenDisplayList ();
  RedrawView ();
  AutoGL_CloseDisplayList ();
#endif
  AutoGL_DrawView ();
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_tri3d_strip_view2  node_file  element_file \n"
       "     strip_file \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadStripFile (argv[3]);

  SetView ();
  PrepareRendering ();
  
#ifdef USE_DISPLAY_LIST
  AutoGL_SetViewRedrawCallback (RedrawView2);
#else
  AutoGL_SetViewRedrawCallback (RedrawView);
#endif

  AutoGL_SetDefaultCallbackInMode3D (NULL);
  AutoGL_EnableDragInMode3D ();



  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode3D ();

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");

  AutoGL_AddBoolean (&Fill, "Fill");

  AutoGL_AddCallback (UpdateModel, "UpdateModel");      
}
