/* advautomesh_h_hexa3d_bucket_renumber.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>



/************************
	 Basic
*************************/

double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][3];
int Node_newId[MAX_NODES];
int Node_oldId[MAX_NODES];



/* Element */

#define N_ELEMENT_NODES 8
#define N_ELEMENT_EDGES 12
#define MAX_ELEMENTS 1000000

const int EdgeNodeMap[N_ELEMENT_EDGES][2] = {
  { 0, 1 }, 
  { 1, 2 }, 
  { 2, 3 }, 
  { 3, 0 }, 
  { 4, 5 }, 
  { 5, 6 },  
  { 6, 7 }, 
  { 7, 4 },
  { 0, 4 }, 
  { 1, 5 }, 
  { 2, 6 }, 
  { 3, 7 }
};

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];
int Element_newId[MAX_ELEMENTS];
int Element_oldId[MAX_ELEMENTS];

AutoRes_Bucket3D ElementBucket;



void Element_getRange
(int elementId, 
 double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int nodeId0;
  int iElementNode;

  assert(0 <= elementId);
  assert(elementId < NElements);

  nodeId0 = Element_nodeIds[elementId][0];
  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[nodeId0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[nodeId0][1];
  *minRangeZ_OUT = *maxRangeZ_OUT = Node_coord[nodeId0][2];
  for (iElementNode = 0; iElementNode < N_ELEMENT_NODES; iElementNode++) {
    int nodeId = Element_nodeIds[elementId][iElementNode];
    double x = Node_coord[nodeId][0];
    double y = Node_coord[nodeId][1];
    double z = Node_coord[nodeId][2];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
    if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
  }
}

double Element_averageEdgeLength
(int elementId)
{
  int iElementEdge;
  double average;

  assert(0 <= elementId);
  assert(elementId < NElements);

  average = 0.0;
  for (iElementEdge = 0; 
       iElementEdge < N_ELEMENT_EDGES; 
       iElementEdge++) {
    int edgeNodeId0 
      = Element_nodeIds[elementId][EdgeNodeMap[iElementEdge][0]];
    int edgeNodeId1 
      = Element_nodeIds[elementId][EdgeNodeMap[iElementEdge][1]];
    
    double x0 = Node_coord[edgeNodeId0][0];
    double y0 = Node_coord[edgeNodeId0][1];
    double z0 = Node_coord[edgeNodeId0][2];
    
    double x1 = Node_coord[edgeNodeId1][0];
    double y1 = Node_coord[edgeNodeId1][1];
    double z1 = Node_coord[edgeNodeId1][2];
    
    average += Distance (x0, y0, z0, 
			 x1, y1, z1);
  }
  average /= N_ELEMENT_EDGES;
  return average;
}
 
void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  *minRangeZ_OUT = *maxRangeZ_OUT = Node_coord[0][2];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    double z = Node_coord[iNode][2];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
    if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
  }
}

void Mesh_createElementBucket (void)
{
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  double elementSize;
  int iElement;
  int xDivisions, yDivisions, zDivisions;
  int nOffsetDivisions = 23;
  double offsetX, offsetY, offsetZ;
#if 1
  int nAverageLinksPerCell = 20;
#else
  int nAverageLinksPerCell = 30;
#endif

  assert(0 < NNodes);

  Mesh_getRange 
    (&minRangeX, &minRangeY, &minRangeZ,
     &maxRangeX, &maxRangeY, &maxRangeZ);

  elementSize = 0.0;
  for (iElement = 0; iElement < NElements; iElement++) {
    elementSize += Element_averageEdgeLength (iElement);
  }
  elementSize /= NElements;

#if 1
  elementSize *= 0.7;
#else
  elementSize *= 0.7 * 2;
#endif

  xDivisions = (maxRangeX - minRangeX) / elementSize;
  if (xDivisions < 1) xDivisions = 1;
  yDivisions = (maxRangeY - minRangeY) / elementSize;
  if (yDivisions < 1) yDivisions = 1;
  zDivisions = (maxRangeZ - minRangeZ) / elementSize;
  if (zDivisions < 1) zDivisions = 1;

  offsetX = (maxRangeX - minRangeX) / nOffsetDivisions;
  offsetY = (maxRangeY - minRangeY) / nOffsetDivisions;
  offsetZ = (maxRangeZ - minRangeZ) / nOffsetDivisions;

  AutoRes_Bucket3D_initialize 
    (&ElementBucket,
     minRangeX - offsetX, minRangeY - offsetY, minRangeZ - offsetZ,
     maxRangeX + offsetX, maxRangeY + offsetY, maxRangeZ + offsetZ,
     xDivisions, yDivisions, zDivisions, 
     nAverageLinksPerCell);
}

void Mesh_addElementsIntoBucket (void)
{
  int iElement;

  for (iElement = 0; iElement < NElements; iElement++) {
    double minRangeX, minRangeY, minRangeZ;
    double maxRangeX, maxRangeY, maxRangeZ;
    
    Element_getRange (iElement,
		      &minRangeX, &minRangeY, &minRangeZ,
		      &maxRangeX, &maxRangeY, &maxRangeZ);
    
    AutoRes_Bucket3D_addItemWithin 
      (&ElementBucket,
       minRangeX, minRangeY, minRangeZ,
       maxRangeX, maxRangeY, maxRangeZ,
       iElement);
  }
}

void Mesh_deleteElementBucket (void)
{
  AutoRes_Bucket3D_terminate 
    (&ElementBucket);
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void WriteNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int oldNodeId = Node_oldId[iNode];

    fprintf (fp, "%d   %22.15e %22.15e %22.15e\n", 
	     iNode,
	     Node_coord[oldNodeId][0],
	     Node_coord[oldNodeId][1],
	     Node_coord[oldNodeId][2]); 
  }

  fclose (fp);
}

void WriteElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NElements);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int oldElementId = Element_oldId[iElement];

    fprintf (fp, "%d   ", 
	     iElement); 
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId = Element_nodeIds[oldElementId][iElementNode];
      int newNodeId = Node_newId[nodeId];

      fprintf (fp, "%d ", 
	       newNodeId); 
    }
    fprintf (fp, "\n");
  }
  
  fclose (fp);
}



/************************
	 Command
*************************/

void SetElementNumber (void)
{
  int iElement;
  int count;
  int xCells, yCells, zCells;
  int cx, cy, cz;
  int linkIndex;

  for (iElement = 0; iElement < NElements; iElement++) {
    Element_newId[iElement] = -1;
  }

  Mesh_createElementBucket ();
  Mesh_addElementsIntoBucket ();

  AutoRes_Bucket3D_getNCells 
    (&ElementBucket,
     &xCells, &yCells, &zCells);
  
  count = 0;
  for (cx = 0; cx < xCells; cx++) {
    for (cy = 0; cy < yCells; cy++) {
      for (cz = 0; cz < zCells; cz++) {
	for (linkIndex = AutoRes_Bucket3D_startLinkIndexIn 
	       (&ElementBucket,
		cx, cy, cz);
	     linkIndex != -1;
	     linkIndex = AutoRes_Bucket3D_nextLinkIndex 
	       (&ElementBucket,
		linkIndex)) {
	  int elementId = AutoRes_Bucket3D_itemId 
	    (&ElementBucket,
	     linkIndex);
	  
	  if (Element_newId[elementId] == -1) {
	    Element_newId[elementId] = count;
	    count++;
	  }
	}
      }
    }
  }

  assert(count == NElements);
  for (iElement = 0; iElement < NElements; iElement++) {
    assert(0 <= Element_newId[iElement]);
    assert(Element_newId[iElement] < NElements);
  }

  for (iElement = 0; iElement < NElements; iElement++) {
    Element_oldId[iElement] = -1;
  }
  for (iElement = 0; iElement < NElements; iElement++) {
    int newId = Element_newId[iElement]; 

    assert(Element_oldId[newId] == -1);
    Element_oldId[newId] = iElement;
  }

  for (iElement = 0; iElement < NElements; iElement++) {
    assert(0 <= Element_oldId[iElement]);
    assert(Element_oldId[iElement] < NElements);
  }

  Mesh_deleteElementBucket ();
}

void SetNodeNumber (void)
{
  int iNode;
  int iElement;
  int iElementNode;
  int count;

  for (iNode = 0; iNode < NNodes; iNode++) {
    Node_newId[iNode] = -1;
  }

  count = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    int oldElementId = Element_oldId[iElement];

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId = Element_nodeIds[oldElementId][iElementNode]; 

      if (Node_newId[nodeId] == -1) {
	Node_newId[nodeId] = count;
	count++;
      }
    }
  }

  assert(count == NNodes);
  for (iNode = 0; iNode < NNodes; iNode++) {
    assert(0 <= Node_newId[iNode]);
    assert(Node_newId[iNode] < NNodes);
  }

  for (iNode = 0; iNode < NNodes; iNode++) {
    Node_oldId[iNode] = -1;
  }
  for (iNode = 0; iNode < NNodes; iNode++) {
    int newId = Node_newId[iNode]; 
    
    assert(Node_oldId[newId] == -1);
    Node_oldId[newId] = iNode;
  }
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    assert(0 <= Node_oldId[iNode]);
    assert(Node_oldId[iNode] < NNodes);
  }
}

void PerformCommand (void)
{
  SetElementNumber ();
  SetNodeNumber ();
}



int main (int argc, char *argv[])
{
  if (argc != 5) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_hexa3d_bucket_renumber  node_file  element_file \n"
       "        node_file_OUT  element_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);

  PerformCommand ();

  WriteNodeFile (argv[3]);
  WriteElementFile (argv[4]);

  return 0;
}


