/* advautomesh_h_hexa3d_face2.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>



/************************
	 Basic
*************************/

void SortId3 (int id0, int id1, int id2,
	      int *minId_OUT, int *midId_OUT, int *maxId_OUT)
{
  assert(id0 != id1);
  assert(id0 != id2);
  assert(id1 != id2);

  if (id0 < id1 && id0 < id2) {
    *minId_OUT = id0;
  }
  if (id1 < id0 && id1 < id2) {
    *minId_OUT = id1;
  }
  if (id2 < id0 && id2 < id1) {
    *minId_OUT = id2;
  }

  if ((id1 < id0 && id0 < id2)
      || (id2 < id0 && id0 < id1)) {
    *midId_OUT = id0;
  }
  if ((id0 < id1 && id1 < id2)
      || (id2 < id1 && id1 < id0)) {
    *midId_OUT = id1;
  }
  if ((id0 < id2 && id2 < id1)
      || (id1 < id2 && id2 < id0)) {
    *midId_OUT = id2;
  }

  if (id1 < id0 && id2 < id0) {
    *maxId_OUT = id0;
  }
  if (id0 < id1 && id2 < id1) {
    *maxId_OUT = id1;
  }
  if (id0 < id2 && id1 < id2) {
    *maxId_OUT = id2;
  }
}

void SortId4 (int id0, int id1, int id2, int id3,
	      int *idA_OUT, int *idB_OUT, int *idC_OUT, int *idD_OUT)
{
  int min, mid, max;

  assert(id0 != id1);
  assert(id0 != id2);
  assert(id0 != id3);
  assert(id1 != id2);
  assert(id1 != id3);
  assert(id2 != id3);

  SortId3 (id0, id1, id2,
	   &min, &mid, &max);

  assert(id3 != min);
  assert(id3 != mid);
  assert(id3 != max);
  
  if (id3 < min) {
    *idA_OUT = id3;
    *idB_OUT = min;
    *idC_OUT = mid;
    *idD_OUT = max;
  } else if (id3 < mid) {
    *idA_OUT = min;
    *idB_OUT = id3;
    *idC_OUT = mid;
    *idD_OUT = max;
  } else if (id3 < max) {
    *idA_OUT = min;
    *idB_OUT = mid;
    *idC_OUT = id3;
    *idD_OUT = max;
  } else {
    assert(max < id3);

    *idA_OUT = min;
    *idB_OUT = mid;
    *idC_OUT = max;
    *idD_OUT = id3;
  }

  assert(*idA_OUT < *idB_OUT);
  assert(*idB_OUT < *idC_OUT);
  assert(*idC_OUT < *idD_OUT);
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;



/* Element */

#define N_ELEMENT_NODES 8
#define N_ELEMENT_FACES 6
#define MAX_ELEMENTS 1000000

const int FaceNodeMap[N_ELEMENT_FACES][4] = {
  { 0, 3, 2, 1 },
  { 4, 5, 6, 7 },
  { 0, 1, 5, 4 },
  { 1, 2, 6, 5 },
  { 2, 3, 7, 6 },
  { 0, 4, 7, 3 }
};

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];
int Element_faceIds[MAX_ELEMENTS][N_ELEMENT_FACES];



/* Face */

#define N_FACE_NODES 4
#define MAX_FACES 6000000

int NFaces;
int Face_nodeIds[MAX_FACES][N_FACE_NODES];
int Face_elementIds[MAX_FACES][2];
int Face_elementFaceIds[MAX_FACES][2];

AutoRes_ListTable FaceTable;



void Face_addElement 
(int faceId, int elementId, int elementFaceId)
{
  assert(0 <= faceId);
  assert(faceId < NFaces);
  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(0 <= elementFaceId);
  assert(elementFaceId < N_ELEMENT_FACES);

  assert(Face_elementIds[faceId][1] == -1);
  if (Face_elementIds[faceId][0] == -1) {
    Face_elementIds[faceId][0] = elementId;
    Face_elementFaceIds[faceId][0] = elementFaceId;
  } else {
    Face_elementIds[faceId][1] = elementId;
    Face_elementFaceIds[faceId][1] = elementFaceId;
  }
}

int Mesh_addNewFace 
(int nodeId0, int nodeId1, int nodeId2, int nodeId3)
{
  int faceId = NFaces;
  int hashTableSize;
  int hashValue;

  assert(0 <= nodeId0);
  assert(nodeId0 < NNodes);
  assert(0 <= nodeId1);
  assert(nodeId1 < NNodes);
  assert(0 <= nodeId2);
  assert(nodeId2 < NNodes);
  assert(0 <= nodeId3);
  assert(nodeId3 < NNodes);

  assert(NFaces < MAX_FACES);
  NFaces++;

  Face_nodeIds[faceId][0] = nodeId0;
  Face_nodeIds[faceId][1] = nodeId1;
  Face_nodeIds[faceId][2] = nodeId2;
  Face_nodeIds[faceId][3] = nodeId3;
  Face_elementIds[faceId][0] = -1;
  Face_elementIds[faceId][1] = -1;
  Face_elementFaceIds[faceId][0] = -1;
  Face_elementFaceIds[faceId][1] = -1;
  
  hashTableSize = AutoRes_ListTable_nLists 
    (&FaceTable);
  hashValue = (nodeId0 + nodeId1 + nodeId2 + nodeId3) % hashTableSize;
  AutoRes_ListTable_addItemIn 
    (&FaceTable,
     hashValue, faceId);

  return faceId;
}

int Mesh_findFace 
(int nodeId0, int nodeId1, int nodeId2, int nodeId3)
{
  int nodeIdA, nodeIdB, nodeIdC, nodeIdD;
  int hashTableSize;
  int hashValue;
  int linkIndex;

  assert(0 <= nodeId0);
  assert(nodeId0 < NNodes);
  assert(0 <= nodeId1);
  assert(nodeId1 < NNodes);
  assert(0 <= nodeId2);
  assert(nodeId2 < NNodes);
  assert(0 <= nodeId3);
  assert(nodeId3 < NNodes);

  SortId4 (nodeId0, nodeId1, nodeId2, nodeId3,
	   &nodeIdA, &nodeIdB, &nodeIdC, &nodeIdD);
  
  hashTableSize = AutoRes_ListTable_nLists 
    (&FaceTable);
  hashValue = (nodeId0 + nodeId1 + nodeId2 + nodeId3) % hashTableSize;
  for (linkIndex = AutoRes_ListTable_startLinkIndexIn 
	 (&FaceTable, 
	  hashValue);
       linkIndex != -1;
       linkIndex = AutoRes_ListTable_nextLinkIndex 
	 (&FaceTable, 
	  linkIndex)) {
    int faceId = AutoRes_ListTable_itemId 
      (&FaceTable, 
       linkIndex);
    int faceNodeId0 = Face_nodeIds[faceId][0];
    int faceNodeId1 = Face_nodeIds[faceId][1];
    int faceNodeId2 = Face_nodeIds[faceId][2];
    int faceNodeId3 = Face_nodeIds[faceId][3];
    int faceNodeIdA, faceNodeIdB, faceNodeIdC, faceNodeIdD; 

    SortId4 (faceNodeId0, faceNodeId1, faceNodeId2, faceNodeId3,
	     &faceNodeIdA, &faceNodeIdB, &faceNodeIdC, &faceNodeIdD);

    if (nodeIdA == faceNodeIdA
	&& nodeIdB == faceNodeIdB
	&& nodeIdC == faceNodeIdC
	&& nodeIdD == faceNodeIdD) {
      return faceId;
    }
  }

  return -1;
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  /* skip coordinate */

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void WriteFaceFile (const char *fileName)
{
  FILE *fp;
  int iFace;
  
  assert(0 < NFaces);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  fprintf (fp, "%d\n", 
	   NFaces);
  
  for (iFace = 0; iFace < NFaces; iFace++) {
    fprintf (fp, "%d   %d %d   %d %d\n", 
	     iFace,
	     Face_elementIds[iFace][0],
	     Face_elementFaceIds[iFace][0],
	     Face_elementIds[iFace][1],
	     Face_elementFaceIds[iFace][1]);
  }

  fclose (fp);
}

void WriteFaceVertexFile (const char *fileName)
{
  FILE *fp;
  int iFace;
  
  assert(0 < NFaces);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  fprintf (fp, "%d\n", 
	   NFaces);
  
  for (iFace = 0; iFace < NFaces; iFace++) {
    fprintf (fp, "%d   %d %d %d %d\n", 
	     iFace,
	     Face_nodeIds[iFace][0],
	     Face_nodeIds[iFace][1],
	     Face_nodeIds[iFace][2],
	     Face_nodeIds[iFace][3]);
  }

  fclose (fp);
}

void WriteElementFaceFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementFace;
  
  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  fprintf (fp, "%d\n", 
	   NElements);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    fprintf (fp, "%d   ", 
	     iElement);

    for (iElementFace = 0; 
	 iElementFace < N_ELEMENT_FACES; 
	 iElementFace++) {
      fprintf (fp, "%d ", 
	       Element_faceIds[iElement][iElementFace]);
    }

    fprintf (fp, "\n");
  }

  fclose (fp);
}



/************************
	 Command
*************************/

void PerformCommand (void)
{
  int hashTableSize = NNodes * 10;
  int nAverageLinksPerList = 2;
  int iElement;
  int iElementFace;
  int iFace;

  AutoRes_ListTable_initialize 
    (&FaceTable,
     hashTableSize, nAverageLinksPerList);

  assert(NFaces == 0);

  for (iElement = 0; iElement < NElements; iElement++) {
    for (iElementFace = 0; 
	 iElementFace < N_ELEMENT_FACES; 
	 iElementFace++) {
      int faceNodeId0 
	= Element_nodeIds[iElement][FaceNodeMap[iElementFace][0]];
      int faceNodeId1 
	= Element_nodeIds[iElement][FaceNodeMap[iElementFace][1]];
      int faceNodeId2 
	= Element_nodeIds[iElement][FaceNodeMap[iElementFace][2]];
      int faceNodeId3 
	= Element_nodeIds[iElement][FaceNodeMap[iElementFace][3]];
      int faceId = Mesh_findFace 
	(faceNodeId0, faceNodeId1, faceNodeId2, faceNodeId3);

      if (faceId == -1) {
	faceId = Mesh_addNewFace 
	  (faceNodeId0, faceNodeId1, faceNodeId2, faceNodeId3);
      }
      Element_faceIds[iElement][iElementFace] = faceId;
      Face_addElement (faceId, iElement, iElementFace);
    }
  }

  for (iFace = 0; iFace < NFaces; iFace++) {
    assert(Face_elementIds[iFace][0] != -1);
    assert(Face_elementIds[iFace][1] != -1
	   || Face_elementIds[iFace][1] == -1);
  }

  AutoRes_ListTable_terminate 
    (&FaceTable);
}



int main (int argc, char *argv[])
{
  if (argc != 6) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_hexa3d_face2  node_file  element_file \n"
       "        face_file_OUT  face_vertex_OUT  element_face_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);

  PerformCommand ();

  WriteFaceFile (argv[3]);
  WriteFaceVertexFile (argv[4]);
  WriteElementFaceFile (argv[5]);

  return 0;
}
