/* advautomesh_h_hexa3d_mesh.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001

double Eps = DEFAULT_EPS;



double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][3];



/* Element */

#define N_ELEMENT_NODES 8
#define MAX_ELEMENTS 1000000

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



int Mesh_findNodeAt (double x, double y, double z)
{
  int iNode;

  for (iNode = 0; iNode < NNodes; iNode++) {
    double nodeX = Node_coord[iNode][0];
    double nodeY = Node_coord[iNode][1];
    double nodeZ = Node_coord[iNode][2];

    if (Distance (x, y, z, 
		  nodeX, nodeY, nodeZ) < Eps) {
      return iNode;
    }
  }

  return -1;
}

int Mesh_addNewNode (double x, double y, double z)
{
  int nodeId = NNodes;

  assert(NNodes < MAX_NODES);
  NNodes++;

  Node_coord[nodeId][0] = x;
  Node_coord[nodeId][1] = y;
  Node_coord[nodeId][2] = z;
  
  return nodeId;
}

int Mesh_addNewElement 
(int nodeIds[N_ELEMENT_NODES])
{
  int elementId = NElements;
  int iElementNode;
  
  for (iElementNode = 0; 
       iElementNode < N_ELEMENT_NODES; 
       iElementNode++) {
    assert(0 <= nodeIds[iElementNode]);
    assert(nodeIds[iElementNode] < NNodes);
  }

  assert(NElements < MAX_ELEMENTS);
  NElements++;

  for (iElementNode = 0; 
       iElementNode < N_ELEMENT_NODES; 
       iElementNode++) {
    Element_nodeIds[elementId][iElementNode] = nodeIds[iElementNode];
  }    

  return elementId;
}

void WriteNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    fprintf (fp, "%d   %22.15e %22.15e %22.15e\n", 
	     iNode,
	     Node_coord[iNode][0],
	     Node_coord[iNode][1],
	     Node_coord[iNode][2]); 
  }

  fclose (fp);
}

void WriteElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NElements);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    fprintf (fp, "%d   ", 
	     iElement); 
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      fprintf (fp, "%d ", 
	       Element_nodeIds[iElement][iElementNode]); 
    }
    fprintf (fp, "\n");
  }
  
  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Hexahedron */

#define MAX_HEXAS 1000000

int NHexas;
double Hexa_coord[MAX_HEXAS][N_ELEMENT_NODES][3];



void ReadHexaFile (const char *fileName)
{
  FILE *fp;
  int iHexa;
  int iElementNode;

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NHexas);
  assert(NHexas <= MAX_HEXAS);
  
  for (iHexa = 0; iHexa < NHexas; iHexa++) {
    int hexaId;

    fscanf (fp, "%d", 
	    &hexaId);
    assert(hexaId == iHexa);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      double x, y, z;
      
      fscanf (fp, "%lf %lf %lf", 
	      &x, &y, &z);

      Hexa_coord[iHexa][iElementNode][0] = x;
      Hexa_coord[iHexa][iElementNode][1] = y;
      Hexa_coord[iHexa][iElementNode][2] = z;
    }
  }

  fclose (fp);
}



/************************
	 Command
*************************/

int NodeAt (double x, double y, double z)
{
  int nodeId = Mesh_findNodeAt (x, y, z);
  
  if (nodeId == -1) {
    nodeId = Mesh_addNewNode (x, y, z);
  }

  return nodeId;
}

void PerformCommand (void)
{
  int iHexa;
  int iElementNode;
  int nodeIds[N_ELEMENT_NODES];
  
  for (iHexa = 0; iHexa < NHexas; iHexa++) {
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      double x = Hexa_coord[iHexa][iElementNode][0];
      double y = Hexa_coord[iHexa][iElementNode][1];
      double z = Hexa_coord[iHexa][iElementNode][2];
      
      int nodeId = NodeAt (x, y, z);
      
      nodeIds[iElementNode] = nodeId;
    }

    Mesh_addNewElement (nodeIds);
  }
}



int main (int argc, char *argv[])
{
  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_hexa3d_mesh  hexa_file \n"
       "     node_file_OUT  element_file_OUT \n");
    exit (1);
  }

  ReadHexaFile (argv[1]);

  PerformCommand ();

  WriteNodeFile (argv[2]);
  WriteElementFile (argv[3]);

  return 0;
}
