/* advautomesh_h_quad2d_edge_on_line.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001

double Eps = DEFAULT_EPS;



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][2];
int Node_isOn[MAX_NODES];



/* Element */

#define N_ELEMENT_NODES 4
#define N_ELEMENT_EDGES 4
#define MAX_ELEMENTS 1000000

const int EdgeNodeMap[N_ELEMENT_EDGES][2] = {
  { 0, 1 },
  { 1, 2 },
  { 2, 3 },
  { 3, 0 }
};

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



/* Edge */

#define MAX_EDGES 2000000

int NEdges;
int Edge_elementIds[MAX_NODES][2];
int Edge_elementEdgeIds[MAX_NODES][2];
int Edge_isOn[MAX_EDGES];



void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y;

    fscanf (fp, "%d %lf %lf", 
	    &nodeId, 
	    &x, &y);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void ReadEdgeFile (const char *fileName)
{
  FILE *fp;
  int iEdge;
  
  assert(NEdges == 0);
  assert(0 < NNodes);
  assert(0 < NElements);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NEdges);
  assert(0 < NEdges);
  assert(NEdges <= MAX_EDGES);
  
  for (iEdge = 0; iEdge < NEdges; iEdge++) {
    int edgeId;
    int elementId0, elementId1;
    int elementEdgeId0, elementEdgeId1;

    fscanf (fp, "%d %d %d %d %d", 
	    &edgeId,
	    &elementId0, &elementEdgeId0,
	    &elementId1, &elementEdgeId1);
    assert(edgeId == iEdge);

    assert(elementId0 != -1);

    assert(0 <= elementId0);
    assert(elementId0 < NElements);
    assert(0 <= elementEdgeId0);
    assert(elementEdgeId0 < N_ELEMENT_EDGES);

    if (0 <= elementId1) {
      assert(elementId1 < NElements);
      assert(0 <= elementEdgeId1);
      assert(elementEdgeId1 < N_ELEMENT_EDGES);
    } else {
      assert(elementId1 == -1);
      assert(elementEdgeId1 == -1);
    }

    Edge_elementIds[iEdge][0] = elementId0;
    Edge_elementIds[iEdge][1] = elementId1;

    Edge_elementEdgeIds[iEdge][0] = elementEdgeId0;
    Edge_elementEdgeIds[iEdge][1] = elementEdgeId1;
  }

  fclose (fp);
}

void WriteEdgeListFile (const char *fileName)
{
  FILE *fp;
  int iEdge;
  int count;

  assert(0 < NEdges);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  count = 0;
  for (iEdge = 0; iEdge < NEdges; iEdge++) {
    if (Edge_isOn[iEdge]) {
      count++;
    }
  }
  
  fprintf (fp, "%d\n", 
	   count);
  
  count = 0;
  for (iEdge = 0; iEdge < NEdges; iEdge++) {
    if (Edge_isOn[iEdge]) {
      fprintf (fp, "%d   %d\n", 
	       count,
	       iEdge);
      count++;
    }
  }

  fclose (fp);
}



/************************
	 Command
*************************/

double LineA, LineB, LineC;

void PerformCommand (void)
{
  int iNode;
  int iEdge;
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    Node_isOn[iNode] = 0;
  }
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    double value = LineA * x + LineB * y + LineC;
    
    if (fabs (value) < Eps) {
      Node_isOn[iNode] = 1;
    }
  }

  for (iEdge = 0; iEdge < NEdges; iEdge++) {
    Edge_isOn[iEdge] = 0;
  }

  for (iEdge = 0; iEdge < NEdges; iEdge++) {
    if (Edge_elementIds[iEdge][1] == -1) {
      int elementId = Edge_elementIds[iEdge][0];
      int elementEdgeId = Edge_elementEdgeIds[iEdge][0];
    
      int nodeId0 
	= Element_nodeIds[elementId][EdgeNodeMap[elementEdgeId][0]];
      int nodeId1 
	= Element_nodeIds[elementId][EdgeNodeMap[elementEdgeId][1]];
    
      if (Node_isOn[nodeId0]
	  && Node_isOn[nodeId1]) {
	Edge_isOn[iEdge] = 1;
      }
    }
  }
}



int main (int argc, char *argv[])
{
  if (argc != 8) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_quad2d_edge_on_line  node_file  element_file \n"
       "        edge_file  a  b  c \n"
       "        edge_list_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadEdgeFile (argv[3]);

  LineA = atof (argv[4]);
  LineB = atof (argv[5]);
  LineC = atof (argv[6]);

  PerformCommand ();

  WriteEdgeListFile (argv[7]);

  return 0;
}
