/* advautomesh_h_quad2d_mesh.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001

double Eps = DEFAULT_EPS;



double Distance (double x0, double y0, 
		 double x1, double y1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) + (y1 - y0) * (y1 - y0));
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][2];



/* Element */

#define N_ELEMENT_NODES 4
#define MAX_ELEMENTS 1000000

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



int Mesh_findNodeAt (double x, double y)
{
  int iNode;

  for (iNode = 0; iNode < NNodes; iNode++) {
    double nodeX = Node_coord[iNode][0];
    double nodeY = Node_coord[iNode][1];

    if (Distance (x, y, 
		  nodeX, nodeY) < Eps) {
      return iNode;
    }
  }

  return -1;
}

int Mesh_addNewNode (double x, double y)
{
  int nodeId = NNodes;

  assert(NNodes < MAX_NODES);
  NNodes++;

  Node_coord[nodeId][0] = x;
  Node_coord[nodeId][1] = y;
  
  return nodeId;
}

int Mesh_addNewElement 
(int nodeId0, int nodeId1, int nodeId2, int nodeId3)
{
  int elementId = NElements;

  assert(0 <= nodeId0);
  assert(nodeId0 < NNodes);
  assert(0 <= nodeId1);
  assert(nodeId1 < NNodes);
  assert(0 <= nodeId2);
  assert(nodeId2 < NNodes);
  assert(0 <= nodeId3);
  assert(nodeId3 < NNodes);

  assert(NElements < MAX_ELEMENTS);
  NElements++;

  Element_nodeIds[elementId][0] = nodeId0;
  Element_nodeIds[elementId][1] = nodeId1;
  Element_nodeIds[elementId][2] = nodeId2;
  Element_nodeIds[elementId][3] = nodeId3;

  return elementId;
}

void WriteNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    fprintf (fp, "%d   %22.15e %22.15e\n", 
	     iNode,
	     Node_coord[iNode][0],
	     Node_coord[iNode][1]); 
  }

  fclose (fp);
}

void WriteElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NElements);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    fprintf (fp, "%d   ", 
	     iElement); 
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      fprintf (fp, "%d ", 
	       Element_nodeIds[iElement][iElementNode]); 
    }
    fprintf (fp, "\n");
  }
  
  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Quadrilateral */

#define MAX_QUADS 100000

int NQuads;
double Quad_coord[MAX_QUADS][N_ELEMENT_NODES][2];



void ReadQuadFile (const char *fileName)
{
  FILE *fp;
  int iQuad;

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NQuads);
  assert(NQuads <= MAX_QUADS);
  
  for (iQuad = 0; iQuad < NQuads; iQuad++) {
    int quadId;
    double x0, y0;
    double x1, y1;
    double x2, y2;
    double x3, y3;

    fscanf (fp, "%d %lf %lf %lf %lf %lf %lf %lf %lf", 
	    &quadId, 
	    &x0, &y0, 
	    &x1, &y1, 
	    &x2, &y2, 
	    &x3, &y3);
    assert(quadId == iQuad);
    assert(Eps < Distance (x0, y0, 
			   x1, y1));
    assert(Eps < Distance (x1, y1, 
			   x2, y2));
    assert(Eps < Distance (x2, y2, 
			   x3, y3));
    assert(Eps < Distance (x3, y3, 
			   x0, y0));
    assert(Eps < Distance (x0, y0, 
			   x2, y2));
    assert(Eps < Distance (x1, y1, 
			   x3, y3));

    Quad_coord[iQuad][0][0] = x0;
    Quad_coord[iQuad][0][1] = y0;

    Quad_coord[iQuad][1][0] = x1;
    Quad_coord[iQuad][1][1] = y1;

    Quad_coord[iQuad][2][0] = x2;
    Quad_coord[iQuad][2][1] = y2;

    Quad_coord[iQuad][3][0] = x3;
    Quad_coord[iQuad][3][1] = y3;
  }

  fclose (fp);
}



/************************
	 Command
*************************/

int NodeAt (double x, double y)
{
  int nodeId = Mesh_findNodeAt (x, y);
  
  if (nodeId == -1) {
    nodeId = Mesh_addNewNode (x, y);
  }

  return nodeId;
}

void PerformCommand (void)
{
  int iQuad;

  for (iQuad = 0; iQuad < NQuads; iQuad++) {
    double x0 = Quad_coord[iQuad][0][0];
    double y0 = Quad_coord[iQuad][0][1];

    double x1 = Quad_coord[iQuad][1][0];
    double y1 = Quad_coord[iQuad][1][1];

    double x2 = Quad_coord[iQuad][2][0];
    double y2 = Quad_coord[iQuad][2][1];

    double x3 = Quad_coord[iQuad][3][0];
    double y3 = Quad_coord[iQuad][3][1];

    int nodeId0 = NodeAt (x0, y0);
    int nodeId1 = NodeAt (x1, y1);
    int nodeId2 = NodeAt (x2, y2);
    int nodeId3 = NodeAt (x3, y3);

    Mesh_addNewElement (nodeId0, nodeId1, nodeId2, nodeId3);
  }
}



int main (int argc, char *argv[])
{
  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_quad2d_mesh  quad_file \n"
       "        node_file_OUT  element_file_OUT \n");
    exit (1);
  }

  ReadQuadFile (argv[1]);

  PerformCommand ();

  WriteNodeFile (argv[2]);
  WriteElementFile (argv[3]);

  return 0;
}
