/* advautomesh_h_tetra3d_isosurface_view.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autogl.h>



/************************
	 Basic
*************************/

double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 2500000

int NNodes;
double Node_coord[MAX_NODES][3];



/* Element */

#define N_ELEMENT_NODES 4
#define N_ELEMENT_EDGES 6
#define MAX_ELEMENTS 12000000

const int EdgeNodeMap[N_ELEMENT_EDGES][2] = {
  { 0, 1 }, 
  { 0, 2 }, 
  { 0, 3 }, 
  { 1, 2 }, 
  { 2, 3 }, 
  { 3, 1 }  
};

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



void Element_getEdgePoint 
(int elementId, 
 int elementEdgeId, double ratio,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  int nodeId0, nodeId1;
  double x0, y0, z0;
  double x1, y1, z1;

  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(0 <= elementEdgeId);
  assert(elementEdgeId < N_ELEMENT_EDGES);

  nodeId0 = Element_nodeIds
    [elementId][EdgeNodeMap[elementEdgeId][0]];
  nodeId1 = Element_nodeIds
    [elementId][EdgeNodeMap[elementEdgeId][1]];

  x0 = Node_coord[nodeId0][0];
  y0 = Node_coord[nodeId0][1];
  z0 = Node_coord[nodeId0][2];

  x1 = Node_coord[nodeId1][0];
  y1 = Node_coord[nodeId1][1];
  z1 = Node_coord[nodeId1][2];
  
  *x_OUT = x0 + (x1 - x0) * ratio;
  *y_OUT = y0 + (y1 - y0) * ratio;
  *z_OUT = z0 + (z1 - z0) * ratio;
}

void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  *minRangeZ_OUT = *maxRangeZ_OUT = Node_coord[0][2];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    double z = Node_coord[iNode][2];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
    if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
  }
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Isosurface */

#define MAX_LEVELS 10000
#define MAX_ISOSURFACES 1000000

int NLevels;
int Level_startIsosurfaceId[MAX_LEVELS];
int Level_nIsosurfaces[MAX_LEVELS];
int NIsosurfaces;
int Isosurface_elementId[MAX_ISOSURFACES];
int Isosurface_elementEdgeIds[MAX_ISOSURFACES][3];
double Isosurface_ratios[MAX_ISOSURFACES][3];

double MinLevel, MaxLevel;
int NDivisions;



/* Frame */

#define MAX_FRAMES 1000000

int NFrames;
double Frame_coord0[MAX_FRAMES][3];
double Frame_coord1[MAX_FRAMES][3];



void ReadIsosurfaceFile (const char *fileName)
{
  FILE *fp;
  int iLevel;
  int iIsosurface;
  int count;
  
  assert(0 < NElements);
  assert(NLevels == 0);
  assert(NIsosurfaces == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NLevels);
  assert(0 < NLevels);
  assert(NLevels <= MAX_LEVELS);
  
  count = 0;
  for (iLevel = 0; iLevel < NLevels; iLevel++) {
    int levelId;
    int startId;
    int nIsosurfaces;

    fscanf (fp, "%d %d", 
	    &levelId,
	    &nIsosurfaces);
    assert(levelId == iLevel);
    assert(0 <= nIsosurfaces);
    startId = count;
    count += nIsosurfaces;
    assert(count <= MAX_ISOSURFACES);
    Level_startIsosurfaceId[iLevel] = startId;
    Level_nIsosurfaces[iLevel] = nIsosurfaces;

    for (iIsosurface = startId; 
	 iIsosurface < startId + nIsosurfaces; 
	 iIsosurface++) {
      int isosurfaceId;
      int elementId;
      int elementEdgeId0, elementEdgeId1, elementEdgeId2;
      double ratio0, ratio1, ratio2;
      
      fscanf (fp, "%d %d %d %lf %d %lf %d %lf", 
	      &isosurfaceId,
	      &elementId,
	      &elementEdgeId0, &ratio0,
	      &elementEdgeId1, &ratio1,
	      &elementEdgeId2, &ratio2);
      assert(isosurfaceId == iIsosurface - startId);
      assert(0 <= elementId);
      assert(elementId < NElements);
      Isosurface_elementId[iIsosurface] = elementId;
      
      assert(0 <= elementEdgeId0);
      assert(elementEdgeId0 < N_ELEMENT_EDGES);
      Isosurface_elementEdgeIds[iIsosurface][0] = elementEdgeId0;
      Isosurface_ratios[iIsosurface][0] = ratio0;
      
      assert(0 <= elementEdgeId1);
      assert(elementEdgeId1 < N_ELEMENT_EDGES);
      Isosurface_elementEdgeIds[iIsosurface][1] = elementEdgeId1;
      Isosurface_ratios[iIsosurface][1] = ratio1;
      
      assert(0 <= elementEdgeId2);
      assert(elementEdgeId2 < N_ELEMENT_EDGES);
      Isosurface_elementEdgeIds[iIsosurface][2] = elementEdgeId2;
      Isosurface_ratios[iIsosurface][2] = ratio2;
    }
  }

  fclose (fp);
}

void ReadFrameFile (const char *fileName)
{
  FILE *fp;
  int iFrame;
  
  assert(NFrames == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NFrames);
  assert(0 <= NFrames);
  assert(NFrames <= MAX_FRAMES);
  
  for (iFrame = 0; iFrame < NFrames; iFrame++) {
    int frameId;
    double x0, y0, z0;
    double x1, y1, z1;

    fscanf (fp, "%d %lf %lf %lf %lf %lf %lf", 
	    &frameId,
	    &x0, &y0, &z0, 
	    &x1, &y1, &z1);
    assert(frameId == iFrame);
    Frame_coord0[iFrame][0] = x0;
    Frame_coord0[iFrame][1] = y0;
    Frame_coord0[iFrame][2] = z0;
    Frame_coord1[iFrame][0] = x1;
    Frame_coord1[iFrame][1] = y1;
    Frame_coord1[iFrame][2] = z1;
  }

  fclose (fp);
}



/********************************
	 Visualization
********************************/

void SetView (void)
{
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  double centerX, centerY, centerZ;
  double size;
  
  Mesh_getRange 
    (&minRangeX, &minRangeY, &minRangeZ,
     &maxRangeX, &maxRangeY, &maxRangeZ);
  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;
  centerZ = (minRangeZ + maxRangeZ) / 2;
  size = Distance (minRangeX, minRangeY, minRangeZ,
		   maxRangeX, maxRangeY, maxRangeZ) / 2;

  AutoGL_SetViewSize (size);
  AutoGL_SetViewCenter (centerX, centerY, centerZ); 
}

void DrawFrame (void)
{
  int iFrame;

  AutoGL_SetColor (1, 1, 1);
  for (iFrame = 0; iFrame < NFrames; iFrame++) {
    AutoGL_DrawLine (Frame_coord0[iFrame][0],
		     Frame_coord0[iFrame][1],
		     Frame_coord0[iFrame][2],
		     Frame_coord1[iFrame][0],
		     Frame_coord1[iFrame][1],
		     Frame_coord1[iFrame][2]);
  }
}

#if 1
#include "../../util/time_measure.c"
#define TIME_MEASURE
#endif

void RedrawView (void) 
{
  int iLevel;
  int iIsosurface;

#ifdef TIME_MEASURE
  StartMeasureTime ();
#endif

  AutoGL_DrawLineContourMap (80, MinLevel, MaxLevel, NDivisions);

  DrawFrame ();

  for (iLevel = 0; iLevel < NLevels; iLevel++) {
    int startId = Level_startIsosurfaceId[iLevel];
    int nIsosurfaces = Level_nIsosurfaces[iLevel];
    double grade = 1.0 / NDivisions * iLevel;
    double red, green, blue;

    AutoGL_GetContourColor (&red, &green, &blue,
			    grade);
    AutoGL_SetColor (red, green, blue);
    
    for (iIsosurface = startId; 
	 iIsosurface < startId + nIsosurfaces; 
	 iIsosurface++) {
      double ex0, ey0, ez0;
      double ex1, ey1, ez1;
      double ex2, ey2, ez2;
      
      Element_getEdgePoint 
	(Isosurface_elementId[iIsosurface], 
	 Isosurface_elementEdgeIds[iIsosurface][0],
	 Isosurface_ratios[iIsosurface][0],
	 &ex0, &ey0, &ez0);
      Element_getEdgePoint 
	(Isosurface_elementId[iIsosurface], 
	 Isosurface_elementEdgeIds[iIsosurface][1],
	 Isosurface_ratios[iIsosurface][1],
	 &ex1, &ey1, &ez1);
      Element_getEdgePoint 
	(Isosurface_elementId[iIsosurface], 
	 Isosurface_elementEdgeIds[iIsosurface][2],
	 Isosurface_ratios[iIsosurface][2],
	 &ex2, &ey2, &ez2);
      
      AutoGL_DrawTriangle (ex0, ey0, ez0,
			   ex1, ey1, ez1,
			   ex2, ey2, ez2);
    }
  }

#ifdef TIME_MEASURE
  EndMeasureTime ();
#endif
}

void UpdateModel (void) 
{
  AutoGL_DrawView ();
}

void Quit (void) 
{
  AutoGL_Terminate ();
  exit (0);
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
  if (argc != 8) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tetra3d_isosurface_view  node_file  element_file \n"
       "     isosurface_file  frame_file \n"
       "     minLevel  maxLevel  nDivisions \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadIsosurfaceFile (argv[3]);
  ReadFrameFile (argv[4]);
  MinLevel = atof (argv[5]);
  MaxLevel = atof (argv[6]);
  NDivisions = atoi (argv[7]);

  SetView ();
  
  AutoGL_SetViewRedrawCallback (RedrawView);

  AutoGL_SetDefaultCallbackInMode3D (NULL);
  AutoGL_EnableDragInMode3D ();



  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode3D ();

  AutoGL_AddCallback (Quit, "Quit");       

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");

  AutoGL_AddCallback (UpdateModel, "UpdateModel");      
}
