/* advautomesh_h_tetra3d_node_list_view.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autogl.h>



/************************
	 Basic
*************************/

double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][3];
int Node_isSelected[MAX_NODES];



/* Element */

#define N_ELEMENT_NODES 4
#define N_ELEMENT_FACES 4
#define MAX_ELEMENTS 1000000

const int FaceNodeMap[N_ELEMENT_FACES][3] = {
  { 1, 2, 3 },
  { 0, 3, 2 },
  { 0, 1, 3 },
  { 0, 2, 1 }
};

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



/* Face */

#define MAX_FACES 1000000

int NFaces;
int Face_elementIds[MAX_NODES][2];
int Face_elementFaceIds[MAX_NODES][2];



void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  *minRangeZ_OUT = *maxRangeZ_OUT = Node_coord[0][2];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    double z = Node_coord[iNode][2];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
    if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
  }
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void ReadFaceFile (const char *fileName)
{
  FILE *fp;
  int iFace;
  
  assert(NFaces == 0);
  assert(0 < NNodes);
  assert(0 < NElements);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NFaces);
  assert(0 < NFaces);
  assert(NFaces <= MAX_FACES);
  
  for (iFace = 0; iFace < NFaces; iFace++) {
    int faceId;
    int elementId0, elementId1;
    int elementFaceId0, elementFaceId1;

    fscanf (fp, "%d %d %d %d %d", 
	    &faceId,
	    &elementId0, &elementFaceId0,
	    &elementId1, &elementFaceId1);
    assert(faceId == iFace);

    assert(elementId0 != -1);

    assert(0 <= elementId0);
    assert(elementId0 < NElements);
    assert(0 <= elementFaceId0);
    assert(elementFaceId0 < N_ELEMENT_FACES);

    if (0 <= elementId1) {
      assert(elementId1 < NElements);
      assert(0 <= elementFaceId1);
      assert(elementFaceId1 < N_ELEMENT_FACES);
    } else {
      assert(elementId1 == -1);
      assert(elementFaceId1 == -1);
    }

    Face_elementIds[iFace][0] = elementId0;
    Face_elementIds[iFace][1] = elementId1;

    Face_elementFaceIds[iFace][0] = elementFaceId0;
    Face_elementFaceIds[iFace][1] = elementFaceId1;
  }

  fclose (fp);
}

void ReadNodeListFile (const char *fileName)
{
  FILE *fp;
  int iNode;
  int nItems;
  int iItem;

  assert(0 < NNodes);

  for (iNode = 0; iNode < NNodes; iNode++) {
    Node_isSelected[iNode] = 0;
  }

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &nItems);
  assert(0 < nItems);
  assert(nItems <= MAX_NODES);
  
  for (iItem = 0; iItem < nItems; iItem++) {
    int itemId;
    int nodeId;

    fscanf (fp, "%d %d", 
	    &itemId,
	    &nodeId);
    assert(itemId == iItem);
    assert(0 <= nodeId);
    assert(nodeId < NNodes);
    Node_isSelected[nodeId] = 1;
  }

  fclose (fp);
}



/********************************
	 Visualization
********************************/

int SelectedNode = 1;

void SetView (void)
{
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  double centerX, centerY, centerZ;
  double size;
  
  Mesh_getRange 
    (&minRangeX, &minRangeY, &minRangeZ,
     &maxRangeX, &maxRangeY, &maxRangeZ);
  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;
  centerZ = (minRangeZ + maxRangeZ) / 2;
  size = Distance (minRangeX, minRangeY, minRangeZ,
		   maxRangeX, maxRangeY, maxRangeZ) / 2;

  AutoGL_SetViewSize (size);
  AutoGL_SetViewCenter (centerX, centerY, centerZ); 
}

void RedrawView (void) 
{
  int iNode;
  int iFace;

  AutoGL_SetColor (1, 1, 1);

  for (iFace = 0; iFace < NFaces; iFace++) {
    if (Face_elementIds[iFace][1] == -1) {
      int elementId = Face_elementIds[iFace][0];
      int elementFaceId = Face_elementFaceIds[iFace][0];

      int nodeId0 
	= Element_nodeIds[elementId][FaceNodeMap[elementFaceId][0]];
      int nodeId1 
	= Element_nodeIds[elementId][FaceNodeMap[elementFaceId][1]];
      int nodeId2 
	= Element_nodeIds[elementId][FaceNodeMap[elementFaceId][2]];

      double nodeX0 = Node_coord[nodeId0][0];
      double nodeY0 = Node_coord[nodeId0][1];
      double nodeZ0 = Node_coord[nodeId0][2];
      
      double nodeX1 = Node_coord[nodeId1][0];
      double nodeY1 = Node_coord[nodeId1][1];
      double nodeZ1 = Node_coord[nodeId1][2];
      
      double nodeX2 = Node_coord[nodeId2][0];
      double nodeY2 = Node_coord[nodeId2][1];
      double nodeZ2 = Node_coord[nodeId2][2];

      AutoGL_DrawLine (nodeX0, nodeY0, nodeZ0,
		       nodeX1, nodeY1, nodeZ1);
      AutoGL_DrawLine (nodeX1, nodeY1, nodeZ1,
		       nodeX2, nodeY2, nodeZ2);
      AutoGL_DrawLine (nodeX2, nodeY2, nodeZ2,
		       nodeX0, nodeY0, nodeZ0);
    }
  }

  if (SelectedNode) {
    AutoGL_SetColor (1, 0, 0);

    for (iNode = 0; iNode < NNodes; iNode++) {
      if (Node_isSelected[iNode]) {
	double x = Node_coord[iNode][0];
	double y = Node_coord[iNode][1]; 
	double z = Node_coord[iNode][2]; 
	
	AutoGL_DrawCrossMark3D (x, y, z,
				5);
      }
    }
  }
}

void UpdateModel (void) 
{
  AutoGL_DrawView ();
}

void Quit (void) 
{
  AutoGL_Terminate ();
  exit (0);
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
  if (argc != 5) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tetra3d_node_list_view  node_file  element_file \n"
       "     face_file  node_list_file \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadFaceFile (argv[3]);
  ReadNodeListFile (argv[4]);

  SetView ();
  
  AutoGL_SetViewRedrawCallback (RedrawView);

  AutoGL_SetDefaultCallbackInMode3D (NULL);
  AutoGL_EnableDragInMode3D ();



  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode3D ();

  AutoGL_AddCallback (Quit, "Quit");       

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");

  AutoGL_AddBoolean (&SelectedNode, "SelectedNode");

  AutoGL_AddCallback (UpdateModel, "UpdateModel");      
}
