/* advautomesh_h_tetra3d_topology_view.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autogl.h>



/************************
	 Basic
*************************/

double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}



/************************
	 Mesh
*************************/

/* Node */

typedef enum {
  NODE_TOPOLOGY_INTERNAL,
  NODE_TOPOLOGY_FACE,
  NODE_TOPOLOGY_EDGE,
  NODE_TOPOLOGY_VERTEX
} NodeTopologyType;

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][3];
NodeTopologyType Node_topologyType[MAX_NODES];



/* Element */

#define N_ELEMENT_NODES 4
#define MAX_ELEMENTS 1000000

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  *minRangeZ_OUT = *maxRangeZ_OUT = Node_coord[0][2];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    double z = Node_coord[iNode][2];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
    if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
  }
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void ReadNodeTopologyFile (const char *fileName)
{
  FILE *fp;
  int count;
  int iNode;

  assert(0 < NNodes);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &count);
  assert(count == NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    NodeTopologyType topologyType;
    
    fscanf (fp, "%d %d", 
	    &nodeId,
	    (int *)&topologyType); 
    assert(nodeId == iNode);
    switch (topologyType) {
    case NODE_TOPOLOGY_INTERNAL:
    case NODE_TOPOLOGY_FACE:
    case NODE_TOPOLOGY_EDGE:
    case NODE_TOPOLOGY_VERTEX:
      Node_topologyType[iNode] = topologyType;
      break;
    default:
      assert(0);
      break;
    }
  }

  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Frame */

#define MAX_FRAMES 1000000

int NFrames;
double Frame_coord0[MAX_FRAMES][3];
double Frame_coord1[MAX_FRAMES][3];



void ReadFrameFile (const char *fileName)
{
  FILE *fp;
  int iFrame;
  
  assert(NFrames == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NFrames);
  assert(0 <= NFrames);
  assert(NFrames <= MAX_FRAMES);
  
  for (iFrame = 0; iFrame < NFrames; iFrame++) {
    int frameId;
    double x0, y0, z0;
    double x1, y1, z1;

    fscanf (fp, "%d %lf %lf %lf %lf %lf %lf", 
	    &frameId,
	    &x0, &y0, &z0, 
	    &x1, &y1, &z1);
    assert(frameId == iFrame);
    Frame_coord0[iFrame][0] = x0;
    Frame_coord0[iFrame][1] = y0;
    Frame_coord0[iFrame][2] = z0;

    Frame_coord1[iFrame][0] = x1;
    Frame_coord1[iFrame][1] = y1;
    Frame_coord1[iFrame][2] = z1;
  }

  fclose (fp);
}



/********************************
	 Visualization
********************************/

int NodeId = 1;
int Mesh = 1;

void SetView (void)
{
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  double centerX, centerY, centerZ;
  double size;
  
  Mesh_getRange 
    (&minRangeX, &minRangeY, &minRangeZ,
     &maxRangeX, &maxRangeY, &maxRangeZ);
  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;
  centerZ = (minRangeZ + maxRangeZ) / 2;
  size = Distance (minRangeX, minRangeY, minRangeZ,
		   maxRangeX, maxRangeY, maxRangeZ) / 2;

  AutoGL_SetViewSize (size);
  AutoGL_SetViewCenter (centerX, centerY, centerZ); 
}

void DrawFrame (void)
{
  int iFrame;

  AutoGL_SetColor (1, 1, 1);
  for (iFrame = 0; iFrame < NFrames; iFrame++) {
    AutoGL_DrawLine (Frame_coord0[iFrame][0],
		     Frame_coord0[iFrame][1],
		     Frame_coord0[iFrame][2],
		     Frame_coord1[iFrame][0],
		     Frame_coord1[iFrame][1],
		     Frame_coord1[iFrame][2]);
  }
}

void RedrawView (void) 
{
  int iNode;
  int iElement;
  char buf[256];

  DrawFrame ();

  if (Mesh) {
    for (iElement = 0; iElement < NElements; iElement++) {
      int nodeId0 = Element_nodeIds[iElement][0];
      int nodeId1 = Element_nodeIds[iElement][1];
      int nodeId2 = Element_nodeIds[iElement][2];
      int nodeId3 = Element_nodeIds[iElement][3];
      
      double nodeX0 = Node_coord[nodeId0][0];
      double nodeY0 = Node_coord[nodeId0][1];
      double nodeZ0 = Node_coord[nodeId0][2];
      
      double nodeX1 = Node_coord[nodeId1][0];
      double nodeY1 = Node_coord[nodeId1][1];
      double nodeZ1 = Node_coord[nodeId1][2];
      
      double nodeX2 = Node_coord[nodeId2][0];
      double nodeY2 = Node_coord[nodeId2][1];
      double nodeZ2 = Node_coord[nodeId2][2];
      
      double nodeX3 = Node_coord[nodeId3][0];
      double nodeY3 = Node_coord[nodeId3][1];
      double nodeZ3 = Node_coord[nodeId3][2];

      AutoGL_SetColor (1, 1, 1);
      AutoGL_DrawLine (nodeX0, nodeY0, nodeZ0,
		       nodeX1, nodeY1, nodeZ1);
      AutoGL_DrawLine (nodeX0, nodeY0, nodeZ0,
		       nodeX2, nodeY2, nodeZ2);
      AutoGL_DrawLine (nodeX0, nodeY0, nodeZ0,
		       nodeX3, nodeY3, nodeZ3);
      AutoGL_DrawLine (nodeX1, nodeY1, nodeZ1,
		       nodeX2, nodeY2, nodeZ2);
      AutoGL_DrawLine (nodeX2, nodeY2, nodeZ2,
		       nodeX3, nodeY3, nodeZ3);
      AutoGL_DrawLine (nodeX3, nodeY3, nodeZ3,
		       nodeX1, nodeY1, nodeZ1);
    }
  }

  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1]; 
    double z = Node_coord[iNode][2]; 

    if (NodeId) {
      sprintf(buf, "%d", iNode);
      AutoGL_SetColor (1, 1, 0);
      AutoGL_DrawString (x, y, z,
			 buf);
    }

    switch (Node_topologyType[iNode]) {
    case NODE_TOPOLOGY_INTERNAL:
      AutoGL_SetColor (0, 0, 1);
      break;
    case NODE_TOPOLOGY_FACE:
      AutoGL_SetColor (0, 1, 1);
      break;
    case NODE_TOPOLOGY_EDGE:
      AutoGL_SetColor (0, 1, 0);
      break;
    case NODE_TOPOLOGY_VERTEX:
      AutoGL_SetColor (1, 1, 0);
      break;
    default:
      assert(0);
      break;
    }
    AutoGL_DrawDiamondMark3D (x, y, z, 
			      5, 1);
  }

}

void UpdateModel (void) 
{
  AutoGL_DrawView ();
}

void Quit (void) 
{
  AutoGL_Terminate ();
  exit (0);
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
  if (argc != 5) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tetra3d_topology_view  node_file  element_file \n"
       "     node_topology_file  frame_file \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadNodeTopologyFile (argv[3]);
  ReadFrameFile (argv[4]);

  SetView ();
  
  AutoGL_SetViewRedrawCallback (RedrawView);

  AutoGL_SetDefaultCallbackInMode3D (NULL);
  AutoGL_EnableDragInMode3D ();



  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode3D ();

  AutoGL_AddCallback (Quit, "Quit");       

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");

  AutoGL_AddBoolean (&NodeId, "NodeId");
  AutoGL_AddBoolean (&Mesh, "Mesh");

  AutoGL_AddCallback (UpdateModel, "UpdateModel");      
}
