/* advautomesh_h_tetra3d_degenerate.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}



/************************
	 Mesh
*************************/

/* Node */

typedef enum {
  NODE_TOPOLOGY_INTERNAL,
  NODE_TOPOLOGY_FACE,
  NODE_TOPOLOGY_EDGE,
  NODE_TOPOLOGY_VERTEX
} NodeTopologyType;

#define MAX_NODES 1000000

int NNodes;
int Node_isDeleted[MAX_NODES];
int Node_isReferred[MAX_NODES];
int Node_actualNodeId[MAX_NODES];
double Node_coord[MAX_NODES][3];
NodeTopologyType Node_topologyType[MAX_NODES];



/* Element */

#define N_ELEMENT_NODES 4
#define N_ELEMENT_EDGES 6
#define MAX_ELEMENTS 1000000

const int EdgeNodeMap[N_ELEMENT_EDGES][2] = {
  { 0, 1 }, 
  { 0, 2 }, 
  { 0, 3 }, 
  { 1, 2 }, 
  { 2, 3 }, 
  { 3, 1 }  
};

int NElements;
int Element_isDeleted[MAX_ELEMENTS];
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



int Element_containsNode (int elementId, int nodeId)
{
  int iElementNode;

  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(0 <= nodeId);
  assert(nodeId < NNodes);

  for (iElementNode = 0; 
       iElementNode < N_ELEMENT_NODES; 
       iElementNode++) {
    if (Element_nodeIds[elementId][iElementNode] == nodeId) {
      return 1;
    }
  }

  return 0;
}

void Mesh_deleteNode (int nodeId)
{
  assert(0 <= nodeId);
  assert(nodeId < NNodes);

  Node_isDeleted[nodeId] = 1;
}

void Mesh_deleteElement (int elementId)
{
  assert(0 <= elementId);
  assert(elementId < NElements);

  Element_isDeleted[elementId] = 1;
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    Node_isDeleted[iNode] = 0;
  }

  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    Element_isDeleted[iElement] = 0;
  }

  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void ReadNodeTopologyFile (const char *fileName)
{
  FILE *fp;
  int count;
  int iNode;

  assert(0 < NNodes);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &count);
  assert(count == NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    NodeTopologyType topologyType;
    
    fscanf (fp, "%d %d", 
	    &nodeId,
	    (int *)&topologyType); 
    assert(nodeId == iNode);
    switch (topologyType) {
    case NODE_TOPOLOGY_INTERNAL:
    case NODE_TOPOLOGY_FACE:
    case NODE_TOPOLOGY_EDGE:
    case NODE_TOPOLOGY_VERTEX:
      Node_topologyType[iNode] = topologyType;
      break;
    default:
      assert(0);
      break;
    }
  }

  fclose (fp);
}

void WriteNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;
  int nActualNodes;
  int count;

  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  nActualNodes = 0;
  for (iNode = 0; iNode < NNodes; iNode++) {
    if (Node_isDeleted[iNode]) continue;

    nActualNodes++;
  }

  fprintf (fp, "%d\n", 
	   nActualNodes);

  count = 0;
  for (iNode = 0; iNode < NNodes; iNode++) {
    if (Node_isDeleted[iNode]) continue;

    fprintf (fp, "%d   %22.15e %22.15e %22.15e\n", 
	     count,
	     Node_coord[iNode][0],
	     Node_coord[iNode][1],
	     Node_coord[iNode][2]); 
    count++;
  }

  fclose (fp);
}

void WriteElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;
  int nActualElements;
  int count;

  assert(0 < NNodes);
  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  nActualElements = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) continue;
    
    nActualElements++;
  }

  fprintf (fp, "%d\n", 
	   nActualElements);
  
  count = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) continue;

    fprintf (fp, "%d   ", 
	     count); 
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId = Element_nodeIds[iElement][iElementNode];

      fprintf (fp, "%d ", 
	       Node_actualNodeId[nodeId]); 
    }
    fprintf (fp, "\n");
    count++;
  }
  
  fclose (fp);
}

void WriteNodeTopologyFile (const char *fileName)
{
  FILE *fp;
  int iNode;
  int nActualNodes;
  int count;
  
  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  nActualNodes = 0;
  for (iNode = 0; iNode < NNodes; iNode++) {
    if (Node_isDeleted[iNode]) continue;

    nActualNodes++;
  }

  fprintf (fp, "%d\n", 
	   nActualNodes);
  
  count = 0;
  for (iNode = 0; iNode < NNodes; iNode++) {
    if (Node_isDeleted[iNode]) continue;

    fprintf (fp, "%d   %d\n", 
	     count, 
	     Node_topologyType[iNode]);
    count++;
  }

  fclose (fp);
}



/************************
	 Command
*************************/

int FindCloseNodePair 
(double minSize,
 int *nodeIdToBeDeleted_OUT, int *nodeIdToBeKept_OUT)
{
  int iElement;
  int iElementEdge;

  *nodeIdToBeDeleted_OUT = -1;
  *nodeIdToBeKept_OUT = -1;

  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) continue;

    for (iElementEdge = 0; 
	 iElementEdge < N_ELEMENT_EDGES; 
	 iElementEdge++) {
      int edgeNodeId0 
	= Element_nodeIds[iElement][EdgeNodeMap[iElementEdge][0]];
      int edgeNodeId1 
	= Element_nodeIds[iElement][EdgeNodeMap[iElementEdge][1]];

      double x0 = Node_coord[edgeNodeId0][0];
      double y0 = Node_coord[edgeNodeId0][1];
      double z0 = Node_coord[edgeNodeId0][2];

      double x1 = Node_coord[edgeNodeId1][0];
      double y1 = Node_coord[edgeNodeId1][1];
      double z1 = Node_coord[edgeNodeId1][2];
      
      if (Distance (x0, y0, z0, 
		    x1, y1, z1) < minSize) {
	if (Node_topologyType[edgeNodeId0] 
	    < Node_topologyType[edgeNodeId1]) {
	  *nodeIdToBeDeleted_OUT = edgeNodeId0;
	  *nodeIdToBeKept_OUT = edgeNodeId1;
	} else {
	  *nodeIdToBeDeleted_OUT = edgeNodeId1;
	  *nodeIdToBeKept_OUT = edgeNodeId0;
	}
	return 1;
      }
    }
  }

  return 0;
}

void ModifyMesh (int nodeIdToBeDeleted, int nodeIdToBeKept)
{
  int iElement;
  int iElementNode;

  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) continue;
    
    if (Element_containsNode (iElement, nodeIdToBeDeleted)) {
      if (Element_containsNode (iElement, nodeIdToBeKept)) {
	Mesh_deleteElement (iElement);
      } else {
	for (iElementNode = 0; 
	     iElementNode < N_ELEMENT_NODES; 
	     iElementNode++) {
	  if (Element_nodeIds[iElement][iElementNode] 
	      == nodeIdToBeDeleted) {
	    Element_nodeIds[iElement][iElementNode] 
	      = nodeIdToBeKept;
	    break;
	  }
	}
      }
    }
  }

  Mesh_deleteNode (nodeIdToBeDeleted);
}

void CheckNodeReference (void)
{
  int iNode;
  int iElement;
  int iElementNode;

  for (iNode = 0; iNode < NNodes; iNode++) {
    if (Node_isDeleted[iNode]) continue;

    Node_isReferred[iNode] = 0;
  }

  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) continue;
    
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId = Element_nodeIds[iElement][iElementNode];

      assert(!Node_isDeleted[nodeId]);
      Node_isReferred[nodeId] = 1;
    }
  }

  for (iNode = 0; iNode < NNodes; iNode++) {
    if (Node_isDeleted[iNode]) continue;

    if (!Node_isReferred[iNode]) {
      Mesh_deleteNode (iNode);
    }
  }
}

void SetActualNodeId (void)
{
  int iNode;
  int count;

  for (iNode = 0; iNode < NNodes; iNode++) {
    Node_actualNodeId[iNode] = -1;
  }

  count = 0;
  for (iNode = 0; iNode < NNodes; iNode++) {
    if (Node_isDeleted[iNode]) continue;

    Node_actualNodeId[iNode] = count;
    count++;
  }
}

void PerformCommand (double minSize)
{
  for (;;) {
    int nodeIdToBeDeleted;
    int nodeIdToBeKept;
    int found = FindCloseNodePair 
      (minSize,
       &nodeIdToBeDeleted, &nodeIdToBeKept);
    
    if (found) {
      ModifyMesh (nodeIdToBeDeleted, nodeIdToBeKept);
    } else {
      break;
    }
  }
  CheckNodeReference ();
  SetActualNodeId ();
}



int main (int argc, char *argv[])
{
  double minSize;

  if (argc != 8) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tetra3d_degenerate  node_file  element_file \n"
       "        node_topology_file  minSize \n"
       "        node_file_OUT  element_file_OUT \n"
       "        node_topology_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadNodeTopologyFile (argv[3]);

  minSize = atof (argv[4]);
  PerformCommand (minSize);

  WriteNodeFile (argv[5]);
  WriteElementFile (argv[6]);
  WriteNodeTopologyFile (argv[7]);

  return 0;
}
