/* advautomesh_h_tetra3d_domain_graded_node.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}



/************************
      Nodal Pattern
*************************/

/* Pattern */

#define MAX_PATTERNS 10000

double BasicNodalDistance;

int NPatterns;
double Pattern_center[MAX_PATTERNS][3];
double Pattern_radius[MAX_PATTERNS];
double Pattern_strength[MAX_PATTERNS];

#define COARSE_RATIO 1.2



void ReadPatternFile (const char *fileName)
{
  FILE *fp;
  int iPattern;

  assert(NPatterns == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NPatterns);
  assert(0 <= NPatterns);
  assert(NPatterns <= MAX_PATTERNS);
  
  for (iPattern = 0; iPattern < NPatterns; iPattern++) {
    int patternId;
    double x, y, z;
    double radius;
    double strength;

    fscanf (fp, "%d %lf %lf %lf %lf %lf", 
	    &patternId,
	    &x, &y, &z,
	    &radius, 
	    &strength);
    assert(patternId == iPattern);

    Pattern_center[iPattern][0] = x;
    Pattern_center[iPattern][1] = y; 
    Pattern_center[iPattern][2] = z; 

    Pattern_radius[iPattern] = radius; 
    Pattern_strength[iPattern] = strength; 
  }

  fclose (fp);
}

double NodalDistanceAt (double x, double y, double z)
{
  double strength;
  int iPattern;

  strength = 1.0;
  for (iPattern = 0; iPattern < NPatterns; iPattern++) {
    double centerX = Pattern_center[iPattern][0];
    double centerY = Pattern_center[iPattern][1]; 
    double centerZ = Pattern_center[iPattern][2]; 
    double radius = Pattern_radius[iPattern]; 
    double centerStrength = Pattern_strength[iPattern]; 
    double distance = Distance (x, y, z, 
				centerX, centerY, centerZ);
    
    if (distance < radius) {
      double strengthOfPattern 
	= (centerStrength - 1.0) / radius * (radius - distance) + 1.0;

      if (strength < strengthOfPattern) {
	strength = strengthOfPattern;
      }
    }
  }

  return BasicNodalDistance / strength * COARSE_RATIO;
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][3];



int Mesh_findNodeNear 
(double x, double y, double z, 
 double dist)
{
  int iNode;

  for (iNode = 0; iNode < NNodes; iNode++) {
    double nodeX = Node_coord[iNode][0];
    double nodeY = Node_coord[iNode][1];
    double nodeZ = Node_coord[iNode][2];

    if (Distance (x, y, z,
		  nodeX, nodeY, nodeZ) < dist) {
      return iNode;
    }
  }

  return -1;
}

int Mesh_addNewNode 
(double x, double y, double z)
{
  int nodeId = NNodes;

  assert(NNodes < MAX_NODES);
  NNodes++;
  
  Node_coord[nodeId][0] = x;
  Node_coord[nodeId][1] = y;
  Node_coord[nodeId][2] = z;
  
  return nodeId;
}

void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  *minRangeZ_OUT = *maxRangeZ_OUT = Node_coord[0][2];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    double z = Node_coord[iNode][2];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
    if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
  }
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void WriteNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    fprintf (fp, "%d   %22.15e %22.15e %22.15e\n", 
	     iNode,
	     Node_coord[iNode][0],
	     Node_coord[iNode][1],
	     Node_coord[iNode][2]); 
  }

  fclose (fp);
}



/************************
	 Command
*************************/

#define EXTENSION_RATIO 0.1
#define N_DOMAIN_DIVISIONS 10

static void PutDomainNodes (void)
{
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  double centerX, centerY, centerZ;
  double originX, originY, originZ;
  double cornerX, cornerY, cornerZ;
  double interval;
  int xDivisions, yDivisions, zDivisions;
  int ix, iy, iz;
  
  Mesh_getRange 
    (&minRangeX, &minRangeY, &minRangeZ,
     &maxRangeX, &maxRangeY, &maxRangeZ);

  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;
  centerZ = (minRangeZ + maxRangeZ) / 2;

  originX = centerX - (maxRangeX - minRangeX) * (0.5 + EXTENSION_RATIO);
  originY = centerY - (maxRangeY - minRangeY) * (0.5 + EXTENSION_RATIO);
  originZ = centerZ - (maxRangeZ - minRangeZ) * (0.5 + EXTENSION_RATIO);

  cornerX = centerX + (maxRangeX - minRangeX) * (0.5 + EXTENSION_RATIO);
  cornerY = centerY + (maxRangeY - minRangeY) * (0.5 + EXTENSION_RATIO);
  cornerZ = centerZ + (maxRangeZ - minRangeZ) * (0.5 + EXTENSION_RATIO);

  interval = BasicNodalDistance / N_DOMAIN_DIVISIONS;

  xDivisions = (int)((cornerX - originX) / interval);
  yDivisions = (int)((cornerY - originY) / interval);
  zDivisions = (int)((cornerZ - originZ) / interval);
  
  for (ix = 0; ix <= xDivisions; ix++) {
    for (iy = 0; iy <= yDivisions; iy++) {
      for (iz = 0; iz <= zDivisions; iz++) {
	double x = originX 
	  + (cornerX - originX) / xDivisions * ix;
	double y = originY 
	  + (cornerY - originY) / yDivisions * iy;
	double z = originZ 
	  + (cornerZ - originZ) / zDivisions * iz;

	double distance = NodalDistanceAt (x, y, z);

	int nodeId = Mesh_findNodeNear (x, y, z, 
					distance);
	
	if (nodeId == -1) {
	  Mesh_addNewNode (x, y, z);
	}
      }
    }
  }
}

void PerformCommand (void)
{
  PutDomainNodes ();
}



int main (int argc, char *argv[])
{
  if (argc != 5) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tetra3d_domain_graded_node \n"
       "        node_file \n"
       "        nodal_distance  pattern_file \n"
       "        node_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  
  BasicNodalDistance = atof (argv[2]);
  ReadPatternFile (argv[3]);

  PerformCommand ();

  WriteNodeFile (argv[4]);

  return 0;
}
