/* advautomesh_h_tetra3d_domain_node2.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][3];



int Mesh_findNodeNear 
(double x, double y, double z, 
 double dist)
{
  int iNode;

  for (iNode = 0; iNode < NNodes; iNode++) {
    double nodeX = Node_coord[iNode][0];
    double nodeY = Node_coord[iNode][1];
    double nodeZ = Node_coord[iNode][2];

    if (Distance (x, y, z,
		  nodeX, nodeY, nodeZ) < dist) {
      return iNode;
    }
  }

  return -1;
}

int Mesh_addNewNode 
(double x, double y, double z)
{
  int nodeId = NNodes;

  assert(NNodes < MAX_NODES);
  NNodes++;
  
  Node_coord[nodeId][0] = x;
  Node_coord[nodeId][1] = y;
  Node_coord[nodeId][2] = z;
  
  return nodeId;
}

void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  *minRangeZ_OUT = *maxRangeZ_OUT = Node_coord[0][2];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    double z = Node_coord[iNode][2];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
    if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
  }
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void WriteNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    fprintf (fp, "%d   %22.15e %22.15e %22.15e\n", 
	     iNode,
	     Node_coord[iNode][0],
	     Node_coord[iNode][1],
	     Node_coord[iNode][2]); 
  }

  fclose (fp);
}



/************************
	 Map
*************************/

#define MAX_MAP_DIVISIONS 500

static int Map_xDivisions, Map_yDivisions, Map_zDivisions;
static double Map_originX, Map_originY, Map_originZ;
static double Map_cornerX, Map_cornerY, Map_cornerZ;
static double Map_interval;
static char Map_grid[MAX_MAP_DIVISIONS][MAX_MAP_DIVISIONS][MAX_MAP_DIVISIONS];

static void Map_initialize 
(double originX, double originY, double originZ,
 double cornerX, double cornerY, double cornerZ,
 double interval,
 int xDivisions, int yDivisions, int zDivisions)
{
  int ix, iy, iz;

  assert(xDivisions < MAX_MAP_DIVISIONS);
  assert(yDivisions < MAX_MAP_DIVISIONS);
  assert(zDivisions < MAX_MAP_DIVISIONS);
  
  Map_originX = originX;
  Map_originY = originY;
  Map_originZ = originZ;

  Map_cornerX = cornerX;
  Map_cornerY = cornerY;
  Map_cornerZ = cornerZ;

  Map_interval = interval;

  Map_xDivisions = xDivisions;
  Map_yDivisions = yDivisions;
  Map_zDivisions = zDivisions;

  for (ix = 0; ix <= Map_xDivisions; ix++) {
    for (iy = 0; iy <= Map_yDivisions; iy++) {
      for (iz = 0; iz <= Map_zDivisions; iz++) {
	Map_grid[ix][iy][iz] = 0;
      }
    }
  }
}

static int Map_ix (double x) 
{
  return (int)((x - Map_originX) / Map_interval);
}

static int Map_iy (double y) 
{
  return (int)((y - Map_originY) / Map_interval);
}

static int Map_iz (double z) 
{
  return (int)((z - Map_originZ) / Map_interval);
}

static double Map_x (int ix) 
{
  return Map_originX 
    + (Map_cornerX - Map_originX) / Map_xDivisions * ix;
}

static double Map_y (int iy) 
{
  return Map_originY 
    + (Map_cornerY - Map_originY) / Map_yDivisions * iy;
}

static double Map_z (int iz) 
{
  return Map_originZ 
    + (Map_cornerZ - Map_originZ) / Map_zDivisions * iz;
}

static void Map_fill 
(int cx, int cy, int cz, 
 double radius)
{
  int r = (int)(radius / Map_interval);
  int ix, iy, iz;

  for (ix = cx - r; ix <= cx + r; ix++) {
    if (ix < 0 || Map_xDivisions < ix) continue;
    
    for (iy = cy - r; iy <= cy + r; iy++) {
      if (iy < 0 || Map_yDivisions < iy) continue;
      
      for (iz = cz - r; iz <= cz + r; iz++) {
	if (iz < 0 || Map_zDivisions < iz) continue;
    
	if ((ix - cx) * (ix - cx) 
	    + (iy - cy) * (iy - cy)
	    + (iz - cz) * (iz - cz) <= r * r) {
	  Map_grid[ix][iy][iz] = 1;
	}
      }
    }
  }
}



/************************
	 Command
*************************/

double NodalDistance;

#define EXTENSION_RATIO 0.1
#define N_DOMAIN_DIVISIONS 10

static void SetMap (void)
{
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  double centerX, centerY, centerZ;
  double originX, originY, originZ;
  double cornerX, cornerY, cornerZ;
  double interval;
  int xDivisions, yDivisions, zDivisions;

  Mesh_getRange 
    (&minRangeX, &minRangeY, &minRangeZ,
     &maxRangeX, &maxRangeY, &maxRangeZ);

  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;
  centerZ = (minRangeZ + maxRangeZ) / 2;

  originX = centerX - (maxRangeX - minRangeX) * (0.5 + EXTENSION_RATIO);
  originY = centerY - (maxRangeY - minRangeY) * (0.5 + EXTENSION_RATIO);
  originZ = centerZ - (maxRangeZ - minRangeZ) * (0.5 + EXTENSION_RATIO);

  cornerX = centerX + (maxRangeX - minRangeX) * (0.5 + EXTENSION_RATIO);
  cornerY = centerY + (maxRangeY - minRangeY) * (0.5 + EXTENSION_RATIO);
  cornerZ = centerZ + (maxRangeZ - minRangeZ) * (0.5 + EXTENSION_RATIO);

  interval = NodalDistance / N_DOMAIN_DIVISIONS;

  xDivisions = (int)((cornerX - originX) / interval);
  yDivisions = (int)((cornerY - originY) / interval);
  zDivisions = (int)((cornerZ - originZ) / interval);
  
  Map_initialize (originX, originY, originZ,
		  cornerX, cornerY, cornerZ,
		  interval,
		  xDivisions, yDivisions, zDivisions);
}

static void FillNodeOnBoundary (void)
{
  int iNode;

  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    double z = Node_coord[iNode][2];

    int ix = Map_ix (x);
    int iy = Map_iy (y);
    int iz = Map_iz (z);
    
    Map_fill (ix, iy, iz, 
	      NodalDistance);
  }
}

static void PutDomainNodes (void)
{
  int ix, iy, iz;
  
  for (ix = 0; ix <= Map_xDivisions; ix++) {
    for (iy = 0; iy <= Map_yDivisions; iy++) {
      for (iz = 0; iz <= Map_zDivisions; iz++) {
	if (!Map_grid[ix][iy][iz]) {
	  double x = Map_x (ix);
	  double y = Map_y (iy);
	  double z = Map_z (iz);
	  
	  Mesh_addNewNode (x, y, z);
	  Map_fill (ix, iy, iz, 
		    NodalDistance);
	}
      }
    }
  }
}

void PerformCommand (void)
{
  SetMap ();
  FillNodeOnBoundary ();
  PutDomainNodes ();
}



int main (int argc, char *argv[])
{
  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tetra3d_domain_node2 \n"
       "        node_file \n"
       "        nodal_distance \n"
       "        node_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  
  NodalDistance = atof (argv[2]);
  PerformCommand ();

  WriteNodeFile (argv[3]);

  return 0;
}
