/* advautomesh_h_tetra3d_face2node_vector_bc.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

void GetVectorProduct 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  *x_OUT = y0 * z1 - z0 * y1;
  *y_OUT = z0 * x1 - x0 * z1;
  *z_OUT = x0 * y1 - y0 * x1;
}

double TriArea
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double x2, double y2, double z2)
{
  double vx, vy, vz;
  
  GetVectorProduct (x1 - x0, y1 - y0, z1 - z0,
		    x2 - x0, y2 - y0, z2 - z0,
		    &vx, &vy, &vz);
  return sqrt (vx * vx + vy * vy + vz * vz) / 2;
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][3];
int Node_hasValue[MAX_NODES];
double Node_vector[MAX_NODES][3];



/* Element */

#define N_ELEMENT_NODES 4
#define N_ELEMENT_FACES 4
#define MAX_ELEMENTS 1000000

const int FaceNodeMap[N_ELEMENT_FACES][3] = {
  { 1, 2, 3 },
  { 0, 3, 2 },
  { 0, 1, 3 },
  { 0, 2, 1 }
};

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



/* Face */

#define MAX_FACES 1000000

int NFaces;
int Face_elementIds[MAX_NODES][2];
int Face_elementFaceIds[MAX_NODES][2];
int Face_hasValue[MAX_FACES];
double Face_vector[MAX_FACES][3];



double Face_area 
(int faceId)
{
  assert(0 <= faceId);
  assert(faceId < NFaces);

  {
    int elementId = Face_elementIds[faceId][0];
    int elementFaceId = Face_elementFaceIds[faceId][0];
    
    int nodeId0 
      = Element_nodeIds[elementId][FaceNodeMap[elementFaceId][0]];
    int nodeId1 
      = Element_nodeIds[elementId][FaceNodeMap[elementFaceId][1]];
    int nodeId2 
      = Element_nodeIds[elementId][FaceNodeMap[elementFaceId][2]];
    
    double x0 = Node_coord[nodeId0][0];
    double y0 = Node_coord[nodeId0][1];
    double z0 = Node_coord[nodeId0][2];
    
    double x1 = Node_coord[nodeId1][0];
    double y1 = Node_coord[nodeId1][1];
    double z1 = Node_coord[nodeId1][2];
    
    double x2 = Node_coord[nodeId2][0];
    double y2 = Node_coord[nodeId2][1];
    double z2 = Node_coord[nodeId2][2];
    
    return TriArea (x0, y0, z0,
		    x1, y1, z1, 
		    x2, y2, z2);
  }
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void ReadFaceFile (const char *fileName)
{
  FILE *fp;
  int iFace;
  
  assert(NFaces == 0);
  assert(0 < NNodes);
  assert(0 < NElements);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NFaces);
  assert(0 < NFaces);
  assert(NFaces <= MAX_FACES);
  
  for (iFace = 0; iFace < NFaces; iFace++) {
    int faceId;
    int elementId0, elementId1;
    int elementFaceId0, elementFaceId1;

    fscanf (fp, "%d %d %d %d %d", 
	    &faceId,
	    &elementId0, &elementFaceId0,
	    &elementId1, &elementFaceId1);
    assert(faceId == iFace);

    assert(elementId0 != -1);

    assert(0 <= elementId0);
    assert(elementId0 < NElements);
    assert(0 <= elementFaceId0);
    assert(elementFaceId0 < N_ELEMENT_FACES);

    if (0 <= elementId1) {
      assert(elementId1 < NElements);
      assert(0 <= elementFaceId1);
      assert(elementFaceId1 < N_ELEMENT_FACES);
    } else {
      assert(elementId1 == -1);
      assert(elementFaceId1 == -1);
    }

    Face_elementIds[iFace][0] = elementId0;
    Face_elementIds[iFace][1] = elementId1;

    Face_elementFaceIds[iFace][0] = elementFaceId0;
    Face_elementFaceIds[iFace][1] = elementFaceId1;
  }

  fclose (fp);
}

void ReadFaceVectorBcFile (const char *fileName)
{
  FILE *fp;
  int iFace;
  int nItems;
  int iItem;

  assert(0 < NFaces);

  for (iFace = 0; iFace < NFaces; iFace++) {
    Face_hasValue[iFace] = 0;
    Face_vector[iFace][0] = 0.0;
    Face_vector[iFace][1] = 0.0;
    Face_vector[iFace][2] = 0.0;
  }

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &nItems);
  assert(0 < nItems);
  
  for (iItem = 0; iItem < nItems; iItem++) {
    int itemId;
    int faceId;
    double valueX, valueY, valueZ;
    
    fscanf (fp, "%d %d %lf %lf %lf",
	    &itemId, 
	    &faceId, &valueX, &valueY, &valueZ);
    assert(itemId == iItem);
    assert(0 <= faceId);
    assert(faceId < NFaces);
    assert(Face_elementIds[faceId][1] == -1);

    Face_hasValue[faceId] = 1;
    Face_vector[faceId][0] = valueX;
    Face_vector[faceId][1] = valueY;
    Face_vector[faceId][2] = valueZ;
  }

  fclose (fp);
}

void WriteNodeVectorBcFile (const char *fileName)
{
  FILE *fp;
  int iNode;
  int count;

  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  count = 0;
  for (iNode = 0; iNode < NNodes; iNode++) {
    if (Node_hasValue[iNode]) {
      count++;
    }
  }
  
  fprintf (fp, "%d\n", 
	   count);
  
  count = 0;
  for (iNode = 0; iNode < NNodes; iNode++) {
    if (Node_hasValue[iNode]) {
      fprintf (fp, "%d   %d %f %f %f\n", 
	       count,
	       iNode, 
	       Node_vector[iNode][0], 
	       Node_vector[iNode][1], 
	       Node_vector[iNode][2]);
      count++;
    }
  }

  fclose (fp);
}



/************************
	 Command
*************************/

void PerformCommand (void)
{
  int iNode;
  int iFace;

  for (iNode = 0; iNode < NNodes; iNode++) {
    Node_hasValue[iNode] = 0;
    Node_vector[iNode][0] = 0.0;
    Node_vector[iNode][1] = 0.0;
    Node_vector[iNode][2] = 0.0;
  }

  for (iFace = 0; iFace < NFaces; iFace++) {
    if (Face_elementIds[iFace][1] == -1) {
      int elementId = Face_elementIds[iFace][0];
      int elementFaceId = Face_elementFaceIds[iFace][0];
      
      int nodeId0 
	= Element_nodeIds[elementId][FaceNodeMap[elementFaceId][0]];
      int nodeId1 
	= Element_nodeIds[elementId][FaceNodeMap[elementFaceId][1]];
      int nodeId2 
	= Element_nodeIds[elementId][FaceNodeMap[elementFaceId][2]];

      if (Face_hasValue[iFace]) {
	double valueX = Face_vector[iFace][0];
	double valueY = Face_vector[iFace][1];
	double valueZ = Face_vector[iFace][2];
	
	double area = Face_area (iFace);
	
	double nodalValueX = valueX * area / 3;
	double nodalValueY = valueY * area / 3;
	double nodalValueZ = valueZ * area / 3;
	
	Node_hasValue[nodeId0] = 1;
	Node_vector[nodeId0][0] += nodalValueX;
	Node_vector[nodeId0][1] += nodalValueY;
	Node_vector[nodeId0][2] += nodalValueZ;
	
	Node_hasValue[nodeId1] = 1;
	Node_vector[nodeId1][0] += nodalValueX;
	Node_vector[nodeId1][1] += nodalValueY;
	Node_vector[nodeId1][2] += nodalValueZ;
	
	Node_hasValue[nodeId2] = 1;
	Node_vector[nodeId2][0] += nodalValueX;
	Node_vector[nodeId2][1] += nodalValueY;
	Node_vector[nodeId2][2] += nodalValueZ;
      }
    }
  }
}



int main (int argc, char *argv[])
{
  if (argc != 6) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tetra3d_face2node_vector_bc  node_file  element_file \n"
       "        face_file  face_vector_bc_file \n"
       "        node_vector_bc_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadFaceFile (argv[3]);
  ReadFaceVectorBcFile (argv[4]);

  PerformCommand ();

  WriteNodeVectorBcFile (argv[5]);

  return 0;
}


