/* advautomesh_h_tetra3d_gm_face_node.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



#if 1
#define PRINT_FACE_ID
#endif



/************************
	 Basic
*************************/

double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}

void GetNormalizedVector 
(double x, double y, double z,
 double *nx_OUT, double *ny_OUT, double *nz_OUT)
{
  double length = sqrt (x * x + y * y + z * z);

  *nx_OUT = x / length;
  *ny_OUT = y / length;
  *nz_OUT = z / length;
}

void GetVectorProduct 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  *x_OUT = y0 * z1 - z0 * y1;
  *y_OUT = z0 * x1 - x0 * z1;
  *z_OUT = x0 * y1 - y0 * x1;
}

void GetPlaneEquation 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double x2, double y2, double z2,
 double *a_OUT, double *b_OUT, double *c_OUT, double *d_OUT)
{
  double nx, ny, nz;

  GetVectorProduct (x1 - x0, y1 - y0, z1 - z0,
		    x2 - x0, y2 - y0, z2 - z0,
		    &nx, &ny, &nz);
  GetNormalizedVector (nx, ny, nz,
		       &nx, &ny, &nz);
  *a_OUT = nx;
  *b_OUT = ny;
  *c_OUT = nz;
  *d_OUT = -(nx * x0 + ny * y0 + nz * z0);
}

void GetAnyPerpendicularDirection
(double x, double y, double z,
 double *nx0_OUT, double *ny0_OUT, double *nz0_OUT,
 double *nx1_OUT, double *ny1_OUT, double *nz1_OUT)
{
  double length;

  GetVectorProduct (x, y, z,
		    1, 0, 0,
		    nx0_OUT, ny0_OUT, nz0_OUT);
  length = sqrt (*nx0_OUT * *nx0_OUT
		 + *ny0_OUT * *ny0_OUT
		 + *nz0_OUT * *nz0_OUT);
  if (length == 0.0) {
    GetVectorProduct (x, y, z,
		      0, 1, 0,
		      nx0_OUT, ny0_OUT, nz0_OUT);
  }
  GetNormalizedVector (*nx0_OUT, *ny0_OUT, *nz0_OUT,
		       nx0_OUT, ny0_OUT, nz0_OUT);
  GetVectorProduct (x, y, z,
		    *nx0_OUT, *ny0_OUT, *nz0_OUT,
		    nx1_OUT, ny1_OUT, nz1_OUT);
  GetNormalizedVector (*nx1_OUT, *ny1_OUT, *nz1_OUT,
		       nx1_OUT, ny1_OUT, nz1_OUT);
}

void GetProjectedPointOnPlane
(double a, double b, double c, double d, 
 double x, double y, double z,
 double *px_OUT, double *py_OUT, double *pz_OUT)
{
  double value = a * x + b * y + c * z + d;
  
  *px_OUT = x - a * value;
  *py_OUT = y - b * value;
  *pz_OUT = z - c * value;
}



/********************************
         Face Data
********************************/

/* Point */

#define MAX_POINTS 10000

int NPoints;
double Point_coord[MAX_POINTS][3];



/* Edge */

#define MAX_EDGES 10000

int NEdges;
int Edge_pointId0[MAX_EDGES];
int Edge_pointId1[MAX_EDGES];



/* Face */

#define MAX_FACES 3000
#define MAX_FACE_LOOPS 100
#define MAX_FACE_VERTEXS 100

int NFaces;
int Face_nLoops[MAX_FACES];
int Face_nVertexs[MAX_FACES][MAX_FACE_LOOPS];
int Face_pointIds[MAX_FACES][MAX_FACE_LOOPS][MAX_FACE_VERTEXS];
int Face_edgeIds[MAX_FACES][MAX_FACE_LOOPS][MAX_FACE_VERTEXS];



void ReadPointFile (const char *fileName)
{
  FILE *fp;
  int iPoint;

  assert(NPoints == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NPoints);
  assert(0 < NPoints);
  assert(NPoints <= MAX_POINTS);
  
  for (iPoint = 0; iPoint < NPoints; iPoint++) {
    int pointId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &pointId,
	    &x, &y, &z);
    assert(pointId == iPoint);

    Point_coord[iPoint][0] = x;
    Point_coord[iPoint][1] = y; 
    Point_coord[iPoint][2] = z; 
  }

  fclose (fp);
}

void ReadEdgeFile (const char *fileName)
{
  FILE *fp;
  int iEdge;
  
  assert(0 < NPoints);
  assert(NEdges == 0);
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NEdges);
  assert(0 < NEdges);
  assert(NEdges <= MAX_EDGES);
  
  for (iEdge = 0; iEdge < NEdges; iEdge++) {
    int edgeId;
    int pointId0, pointId1;

    fscanf (fp, "%d %d %d", 
	    &edgeId,
	    &pointId0, &pointId1);
    assert(edgeId == iEdge);
    assert(0 <= pointId0);
    assert(pointId0 < NPoints);
    assert(0 <= pointId1);
    assert(pointId1 < NPoints);

    Edge_pointId0[iEdge] = pointId0;
    Edge_pointId1[iEdge] = pointId1;
  }

  fclose (fp);
}

void ReadFaceFile (const char *fileName)
{
  FILE *fp;
  int iFace;
  int iLoop;
  int iVertex;

  assert(0 < NPoints);
  assert(0 < NEdges);
  assert(NFaces == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NFaces);
  assert(0 < NFaces);
  assert(NFaces <= MAX_FACES);
  
  for (iFace = 0; iFace < NFaces; iFace++) {
    int faceId;
    int nLoops;
    
    fscanf (fp, "%d %d", 
	    &faceId,
	    &nLoops);
    assert(faceId == iFace);
    assert(0 < nLoops);
    
    Face_nLoops[iFace] = nLoops;
    
    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs;

      fscanf (fp, "%d", 
	      &nVertexs);
      assert(2 < nVertexs);
      
      Face_nVertexs[iFace][iLoop] = nVertexs;
      
      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	int pointId;
	
	fscanf (fp, "%d", 
		&pointId);
	assert(0 <= pointId);
	assert(pointId < NPoints);
	
	Face_pointIds[iFace][iLoop][iVertex] = pointId;
      }

      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	int edgeId;

	fscanf (fp, "%d", 
		&edgeId);
	assert(0 <= edgeId);
	assert(edgeId < NEdges);
	
	Face_edgeIds[iFace][iLoop][iVertex] = edgeId;
      }

      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	int v0 = iVertex;
	int v1 = (iVertex + 1) % nVertexs;

	int p0 = Face_pointIds[iFace][iLoop][v0];
	int p1 = Face_pointIds[iFace][iLoop][v1];

	int edgeId = Face_edgeIds[iFace][iLoop][iVertex];
	int ep0 = Edge_pointId0[edgeId];
	int ep1 = Edge_pointId1[edgeId];

	assert((p0 == ep0 && p1 == ep1)
	       || (p0 == ep1 && p1 == ep0));
      }
    }
  }

  fclose (fp);
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

typedef enum {
  NODE_TOPOLOGY_UNKNOWN,
  NODE_TOPOLOGY_OUTSIDE,
  NODE_TOPOLOGY_INSIDE,
  NODE_TOPOLOGY_VERTEX,
  NODE_TOPOLOGY_EDGE,
  NODE_TOPOLOGY_FACE
} NodeTopologyType;

int NNodes;
double Node_coord[MAX_NODES][3];
NodeTopologyType Node_topologyType[MAX_NODES];
int Node_topologyId[MAX_NODES];



int Mesh_addNewNode 
(double x, double y, double z,
 NodeTopologyType topologyType,
 int topologyId)
{
  int nodeId = NNodes;

  assert(NNodes < MAX_NODES);
  NNodes++;

  Node_coord[nodeId][0] = x;
  Node_coord[nodeId][1] = y;
  Node_coord[nodeId][2] = z;
  
  Node_topologyType[nodeId] = topologyType;
  Node_topologyId[nodeId] = topologyId;

  return nodeId;
}

void WriteNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    fprintf (fp, "%d   %22.15e %22.15e %22.15e\n", 
	     iNode,
	     Node_coord[iNode][0],
	     Node_coord[iNode][1],
	     Node_coord[iNode][2]); 
  }

  fclose (fp);
}



/************************
	 Command
*************************/

double NodalDistance;

static void MakeNodesOnVertexs (void)
{
  int iPoint;

  for (iPoint = 0; iPoint < NPoints; iPoint++) {
    double x = Point_coord[iPoint][0];
    double y = Point_coord[iPoint][1];
    double z = Point_coord[iPoint][2];
    
    Mesh_addNewNode 
      (x, y, z,
       NODE_TOPOLOGY_VERTEX, iPoint);
  }
}

static void MakeNodesOnEdges (void)
{
  int iEdge;
  int iNode;

  for (iEdge = 0; iEdge < NEdges; iEdge++) {
    int pointId0 = Edge_pointId0[iEdge];
    int pointId1 = Edge_pointId1[iEdge];
    
    double x0 = Point_coord[pointId0][0];
    double y0 = Point_coord[pointId0][1];
    double z0 = Point_coord[pointId0][2];
    
    double x1 = Point_coord[pointId1][0];
    double y1 = Point_coord[pointId1][1];
    double z1 = Point_coord[pointId1][2];

    double length = Distance (x0, y0, z0,
			      x1, y1, z1);
    int nNodes = (int)(length / NodalDistance) + 2;
    
    for (iNode = 1; iNode < nNodes; iNode++) {
      double x = x0 + (x1 - x0) / nNodes * iNode;
      double y = y0 + (y1 - y0) / nNodes * iNode;
      double z = z0 + (z1 - z0) / nNodes * iNode;

      Mesh_addNewNode 
	(x, y, z,
	 NODE_TOPOLOGY_EDGE, iEdge);
    }
  }
}



static int CurrentFaceId;

static double Plane_a, Plane_b, Plane_c, Plane_d;
static double Plane_ox, Plane_oy, Plane_oz;
static double Plane_ux, Plane_uy, Plane_uz;
static double Plane_vx, Plane_vy, Plane_vz;

#define MAX_FACE_NODES 100000
static int NFaceNodes;
static int FaceNode_nodeId[MAX_FACE_NODES];

static void SetUVCoordinateOnFace (void)
{
  const int outerLoopId = 0;
  int nVertexs = Face_nVertexs[CurrentFaceId][outerLoopId];
  int pointId0, pointId1, pointId2;
  double x0, y0, z0;
  double x1, y1, z1;
  double x2, y2, z2;

  pointId0 = Face_pointIds[CurrentFaceId][outerLoopId][0];
  pointId1 = Face_pointIds[CurrentFaceId][outerLoopId][1];
  pointId2 = Face_pointIds[CurrentFaceId][outerLoopId][nVertexs - 1];

  x0 = Point_coord[pointId0][0];
  y0 = Point_coord[pointId0][1];
  z0 = Point_coord[pointId0][2];
  
  x1 = Point_coord[pointId1][0];
  y1 = Point_coord[pointId1][1];
  z1 = Point_coord[pointId1][2];

  x2 = Point_coord[pointId2][0];
  y2 = Point_coord[pointId2][1];
  z2 = Point_coord[pointId2][2];

  GetPlaneEquation (x0, y0, z0,
		    x1, y1, z1,
		    x2, y2, z2,
		    &Plane_a, &Plane_b, &Plane_c, &Plane_d);
  GetProjectedPointOnPlane
    (Plane_a, Plane_b, Plane_c, Plane_d,
     x0, y0, z0, 
     &Plane_ox, &Plane_oy, &Plane_oz);
  GetAnyPerpendicularDirection
    (Plane_a, Plane_b, Plane_c,
     &Plane_ux, &Plane_uy, &Plane_uz,
     &Plane_vx, &Plane_vy, &Plane_vz);
  
}

static void InitializeMeshGenerationOnFace (int faceId)
{
  CurrentFaceId = faceId;
  NFaceNodes = 0;

  SetUVCoordinateOnFace ();
}

static void AddNodeOnFace (int nodeId)
{
  assert(NFaceNodes < MAX_FACE_NODES); 
  
  FaceNode_nodeId[NFaceNodes] = nodeId;
  NFaceNodes++;
}

static void GetUVOfCoordinateOnFace 
(double x, double y, double z,
 double *u_OUT, double *v_OUT)
{
  double u 
    = (x - Plane_ox) * Plane_ux
    + (y - Plane_oy) * Plane_uy
    + (z - Plane_oz) * Plane_uz;
  double v 
    = (x - Plane_ox) * Plane_vx
    + (y - Plane_oy) * Plane_vy
    + (z - Plane_oz) * Plane_vz;

  *u_OUT = u;
  *v_OUT = v;
}

static void GetCoordinateOfUVOnFace 
(double u, double v,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  double x = Plane_ox + Plane_ux * u + Plane_vx * v;
  double y = Plane_oy + Plane_uy * u + Plane_vy * v;
  double z = Plane_oz + Plane_uz * u + Plane_vz * v;

  *x_OUT = x;
  *y_OUT = y;
  *z_OUT = z;
}



#define GEOMETRY_BASE_NAME "___gm_face_node__geometry"
#define MESH_BASE_NAME "___gm_face_node__mesh"
#define BOUNDARY_MESH_BASE_NAME "___gm_face_node__boundary"

#define MESH2D_COMMAND "advautomesh_h_tri2d_delaunay_domain_mesh.sh"

static void CollectVertexNodesOnFace (void)
{
  int nLoops = Face_nLoops[CurrentFaceId];
  int iLoop;
  int iVertex;
  int iNode;

  for (iLoop = 0; iLoop < nLoops; iLoop++) {
    int nVertexs = Face_nVertexs[CurrentFaceId][iLoop];
    
    for (iVertex = 0; iVertex < nVertexs; iVertex++) {
      int pointId = Face_pointIds[CurrentFaceId][iLoop][iVertex];

      for (iNode = 0; iNode < NNodes; iNode++) {
	if (Node_topologyType[iNode] == NODE_TOPOLOGY_VERTEX
	    && Node_topologyId[iNode] == pointId) {
	  AddNodeOnFace (iNode);
	}
      }
    }
  }
}

static void CollectEdgeNodesOnFace (void)
{
  int nLoops = Face_nLoops[CurrentFaceId];
  int iLoop;
  int iVertex;
  int iNode;

  for (iLoop = 0; iLoop < nLoops; iLoop++) {
    int nVertexs = Face_nVertexs[CurrentFaceId][iLoop];
    
    for (iVertex = 0; iVertex < nVertexs; iVertex++) {
      int edgeId = Face_edgeIds[CurrentFaceId][iLoop][iVertex];
      
      for (iNode = 0; iNode < NNodes; iNode++) {
	if (Node_topologyType[iNode] == NODE_TOPOLOGY_EDGE
	    && Node_topologyId[iNode] == edgeId) {
	  AddNodeOnFace (iNode);
	}
      }
    }
  }
}

static void WriteBoundaryNodeFileOnFace (void)
{
  char FileName[256];
  FILE *fp;
  int pointId;
  int iFaceNode;
  
  sprintf (FileName, "%s%d.node",
	   BOUNDARY_MESH_BASE_NAME, CurrentFaceId);
  
  fp = fopen (FileName, "w");
  assert(fp != NULL);
  
  fprintf (fp, "%d\n", NFaceNodes);

  pointId = 0;
  for (iFaceNode = 0; iFaceNode < NFaceNodes; iFaceNode++) {
    int nodeId = FaceNode_nodeId[iFaceNode];
    
    double x = Node_coord[nodeId][0];
    double y = Node_coord[nodeId][1];
    double z = Node_coord[nodeId][2];
    
    double u, v;

    GetUVOfCoordinateOnFace (x, y, z,
			     &u, &v);
    
    fprintf (fp, "%d   %22.15e %22.15e\n", 
	     pointId,
	     u, v);
    pointId++;
  }

  fclose (fp);
}

static void WritePointFileOnFace (void)
{
  int nLoops = Face_nLoops[CurrentFaceId];
  int iLoop;
  int iVertex;
  char FileName[256];
  FILE *fp;
  int pointCount;
  int pointId;
  
  sprintf (FileName, "%s%d.gm_point",
	   GEOMETRY_BASE_NAME, CurrentFaceId);

  fp = fopen (FileName, "w");
  assert(fp != NULL);
  
  pointCount = 0;
  for (iLoop = 0; iLoop < nLoops; iLoop++) {
    int nVertexs = Face_nVertexs[CurrentFaceId][iLoop];
    
    pointCount += nVertexs;
  }

  fprintf (fp, "%d\n", pointCount);

  pointId = 0;
  for (iLoop = 0; iLoop < nLoops; iLoop++) {
    int nVertexs = Face_nVertexs[CurrentFaceId][iLoop];
    
    for (iVertex = 0; iVertex < nVertexs; iVertex++) {
      int pointId = Face_pointIds[CurrentFaceId][iLoop][iVertex];
      
      double x = Point_coord[pointId][0];
      double y = Point_coord[pointId][1];
      double z = Point_coord[pointId][2];
      
      double u, v;

      GetUVOfCoordinateOnFace (x, y, z,
			       &u, &v);
      
      fprintf (fp, "%d   %22.15e %22.15e\n", 
	       pointId,
	       u, v);
      pointId++;
    }
  }
  
  fclose (fp);
}

static void WriteSegFileOnFace (void)
{
  int nLoops = Face_nLoops[CurrentFaceId];
  int iLoop;
  int iVertex;
  char FileName[256];
  FILE *fp;
  int segCount;
  int segId;

  sprintf (FileName, "%s%d.gm_seg",
	   GEOMETRY_BASE_NAME, CurrentFaceId);

  fp = fopen (FileName, "w");
  assert(fp != NULL);

  segCount = 0;
  for (iLoop = 0; iLoop < nLoops; iLoop++) {
    int nVertexs = Face_nVertexs[CurrentFaceId][iLoop];
    
    segCount += nVertexs;
  }

  fprintf (fp, "%d\n", segCount);

  segId = 0;
  for (iLoop = 0; iLoop < nLoops; iLoop++) {
    int nVertexs = Face_nVertexs[CurrentFaceId][iLoop];
    
    for (iVertex = 0; iVertex < nVertexs; iVertex++) {
      int edgeId = Face_edgeIds[CurrentFaceId][iLoop][iVertex];
      
      int pointId0 = Edge_pointId0[edgeId];
      int pointId1 = Edge_pointId1[edgeId];
      
      double x0 = Point_coord[pointId0][0];
      double y0 = Point_coord[pointId0][1];
      double z0 = Point_coord[pointId0][2];
      
      double x1 = Point_coord[pointId1][0];
      double y1 = Point_coord[pointId1][1];
      double z1 = Point_coord[pointId1][2];

      double u0, v0;
      double u1, v1;

      GetUVOfCoordinateOnFace (x0, y0, z0,
			       &u0, &v0);
      GetUVOfCoordinateOnFace (x1, y1, z1,
			       &u1, &v1);

      fprintf (fp, "%d   %22.15e %22.15e   %22.15e %22.15e\n", 
	       segId,
	       u0, v0, 
	       u1, v1);

      segId++;
    }
  }
  
  fclose (fp);
}

static void GenerateMeshOnFace (void)
{
  char buf[2048];
  char geometryBaseName[256];
  char meshBaseName[256];
  char boundaryMeshBaseName[256];
  
  sprintf (geometryBaseName, "%s%d",
	   GEOMETRY_BASE_NAME, CurrentFaceId);
  sprintf (meshBaseName, "%s%d",
	   MESH_BASE_NAME, CurrentFaceId);
  sprintf (boundaryMeshBaseName, "%s%d",
	   BOUNDARY_MESH_BASE_NAME, CurrentFaceId);

  sprintf (buf, "%s %s %s %22.15e %s",
	   MESH2D_COMMAND,
	   geometryBaseName,
	   boundaryMeshBaseName,
	   NodalDistance,
	   meshBaseName);
  system (buf);
}

static void ReadNodeFileOnFace (void)
{
  int nBoundaryNodes = NFaceNodes;
  char fileName[256];
  FILE *fp;
  int nFaceNodes;
  int iFaceNode;

  sprintf (fileName, "%s%d.node",
	   MESH_BASE_NAME, CurrentFaceId);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &nFaceNodes);
  assert(0 < nFaceNodes);
  
  for (iFaceNode = 0; iFaceNode < nFaceNodes; iFaceNode++) {
    int faceNodeId;
    double u, v;
    double x, y, z;

    fscanf (fp, "%d %lf %lf", 
	    &faceNodeId, 
	    &u, &v);
    assert(faceNodeId == iFaceNode);

    GetCoordinateOfUVOnFace (u, v,
			     &x, &y, &z);
    
    if (nBoundaryNodes <= faceNodeId) {
      int nodeId;
      
      nodeId = Mesh_addNewNode 
	(x, y, z,
	 NODE_TOPOLOGY_FACE, CurrentFaceId);
      AddNodeOnFace (nodeId);
    }
  }

  fclose (fp);

  assert(NFaceNodes == nFaceNodes);
}

void PerformCommand (void)
{
  int iFace;
  
  MakeNodesOnVertexs ();
  MakeNodesOnEdges ();
  
  for (iFace = 0; iFace < NFaces; iFace++) {

#ifdef PRINT_FACE_ID
    printf (" iFace %d \n",
	    iFace);
#endif

    InitializeMeshGenerationOnFace (iFace);
    CollectVertexNodesOnFace ();
    CollectEdgeNodesOnFace ();
    
    WriteBoundaryNodeFileOnFace ();
    WritePointFileOnFace ();
    WriteSegFileOnFace ();
    
    GenerateMeshOnFace ();

    ReadNodeFileOnFace ();
  }
}



int main (int argc, char *argv[])
{
  if (argc != 6) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tetra3d_gm_face_node  \n"
       "       gm_point_file  gm_edge_file  gm_face_file \n"
       "       nodal_distance \n"
       "       node_file_OUT \n");
    exit (1);
  }

  ReadPointFile (argv[1]);
  ReadEdgeFile (argv[2]);
  ReadFaceFile (argv[3]);

  NodalDistance = atof (argv[4]);
  PerformCommand ();

  WriteNodeFile (argv[5]);

  return 0;
}
