/* advautomesh_h_tetra3d_graded_refine.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001

double Eps = DEFAULT_EPS;



double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}



/************************
      Nodal Pattern
*************************/

/* Pattern */

#define MAX_PATTERNS 10000

double BasicNodalDistance;

int NPatterns;
double Pattern_center[MAX_PATTERNS][3];
double Pattern_radius[MAX_PATTERNS];
double Pattern_strength[MAX_PATTERNS];



void ReadPatternFile (const char *fileName)
{
  FILE *fp;
  int iPattern;

  assert(NPatterns == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NPatterns);
  assert(0 <= NPatterns);
  assert(NPatterns <= MAX_PATTERNS);
  
  for (iPattern = 0; iPattern < NPatterns; iPattern++) {
    int patternId;
    double x, y, z;
    double radius;
    double strength;

    fscanf (fp, "%d %lf %lf %lf %lf %lf", 
	    &patternId,
	    &x, &y, &z,
	    &radius, 
	    &strength);
    assert(patternId == iPattern);

    Pattern_center[iPattern][0] = x;
    Pattern_center[iPattern][1] = y; 
    Pattern_center[iPattern][2] = z; 

    Pattern_radius[iPattern] = radius; 
    Pattern_strength[iPattern] = strength; 
  }

  fclose (fp);
}

double MinNodalDistance (void)
{
  return BasicNodalDistance * 0.2;
}

double NodalDistanceAt (double x, double y, double z)
{
  double strength;
  int iPattern;

  strength = 1.0;
  for (iPattern = 0; iPattern < NPatterns; iPattern++) {
    double centerX = Pattern_center[iPattern][0];
    double centerY = Pattern_center[iPattern][1]; 
    double centerZ = Pattern_center[iPattern][2]; 
    double radius = Pattern_radius[iPattern]; 
    double centerStrength = Pattern_strength[iPattern]; 
    double distance = Distance (x, y, z, 
				centerX, centerY, centerZ);
    
    if (distance < radius) {
      double strengthOfPattern 
	= (centerStrength - 1.0) / radius * (radius - distance) + 1.0;

      if (strength < strengthOfPattern) {
	strength = strengthOfPattern;
      }
    }
  }

  return BasicNodalDistance / strength;
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][3];



/* Element */

#define N_ELEMENT_NODES 4
#define MAX_ELEMENTS 5000000

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];
int Element_isRefined[MAX_ELEMENTS];



void Element_getCenter 
(int elementId, 
 double *centerX_OUT, double *centerY_OUT, double *centerZ_OUT)
{
  int iElementNode;

  assert(0 <= elementId);
  assert(elementId < NElements);

  *centerX_OUT = 0.0;
  *centerY_OUT = 0.0;
  *centerZ_OUT = 0.0;
  for (iElementNode = 0; 
       iElementNode < N_ELEMENT_NODES; 
       iElementNode++) {
    int nodeId = Element_nodeIds[elementId][iElementNode];
    double x = Node_coord[nodeId][0];
    double y = Node_coord[nodeId][1];
    double z = Node_coord[nodeId][2];
    
    *centerX_OUT += x;
    *centerY_OUT += y;
    *centerZ_OUT += z;
  }
  *centerX_OUT /= N_ELEMENT_NODES;
  *centerY_OUT /= N_ELEMENT_NODES;
  *centerZ_OUT /= N_ELEMENT_NODES;
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void WriteNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    fprintf (fp, "%d   %22.15e %22.15e %22.15e\n", 
	     iNode,
	     Node_coord[iNode][0],
	     Node_coord[iNode][1],
	     Node_coord[iNode][2]); 
  }

  fclose (fp);
}

void WriteElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NElements);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    fprintf (fp, "%d   ", 
	     iElement); 
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      fprintf (fp, "%d ", 
	       Element_nodeIds[iElement][iElementNode]); 
    }
    fprintf (fp, "\n");
  }
  
  fclose (fp);
}

void WriteRefinedElementListFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int count;
  
  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  count = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isRefined[iElement]) {
      count++;
    }
  }

  fprintf (fp, "%d\n", 
	   count);

  count = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isRefined[iElement]) {
      fprintf (fp, "%d   %d\n", 
	       count, iElement);
      count++;
    }
  }

  fclose (fp);
}



/************************
	 Command
*************************/

static int MakeRefinedElementList (double ratioThreshold)
{
  int exists;
  int iElement;
  
  exists = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    Element_isRefined[iElement] = 0;
  }

  for (iElement = 0; iElement < NElements; iElement++) {
    double centerX, centerY, centerZ;
    double ratio;
    
    Element_getCenter (iElement, 
		       &centerX, &centerY, &centerZ);

    ratio = NodalDistanceAt (centerX, centerY, centerZ) / BasicNodalDistance;

    if (ratio + Eps < ratioThreshold) {
      Element_isRefined[iElement] = 1;
      exists = 1;
    }
  }

  return exists;
}

static void RemoveFile (const char *fileName)
{
  char buf[1024];

  sprintf (buf, "rm -f   %s",
	   fileName);
  system (buf);
}



#define IN_NODE_FILE_NAME         "___graded_refine__in.node"
#define IN_ELEMENT_FILE_NAME      "___graded_refine__in.element"
#define IN_EDGE_FILE_NAME         "___graded_refine__in.edge"
#define IN_ELEMENT_EDGE_FILE_NAME "___graded_refine__in.el_ed"
#define IN_LIST_FILE_NAME         "___graded_refine__in.rel"

#define OUT_NODE_FILE_NAME        "___graded_refine__out.node"
#define OUT_ELEMENT_FILE_NAME     "___graded_refine__out.element"

#define EDGE_COMMAND "advautomesh_h_tetra3d_edge2"
#define REFINE_COMMAND "advautomesh_h_tetra3d_refine"

static void RefineMesh (void)
{
  char buf[2048];

  WriteNodeFile (IN_NODE_FILE_NAME);
  WriteElementFile (IN_ELEMENT_FILE_NAME);
  WriteRefinedElementListFile (IN_LIST_FILE_NAME);
  
  sprintf (buf, "%s   %s %s   %s %s",
	   EDGE_COMMAND,
	   IN_NODE_FILE_NAME,
	   IN_ELEMENT_FILE_NAME,
	   
	   IN_EDGE_FILE_NAME,
	   IN_ELEMENT_EDGE_FILE_NAME);
  system (buf);
  
  sprintf (buf, "%s   %s %s %s %s %s   %s %s",
	   REFINE_COMMAND,
	   IN_NODE_FILE_NAME,
	   IN_ELEMENT_FILE_NAME,
	   IN_EDGE_FILE_NAME,
	   IN_ELEMENT_EDGE_FILE_NAME,
	   IN_LIST_FILE_NAME,
	   
	   OUT_NODE_FILE_NAME,
	   OUT_ELEMENT_FILE_NAME);
  system (buf);
  
  ReadNodeFile (OUT_NODE_FILE_NAME);
  ReadElementFile (OUT_ELEMENT_FILE_NAME);
  
  RemoveFile (IN_NODE_FILE_NAME);
  RemoveFile (IN_ELEMENT_FILE_NAME);
  RemoveFile (IN_EDGE_FILE_NAME);
  RemoveFile (IN_ELEMENT_EDGE_FILE_NAME);
  RemoveFile (IN_LIST_FILE_NAME);
  
  RemoveFile (OUT_NODE_FILE_NAME);
  RemoveFile (OUT_ELEMENT_FILE_NAME);
}



void PerformCommand (void)
{
  double ratioThreshold;
  
  ratioThreshold = 1.0;
  for (;;) {
    int exists = MakeRefinedElementList (ratioThreshold);
    
    if (!exists) {
      break;
    }
    
    RefineMesh ();
    
    ratioThreshold /= 2;
  }
}



int main (int argc, char *argv[])
{
  if (argc != 7) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tetra3d_graded_refine  node_file  element_file \n"
       "       nodal_distance  pattern_file \n"
       "       node_file_OUT  element_file_OUT \n");
    exit (1);
  }
  
  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);

  BasicNodalDistance = atof (argv[3]);
  ReadPatternFile (argv[4]);

  PerformCommand ();

  WriteNodeFile (argv[5]);
  WriteElementFile (argv[6]);

  return 0;
}
