/* advautomesh_h_tetra3d_isosurface.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 2500000

int NNodes;
double Node_scalar[MAX_NODES];



/* Element */

#define N_ELEMENT_NODES 4
#define N_ELEMENT_EDGES 6
#define MAX_ELEMENTS 12000000

const int EdgeNodeMap[N_ELEMENT_EDGES][2] = {
  { 0, 1 }, 
  { 0, 2 }, 
  { 0, 3 }, 
  { 1, 2 }, 
  { 2, 3 }, 
  { 3, 1 }  
};

const int NIsosurfaceEdges[8] = {
  0, 3, 3, 4, 3, 4, 4, 3
};

const int IsosurfaceEdgeIds[8][4] = {
  { -1, -1, -1, -1 },
  { 0, 1, 2, -1 },
  { 0, 3, 5, -1 },
  { 3, 1, 2, 5 },
  { 3, 1, 4, -1 },
  { 0, 3, 4, 2 },
  { 0, 5, 4, 1 },
  { 2, 5, 4, -1 }
};

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



void ReadNodeFile (const char *fileName)
{
  FILE *fp;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  /* skip coordinate */

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void ReadNodeScalarFile (const char *fileName)
{
  FILE *fp;
  int count;
  int iNode;
  
  assert(0 < NNodes);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &count);
  assert(count == NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double scalar;

    fscanf (fp, "%d %lf", 
	    &nodeId,
	    &scalar);
    assert(nodeId == iNode);
    Node_scalar[iNode] = scalar;
  }

  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Isosurface */

#define MAX_LEVELS 10000
#define MAX_ISOSURFACES 1000000

int NLevels;
int Level_startIsosurfaceId[MAX_LEVELS];
int Level_nIsosurfaces[MAX_LEVELS];
int NIsosurfaces;
int Isosurface_elementId[MAX_ISOSURFACES];
int Isosurface_elementEdgeIds[MAX_ISOSURFACES][3];
double Isosurface_ratios[MAX_ISOSURFACES][3];



void IsosurfaceGroup_addLevel (void)
{
  assert(NLevels < MAX_LEVELS);

  Level_startIsosurfaceId[NLevels] = NIsosurfaces;
  Level_nIsosurfaces[NLevels] = 0;
  NLevels++;
}

void IsosurfaceGroup_addIsosurface 
(int elementId, 
 int elementEdgeId0, double ratio0,
 int elementEdgeId1, double ratio1,
 int elementEdgeId2, double ratio2)
{
  assert(0 < NLevels);
  assert(NIsosurfaces < MAX_ISOSURFACES);
  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(0 <= elementEdgeId0);
  assert(elementEdgeId0 < N_ELEMENT_EDGES);
  assert(0 <= elementEdgeId1);
  assert(elementEdgeId1 < N_ELEMENT_EDGES);
  assert(0 <= elementEdgeId2);
  assert(elementEdgeId2 < N_ELEMENT_EDGES);

  Level_nIsosurfaces[NLevels - 1]++;

  Isosurface_elementId[NIsosurfaces] = elementId;
  Isosurface_elementEdgeIds[NIsosurfaces][0] = elementEdgeId0;
  Isosurface_elementEdgeIds[NIsosurfaces][1] = elementEdgeId1;
  Isosurface_elementEdgeIds[NIsosurfaces][2] = elementEdgeId2;
  Isosurface_ratios[NIsosurfaces][0] = ratio0;
  Isosurface_ratios[NIsosurfaces][1] = ratio1;
  Isosurface_ratios[NIsosurfaces][2] = ratio2;

  NIsosurfaces++;
}

void WriteIsosurfaceFile (const char *fileName)
{
  FILE *fp;
  int iLevel;
  int iIsosurface;
  
  assert(0 < NLevels);
  assert(0 < NIsosurfaces);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NLevels);
  
  for (iLevel = 0; iLevel < NLevels; iLevel++) {
    int startId = Level_startIsosurfaceId[iLevel];
    int nIsosurfaces = Level_nIsosurfaces[iLevel];
    
    fprintf (fp, "%d   %d\n", 
	     iLevel,
	     nIsosurfaces);

    for (iIsosurface = startId; 
	 iIsosurface < startId + nIsosurfaces; 
	 iIsosurface++) {
      fprintf (fp, "%d   %d   %d %22.15e   %d %22.15e   %d %22.15e\n", 
	       iIsosurface - startId,
	       Isosurface_elementId[iIsosurface],
	       Isosurface_elementEdgeIds[iIsosurface][0],
	       Isosurface_ratios[iIsosurface][0],
	       Isosurface_elementEdgeIds[iIsosurface][1],
	       Isosurface_ratios[iIsosurface][1],
	       Isosurface_elementEdgeIds[iIsosurface][2],
	       Isosurface_ratios[iIsosurface][2]);
    }
  }

  fclose (fp);
}



/************************
	 Command
*************************/

int IsosurfaceMask (int elementId, double levelValue)
{
  int nodeId0 = Element_nodeIds[elementId][0];
  int nodeId1 = Element_nodeIds[elementId][1];
  int nodeId2 = Element_nodeIds[elementId][2];
  int nodeId3 = Element_nodeIds[elementId][3];
  unsigned int mask = 0;
  
  if (levelValue < Node_scalar[nodeId0]) mask |= 1;
  if (levelValue < Node_scalar[nodeId1]) mask |= 2;
  if (levelValue < Node_scalar[nodeId2]) mask |= 4;
  if (levelValue < Node_scalar[nodeId3]) mask |= 8;
  if ((mask & 8) != 0) mask = (~mask) & 15;

  assert(0 <= mask);
  assert(mask < 8);

  return mask;
}

int GetIsosurface 
(int elementId, double levelValue, 
 int *nEdges,
 int elementEdgeIds_OUT[4], double ratios_OUT[4])
{
  unsigned int mask = IsosurfaceMask (elementId, levelValue);
  int iEdge;

  *nEdges = NIsosurfaceEdges[mask];
  if (*nEdges == 0) return 0;
  
  assert(*nEdges == 3 || *nEdges == 4);

  for (iEdge = 0; iEdge < *nEdges; iEdge++) {
    int elementEdgeId = IsosurfaceEdgeIds[mask][iEdge];

    int elementNodeId0 = EdgeNodeMap[elementEdgeId][0];
    int elementNodeId1 = EdgeNodeMap[elementEdgeId][1];

    int nodeId0 = Element_nodeIds[elementId][elementNodeId0];
    int nodeId1 = Element_nodeIds[elementId][elementNodeId1];

    double ratio = (levelValue - Node_scalar[nodeId0]) 
      / (Node_scalar[nodeId1] - Node_scalar[nodeId0]);
    
    elementEdgeIds_OUT[iEdge] = elementEdgeId; 
    ratios_OUT[iEdge] = ratio;
  }    

  return 1;
}

void PerformCommand 
(double minLevel, double maxLevel, int nDivisions)
{
  int iLevel;
  int iElement;

  for (iLevel = 0; iLevel <= nDivisions; iLevel++) {
    double levelValue = minLevel 
      + (maxLevel - minLevel) / nDivisions * iLevel;

    IsosurfaceGroup_addLevel ();

    for (iElement = 0; iElement < NElements; iElement++) {
      int nEdges;
      int elementEdgeIds[4];
      double ratios[4];
      
      if (GetIsosurface
	  (iElement, levelValue,
	   &nEdges, elementEdgeIds, ratios)) {
	assert(nEdges == 3 || nEdges == 4);
	IsosurfaceGroup_addIsosurface 
	  (iElement, 
	   elementEdgeIds[0], ratios[0],
	   elementEdgeIds[1], ratios[1],
	   elementEdgeIds[2], ratios[2]);
	if (nEdges == 4) {
	  IsosurfaceGroup_addIsosurface 
	    (iElement, 
	     elementEdgeIds[2], ratios[2],
	     elementEdgeIds[3], ratios[3],
	     elementEdgeIds[0], ratios[0]);
	}
      }
    }
  }
}



#if 1
#include "../../util/time_measure.c"
#define TIME_MEASURE
#endif

int main (int argc, char *argv[])
{
  double minLevel, maxLevel;
  int nDivisions;
  
  if (argc != 8) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tetra3d_isosurface  node_file  element_file \n"
       "        node_scalar_file  minLevel  maxLevel  nDivisions \n"
       "        isosurface_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadNodeScalarFile (argv[3]);
  minLevel = atof (argv[4]);
  maxLevel = atof (argv[5]);
  nDivisions = atoi (argv[6]);

#ifdef TIME_MEASURE
  StartMeasureTime ();
#endif

  PerformCommand (minLevel, maxLevel, nDivisions);

#ifdef TIME_MEASURE
  EndMeasureTime ();
#endif

  WriteIsosurfaceFile (argv[7]);

  return 0;
}
