/* advautomesh_h_tetra3d_mesh.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001

double Eps = DEFAULT_EPS;



double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][3];



/* Element */

#define N_ELEMENT_NODES 4
#define MAX_ELEMENTS 1000000

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



int Mesh_findNodeAt (double x, double y, double z)
{
  int iNode;

  for (iNode = 0; iNode < NNodes; iNode++) {
    double nodeX = Node_coord[iNode][0];
    double nodeY = Node_coord[iNode][1];
    double nodeZ = Node_coord[iNode][2];

    if (Distance (x, y, z, 
		  nodeX, nodeY, nodeZ) < Eps) {
      return iNode;
    }
  }

  return -1;
}

int Mesh_addNewNode (double x, double y, double z)
{
  int nodeId = NNodes;

  assert(NNodes < MAX_NODES);
  NNodes++;

  Node_coord[nodeId][0] = x;
  Node_coord[nodeId][1] = y;
  Node_coord[nodeId][2] = z;
  
  return nodeId;
}

int Mesh_addNewElement (int nodeId0, int nodeId1, int nodeId2, int nodeId3)
{
  int elementId = NElements;

  assert(0 <= nodeId0);
  assert(nodeId0 < NNodes);
  assert(0 <= nodeId1);
  assert(nodeId1 < NNodes);
  assert(0 <= nodeId2);
  assert(nodeId2 < NNodes);
  assert(0 <= nodeId3);
  assert(nodeId3 < NNodes);

  assert(NElements < MAX_ELEMENTS);
  NElements++;

  Element_nodeIds[elementId][0] = nodeId0;
  Element_nodeIds[elementId][1] = nodeId1;
  Element_nodeIds[elementId][2] = nodeId2;
  Element_nodeIds[elementId][3] = nodeId3;

  return elementId;
}

void WriteNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    fprintf (fp, "%d   %22.15e %22.15e %22.15e\n", 
	     iNode,
	     Node_coord[iNode][0],
	     Node_coord[iNode][1],
	     Node_coord[iNode][2]); 
  }

  fclose (fp);
}

void WriteElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NElements);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    fprintf (fp, "%d   ", 
	     iElement); 
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      fprintf (fp, "%d ", 
	       Element_nodeIds[iElement][iElementNode]); 
    }
    fprintf (fp, "\n");
  }
  
  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Tetrahedron */

#define MAX_TETRAS 1000000

int NTetras;
double Tetra_coord[MAX_TETRAS][N_ELEMENT_NODES][3];



void ReadTetraFile (const char *fileName)
{
  FILE *fp;
  int iTetra;

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NTetras);
  assert(NTetras <= MAX_TETRAS);
  
  for (iTetra = 0; iTetra < NTetras; iTetra++) {
    int tetraId;
    double x0, y0, z0;
    double x1, y1, z1;
    double x2, y2, z2;
    double x3, y3, z3;

    fscanf (fp, "%d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf", 
	    &tetraId, 
	    &x0, &y0, &z0, 
	    &x1, &y1, &z1, 
	    &x2, &y2, &z2, 
	    &x3, &y3, &z3);
    assert(tetraId == iTetra);
    assert(Eps < Distance (x0, y0, z0, 
			   x1, y1, z1));
    assert(Eps < Distance (x0, y0, z0, 
			   x2, y2, z2));
    assert(Eps < Distance (x0, y0, z0, 
			   x3, y3, z3));
    assert(Eps < Distance (x1, y1, z1, 
			   x2, y2, z2));
    assert(Eps < Distance (x2, y2, z2, 
			   x3, y3, z3));
    assert(Eps < Distance (x3, y3, z3, 
			   x1, y1, z1));

    Tetra_coord[iTetra][0][0] = x0;
    Tetra_coord[iTetra][0][1] = y0;
    Tetra_coord[iTetra][0][2] = z0;

    Tetra_coord[iTetra][1][0] = x1;
    Tetra_coord[iTetra][1][1] = y1;
    Tetra_coord[iTetra][1][2] = z1;

    Tetra_coord[iTetra][2][0] = x2;
    Tetra_coord[iTetra][2][1] = y2;
    Tetra_coord[iTetra][2][2] = z2;

    Tetra_coord[iTetra][3][0] = x3;
    Tetra_coord[iTetra][3][1] = y3;
    Tetra_coord[iTetra][3][2] = z3;
  }

  fclose (fp);
}



/************************
	 Command
*************************/

int NodeAt (double x, double y, double z)
{
  int nodeId = Mesh_findNodeAt (x, y, z);
  
  if (nodeId == -1) {
    nodeId = Mesh_addNewNode (x, y, z);
  }

  return nodeId;
}

void PerformCommand (void)
{
  int iTetra;

  for (iTetra = 0; iTetra < NTetras; iTetra++) {
    double x0 = Tetra_coord[iTetra][0][0];
    double y0 = Tetra_coord[iTetra][0][1];
    double z0 = Tetra_coord[iTetra][0][2];

    double x1 = Tetra_coord[iTetra][1][0];
    double y1 = Tetra_coord[iTetra][1][1];
    double z1 = Tetra_coord[iTetra][1][2];

    double x2 = Tetra_coord[iTetra][2][0];
    double y2 = Tetra_coord[iTetra][2][1];
    double z2 = Tetra_coord[iTetra][2][2];

    double x3 = Tetra_coord[iTetra][3][0];
    double y3 = Tetra_coord[iTetra][3][1];
    double z3 = Tetra_coord[iTetra][3][2];

    int nodeId0 = NodeAt (x0, y0, z0);
    int nodeId1 = NodeAt (x1, y1, z1);
    int nodeId2 = NodeAt (x2, y2, z2);
    int nodeId3 = NodeAt (x3, y3, z3);

    Mesh_addNewElement (nodeId0, nodeId1, nodeId2, nodeId3);
  }
}



int main (int argc, char *argv[])
{
  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tetra3d_mesh  tetra_file \n"
       "     node_file_OUT  element_file_OUT \n");
    exit (1);
  }

  ReadTetraFile (argv[1]);

  PerformCommand ();

  WriteNodeFile (argv[2]);
  WriteElementFile (argv[3]);

  return 0;
}
