/* advautomesh_h_tetra3d_refine.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][3];



/* Element */

#define N_ELEMENT_NODES 4
#define N_ELEMENT_EDGES 6
#define N_ELEMENT_FACES 4
#define MAX_ELEMENTS 1000000

const int FaceNodeMap[N_ELEMENT_FACES][3] = {
  { 1, 2, 3 },
  { 0, 3, 2 },
  { 0, 1, 3 },
  { 0, 2, 1 }
};

const int FaceEdgeMap[N_ELEMENT_FACES][3] = {
  { 3, 4, 5 },
  { 2, 4, 1 },
  { 0, 5, 2 },
  { 1, 3, 0 }
};

int NElements;
int Element_isDeleted[MAX_ELEMENTS];
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];
int Element_edgeIds[MAX_ELEMENTS][N_ELEMENT_EDGES];
int Element_isRefined[MAX_ELEMENTS];



/* Edge */

#define MAX_EDGES 1000000

int NEdges;
int Edge_nodeIds[MAX_EDGES][2];
int Edge_isDivided[MAX_EDGES];
int Edge_midNodeId[MAX_EDGES];



void Edge_getMidPoint 
(int edgeId, 
 double *midX_OUT, double *midY_OUT, double *midZ_OUT)
{
  int edgeNodeId0, edgeNodeId1;
  double x0, y0, z0;
  double x1, y1, z1;

  assert(0 <= edgeId);
  assert(edgeId < NEdges);

  edgeNodeId0 = Edge_nodeIds[edgeId][0];
  edgeNodeId1 = Edge_nodeIds[edgeId][1];
  
  x0 = Node_coord[edgeNodeId0][0];
  y0 = Node_coord[edgeNodeId0][1];
  z0 = Node_coord[edgeNodeId0][2];
  
  x1 = Node_coord[edgeNodeId1][0];
  y1 = Node_coord[edgeNodeId1][1];
  z1 = Node_coord[edgeNodeId1][2];

  *midX_OUT = (x0 + x1) / 2;
  *midY_OUT = (y0 + y1) / 2;
  *midZ_OUT = (z0 + z1) / 2;
}

int Mesh_addNewNode (double x, double y, double z)
{
  int nodeId = NNodes;

  assert(NNodes < MAX_NODES);
  NNodes++;

  Node_coord[nodeId][0] = x;
  Node_coord[nodeId][1] = y;
  Node_coord[nodeId][2] = z;
  
  return nodeId;
}

int Mesh_addNewElement (int nodeId0, int nodeId1, int nodeId2, int nodeId3)
{
  int elementId = NElements;

  assert(0 <= nodeId0);
  assert(nodeId0 < NNodes);
  assert(0 <= nodeId1);
  assert(nodeId1 < NNodes);
  assert(0 <= nodeId2);
  assert(nodeId2 < NNodes);
  assert(0 <= nodeId3);
  assert(nodeId3 < NNodes);

  assert(NElements < MAX_ELEMENTS);
  NElements++;

  Element_isDeleted[elementId] = 0;

  Element_nodeIds[elementId][0] = nodeId0;
  Element_nodeIds[elementId][1] = nodeId1;
  Element_nodeIds[elementId][2] = nodeId2;
  Element_nodeIds[elementId][3] = nodeId3;

  Element_edgeIds[elementId][0] = -1;
  Element_edgeIds[elementId][1] = -1;
  Element_edgeIds[elementId][2] = -1;
  Element_edgeIds[elementId][3] = -1;
  Element_edgeIds[elementId][4] = -1;
  Element_edgeIds[elementId][5] = -1;

  Element_isRefined[elementId] = 0;

  return elementId;
}

void Mesh_deleteElement (int elementId)
{
  assert(0 <= elementId);
  assert(elementId < NElements);

  Element_isDeleted[elementId] = 1;
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    Element_isDeleted[iElement] = 0;
  }

  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void ReadEdgeVertexFile (const char *fileName)
{
  FILE *fp;
  int iEdge;
  
  assert(0 < NNodes);
  assert(0 < NElements);
  assert(NEdges == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NEdges);
  assert(NEdges <= MAX_EDGES);
  
  for (iEdge = 0; iEdge < NEdges; iEdge++) {
    int edgeId;
    int nodeId0, nodeId1;
    
    fscanf (fp, "%d %d %d", 
	    &edgeId,
	    &nodeId0, &nodeId1);
    assert(edgeId == iEdge);
    assert(0 <= nodeId0);
    assert(nodeId0 < NNodes);
    assert(0 <= nodeId1);
    assert(nodeId1 < NNodes);
    Edge_nodeIds[iEdge][0] = nodeId0;
    Edge_nodeIds[iEdge][1] = nodeId1;
  }

  fclose (fp);
}

void ReadElementEdgeFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementEdge;
  int count;
  
  assert(0 < NElements);
  assert(0 < NEdges);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &count);
  assert(count == NElements);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId);
    assert(elementId == iElement);
    
    for (iElementEdge = 0; 
	 iElementEdge < N_ELEMENT_EDGES; 
	 iElementEdge++) {
      int edgeId;

      fscanf (fp, "%d", 
	      &edgeId);
      assert(0 <= edgeId);
      assert(edgeId < NEdges);
      Element_edgeIds[iElement][iElementEdge] = edgeId;
    }
  }

  fclose (fp);
}

void ReadRefinedElementListFile (const char *fileName)
{
  FILE *fp;
  int nItems;
  int iItem;
  int iElement;
  
  assert(0 < NElements);

  for (iElement = 0; iElement < NElements; iElement++) {
    Element_isRefined[iElement] = 0;
  }

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &nItems);
  assert(0 < nItems);
  
  for (iItem = 0; iItem < nItems; iItem++) {
    int itemId;
    int elementId;

    fscanf (fp, "%d %d", 
	    &itemId,
	    &elementId);
    assert(itemId == iItem);
    assert(0 <= elementId);
    assert(elementId < NElements);
    Element_isRefined[elementId] = 1;
  }

  fclose (fp);
}

void WriteNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    fprintf (fp, "%d   %22.15e %22.15e %22.15e\n", 
	     iNode,
	     Node_coord[iNode][0],
	     Node_coord[iNode][1],
	     Node_coord[iNode][2]); 
  }

  fclose (fp);
}

void WriteElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;
  int nActualElements;
  int count;

  assert(0 < NNodes);
  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  nActualElements = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) continue;
    
    nActualElements++;
  }

  fprintf (fp, "%d\n", 
	   nActualElements);
  
  count = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) continue;

    fprintf (fp, "%d   ", 
	     count); 
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      fprintf (fp, "%d ", 
	       Element_nodeIds[iElement][iElementNode]); 
    }
    fprintf (fp, "\n");
    count++;
  }
  
  fclose (fp);
}



/************************
	 Command
*************************/

double DistanceBetweenNodes (int nodeId0, int nodeId1)
{
  return Distance 
    (Node_coord[nodeId0][0], 
     Node_coord[nodeId0][1],
     Node_coord[nodeId0][2],
     Node_coord[nodeId1][0], 
     Node_coord[nodeId1][1],
     Node_coord[nodeId1][2]);
}

int MissingEdgeOfPartiallyDividedElementFace (int elementId)
{
  int iElementFace;
  
  for (iElementFace = 0; 
       iElementFace < N_ELEMENT_FACES; 
       iElementFace++) {
    int edgeId0 = Element_edgeIds[elementId][FaceEdgeMap[iElementFace][0]];
    int edgeId1 = Element_edgeIds[elementId][FaceEdgeMap[iElementFace][1]];
    int edgeId2 = Element_edgeIds[elementId][FaceEdgeMap[iElementFace][2]];
    
    if (Edge_isDivided[edgeId0] && Edge_isDivided[edgeId1]) {
      return edgeId2;
    } else if (Edge_isDivided[edgeId1] && Edge_isDivided[edgeId2]) {
      return edgeId0;
    } else if (Edge_isDivided[edgeId2] && Edge_isDivided[edgeId0]) {
      return edgeId1;
    }
  }
  return -1;
}

int ElementEdgeOfDividedElement (int elementId)
{
  int iElementEdge;
  
  for (iElementEdge = 0; 
       iElementEdge < N_ELEMENT_EDGES; 
       iElementEdge++) {
    int edgeId = Element_edgeIds[elementId][iElementEdge];
    
    if (Edge_isDivided[edgeId]) return iElementEdge;
  }
  return -1;
}

int ElementFaceOfDividedElement (int elementId)
{
  int iElementFace;
  
  for (iElementFace = 0; 
       iElementFace < N_ELEMENT_FACES; 
       iElementFace++) {
    int edgeId0 = Element_edgeIds[elementId][FaceEdgeMap[iElementFace][0]];
    int edgeId1 = Element_edgeIds[elementId][FaceEdgeMap[iElementFace][1]];
    int edgeId2 = Element_edgeIds[elementId][FaceEdgeMap[iElementFace][2]];
    
    if (Edge_isDivided[edgeId0]
	&& Edge_isDivided[edgeId1]
	&& Edge_isDivided[edgeId2]) {
      return iElementFace;
    }
  }
  return -1;
}

void DivideEdges (void)
{
  int iEdge;
  int iElement;
  int iElementEdge;
  
  for (iEdge = 0; iEdge < NEdges; iEdge++) {
    Edge_isDivided[iEdge] = 0;
  }

  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isRefined[iElement]) {
      for (iElementEdge = 0; 
	   iElementEdge < N_ELEMENT_EDGES; 
	   iElementEdge++) {
	int edgeId = Element_edgeIds[iElement][iElementEdge];
	
	Edge_isDivided[edgeId] = 1;
      }
    }
  }
}

void PropergateRefinement (void)
{
  int iElement;
  int iElementEdge;

  for (;;) {
    int noChange = 1;

    for (iElement = 0; iElement < NElements; iElement++) {
      int nDividedEdges = 0;

      for (iElementEdge = 0; 
	   iElementEdge < N_ELEMENT_EDGES; 
	   iElementEdge++) {
	int edgeId = Element_edgeIds[iElement][iElementEdge];
	
	if (Edge_isDivided[edgeId]) nDividedEdges++;
      }

      if (nDividedEdges == 2) {
	/* face or full */
	int missingEdgeId 
	  = MissingEdgeOfPartiallyDividedElementFace (iElement);

	if (missingEdgeId != -1) {
	  /* face */
	  Edge_isDivided[missingEdgeId] = 1;
	  noChange = 0;
	} else {
	  /* full */
	  for (iElementEdge = 0; 
	       iElementEdge < N_ELEMENT_EDGES; 
	       iElementEdge++) {
	    int edgeId = Element_edgeIds[iElement][iElementEdge];
	    
	    Edge_isDivided[edgeId] = 1;
	  }
	  noChange = 0;
	}

      } else if (nDividedEdges == 3) {
	/* face or full */

	if (ElementFaceOfDividedElement (iElement) == -1) {
	  /* full */
	  for (iElementEdge = 0; 
	       iElementEdge < N_ELEMENT_EDGES; 
	       iElementEdge++) {
	    int edgeId = Element_edgeIds[iElement][iElementEdge];
	    
	    Edge_isDivided[edgeId] = 1;
	  }
	  noChange = 0;
	}

      } else if (nDividedEdges == 4
		 || nDividedEdges == 5) {
	/* full */

	for (iElementEdge = 0; 
	     iElementEdge < N_ELEMENT_EDGES; 
	     iElementEdge++) {
	  int edgeId = Element_edgeIds[iElement][iElementEdge];
	  
	  Edge_isDivided[edgeId] = 1;
	}
	noChange = 0;

      }
    }

    if (noChange) break;
  }
}

void MakeMidNodeOnDividedEdges (void)
{
  int iEdge;
  
  for (iEdge = 0; iEdge < NEdges; iEdge++) {
    Edge_midNodeId[iEdge] = -1;
  }

  for (iEdge = 0; iEdge < NEdges; iEdge++) {
    if (Edge_isDivided[iEdge]) {
      double midX, midY, midZ;

      Edge_getMidPoint (iEdge,
			&midX, &midY, &midZ);
      Edge_midNodeId[iEdge] = Mesh_addNewNode (midX, midY, midZ);
    }
  }
}

void DivideElementEdge (int elementId, int elementEdgeId)
{
  int nodeId0 = Element_nodeIds[elementId][0];
  int nodeId1 = Element_nodeIds[elementId][1];
  int nodeId2 = Element_nodeIds[elementId][2];
  int nodeId3 = Element_nodeIds[elementId][3];
  
  int edgeId0 = Element_edgeIds[elementId][0];
  int edgeId1 = Element_edgeIds[elementId][1];
  int edgeId2 = Element_edgeIds[elementId][2];
  int edgeId3 = Element_edgeIds[elementId][3];
  int edgeId4 = Element_edgeIds[elementId][4];
  int edgeId5 = Element_edgeIds[elementId][5];
  
  int edgeNodeId0 = Edge_midNodeId[edgeId0];
  int edgeNodeId1 = Edge_midNodeId[edgeId1];
  int edgeNodeId2 = Edge_midNodeId[edgeId2];
  int edgeNodeId3 = Edge_midNodeId[edgeId3];
  int edgeNodeId4 = Edge_midNodeId[edgeId4];
  int edgeNodeId5 = Edge_midNodeId[edgeId5];

  switch (elementEdgeId) {
  case 0:
    assert(edgeNodeId0 != -1);
    Mesh_addNewElement (edgeNodeId0, nodeId2, nodeId0, nodeId3);
    Mesh_addNewElement (edgeNodeId0, nodeId1, nodeId2, nodeId3);
    break;
  case 1:
    assert(edgeNodeId1 != -1);
    Mesh_addNewElement (edgeNodeId1, nodeId3, nodeId0, nodeId1);
    Mesh_addNewElement (edgeNodeId1, nodeId2, nodeId3, nodeId1);
    break;
  case 2:
    assert(edgeNodeId2 != -1);
    Mesh_addNewElement (edgeNodeId2, nodeId1, nodeId0, nodeId2);
    Mesh_addNewElement (edgeNodeId2, nodeId3, nodeId1, nodeId2);
    break;
  case 3:
    assert(edgeNodeId3 != -1);
    Mesh_addNewElement (edgeNodeId3, nodeId0, nodeId1, nodeId3);
    Mesh_addNewElement (edgeNodeId3, nodeId2, nodeId0, nodeId3);
    break;
  case 4:
    assert(edgeNodeId4 != -1);
    Mesh_addNewElement (edgeNodeId4, nodeId0, nodeId2, nodeId1);
    Mesh_addNewElement (edgeNodeId4, nodeId3, nodeId0, nodeId1);
    break;
  case 5:
    assert(edgeNodeId5 != -1);
    Mesh_addNewElement (edgeNodeId5, nodeId0, nodeId3, nodeId2);
    Mesh_addNewElement (edgeNodeId5, nodeId1, nodeId0, nodeId2);
    break;
  default:
    assert(0);
    break;
  }
  Mesh_deleteElement (elementId);
}

void DivideElementFace (int elementId, int elementFaceId)
{
  int nodeId0 = Element_nodeIds[elementId][0];
  int nodeId1 = Element_nodeIds[elementId][1];
  int nodeId2 = Element_nodeIds[elementId][2];
  int nodeId3 = Element_nodeIds[elementId][3];
  
  int edgeId0 = Element_edgeIds[elementId][0];
  int edgeId1 = Element_edgeIds[elementId][1];
  int edgeId2 = Element_edgeIds[elementId][2];
  int edgeId3 = Element_edgeIds[elementId][3];
  int edgeId4 = Element_edgeIds[elementId][4];
  int edgeId5 = Element_edgeIds[elementId][5];
  
  int edgeNodeId0 = Edge_midNodeId[edgeId0];
  int edgeNodeId1 = Edge_midNodeId[edgeId1];
  int edgeNodeId2 = Edge_midNodeId[edgeId2];
  int edgeNodeId3 = Edge_midNodeId[edgeId3];
  int edgeNodeId4 = Edge_midNodeId[edgeId4];
  int edgeNodeId5 = Edge_midNodeId[edgeId5];
    
  switch (elementFaceId) {
  case 0:
    assert(edgeNodeId3 != -1);
    assert(edgeNodeId4 != -1);
    assert(edgeNodeId5 != -1);
    Mesh_addNewElement (nodeId1, edgeNodeId5, edgeNodeId3, nodeId0);
    Mesh_addNewElement (nodeId2, edgeNodeId3, edgeNodeId4, nodeId0);
    Mesh_addNewElement (nodeId3, edgeNodeId4, edgeNodeId5, nodeId0);
    Mesh_addNewElement (edgeNodeId5, edgeNodeId4, edgeNodeId3, nodeId0);
    break;
  case 1:
    assert(edgeNodeId2 != -1);
    assert(edgeNodeId4 != -1);
    assert(edgeNodeId1 != -1);
    Mesh_addNewElement (nodeId0, edgeNodeId1, edgeNodeId2, nodeId1);
    Mesh_addNewElement (nodeId3, edgeNodeId2, edgeNodeId4, nodeId1);
    Mesh_addNewElement (nodeId2, edgeNodeId4, edgeNodeId1, nodeId1);
    Mesh_addNewElement (edgeNodeId1, edgeNodeId4, edgeNodeId2, nodeId1);
    break;
  case 2:
    assert(edgeNodeId0 != -1);
    assert(edgeNodeId5 != -1);
    assert(edgeNodeId2 != -1);
    Mesh_addNewElement (nodeId0, edgeNodeId2, edgeNodeId0, nodeId2);
    Mesh_addNewElement (nodeId1, edgeNodeId0, edgeNodeId5, nodeId2);
    Mesh_addNewElement (nodeId3, edgeNodeId5, edgeNodeId2, nodeId2);
    Mesh_addNewElement (edgeNodeId2, edgeNodeId5, edgeNodeId0, nodeId2);
    break;
  case 3:
    assert(edgeNodeId1 != -1);
    assert(edgeNodeId3 != -1);
    assert(edgeNodeId0 != -1);
    Mesh_addNewElement (nodeId0, edgeNodeId0, edgeNodeId1, nodeId3);
    Mesh_addNewElement (nodeId2, edgeNodeId1, edgeNodeId3, nodeId3);
    Mesh_addNewElement (nodeId1, edgeNodeId3, edgeNodeId0, nodeId3);
    Mesh_addNewElement (edgeNodeId0, edgeNodeId3, edgeNodeId1, nodeId3);
    break;
  default:
    assert(0);
    break;
  }
  Mesh_deleteElement (elementId);
}

void DivideElement (int elementId)
{
  int nodeId0 = Element_nodeIds[elementId][0];
  int nodeId1 = Element_nodeIds[elementId][1];
  int nodeId2 = Element_nodeIds[elementId][2];
  int nodeId3 = Element_nodeIds[elementId][3];
  
  int edgeId0 = Element_edgeIds[elementId][0];
  int edgeId1 = Element_edgeIds[elementId][1];
  int edgeId2 = Element_edgeIds[elementId][2];
  int edgeId3 = Element_edgeIds[elementId][3];
  int edgeId4 = Element_edgeIds[elementId][4];
  int edgeId5 = Element_edgeIds[elementId][5];
  
  int edgeNodeId0 = Edge_midNodeId[edgeId0];
  int edgeNodeId1 = Edge_midNodeId[edgeId1];
  int edgeNodeId2 = Edge_midNodeId[edgeId2];
  int edgeNodeId3 = Edge_midNodeId[edgeId3];
  int edgeNodeId4 = Edge_midNodeId[edgeId4];
  int edgeNodeId5 = Edge_midNodeId[edgeId5];
    
  double dist04 = DistanceBetweenNodes (edgeNodeId0, edgeNodeId4);
  double dist15 = DistanceBetweenNodes (edgeNodeId1, edgeNodeId5);
  double dist23 = DistanceBetweenNodes (edgeNodeId2, edgeNodeId3);
  
  assert(edgeNodeId0 != -1);
  assert(edgeNodeId1 != -1);
  assert(edgeNodeId2 != -1);
  assert(edgeNodeId3 != -1);
  assert(edgeNodeId4 != -1);
  assert(edgeNodeId5 != -1);
  
  Mesh_addNewElement (edgeNodeId2, edgeNodeId1, edgeNodeId0, nodeId0);
  Mesh_addNewElement (edgeNodeId0, edgeNodeId3, edgeNodeId5, nodeId1);
  Mesh_addNewElement (edgeNodeId1, edgeNodeId4, edgeNodeId3, nodeId2);
  Mesh_addNewElement (edgeNodeId2, edgeNodeId5, edgeNodeId4, nodeId3);
  
  if (dist04 < dist15 && dist04 < dist23) {
    /* e0 - e4 */
    Mesh_addNewElement (edgeNodeId0, edgeNodeId4, edgeNodeId1, edgeNodeId2);
    Mesh_addNewElement (edgeNodeId0, edgeNodeId4, edgeNodeId2, edgeNodeId5);
    Mesh_addNewElement (edgeNodeId0, edgeNodeId4, edgeNodeId5, edgeNodeId3);
    Mesh_addNewElement (edgeNodeId0, edgeNodeId4, edgeNodeId3, edgeNodeId1);
  } else if (dist15 < dist04 && dist15 < dist23) {
    /* e1 - e5 */
    Mesh_addNewElement (edgeNodeId1, edgeNodeId5, edgeNodeId4, edgeNodeId2);
    Mesh_addNewElement (edgeNodeId1, edgeNodeId5, edgeNodeId2, edgeNodeId0);
    Mesh_addNewElement (edgeNodeId1, edgeNodeId5, edgeNodeId0, edgeNodeId3);
    Mesh_addNewElement (edgeNodeId1, edgeNodeId5, edgeNodeId3, edgeNodeId4);
  } else {
    /* e2 - e3 */
    Mesh_addNewElement (edgeNodeId2, edgeNodeId3, edgeNodeId0, edgeNodeId1);
    Mesh_addNewElement (edgeNodeId2, edgeNodeId3, edgeNodeId1, edgeNodeId4);
    Mesh_addNewElement (edgeNodeId2, edgeNodeId3, edgeNodeId4, edgeNodeId5);
    Mesh_addNewElement (edgeNodeId2, edgeNodeId3, edgeNodeId5, edgeNodeId0);
  }
  
  Mesh_deleteElement (elementId);
}

void RefineElements (void)
{
  int iElement;
  int iElementEdge; 
  int nOldElements = NElements;

  for (iElement = 0; iElement < nOldElements; iElement++) {
    int nDividedEdges = 0;
    
    for (iElementEdge = 0; 
	 iElementEdge < N_ELEMENT_EDGES; 
	 iElementEdge++) {
      int edgeId = Element_edgeIds[iElement][iElementEdge];
      
      if (Edge_isDivided[edgeId]) nDividedEdges++;
    }
    
    assert(nDividedEdges == 0
	   || nDividedEdges == 1
	   || nDividedEdges == 3
	   || nDividedEdges == 6);
    
    if (nDividedEdges == 1) {
      int elementEdgeId = ElementEdgeOfDividedElement (iElement);

      assert(elementEdgeId != -1);
      
      DivideElementEdge (iElement, elementEdgeId);
    } else if (nDividedEdges == 3) {
      int elementFaceId = ElementFaceOfDividedElement (iElement);
      
      assert(elementFaceId != -1);

      DivideElementFace (iElement, elementFaceId);
    } else if (nDividedEdges == 6) {
      DivideElement (iElement);
    }
  }
}

void PerformCommand (void)
{
  DivideEdges ();
  PropergateRefinement ();
  MakeMidNodeOnDividedEdges ();
  RefineElements ();
}



int main (int argc, char *argv[])
{
  if (argc != 8) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tetra3d_refine  node_file  element_file \n"
       "        edge_vertex_file  element_edge_file \n"
       "        refined_element_list_file \n"
       "        node_file_OUT  element_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadEdgeVertexFile (argv[3]);
  ReadElementEdgeFile (argv[4]);
  ReadRefinedElementListFile (argv[5]);

  PerformCommand ();

  WriteNodeFile (argv[6]);
  WriteElementFile (argv[7]);

  return 0;
}


