/* advautomesh_h_tetra3d_sc_refine.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001

double Eps = DEFAULT_EPS;



double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}

void GetNormalizedVector 
(double x, double y, double z,
 double *nx_OUT, double *ny_OUT, double *nz_OUT)
{
  double length = sqrt (x * x + y * y + z * z);

  *nx_OUT = x / length;
  *ny_OUT = y / length;
  *nz_OUT = z / length;
}

void GetVectorProduct 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  *x_OUT = y0 * z1 - z0 * y1;
  *y_OUT = z0 * x1 - x0 * z1;
  *z_OUT = x0 * y1 - y0 * x1;
}

void GetPlaneEquation 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double x2, double y2, double z2,
 double *a_OUT, double *b_OUT, double *c_OUT, double *d_OUT)
{
  double nx, ny, nz;

  GetVectorProduct (x1 - x0, y1 - y0, z1 - z0,
		    x2 - x0, y2 - y0, z2 - z0,
		    &nx, &ny, &nz);
  GetNormalizedVector (nx, ny, nz,
		       &nx, &ny, &nz);
  *a_OUT = nx;
  *b_OUT = ny;
  *c_OUT = nz;
  *d_OUT = -(nx * x0 + ny * y0 + nz * z0);
}

void GetPlaneEquationToward 
(double nx, double ny, double nz,
 double x, double y, double z,
 double *a_OUT, double *b_OUT, double *c_OUT, double *d_OUT)
{
  double nx2, ny2, nz2;

  GetNormalizedVector (nx, ny, nz,
		       &nx2, &ny2, &nz2);
  *a_OUT = nx2;
  *b_OUT = ny2;
  *c_OUT = nz2;
  *d_OUT = -(nx2 * x + ny2 * y + nz2 * z);
}

double DistanceFromLine 
(double originX, double originY, double originZ,
 double dirX, double dirY, double dirZ,
 double x, double y, double z)
{
  double t = (x - originX) * dirX
    + (y - originY) * dirY
    + (z - originZ) * dirZ;
  
  return Distance (x, y, z,
		   originX + dirX * t,
		   originY + dirY * t,
		   originZ + dirZ * t);
}

void GetProjectedPointOnPlane
(double a, double b, double c, double d, 
 double x, double y, double z,
 double *px_OUT, double *py_OUT, double *pz_OUT)
{
  double value = a * x + b * y + c * z + d;
  
  *px_OUT = x - a * value;
  *py_OUT = y - b * value;
  *pz_OUT = z - c * value;
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][3];



/* Element */

#define N_ELEMENT_NODES 4
#define MAX_ELEMENTS 1000000

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];
int Element_isRefined[MAX_ELEMENTS];



void Element_getCenter 
(int elementId, 
 double *centerX_OUT, double *centerY_OUT, double *centerZ_OUT)
{
  int iElementNode;

  assert(0 <= elementId);
  assert(elementId < NElements);

  *centerX_OUT = 0.0;
  *centerY_OUT = 0.0;
  *centerZ_OUT = 0.0;
  for (iElementNode = 0; 
       iElementNode < N_ELEMENT_NODES; 
       iElementNode++) {
    int nodeId = Element_nodeIds[elementId][iElementNode];
    double x = Node_coord[nodeId][0];
    double y = Node_coord[nodeId][1];
    double z = Node_coord[nodeId][2];
    
    *centerX_OUT += x;
    *centerY_OUT += y;
    *centerZ_OUT += z;
  }
  *centerX_OUT /= N_ELEMENT_NODES;
  *centerY_OUT /= N_ELEMENT_NODES;
  *centerZ_OUT /= N_ELEMENT_NODES;
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void WriteRefinedElementListFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int count;
  
  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  count = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isRefined[iElement]) {
      count++;
    }
  }

  fprintf (fp, "%d\n", 
	   count);
  
  count = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isRefined[iElement]) {
      fprintf (fp, "%d   %d\n", 
	       count, iElement);
      count++;
    }
  }

  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Crack */

double Center[3];
double XDir[3], YDir[3], ZDir[3];
double XRadius, YRadius;

#define N_DIVISIONS 100

double CrackPlaneA, CrackPlaneB, CrackPlaneC, CrackPlaneD;



void ReadSurfaceCrackFile (const char *fileName)
{
  FILE *fp;
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%lf %lf %lf", 
	  &Center[0], &Center[1], &Center[2]);
  fscanf (fp, "%lf %lf %lf", 
	  &ZDir[0], &ZDir[1], &ZDir[2]);
  fscanf (fp, "%lf %lf %lf", 
	  &XDir[0], &XDir[1], &XDir[2]);
  fscanf (fp, "%lf %lf", 
	  &XRadius, &YRadius);

  GetNormalizedVector
    (XDir[0], XDir[1], XDir[2],
     &XDir[0], &XDir[1], &XDir[2]);
  GetNormalizedVector
    (ZDir[0], ZDir[1], ZDir[2],
     &ZDir[0], &ZDir[1], &ZDir[2]);
  GetVectorProduct 
    (ZDir[0], ZDir[1], ZDir[2],
     XDir[0], XDir[1], XDir[2],
     &YDir[0], &YDir[1], &YDir[2]);

  fclose (fp);

  GetPlaneEquationToward 
    (ZDir[0], ZDir[1], ZDir[2],
     Center[0], Center[1], Center[2],
     &CrackPlaneA, &CrackPlaneB, &CrackPlaneC, &CrackPlaneD);
}



/************************
	 Command
*************************/

void PerformCommand (double radius, int crackTipOnlyFlag)
{
  int iElement;
  int i;

  for (iElement = 0; iElement < NElements; iElement++) {
    Element_isRefined[iElement] = 0;
  }

  for (iElement = 0; iElement < NElements; iElement++) {
    double centerX, centerY, centerZ;
    
    Element_getCenter (iElement, 
		       &centerX, &centerY, &centerZ);

    for (i = 0; i < N_DIVISIONS * 4; i++) {
      double angle = M_PI / 2 / N_DIVISIONS * i;
      double s = sin (angle);
      double c = cos (angle);
      
      double x = Center[0] 
	+ XDir[0] * XRadius * c + YDir[0] * YRadius * s;
      double y = Center[1] 
	+ XDir[1] * XRadius * c + YDir[1] * YRadius * s;
      double z = Center[2] 
	+ XDir[2] * XRadius * c + YDir[2] * YRadius * s;

      double distance = Distance 
	(x, y, z,
	 centerX, centerY, centerZ);

      if (distance < radius - Eps) {
	Element_isRefined[iElement] = 1;
      }
    }

    if (!crackTipOnlyFlag) {
      double crackPlaneValue
	= CrackPlaneA * centerX
	+ CrackPlaneB * centerY
	+ CrackPlaneC * centerZ 
	+ CrackPlaneD;

      if (fabs (crackPlaneValue) < radius - Eps) {
	double px, py, pz;
	double distance;
	
	GetProjectedPointOnPlane 
	  (CrackPlaneA, CrackPlaneB, CrackPlaneC, CrackPlaneD,
	   centerX, centerY, centerZ,
	   &px, &py, &pz);
	distance = Distance 
	  (px, py, pz,
	   Center[0], Center[1], Center[2]);

	if (distance < XRadius - Eps
	    || distance < YRadius - Eps) {
	  Element_isRefined[iElement] = 1;
	}
      }
    }
  }
}



int main (int argc, char *argv[])
{
  double radius;
  int crackTipOnlyFlag;

  if (argc != 7) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tetra3d_sc_refine  node_file  element_file \n"
       "        surface_crack_file  radius  crackTipOnlyFlag \n"
       "        refined_element_list_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadSurfaceCrackFile (argv[3]);

  radius = atof (argv[4]);
  crackTipOnlyFlag = atoi (argv[5]);
  PerformCommand (radius, crackTipOnlyFlag);

  WriteRefinedElementListFile (argv[6]);

  return 0;
}
