/* advautomesh_h_tetra3d_section.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 2500000

int NNodes;
double Node_coord[MAX_NODES][3];
double Node_sectionValue[MAX_NODES];



/* Element */

#define N_ELEMENT_NODES 4
#define N_ELEMENT_EDGES 6
#define MAX_ELEMENTS 12000000

const int EdgeNodeMap[N_ELEMENT_EDGES][2] = {
  { 0, 1 }, 
  { 0, 2 }, 
  { 0, 3 }, 
  { 1, 2 }, 
  { 2, 3 }, 
  { 3, 1 }  
};

const int NSectionEdges[8] = {
  0, 3, 3, 4, 3, 4, 4, 3
};

const int SectionEdgeIds[8][4] = {
  { -1, -1, -1, -1 },
  { 0, 1, 2, -1 },
  { 0, 3, 5, -1 },
  { 3, 1, 2, 5 },
  { 3, 1, 4, -1 },
  { 0, 3, 4, 2 },
  { 0, 5, 4, 1 },
  { 2, 5, 4, -1 }
};

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Section */

#define MAX_SECTIONS 1000000

int NSections;
int Section_elementId[MAX_SECTIONS];
int Section_elementEdgeIds[MAX_SECTIONS][3];
double Section_ratios[MAX_SECTIONS][3];

double PlaneA, PlaneB, PlaneC, PlaneD;



void SectionGroup_addSection 
(int elementId, 
 int elementEdgeId0, double ratio0,
 int elementEdgeId1, double ratio1,
 int elementEdgeId2, double ratio2)
{
  assert(NSections < MAX_SECTIONS);
  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(0 <= elementEdgeId0);
  assert(elementEdgeId0 < N_ELEMENT_EDGES);
  assert(0 <= elementEdgeId1);
  assert(elementEdgeId1 < N_ELEMENT_EDGES);
  assert(0 <= elementEdgeId2);
  assert(elementEdgeId2 < N_ELEMENT_EDGES);

  Section_elementId[NSections] = elementId;
  Section_elementEdgeIds[NSections][0] = elementEdgeId0;
  Section_elementEdgeIds[NSections][1] = elementEdgeId1;
  Section_elementEdgeIds[NSections][2] = elementEdgeId2;
  Section_ratios[NSections][0] = ratio0;
  Section_ratios[NSections][1] = ratio1;
  Section_ratios[NSections][2] = ratio2;

  NSections++;
}

void WriteSectionFile (const char *fileName)
{
  FILE *fp;
  int iSection;
  
  assert(0 < NSections);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NSections);
  
  for (iSection = 0; iSection < NSections; iSection++) {
    fprintf (fp, "%d   %d   %d %22.15e   %d %22.15e   %d %22.15e\n", 
	     iSection,
	     Section_elementId[iSection],
	     Section_elementEdgeIds[iSection][0],
	     Section_ratios[iSection][0],
	     Section_elementEdgeIds[iSection][1],
	     Section_ratios[iSection][1],
	     Section_elementEdgeIds[iSection][2],
	     Section_ratios[iSection][2]);
  }

  fclose (fp);
}



/************************
	 Command
*************************/

int SectionMask (int elementId)
{
  int nodeId0 = Element_nodeIds[elementId][0];
  int nodeId1 = Element_nodeIds[elementId][1];
  int nodeId2 = Element_nodeIds[elementId][2];
  int nodeId3 = Element_nodeIds[elementId][3];
  unsigned int mask = 0;
  
  if (0.0 < Node_sectionValue[nodeId0]) mask |= 1;
  if (0.0 < Node_sectionValue[nodeId1]) mask |= 2;
  if (0.0 < Node_sectionValue[nodeId2]) mask |= 4;
  if (0.0 < Node_sectionValue[nodeId3]) mask |= 8;
  if ((mask & 8) != 0) mask = (~mask) & 15;

  assert(0 <= mask);
  assert(mask < 8);

  return mask;
}

int GetSection 
(int elementId, 
 int *nEdges_OUT,
 int elementEdgeIds_OUT[4], double ratios_OUT[4])
{
  unsigned int mask = SectionMask (elementId);
  int iEdge;

  *nEdges_OUT = NSectionEdges[mask];
  if (*nEdges_OUT == 0) return 0;
  
  assert(*nEdges_OUT == 3 || *nEdges_OUT == 4);

  for (iEdge = 0; iEdge < *nEdges_OUT; iEdge++) {
    int elementEdgeId = SectionEdgeIds[mask][iEdge];

    int elementNodeId0 = EdgeNodeMap[elementEdgeId][0];
    int elementNodeId1 = EdgeNodeMap[elementEdgeId][1];

    int nodeId0 = Element_nodeIds[elementId][elementNodeId0];
    int nodeId1 = Element_nodeIds[elementId][elementNodeId1];

    double ratio = (0.0 - Node_sectionValue[nodeId0]) 
      / (Node_sectionValue[nodeId1] - Node_sectionValue[nodeId0]);
    
    elementEdgeIds_OUT[iEdge] = elementEdgeId; 
    ratios_OUT[iEdge] = ratio;
  }    

  return 1;
}

void PerformCommand (void)
{
  int iNode;
  int iElement;

  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    double z = Node_coord[iNode][2];
    
    Node_sectionValue[iNode] 
      = PlaneA * x + PlaneB * y + PlaneC * z + PlaneD;
  }

  for (iElement = 0; iElement < NElements; iElement++) {
    int nEdges;
    int elementEdgeIds[4];
    double ratios[4];
    
    if (GetSection
	(iElement,
	 &nEdges, elementEdgeIds, ratios)) {
      assert(nEdges == 3 || nEdges == 4);
      SectionGroup_addSection 
	(iElement, 
	 elementEdgeIds[0], ratios[0],
	 elementEdgeIds[1], ratios[1],
	 elementEdgeIds[2], ratios[2]);
      if (nEdges == 4) {
	SectionGroup_addSection 
	  (iElement, 
	   elementEdgeIds[2], ratios[2],
	   elementEdgeIds[3], ratios[3],
	   elementEdgeIds[0], ratios[0]);
      }
    }
  }
}



#if 1
#include "../../util/time_measure.c"
#define TIME_MEASURE
#endif

int main (int argc, char *argv[])
{
  if (argc != 8) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tetra3d_section  node_file  element_file \n"
       "        a  b  c  d  \n"
       "        section_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  PlaneA = atof (argv[3]);
  PlaneB = atof (argv[4]);
  PlaneC = atof (argv[5]);
  PlaneD = atof (argv[6]);

#ifdef TIME_MEASURE
  StartMeasureTime ();
#endif

  PerformCommand ();

#ifdef TIME_MEASURE
  EndMeasureTime ();
#endif

  WriteSectionFile (argv[7]);

  return 0;
}
