/* advautomesh_h_tetra3d_section_grid.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

#define VOLUME_COORD_EPS 0.00001



double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}

void GetNormalizedVector 
(double x, double y, double z,
 double *nx_OUT, double *ny_OUT, double *nz_OUT)
{
  double length = sqrt (x * x + y * y + z * z);

  *nx_OUT = x / length;
  *ny_OUT = y / length;
  *nz_OUT = z / length;
}

void GetVectorProduct 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  *x_OUT = y0 * z1 - z0 * y1;
  *y_OUT = z0 * x1 - x0 * z1;
  *z_OUT = x0 * y1 - y0 * x1;
}

void GetAnyPerpendicularDirection
(double x, double y, double z,
 double *nx0_OUT, double *ny0_OUT, double *nz0_OUT,
 double *nx1_OUT, double *ny1_OUT, double *nz1_OUT)
{
  double length;

  GetVectorProduct (x, y, z,
		    1, 0, 0,
		    nx0_OUT, ny0_OUT, nz0_OUT);
  length = sqrt (*nx0_OUT * *nx0_OUT
		 + *ny0_OUT * *ny0_OUT
		 + *nz0_OUT * *nz0_OUT);
  if (length == 0.0) {
    GetVectorProduct (x, y, z,
		      0, 1, 0,
		      nx0_OUT, ny0_OUT, nz0_OUT);
  }
  GetNormalizedVector (*nx0_OUT, *ny0_OUT, *nz0_OUT,
		       nx0_OUT, ny0_OUT, nz0_OUT);
  GetVectorProduct (x, y, z,
		    *nx0_OUT, *ny0_OUT, *nz0_OUT,
		    nx1_OUT, ny1_OUT, nz1_OUT);
  GetNormalizedVector (*nx1_OUT, *ny1_OUT, *nz1_OUT,
		       nx1_OUT, ny1_OUT, nz1_OUT);
}

void GetProjectedPointOnPlane
(double a, double b, double c, double d, 
 double x, double y, double z,
 double *px_OUT, double *py_OUT, double *pz_OUT)
{
  double value = a * x + b * y + c * z + d;
  
  *px_OUT = x - a * value;
  *py_OUT = y - b * value;
  *pz_OUT = z - c * value;
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 2500000

int NNodes;
double Node_coord[MAX_NODES][3];



void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  *minRangeZ_OUT = *maxRangeZ_OUT = Node_coord[0][2];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    double z = Node_coord[iNode][2];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
    if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
  }
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Section Grid */

double PlaneA, PlaneB, PlaneC, PlaneD;
double Interval;



/* Point */

#define MAX_POINTS 1000000

int NPoints;
double Point_coord[MAX_POINTS][3];



void SectionGrid_setPlane (double a, double b, double c, double d)
{
  double length = sqrt (a * a + b * b + c * c);
  
  PlaneA = a / length;
  PlaneB = b / length;
  PlaneC = c / length;
  PlaneD = d / length;
}

void PointGroup_addPoint 
(double x, double y, double z)
{
  assert(NPoints < MAX_POINTS);
  
  Point_coord[NPoints][0] = x;
  Point_coord[NPoints][1] = y;
  Point_coord[NPoints][2] = z;
  
  NPoints++;
}

void WritePointFile (const char *fileName)
{
  FILE *fp;
  int iPoint;

  assert(0 <= NPoints);
  
  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  fprintf (fp, "%d\n", 
	   NPoints);
  
  for (iPoint = 0; iPoint < NPoints; iPoint++) {
    fprintf (fp, "%d   %22.15e %22.15e %22.15e\n", 
	     iPoint,
	     Point_coord[iPoint][0], 
	     Point_coord[iPoint][1], 
	     Point_coord[iPoint][2]);
  }

  fclose (fp);
}



/************************
	 Command
*************************/

void PerformCommand (void)
{
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  double centerX, centerY, centerZ;
  double size;
  double sectionCenterX, sectionCenterY, sectionCenterZ;
  double uAxisX, uAxisY, uAxisZ;
  double vAxisX, vAxisY, vAxisZ;
  int nDivisions;
  int iu, iv;

  Mesh_getRange
    (&minRangeX, &minRangeY, &minRangeZ,
     &maxRangeX, &maxRangeY, &maxRangeZ);
  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;
  centerZ = (minRangeZ + maxRangeZ) / 2;
  size = Distance (minRangeX, minRangeY, minRangeZ,
		   maxRangeX, maxRangeY, maxRangeZ) / 2;
  GetProjectedPointOnPlane
    (PlaneA, PlaneB, PlaneC, PlaneD,
     centerX, centerY, centerZ, 
     &sectionCenterX, &sectionCenterY, &sectionCenterZ);
  GetAnyPerpendicularDirection
    (PlaneA, PlaneB, PlaneC,
     &uAxisX, &uAxisY, &uAxisZ,
     &vAxisX, &vAxisY, &vAxisZ);
  nDivisions = size / Interval;
  
  for (iu = -nDivisions; iu <= nDivisions; iu++) {
    for (iv = -nDivisions; iv <= nDivisions; iv++) {
      double x = sectionCenterX 
	+ uAxisX * Interval * iu + vAxisX * Interval * iv;
      double y = sectionCenterY 
	+ uAxisY * Interval * iu + vAxisY * Interval * iv;
      double z = sectionCenterZ 
	+ uAxisZ * Interval * iu + vAxisZ * Interval * iv;
      
      PointGroup_addPoint (x, y, z);
    }
  }
}


int main (int argc, char *argv[])
{
  double a, b, c, d;

  if (argc != 8) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tetra3d_section_grid  node_file \n"
       "        a  b  c  d  interval \n"
       "        point_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);

  a = atof (argv[2]);
  b = atof (argv[3]);
  c = atof (argv[4]);
  d = atof (argv[5]);
  SectionGrid_setPlane (a, b, c, d);
  
  Interval = atof (argv[6]);

  PerformCommand ();

  WritePointFile (argv[7]);

  return 0;
}
