/* advautomesh_h_tetra3d_section_isoline.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 2500000

int NNodes;
double Node_coord[MAX_NODES][3];
double Node_scalar[MAX_NODES];



/* Element */

#define N_ELEMENT_NODES 4
#define N_ELEMENT_EDGES 6
#define MAX_ELEMENTS 12000000

const int EdgeNodeMap[N_ELEMENT_EDGES][2] = {
  { 0, 1 }, 
  { 0, 2 }, 
  { 0, 3 }, 
  { 1, 2 }, 
  { 2, 3 }, 
  { 3, 1 }  
};

const int NIsolineEdges[4] = {
  0, 2, 2, 2
};

const int IsolineEdgeIds[4][2] = {
  { -1, -1 },
  { 1, 2 },
  { 2, 0 },
  { 0, 1 }
};

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



void Element_getEdgePoint 
(int elementId, 
 int elementEdgeId, double ratio,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  int nodeId0, nodeId1;
  double x0, y0, z0;
  double x1, y1, z1;

  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(0 <= elementEdgeId);
  assert(elementEdgeId < N_ELEMENT_EDGES);

  nodeId0 = Element_nodeIds
    [elementId][EdgeNodeMap[elementEdgeId][0]];
  nodeId1 = Element_nodeIds
    [elementId][EdgeNodeMap[elementEdgeId][1]];

  x0 = Node_coord[nodeId0][0];
  y0 = Node_coord[nodeId0][1];
  z0 = Node_coord[nodeId0][2];

  x1 = Node_coord[nodeId1][0];
  y1 = Node_coord[nodeId1][1];
  z1 = Node_coord[nodeId1][2];
  
  *x_OUT = x0 + (x1 - x0) * ratio;
  *y_OUT = y0 + (y1 - y0) * ratio;
  *z_OUT = z0 + (z1 - z0) * ratio;
}

double Element_edgeScalar 
(int elementId, 
 int elementEdgeId, double ratio)
{
  int nodeId0, nodeId1;
  double scalar0, scalar1;

  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(0 <= elementEdgeId);
  assert(elementEdgeId < N_ELEMENT_EDGES);

  nodeId0 = Element_nodeIds
    [elementId][EdgeNodeMap[elementEdgeId][0]];
  nodeId1 = Element_nodeIds
    [elementId][EdgeNodeMap[elementEdgeId][1]];

  scalar0 = Node_scalar[nodeId0];
  scalar1 = Node_scalar[nodeId1];
  
  return scalar0 + (scalar1 - scalar0) * ratio;
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void ReadNodeScalarFile (const char *fileName)
{
  FILE *fp;
  int count;
  int iNode;
  
  assert(0 < NNodes);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &count);
  assert(count == NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double scalar;

    fscanf (fp, "%d %lf", 
	    &nodeId,
	    &scalar);
    assert(nodeId == iNode);
    Node_scalar[iNode] = scalar;
  }

  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Section */

#define MAX_SECTIONS 1000000

int NSections;
int Section_elementId[MAX_SECTIONS];
int Section_elementEdgeIds[MAX_SECTIONS][3];
double Section_ratios[MAX_SECTIONS][3];



/* Isoline */

#define MAX_LEVELS 10000
#define MAX_ISOLINES 1000000

int NLevels;
int Level_startIsolineId[MAX_LEVELS];
int Level_nIsolines[MAX_LEVELS];
int NIsolines;
double Isoline_coord0[MAX_ISOLINES][3];
double Isoline_coord1[MAX_ISOLINES][3];



void IsolineGroup_addLevel (void)
{
  assert(NLevels < MAX_LEVELS);

  Level_startIsolineId[NLevels] = NIsolines;
  Level_nIsolines[NLevels] = 0;
  NLevels++;
}

void IsolineGroup_addIsoline 
(double x0, double y0, double z0,
 double x1, double y1, double z1)
{
  assert(0 < NLevels);
  assert(NIsolines < MAX_ISOLINES);

  Level_nIsolines[NLevels - 1]++;

  Isoline_coord0[NIsolines][0] = x0;
  Isoline_coord0[NIsolines][1] = y0;
  Isoline_coord0[NIsolines][2] = z0;

  Isoline_coord1[NIsolines][0] = x1;
  Isoline_coord1[NIsolines][1] = y1;
  Isoline_coord1[NIsolines][2] = z1;

  NIsolines++;
}

void ReadSectionFile (const char *fileName)
{
  FILE *fp;
  int iSection;
  
  assert(0 < NElements);
  assert(NSections == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NSections);
  assert(0 < NSections);
  assert(NSections <= MAX_SECTIONS);
  
  for (iSection = 0; iSection < NSections; iSection++) {
    int sectionId;
    int elementId;
    int elementEdgeId0, elementEdgeId1, elementEdgeId2;
    double ratio0, ratio1, ratio2;
    
    fscanf (fp, "%d %d %d %lf %d %lf %d %lf", 
	    &sectionId,
	    &elementId,
	    &elementEdgeId0, &ratio0,
	    &elementEdgeId1, &ratio1,
	    &elementEdgeId2, &ratio2);
    assert(sectionId == iSection);
    assert(0 <= elementId);
    assert(elementId < NElements);
    Section_elementId[iSection] = elementId;
    
    assert(0 <= elementEdgeId0);
    assert(elementEdgeId0 < N_ELEMENT_EDGES);
    Section_elementEdgeIds[iSection][0] = elementEdgeId0;
    Section_ratios[iSection][0] = ratio0;
    
    assert(0 <= elementEdgeId1);
    assert(elementEdgeId1 < N_ELEMENT_EDGES);
    Section_elementEdgeIds[iSection][1] = elementEdgeId1;
    Section_ratios[iSection][1] = ratio1;
    
    assert(0 <= elementEdgeId2);
    assert(elementEdgeId2 < N_ELEMENT_EDGES);
    Section_elementEdgeIds[iSection][2] = elementEdgeId2;
    Section_ratios[iSection][2] = ratio2;
  }

  fclose (fp);
}

void WriteSectionIsolineFile (const char *fileName)
{
  FILE *fp;
  int iLevel;
  int iIsoline;
  
  assert(0 < NLevels);
  assert(0 < NIsolines);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NLevels);
  
  for (iLevel = 0; iLevel < NLevels; iLevel++) {
    int startId = Level_startIsolineId[iLevel];
    int nIsolines = Level_nIsolines[iLevel];
    
    fprintf (fp, "%d   %d\n", 
	     iLevel,
	     nIsolines);

    for (iIsoline = startId; 
	 iIsoline < startId + nIsolines; 
	 iIsoline++) {
      fprintf (fp, "%d   %22.15e %22.15e %22.15e   %22.15e %22.15e %22.15e\n", 
	       iIsoline - startId,
	       Isoline_coord0[iIsoline][0],
	       Isoline_coord0[iIsoline][1],
	       Isoline_coord0[iIsoline][2],
	       Isoline_coord1[iIsoline][0],
	       Isoline_coord1[iIsoline][1],
	       Isoline_coord1[iIsoline][2]);
    }
  }
  
  fclose (fp);
}



/************************
	 Command
*************************/

int IsolineMask 
(double scalar0, double scalar1, double scalar2, 
 double levelValue)
{
  unsigned int mask = 0;
  
  if (levelValue < scalar0) mask |= 1;
  if (levelValue < scalar1) mask |= 2;
  if (levelValue < scalar2) mask |= 4;
  if ((mask & 4) != 0) mask = (~mask) & 7;

  assert(0 <= mask);
  assert(mask < 4);

  return mask;
}

int GetIsoline 
(int sectionId, double levelValue, 
 double *ix0_OUT, double *iy0_OUT, double *iz0_OUT, 
 double *ix1_OUT, double *iy1_OUT, double *iz1_OUT)
{
  double scalar0, scalar1, scalar2;
  unsigned int mask;
  int nEdges;
  double ex0, ey0, ez0;
  double ex1, ey1, ez1;
  double ex2, ey2, ez2;

  scalar0 = Element_edgeScalar 
    (Section_elementId[sectionId], 
     Section_elementEdgeIds[sectionId][0],
     Section_ratios[sectionId][0]);
  scalar1 = Element_edgeScalar 
    (Section_elementId[sectionId], 
     Section_elementEdgeIds[sectionId][1],
     Section_ratios[sectionId][1]);
  scalar2 = Element_edgeScalar 
    (Section_elementId[sectionId], 
     Section_elementEdgeIds[sectionId][2],
     Section_ratios[sectionId][2]);

  mask = IsolineMask (scalar0, scalar1, scalar2, levelValue);
  nEdges = NIsolineEdges[mask];;
  if (nEdges == 0) return 0;
  
  assert(nEdges == 2);

  Element_getEdgePoint 
    (Section_elementId[sectionId], 
     Section_elementEdgeIds[sectionId][0],
     Section_ratios[sectionId][0],
     &ex0, &ey0, &ez0);
  Element_getEdgePoint 
    (Section_elementId[sectionId], 
     Section_elementEdgeIds[sectionId][1],
     Section_ratios[sectionId][1],
     &ex1, &ey1, &ez1);
  Element_getEdgePoint 
    (Section_elementId[sectionId], 
     Section_elementEdgeIds[sectionId][2],
     Section_ratios[sectionId][2],
     &ex2, &ey2, &ez2);
  
  {
    int sectionEdgeId = IsolineEdgeIds[mask][0];
    double ratio;

    switch (sectionEdgeId) {
    case 0:
      ratio = (levelValue - scalar1) / (scalar2 - scalar1);
      *ix0_OUT = ex1 + (ex2 - ex1)  * ratio;
      *iy0_OUT = ey1 + (ey2 - ey1)  * ratio;
      *iz0_OUT = ez1 + (ez2 - ez1)  * ratio;
      break;
    case 1:
      ratio = (levelValue - scalar2) / (scalar0 - scalar2);
      *ix0_OUT = ex2 + (ex0 - ex2)  * ratio;
      *iy0_OUT = ey2 + (ey0 - ey2)  * ratio;
      *iz0_OUT = ez2 + (ez0 - ez2)  * ratio;
      break;
    case 2:
      ratio = (levelValue - scalar0) / (scalar1 - scalar0);
      *ix0_OUT = ex0 + (ex1 - ex0)  * ratio;
      *iy0_OUT = ey0 + (ey1 - ey0)  * ratio;
      *iz0_OUT = ez0 + (ez1 - ez0)  * ratio;
      break;
    default:
      assert(0);
      break;
    }
  }
    
  {
    int sectionEdgeId = IsolineEdgeIds[mask][1];
    double ratio;

    switch (sectionEdgeId) {
    case 0:
      ratio = (levelValue - scalar1) / (scalar2 - scalar1);
      *ix1_OUT = ex1 + (ex2 - ex1)  * ratio;
      *iy1_OUT = ey1 + (ey2 - ey1)  * ratio;
      *iz1_OUT = ez1 + (ez2 - ez1)  * ratio;
      break;
    case 1:
      ratio = (levelValue - scalar2) / (scalar0 - scalar2);
      *ix1_OUT = ex2 + (ex0 - ex2)  * ratio;
      *iy1_OUT = ey2 + (ey0 - ey2)  * ratio;
      *iz1_OUT = ez2 + (ez0 - ez2)  * ratio;
      break;
    case 2:
      ratio = (levelValue - scalar0) / (scalar1 - scalar0);
      *ix1_OUT = ex0 + (ex1 - ex0)  * ratio;
      *iy1_OUT = ey0 + (ey1 - ey0)  * ratio;
      *iz1_OUT = ez0 + (ez1 - ez0)  * ratio;
      break;
    default:
      assert(0);
      break;
    }
  }
    
  return 1;
}

void PerformCommand 
(double minLevel, double maxLevel, int nDivisions)
{
  int iLevel;
  int iSection;

  for (iLevel = 0; iLevel <= nDivisions; iLevel++) {
    double levelValue = minLevel 
      + (maxLevel - minLevel) / nDivisions * iLevel;

    IsolineGroup_addLevel ();

    for (iSection = 0; iSection < NSections; iSection++) {
      double ix0, iy0, iz0;
      double ix1, iy1, iz1;
      
      if (GetIsoline
	  (iSection, levelValue,
	   &ix0, &iy0, &iz0,
	   &ix1, &iy1, &iz1)) {
	IsolineGroup_addIsoline 
	  (ix0, iy0, iz0,
	   ix1, iy1, iz1);
      }
    }
  }
}



int main (int argc, char *argv[])
{
  double minLevel, maxLevel;
  int nDivisions;
  
  if (argc != 9) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tetra3d_section_isoline  node_file  element_file \n"
       "        node_scalar_file  section_file \n"
       "        minLevel  maxLevel  nDivisions \n"
       "        section_isoline_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadNodeScalarFile (argv[3]);
  ReadSectionFile (argv[4]);

  minLevel = atof (argv[5]);
  maxLevel = atof (argv[6]);
  nDivisions = atoi (argv[7]);

  PerformCommand (minLevel, maxLevel, nDivisions);

  WriteSectionIsolineFile (argv[8]);

  return 0;
}

