/* advautomesh_h_tetra3d_sliver.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001

double Eps = DEFAULT_EPS;



int Equal (double value0, double value1) 
{
  return fabs (value1 - value0) < Eps;
}

int Equal2 (double value0, double value1) 
{
  return fabs (value1 - value0) < Eps * Eps;
}

double SignedTetraVolume
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double x2, double y2, double z2,
 double x3, double y3, double z3)
{
  double x10 = x1 - x0;
  double y10 = y1 - y0;
  double z10 = z1 - z0;

  double x20 = x2 - x0;
  double y20 = y2 - y0;
  double z20 = z2 - z0;
  
  double x30 = x3 - x0;
  double y30 = y3 - y0;
  double z30 = z3 - z0;
  
  double d23 = y20 * z30 - z20 * y30;
  double d31 = y30 * z10 - z30 * y10;
  double d12 = y10 * z20 - z10 * y20;
  
  return (x10 * d23 + x20 * d31 + x30 * d12) / 6;
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][3];



/* Element */

#define N_ELEMENT_NODES 4
#define N_ELEMENT_FACES 4
#define MAX_ELEMENTS 5000000

const int FaceNodeMap[N_ELEMENT_FACES][3] = {
  { 1, 2, 3 },
  { 0, 3, 2 },
  { 0, 1, 3 },
  { 0, 2, 1 }
};

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];
int Element_isDeleted[MAX_ELEMENTS];




int Element_containsNode 
(int elementId, 
 int nodeId)
{
  int iElementNode;
  
  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(!Element_isDeleted[elementId]);
  
  for (iElementNode = 0;
       iElementNode < N_ELEMENT_NODES;
       iElementNode++) {
    if (Element_nodeIds[elementId][iElementNode] == nodeId) {
      return 1;
    }
  }
  return 0;
}

int Element_otherNode 
(int elementId, 
 int nodeId0, int nodeId1, int nodeId2)
{
  int iElementNode;
  
  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(!Element_isDeleted[elementId]);
  
  if (Element_containsNode (elementId, nodeId0)
      && Element_containsNode (elementId, nodeId1)
      && Element_containsNode (elementId, nodeId2)) {
    for (iElementNode = 0;
	 iElementNode < N_ELEMENT_NODES;
	 iElementNode++) {
      int nodeId = Element_nodeIds[elementId][iElementNode];
      
      if (nodeId != nodeId0
	  && nodeId != nodeId1
	  && nodeId != nodeId2) {
	return nodeId;
      }
    }
    assert(0);
    return -1;
  }
  return -1;
}

double Element_volume (int elementId)
{
  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(!Element_isDeleted[elementId]);

  {
    int nodeId0 = Element_nodeIds[elementId][0];
    int nodeId1 = Element_nodeIds[elementId][1];
    int nodeId2 = Element_nodeIds[elementId][2];
    int nodeId3 = Element_nodeIds[elementId][3];
    
    double x0 = Node_coord[nodeId0][0];
    double y0 = Node_coord[nodeId0][1];
    double z0 = Node_coord[nodeId0][2];
    
    double x1 = Node_coord[nodeId1][0];
    double y1 = Node_coord[nodeId1][1];
    double z1 = Node_coord[nodeId1][2];
    
    double x2 = Node_coord[nodeId2][0];
    double y2 = Node_coord[nodeId2][1];
    double z2 = Node_coord[nodeId2][2];
    
    double x3 = Node_coord[nodeId3][0];
    double y3 = Node_coord[nodeId3][1];
    double z3 = Node_coord[nodeId3][2];
    
    double volume 
      = SignedTetraVolume (x0, y0, z0, 
			   x1, y1, z1, 
			   x2, y2, z2, 
			   x3, y3, z3);

    return volume;
  }
}

int Mesh_addNewElement (int nodeId0, int nodeId1, int nodeId2, int nodeId3)
{
  int elementId = NElements;

  assert(0 <= nodeId0);
  assert(nodeId0 < NNodes);
  assert(0 <= nodeId1);
  assert(nodeId1 < NNodes);
  assert(0 <= nodeId2);
  assert(nodeId2 < NNodes);
  assert(0 <= nodeId3);
  assert(nodeId3 < NNodes);

  /* check orientation of tetrahedron */
  {
    double x0 = Node_coord[nodeId0][0];
    double y0 = Node_coord[nodeId0][1];
    double z0 = Node_coord[nodeId0][2];
    
    double x1 = Node_coord[nodeId1][0];
    double y1 = Node_coord[nodeId1][1];
    double z1 = Node_coord[nodeId1][2];
    
    double x2 = Node_coord[nodeId2][0];
    double y2 = Node_coord[nodeId2][1];
    double z2 = Node_coord[nodeId2][2];
    
    double x3 = Node_coord[nodeId3][0];
    double y3 = Node_coord[nodeId3][1];
    double z3 = Node_coord[nodeId3][2];
    
    double volume = SignedTetraVolume (x0, y0, z0, 
				       x1, y1, z1, 
				       x2, y2, z2, 
				       x3, y3, z3);

#if 0
    assert(!Equal2 (volume, 0.0));
#endif

    if (volume < 0.0) {
      int dummy = nodeId0;
      
      nodeId0 = nodeId1;
      nodeId1 = dummy;
    }
  }

  assert(NElements < MAX_ELEMENTS);
  NElements++;

  Element_isDeleted[elementId] = 0;

  Element_nodeIds[elementId][0] = nodeId0;
  Element_nodeIds[elementId][1] = nodeId1;
  Element_nodeIds[elementId][2] = nodeId2;
  Element_nodeIds[elementId][3] = nodeId3;

  return elementId;
}

void Mesh_deleteElement (int elementId)
{
  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(!Element_isDeleted[elementId]);

  Element_isDeleted[elementId] = 1;
}

int Mesh_findOtherElementWithThreeNodes 
(int elementId,
 int nodeId0, int nodeId1, int nodeId2)
{
  int iElement;
  
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) {
      continue;
    }
    
    if (iElement == elementId) {
      continue;
    }
    
    if (Element_containsNode (iElement, nodeId0)
	&& Element_containsNode (iElement, nodeId1)
	&& Element_containsNode (iElement, nodeId2)) {
      return iElement;
    }
  }

  return -1;
}

int Mesh_findElementWithFourNodes 
(int nodeId0, int nodeId1, int nodeId2, int nodeId3)
{
  int iElement;
  
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) {
      continue;
    }
    
    if (Element_containsNode (iElement, nodeId0)
	&& Element_containsNode (iElement, nodeId1)
	&& Element_containsNode (iElement, nodeId2)
	&& Element_containsNode (iElement, nodeId3)) {
      return iElement;
    }
  }

  return -1;
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
    
    Element_isDeleted[iElement] = 0;
  }
  
  fclose (fp);
}

void WriteNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    fprintf (fp, "%d   %22.15e %22.15e %22.15e\n", 
	     iNode,
	     Node_coord[iNode][0],
	     Node_coord[iNode][1],
	     Node_coord[iNode][2]); 
  }

  fclose (fp);
}

void WriteElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;
  int nAllElements;
  int count;

  assert(0 < NNodes);
  assert(0 < NElements);

  nAllElements = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) {
      continue;
    }
    
    nAllElements++;
  }

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   nAllElements);
  
  count = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) {
      continue;
    }

    fprintf (fp, "%d   ", 
	     count); 
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      fprintf (fp, "%d ", 
	       Element_nodeIds[iElement][iElementNode]); 
    }
    fprintf (fp, "\n");
    
    count++;
  }
  assert(count == nAllElements);
  
  fclose (fp);
}



/************************
	 Command
*************************/

static int SliverElementId;
static int FaceNodeIds[N_ELEMENT_FACES][3];
static int NeighborIds[N_ELEMENT_FACES];
static int NeighborNodeIds[N_ELEMENT_FACES];

static int CommonNodeId0;
static int CommonNodeId1;
static int OppositeNodeId0;
static int OppositeNodeId1;

static int IsFinalTurn;

static void SetCommonNodes (int faceId0, int faceId1)
{
  int iElementNode;
  int jElementNode;

  CommonNodeId0 = -1;
  for (iElementNode = 0;
       iElementNode < N_ELEMENT_NODES;
       iElementNode++) {
    int nodeId = FaceNodeIds[faceId0][iElementNode];
    
    for (jElementNode = 0;
	 jElementNode < N_ELEMENT_NODES;
	 jElementNode++) {
      if (FaceNodeIds[faceId1][jElementNode] == nodeId) {
	CommonNodeId0 = nodeId;
	break;
      }
    }
    
    if (CommonNodeId0 != -1) {
      break;
    }
  }
  assert(CommonNodeId0 != -1);

  CommonNodeId1 = -1;
  for (iElementNode = 0;
       iElementNode < N_ELEMENT_NODES;
       iElementNode++) {
    int nodeId = FaceNodeIds[faceId0][iElementNode];

    if (nodeId == CommonNodeId0) {
      continue;
    }

    for (jElementNode = 0;
	 jElementNode < N_ELEMENT_NODES;
	 jElementNode++) {
      if (FaceNodeIds[faceId1][jElementNode] == nodeId) {
	CommonNodeId1 = nodeId;
	break;
      }
    }
    
    if (CommonNodeId1 != -1) {
      break;
    }
  }
  assert(CommonNodeId1 != -1);

  OppositeNodeId0 = -1;
  for (iElementNode = 0;
       iElementNode < N_ELEMENT_NODES;
       iElementNode++) {
    int nodeId = FaceNodeIds[faceId0][iElementNode];

    if (nodeId != CommonNodeId0
	&& nodeId != CommonNodeId1) {
      OppositeNodeId0 = nodeId;
      break;
    }
  }
  assert(OppositeNodeId0 != -1);

  OppositeNodeId1 = -1;
  for (iElementNode = 0;
       iElementNode < N_ELEMENT_NODES;
       iElementNode++) {
    int nodeId = FaceNodeIds[faceId1][iElementNode];

    if (nodeId != CommonNodeId0
	&& nodeId != CommonNodeId1) {
      OppositeNodeId1 = nodeId;
      break;
    }
  }
  assert(OppositeNodeId1 != -1);
}

static int CheckPatternA (int faceId0, int faceId1)
{
  assert(faceId0 != faceId1);

  if (NeighborNodeIds[faceId0] == NeighborNodeIds[faceId1]) {
    int commonNeighborNodeId = NeighborNodeIds[faceId0];

    SetCommonNodes (faceId0, faceId1);

    Mesh_addNewElement (commonNeighborNodeId,
			OppositeNodeId0, OppositeNodeId1,
			CommonNodeId0);
    Mesh_addNewElement (commonNeighborNodeId,
			OppositeNodeId0, OppositeNodeId1,
			CommonNodeId1);

    Mesh_deleteElement (SliverElementId);
    Mesh_deleteElement (NeighborIds[faceId0]);
    Mesh_deleteElement (NeighborIds[faceId1]);

    fprintf (stderr, " pattern A \n");

    return 1;
  }

  return 0;
}

static int CheckPatternB (int faceId0, int faceId1)
{
  int neighborNodeId0 = NeighborNodeIds[faceId0];
  int neighborNodeId1 = NeighborNodeIds[faceId1];
  int commonElementId;

  assert(faceId0 != faceId1);
  assert(neighborNodeId0 != neighborNodeId1);
  
  SetCommonNodes (faceId0, faceId1);
  
  commonElementId = Mesh_findElementWithFourNodes
    (CommonNodeId0, CommonNodeId1, 
     neighborNodeId0, neighborNodeId1);
  
  if (commonElementId != -1) {
    Mesh_addNewElement (neighborNodeId0,
			OppositeNodeId0, 
			OppositeNodeId1,
			CommonNodeId0);
    Mesh_addNewElement (OppositeNodeId1,
			neighborNodeId1,
			neighborNodeId0,
			CommonNodeId0);

    Mesh_addNewElement (neighborNodeId0,
			OppositeNodeId0, 
			OppositeNodeId1,
			CommonNodeId1);
    Mesh_addNewElement (OppositeNodeId1,
			neighborNodeId1,
			neighborNodeId0,
			CommonNodeId1);

    Mesh_deleteElement (SliverElementId);
    Mesh_deleteElement (NeighborIds[faceId0]);
    Mesh_deleteElement (NeighborIds[faceId1]);
    Mesh_deleteElement (commonElementId);

    fprintf (stderr, " pattern B \n");

    return 1;
  }

  return 0;
}

static void RemoveSliver (int elementId)
{
  int iFace;

  SliverElementId = elementId;

  fprintf (stderr, " sliver is found %d \n",
	   SliverElementId);

  for (iFace = 0; iFace < N_ELEMENT_FACES; iFace++) {
    int nodeId0 = Element_nodeIds[SliverElementId][FaceNodeMap[iFace][0]];
    int nodeId1 = Element_nodeIds[SliverElementId][FaceNodeMap[iFace][1]];
    int nodeId2 = Element_nodeIds[SliverElementId][FaceNodeMap[iFace][2]];
    
    FaceNodeIds[iFace][0] = nodeId0;
    FaceNodeIds[iFace][1] = nodeId1;
    FaceNodeIds[iFace][2] = nodeId2;
    
    NeighborIds[iFace] = Mesh_findOtherElementWithThreeNodes 
      (SliverElementId, 
       nodeId0, nodeId1, nodeId2);
    
    if (NeighborIds[iFace] != -1) {
      NeighborNodeIds[iFace] 
	= Element_otherNode (NeighborIds[iFace], 
			     nodeId0, nodeId1, nodeId2);
    } else {
      NeighborNodeIds[iFace] = -1;
    }
  }

  for (iFace = 0; iFace < N_ELEMENT_FACES; iFace++) {
    if (NeighborIds[iFace] == -1) {
      fprintf (stderr, " sliver without neighbor is found \n");

      if (IsFinalTurn) {
	Mesh_deleteElement (SliverElementId);
      }
      return;
    }
  }

  for (iFace = 0; iFace < N_ELEMENT_FACES; iFace++) {
    assert(NeighborNodeIds[iFace] != -1);
  }

  if (CheckPatternA (0, 1)) {
    return;
  }
  if (CheckPatternA (0, 2)) {
    return;
  }
  if (CheckPatternA (0, 3)) {
    return;
  }
  if (CheckPatternA (1, 2)) {
    return;
  }
  if (CheckPatternA (1, 3)) {
    return;
  }
  if (CheckPatternA (2, 3)) {
    return;
  }

  if (CheckPatternB (0, 1)) {
    return;
  }
  if (CheckPatternB (0, 2)) {
    return;
  }
  if (CheckPatternB (0, 3)) {
    return;
  }
  if (CheckPatternB (1, 2)) {
    return;
  }
  if (CheckPatternB (1, 3)) {
    return;
  }
  if (CheckPatternB (2, 3)) {
    return;
  }

  fprintf (stderr, " failed to remove sliver \n");
}

void PerformCommand (void)
{
  int iElement;
  
  IsFinalTurn = 0;

  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) {
      continue;
    }

    {
      double volume = Element_volume (iElement);
      
#if 0
      assert(0.0 < volume);
#endif
      
      if (Equal2 (volume, 0.0)) {
	RemoveSliver (iElement);
      }
    }
  }

  fprintf (stderr, " first turn finished \n");

  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) {
      continue;
    }

    {
      double volume = Element_volume (iElement);
      
#if 0
      assert(0.0 < volume);
#endif
      
      if (Equal2 (volume, 0.0)) {
	RemoveSliver (iElement);
      }
    }
  }

  fprintf (stderr, " second turn finished \n");

  IsFinalTurn = 1;

  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) {
      continue;
    }

    {
      double volume = Element_volume (iElement);
      
#if 0
      assert(0.0 < volume);
#endif
      
      if (Equal2 (volume, 0.0)) {
	RemoveSliver (iElement);
      }
    }
  }

  fprintf (stderr, " third turn finished \n");
}



int main (int argc, char *argv[])
{
  if (argc != 5) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tetra3d_sliver  node_file  element_file \n"
       "          node_file_OUT  element_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);

  PerformCommand ();

  WriteNodeFile (argv[3]);
  WriteElementFile (argv[4]);

  return 0;
}


