/* advautomesh_h_tetra3d_swap_boundary.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001
#define AREA_COORD_EPS 0.00001
#define VOLUME_COORD_EPS 0.00001

double Eps = DEFAULT_EPS;



int Equal (double value0, double value1) 
{
  return fabs (value1 - value0) < Eps;
}

int Equal2 (double value0, double value1) 
{
  return fabs (value1 - value0) < Eps * Eps;
}

int Between (double value, double range0, double range1) 
{
  return (range0 - Eps < value && value < range1 + Eps)
    || (range1 - Eps < value && value < range0 + Eps);
}

double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}

double SignedTetraVolume
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double x2, double y2, double z2,
 double x3, double y3, double z3)
{
  double x10 = x1 - x0;
  double y10 = y1 - y0;
  double z10 = z1 - z0;

  double x20 = x2 - x0;
  double y20 = y2 - y0;
  double z20 = z2 - z0;
  
  double x30 = x3 - x0;
  double y30 = y3 - y0;
  double z30 = z3 - z0;
  
  double d23 = y20 * z30 - z20 * y30;
  double d31 = y30 * z10 - z30 * y10;
  double d12 = y10 * z20 - z10 * y20;
  
  return (x10 * d23 + x20 * d31 + x30 * d12) / 6;
}

void GetNormalizedVector 
(double x, double y, double z,
 double *nx_OUT, double *ny_OUT, double *nz_OUT)
{
  double length = sqrt (x * x + y * y + z * z);

  *nx_OUT = x / length;
  *ny_OUT = y / length;
  *nz_OUT = z / length;
}

void GetVectorProduct 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  *x_OUT = y0 * z1 - z0 * y1;
  *y_OUT = z0 * x1 - x0 * z1;
  *z_OUT = x0 * y1 - y0 * x1;
}

void GetPlaneEquation 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double x2, double y2, double z2,
 double *a_OUT, double *b_OUT, double *c_OUT, double *d_OUT)
{
  double nx, ny, nz;

  GetVectorProduct (x1 - x0, y1 - y0, z1 - z0,
		    x2 - x0, y2 - y0, z2 - z0,
		    &nx, &ny, &nz);
  GetNormalizedVector (nx, ny, nz,
		       &nx, &ny, &nz);
  *a_OUT = nx;
  *b_OUT = ny;
  *c_OUT = nz;
  *d_OUT = -(nx * x0 + ny * y0 + nz * z0);
}

int GetIntersectionBetweenPlaneAndSeg 
(double a, double b, double c, double d, 
 double sx0, double sy0, double sz0, 
 double sx1, double sy1, double sz1, 
 double *intX_OUT, double *intY_OUT, double *intZ_OUT) 
{
  double value0 = a * sx0 + b * sy0 + c * sz0 + d;
  double value1 = a * sx1 + b * sy1 + c * sz1 + d;

  if (fabs (value0 - value1) < Eps) return 0;

  if (fabs (value0) < Eps) {
    *intX_OUT = sx0;
    *intY_OUT = sy0;
    *intZ_OUT = sz0;
    return 1;
  } else if (fabs (value1) < Eps) {
    *intX_OUT = sx1;
    *intY_OUT = sy1;
    *intZ_OUT = sz1;
    return 1;
  } else if (value0 * value1 < Eps * Eps) {
    *intX_OUT = (sx1 * value0 - sx0 * value1) / (value0 - value1);
    *intY_OUT = (sy1 * value0 - sy0 * value1) / (value0 - value1);
    *intZ_OUT = (sz1 * value0 - sz0 * value1) / (value0 - value1);
    return 1;
  }
  return 0;
}

double TriArea
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double x2, double y2, double z2)
{
  double vx, vy, vz;
  
  GetVectorProduct (x1 - x0, y1 - y0, z1 - z0,
		    x2 - x0, y2 - y0, z2 - z0,
		    &vx, &vy, &vz);
  return sqrt (vx * vx + vy * vy + vz * vz) / 2;
}

int TriContainsPoint 
(double tx0, double ty0, double tz0, 
 double tx1, double ty1, double tz1, 
 double tx2, double ty2, double tz2, 
 double x, double y, double z)
{
  double area = TriArea (tx0, ty0, tz0, 
			 tx1, ty1, tz1, 
			 tx2, ty2, tz2);
  double area0 = TriArea (tx1, ty1, tz1, 
			  tx2, ty2, tz2, 
			  x, y, z);
  double area1 = TriArea (tx2, ty2, tz2, 
			  tx0, ty0, tz0, 
			  x, y, z);
  double area2 = TriArea (tx0, ty0, tz0, 
			  tx1, ty1, tz1, 
			  x, y, z);
  
  if (fabs ((area0 + area1 + area2) / area - 1.0) < AREA_COORD_EPS) {
    return 1;
  }
  return 0;
}

int GetIntersectionBetweenTriAndSeg 
(double tx0, double ty0, double tz0, 
 double tx1, double ty1, double tz1, 
 double tx2, double ty2, double tz2, 
 double sx0, double sy0, double sz0, 
 double sx1, double sy1, double sz1, 
 double *intX_OUT, double *intY_OUT, double *intZ_OUT) 
{
  double a, b, c, d;
  int flag;
  
  GetPlaneEquation (tx0, ty0, tz0, 
		    tx1, ty1, tz1, 
		    tx2, ty2, tz2,
		    &a, &b, &c, &d);
  flag = GetIntersectionBetweenPlaneAndSeg 
    (a, b, c, d,
     sx0, sy0, sz0, 
     sx1, sy1, sz1, 
     intX_OUT, intY_OUT, intZ_OUT);
  if (flag) {
    if (TriContainsPoint (tx0, ty0, tz0, 
			  tx1, ty1, tz1, 
			  tx2, ty2, tz2,
			  *intX_OUT, *intY_OUT, *intZ_OUT)) {
      return 1;
    }
  }

  return 0;
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][3];
int Node_isAdded[MAX_NODES];



/* Element */

#define N_ELEMENT_NODES 4
#define N_ELEMENT_EDGES 6
#define MAX_ELEMENTS 5000000

const int EdgeNodeMap[N_ELEMENT_EDGES][2] = {
  { 0, 1 }, 
  { 0, 2 }, 
  { 0, 3 }, 
  { 1, 2 }, 
  { 2, 3 }, 
  { 3, 1 }  
};

int NElements;
int Element_isDeleted[MAX_ELEMENTS];
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];
int Element_intersectionIsChecked[MAX_ELEMENTS];



double Element_volume (int elementId)
{
  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(!Element_isDeleted[elementId]);

  {
    int nodeId0 = Element_nodeIds[elementId][0];
    int nodeId1 = Element_nodeIds[elementId][1];
    int nodeId2 = Element_nodeIds[elementId][2];
    int nodeId3 = Element_nodeIds[elementId][3];
    
    double x0 = Node_coord[nodeId0][0];
    double y0 = Node_coord[nodeId0][1];
    double z0 = Node_coord[nodeId0][2];
    
    double x1 = Node_coord[nodeId1][0];
    double y1 = Node_coord[nodeId1][1];
    double z1 = Node_coord[nodeId1][2];
    
    double x2 = Node_coord[nodeId2][0];
    double y2 = Node_coord[nodeId2][1];
    double z2 = Node_coord[nodeId2][2];
    
    double x3 = Node_coord[nodeId3][0];
    double y3 = Node_coord[nodeId3][1];
    double z3 = Node_coord[nodeId3][2];
    
    double volume 
      = SignedTetraVolume (x0, y0, z0, 
			   x1, y1, z1, 
			   x2, y2, z2, 
			   x3, y3, z3);

    assert(!Equal2 (volume, 0.0));
    assert(0.0 < volume);

    return volume;
  }
}

int Element_containsPoint 
(int elementId, 
 double x, double y, double z,
 double *L0_OUT, double *L1_OUT, double *L2_OUT, double *L3_OUT)
{
  int nodeId0, nodeId1, nodeId2, nodeId3;
  double x0, y0, z0;
  double x1, y1, z1;
  double x2, y2, z2;
  double x3, y3, z3;
  double volume;
  double volume0, volume1, volume2, volume3;

  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(!Element_isDeleted[elementId]);

  nodeId0 = Element_nodeIds[elementId][0];
  nodeId1 = Element_nodeIds[elementId][1];
  nodeId2 = Element_nodeIds[elementId][2];
  nodeId3 = Element_nodeIds[elementId][3];

  x0 = Node_coord[nodeId0][0];
  y0 = Node_coord[nodeId0][1];
  z0 = Node_coord[nodeId0][2];

  x1 = Node_coord[nodeId1][0];
  y1 = Node_coord[nodeId1][1];
  z1 = Node_coord[nodeId1][2];

  x2 = Node_coord[nodeId2][0];
  y2 = Node_coord[nodeId2][1];
  z2 = Node_coord[nodeId2][2];
  
  x3 = Node_coord[nodeId3][0];
  y3 = Node_coord[nodeId3][1];
  z3 = Node_coord[nodeId3][2];
  
  volume = SignedTetraVolume (x0, y0, z0, 
			      x1, y1, z1, 
			      x2, y2, z2, 
			      x3, y3, z3);
  volume0 = SignedTetraVolume (x, y, z, 
			       x1, y1, z1, 
			       x2, y2, z2, 
			       x3, y3, z3);
  volume1 = SignedTetraVolume (x0, y0, z0, 
			       x, y, z, 
			       x2, y2, z2, 
			       x3, y3, z3);
  volume2 = SignedTetraVolume (x0, y0, z0, 
			       x1, y1, z1, 
			       x, y, z, 
			       x3, y3, z3);
  volume3 = SignedTetraVolume (x0, y0, z0, 
			       x1, y1, z1, 
			       x2, y2, z2, 
			       x, y, z);

  *L0_OUT = volume0 / volume;
  *L1_OUT = volume1 / volume;
  *L2_OUT = volume2 / volume;
  *L3_OUT = volume3 / volume;

  if (0.0 - VOLUME_COORD_EPS < *L0_OUT 
      && *L0_OUT < 1.0 + VOLUME_COORD_EPS
      && 0.0 - VOLUME_COORD_EPS < *L1_OUT 
      && *L1_OUT < 1.0 + VOLUME_COORD_EPS
      && 0.0 - VOLUME_COORD_EPS < *L2_OUT 
      && *L2_OUT < 1.0 + VOLUME_COORD_EPS
      && 0.0 - VOLUME_COORD_EPS < *L3_OUT 
      && *L3_OUT < 1.0 + VOLUME_COORD_EPS) {
    return 1;
  }
  return 0;
}

int Element_containsNode 
(int elementId, 
 int nodeId)
{
  int iElementNode;
  
  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(!Element_isDeleted[elementId]);
  
  for (iElementNode = 0;
       iElementNode < N_ELEMENT_NODES;
       iElementNode++) {
    if (Element_nodeIds[elementId][iElementNode] == nodeId) {
      return 1;
    }
  }
  return 0;
}

int Element_elementNodeId 
(int elementId, 
 int nodeId)
{
  int iElementNode;
  
  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(!Element_isDeleted[elementId]);
  
  for (iElementNode = 0;
       iElementNode < N_ELEMENT_NODES;
       iElementNode++) {
    if (Element_nodeIds[elementId][iElementNode] == nodeId) {
      return iElementNode;
    }
  }
  return -1;
}

void Element_getOtherTwoNodes 
(int elementId, 
 int nodeId0, int nodeId1,
 int *otherNodeId0_OUT, int *otherNodeId1_OUT)
{
  int otherNodeId0, otherNodeId1;
  int iElementEdge;
  
  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(!Element_isDeleted[elementId]);
  
  assert(Element_containsNode (elementId, nodeId0));
  assert(Element_containsNode (elementId, nodeId1));

  otherNodeId0 = -1;
  otherNodeId1 = -1;
  for (iElementEdge = 0; 
       iElementEdge < N_ELEMENT_EDGES; 
       iElementEdge++) {
    int n0 = Element_nodeIds[elementId][EdgeNodeMap[iElementEdge][0]];
    int n1 = Element_nodeIds[elementId][EdgeNodeMap[iElementEdge][1]];
    
    if (n0 != nodeId0
	&& n0 != nodeId1
	&& n1 != nodeId0
	&& n1 != nodeId1) {
      otherNodeId0 = n0;
      otherNodeId1 = n1;
      break;
    }
  }
  assert(otherNodeId0 != -1);
  assert(otherNodeId1 != -1);

  *otherNodeId0_OUT = otherNodeId0;
  *otherNodeId1_OUT = otherNodeId1;
}

int Element_otherNode 
(int elementId, 
 int nodeId0, int nodeId1, int nodeId2)
{
  int iElementNode;
  
  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(!Element_isDeleted[elementId]);
  
  if (Element_containsNode (elementId, nodeId0)
      && Element_containsNode (elementId, nodeId1)
      && Element_containsNode (elementId, nodeId2)) {
    for (iElementNode = 0;
	 iElementNode < N_ELEMENT_NODES;
	 iElementNode++) {
      int nodeId = Element_nodeIds[elementId][iElementNode];
      
      if (nodeId != nodeId0
	  && nodeId != nodeId1
	  && nodeId != nodeId2) {
	return nodeId;
      }
    }
    assert(0);
    return -1;
  }
  return -1;
}

void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  *minRangeZ_OUT = *maxRangeZ_OUT = Node_coord[0][2];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    double z = Node_coord[iNode][2];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
    if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
  }
}

int Mesh_findNodeAt (double x, double y, double z)
{
  int iNode;

  for (iNode = 0; iNode < NNodes; iNode++) {
    double nodeX = Node_coord[iNode][0];
    double nodeY = Node_coord[iNode][1];
    double nodeZ = Node_coord[iNode][2];

    if (Distance (x, y, z, 
		  nodeX, nodeY, nodeZ) < Eps) {
      return iNode;
    }
  }

  return -1;
}

int Mesh_addNewNode 
(double x, double y, double z)
{
  int nodeId = NNodes;

  assert(NNodes < MAX_NODES);
  NNodes++;

  Node_coord[nodeId][0] = x;
  Node_coord[nodeId][1] = y;
  Node_coord[nodeId][2] = z;

  Node_isAdded[nodeId] = 1;
  
  return nodeId;
}

int Mesh_addNewElement (int nodeId0, int nodeId1, int nodeId2, int nodeId3)
{
  int elementId = NElements;

  assert(0 <= nodeId0);
  assert(nodeId0 < NNodes);
  assert(0 <= nodeId1);
  assert(nodeId1 < NNodes);
  assert(0 <= nodeId2);
  assert(nodeId2 < NNodes);
  assert(0 <= nodeId3);
  assert(nodeId3 < NNodes);

  /* check orientation of tetrahedron */
  {
    double x0 = Node_coord[nodeId0][0];
    double y0 = Node_coord[nodeId0][1];
    double z0 = Node_coord[nodeId0][2];
    
    double x1 = Node_coord[nodeId1][0];
    double y1 = Node_coord[nodeId1][1];
    double z1 = Node_coord[nodeId1][2];
    
    double x2 = Node_coord[nodeId2][0];
    double y2 = Node_coord[nodeId2][1];
    double z2 = Node_coord[nodeId2][2];
    
    double x3 = Node_coord[nodeId3][0];
    double y3 = Node_coord[nodeId3][1];
    double z3 = Node_coord[nodeId3][2];
    
    double volume = SignedTetraVolume (x0, y0, z0, 
				       x1, y1, z1, 
				       x2, y2, z2, 
				       x3, y3, z3);

    assert(!Equal2 (volume, 0.0));

    if (volume < 0.0) {
      int dummy = nodeId0;
      
      nodeId0 = nodeId1;
      nodeId1 = dummy;
    }
  }

  assert(NElements < MAX_ELEMENTS);
  NElements++;

  Element_isDeleted[elementId] = 0;

  Element_nodeIds[elementId][0] = nodeId0;
  Element_nodeIds[elementId][1] = nodeId1;
  Element_nodeIds[elementId][2] = nodeId2;
  Element_nodeIds[elementId][3] = nodeId3;

  Element_intersectionIsChecked[elementId] = 0;

  return elementId;
}

void Mesh_deleteElement (int elementId)
{
  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(!Element_isDeleted[elementId]);

  Element_isDeleted[elementId] = 1;
}

double Mesh_volume (void)
{
  int iElement;
  double totalVolume;

  totalVolume = 0.0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) continue;

    totalVolume += Element_volume (iElement);
  }

  return totalVolume;
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 

    Node_isAdded[iNode] = 0;
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }

    Element_isDeleted[iElement] = 0;
    Element_intersectionIsChecked[iElement] = 0;
  }
  
  fclose (fp);
}

void WriteNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    fprintf (fp, "%d   %22.15e %22.15e %22.15e\n", 
	     iNode,
	     Node_coord[iNode][0],
	     Node_coord[iNode][1],
	     Node_coord[iNode][2]); 
  }

  fclose (fp);
}

void WriteElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;
  int nActualElements;
  int count;

  assert(0 < NNodes);
  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  nActualElements = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) continue;
    
    nActualElements++;
  }

  fprintf (fp, "%d\n", 
	   nActualElements);
  
  count = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) continue;

    fprintf (fp, "%d   ", 
	     count); 
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      fprintf (fp, "%d ", 
	       Element_nodeIds[iElement][iElementNode]); 
    }
    fprintf (fp, "\n");
    count++;
  }
  
  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Triangle */

#define MAX_TRIS 100000

int NTris;
double Tri_coord0[MAX_TRIS][3];
double Tri_coord1[MAX_TRIS][3];
double Tri_coord2[MAX_TRIS][3];



void ReadTriFile (const char *fileName)
{
  FILE *fp;
  int iTri;

  assert(NTris == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NTris);
  assert(0 <= NTris);
  assert(NTris <= MAX_TRIS);
  
  for (iTri = 0; iTri < NTris; iTri++) {
    int triId;
    double x0, y0, z0, x1, y1, z1, x2, y2, z2;

    fscanf (fp, "%d %lf %lf %lf %lf %lf %lf %lf %lf %lf", 
	    &triId,
	    &x0, &y0, &z0, 
	    &x1, &y1, &z1, 
	    &x2, &y2, &z2);
    assert(triId == iTri);

    Tri_coord0[iTri][0] = x0;
    Tri_coord0[iTri][1] = y0; 
    Tri_coord0[iTri][2] = z0; 

    Tri_coord1[iTri][0] = x1;
    Tri_coord1[iTri][1] = y1; 
    Tri_coord1[iTri][2] = z1; 

    Tri_coord2[iTri][0] = x2;
    Tri_coord2[iTri][1] = y2; 
    Tri_coord2[iTri][2] = z2; 
  }

  fclose (fp);
}



/************************
	 Kernel
*************************/

#define MAX_KERNELS 1000000
#define MAX_KERNEL_ELEMENTS 10

int NKernels;
int Kernel_axisNodeId0[MAX_KERNELS];
int Kernel_axisNodeId1[MAX_KERNELS];
int Kernel_nElements[MAX_KERNELS];
int Kernel_elementIds[MAX_KERNELS][MAX_KERNEL_ELEMENTS];
int Kernel_sortedElementIds[MAX_KERNELS][MAX_KERNEL_ELEMENTS];
int Kernel_sortedNodeIds[MAX_KERNELS][MAX_KERNEL_ELEMENTS];
int Kernel_isConverted[MAX_KERNELS];

int Kernel_isValid 
(int kernelId)
{
  int nKernelElements;
  int iKernelElement;

  assert(0 <= kernelId);
  assert(kernelId < NKernels);

  nKernelElements = Kernel_nElements[kernelId];

  for (iKernelElement = 0;
       iKernelElement < nKernelElements;
       iKernelElement++) {
    int elementId = Kernel_elementIds[kernelId][iKernelElement];

    if (Element_isDeleted[elementId]) {
      return 0;
    }
  }
  return 1;
}

void Kernel_Clear (void)
{
  NKernels = 0;
} 

int Kernel_AddNewKernel 
(int nodeId0, int nodeId1)
{
  int kernelId;

  assert(0 <= nodeId0);
  assert(nodeId0 < NNodes);
  assert(0 <= nodeId1);
  assert(nodeId1 < NNodes);
  assert(nodeId0 != nodeId1);

  kernelId = NKernels;

  assert(NKernels < MAX_KERNELS);
  NKernels++;

  Kernel_axisNodeId0[kernelId] = nodeId0;
  Kernel_axisNodeId1[kernelId] = nodeId1;
  Kernel_nElements[kernelId] = 0;

  Kernel_isConverted[kernelId] = 0;

  assert(!Kernel_isConverted[kernelId]);
  assert(Kernel_isValid (kernelId));

  return kernelId;
}

static void AddElementIntoKernel
(int kernelId,
 int elementId)
{
  int kernelElementId;

  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(!Element_isDeleted[elementId]);

  assert(Element_containsNode (elementId, 
			       Kernel_axisNodeId0[kernelId]));
  assert(Element_containsNode (elementId, 
			       Kernel_axisNodeId1[kernelId]));

  kernelElementId = Kernel_nElements[kernelId];

  assert(Kernel_nElements[kernelId] < MAX_KERNEL_ELEMENTS);
  Kernel_nElements[kernelId]++;

  Kernel_elementIds[kernelId][kernelElementId] = elementId;
}

static void CollectKernelElement 
(int kernelId)
{
  int axisNodeId0 = Kernel_axisNodeId0[kernelId];
  int axisNodeId1 = Kernel_axisNodeId1[kernelId];
  int iElement;

  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) continue;
    
    if (Element_containsNode (iElement, axisNodeId0)
	&& Element_containsNode (iElement, axisNodeId1)) {
      AddElementIntoKernel (kernelId, iElement);
    }
  }

  assert(Kernel_isValid (kernelId));
}

static void SortKernelElement 
(int kernelId)
{
  int axisNodeId0 = Kernel_axisNodeId0[kernelId];
  int axisNodeId1 = Kernel_axisNodeId1[kernelId];
  int nKernelElements = Kernel_nElements[kernelId];
  int iKernelElement;
  int jKernelElement;
  int firstElementId;
  int firstNodeId;
  int secondNodeId;
  int prevElementId;
  int prevNodeId;
  
  assert(2 < nKernelElements);

  for (iKernelElement = 0;
       iKernelElement < nKernelElements;
       iKernelElement++) {
    Kernel_sortedElementIds[kernelId][iKernelElement] = -1;
    Kernel_sortedNodeIds[kernelId][iKernelElement] = -1;
  }

  firstElementId = Kernel_elementIds[kernelId][0];

  {
    int otherNodeId0, otherNodeId1;
    
    Element_getOtherTwoNodes 
      (firstElementId,
       axisNodeId0, axisNodeId1,
       &otherNodeId0, &otherNodeId1);
    
    firstNodeId = otherNodeId0;
    secondNodeId = otherNodeId1;
  }
  
  Kernel_sortedElementIds[kernelId][0] = firstElementId;

  Kernel_sortedNodeIds[kernelId][0] = firstNodeId;
  Kernel_sortedNodeIds[kernelId][1] = secondNodeId;

  prevElementId = firstElementId;
  prevNodeId = secondNodeId;

  for (iKernelElement = 1;
       iKernelElement < nKernelElements;
       iKernelElement++) {
    int elementIdFound;
    int nodeIdFound;

    elementIdFound = -1;
    nodeIdFound = -1;
    for (jKernelElement = 0;
	 jKernelElement < nKernelElements;
	 jKernelElement++) {
      int elementId = Kernel_elementIds[kernelId][jKernelElement];
      
      if (elementId == prevElementId) continue;
      
      nodeIdFound = Element_otherNode 
	(elementId, 
	 axisNodeId0, axisNodeId1, prevNodeId);
      if (nodeIdFound != -1) {
	elementIdFound = elementId;
	break;
      }
    }
    assert(elementIdFound != -1);
    assert(nodeIdFound != -1);

    if (iKernelElement == nKernelElements - 1) {
      assert(nodeIdFound == firstNodeId);

      Kernel_sortedElementIds[kernelId][iKernelElement] = elementIdFound;
    } else {
      assert(iKernelElement < nKernelElements - 1);

      Kernel_sortedElementIds[kernelId][iKernelElement] = elementIdFound;
      Kernel_sortedNodeIds[kernelId][iKernelElement + 1] = nodeIdFound;
    }
    
    prevNodeId = nodeIdFound;
    prevElementId = elementIdFound;
  }

  /* verification */
  for (iKernelElement = 0;
       iKernelElement < nKernelElements;
       iKernelElement++) {
    int sortedElementId = Kernel_sortedElementIds[kernelId][iKernelElement];
    int found;

    found = 0;
    for (jKernelElement = 0;
	 jKernelElement < nKernelElements;
	 jKernelElement++) {
      int elementId = Kernel_elementIds[kernelId][jKernelElement];
      
      if (elementId == sortedElementId) {
	found = 1;
	break;
      }
    }
    assert(found);

    {
      int v0 = iKernelElement;
      int v1 = (iKernelElement + 1) % nKernelElements;
      
      int nodeId0 = Kernel_sortedNodeIds[kernelId][v0];
      int nodeId1 = Kernel_sortedNodeIds[kernelId][v1];

      assert(Element_containsNode (sortedElementId, nodeId0));
      assert(Element_containsNode (sortedElementId, nodeId1));
    }
  }

  assert(Kernel_isValid (kernelId));
}

void Kernel_collectElement 
(int kernelId)
{
  assert(0 <= kernelId);
  assert(kernelId < NKernels);
  
  assert(!Kernel_isConverted[kernelId]);
  assert(Kernel_isValid (kernelId));

  CollectKernelElement (kernelId);
  SortKernelElement (kernelId);

  assert(!Kernel_isConverted[kernelId]);
  assert(Kernel_isValid (kernelId));
}

static void DeleteKernelElement 
(int kernelId)
{
  int nKernelElements = Kernel_nElements[kernelId];
  int iKernelElement;
  
  for (iKernelElement = 0;
       iKernelElement < nKernelElements;
       iKernelElement++) {
    int elementId = Kernel_sortedElementIds[kernelId][iKernelElement];

    Mesh_deleteElement (elementId);
  }
}

static int KernelContainsPoint 
(int kernelId, 
 double px, double py, double pz)
{
  int nKernelElements = Kernel_nElements[kernelId];
  int iKernelElement;
  
  for (iKernelElement = 0;
       iKernelElement < nKernelElements;
       iKernelElement++) {
    int elementId = Kernel_sortedElementIds[kernelId][iKernelElement];
    double L0, L1, L2, L3;

    if (Element_containsPoint (elementId,
			       px, py, pz,
			       &L0, &L1, &L2, &L3)) {
      return 1;
    }
  }

  return 0;
}

static int VolumeIsNotZero 
(int nodeId0, int nodeId1, int nodeId2, int nodeId3)
{
  double x0 = Node_coord[nodeId0][0];
  double y0 = Node_coord[nodeId0][1];
  double z0 = Node_coord[nodeId0][2];
  
  double x1 = Node_coord[nodeId1][0];
  double y1 = Node_coord[nodeId1][1];
  double z1 = Node_coord[nodeId1][2];
  
  double x2 = Node_coord[nodeId2][0];
  double y2 = Node_coord[nodeId2][1];
  double z2 = Node_coord[nodeId2][2];
  
  double x3 = Node_coord[nodeId3][0];
  double y3 = Node_coord[nodeId3][1];
  double z3 = Node_coord[nodeId3][2];
  
  double volume 
    = SignedTetraVolume (x0, y0, z0, 
			 x1, y1, z1, 
			 x2, y2, z2, 
			 x3, y3, z3);
  
  if (Equal2 (volume, 0.0)) {
    return 0;
  }
  return 1;
}

static void Convert3NodeKernel 
(int kernelId)
{
  int nKernelElements = Kernel_nElements[kernelId];
  int axisNodeId0;
  int axisNodeId1;
  int nodeId0, nodeId1, nodeId2;

  assert(nKernelElements == 3);

  axisNodeId0 = Kernel_axisNodeId0[kernelId];
  axisNodeId1 = Kernel_axisNodeId1[kernelId];

  nodeId0 = Kernel_sortedNodeIds[kernelId][0];
  nodeId1 = Kernel_sortedNodeIds[kernelId][1];
  nodeId2 = Kernel_sortedNodeIds[kernelId][2];

  DeleteKernelElement (kernelId);

  Mesh_addNewElement (axisNodeId0, nodeId0, nodeId1, nodeId2);
  Mesh_addNewElement (axisNodeId1, nodeId0, nodeId1, nodeId2);

  Kernel_isConverted[kernelId] = 1;
}

#define CHECK_RATIO 0.99

static void Convert4NodeKernel 
(int kernelId)
{
  int nKernelElements = Kernel_nElements[kernelId];
  int axisNodeId0;
  int axisNodeId1;
  int iKernelNode;
  int success;

  assert(nKernelElements == 4);

  axisNodeId0 = Kernel_axisNodeId0[kernelId];
  axisNodeId1 = Kernel_axisNodeId1[kernelId];

  success = 0;
  for (iKernelNode = 0;
       iKernelNode < 2;
       iKernelNode++) {
    int v0 = (0 + iKernelNode) % nKernelElements;
    int v1 = (1 + iKernelNode) % nKernelElements;
    int v2 = (2 + iKernelNode) % nKernelElements;
    int v3 = (3 + iKernelNode) % nKernelElements;

    int nodeId0 = Kernel_sortedNodeIds[kernelId][v0];
    int nodeId1 = Kernel_sortedNodeIds[kernelId][v1];
    int nodeId2 = Kernel_sortedNodeIds[kernelId][v2];
    int nodeId3 = Kernel_sortedNodeIds[kernelId][v3];

    double x0 = Node_coord[nodeId0][0];
    double y0 = Node_coord[nodeId0][1];
    double z0 = Node_coord[nodeId0][2];

    double x2 = Node_coord[nodeId2][0];
    double y2 = Node_coord[nodeId2][1];
    double z2 = Node_coord[nodeId2][2];

    double s02x0 = x0 + (x2 - x0) * CHECK_RATIO;
    double s02y0 = y0 + (y2 - y0) * CHECK_RATIO;
    double s02z0 = z0 + (z2 - z0) * CHECK_RATIO;

    double s02x1 = x2 + (x0 - x2) * CHECK_RATIO;
    double s02y1 = y2 + (y0 - y2) * CHECK_RATIO;
    double s02z1 = z2 + (z0 - z2) * CHECK_RATIO;

    if (KernelContainsPoint (kernelId, 
			     s02x0, s02y0, s02z0)
	&& KernelContainsPoint (kernelId, 
				s02x1, s02y1, s02z1)
	&& VolumeIsNotZero (axisNodeId0, nodeId0, nodeId1, nodeId2)
	&& VolumeIsNotZero (axisNodeId0, nodeId2, nodeId3, nodeId0)
	&& VolumeIsNotZero (axisNodeId1, nodeId0, nodeId1, nodeId2)
	&& VolumeIsNotZero (axisNodeId1, nodeId2, nodeId3, nodeId0)) {
      
      DeleteKernelElement (kernelId);
      
      Mesh_addNewElement (axisNodeId0, nodeId0, nodeId1, nodeId2);
      Mesh_addNewElement (axisNodeId0, nodeId2, nodeId3, nodeId0);
      
      Mesh_addNewElement (axisNodeId1, nodeId0, nodeId1, nodeId2);
      Mesh_addNewElement (axisNodeId1, nodeId2, nodeId3, nodeId0);
      
      success = 1;
      break;
    }
  }

  Kernel_isConverted[kernelId] = 1;
}

static void Convert5NodeKernel 
(int kernelId)
{
  int nKernelElements = Kernel_nElements[kernelId];
  int axisNodeId0;
  int axisNodeId1;
  int iKernelNode;
  int success;

  assert(nKernelElements == 5);

  axisNodeId0 = Kernel_axisNodeId0[kernelId];
  axisNodeId1 = Kernel_axisNodeId1[kernelId];

  success = 0;
  for (iKernelNode = 0;
       iKernelNode < nKernelElements;
       iKernelNode++) {
    int v0 = (0 + iKernelNode) % nKernelElements;
    int v1 = (1 + iKernelNode) % nKernelElements;
    int v2 = (2 + iKernelNode) % nKernelElements;
    int v3 = (3 + iKernelNode) % nKernelElements;
    int v4 = (4 + iKernelNode) % nKernelElements;

    int nodeId0 = Kernel_sortedNodeIds[kernelId][v0];
    int nodeId1 = Kernel_sortedNodeIds[kernelId][v1];
    int nodeId2 = Kernel_sortedNodeIds[kernelId][v2];
    int nodeId3 = Kernel_sortedNodeIds[kernelId][v3];
    int nodeId4 = Kernel_sortedNodeIds[kernelId][v4];

    double x0 = Node_coord[nodeId0][0];
    double y0 = Node_coord[nodeId0][1];
    double z0 = Node_coord[nodeId0][2];

    double x2 = Node_coord[nodeId2][0];
    double y2 = Node_coord[nodeId2][1];
    double z2 = Node_coord[nodeId2][2];

    double x4 = Node_coord[nodeId4][0];
    double y4 = Node_coord[nodeId4][1];
    double z4 = Node_coord[nodeId4][2];

    double s02x0 = x0 + (x2 - x0) * CHECK_RATIO;
    double s02y0 = y0 + (y2 - y0) * CHECK_RATIO;
    double s02z0 = z0 + (z2 - z0) * CHECK_RATIO;

    double s02x1 = x2 + (x0 - x2) * CHECK_RATIO;
    double s02y1 = y2 + (y0 - y2) * CHECK_RATIO;
    double s02z1 = z2 + (z0 - z2) * CHECK_RATIO;

    double s24x0 = x2 + (x4 - x2) * CHECK_RATIO;
    double s24y0 = y2 + (y4 - y2) * CHECK_RATIO;
    double s24z0 = z2 + (z4 - z2) * CHECK_RATIO;

    double s24x1 = x4 + (x2 - x4) * CHECK_RATIO;
    double s24y1 = y4 + (y2 - y4) * CHECK_RATIO;
    double s24z1 = z4 + (z2 - z4) * CHECK_RATIO;

    if (KernelContainsPoint (kernelId, 
			     s02x0, s02y0, s02z0)
	&& KernelContainsPoint (kernelId, 
				s02x1, s02y1, s02z1)
	&& KernelContainsPoint (kernelId, 
				s24x0, s24y0, s24z0)
	&& KernelContainsPoint (kernelId, 
				s24x1, s24y1, s24z1)
	&& VolumeIsNotZero (axisNodeId0, nodeId0, nodeId1, nodeId2)
	&& VolumeIsNotZero (axisNodeId0, nodeId2, nodeId4, nodeId0)
	&& VolumeIsNotZero (axisNodeId0, nodeId2, nodeId3, nodeId4)
	&& VolumeIsNotZero (axisNodeId1, nodeId0, nodeId1, nodeId2)
	&& VolumeIsNotZero (axisNodeId1, nodeId2, nodeId4, nodeId0)
	&& VolumeIsNotZero (axisNodeId1, nodeId2, nodeId3, nodeId4)) {
      
      DeleteKernelElement (kernelId);
      
      Mesh_addNewElement (axisNodeId0, nodeId0, nodeId1, nodeId2);
      Mesh_addNewElement (axisNodeId0, nodeId2, nodeId4, nodeId0);
      Mesh_addNewElement (axisNodeId0, nodeId2, nodeId3, nodeId4);
      
      Mesh_addNewElement (axisNodeId1, nodeId0, nodeId1, nodeId2);
      Mesh_addNewElement (axisNodeId1, nodeId2, nodeId4, nodeId0);
      Mesh_addNewElement (axisNodeId1, nodeId2, nodeId3, nodeId4);
      
      success = 1;
      break;
    }
  }

  Kernel_isConverted[kernelId] = 1;
}

static void Convert6NodeKernel 
(int kernelId)
{
  int nKernelElements = Kernel_nElements[kernelId];
  int axisNodeId0;
  int axisNodeId1;
  int iKernelNode;
  int success;

  assert(nKernelElements == 6);

  axisNodeId0 = Kernel_axisNodeId0[kernelId];
  axisNodeId1 = Kernel_axisNodeId1[kernelId];

  success = 0;
  for (iKernelNode = 0;
       iKernelNode < nKernelElements;
       iKernelNode++) {
    int v0 = (0 + iKernelNode) % nKernelElements;
    int v1 = (1 + iKernelNode) % nKernelElements;
    int v2 = (2 + iKernelNode) % nKernelElements;
    int v3 = (3 + iKernelNode) % nKernelElements;
    int v4 = (4 + iKernelNode) % nKernelElements;
    int v5 = (5 + iKernelNode) % nKernelElements;

    int nodeId0 = Kernel_sortedNodeIds[kernelId][v0];
    int nodeId1 = Kernel_sortedNodeIds[kernelId][v1];
    int nodeId2 = Kernel_sortedNodeIds[kernelId][v2];
    int nodeId3 = Kernel_sortedNodeIds[kernelId][v3];
    int nodeId4 = Kernel_sortedNodeIds[kernelId][v4];
    int nodeId5 = Kernel_sortedNodeIds[kernelId][v5];

    double x0 = Node_coord[nodeId0][0];
    double y0 = Node_coord[nodeId0][1];
    double z0 = Node_coord[nodeId0][2];

    double x2 = Node_coord[nodeId2][0];
    double y2 = Node_coord[nodeId2][1];
    double z2 = Node_coord[nodeId2][2];

    double x3 = Node_coord[nodeId3][0];
    double y3 = Node_coord[nodeId3][1];
    double z3 = Node_coord[nodeId3][2];

    double x5 = Node_coord[nodeId5][0];
    double y5 = Node_coord[nodeId5][1];
    double z5 = Node_coord[nodeId5][2];

    double s02x0 = x0 + (x2 - x0) * CHECK_RATIO;
    double s02y0 = y0 + (y2 - y0) * CHECK_RATIO;
    double s02z0 = z0 + (z2 - z0) * CHECK_RATIO;

    double s02x1 = x2 + (x0 - x2) * CHECK_RATIO;
    double s02y1 = y2 + (y0 - y2) * CHECK_RATIO;
    double s02z1 = z2 + (z0 - z2) * CHECK_RATIO;

    double s03x0 = x0 + (x3 - x0) * CHECK_RATIO;
    double s03y0 = y0 + (y3 - y0) * CHECK_RATIO;
    double s03z0 = z0 + (z3 - z0) * CHECK_RATIO;

    double s03x1 = x3 + (x0 - x3) * CHECK_RATIO;
    double s03y1 = y3 + (y0 - y3) * CHECK_RATIO;
    double s03z1 = z3 + (z0 - z3) * CHECK_RATIO;

    double s35x0 = x3 + (x5 - x3) * CHECK_RATIO;
    double s35y0 = y3 + (y5 - y3) * CHECK_RATIO;
    double s35z0 = z3 + (z5 - z3) * CHECK_RATIO;

    double s35x1 = x5 + (x3 - x5) * CHECK_RATIO;
    double s35y1 = y5 + (y3 - y5) * CHECK_RATIO;
    double s35z1 = z5 + (z3 - z5) * CHECK_RATIO;

    if (KernelContainsPoint (kernelId, 
			     s02x0, s02y0, s02z0)
	&& KernelContainsPoint (kernelId, 
				s02x1, s02y1, s02z1)
	&& KernelContainsPoint (kernelId, 
				s03x0, s03y0, s03z0)
	&& KernelContainsPoint (kernelId, 
				s03x1, s03y1, s03z1)
	&& KernelContainsPoint (kernelId, 
				s35x0, s35y0, s35z0)
	&& KernelContainsPoint (kernelId, 
				s35x1, s35y1, s35z1)
	&& VolumeIsNotZero (axisNodeId0, nodeId0, nodeId1, nodeId2)
	&& VolumeIsNotZero (axisNodeId0, nodeId2, nodeId3, nodeId0)
	&& VolumeIsNotZero (axisNodeId0, nodeId3, nodeId5, nodeId0)
	&& VolumeIsNotZero (axisNodeId0, nodeId3, nodeId4, nodeId5)
	&& VolumeIsNotZero (axisNodeId1, nodeId0, nodeId1, nodeId2)
	&& VolumeIsNotZero (axisNodeId1, nodeId2, nodeId3, nodeId0)
	&& VolumeIsNotZero (axisNodeId1, nodeId3, nodeId5, nodeId0)
	&& VolumeIsNotZero (axisNodeId1, nodeId3, nodeId4, nodeId5)) {
      
      DeleteKernelElement (kernelId);
      
      Mesh_addNewElement (axisNodeId0, nodeId0, nodeId1, nodeId2);
      Mesh_addNewElement (axisNodeId0, nodeId2, nodeId3, nodeId0);
      Mesh_addNewElement (axisNodeId0, nodeId3, nodeId5, nodeId0);
      Mesh_addNewElement (axisNodeId0, nodeId3, nodeId4, nodeId5);
      
      Mesh_addNewElement (axisNodeId1, nodeId0, nodeId1, nodeId2);
      Mesh_addNewElement (axisNodeId1, nodeId2, nodeId3, nodeId0);
      Mesh_addNewElement (axisNodeId1, nodeId3, nodeId5, nodeId0);
      Mesh_addNewElement (axisNodeId1, nodeId3, nodeId4, nodeId5);

      success = 1;
      break;
    }
  }

  Kernel_isConverted[kernelId] = 1;
}

void Kernel_convert 
(int kernelId)
{
  int nKernelElements;

  assert(0 <= kernelId);
  assert(kernelId < NKernels);
  
  assert(!Kernel_isConverted[kernelId]);
  assert(Kernel_isValid (kernelId));

  nKernelElements = Kernel_nElements[kernelId];
  assert(2 < nKernelElements);

  switch (nKernelElements) {
  case 3:
    Convert3NodeKernel (kernelId);
    break;
  case 4:
    Convert4NodeKernel (kernelId);
    break;
  case 5:
    Convert5NodeKernel (kernelId);
    break;
  case 6:
    Convert6NodeKernel (kernelId);
    break;
  case 7:
  case 8:
  case 9:
  case 10:
    fprintf (stderr, " 7+ node kernel is not supported \n");
    break;
  default:
    assert(0);
    break;
  }
}

void Kernel_Collect (void)
{
  int iKernel;

  for (iKernel = 0; iKernel < NKernels; iKernel++) {
    Kernel_collectElement (iKernel);
  }
}

void Kernel_Convert (void)
{
  int iKernel;

  for (iKernel = 0; iKernel < NKernels; iKernel++) {
    if (Kernel_isValid (iKernel)) {
      Kernel_convert (iKernel);
    }
  }
}



/************************
	 Command
*************************/

void CheckTri 
(int triId)
{
  double tx0 = Tri_coord0[triId][0];
  double ty0 = Tri_coord0[triId][1];
  double tz0 = Tri_coord0[triId][2];
  
  double tx1 = Tri_coord1[triId][0];
  double ty1 = Tri_coord1[triId][1];
  double tz1 = Tri_coord1[triId][2];
  
  double tx2 = Tri_coord2[triId][0];
  double ty2 = Tri_coord2[triId][1];
  double tz2 = Tri_coord2[triId][2];
    
  int iElement;
  int iElementEdge;

  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) continue;
    
    if (Element_intersectionIsChecked[iElement]) continue;
    
    for (iElementEdge = 0; 
	 iElementEdge < N_ELEMENT_EDGES; 
	 iElementEdge++) {
      int nodeId0 
	= Element_nodeIds[iElement][EdgeNodeMap[iElementEdge][0]];
      int nodeId1 
	= Element_nodeIds[iElement][EdgeNodeMap[iElementEdge][1]];

      double ex0 = Node_coord[nodeId0][0];
      double ey0 = Node_coord[nodeId0][1];
      double ez0 = Node_coord[nodeId0][2];

      double ex1 = Node_coord[nodeId1][0];
      double ey1 = Node_coord[nodeId1][1];
      double ez1 = Node_coord[nodeId1][2];

      double intX, intY, intZ;

      if (GetIntersectionBetweenTriAndSeg 
	  (tx0, ty0, tz0, 
	   tx1, ty1, tz1, 
	   tx2, ty2, tz2, 
	   ex0, ey0, ez0, 
	   ex1, ey1, ez1, 
	   &intX, &intY, &intZ)) {
	int flag;
	double L0, L1, L2, L3;
	
	flag = Element_containsPoint 
	  (iElement, intX, intY, intZ,
	   &L0, &L1, &L2, &L3);
	assert(flag);
	
	if (L0 < 1.0 - VOLUME_COORD_EPS
	    && L1 < 1.0 - VOLUME_COORD_EPS
	    && L2 < 1.0 - VOLUME_COORD_EPS
	    && L3 < 1.0 - VOLUME_COORD_EPS) {
	  Kernel_AddNewKernel (nodeId0, nodeId1);
	}
      }
    }
  }
}

void MakeKernelByTri (void)
{
  int iTri;

  for (iTri = 0; iTri < NTris; iTri++) {
    CheckTri (iTri);
  }
}

void CompleteIntersectionCheck (void)
{
  int iElement;

  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) continue;

    Element_intersectionIsChecked[iElement] = 1;
  }
}

void PerformCommand (void)
{
  int nPrevKernels;

  {
    double volume = Mesh_volume ();
    
    fprintf (stderr, " volume %f\n", volume);
  }
  
  nPrevKernels = -1;
  for (;;) {
    Kernel_Clear ();
    MakeKernelByTri ();
    CompleteIntersectionCheck ();

    fprintf (stderr, " %d kernels are found \n",
	     NKernels);

    Kernel_Collect ();
    Kernel_Convert ();

    {
      double volume = Mesh_volume ();
      
      fprintf (stderr, " volume %f\n", volume);
    }
    
    if (NKernels == 0) {
      fprintf (stderr, " complete \n");
      break;
    } else if (NKernels == nPrevKernels) {
      fprintf (stderr, " give up \n");
      break;
    }
    nPrevKernels = NKernels;
  }
}



int main (int argc, char *argv[])
{
  if (argc != 6) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tetra3d_swap_boundary  node_file  element_file \n"
       "        tri_file \n"
       "        node_file_OUT  element_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadTriFile (argv[3]);

  PerformCommand ();

  WriteNodeFile (argv[4]);
  WriteElementFile (argv[5]);

  return 0;
}
