/* advautomesh_h_tri2d_isoline_view.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autogl.h>



/************************
	 Basic
*************************/

double Distance (double x0, double y0,
		 double x1, double y1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0));
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][2];



/* Element */

#define N_ELEMENT_NODES 3
#define N_ELEMENT_EDGES 3
#define MAX_ELEMENTS 1000000

const int EdgeNodeMap[N_ELEMENT_EDGES][2] = {
  { 1, 2 },
  { 2, 0 },
  { 0, 1 }
};

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



void Element_getEdgePoint 
(int elementId, 
 int elementEdgeId, double ratio,
 double *x_OUT, double *y_OUT)
{
  int nodeId0, nodeId1;
  double x0, y0;
  double x1, y1;

  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(0 <= elementEdgeId);
  assert(elementEdgeId < N_ELEMENT_EDGES);

  nodeId0 = Element_nodeIds
    [elementId][EdgeNodeMap[elementEdgeId][0]];
  nodeId1 = Element_nodeIds
    [elementId][EdgeNodeMap[elementEdgeId][1]];

  x0 = Node_coord[nodeId0][0];
  y0 = Node_coord[nodeId0][1];

  x1 = Node_coord[nodeId1][0];
  y1 = Node_coord[nodeId1][1];
  
  *x_OUT = x0 + (x1 - x0) * ratio;
  *y_OUT = y0 + (y1 - y0) * ratio;
}

void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
  }
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y;

    fscanf (fp, "%d %lf %lf", 
	    &nodeId, 
	    &x, &y);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Isoline */

#define MAX_LEVELS 1000000
#define MAX_ISOLINES 1000000

int NLevels;
int Level_startIsolineId[MAX_LEVELS];
int Level_nIsolines[MAX_LEVELS];
int NIsolines;
int Isoline_elementId[MAX_ISOLINES];
int Isoline_elementEdgeIds[MAX_ISOLINES][2];
double Isoline_ratios[MAX_ISOLINES][2];

double MinLevel, MaxLevel;
int NDivisions;



/* Frame */

#define MAX_FRAMES 1000000

int NFrames;
double Frame_coord0[MAX_FRAMES][2];
double Frame_coord1[MAX_FRAMES][2];



void ReadIsolineFile (const char *fileName)
{
  FILE *fp;
  int iLevel;
  int iIsoline;
  int count;
  
  assert(0 < NElements);
  assert(NLevels == 0);
  assert(NIsolines == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NLevels);
  assert(0 < NLevels);
  assert(NLevels <= MAX_LEVELS);
  
  count = 0;
  for (iLevel = 0; iLevel < NLevels; iLevel++) {
    int levelId;
    int startId;
    int nIsolines;

    fscanf (fp, "%d %d", 
	    &levelId,
	    &nIsolines);
    assert(levelId == iLevel);
    assert(0 <= nIsolines);
    startId = count;
    count += nIsolines;
    assert(count <= MAX_ISOLINES);
    Level_startIsolineId[iLevel] = startId;
    Level_nIsolines[iLevel] = nIsolines;

    for (iIsoline = startId; 
	 iIsoline < startId + nIsolines; 
	 iIsoline++) {
      int isolineId;
      int elementId;
      int elementEdgeId0, elementEdgeId1;
      double ratio0, ratio1;
      
      fscanf (fp, "%d %d %d %lf %d %lf", 
	      &isolineId,
	      &elementId,
	      &elementEdgeId0, &ratio0,
	      &elementEdgeId1, &ratio1);
      assert(isolineId == iIsoline - startId);
      assert(0 <= elementId);
      assert(elementId < NElements);
      Isoline_elementId[iIsoline] = elementId;
      
      assert(0 <= elementEdgeId0);
      assert(elementEdgeId0 < N_ELEMENT_EDGES);
      Isoline_elementEdgeIds[iIsoline][0] = elementEdgeId0;
      Isoline_ratios[iIsoline][0] = ratio0;
      
      assert(0 <= elementEdgeId1);
      assert(elementEdgeId1 < N_ELEMENT_EDGES);
      Isoline_elementEdgeIds[iIsoline][1] = elementEdgeId1;
      Isoline_ratios[iIsoline][1] = ratio1;
    }
  }

  fclose (fp);
}

void ReadFrameFile (const char *fileName)
{
  FILE *fp;
  int iFrame;
  
  assert(NFrames == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NFrames);
  assert(0 <= NFrames);
  assert(NFrames <= MAX_FRAMES);
  
  for (iFrame = 0; iFrame < NFrames; iFrame++) {
    int frameId;
    double x0, y0;
    double x1, y1;

    fscanf (fp, "%d %lf %lf %lf %lf", 
	    &frameId,
	    &x0, &y0, 
	    &x1, &y1);
    assert(frameId == iFrame);
    Frame_coord0[iFrame][0] = x0;
    Frame_coord0[iFrame][1] = y0;
    Frame_coord1[iFrame][0] = x1;
    Frame_coord1[iFrame][1] = y1;
  }

  fclose (fp);
}



/********************************
	 Visualization
********************************/

void SetView (void)
{
  double minRangeX, minRangeY;
  double maxRangeX, maxRangeY;
  double centerX, centerY;
  double size;
  
  Mesh_getRange 
    (&minRangeX, &minRangeY,
     &maxRangeX, &maxRangeY);
  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;
  size = Distance (minRangeX, minRangeY,
		   maxRangeX, maxRangeY) / 2;

  AutoGL_SetViewSize (size);
  AutoGL_SetViewCenter (centerX, centerY, 0.0); 
}

void DrawFrame (void)
{
  int iFrame;

  AutoGL_SetColor (1, 1, 1);
  for (iFrame = 0; iFrame < NFrames; iFrame++) {
    AutoGL_DrawLine (Frame_coord0[iFrame][0],
		     Frame_coord0[iFrame][1],
		     0.0,
		     Frame_coord1[iFrame][0],
		     Frame_coord1[iFrame][1],
		     0.0);
  }
}

void RedrawView (void) 
{
  int iLevel;
  int iIsoline;

  AutoGL_DrawLineContourMap (80, MinLevel, MaxLevel, NDivisions);

  DrawFrame ();

  for (iLevel = 0; iLevel < NLevels; iLevel++) {
    int startId = Level_startIsolineId[iLevel];
    int nIsolines = Level_nIsolines[iLevel];
    double grade = 1.0 / NDivisions * iLevel;
    double red, green, blue;

    AutoGL_GetContourColor (&red, &green, &blue,
			    grade);
    AutoGL_SetColor (red, green, blue);

    for (iIsoline = startId; 
	 iIsoline < startId + nIsolines; 
	 iIsoline++) {
      double ex0, ey0;
      double ex1, ey1;
      
      Element_getEdgePoint 
	(Isoline_elementId[iIsoline], 
	 Isoline_elementEdgeIds[iIsoline][0],
	 Isoline_ratios[iIsoline][0],
	 &ex0, &ey0);
      Element_getEdgePoint 
	(Isoline_elementId[iIsoline], 
	 Isoline_elementEdgeIds[iIsoline][1],
	 Isoline_ratios[iIsoline][1],
	 &ex1, &ey1);
      
      AutoGL_DrawLine (ex0, ey0, 0.0,
		       ex1, ey1, 0.0);
    }
  }
}

void UpdateModel (void) 
{
  AutoGL_DrawView ();
}

void Quit (void) 
{
  AutoGL_Terminate ();
  exit (0);
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
  if (argc != 8) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tri2d_isoline_view  node_file  element_file \n"
       "     isoline_file  frame_file \n"
       "     minLevel  maxLevel  nDivisions \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadIsolineFile (argv[3]);
  ReadFrameFile (argv[4]);
  MinLevel = atof (argv[5]);
  MaxLevel = atof (argv[6]);
  NDivisions = atoi (argv[7]);

  SetView ();
  
  AutoGL_SetViewRedrawCallback (RedrawView);

  AutoGL_SetDefaultCallbackInMode2D (NULL);
  AutoGL_EnableDragInMode2D ();



  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode2D ();

  AutoGL_AddCallback (Quit, "Quit");       

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");

  AutoGL_AddCallback (UpdateModel, "UpdateModel");      
}
