/* advautomesh_h_tri2d_boundary_node.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

double Distance (double x0, double y0, 
		 double x1, double y1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) + (y1 - y0) * (y1 - y0));
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][2];
int Node_isOnVertex[MAX_NODES];



int Mesh_findNodeNear (double x, double y, double dist)
{
  int iNode;

  for (iNode = 0; iNode < NNodes; iNode++) {
    double nodeX = Node_coord[iNode][0];
    double nodeY = Node_coord[iNode][1];

    if (Distance (x, y, 
		  nodeX, nodeY) < dist) {
      return iNode;
    }
  }

  return -1;
}

int Mesh_addNewNode 
(double x, double y,
 int isOnVertex)
{
  int nodeId = NNodes;

  assert(NNodes < MAX_NODES);
  NNodes++;
  
  Node_coord[nodeId][0] = x;
  Node_coord[nodeId][1] = y;
  
  Node_isOnVertex[nodeId] = isOnVertex;
  
  return nodeId;
}

void WriteNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    fprintf (fp, "%d   %22.15e %22.15e\n", 
	     iNode,
	     Node_coord[iNode][0],
	     Node_coord[iNode][1]); 
  }

  fclose (fp);
}



/************************
      Geometry Data
*************************/

/* Point */

#define MAX_POINTS 100000

int NPoints;
double Point_coord[MAX_POINTS][2];



/* Segment */

#define MAX_SEGS 100000

int NSegs;
double Seg_coord0[MAX_SEGS][2];
double Seg_coord1[MAX_SEGS][2];



void ReadPointFile (const char *fileName)
{
  FILE *fp;
  int iPoint;

  assert(NPoints == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NPoints);
  assert(NPoints <= MAX_POINTS);
  
  for (iPoint = 0; iPoint < NPoints; iPoint++) {
    int pointId;
    double x, y;

    fscanf (fp, "%d %lf %lf", 
	    &pointId,
	    &x, &y);
    assert(pointId == iPoint);
    Point_coord[iPoint][0] = x;
    Point_coord[iPoint][1] = y; 
  }

  fclose (fp);
}

void ReadSegFile (const char *fileName)
{
  FILE *fp;
  int iSeg;

  assert(NSegs == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NSegs);
  assert(NSegs <= MAX_SEGS);
  
  for (iSeg = 0; iSeg < NSegs; iSeg++) {
    int segId;
    double x0, y0, x1, y1;

    fscanf (fp, "%d %lf %lf %lf %lf", 
	    &segId,
	    &x0, &y0, 
	    &x1, &y1);
    assert(segId == iSeg);

    Seg_coord0[iSeg][0] = x0;
    Seg_coord0[iSeg][1] = y0; 

    Seg_coord1[iSeg][0] = x1;
    Seg_coord1[iSeg][1] = y1; 
  }

  fclose (fp);
}

void Seg_getRange
(double *minRangeX_OUT, double *minRangeY_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT)
{
  int iSeg;

  assert(0 < NSegs);
  
  *minRangeX_OUT = *maxRangeX_OUT = Seg_coord0[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Seg_coord0[0][1];
  
  for (iSeg = 0; iSeg < NSegs; iSeg++) {
    double x0 = Seg_coord0[iSeg][0];
    double y0 = Seg_coord0[iSeg][1];
    
    double x1 = Seg_coord1[iSeg][0];
    double y1 = Seg_coord1[iSeg][1];
    
    if (x0 < *minRangeX_OUT) *minRangeX_OUT = x0;
    if (y0 < *minRangeY_OUT) *minRangeY_OUT = y0;

    if (*maxRangeX_OUT < x0) *maxRangeX_OUT = x0;
    if (*maxRangeY_OUT < y0) *maxRangeY_OUT = y0;

    if (x1 < *minRangeX_OUT) *minRangeX_OUT = x1;
    if (y1 < *minRangeY_OUT) *minRangeY_OUT = y1;

    if (*maxRangeX_OUT < x1) *maxRangeX_OUT = x1;
    if (*maxRangeY_OUT < y1) *maxRangeY_OUT = y1;
  }
}



/************************
	 Command
*************************/

double NodalDistance;
double SortOriginX, SortOriginY;

int SortedSeg_segId[MAX_SEGS];
double Seg_distance[MAX_SEGS];

static void SetSortOrigin (void)
{
  double minRangeX, minRangeY;
  double maxRangeX, maxRangeY;
  double centerX, centerY;

  Seg_getRange 
    (&minRangeX, &minRangeY,
     &maxRangeX, &maxRangeY);
  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;

  SortOriginX = centerX - (maxRangeX - minRangeX);
  SortOriginY = centerY - (maxRangeY - minRangeY);
}

static void PutNodeOnPoint (int pointId)
{
  const int isOnVertex = 1;
  double px = Point_coord[pointId][0];
  double py = Point_coord[pointId][1];

  Mesh_addNewNode (px, py, isOnVertex);
}

#define N_SEG_DIVISIONS 10

static void PutNodesOnSeg (int segId)
{
  const int isOnVertex = 0;

  double sx0 = Seg_coord0[segId][0];
  double sy0 = Seg_coord0[segId][1];
  
  double sx1 = Seg_coord1[segId][0];
  double sy1 = Seg_coord1[segId][1];

  double length = Distance (sx0, sy0,
			    sx1, sy1);
  double interval = NodalDistance / N_SEG_DIVISIONS;
  int nDivisions = (int)(length / interval);
  double dist0 = Distance (sx0, sy0, 
			   SortOriginX, SortOriginY);
  double dist1 = Distance (sx1, sy1, 
			   SortOriginX, SortOriginY);

  double startX, startY;
  double endX, endY;
  int iDivision;
  
  if (dist0 < dist1) {
    startX = sx0;
    startY = sy0;
    endX = sx1;
    endY = sy1;
  } else {
    startX = sx1;
    startY = sy1;
    endX = sx0;
    endY = sy0;
  }

  for (iDivision = 0; iDivision <= nDivisions; iDivision++) {
    double x = startX + (endX - startX) / nDivisions * iDivision;
    double y = startY + (endY - startY) / nDivisions * iDivision;

    int nodeId = Mesh_findNodeNear (x, y, NodalDistance);
    
    if (nodeId == -1) {
      Mesh_addNewNode (x, y, isOnVertex);
    } else if (Node_isOnVertex[nodeId]) {
      nodeId = Mesh_findNodeNear (x, y, NodalDistance / 2);
      
      if (nodeId == -1) {
	Mesh_addNewNode (x, y, isOnVertex);
      }
    }
  }
}

static int CompareSegs (const void *seg0, const void *seg1)
{
  int segId0 = *(int *)seg0;
  int segId1 = *(int *)seg1;
  double dist0 = Seg_distance[segId0];
  double dist1 = Seg_distance[segId1];

  if (dist0 < dist1) {
    return -1;
  } else {
    return 1;
  }
}

static void PutNodesOnSegs (void)
{
  int iSeg;

  for (iSeg = 0; iSeg < NSegs; iSeg++) {
    double sx0 = Seg_coord0[iSeg][0];
    double sy0 = Seg_coord0[iSeg][1];
    
    double sx1 = Seg_coord1[iSeg][0];
    double sy1 = Seg_coord1[iSeg][1];

    double dist0 = Distance (sx0, sy0,
			     SortOriginX, SortOriginY);
    double dist1 = Distance (sx1, sy1,
			     SortOriginX, SortOriginY);

    SortedSeg_segId[iSeg] = iSeg;

    if (dist0 < dist1) {
      Seg_distance[iSeg] = dist0;
    } else {
      Seg_distance[iSeg] = dist1;
    }
  }

  qsort (SortedSeg_segId, NSegs, sizeof (int), CompareSegs);

  for (iSeg = 0; iSeg < NSegs; iSeg++) {
    int segId = SortedSeg_segId[iSeg];

    PutNodesOnSeg (segId);
  }
}

void PerformCommand (void)
{
  int iPoint;

  SetSortOrigin ();
  for (iPoint = 0; iPoint < NPoints; iPoint++) {
    PutNodeOnPoint (iPoint);
  }
  PutNodesOnSegs ();
}



int main (int argc, char *argv[])
{
  if (argc != 5) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tri2d_boundary_node \n"
       "        point_file  seg_file \n"
       "        nodal_distance \n"
       "        node_file_OUT \n");
    exit (1);
  }

  ReadPointFile (argv[1]);
  ReadSegFile (argv[2]);
  
  NodalDistance = atof (argv[3]);
  PerformCommand ();

  WriteNodeFile (argv[4]);

  return 0;
}
