/* advautomesh_h_tri2d_delaunay.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



#if 0
#define PRINT
#endif



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001

double Eps = DEFAULT_EPS;



int LessOrEqual (double value0, double value1) 
{
  return value0 < value1 + Eps;
}

double Distance (double x0, double y0,
		 double x1, double y1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0));
}

void GetTriCircumcenter 
(double x0, double y0, 
 double x1, double y1, 
 double x2, double y2,
 double *centerX_OUT, double *centerY_OUT)
{
  double a, b;

  a = ((x0 * x0 - x1 * x1 + y0 * y0 - y1 * y1) * (y0 - y2)
       - (x0 * x0 - x2 * x2 + y0 * y0 - y2 * y2) * (y0 - y1))
    / ((x0 - x1) * (y0 - y2) - (x0 - x2) * (y0 - y1));
  a /= 2;
  b = ((x0 * x0 - x1 * x1 + y0 * y0 - y1 * y1) * (x0 - x2)
       - (x0 * x0 - x2 * x2 + y0 * y0 - y2 * y2) * (x0 - x1))
    / ((y0 - y1) * (x0 - x2) - (y0 - y2) * (x0 - x1));
  b /= 2;

  *centerX_OUT = a;
  *centerY_OUT = b;
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
int NAllNodes;
double Node_coord[MAX_NODES][2];



/* Element */

#define N_ELEMENT_NODES 3
#define MAX_ELEMENTS 1000000

int NElements;
int Element_isDeleted[MAX_ELEMENTS];
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];
double Element_circumcenter[MAX_ELEMENTS][2];
double Element_circumradius[MAX_ELEMENTS];



void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
  }
}

int Mesh_addDummyNode (double x, double y)
{
  int nodeId = NAllNodes;
  
  assert(NAllNodes < MAX_NODES);
  NAllNodes++;
  
  Node_coord[nodeId][0] = x;
  Node_coord[nodeId][1] = y;

  return nodeId;
}

int Mesh_nodeIsDummy (int nodeId)
{
  assert(0 <= nodeId);
  assert(nodeId < NAllNodes);
  
  return NNodes <= nodeId;
}

static void CalcElementCircumcircle (int elementId)
{
  int n0 = Element_nodeIds[elementId][0];
  int n1 = Element_nodeIds[elementId][1];
  int n2 = Element_nodeIds[elementId][2];
  
  double x0 = Node_coord[n0][0];
  double y0 = Node_coord[n0][1];

  double x1 = Node_coord[n1][0];
  double y1 = Node_coord[n1][1];

  double x2 = Node_coord[n2][0];
  double y2 = Node_coord[n2][1];

  double centerX, centerY;
  double radius;

  GetTriCircumcenter 
    (x0, y0, 
     x1, y1, 
     x2, y2,
     &centerX, &centerY);
  radius = Distance (centerX, centerY,
		     x0, y0);

  Element_circumcenter[elementId][0] = centerX;
  Element_circumcenter[elementId][1] = centerY;
  Element_circumradius[elementId] = radius;
}

int Mesh_addNewElement (int nodeId0, int nodeId1, int nodeId2)
{
  int elementId = NElements;

  assert(0 <= nodeId0);
  assert(nodeId0 < NAllNodes);
  assert(0 <= nodeId1);
  assert(nodeId1 < NAllNodes);
  assert(0 <= nodeId2);
  assert(nodeId2 < NAllNodes);

  assert(NElements < MAX_ELEMENTS);
  NElements++;

  Element_isDeleted[elementId] = 0;

  Element_nodeIds[elementId][0] = nodeId0;
  Element_nodeIds[elementId][1] = nodeId1;
  Element_nodeIds[elementId][2] = nodeId2;

  CalcElementCircumcircle (elementId);

  return elementId;
}

void Mesh_markElementAsDeleted (int elementId)
{
  assert(0 <= elementId);
  assert(elementId < NElements);
  
  Element_isDeleted[elementId] = 1;
}

static void MoveElement (int srcId, int destId)
{
  int iElementNode;

  for (iElementNode = 0; 
       iElementNode < N_ELEMENT_NODES; 
       iElementNode++) {
    Element_nodeIds[destId][iElementNode] 
      = Element_nodeIds[srcId][iElementNode];
  }

  Element_circumcenter[destId][0]
    = Element_circumcenter[srcId][0];
  Element_circumcenter[destId][1]
    = Element_circumcenter[srcId][1];

  Element_circumradius[destId]
    = Element_circumradius[srcId];
}

void Mesh_deleteAllMarkedElements (void)
{
  int destId;
  int iElement;
  
  destId = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (!Element_isDeleted[iElement]) {
      MoveElement (iElement, destId);
      destId++;
    }
  }
  NElements = destId;

  for (iElement = 0; iElement < NElements; iElement++) {
    Element_isDeleted[iElement] = 0;
  }
}

int Mesh_elementIsDummy (int elementId)
{
  int n0, n1, n2;

  assert(0 <= elementId);
  assert(elementId < NElements);

  n0 = Element_nodeIds[elementId][0];
  n1 = Element_nodeIds[elementId][1];
  n2 = Element_nodeIds[elementId][2];

  return Mesh_nodeIsDummy (n0)
    || Mesh_nodeIsDummy (n1)
    || Mesh_nodeIsDummy (n2);
}

void Mesh_printNodes (void)
{
  int iNode;
  
  printf ("Nodes\n");
  printf ("%d\n", NNodes);
  printf (" NAllNodes %d\n", NAllNodes);
  
  for (iNode = 0; iNode < NAllNodes; iNode++) {
    printf ("%d x %f y %f ", 
	    iNode, 
	    Node_coord[iNode][0], 
	    Node_coord[iNode][1]);
    if (Mesh_nodeIsDummy (iNode)) {
      printf (" isDummy ");
    }
    printf ("\n");
  }
}

void Mesh_printElements (void)
{
  int iElement;

  printf ("Elements\n");
  printf ("%d\n", NElements);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int n0 = Element_nodeIds[iElement][0];
    int n1 = Element_nodeIds[iElement][1];
    int n2 = Element_nodeIds[iElement][2];

    printf ("%d nodeIds %d %d %d ", 
	    iElement, 
	    n0, n1, n2); 
    printf ("center %f %f radius %f ", 
	    Element_circumcenter[iElement][0], 
	    Element_circumcenter[iElement][1], 
	    Element_circumradius[iElement]); 
    if (Element_isDeleted[iElement]) {
      printf (" isDeleted ");
    }
    printf ("\n");
  }
}



void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y;

    fscanf (fp, "%d %lf %lf", 
	    &nodeId, 
	    &x, &y);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
  }

  fclose (fp);

  NAllNodes = NNodes;
}

void WriteElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NElements);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    fprintf (fp, "%d   ", 
	     iElement); 
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      fprintf (fp, "%d ", 
	       Element_nodeIds[iElement][iElementNode]); 
    }
    fprintf (fp, "\n");
  }
  
  fclose (fp);
}



/************************
	 Other Data
*************************/

#define MAX_COLLECTED_ELEMENTS 10000

int NCollectedElements;
int CollectedElement_elementId[MAX_COLLECTED_ELEMENTS];

#define MAX_EDGES 10000

int NEdges;
int Edge_nodeIds[MAX_EDGES][2];

void CollectElements (int nodeId)
{
  int iElement;
  
  NCollectedElements = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    double centerX = Element_circumcenter[iElement][0];
    double centerY = Element_circumcenter[iElement][1];
    double radius = Element_circumradius[iElement];
    
    double x = Node_coord[nodeId][0];
    double y = Node_coord[nodeId][1];
    
    double d = Distance (x, y, 
			 centerX, centerY);

    if (LessOrEqual (d, radius)) {
      assert(NCollectedElements < MAX_COLLECTED_ELEMENTS);
      
      CollectedElement_elementId[NCollectedElements] = iElement;
      NCollectedElements++;
    }
  }
}

static void AddEdge (int nodeId0, int nodeId1)
{
  int sameEdgeId;
  int iEdge;

  sameEdgeId = -1;
  for (iEdge = 0; iEdge < NEdges; iEdge++) {
    int n0 = Edge_nodeIds[iEdge][0];
    int n1 = Edge_nodeIds[iEdge][1];

    if ((n0 == nodeId0 && n1 == nodeId1)
	|| (n0 == nodeId1 && n1 == nodeId0)) {
      sameEdgeId = iEdge;
      break;
    }
  }

  if (0 <= sameEdgeId) {
    /* delete edge <sameEdgeId> */

    for (iEdge = sameEdgeId + 1; iEdge < NEdges; iEdge++) {
      Edge_nodeIds[iEdge - 1][0] = Edge_nodeIds[iEdge][0];
      Edge_nodeIds[iEdge - 1][1] = Edge_nodeIds[iEdge][1];
    }
    NEdges--;
  } else {
    /* add new edge <nodeId0, nodeId1> */

    assert(NEdges < MAX_EDGES);
    
    Edge_nodeIds[NEdges][0] = nodeId0;
    Edge_nodeIds[NEdges][1] = nodeId1;
    NEdges++;
  }
}

void CollectEdges (void)
{
  int iCe;

  NEdges = 0;
  for (iCe = 0; iCe < NCollectedElements; iCe++) {
    int elementId = CollectedElement_elementId[iCe];

    int n0 = Element_nodeIds[elementId][0];
    int n1 = Element_nodeIds[elementId][1];
    int n2 = Element_nodeIds[elementId][2];

    AddEdge (n0, n1);
    AddEdge (n1, n2);
    AddEdge (n2, n0);
  }
}

void DeleteOldElements (void)
{
  int iCe;

  for (iCe = 0; iCe < NCollectedElements; iCe++) {
    Mesh_markElementAsDeleted (CollectedElement_elementId[iCe]);
  }
  Mesh_deleteAllMarkedElements ();
}

void AddNewElements (int nodeId)
{
  int iEdge;

  for (iEdge = 0; iEdge < NEdges; iEdge++) {
    int n0 = Edge_nodeIds[iEdge][0];
    int n1 = Edge_nodeIds[iEdge][1];

    Mesh_addNewElement (n0, n1, nodeId);
  }
}

void PrintCollectedElements (void)
{
  int iCe;

  printf ("CollectedElements\n");
  printf ("%d\n", NCollectedElements);
  
  for (iCe = 0; iCe < NCollectedElements; iCe++) {
    int elementId = CollectedElement_elementId[iCe];

    int n0 = Element_nodeIds[elementId][0];
    int n1 = Element_nodeIds[elementId][1];
    int n2 = Element_nodeIds[elementId][2];

    printf ("%d elementId %d nodeIds %d %d %d\n", 
	    iCe,
	    elementId, n0, n1, n2); 
  }
}

void PrintEdges (void)
{
  int iEdge;

  printf ("Edges\n");
  printf ("%d\n", NEdges);
  
  for (iEdge = 0; iEdge < NEdges; iEdge++) {
    int n0 = Edge_nodeIds[iEdge][0];
    int n1 = Edge_nodeIds[iEdge][1];

    printf ("%d nodeIds %d %d\n", 
	    iEdge,
	    n0, n1); 
  }
}



/************************
	 Command
*************************/

static void MakeInitialMesh (void)
{
  double minRangeX, minRangeY;
  double maxRangeX, maxRangeY;
  double centerX, centerY;
  double x0, y0;
  double x1, y1;
  int n0, n1, n2, n3;

  Mesh_getRange 
    (&minRangeX, &minRangeY,
     &maxRangeX, &maxRangeY);
  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;

  /* corners of extended region */
  x0 = centerX - (maxRangeX - minRangeX);
  y0 = centerY - (maxRangeY - minRangeY);

  x1 = centerX + (maxRangeX - minRangeX);
  y1 = centerY + (maxRangeY - minRangeY);

  /* 4 dummy nodes on corners of extended region */
  n0 = Mesh_addDummyNode (x0, y0);
  n1 = Mesh_addDummyNode (x1, y0);
  n2 = Mesh_addDummyNode (x1, y1);
  n3 = Mesh_addDummyNode (x0, y1);

  /* 2 initial triangles */
  Mesh_addNewElement (n0, n1, n2);
  Mesh_addNewElement (n2, n3, n0);

#ifdef PRINT
  printf (" MakeInitialMesh \n");
  Mesh_printNodes ();
  Mesh_printElements ();
#endif
}

static void InsertNode (int nodeId)
{
#ifdef PRINT
  printf (" InsertNode %d %f %f\n",
	  nodeId,
	  Node_coord[nodeId][0],
	  Node_coord[nodeId][1]);
#endif

  CollectElements (nodeId);

#ifdef PRINT
  printf (" CollectElements \n");
  PrintCollectedElements ();
#endif

  CollectEdges ();

#ifdef PRINT
  printf (" CollectEdges \n");
  PrintEdges ();
#endif

  DeleteOldElements ();

#ifdef PRINT
  printf (" DeleteOldElements \n");
  Mesh_printElements ();
#endif

  AddNewElements (nodeId);

#ifdef PRINT
  printf (" AddNewElements \n");
  Mesh_printElements ();
#endif
}

static void DeleteDummyElements (void)
{
  int iElement;
  
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Mesh_elementIsDummy (iElement)) {
      Mesh_markElementAsDeleted (iElement);
    }
  }
  Mesh_deleteAllMarkedElements ();

#ifdef PRINT
  printf (" DeleteDummyElements \n");
  Mesh_printNodes ();
  Mesh_printElements ();
#endif
}

void PerformCommand (void)
{
  int iNode;

  MakeInitialMesh ();

  for (iNode = 0; iNode < NNodes; iNode++) {
    InsertNode (iNode);
  }

  DeleteDummyElements ();
}



int main (int argc, char *argv[])
{
  if (argc != 3) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tri2d_delaunay  node_file \n"
       "        element_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);

  PerformCommand ();

  WriteElementFile (argv[2]);

  return 0;
}
