/* advautomesh_h_tri2d_delaunay3.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



#if 0
#define PRINT
#endif
#if 0
#define PRINT_NODE_ID
#endif



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001

double Eps = DEFAULT_EPS;



void SortId2 (int id0, int id1,
	      int *minId_OUT, int *maxId_OUT)
{
  assert(id0 != id1);

  if (id0 < id1) {
    *minId_OUT = id0;
    *maxId_OUT = id1;
  } else {
    *minId_OUT = id1;
    *maxId_OUT = id0;
  }
}

int LessOrEqual (double value0, double value1) 
{
  return value0 < value1 + Eps;
}

double Distance (double x0, double y0,
		 double x1, double y1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0));
}

void GetTriCircumcenter 
(double x0, double y0, 
 double x1, double y1, 
 double x2, double y2,
 double *centerX_OUT, double *centerY_OUT)
{
  double a, b;

  a = ((x0 * x0 - x1 * x1 + y0 * y0 - y1 * y1) * (y0 - y2)
       - (x0 * x0 - x2 * x2 + y0 * y0 - y2 * y2) * (y0 - y1))
    / ((x0 - x1) * (y0 - y2) - (x0 - x2) * (y0 - y1));
  a /= 2;
  b = ((x0 * x0 - x1 * x1 + y0 * y0 - y1 * y1) * (x0 - x2)
       - (x0 * x0 - x2 * x2 + y0 * y0 - y2 * y2) * (x0 - x1))
    / ((y0 - y1) * (x0 - x2) - (y0 - y2) * (x0 - x1));
  b /= 2;

  *centerX_OUT = a;
  *centerY_OUT = b;
}

double SignedTriArea
(double x0, double y0,
 double x1, double y1,
 double x2, double y2)
{
  return ((x1 - x0) * (y2 - y0) - (y1 - y0) * (x2 - x0)) / 2;
}

void GetLineEquation 
(double x0, double y0, 
 double x1, double y1, 
 double *a_OUT, double *b_OUT, double *c_OUT) 
{
  double dx = x1 - x0;
  double dy = y1 - y0;

  double a0 = -dy;
  double b0 = dx;
  double c0 = dy * x0 - y0 * dx;

  double len = sqrt (a0 * a0 + b0 * b0);

  *a_OUT = a0 / len;  
  *b_OUT = b0 / len;  
  *c_OUT = c0 / len;
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
int NAllNodes;
double Node_coord[MAX_NODES][2];



/* Element */

#define N_ELEMENT_NODES 3
#define N_ELEMENT_EDGES 3
#define MAX_ELEMENTS 1000000

const int EdgeNodeMap[N_ELEMENT_EDGES][2] = {
  { 1, 2 },
  { 2, 0 },
  { 0, 1 }
};

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];
int Element_neighborIds[MAX_ELEMENTS][N_ELEMENT_EDGES];
double Element_circumcenter[MAX_ELEMENTS][2];
double Element_circumradius[MAX_ELEMENTS];

#define MAX_DELETED_ELEMENTS 10000

int NDeletedElements;
int DeletedElement_elementId[MAX_DELETED_ELEMENTS];



void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
  }
}

int Mesh_addDummyNode (double x, double y)
{
  int nodeId = NAllNodes;
  
  assert(NAllNodes < MAX_NODES);
  NAllNodes++;
  
  Node_coord[nodeId][0] = x;
  Node_coord[nodeId][1] = y;

  return nodeId;
}

int Mesh_nodeIsDummy (int nodeId)
{
  assert(0 <= nodeId);
  assert(nodeId < NAllNodes);
  
  return NNodes <= nodeId;
}

static void CalcElementCircumcircle (int elementId)
{
  int n0 = Element_nodeIds[elementId][0];
  int n1 = Element_nodeIds[elementId][1];
  int n2 = Element_nodeIds[elementId][2];
  
  double x0 = Node_coord[n0][0];
  double y0 = Node_coord[n0][1];

  double x1 = Node_coord[n1][0];
  double y1 = Node_coord[n1][1];

  double x2 = Node_coord[n2][0];
  double y2 = Node_coord[n2][1];

  double centerX, centerY;
  double radius;

  GetTriCircumcenter 
    (x0, y0, 
     x1, y1, 
     x2, y2,
     &centerX, &centerY);
  radius = Distance (centerX, centerY,
		     x0, y0);

  Element_circumcenter[elementId][0] = centerX;
  Element_circumcenter[elementId][1] = centerY;
  Element_circumradius[elementId] = radius;
}

void Mesh_deleteElement (int elementId)
{
  assert(0 <= elementId);
  assert(elementId < NElements);

  assert(NDeletedElements < MAX_DELETED_ELEMENTS);
  
  DeletedElement_elementId[NDeletedElements] = elementId;
  NDeletedElements++;
}

static int GetElementIdFromDeletedPool (void)
{
  int elementId = -1;

  if (0 < NDeletedElements) {
    elementId = DeletedElement_elementId[NDeletedElements - 1];
    NDeletedElements--;

    assert(0 <= elementId);
    assert(elementId < NElements);
  }
  return elementId;
}

int Mesh_elementIsDeleted (int elementId)
{
  int iDeletedElement;

  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(NDeletedElements < MAX_DELETED_ELEMENTS);
  
  for (iDeletedElement = 0; 
       iDeletedElement < NDeletedElements; 
       iDeletedElement++) {
    if (DeletedElement_elementId[iDeletedElement] == elementId) {
      return 1;
    }
  }
  return 0;
}

int Mesh_addNewElement 
(int nodeId0, int nodeId1, int nodeId2)
{
  int elementId;
  int iEdge;

  assert(0 <= nodeId0);
  assert(nodeId0 < NAllNodes);
  assert(0 <= nodeId1);
  assert(nodeId1 < NAllNodes);
  assert(0 <= nodeId2);
  assert(nodeId2 < NAllNodes);

  elementId = GetElementIdFromDeletedPool ();
  if (elementId == -1) {
    elementId = NElements;
    
    assert(NElements < MAX_ELEMENTS);
    NElements++;
  }

  Element_nodeIds[elementId][0] = nodeId0;
  Element_nodeIds[elementId][1] = nodeId1;
  Element_nodeIds[elementId][2] = nodeId2;

  for (iEdge = 0; iEdge < N_ELEMENT_EDGES; iEdge++) {
    Element_neighborIds[elementId][iEdge] = -1;
  }

  CalcElementCircumcircle (elementId);

  return elementId;
}

void Mesh_overwriteElement 
(int elementId,
 int nodeId0, int nodeId1, int nodeId2)
{
  int iEdge;

  assert(0 <= elementId);
  assert(elementId < NElements);

  assert(0 <= nodeId0);
  assert(nodeId0 < NAllNodes);
  assert(0 <= nodeId1);
  assert(nodeId1 < NAllNodes);
  assert(0 <= nodeId2);
  assert(nodeId2 < NAllNodes);

  Element_nodeIds[elementId][0] = nodeId0;
  Element_nodeIds[elementId][1] = nodeId1;
  Element_nodeIds[elementId][2] = nodeId2;

  for (iEdge = 0; iEdge < N_ELEMENT_EDGES; iEdge++) {
    Element_neighborIds[elementId][iEdge] = -1;
  }

  CalcElementCircumcircle (elementId);
}

int Mesh_elementIsDummy (int elementId)
{
  int n0, n1, n2;

  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(!Mesh_elementIsDeleted (elementId));

  n0 = Element_nodeIds[elementId][0];
  n1 = Element_nodeIds[elementId][1];
  n2 = Element_nodeIds[elementId][2];

  return Mesh_nodeIsDummy (n0)
    || Mesh_nodeIsDummy (n1)
    || Mesh_nodeIsDummy (n2);
}

int Mesh_elementIsWithinCircumcircle 
(int elementId, 
 double x, double y) 
{
  double centerX = Element_circumcenter[elementId][0];
  double centerY = Element_circumcenter[elementId][1];
  double radius = Element_circumradius[elementId];
  
  double dx = x - centerX;
  double dy = y - centerY;
  double d2 = dx * dx + dy * dy;
  
  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(!Mesh_elementIsDeleted (elementId));

  /* WARNING */
  /* meaning of tolerance changed to r^2 */
  return (d2 < radius * radius + Eps);
}

void Mesh_checkOrientation (int elementId)
{
  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(!Mesh_elementIsDeleted (elementId));

  {
    int n0 = Element_nodeIds[elementId][0];
    int n1 = Element_nodeIds[elementId][1];
    int n2 = Element_nodeIds[elementId][2];

    double x0 = Node_coord[n0][0];
    double y0 = Node_coord[n0][1];
    
    double x1 = Node_coord[n1][0];
    double y1 = Node_coord[n1][1];
    
    double x2 = Node_coord[n2][0];
    double y2 = Node_coord[n2][1];
    
    double area = SignedTriArea 
      (x0, y0,
       x1, y1,
       x2, y2);
    
    if (area < 0.0) {
      int n0 = Element_nodeIds[elementId][0];
      int n1 = Element_nodeIds[elementId][1];
      
      Element_nodeIds[elementId][0] = n1;
      Element_nodeIds[elementId][1] = n0;



      fprintf (stderr, " checkOrientation switched \n");
    }
  }
}

void Mesh_tryToSetNeighbor 
(int elementId, 
 int theOtherElementId)
{
  int iEdge;
  int jEdge;
  int found;

  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(!Mesh_elementIsDeleted (elementId));
  
  assert(0 <= theOtherElementId);
  assert(theOtherElementId < NElements);
  assert(!Mesh_elementIsDeleted (theOtherElementId));
  
  if (elementId == theOtherElementId) {
    return;
  }
  assert(elementId != theOtherElementId);

  found = 0;
  for (iEdge = 0; iEdge < N_ELEMENT_EDGES; iEdge++) {
    int ien0, ien1;
    
    {
      int nodeId0 = Element_nodeIds[elementId][EdgeNodeMap[iEdge][0]];
      int nodeId1 = Element_nodeIds[elementId][EdgeNodeMap[iEdge][1]];
      
      SortId2 (nodeId0, nodeId1,
	       &ien0, &ien1);
    }
    
    for (jEdge = 0; jEdge < N_ELEMENT_EDGES; jEdge++) {
      int jen0, jen1;
      
      {
	int nodeId0 = Element_nodeIds[theOtherElementId][EdgeNodeMap[jEdge][0]];
	int nodeId1 = Element_nodeIds[theOtherElementId][EdgeNodeMap[jEdge][1]];
	
	SortId2 (nodeId0, nodeId1,
		 &jen0, &jen1);
      }
      
      if (ien0 == jen0 && ien1 == jen1) {
	Element_neighborIds[elementId][iEdge] = theOtherElementId;
	Element_neighborIds[theOtherElementId][jEdge] = elementId;
	
	found = 1;
	break;
      }
    }

    if (found) {
      break;
    }
  }
}

void Mesh_setNeighborForAll (void)
{
  int iElement;
  int jElement;

  for (iElement = 0; iElement < NElements; iElement++) {
    for (jElement = iElement + 1; jElement < NElements; jElement++) {
      Mesh_tryToSetNeighbor (iElement, jElement);
    }
  }
}

void Mesh_printNodes (void)
{
  int iNode;
  
  printf ("Nodes\n");
  printf ("%d\n", NNodes);
  printf (" NAllNodes %d\n", NAllNodes);
  
  for (iNode = 0; iNode < NAllNodes; iNode++) {
    printf ("%d x %f y %f ", 
	    iNode, 
	    Node_coord[iNode][0], 
	    Node_coord[iNode][1]);
    if (Mesh_nodeIsDummy (iNode)) {
      printf (" isDummy ");
    }
    printf ("\n");
  }
}

void Mesh_printElements (void)
{
  int iElement;

  printf ("Elements\n");
  printf ("%d\n", 
	  NElements);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int n0 = Element_nodeIds[iElement][0];
    int n1 = Element_nodeIds[iElement][1];
    int n2 = Element_nodeIds[iElement][2];

    printf ("%d nodeIds %d %d %d ", 
	    iElement, 
	    n0, n1, n2); 
    printf ("center %f %f radius %f ", 
	    Element_circumcenter[iElement][0], 
	    Element_circumcenter[iElement][1], 
	    Element_circumradius[iElement]); 

    printf ("neighborIds %d %d %d ", 
	    Element_neighborIds[iElement][0],
	    Element_neighborIds[iElement][1],
	    Element_neighborIds[iElement][2]);

    if (Mesh_elementIsDeleted (iElement)) {
      printf (" isDeleted ");
    } else {
      printf (" isAlive ");
    }
    if (Mesh_elementIsDummy (iElement)) {
      printf (" isDummy ");
    } else {
      printf (" notDummy ");
    }

    printf ("\n");
  }
}



void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y;

    fscanf (fp, "%d %lf %lf", 
	    &nodeId, 
	    &x, &y);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
  }

  fclose (fp);

  NAllNodes = NNodes;
}

void WriteElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;
  int count;

  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  count = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (!Mesh_elementIsDeleted (iElement)
	&& !Mesh_elementIsDummy (iElement)) {
      count++;
    }
  }
  
  fprintf (fp, "%d\n", 
	   count);

  count = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (!Mesh_elementIsDeleted (iElement)
	&& !Mesh_elementIsDummy (iElement)) {
      fprintf (fp, "%d   ", 
	       count); 
      for (iElementNode = 0; 
	   iElementNode < N_ELEMENT_NODES; 
	   iElementNode++) {
	fprintf (fp, "%d ", 
		 Element_nodeIds[iElement][iElementNode]); 
      }
      fprintf (fp, "\n");
      
      count++;
    }
  }
  
  fclose (fp);
}



/************************
       Other Data
*************************/

#define MAX_COLLECTED_ELEMENTS 10000

int NCollectedElements;
int CollectedElement_elementId[MAX_COLLECTED_ELEMENTS];

#define MAX_COLLECTED_EDGES 10000

int NCollectedEdges;
int CollectedEdge_nodeIds[MAX_COLLECTED_EDGES][2];
int CollectedEdge_insideElementId[MAX_COLLECTED_EDGES];
int CollectedEdge_insideElementEdgeId[MAX_COLLECTED_EDGES];
int CollectedEdge_outsideElementId[MAX_COLLECTED_EDGES];

#define MAX_NEW_ELEMENTS 10000

int NNewElements;
int NewElement_elementId[MAX_NEW_ELEMENTS];

void ClearElementCollection (void)
{
  NCollectedElements = 0;
}

void AddElementIntoCollection (int elementId)
{
  int ceId;

  assert(NCollectedElements < MAX_COLLECTED_ELEMENTS);
  ceId = NCollectedElements;
  NCollectedElements++;
  
  CollectedElement_elementId[ceId] = elementId;
}

int ElementIsInCollection (int elementId)
{
  int iCe;

  for (iCe = 0; iCe < NCollectedElements; iCe++) {
    if (CollectedElement_elementId[iCe] == elementId) {
      return 1;
    }
  }
  return 0;
}

void ClearEdgeCollection (void)
{
  NCollectedEdges = 0;
}

void AddEdgeIntoCollection (int elementId, int elementEdgeId)
{
  int ceId;
  int nodeId0 = Element_nodeIds[elementId][EdgeNodeMap[elementEdgeId][0]];
  int nodeId1 = Element_nodeIds[elementId][EdgeNodeMap[elementEdgeId][1]];

  assert(NCollectedEdges < MAX_COLLECTED_EDGES);
  ceId = NCollectedEdges;
  NCollectedEdges++;

  CollectedEdge_insideElementId[ceId] = elementId;
  CollectedEdge_insideElementEdgeId[ceId] = elementEdgeId;
  CollectedEdge_nodeIds[ceId][0] = nodeId0;
  CollectedEdge_nodeIds[ceId][1] = nodeId1;
}

void FindNeighborElement 
(int elementId,
 double x, double y)
{
  int iEdge;

  assert(elementId != -1);

  for (iEdge = 0; iEdge < N_ELEMENT_EDGES; iEdge++) {
    int neighborId = Element_neighborIds[elementId][iEdge];

    if (neighborId == -1) {
      /* reach boundary of element collection */

      AddEdgeIntoCollection (elementId, iEdge);

      continue;
    }

    assert(!Mesh_elementIsDeleted (neighborId));

    if (ElementIsInCollection (neighborId)) {
      continue;
    }

    if (Mesh_elementIsWithinCircumcircle (neighborId, x, y)) {
      /* element <neighborId> is inside collection */
      
      AddElementIntoCollection (neighborId);
      FindNeighborElement (neighborId, x, y);
    } else {
      /* element <neighborId> is outside collection */
      /* reach boundary of element collection */
      
      AddEdgeIntoCollection (elementId, iEdge);
    }
  }
}

void FindFirstElement (double x, double y)
{
  int iElement;
  int elementIdFound;
  
  elementIdFound = -1;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (!Mesh_elementIsDeleted (iElement)) {
      if (Mesh_elementIsWithinCircumcircle (iElement, x, y)) {
	elementIdFound = iElement;
	break;
      }
    }
  }
  assert(elementIdFound != -1);
  
  AddElementIntoCollection (elementIdFound);
  FindNeighborElement (elementIdFound, x, y);
}

void CollectElements (int nodeId)
{
  double x = Node_coord[nodeId][0];
  double y = Node_coord[nodeId][1];

  ClearElementCollection ();
  ClearEdgeCollection ();

  FindFirstElement (x, y);
}

int NodeIsOnEdge (int nodeId)
{
  double x = Node_coord[nodeId][0];
  double y = Node_coord[nodeId][1];

  int iCd;

  for (iCd = 0; iCd < NCollectedEdges; iCd++) {
    int n0 = CollectedEdge_nodeIds[iCd][0];
    int n1 = CollectedEdge_nodeIds[iCd][1];

    double a, b, c;
    double equationValue;

    GetLineEquation
      (Node_coord[n0][0], Node_coord[n0][1],
       Node_coord[n1][0], Node_coord[n1][1],
       &a, &b, &c);
    equationValue = a * x + b * y + c;
    
    if (fabs (equationValue) < Eps) {
      fprintf (stderr, " node is on edge %d \n",
	       nodeId);

      return 1;
    }
  }
  return 0;
}

void SetOutsideElementOfCollectedEdges (void)
{
  int iCd;

  for (iCd = 0; iCd < NCollectedEdges; iCd++) {
    int insideElementId = CollectedEdge_insideElementId[iCd];
    int insideElementEdgeId = CollectedEdge_insideElementEdgeId[iCd];
    
    int outsideElementId 
      = Element_neighborIds[insideElementId][insideElementEdgeId];

    assert(outsideElementId != -1
	   || outsideElementId == -1);
    
    CollectedEdge_outsideElementId[iCd] = outsideElementId;
  }
}

void DeleteOldElementsAndAddNewElements (int nodeId)
{
  int iCd;
  int iCe;

  NNewElements = 0;
  for (iCd = 0; iCd < NCollectedEdges; iCd++) {
    int n0 = CollectedEdge_nodeIds[iCd][0];
    int n1 = CollectedEdge_nodeIds[iCd][1];
    
    iCe = iCd;

    if (iCe < NCollectedElements) {
      int elementId = CollectedElement_elementId[iCe];
      
      Mesh_overwriteElement (elementId, 
			     n0, n1, nodeId);

      assert(NNewElements < MAX_NEW_ELEMENTS);
      NewElement_elementId[NNewElements] = elementId;
      NNewElements++;
    } else {
      int elementId = Mesh_addNewElement (n0, n1, nodeId);

      assert(NNewElements < MAX_NEW_ELEMENTS);
      NewElement_elementId[NNewElements] = elementId;
      NNewElements++;
    }
  }

  for (iCe = 0; iCe < NCollectedElements; iCe++) {
    iCd = iCe;

    if (iCd < NCollectedEdges) {
      /* do nothing */
    } else {
      int elementId = CollectedElement_elementId[iCe];

      Mesh_deleteElement (elementId);
    }
  }
}

void ModifyNeighbors (void)
{
  int iNe;
  int jNe;
  int iCd;

  for (iNe = 0; iNe < NNewElements; iNe++) {
    for (jNe = iNe + 1; jNe < NNewElements; jNe++) {
      Mesh_tryToSetNeighbor (NewElement_elementId[iNe], 
			     NewElement_elementId[jNe]);
    }
  }
  
  for (iCd = 0; iCd < NCollectedEdges; iCd++) {
    int outsideElementId = CollectedEdge_outsideElementId[iCd];
    
    if (outsideElementId != -1) {
      for (iNe = 0; iNe < NNewElements; iNe++) {
	Mesh_tryToSetNeighbor (outsideElementId, 
			       NewElement_elementId[iNe]);
      }
    }
  }
}

void PrintCollectedElements (void)
{
  int iCe;

  printf ("CollectedElements\n");
  printf ("%d\n", NCollectedElements);
  
  for (iCe = 0; iCe < NCollectedElements; iCe++) {
    int elementId = CollectedElement_elementId[iCe];

    int n0 = Element_nodeIds[elementId][0];
    int n1 = Element_nodeIds[elementId][1];
    int n2 = Element_nodeIds[elementId][2];

    printf ("%d elementId %d nodeIds %d %d %d\n", 
	    iCe,
	    elementId, n0, n1, n2); 
  }
}

void PrintCollectedEdges (void)
{
  int iCd;

  printf ("CollectedEdges\n");
  printf ("%d\n", NCollectedEdges);
  
  for (iCd = 0; iCd < NCollectedEdges; iCd++) {
    int n0 = CollectedEdge_nodeIds[iCd][0];
    int n1 = CollectedEdge_nodeIds[iCd][1];

    printf ("%d nodeIds %d %d\n", 
	    iCd,
	    n0, n1); 
  }
}



/************************
	 Command
*************************/

static void MakeInitialMesh (void)
{
  double minRangeX, minRangeY;
  double maxRangeX, maxRangeY;
  double centerX, centerY;
  double x0, y0;
  double x1, y1;
  int n0, n1, n2, n3;

  Mesh_getRange 
    (&minRangeX, &minRangeY,
     &maxRangeX, &maxRangeY);
  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;

  /* corners of extended region */
  x0 = centerX - (maxRangeX - minRangeX);
  y0 = centerY - (maxRangeY - minRangeY);

  x1 = centerX + (maxRangeX - minRangeX);
  y1 = centerY + (maxRangeY - minRangeY);

  /* 4 dummy nodes on corners of extended region */
  n0 = Mesh_addDummyNode (x0, y0);
  n1 = Mesh_addDummyNode (x1, y0);
  n2 = Mesh_addDummyNode (x1, y1);
  n3 = Mesh_addDummyNode (x0, y1);

  /* 2 initial triangles */
  Mesh_addNewElement (n0, n1, n2);
  Mesh_addNewElement (n2, n3, n0);

  Mesh_setNeighborForAll ();

#ifdef PRINT
  printf (" MakeInitialMesh \n");
  Mesh_printNodes ();
  Mesh_printElements ();
#endif
}

static int InsertNode (int nodeId)
{
#ifdef PRINT
  printf (" InsertNode %d %f %f\n",
	  nodeId,
	  Node_coord[nodeId][0],
	  Node_coord[nodeId][1]);
#endif

  CollectElements (nodeId);
  if (NodeIsOnEdge (nodeId)) {
    return 0;
  }

  SetOutsideElementOfCollectedEdges ();

#ifdef PRINT
  printf (" CollectElements \n");
  PrintCollectedElements ();
  PrintCollectedEdges ();
#endif

  DeleteOldElementsAndAddNewElements (nodeId);
  ModifyNeighbors ();

#ifdef PRINT
  printf (" DeleteOldElementsAndAddNewElements \n");
#if 0
  Mesh_printElements ();
#endif
#endif

  return 1;
}

void CheckOrientation (void)
{
  int iElement;
  
  for (iElement = 0; iElement < NElements; iElement++) {
    if (!Mesh_elementIsDeleted (iElement)) {
      Mesh_checkOrientation (iElement);
    }
  }
  
#ifdef PRINT
  printf (" CheckOrientation \n");
#if 0
  Mesh_printElements ();
#endif
#endif
}

static int Node_isDone[MAX_NODES];

void PerformCommand (void)
{
  int iNode;
  int flag;

  MakeInitialMesh ();

  for (iNode = 0; iNode < NNodes; iNode++) {
    Node_isDone[iNode] = 0;
  }

#ifdef PRINT_NODE_ID
  printf (" the first 100, including vertexs \n");
#endif

  for (iNode = 0; iNode < 100; iNode++) {
    if (iNode < NNodes) {
      if (!Node_isDone[iNode]) {
	flag = InsertNode (iNode);
	
	if (flag) {
	  Node_isDone[iNode] = 1;
	}
      }
    }
  }

#ifdef PRINT_NODE_ID
  printf (" each 1000 \n");
#endif

  for (iNode = 0; iNode < NNodes; iNode++) {
    if (iNode % 1000 != 0) continue;
    
    if (!Node_isDone[iNode]) {
      flag = InsertNode (iNode);
      
      if (flag) {
	Node_isDone[iNode] = 1;
      }
    }
  }

#ifdef PRINT_NODE_ID
  printf (" each 100 \n");
#endif

  for (iNode = 0; iNode < NNodes; iNode++) {
    if (iNode % 100 != 0) continue;

    if (!Node_isDone[iNode]) {
      flag = InsertNode (iNode);
      
      if (flag) {
	Node_isDone[iNode] = 1;
      }
    }
  }

#ifdef PRINT_NODE_ID
  printf (" each 10 \n");
#endif

  for (iNode = 0; iNode < NNodes; iNode++) {
    if (iNode % 10 != 0) continue;

    if (!Node_isDone[iNode]) {
      flag = InsertNode (iNode);
      
      if (flag) {
	Node_isDone[iNode] = 1;
      }
    }
  }

#ifdef PRINT_NODE_ID
  printf (" the rest \n");
#endif

  for (iNode = 0; iNode < NNodes; iNode++) {
    if (!Node_isDone[iNode]) {
      flag = InsertNode (iNode);
      
      if (flag) {
	Node_isDone[iNode] = 1;
      }
    }
    
#ifdef PRINT_NODE_ID
    if (iNode % 1000 == 1) {
      printf (" iNode %d \n",
	      iNode);
    }
#endif    
  }

  for (iNode = 0; iNode < NNodes; iNode++) {
    if (!Node_isDone[iNode]) {
      fprintf (stderr, " fail to insert node %d \n",
	       iNode);
    }
  }

  CheckOrientation ();
}



int main (int argc, char *argv[])
{
  if (argc != 3) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tri2d_delaunay3  node_file \n"
       "        element_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);

  PerformCommand ();

  WriteElementFile (argv[2]);

  return 0;
}
