/* advautomesh_h_tri2d_domain_node.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

double Distance (double x0, double y0, 
		 double x1, double y1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) + (y1 - y0) * (y1 - y0));
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][2];
int NBoundaryNodes;



int Mesh_findNodeNear 
(double x, double y, 
 double dist)
{
  int iNode;

  for (iNode = 0; iNode < NNodes; iNode++) {
    double nodeX = Node_coord[iNode][0];
    double nodeY = Node_coord[iNode][1];

    if (Distance (x, y, 
		  nodeX, nodeY) < dist) {
      return iNode;
    }
  }

  return -1;
}

int Mesh_addNewNode 
(double x, double y)
{
  int nodeId = NNodes;

  assert(NNodes < MAX_NODES);
  NNodes++;
  
  Node_coord[nodeId][0] = x;
  Node_coord[nodeId][1] = y;
  
  return nodeId;
}

void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
  }
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y;

    fscanf (fp, "%d %lf %lf", 
	    &nodeId, 
	    &x, &y);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
  }

  fclose (fp);

  NBoundaryNodes = NNodes;
}

void WriteNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(0 < NNodes);
  assert(NBoundaryNodes <= NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    fprintf (fp, "%d   %22.15e %22.15e\n", 
	     iNode,
	     Node_coord[iNode][0],
	     Node_coord[iNode][1]); 
  }

  fclose (fp);
}



/************************
	 Command
*************************/

double NodalDistance;

#define EXTENSION_RATIO 0.1
#define N_DOMAIN_DIVISIONS 10

static void PutDomainNodes (void)
{
  double minRangeX, minRangeY;
  double maxRangeX, maxRangeY;
  double centerX, centerY;
  double originX, originY;
  double cornerX, cornerY;
  double interval;
  int xDivisions, yDivisions;
  int ix, iy;
  
  Mesh_getRange 
    (&minRangeX, &minRangeY,
     &maxRangeX, &maxRangeY);

  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;

  originX = centerX - (maxRangeX - minRangeX) * (0.5 + EXTENSION_RATIO);
  originY = centerY - (maxRangeY - minRangeY) * (0.5 + EXTENSION_RATIO);

  cornerX = centerX + (maxRangeX - minRangeX) * (0.5 + EXTENSION_RATIO);
  cornerY = centerY + (maxRangeY - minRangeY) * (0.5 + EXTENSION_RATIO);

  interval = NodalDistance / N_DOMAIN_DIVISIONS;

  xDivisions = (int)((cornerX - originX) / interval);
  yDivisions = (int)((cornerY - originY) / interval);
  
  for (ix = 0; ix <= xDivisions; ix++) {
    for (iy = 0; iy <= yDivisions; iy++) {
      double x = originX 
	+ (cornerX - originX) / xDivisions * ix;
      double y = originY 
	+ (cornerY - originY) / yDivisions * iy;

      int nodeId = Mesh_findNodeNear (x, y, 
				      NodalDistance);
      
      if (nodeId == -1) {
	Mesh_addNewNode (x, y);
      }
    }
  }
}

void PerformCommand (void)
{
  PutDomainNodes ();
}



int main (int argc, char *argv[])
{
  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tri2d_domain_node \n"
       "        node_file \n"
       "        nodal_distance \n"
       "        node_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  
  NodalDistance = atof (argv[2]);
  PerformCommand ();

  WriteNodeFile (argv[3]);

  return 0;
}
