/* advautomesh_h_tri2d_domain_node2.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

double Distance (double x0, double y0, 
		 double x1, double y1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) + (y1 - y0) * (y1 - y0));
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][2];
int NBoundaryNodes;



int Mesh_addNewNode 
(double x, double y)
{
  int nodeId = NNodes;

  assert(NNodes < MAX_NODES);
  NNodes++;
  
  Node_coord[nodeId][0] = x;
  Node_coord[nodeId][1] = y;
  
  return nodeId;
}

void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
  }
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y;

    fscanf (fp, "%d %lf %lf", 
	    &nodeId, 
	    &x, &y);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
  }

  fclose (fp);

  NBoundaryNodes = NNodes;
}

void WriteNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(0 < NNodes);
  assert(NBoundaryNodes <= NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    fprintf (fp, "%d   %22.15e %22.15e\n", 
	     iNode,
	     Node_coord[iNode][0],
	     Node_coord[iNode][1]); 
  }

  fclose (fp);
}



/************************
	 Map
*************************/

#define MAX_MAP_DIVISIONS 10000

static int Map_xDivisions, Map_yDivisions;
static double Map_originX, Map_originY;
static double Map_cornerX, Map_cornerY;
static double Map_interval;
static char Map_grid[MAX_MAP_DIVISIONS][MAX_MAP_DIVISIONS];

static void Map_initialize 
(double originX, double originY,
 double cornerX, double cornerY,
 double interval,
 int xDivisions, int yDivisions)
{
  int ix, iy;

  assert(xDivisions < MAX_MAP_DIVISIONS);
  assert(yDivisions < MAX_MAP_DIVISIONS);
  
  Map_originX = originX;
  Map_originY = originY;

  Map_cornerX = cornerX;
  Map_cornerY = cornerY;

  Map_interval = interval;

  Map_xDivisions = xDivisions;
  Map_yDivisions = yDivisions;

  for (ix = 0; ix <= Map_xDivisions; ix++) {
    for (iy = 0; iy <= Map_yDivisions; iy++) {
      Map_grid[ix][iy] = 0;
    }
  }
}

static int Map_ix (double x) 
{
  return (int)((x - Map_originX) / Map_interval);
}

static int Map_iy (double y) 
{
  return (int)((y - Map_originY) / Map_interval);
}

static double Map_x (int ix) 
{
  return Map_originX 
    + (Map_cornerX - Map_originX) / Map_xDivisions * ix;
}

static double Map_y (int iy) 
{
  return Map_originY 
    + (Map_cornerY - Map_originY) / Map_yDivisions * iy;
}

static void Map_fill 
(int cx, int cy, 
 double radius)
{
  int r = (int)(radius / Map_interval);
  int ix, iy;

  for (ix = cx - r; ix <= cx + r; ix++) {
    if (ix < 0 || Map_xDivisions < ix) continue;
    
    for (iy = cy - r; iy <= cy + r; iy++) {
      if (iy < 0 || Map_yDivisions < iy) continue;
    
      if ((ix - cx) * (ix - cx) + (iy - cy) * (iy - cy) <= r * r) {
	Map_grid[ix][iy] = 1;
      }
    }
  }
}



/************************
	 Command
*************************/

double NodalDistance;

#define EXTENSION_RATIO 0.1
#define N_DOMAIN_DIVISIONS 10

static void SetMap (void)
{
  double minRangeX, minRangeY;
  double maxRangeX, maxRangeY;
  double centerX, centerY;
  double originX, originY;
  double cornerX, cornerY;
  double interval;
  int xDivisions, yDivisions;

  Mesh_getRange 
    (&minRangeX, &minRangeY,
     &maxRangeX, &maxRangeY);

  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;

  originX = centerX - (maxRangeX - minRangeX) * (0.5 + EXTENSION_RATIO);
  originY = centerY - (maxRangeY - minRangeY) * (0.5 + EXTENSION_RATIO);

  cornerX = centerX + (maxRangeX - minRangeX) * (0.5 + EXTENSION_RATIO);
  cornerY = centerY + (maxRangeY - minRangeY) * (0.5 + EXTENSION_RATIO);

  interval = NodalDistance / N_DOMAIN_DIVISIONS;

  xDivisions = (int)((cornerX - originX) / interval);
  yDivisions = (int)((cornerY - originY) / interval);
  
  Map_initialize (originX, originY,
		  cornerX, cornerY,
		  interval,
		  xDivisions, yDivisions);
}

static void FillNodeOnBoundary (void)
{
  int iNode;

  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];

    int ix = Map_ix (x);
    int iy = Map_iy (y);
    
    Map_fill (ix, iy, 
	      NodalDistance);
  }
}

static void PutDomainNodes (void)
{
  int ix, iy;

  for (ix = 0; ix <= Map_xDivisions; ix++) {
    for (iy = 0; iy <= Map_yDivisions; iy++) {
      if (!Map_grid[ix][iy]) {
	double x = Map_x (ix);
	double y = Map_y (iy);
	
	Mesh_addNewNode (x, y);
	Map_fill (ix, iy, 
		  NodalDistance);
      }
    }
  }
}

void PerformCommand (void)
{
  SetMap ();
  FillNodeOnBoundary ();
  PutDomainNodes ();
}



int main (int argc, char *argv[])
{
  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tri2d_domain_node2 \n"
       "        node_file \n"
       "        nodal_distance \n"
       "        node_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  
  NodalDistance = atof (argv[2]);
  PerformCommand ();

  WriteNodeFile (argv[3]);

  return 0;
}
