/* advautomesh_h_tri2d_ec_refine.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001

double Eps = DEFAULT_EPS;



double Distance (double x0, double y0, 
		 double x1, double y1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) + (y1 - y0) * (y1 - y0));
}

int JustBetween (double value, double range0, double range1) 
{
  return (range0 + Eps < value && value < range1 - Eps)
    || (range1 + Eps < value && value < range0 - Eps);
}

void GetLineEquation 
(double x0, double y0, 
 double x1, double y1, 
 double *a_OUT, double *b_OUT, double *c_OUT) 
{
  double dx = x1 - x0;
  double dy = y1 - y0;

  double a0 = -dy;
  double b0 = dx;
  double c0 = dy * x0 - y0 * dx;

  double len = sqrt (a0 * a0 + b0 * b0);

  *a_OUT = a0 / len;  
  *b_OUT = b0 / len;  
  *c_OUT = c0 / len;
}

void GetProjectedPointOnLine 
(double a, double b, double c,
 double x, double y, 
 double *px_OUT, double *py_OUT)
{
  double value = a * x + b * y + c;

  *px_OUT = x - a * value;
  *py_OUT = y - b * value;
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][2];



/* Element */

#define N_ELEMENT_NODES 3
#define MAX_ELEMENTS 1000000

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];
int Element_isRefined[MAX_ELEMENTS];



void Element_getCenter 
(int elementId, 
 double *centerX_OUT, double *centerY_OUT)
{
  int iElementNode;

  assert(0 <= elementId);
  assert(elementId < NElements);

  *centerX_OUT = 0.0;
  *centerY_OUT = 0.0;
  for (iElementNode = 0; 
       iElementNode < N_ELEMENT_NODES; 
       iElementNode++) {
    int nodeId = Element_nodeIds[elementId][iElementNode];
    double x = Node_coord[nodeId][0];
    double y = Node_coord[nodeId][1];
    
    *centerX_OUT += x;
    *centerY_OUT += y;
  }
  *centerX_OUT /= N_ELEMENT_NODES;
  *centerY_OUT /= N_ELEMENT_NODES;
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y;

    fscanf (fp, "%d %lf %lf", 
	    &nodeId, 
	    &x, &y);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void WriteRefinedElementListFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int count;
  
  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  count = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isRefined[iElement]) {
      count++;
    }
  }

  fprintf (fp, "%d\n", 
	   count);

  count = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isRefined[iElement]) {
      fprintf (fp, "%d   %d\n", 
	       count, iElement);
      count++;
    }
  }

  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Crack */

double CrackTipX, CrackTipY;
double CrackEndX, CrackEndY;



void ReadEdgeCrackFile (const char *fileName)
{
  FILE *fp;
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%lf %lf", 
	  &CrackTipX, &CrackTipY);
  fscanf (fp, "%lf %lf", 
	  &CrackEndX, &CrackEndY);

  fclose (fp);
}



/************************
	 Command
*************************/

void PerformCommand (double radius, int crackTipOnlyFlag)
{
  int iElement;

  for (iElement = 0; iElement < NElements; iElement++) {
    Element_isRefined[iElement] = 0;
  }

  for (iElement = 0; iElement < NElements; iElement++) {
    double centerX, centerY;
    
    Element_getCenter (iElement, 
		       &centerX, &centerY);
    if (Distance (centerX, centerY, 
		  CrackTipX, CrackTipY) < radius - Eps) {
      Element_isRefined[iElement] = 1;
    }

    if (!crackTipOnlyFlag) {
      double a, b, c;
      
      GetLineEquation (CrackTipX, CrackTipY, CrackEndX, CrackEndY,
		       &a, &b, &c);

      if (fabs (a * centerX + b * centerY + c) < radius - Eps) {
	double px, py;
	
	GetProjectedPointOnLine 
	  (a, b, c,
	   centerX, centerY, 
	   &px, &py);
	if (JustBetween (centerX, CrackTipX, CrackEndX) 
	    || JustBetween (centerY, CrackTipY, CrackEndY)) {
	  Element_isRefined[iElement] = 1;
	}
      }
    }
  }
}



int main (int argc, char *argv[])
{
  double radius;
  int crackTipOnlyFlag;

  if (argc != 7) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tri2d_ec_refine  node_file  element_file \n"
       "        edge_crack_file  radius  crackTipOnlyFlag \n"
       "        refined_element_list_file_OUT \n");
    exit (1);
  }
  
  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadEdgeCrackFile (argv[3]);
  
  radius = atof (argv[4]);
  crackTipOnlyFlag = atoi (argv[5]);
  PerformCommand (radius, crackTipOnlyFlag);

  WriteRefinedElementListFile (argv[6]);

  return 0;
}
