/* advautomesh_h_tri2d_insert.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001
#define AREA_COORD_EPS 0.00001

double Eps = DEFAULT_EPS;



double Distance (double x0, double y0, 
		 double x1, double y1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) + (y1 - y0) * (y1 - y0));
}

double SignedTriArea
(double x0, double y0,
 double x1, double y1,
 double x2, double y2)
{
  return ((x1 - x0) * (y2 - y0) - (y1 - y0) * (x2 - x0)) / 2;
}

void GetLineEquation 
(double x0, double y0, 
 double x1, double y1, 
 double *a_OUT, double *b_OUT, double *c_OUT) 
{
  double dx = x1 - x0;
  double dy = y1 - y0;

  double a0 = -dy;
  double b0 = dx;
  double c0 = dy * x0 - y0 * dx;

  double len = sqrt (a0 * a0 + b0 * b0);

  *a_OUT = a0 / len;  
  *b_OUT = b0 / len;  
  *c_OUT = c0 / len;
}

int SolveLinearEquation
(double a00, double a01,
 double a10, double a11, 
 double b0, double b1,
 double *x_OUT, double *y_OUT)
{
  double det = a00 * a11 - a01 * a10;

  /* a00 x + a01 y + b0 = 0 */
  /* a10 x + a11 y + b1 = 0 */

  if (fabs (det) < Eps * Eps) return 0;

  *x_OUT = (a01 * b1 - a11 * b0) / det;
  *y_OUT = (a10 * b0 - a00 * b1) / det;
  return 1;
}

int Between (double value, double range0, double range1) 
{
  return (range0 - Eps < value && value < range1 + Eps)
    || (range1 - Eps < value && value < range0 + Eps);
}

int GetIntersectionBetweenSegs 
(double seg0_x0, double seg0_y0, 
 double seg0_x1, double seg0_y1, 
 double seg1_x0, double seg1_y0, 
 double seg1_x1, double seg1_y1,
 double *intX_OUT, double *intY_OUT) 
{
  double a0, b0, c0;
  double a1, b1, c1;
  int flag;
  
  GetLineEquation (seg0_x0, seg0_y0, 
		   seg0_x1, seg0_y1, 
		   &a0, &b0, &c0);
  GetLineEquation (seg1_x0, seg1_y0, 
		   seg1_x1, seg1_y1, 
		   &a1, &b1, &c1);
  flag = SolveLinearEquation 
    (a0, b0, 
     a1, b1, 
     c0, c1,
     intX_OUT, intY_OUT);
  if (flag) {
    if (Between (*intX_OUT, seg0_x0, seg0_x1)
	&& Between (*intY_OUT, seg0_y0, seg0_y1)) {
      if (Between (*intX_OUT, seg1_x0, seg1_x1)
	  && Between (*intY_OUT, seg1_y0, seg1_y1)) {
	return 1;
      }
    }
  }
  return 0;
}



/************************
	 Mesh
*************************/

/* Node */

typedef enum {
  NODE_TOPOLOGY_INTERNAL,
  NODE_TOPOLOGY_EDGE,
  NODE_TOPOLOGY_VERTEX
} NodeTopologyType;

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][2];
NodeTopologyType Node_topologyType[MAX_NODES];



/* Element */

#define N_ELEMENT_NODES 3
#define N_ELEMENT_EDGES 3
#define MAX_ELEMENTS 1000000

const int EdgeNodeMap[N_ELEMENT_EDGES][2] = {
  { 1, 2 },
  { 2, 0 },
  { 0, 1 }
};

int NElements;
int Element_isDeleted[MAX_ELEMENTS];
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



void Node_setTopology 
(int nodeId, 
 NodeTopologyType topologyType)
{
  assert(0 <= nodeId);
  assert(nodeId < NNodes);

  switch (topologyType) {
  case NODE_TOPOLOGY_INTERNAL:
    if (Node_topologyType[nodeId] != NODE_TOPOLOGY_VERTEX
	&& Node_topologyType[nodeId] != NODE_TOPOLOGY_EDGE) {
      Node_topologyType[nodeId] = NODE_TOPOLOGY_INTERNAL;
    }
    break;
  case NODE_TOPOLOGY_EDGE:
    if (Node_topologyType[nodeId] != NODE_TOPOLOGY_VERTEX) {
      Node_topologyType[nodeId] = NODE_TOPOLOGY_EDGE;
    }
    break;
  case NODE_TOPOLOGY_VERTEX:
    Node_topologyType[nodeId] = NODE_TOPOLOGY_VERTEX;
    break;
  default:
    assert(0);
    break;
  }
}

int Element_containsPoint 
(int elementId, 
 double x, double y,
 double *L0_OUT, double *L1_OUT, double *L2_OUT)
{
  int nodeId0, nodeId1, nodeId2;
  double x0, y0, x1, y1, x2, y2;
  double area;
  double area0, area1, area2;

  assert(0 <= elementId);
  assert(elementId < NElements);

  nodeId0 = Element_nodeIds[elementId][0];
  nodeId1 = Element_nodeIds[elementId][1];
  nodeId2 = Element_nodeIds[elementId][2];

  x0 = Node_coord[nodeId0][0];
  y0 = Node_coord[nodeId0][1];

  x1 = Node_coord[nodeId1][0];
  y1 = Node_coord[nodeId1][1];

  x2 = Node_coord[nodeId2][0];
  y2 = Node_coord[nodeId2][1];
  
  area = SignedTriArea (x0, y0, 
			x1, y1, 
			x2, y2);
  area0 = SignedTriArea (x1, y1, 
			 x2, y2, 
			 x, y);
  area1 = SignedTriArea (x2, y2, 
			 x0, y0, 
			 x, y);
  area2 = SignedTriArea (x0, y0, 
			 x1, y1, 
			 x, y);

  *L0_OUT = area0 / area;
  *L1_OUT = area1 / area;
  *L2_OUT = area2 / area;

  if (0.0 - AREA_COORD_EPS < *L0_OUT 
      && *L0_OUT < 1.0 + AREA_COORD_EPS
      && 0.0 - AREA_COORD_EPS < *L1_OUT 
      && *L1_OUT < 1.0 + AREA_COORD_EPS
      && 0.0 - AREA_COORD_EPS < *L2_OUT 
      && *L2_OUT < 1.0 + AREA_COORD_EPS) {
    return 1;
  }
  return 0;
}

int Mesh_findNodeAt (double x, double y)
{
  int iNode;

  for (iNode = 0; iNode < NNodes; iNode++) {
    double nodeX = Node_coord[iNode][0];
    double nodeY = Node_coord[iNode][1];

    if (Distance (x, y, 
		  nodeX, nodeY) < Eps) {
      return iNode;
    }
  }

  return -1;
}

int Mesh_addNewNode 
(double x, double y,
 NodeTopologyType topologyType)
{
  int nodeId = NNodes;

  assert(NNodes < MAX_NODES);
  NNodes++;

  Node_coord[nodeId][0] = x;
  Node_coord[nodeId][1] = y;

  Node_topologyType[nodeId] = topologyType;
  
  return nodeId;
}

int Mesh_addNewElement (int nodeId0, int nodeId1, int nodeId2)
{
  int elementId = NElements;

  assert(0 <= nodeId0);
  assert(nodeId0 < NNodes);
  assert(0 <= nodeId1);
  assert(nodeId1 < NNodes);
  assert(0 <= nodeId2);
  assert(nodeId2 < NNodes);

  assert(NElements < MAX_ELEMENTS);
  NElements++;

  Element_isDeleted[elementId] = 0;

  Element_nodeIds[elementId][0] = nodeId0;
  Element_nodeIds[elementId][1] = nodeId1;
  Element_nodeIds[elementId][2] = nodeId2;

  return elementId;
}

void Mesh_deleteElement (int elementId)
{
  assert(0 <= elementId);
  assert(elementId < NElements);

  Element_isDeleted[elementId] = 1;
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y;

    fscanf (fp, "%d %lf %lf", 
	    &nodeId, 
	    &x, &y);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    Element_isDeleted[iElement] = 0;
  }

  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void ReadNodeTopologyFile (const char *fileName)
{
  FILE *fp;
  int count;
  int iNode;

  assert(0 < NNodes);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &count);
  assert(count == NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    NodeTopologyType topologyType;
    
    fscanf (fp, "%d %d", 
	    &nodeId,
	    (int *)&topologyType); 
    assert(nodeId == iNode);
    switch (topologyType) {
    case NODE_TOPOLOGY_INTERNAL:
    case NODE_TOPOLOGY_EDGE:
    case NODE_TOPOLOGY_VERTEX:
      Node_topologyType[iNode] = topologyType;
      break;
    default:
      assert(0);
      break;
    }
  }

  fclose (fp);
}

void WriteNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    fprintf (fp, "%d   %22.15e %22.15e\n", 
	     iNode,
	     Node_coord[iNode][0],
	     Node_coord[iNode][1]); 
  }

  fclose (fp);
}

void WriteElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;
  int nActualElements;
  int count;

  assert(0 < NNodes);
  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  nActualElements = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) continue;
    
    nActualElements++;
  }

  fprintf (fp, "%d\n", 
	   nActualElements);
  
  count = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) continue;

    fprintf (fp, "%d   ", 
	     count); 
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      fprintf (fp, "%d ", 
	       Element_nodeIds[iElement][iElementNode]); 
    }
    fprintf (fp, "\n");
    count++;
  }
  
  fclose (fp);
}

void WriteNodeTopologyFile (const char *fileName)
{
  FILE *fp;
  int iNode;
  
  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    fprintf (fp, "%d   %d\n", 
	     iNode, Node_topologyType[iNode]);
  }

  fclose (fp);
}



/************************
     Geometry  Data
*************************/

/* Point */

#define MAX_POINTS 100000

int NPoints;
double Point_coord[MAX_POINTS][2];



/* Segment */

#define MAX_SEGS 100000

int NSegs;
double Seg_coord0[MAX_SEGS][2];
double Seg_coord1[MAX_SEGS][2];



void ReadPointFile (const char *fileName)
{
  FILE *fp;
  int iPoint;

  assert(NPoints == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NPoints);
  assert(NPoints <= MAX_POINTS);
  
  for (iPoint = 0; iPoint < NPoints; iPoint++) {
    int pointId;
    double x, y;

    fscanf (fp, "%d %lf %lf", 
	    &pointId,
	    &x, &y);
    assert(pointId == iPoint);
    Point_coord[iPoint][0] = x;
    Point_coord[iPoint][1] = y; 
  }

  fclose (fp);
}

void ReadSegFile (const char *fileName)
{
  FILE *fp;
  int iSeg;

  assert(NSegs == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NSegs);
  assert(NSegs <= MAX_SEGS);
  
  for (iSeg = 0; iSeg < NSegs; iSeg++) {
    int segId;
    double x0, y0, x1, y1;

    fscanf (fp, "%d %lf %lf %lf %lf", 
	    &segId,
	    &x0, &y0, 
	    &x1, &y1);
    assert(segId == iSeg);

    Seg_coord0[iSeg][0] = x0;
    Seg_coord0[iSeg][1] = y0; 

    Seg_coord1[iSeg][0] = x1;
    Seg_coord1[iSeg][1] = y1; 
  }

  fclose (fp);
}



/************************
	 Command
*************************/

int FindIntersectionWithSeg 
(double sx0, double sy0, 
 double sx1, double sy1,
 double *intX_OUT, double *intY_OUT)
{
  int iElement;
  int iElementEdge;
  
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) continue;
    
    for (iElementEdge = 0; 
	 iElementEdge < N_ELEMENT_EDGES; 
	 iElementEdge++) {
      int edgeNodeId0 
	= Element_nodeIds[iElement][EdgeNodeMap[iElementEdge][0]];
      int edgeNodeId1 
	= Element_nodeIds[iElement][EdgeNodeMap[iElementEdge][1]];

      double ex0 = Node_coord[edgeNodeId0][0];
      double ey0 = Node_coord[edgeNodeId0][1];

      double ex1 = Node_coord[edgeNodeId1][0];
      double ey1 = Node_coord[edgeNodeId1][1];

      if (GetIntersectionBetweenSegs 
	  (ex0, ey0, 
	   ex1, ey1, 
	   sx0, sy0, 
	   sx1, sy1, 
	   intX_OUT, intY_OUT)) {
	int flag;
	double L0, L1, L2;
	
	flag = Element_containsPoint 
	  (iElement, *intX_OUT, *intY_OUT,
	   &L0, &L1, &L2);
	assert(flag);
	
	if (L0 < 1.0 - AREA_COORD_EPS
	    && L1 < 1.0 - AREA_COORD_EPS
	    && L2 < 1.0 - AREA_COORD_EPS) {
	  return 1;
	}
      }
    }
  }
  return 0;
}

void InsertPointIntoElement 
(int elementId, int nodeId,
 double L0, double L1, double L2)
{
  int nodeId0 = Element_nodeIds[elementId][0];
  int nodeId1 = Element_nodeIds[elementId][1];
  int nodeId2 = Element_nodeIds[elementId][2];
    
  if (fabs (L0 - 1.0) < AREA_COORD_EPS
      || fabs (L1 - 1.0) < AREA_COORD_EPS
      || fabs (L2 - 1.0) < AREA_COORD_EPS) return;

  if (fabs (L0) < AREA_COORD_EPS) {
    /* edge 0 */
    Mesh_addNewElement (nodeId2, nodeId0, nodeId);
    Mesh_addNewElement (nodeId0, nodeId1, nodeId);
    
  } else if (fabs (L1) < AREA_COORD_EPS) {
    /* edge 1 */
    Mesh_addNewElement (nodeId0, nodeId1, nodeId);
    Mesh_addNewElement (nodeId1, nodeId2, nodeId);
    
  } else if (fabs (L2) < AREA_COORD_EPS) {
    /* edge 2 */
    Mesh_addNewElement (nodeId1, nodeId2, nodeId);
    Mesh_addNewElement (nodeId2, nodeId0, nodeId);
    
  } else {
    /* internal */
    Mesh_addNewElement (nodeId1, nodeId2, nodeId);
    Mesh_addNewElement (nodeId2, nodeId0, nodeId);
    Mesh_addNewElement (nodeId0, nodeId1, nodeId);
  }
  Mesh_deleteElement (elementId);
}

void Divide 
(double intX, double intY,
 NodeTopologyType topologyType)
{
  int iElement;
  
  for (iElement = 0; iElement < NElements; iElement++) {
    double L0, L1, L2;

    if (Element_isDeleted[iElement]) continue;
    
    if (Element_containsPoint (iElement, intX, intY,
			       &L0, &L1, &L2)) {
      if (L0 < 1.0 - AREA_COORD_EPS
	  && L1 < 1.0 - AREA_COORD_EPS
	  && L2 < 1.0 - AREA_COORD_EPS) {
	int intNodeId;
	
	intNodeId = Mesh_findNodeAt (intX, intY);
	if (intNodeId != -1) {
	  Node_setTopology (intNodeId, topologyType);
	} else {
	  intNodeId = Mesh_addNewNode (intX, intY, topologyType);
	}

	InsertPointIntoElement (iElement, intNodeId, 
				L0, L1, L2);
      } else {
	int intNodeId = Mesh_findNodeAt (intX, intY);

	assert(intNodeId != -1);
	Node_setTopology (intNodeId, topologyType);
      }
    }
  }
}

void InsertSeg (double sx0, double sy0, 
		double sx1, double sy1)
{
  for (;;) {
    double intX, intY;
    int found = FindIntersectionWithSeg 
      (sx0, sy0,
       sx1, sy1,
       &intX, &intY);
    
    if (found) {
      Divide (intX, intY, NODE_TOPOLOGY_EDGE);
    } else {
      break;
    }
  }
}

void PerformCommand (void)
{
  int iPoint;
  int iSeg;

  for (iPoint = 0; iPoint < NPoints; iPoint++) {
    double px = Point_coord[iPoint][0];
    double py = Point_coord[iPoint][1];
    
    Divide (px, py, NODE_TOPOLOGY_VERTEX);
  }

  for (iSeg = 0; iSeg < NSegs; iSeg++) {
    double sx0 = Seg_coord0[iSeg][0];
    double sy0 = Seg_coord0[iSeg][1];

    double sx1 = Seg_coord1[iSeg][0];
    double sy1 = Seg_coord1[iSeg][1];
    
    InsertSeg (sx0, sy0, 
	       sx1, sy1);
  }
}



int main (int argc, char *argv[])
{
  if (argc != 9) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tri2d_insert  node_file  element_file \n"
       "        node_topology_file \n"
       "        point_file  seg_file \n"
       "        node_file_OUT  element_file_OUT \n"
       "        node_topology_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadNodeTopologyFile (argv[3]);
  ReadPointFile (argv[4]);
  ReadSegFile (argv[5]);

  PerformCommand ();

  WriteNodeFile (argv[6]);
  WriteElementFile (argv[7]);
  WriteNodeTopologyFile (argv[8]);

  return 0;
}
