/* advautomesh_h_tri2d_laplacian.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Mesh
*************************/

/* Node */

typedef enum {
  NODE_TOPOLOGY_INTERNAL,
  NODE_TOPOLOGY_EDGE,
  NODE_TOPOLOGY_VERTEX
} NodeTopologyType;

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][2];
NodeTopologyType Node_topologyType[MAX_NODES];
int Node_count[MAX_NODES];
double Node_newCoord[MAX_NODES][2];



/* Element */

#define N_ELEMENT_NODES 3
#define MAX_ELEMENTS 1000000

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



void Element_getCenter 
(int elementId, 
 double *centerX_OUT, double *centerY_OUT)
{
  int iElementNode;

  assert(0 <= elementId);
  assert(elementId < NElements);

  *centerX_OUT = 0.0;
  *centerY_OUT = 0.0;
  for (iElementNode = 0; 
       iElementNode < N_ELEMENT_NODES; 
       iElementNode++) {
    int nodeId = Element_nodeIds[elementId][iElementNode];
    double x = Node_coord[nodeId][0];
    double y = Node_coord[nodeId][1];
    
    *centerX_OUT += x;
    *centerY_OUT += y;
  }
  *centerX_OUT /= N_ELEMENT_NODES;
  *centerY_OUT /= N_ELEMENT_NODES;
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y;

    fscanf (fp, "%d %lf %lf", 
	    &nodeId, 
	    &x, &y);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void ReadNodeTopologyFile (const char *fileName)
{
  FILE *fp;
  int count;
  int iNode;

  assert(0 < NNodes);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &count);
  assert(count == NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    NodeTopologyType topologyType;
    
    fscanf (fp, "%d %d", 
	    &nodeId,
	    (int *)&topologyType); 
    assert(nodeId == iNode);
    switch (topologyType) {
    case NODE_TOPOLOGY_INTERNAL:
    case NODE_TOPOLOGY_EDGE:
    case NODE_TOPOLOGY_VERTEX:
      Node_topologyType[iNode] = topologyType;
      break;
    default:
      assert(0);
      break;
    }
  }

  fclose (fp);
}

void WriteNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    fprintf (fp, "%d   %22.15e %22.15e\n", 
	     iNode,
	     Node_coord[iNode][0],
	     Node_coord[iNode][1]); 
  }

  fclose (fp);
}

void WriteElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NElements);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    fprintf (fp, "%d   ", 
	     iElement); 
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      fprintf (fp, "%d ", 
	       Element_nodeIds[iElement][iElementNode]); 
    }
    fprintf (fp, "\n");
  }
  
  fclose (fp);
}



/************************
	 Command
*************************/

void SetNodeCount (void)
{
  int iNode;
  int iElement;
  int iElementNode;

  for (iNode = 0; iNode < NNodes; iNode++) {
    Node_count[iNode] = 0;
  }

  for (iElement = 0; iElement < NElements; iElement++) {
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId = Element_nodeIds[iElement][iElementNode]; 
      
      Node_count[nodeId]++;
    }
  }

  for (iNode = 0; iNode < NNodes; iNode++) {
    assert(0 < Node_count[iNode]);
  }
}

void DoSmoothing (void)
{
  int iNode;
  int iElement;
  int iElementNode;

  for (iNode = 0; iNode < NNodes; iNode++) {
    Node_newCoord[iNode][0] = 0.0;
    Node_newCoord[iNode][1] = 0.0;
  }
  
  for (iElement = 0; iElement < NElements; iElement++) {
    double centerX, centerY;
    
    Element_getCenter (iElement, 
		       &centerX, &centerY);
    
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId = Element_nodeIds[iElement][iElementNode]; 
      
      Node_newCoord[nodeId][0] += centerX;
      Node_newCoord[nodeId][1] += centerY;
    }
  }
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int count = Node_count[iNode];

    Node_newCoord[iNode][0] /= count;
    Node_newCoord[iNode][1] /= count;
  }
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    if (Node_topologyType[iNode] == NODE_TOPOLOGY_INTERNAL) {
      Node_coord[iNode][0] = Node_newCoord[iNode][0];
      Node_coord[iNode][1] = Node_newCoord[iNode][1];
    }
  }
}

void PerformCommand (int nIterations)
{
  int iIteration;

  SetNodeCount ();
  for (iIteration = 0; iIteration < nIterations; iIteration++) {
    DoSmoothing ();
  }
}



int main (int argc, char *argv[])
{
  int nIterations;

  if (argc != 7) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tri2d_laplacian  node_file  element_file \n"
       "        node_topology_file  nIterations \n"
       "        node_file_OUT  element_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadNodeTopologyFile (argv[3]);

  nIterations = atoi (argv[4]);
  PerformCommand (nIterations);

  WriteNodeFile (argv[5]);
  WriteElementFile (argv[6]);

  return 0;
}
