/* advautomesh_h_tri2d_particle.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

#define AREA_COORD_EPS 0.00001



double SignedTriArea
(double x0, double y0,
 double x1, double y1,
 double x2, double y2)
{
  return ((x1 - x0) * (y2 - y0) - (y1 - y0) * (x2 - x0)) / 2;
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][2];
double Node_vector[MAX_NODES][2];



/* Element */

#define N_ELEMENT_NODES 3
#define N_ELEMENT_EDGES 3
#define MAX_ELEMENTS 1000000

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];
int Element_neighborIds[MAX_ELEMENTS][N_ELEMENT_EDGES];



void Element_getVector
(int elementId,
 double L0, double L1, double L2,
 double *vectorX_OUT, double *vectorY_OUT)
{
  int nodeId0, nodeId1, nodeId2;

  assert(0 <= elementId);
  assert(elementId < NElements);

  nodeId0 = Element_nodeIds[elementId][0];
  nodeId1 = Element_nodeIds[elementId][1];
  nodeId2 = Element_nodeIds[elementId][2];
  
  *vectorX_OUT 
    = Node_vector[nodeId0][0] * L0
    + Node_vector[nodeId1][0] * L1
    + Node_vector[nodeId2][0] * L2;
  *vectorY_OUT
    = Node_vector[nodeId0][1] * L0
    + Node_vector[nodeId1][1] * L1
    + Node_vector[nodeId2][1] * L2;
}

int Element_containsPoint 
(int elementId, 
 double x, double y,
 double *L0_OUT, double *L1_OUT, double *L2_OUT)
{
  int nodeId0, nodeId1, nodeId2;
  double x0, y0, x1, y1, x2, y2;
  double area;
  double area0, area1, area2;

  assert(0 <= elementId);
  assert(elementId < NElements);

  nodeId0 = Element_nodeIds[elementId][0];
  nodeId1 = Element_nodeIds[elementId][1];
  nodeId2 = Element_nodeIds[elementId][2];

  x0 = Node_coord[nodeId0][0];
  y0 = Node_coord[nodeId0][1];

  x1 = Node_coord[nodeId1][0];
  y1 = Node_coord[nodeId1][1];

  x2 = Node_coord[nodeId2][0];
  y2 = Node_coord[nodeId2][1];
  
  area = SignedTriArea (x0, y0, 
			x1, y1, 
			x2, y2);
  area0 = SignedTriArea (x1, y1, 
			 x2, y2, 
			 x, y);
  area1 = SignedTriArea (x2, y2, 
			 x0, y0, 
			 x, y);
  area2 = SignedTriArea (x0, y0, 
			 x1, y1, 
			 x, y);

  *L0_OUT = area0 / area;
  *L1_OUT = area1 / area;
  *L2_OUT = area2 / area;

  if (0.0 - AREA_COORD_EPS < *L0_OUT 
      && *L0_OUT < 1.0 + AREA_COORD_EPS
      && 0.0 - AREA_COORD_EPS < *L1_OUT 
      && *L1_OUT < 1.0 + AREA_COORD_EPS
      && 0.0 - AREA_COORD_EPS < *L2_OUT 
      && *L2_OUT < 1.0 + AREA_COORD_EPS) {
    return 1;
  }
  return 0;
}

int Mesh_findElementAt 
(double x, double y,
 double *L0_OUT, double *L1_OUT, double *L2_OUT)
{
  int iElement;

  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_containsPoint (iElement, x, y,
			       L0_OUT, L1_OUT, L2_OUT)) {
      return iElement;
    }
  }
  return -1;
}

int Mesh_findElementNear 
(int elementId,
 double x, double y, 
 double *L0_OUT, double *L1_OUT, double *L2_OUT)
{
  int iElementEdge;
  int jElementEdge;
  int kElementEdge;
  
  assert(elementId != -1);

  if (Element_containsPoint (elementId, x, y,
			     L0_OUT, L1_OUT, L2_OUT)) {
    return elementId;
  }

  for (iElementEdge = 0; 
       iElementEdge < N_ELEMENT_EDGES; 
       iElementEdge++) {
    int neighborId = Element_neighborIds[elementId][iElementEdge];
    
    if (neighborId == -1) continue;

    if (Element_containsPoint (neighborId, x, y,
			       L0_OUT, L1_OUT, L2_OUT)) {
      return neighborId;
    }
  }

  for (iElementEdge = 0; 
       iElementEdge < N_ELEMENT_EDGES; 
       iElementEdge++) {
    int neighborIdI = Element_neighborIds[elementId][iElementEdge];
    
    if (neighborIdI == -1) continue;

    for (jElementEdge = 0; 
	 jElementEdge < N_ELEMENT_EDGES; 
	 jElementEdge++) {
      int neighborIdJ = Element_neighborIds[neighborIdI][jElementEdge];
      
      if (neighborIdJ == -1) continue;
      if (neighborIdJ == elementId) continue;

      if (Element_containsPoint (neighborIdJ, x, y,
				 L0_OUT, L1_OUT, L2_OUT)) {
	return neighborIdJ;
      }
    }
  }

  for (iElementEdge = 0; 
       iElementEdge < N_ELEMENT_EDGES; 
       iElementEdge++) {
    int neighborIdI = Element_neighborIds[elementId][iElementEdge];
    
    if (neighborIdI == -1) continue;

    for (jElementEdge = 0; 
	 jElementEdge < N_ELEMENT_EDGES; 
	 jElementEdge++) {
      int neighborIdJ = Element_neighborIds[neighborIdI][jElementEdge];
      
      if (neighborIdJ == -1) continue;
      if (neighborIdJ == elementId) continue;

      for (kElementEdge = 0; 
	   kElementEdge < N_ELEMENT_EDGES; 
	   kElementEdge++) {
	int neighborIdK = Element_neighborIds[neighborIdJ][kElementEdge];
	
	if (neighborIdK == -1) continue;
	if (neighborIdK == elementId) continue;
	if (neighborIdK == neighborIdI) continue;
	
	if (Element_containsPoint (neighborIdK, x, y,
				   L0_OUT, L1_OUT, L2_OUT)) {
	  return neighborIdK;
	}
      }
    }
  }

  return -1;
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y;

    fscanf (fp, "%d %lf %lf", 
	    &nodeId, 
	    &x, &y);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void ReadElementNeighborFile (const char *fileName)
{
  FILE *fp;
  int count;
  int iElement;
  int iElementEdge;
  
  assert(0 < NElements);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &count);
  assert(count == NElements);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId);
    assert(elementId == iElement);

    for (iElementEdge = 0; 
	 iElementEdge < N_ELEMENT_EDGES; 
	 iElementEdge++) {
      int neighborId;

      fscanf (fp, "%d", 
	      &neighborId);
      assert((0 <= neighborId && neighborId < NElements) 
	     || neighborId == -1);
      Element_neighborIds[iElement][iElementEdge] = neighborId;
    }
  }

  fclose (fp);
}

void ReadNodeVectorFile (const char *fileName)
{
  FILE *fp;
  int count;
  int iNode;
  
  assert(0 < NNodes);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &count);
  assert(count == NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double vectorX, vectorY;

    fscanf (fp, "%d %lf %lf", 
	    &nodeId,
	    &vectorX, &vectorY);
    assert(nodeId == iNode);
    Node_vector[iNode][0] = vectorX;
    Node_vector[iNode][1] = vectorY;
  }

  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Particle */

#define MAX_PARTICLES 1000
#define MAX_STEPS 1000

int NParticles;
int NSteps;
double DeltaT;
double Particle_startCoord[MAX_PARTICLES][2]; 

double Point_coord[MAX_PARTICLES][MAX_STEPS][2];
int Point_elementId[MAX_PARTICLES][MAX_STEPS];
double Point_L0[MAX_PARTICLES][MAX_STEPS];
double Point_L1[MAX_PARTICLES][MAX_STEPS];
double Point_L2[MAX_PARTICLES][MAX_STEPS];



void ParticleGroup_setParticlePoint 
(int particleId, int stepId,
 double x, double y,
 int elementId, 
 double L0, double L1, double L2)
{
  assert(0 <= particleId);
  assert(particleId < NParticles);
  assert(0 <= stepId);
  assert(stepId < NSteps);
  assert(0 <= elementId || elementId == -1);
  assert(elementId < NElements);

  Point_coord[particleId][stepId][0] = x;
  Point_coord[particleId][stepId][1] = y;

  Point_elementId[particleId][stepId] = elementId;

  Point_L0[particleId][stepId] = L0;
  Point_L1[particleId][stepId] = L1;
  Point_L2[particleId][stepId] = L2;
}

void ReadParticleFile (const char *fileName)
{
  FILE *fp;
  int iParticle;
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d %d %lf", 
	  &NParticles,
	  &NSteps,
	  &DeltaT);
  assert(0 < NParticles);
  assert(NParticles <= MAX_PARTICLES);
  assert(1 < NSteps);
  assert(NSteps <= MAX_STEPS);

  for (iParticle = 0; iParticle < NParticles; iParticle++) {
    int particleId;
    double x, y;
    
    fscanf (fp, "%d %lf %lf", 
	    &particleId,
	    &x, &y);
    assert(particleId == iParticle);
    Particle_startCoord[iParticle][0] = x;
    Particle_startCoord[iParticle][1] = y;
  }

  fclose (fp);
}

void WriteParticleTraceFile (const char *fileName)
{
  FILE *fp;
  int iParticle;
  int iStep;
  
  assert(0 < NParticles);
  assert(1 < NSteps);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d %d\n", 
	   NParticles, NSteps);
  
  for (iParticle = 0; iParticle < NParticles; iParticle++) {
    fprintf (fp, "%d\n", 
	     iParticle);
    
    for (iStep = 0; iStep < NSteps; iStep++) {
      fprintf (fp, "%d   %22.15e %22.15e   %d   %22.15e %22.15e %22.15e\n", 
	       iStep,
	       Point_coord[iParticle][iStep][0],
	       Point_coord[iParticle][iStep][1],
	       Point_elementId[iParticle][iStep],
	       Point_L0[iParticle][iStep],
	       Point_L1[iParticle][iStep],
	       Point_L2[iParticle][iStep]);
    }
  }

  fclose (fp);
}



/************************
	 Command
*************************/

void PerformCommand (void)
{
  int iParticle;
  int iStep;

  assert(0 < NParticles); 
  assert(1 < NSteps); 

  iStep = 0;
  for (iParticle = 0; iParticle < NParticles; iParticle++) {
    double x = Particle_startCoord[iParticle][0];
    double y = Particle_startCoord[iParticle][1];
    int elementId;
    double L0, L1, L2;

    elementId = Mesh_findElementAt (x, y,
				    &L0, &L1, &L2);
    if (elementId != -1) {
      ParticleGroup_setParticlePoint 
	(iParticle, iStep,
	 x, y, 
	 elementId, L0, L1, L2);
    } else {
      ParticleGroup_setParticlePoint 
	(iParticle, iStep,
	 x, y, 
	 -1, 0.0, 0.0, 0.0);
    }
  }

  for (iStep = 1; iStep < NSteps; iStep++) {
    for (iParticle = 0; iParticle < NParticles; iParticle++) {
      double x = Point_coord[iParticle][iStep - 1][0];
      double y = Point_coord[iParticle][iStep - 1][1];
      int elementId = Point_elementId[iParticle][iStep - 1];
      double L0 = Point_L0[iParticle][iStep - 1];
      double L1 = Point_L1[iParticle][iStep - 1];
      double L2 = Point_L2[iParticle][iStep - 1];
      
      if (elementId != -1) {
	double vectorX, vectorY;

	Element_getVector (elementId, L0, L1, L2,
			   &vectorX, &vectorY);
	x += vectorX * DeltaT;
	y += vectorY * DeltaT;

	elementId = Mesh_findElementNear (elementId, x, y, 
					  &L0, &L1, &L2);
	if (elementId == -1) {
	  elementId = Mesh_findElementAt (x, y,
					  &L0, &L1, &L2);
	}
      }

      if (elementId != -1) {
	ParticleGroup_setParticlePoint 
	  (iParticle, iStep,
	   x, y, 
	   elementId, L0, L1, L2);
      } else {
	ParticleGroup_setParticlePoint 
	  (iParticle, iStep,
	   x, y, 
	   -1, 0.0, 0.0, 0.0);
      }
    }
  }
}



int main (int argc, char *argv[])
{
  if (argc != 7) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tri2d_particle  node_file  element_file \n"
       "        element_neighbor_file  node_vector_file  particle_file \n"
       "        particle_trace_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadElementNeighborFile (argv[3]);
  ReadNodeVectorFile (argv[4]);
  ReadParticleFile (argv[5]);

  PerformCommand ();

  WriteParticleTraceFile (argv[6]);

  return 0;
}

