/* advautomesh_h_tri2d_property.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

double Distance (double x0, double y0, 
		 double x1, double y1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) + (y1 - y0) * (y1 - y0));
}

double SignedTriArea
(double x0, double y0,
 double x1, double y1,
 double x2, double y2)
{
  return ((x1 - x0) * (y2 - y0) - (y1 - y0) * (x2 - x0)) / 2;
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][2];



/* Element */

#define N_ELEMENT_NODES 3
#define N_ELEMENT_EDGES 3
#define MAX_ELEMENTS 1000000

const int EdgeNodeMap[N_ELEMENT_EDGES][2] = {
  { 1, 2 },
  { 2, 0 },
  { 0, 1 }
};

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



double Element_signedArea 
(int elementId)
{
  int nodeId0, nodeId1, nodeId2;
  double x0, y0;
  double x1, y1;
  double x2, y2;

  assert(0 <= elementId);
  assert(elementId < NElements);

  nodeId0 = Element_nodeIds[elementId][0];
  nodeId1 = Element_nodeIds[elementId][1];
  nodeId2 = Element_nodeIds[elementId][2];
  
  x0 = Node_coord[nodeId0][0];
  y0 = Node_coord[nodeId0][1];
  
  x1 = Node_coord[nodeId1][0];
  y1 = Node_coord[nodeId1][1];
  
  x2 = Node_coord[nodeId2][0];
  y2 = Node_coord[nodeId2][1];
  
  return SignedTriArea (x0, y0, 
			x1, y1, 
			x2, y2);
}

void Element_getEdgeLength
(int elementId,
 double *minLength_OUT, double *maxLength_OUT)
{
  int isValid;
  int iElementEdge;

  assert(0 <= elementId);
  assert(elementId < NElements);

  isValid = 0;
  for (iElementEdge = 0; 
       iElementEdge < N_ELEMENT_EDGES; 
       iElementEdge++) {
    int edgeNodeId0 
      = Element_nodeIds[elementId][EdgeNodeMap[iElementEdge][0]];
    int edgeNodeId1 
      = Element_nodeIds[elementId][EdgeNodeMap[iElementEdge][1]];
    
    double x0 = Node_coord[edgeNodeId0][0];
    double y0 = Node_coord[edgeNodeId0][1];
    
    double x1 = Node_coord[edgeNodeId1][0];
    double y1 = Node_coord[edgeNodeId1][1];
    
    double length = Distance (x0, y0, 
			      x1, y1);
    
    if (!isValid) {
      *minLength_OUT = length;
      *maxLength_OUT = length;
      isValid = 1;
    }
    if (length < *minLength_OUT) *minLength_OUT = length;
    if (*maxLength_OUT < length) *maxLength_OUT = length;
  }
  assert(isValid);
}
 
double Element_averageEdgeLength
(int elementId)
{
  int iElementEdge;
  double average;

  assert(0 <= elementId);
  assert(elementId < NElements);

  average = 0.0;
  for (iElementEdge = 0; 
       iElementEdge < N_ELEMENT_EDGES; 
       iElementEdge++) {
    int edgeNodeId0 
      = Element_nodeIds[elementId][EdgeNodeMap[iElementEdge][0]];
    int edgeNodeId1 
      = Element_nodeIds[elementId][EdgeNodeMap[iElementEdge][1]];
    
    double x0 = Node_coord[edgeNodeId0][0];
    double y0 = Node_coord[edgeNodeId0][1];
    
    double x1 = Node_coord[edgeNodeId1][0];
    double y1 = Node_coord[edgeNodeId1][1];
    
    average += Distance (x0, y0, 
			 x1, y1);
  }
  average /= N_ELEMENT_EDGES;
  return average;
}
 
void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
  }
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y;

    fscanf (fp, "%d %lf %lf", 
	    &nodeId, 
	    &x, &y);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Property */

double Area;
double MinRangeX, MinRangeY;
double MaxRangeX, MaxRangeY;
double MinSize, MaxSize, AverageSize;



void WritePropertyFile (const char *fileName)
{
  FILE *fp;

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%22.15e\n", 
	   Area);
  fprintf (fp, "%22.15e %22.15e\n", 
	   MinRangeX, MinRangeY);
  fprintf (fp, "%22.15e %22.15e\n", 
	   MaxRangeX, MaxRangeY);
  fprintf (fp, "%22.15e %22.15e %22.15e\n", 
	   MinSize, MaxSize, AverageSize);

  fclose (fp);
}



/************************
	 Command
*************************/

void PerformCommand (void)
{
  int iElement;
  int isValid;

  assert(0 < NElements);
  
  Mesh_getRange
    (&MinRangeX, &MinRangeY,
     &MaxRangeX, &MaxRangeY);

  isValid = 0;
  AverageSize = 0.0;
  Area = 0.0;
  for (iElement = 0; iElement < NElements; iElement++) {
    double min, max;
    
    Element_getEdgeLength (iElement,
			   &min, &max);
    if (!isValid) {
      MinSize = min;
      MaxSize = max;
      isValid = 1;
    }
    if (min < MinSize) MinSize = min;
    if (MaxSize < max) MaxSize = max;
    
    AverageSize += Element_averageEdgeLength (iElement);
    Area += fabs (Element_signedArea (iElement));
  }
  assert(isValid);
  AverageSize /= NElements;
}



int main (int argc, char *argv[])
{
  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tri2d_property  node_file  element_file \n"
       "        property_file_OUT \n");
    fprintf 
      (stderr, 
       " The contents of <property_file_OUT> are, \n"
       "   area \n"
       "   minRangeX minRangeY \n"
       "   maxRangeX maxRangeY \n"
       "   minSize maxSize averageSize \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);

  PerformCommand ();

  WritePropertyFile (argv[3]);

  return 0;
}
