/* advautomesh_h_tri2d_swap.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

double Distance (double x0, double y0, 
		 double x1, double y1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) + (y1 - y0) * (y1 - y0));
}

void GetLineEquation 
(double x0, double y0, 
 double x1, double y1, 
 double *a_OUT, double *b_OUT, double *c_OUT) 
{
  double dx = x1 - x0;
  double dy = y1 - y0;

  double a0 = -dy;
  double b0 = dx;
  double c0 = dy * x0 - y0 * dx;

  double len = sqrt (a0 * a0 + b0 * b0);

  *a_OUT = a0 / len;  
  *b_OUT = b0 / len;  
  *c_OUT = c0 / len;
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][2];



/* Element */

#define N_ELEMENT_NODES 3
#define N_ELEMENT_EDGES 3
#define MAX_ELEMENTS 1000000

const int EdgeNodeMap[N_ELEMENT_EDGES][2] = {
  { 1, 2 },
  { 2, 0 },
  { 0, 1 }
};

const int EdgeOppositeNodeMap[N_ELEMENT_EDGES] = { 
  0, 1, 2 
};

int NElements;
int Element_isDeleted[MAX_ELEMENTS];
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



int Element_containsNode 
(int elementId, 
 int nodeId)
{
  int iElementNode;

  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(0 <= nodeId);
  assert(nodeId < NNodes);

  for (iElementNode = 0; 
       iElementNode < N_ELEMENT_NODES; 
       iElementNode++) {
    if (Element_nodeIds[elementId][iElementNode] == nodeId) {
      return 1;
    }
  }

  return 0;
}

int Mesh_addNewElement (int nodeId0, int nodeId1, int nodeId2)
{
  int elementId = NElements;

  assert(0 <= nodeId0);
  assert(nodeId0 < NNodes);
  assert(0 <= nodeId1);
  assert(nodeId1 < NNodes);
  assert(0 <= nodeId2);
  assert(nodeId2 < NNodes);

  assert(NElements < MAX_ELEMENTS);
  NElements++;

  Element_isDeleted[elementId] = 0;

  Element_nodeIds[elementId][0] = nodeId0;
  Element_nodeIds[elementId][1] = nodeId1;
  Element_nodeIds[elementId][2] = nodeId2;

  return elementId;
}

void Mesh_deleteElement (int elementId)
{
  assert(0 <= elementId);
  assert(elementId < NElements);

  Element_isDeleted[elementId] = 1;
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y;

    fscanf (fp, "%d %lf %lf", 
	    &nodeId, 
	    &x, &y);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    Element_isDeleted[iElement] = 0;
  }

  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void WriteNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    fprintf (fp, "%d   %22.15e %22.15e\n", 
	     iNode,
	     Node_coord[iNode][0],
	     Node_coord[iNode][1]); 
  }

  fclose (fp);
}

void WriteElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;
  int nActualElements;
  int count;

  assert(0 < NNodes);
  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  nActualElements = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) continue;
    
    nActualElements++;
  }

  fprintf (fp, "%d\n", 
	   nActualElements);
  
  count = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) continue;

    fprintf (fp, "%d   ", 
	     count); 
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      fprintf (fp, "%d ", 
	       Element_nodeIds[iElement][iElementNode]); 
    }
    fprintf (fp, "\n");
    count++;
  }
  
  fclose (fp);
}



/************************
	 Command
*************************/

int ShapeIsBad 
(int edgeNodeId0, int edgeNodeId1, int oppositeNodeId,
 double ratio)
{
  double ex0 = Node_coord[edgeNodeId0][0];
  double ey0 = Node_coord[edgeNodeId0][1];

  double ex1 = Node_coord[edgeNodeId1][0];
  double ey1 = Node_coord[edgeNodeId1][1];

  double ox = Node_coord[oppositeNodeId][0];
  double oy = Node_coord[oppositeNodeId][1];

  double edgeLength = Distance (ex0, ey0, 
				ex1, ey1);
  double a, b, c;
  double distance;

  GetLineEquation (ex0, ey0, ex1, ey1,
		   &a, &b, &c);
  distance = fabs (a * ox + b * oy + c);
  if (distance < edgeLength * ratio) {
    return 1;
  }

  return 0;
}

int GetNeighbor 
(int elementId, int edgeNodeId0, int edgeNodeId1,
 int *neighborElementId_OUT, int *oppositeNodeId_OUT)
{
  int iElement;
  int iElementNode;

  *neighborElementId_OUT = -1;
  *oppositeNodeId_OUT = -1;

  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isDeleted[iElement]) continue;

    if (iElement == elementId) continue;
    
    if (Element_containsNode (iElement, edgeNodeId0)
	&& Element_containsNode (iElement, edgeNodeId1)) {
      for (iElementNode = 0;
	   iElementNode < N_ELEMENT_NODES;
	   iElementNode++) {
	int nodeId = Element_nodeIds[iElement][iElementNode];
	
	if (nodeId != edgeNodeId0
	    && nodeId != edgeNodeId1) {
	  *neighborElementId_OUT = iElement;
	  *oppositeNodeId_OUT = nodeId;
	  return 1;
	}
      }
    }
  }
  return 0;
}

void PerformCommand (double ratio)
{
  int iElement;
  int iElementEdge;
  int nOldElements = NElements;
  
  for (iElement = 0; iElement < nOldElements; iElement++) {
    if (Element_isDeleted[iElement]) continue;
    
    for (iElementEdge = 0; 
	 iElementEdge < N_ELEMENT_EDGES; 
	 iElementEdge++) {
      int edgeNodeId0 
	= Element_nodeIds[iElement][EdgeNodeMap[iElementEdge][0]];
      int edgeNodeId1 
	= Element_nodeIds[iElement][EdgeNodeMap[iElementEdge][1]];
      int oppositeNodeId 
	= Element_nodeIds[iElement][EdgeOppositeNodeMap[iElementEdge]];

      if (ShapeIsBad (edgeNodeId0, edgeNodeId1, 
		      oppositeNodeId, ratio)) {
	int neighborElementId;
	int the4thNodeId;

	if (GetNeighbor (iElement, edgeNodeId0, edgeNodeId1,
			 &neighborElementId, &the4thNodeId)) {
	  Mesh_addNewElement (oppositeNodeId, edgeNodeId0, the4thNodeId);
	  Mesh_addNewElement (the4thNodeId, edgeNodeId1, oppositeNodeId);
	  Mesh_deleteElement (iElement);
	  Mesh_deleteElement (neighborElementId);
	  break;
	}
      }
    }
  }
}



int main (int argc, char *argv[])
{
  double ratio;

  if (argc != 6) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tri2d_swap  node_file  element_file  ratio \n"
       "        node_file_OUT  element_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);

  ratio = atof (argv[3]);
  PerformCommand (ratio);

  WriteNodeFile (argv[4]);
  WriteElementFile (argv[5]);

  return 0;
}
