/* advautomesh_h_tri3d_deform_view.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autogl.h>



/************************
	 Basic
*************************/

double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][3];
double Node_vector[MAX_NODES][3];



/* Element */

#define N_ELEMENT_NODES 3
#define MAX_ELEMENTS 2000000

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  *minRangeZ_OUT = *maxRangeZ_OUT = Node_coord[0][2];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    double z = Node_coord[iNode][2];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
    if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
  }
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void ReadNodeVectorFile (const char *fileName)
{
  FILE *fp;
  int count;
  int iNode;
  
  assert(0 < NNodes);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &count);
  assert(count == NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double vectorX, vectorY, vectorZ;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId,
	    &vectorX, &vectorY, &vectorZ);
    assert(nodeId == iNode);
    Node_vector[iNode][0] = vectorX;
    Node_vector[iNode][1] = vectorY;
    Node_vector[iNode][2] = vectorZ;
  }

  fclose (fp);
}



/********************************
	 Visualization
********************************/

int OriginalEdge = 0;
int OriginalFill = 0;
int DeformedEdge = 1;
int DeformedFill = 1;
double Scale;

void SetView (void)
{
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  double centerX, centerY, centerZ;
  double size;
  
  Mesh_getRange 
    (&minRangeX, &minRangeY, &minRangeZ,
     &maxRangeX, &maxRangeY, &maxRangeZ);
  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;
  centerZ = (minRangeZ + maxRangeZ) / 2;
  size = Distance (minRangeX, minRangeY, minRangeZ,
		   maxRangeX, maxRangeY, maxRangeZ) / 2;

  AutoGL_SetViewSize (size);
  AutoGL_SetViewCenter (centerX, centerY, centerZ); 
}

void SetVectorScale (void)
{
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  double centerX, centerY, centerZ;
  double size;
  double maxLength;
  int iNode;

  assert(0 < NNodes);

  Mesh_getRange 
    (&minRangeX, &minRangeY, &minRangeZ,
     &maxRangeX, &maxRangeY, &maxRangeZ);
  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;
  centerZ = (minRangeZ + maxRangeZ) / 2;
  size = Distance (minRangeX, minRangeY, minRangeZ,
		   maxRangeX, maxRangeY, maxRangeZ) / 2;

  maxLength = 0.0;
  for (iNode = 0; iNode < NNodes; iNode++) {
    double vx = Node_vector[iNode][0];
    double vy = Node_vector[iNode][1];
    double vz = Node_vector[iNode][2];
    double length = sqrt (vx * vx + vy * vy + vz * vz);
    
    if (maxLength < length) maxLength = length;
  }

  Scale = size / maxLength / 5;
}

#if 0
#include "../../util/time_measure.c"
#define TIME_MEASURE
#endif

void RedrawView (void) 
{
  int iElement;

#ifdef TIME_MEASURE
  StartMeasureTime ();
#endif

  for (iElement = 0; iElement < NElements; iElement++) {
    int nodeId0 = Element_nodeIds[iElement][0];
    int nodeId1 = Element_nodeIds[iElement][1];
    int nodeId2 = Element_nodeIds[iElement][2];

    double nodeX0 = Node_coord[nodeId0][0];
    double nodeY0 = Node_coord[nodeId0][1];
    double nodeZ0 = Node_coord[nodeId0][2];

    double nodeX1 = Node_coord[nodeId1][0];
    double nodeY1 = Node_coord[nodeId1][1];
    double nodeZ1 = Node_coord[nodeId1][2];

    double nodeX2 = Node_coord[nodeId2][0];
    double nodeY2 = Node_coord[nodeId2][1];
    double nodeZ2 = Node_coord[nodeId2][2];

    double vectorX0 = Node_vector[nodeId0][0];
    double vectorY0 = Node_vector[nodeId0][1];
    double vectorZ0 = Node_vector[nodeId0][2];

    double vectorX1 = Node_vector[nodeId1][0];
    double vectorY1 = Node_vector[nodeId1][1];
    double vectorZ1 = Node_vector[nodeId1][2];

    double vectorX2 = Node_vector[nodeId2][0];
    double vectorY2 = Node_vector[nodeId2][1];
    double vectorZ2 = Node_vector[nodeId2][2];

    double tipX0 = nodeX0 + vectorX0 * Scale;
    double tipY0 = nodeY0 + vectorY0 * Scale;
    double tipZ0 = nodeZ0 + vectorZ0 * Scale;

    double tipX1 = nodeX1 + vectorX1 * Scale;
    double tipY1 = nodeY1 + vectorY1 * Scale;
    double tipZ1 = nodeZ1 + vectorZ1 * Scale;

    double tipX2 = nodeX2 + vectorX2 * Scale;
    double tipY2 = nodeY2 + vectorY2 * Scale;
    double tipZ2 = nodeZ2 + vectorZ2 * Scale;

    if (OriginalEdge) {
      AutoGL_SetColor (1, 0, 0);
      AutoGL_DrawLine (nodeX0, nodeY0, nodeZ0,
		       nodeX1, nodeY1, nodeZ1);
      AutoGL_DrawLine (nodeX1, nodeY1, nodeZ1,
		       nodeX2, nodeY2, nodeZ2);
      AutoGL_DrawLine (nodeX2, nodeY2, nodeZ2,
		       nodeX0, nodeY0, nodeZ0);
    }
    if (OriginalFill) {
      AutoGL_SetColor (0, 1, 1);
      AutoGL_DrawTriangle (nodeX0, nodeY0, nodeZ0,
			   nodeX1, nodeY1, nodeZ1,
			   nodeX2, nodeY2, nodeZ2);
    }

    if (DeformedEdge) {
      AutoGL_SetColor (0, 1, 0);
      AutoGL_DrawLine (tipX0, tipY0, tipZ0,
		       tipX1, tipY1, tipZ1);
      AutoGL_DrawLine (tipX1, tipY1, tipZ1,
		       tipX2, tipY2, tipZ2);
      AutoGL_DrawLine (tipX2, tipY2, tipZ2,
		       tipX0, tipY0, tipZ0);
    }
    if (DeformedFill) {
      AutoGL_SetColor (1, 0, 1);
      AutoGL_DrawTriangle (tipX0, tipY0, tipZ0,
			   tipX1, tipY1, tipZ1,
			   tipX2, tipY2, tipZ2);
    }
  }

#ifdef TIME_MEASURE
  EndMeasureTime ();
#endif
}

void UpdateModel (void) 
{
  AutoGL_DrawView ();
}

void Quit (void) 
{
  AutoGL_Terminate ();
  exit (0);
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tri3d_deform_view  node_file  element_file \n"
       "        node_vector_file \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadNodeVectorFile (argv[3]);

  SetView ();
  SetVectorScale ();
  
  AutoGL_SetViewRedrawCallback (RedrawView);

  AutoGL_SetDefaultCallbackInMode3D (NULL);
  AutoGL_EnableDragInMode3D ();



  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode3D ();

  AutoGL_AddCallback (Quit, "Quit");       

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");

  AutoGL_AddReal (&Scale, "Scale");
  AutoGL_AddBoolean (&DeformedEdge, "DeformedEdge");
  AutoGL_AddBoolean (&DeformedFill, "DeformedFill");
  AutoGL_AddBoolean (&OriginalEdge, "OriginalEdge");
  AutoGL_AddBoolean (&OriginalFill, "OriginalFill");

  AutoGL_AddCallback (UpdateModel, "UpdateModel");      
}
