/* advautomesh_h_tri3d_isoline2.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_scalar[MAX_NODES];



/* Element */

#define N_ELEMENT_NODES 3
#define N_ELEMENT_EDGES 3
#define MAX_ELEMENTS 2000000

const int EdgeNodeMap[N_ELEMENT_EDGES][2] = {
  { 1, 2 },
  { 2, 0 },
  { 0, 1 }
};

const int NIsolineEdges[4] = {
  0, 2, 2, 2
};

const int IsolineEdgeIds[4][2] = {
  { -1, -1 },
  { 1, 2 },
  { 2, 0 },
  { 0, 1 }
};

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



void ReadNodeFile (const char *fileName)
{
  FILE *fp;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  /* skip coordinate */

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void ReadNodeScalarFile (const char *fileName)
{
  FILE *fp;
  int count;
  int iNode;
  
  assert(0 < NNodes);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &count);
  assert(count == NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double scalar;

    fscanf (fp, "%d %lf", 
	    &nodeId,
	    &scalar);
    assert(nodeId == iNode);
    Node_scalar[iNode] = scalar;
  }

  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Isoline */

#define MAX_LEVELS 10000
#define MAX_ISOLINES 1000000

int NLevels;
int Level_startIsolineId[MAX_LEVELS];
int Level_nIsolines[MAX_LEVELS];
int NIsolines;
int Isoline_elementId[MAX_ISOLINES];
int Isoline_elementEdgeIds[MAX_ISOLINES][2];
double Isoline_ratios[MAX_ISOLINES][2];



void IsolineGroup_addLevel (void)
{
  assert(NLevels < MAX_LEVELS);

  Level_startIsolineId[NLevels] = NIsolines;
  Level_nIsolines[NLevels] = 0;
  NLevels++;
}

void IsolineGroup_addIsoline 
(int elementId, 
 int elementEdgeId0, double ratio0,
 int elementEdgeId1, double ratio1)
{
  assert(0 < NLevels);
  assert(NIsolines < MAX_ISOLINES);
  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(0 <= elementEdgeId0);
  assert(elementEdgeId0 < N_ELEMENT_EDGES);
  assert(0 <= elementEdgeId1);
  assert(elementEdgeId1 < N_ELEMENT_EDGES);

  Level_nIsolines[NLevels - 1]++;

  Isoline_elementId[NIsolines] = elementId;
  Isoline_elementEdgeIds[NIsolines][0] = elementEdgeId0;
  Isoline_elementEdgeIds[NIsolines][1] = elementEdgeId1;
  Isoline_ratios[NIsolines][0] = ratio0;
  Isoline_ratios[NIsolines][1] = ratio1;

  NIsolines++;
}

void WriteIsolineFile (const char *fileName)
{
  FILE *fp;
  int iLevel;
  int iIsoline;
  
  assert(0 < NLevels);
  assert(0 < NIsolines);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NLevels);
  
  for (iLevel = 0; iLevel < NLevels; iLevel++) {
    int startId = Level_startIsolineId[iLevel];
    int nIsolines = Level_nIsolines[iLevel];
    
    fprintf (fp, "%d   %d\n", 
	     iLevel,
	     nIsolines);

    for (iIsoline = startId; 
	 iIsoline < startId + nIsolines; 
	 iIsoline++) {
      fprintf (fp, "%d   %d   %d %22.15e   %d %22.15e\n", 
	       iIsoline - startId,
	       Isoline_elementId[iIsoline],
	       Isoline_elementEdgeIds[iIsoline][0],
	       Isoline_ratios[iIsoline][0],
	       Isoline_elementEdgeIds[iIsoline][1],
	       Isoline_ratios[iIsoline][1]);
    }
  }
  
  fclose (fp);
}



/************************
	 Command
*************************/

AutoRes_Bucket1D IsolineBucket;

void CreateIsolineBucket (void)
{
  int iNode;
  double minRange, maxRange;
  int nDivisions;
  double offset;
  int nAverageLinksPerCell;
  int nOffsetDivisions = 23;
  
  minRange = maxRange = Node_scalar[0];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double value = Node_scalar[iNode];
    
    if (value < minRange) minRange = value;
    if (maxRange < value) maxRange = value;
  }

  nDivisions = (int)(pow (NNodes, 0.5));
  if (nDivisions < 1) nDivisions = 1;
#if 0
  nAverageLinksPerCell = (int)((double)NElements / nDivisions * 3.0);
#else
  nAverageLinksPerCell = (int)((double)NElements / nDivisions * 5.0);
#endif

  offset = (maxRange - minRange) / nOffsetDivisions;
  
  AutoRes_Bucket1D_initialize 
    (&IsolineBucket,
     minRange - offset,
     maxRange + offset,
     nDivisions, 
     nAverageLinksPerCell);
}

void AddElementsIntoIsolineBucket (void)
{
  int iElement;
  int iElementNode;

  for (iElement = 0; iElement < NElements; iElement++) {
    int nodeId0 = Element_nodeIds[iElement][0];
    double minRange, maxRange;
    
    minRange = maxRange = Node_scalar[nodeId0];
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId = Element_nodeIds[iElement][iElementNode]; 
      double value = Node_scalar[nodeId];
      
      if (value < minRange) minRange = value;
      if (maxRange < value) maxRange = value;
    }
    
    AutoRes_Bucket1D_addItemWithin 
      (&IsolineBucket,
       minRange,
       maxRange,
       iElement);
  }
}

void DeleteIsolineBucket (void)
{
  AutoRes_Bucket1D_terminate 
    (&IsolineBucket);
}

int IsolineMask (int elementId, double levelValue)
{
  int nodeId0 = Element_nodeIds[elementId][0];
  int nodeId1 = Element_nodeIds[elementId][1];
  int nodeId2 = Element_nodeIds[elementId][2];
  unsigned int mask = 0;
  
  if (levelValue < Node_scalar[nodeId0]) mask |= 1;
  if (levelValue < Node_scalar[nodeId1]) mask |= 2;
  if (levelValue < Node_scalar[nodeId2]) mask |= 4;
  if ((mask & 4) != 0) mask = (~mask) & 7;

  assert(0 <= mask);
  assert(mask < 4);

  return mask;
}

int GetIsoline 
(int elementId, double levelValue, 
 int elementEdgeIds_OUT[2], double ratios_OUT[2])
{
  unsigned int mask = IsolineMask (elementId, levelValue);
  int nEdges;
  int iEdge;

  nEdges = NIsolineEdges[mask];
  if (nEdges == 0) return 0;
  
  assert(nEdges == 2);

  for (iEdge = 0; iEdge < nEdges; iEdge++) {
    int elementEdgeId = IsolineEdgeIds[mask][iEdge];

    int elementNodeId0 = EdgeNodeMap[elementEdgeId][0];
    int elementNodeId1 = EdgeNodeMap[elementEdgeId][1];

    int nodeId0 = Element_nodeIds[elementId][elementNodeId0];
    int nodeId1 = Element_nodeIds[elementId][elementNodeId1];

    double ratio = (levelValue - Node_scalar[nodeId0]) 
      / (Node_scalar[nodeId1] - Node_scalar[nodeId0]);
    
    elementEdgeIds_OUT[iEdge] = elementEdgeId; 
    ratios_OUT[iEdge] = ratio;
  }    

  return 1;
}

void CheckElementIsoline (int elementId, double levelValue)
{
  int elementEdgeIds[2];
  double ratios[2];
  
  if (GetIsoline
      (elementId, levelValue,
       elementEdgeIds, ratios)) {
    IsolineGroup_addIsoline 
      (elementId, 
       elementEdgeIds[0], ratios[0],
       elementEdgeIds[1], ratios[1]);
  }
}

void MakeIsoline (double levelValue)
{
  int cx;
  int linkIndex;
  
  AutoRes_Bucket1D_getIndexAt 
    (&IsolineBucket, 
     levelValue,
     &cx);
  
  if (AutoRes_Bucket1D_isOutOfRange 
      (&IsolineBucket, 
       cx)) {
    return;
  } else {
    for (linkIndex = AutoRes_Bucket1D_startLinkIndexIn 
	   (&IsolineBucket, 
	    cx);
	 linkIndex != -1;
	 linkIndex = AutoRes_Bucket1D_nextLinkIndex 
	   (&IsolineBucket, 
	    linkIndex)) {
      int elementId = AutoRes_Bucket1D_itemId 
	(&IsolineBucket, 
	 linkIndex);
      
      CheckElementIsoline (elementId, levelValue);
    }
  }
}

void PerformCommand 
(double minLevel, double maxLevel, int nDivisions)
{
  int iLevel;

  CreateIsolineBucket ();
  AddElementsIntoIsolineBucket ();

  for (iLevel = 0; iLevel <= nDivisions; iLevel++) {
    double levelValue = minLevel 
      + (maxLevel - minLevel) / nDivisions * iLevel;
    
    IsolineGroup_addLevel ();
    MakeIsoline (levelValue);
  }

  DeleteIsolineBucket ();
}



int main (int argc, char *argv[])
{
  double minLevel, maxLevel;
  int nDivisions;
  
  if (argc != 8) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tri3d_isoline2  node_file  element_file \n"
       "        node_scalar_file  minLevel  maxLevel  nDivisions \n"
       "        isoline_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadNodeScalarFile (argv[3]);
  minLevel = atof (argv[4]);
  maxLevel = atof (argv[5]);
  nDivisions = atoi (argv[6]);

  PerformCommand (minLevel, maxLevel, nDivisions);

  WriteIsolineFile (argv[7]);

  return 0;
}
