/* advautomesh_h_tri3d_frame.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

void GetNormalizedVector 
(double x, double y, double z,
 double *nx_OUT, double *ny_OUT, double *nz_OUT)
{
  double length = sqrt (x * x + y * y + z * z);

  *nx_OUT = x / length;
  *ny_OUT = y / length;
  *nz_OUT = z / length;
}

void GetVectorProduct 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  *x_OUT = y0 * z1 - z0 * y1;
  *y_OUT = z0 * x1 - x0 * z1;
  *z_OUT = x0 * y1 - y0 * x1;
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][3];



/* Element */

#define N_ELEMENT_NODES 3
#define N_ELEMENT_EDGES 3
#define MAX_ELEMENTS 2000000

const int EdgeNodeMap[N_ELEMENT_EDGES][2] = {
  { 1, 2 },
  { 2, 0 },
  { 0, 1 }
};

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];
int Element_neighborIds[MAX_ELEMENTS][N_ELEMENT_EDGES];



void Element_getNormal 
(int elementId,
 double *nx_OUT, double *ny_OUT, double *nz_OUT)
{
  int nodeId0 = Element_nodeIds[elementId][0];
  int nodeId1 = Element_nodeIds[elementId][1];
  int nodeId2 = Element_nodeIds[elementId][2];

  double x0 = Node_coord[nodeId0][0];
  double y0 = Node_coord[nodeId0][1];
  double z0 = Node_coord[nodeId0][2];

  double x1 = Node_coord[nodeId1][0];
  double y1 = Node_coord[nodeId1][1];
  double z1 = Node_coord[nodeId1][2];

  double x2 = Node_coord[nodeId2][0];
  double y2 = Node_coord[nodeId2][1];
  double z2 = Node_coord[nodeId2][2];

  GetVectorProduct (x0 - x1, y0 - y1, z0 - z1,
		    x0 - x2, y0 - y2, z0 - z2,
		    nx_OUT, ny_OUT, nz_OUT);
  GetNormalizedVector (*nx_OUT, *ny_OUT, *nz_OUT,
		       nx_OUT, ny_OUT, nz_OUT);
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void ReadElementNeighborFile (const char *fileName)
{
  FILE *fp;
  int count;
  int iElement;
  int iElementEdge;
  
  assert(0 < NElements);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &count);
  assert(count == NElements);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId);
    assert(elementId == iElement);

    for (iElementEdge = 0; 
	 iElementEdge < N_ELEMENT_EDGES; 
	 iElementEdge++) {
      int neighborId;

      fscanf (fp, "%d", 
	      &neighborId);
      assert((0 <= neighborId && neighborId < NElements) 
	     || neighborId == -1);
      Element_neighborIds[iElement][iElementEdge] = neighborId;
    }
  }

  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Frame */

#define MAX_FRAMES 1000000

int NFrames;
double Frame_coord0[MAX_FRAMES][3];
double Frame_coord1[MAX_FRAMES][3];



void FrameGroup_addFrame 
(double x0, double y0, double z0, 
 double x1, double y1, double z1)
{
  assert(NFrames < MAX_FRAMES);

  Frame_coord0[NFrames][0] = x0;
  Frame_coord0[NFrames][1] = y0;
  Frame_coord0[NFrames][2] = z0;

  Frame_coord1[NFrames][0] = x1;
  Frame_coord1[NFrames][1] = y1;
  Frame_coord1[NFrames][2] = z1;

  NFrames++;
}

void WriteFrameFile (const char *fileName)
{
  FILE *fp;
  int iFrame;
  
  assert(0 < NFrames);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NFrames);
  
  for (iFrame = 0; iFrame < NFrames; iFrame++) {
    fprintf (fp, "%d   %22.15e %22.15e %22.15e   %22.15e %22.15e %22.15e\n", 
	     iFrame,
	     Frame_coord0[iFrame][0],
	     Frame_coord0[iFrame][1],
	     Frame_coord0[iFrame][2],
	     Frame_coord1[iFrame][0],
	     Frame_coord1[iFrame][1],
	     Frame_coord1[iFrame][2]);
  }

  fclose (fp);
}



/************************
	 Command
*************************/

void PerformCommand (double angle)
{
  int iElement;
  int iElementEdge;
  double radian = M_PI * angle / 180;

  for (iElement = 0; iElement < NElements; iElement++) {
    double nx, ny, nz;

    Element_getNormal (iElement,
		       &nx, &ny, &nz);

    for (iElementEdge = 0; 
	 iElementEdge < N_ELEMENT_EDGES; 
	 iElementEdge++) {
      int nodeId0 = Element_nodeIds
	[iElement][EdgeNodeMap[iElementEdge][0]];
      int nodeId1 = Element_nodeIds
	[iElement][EdgeNodeMap[iElementEdge][1]];
      int neighborId 
	= Element_neighborIds[iElement][iElementEdge];

      if (neighborId != -1) {
	double nx2, ny2, nz2;
	
	Element_getNormal (neighborId,
			   &nx2, &ny2, &nz2);
	if (nx * nx2 + ny * ny2 + nz * nz2 <= cos (radian)) {
	  FrameGroup_addFrame 
	    (Node_coord[nodeId0][0],
	     Node_coord[nodeId0][1],
	     Node_coord[nodeId0][2],
	     Node_coord[nodeId1][0],
	     Node_coord[nodeId1][1],
	     Node_coord[nodeId1][2]);
	}
      } else {
	FrameGroup_addFrame 
	  (Node_coord[nodeId0][0],
	   Node_coord[nodeId0][1],
	   Node_coord[nodeId0][2],
	   Node_coord[nodeId1][0],
	   Node_coord[nodeId1][1],
	   Node_coord[nodeId1][2]);
      }
    }
  }
}



int main (int argc, char *argv[])
{
  double angle;

  if (argc != 6) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tri3d_frame  node_file  element_file \n"
       "        element_neighbor_file  angle \n"
       "        frame_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadElementNeighborFile (argv[3]);
  angle = atof (argv[4]);

  PerformCommand (angle);

  WriteFrameFile (argv[5]);

  return 0;
}
