/* advautomesh_h_tri3d_renumber_node.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}



/************************
	 Mesh
*************************/

/* Node */

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][3];
int Node_newId[MAX_NODES];
int Node_oldId[MAX_NODES];



/* Element */

#define N_ELEMENT_NODES 3
#define MAX_ELEMENTS 2000000

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];



void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  *minRangeZ_OUT = *maxRangeZ_OUT = Node_coord[0][2];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    double z = Node_coord[iNode][2];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
    if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
  }
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void WriteNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int oldNodeId = Node_oldId[iNode];

    fprintf (fp, "%d   %22.15e %22.15e %22.15e\n", 
	     iNode,
	     Node_coord[oldNodeId][0],
	     Node_coord[oldNodeId][1],
	     Node_coord[oldNodeId][2]); 
  }

  fclose (fp);
}

void WriteElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NElements);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    fprintf (fp, "%d   ", 
	     iElement); 
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId = Element_nodeIds[iElement][iElementNode];
      int newNodeId = Node_newId[nodeId];

      fprintf (fp, "%d ", 
	       newNodeId); 
    }
    fprintf (fp, "\n");
  }
  
  fclose (fp);
}



/************************
	 Command
*************************/

int OuterNode (void)
{
  int iNode;
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  double centerX, centerY, centerZ;
  int outerNodeId;
  double currentDistance;

  Mesh_getRange
    (&minRangeX, &minRangeY, &minRangeZ,
     &maxRangeX, &maxRangeY, &maxRangeZ);

  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;
  centerZ = (minRangeZ + maxRangeZ) / 2;
  outerNodeId = 0;
  currentDistance = Distance 
    (centerX, centerY, centerZ, 
     Node_coord[0][0], Node_coord[0][1], Node_coord[0][2]);
  for (iNode = 0; iNode < NNodes; iNode++) {
    double distance = Distance 
      (centerX, centerY, centerZ, 
       Node_coord[iNode][0], Node_coord[iNode][1], Node_coord[iNode][2]);

    if (currentDistance < distance) {
      outerNodeId = iNode;
      currentDistance = distance;
    }
  }

  return outerNodeId;
}

void SetNumber (int outerNodeId)
{
  int iNode;
  int iElement;
  int iElementNode;
  int count;

  for (iNode = 0; iNode < NNodes; iNode++) {
    Node_newId[iNode] = -1;
  }

  count = 0;

  Node_newId[outerNodeId] = count;
  count++;

  for (;;) {
    int noChange = 1;

    for (iElement = 0; iElement < NElements; iElement++) {
      int elementHasNewId = 0;
      
      for (iElementNode = 0; 
	   iElementNode < N_ELEMENT_NODES; 
	   iElementNode++) {
	int nodeId = Element_nodeIds[iElement][iElementNode]; 
	
	if (Node_newId[nodeId] != -1) {
	  elementHasNewId = 1;
	  break;
	}
      }
      if (elementHasNewId) {
	for (iElementNode = 0; 
	     iElementNode < N_ELEMENT_NODES; 
	     iElementNode++) {
	  int nodeId = Element_nodeIds[iElement][iElementNode]; 
	  
	  if (Node_newId[nodeId] == -1) {
	    Node_newId[nodeId] = count;
	    count++;
	    noChange = 0;
	  }
	}
      }
    }
    
    if (noChange) break;
  }

  assert(count == NNodes);
  for (iNode = 0; iNode < NNodes; iNode++) {
    assert(0 <= Node_newId[iNode]);
    assert(Node_newId[iNode] < NNodes);
  }

  for (iNode = 0; iNode < NNodes; iNode++) {
    Node_oldId[iNode] = -1;
  }
  for (iNode = 0; iNode < NNodes; iNode++) {
    int newId = Node_newId[iNode]; 
    
    assert(Node_oldId[newId] == -1);
    Node_oldId[newId] = iNode;
  }
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    assert(0 <= Node_oldId[iNode]);
    assert(Node_oldId[iNode] < NNodes);
  }
}

void PerformCommand (void)
{
  int outerNodeId = OuterNode ();
  
  SetNumber (outerNodeId);
}



int main (int argc, char *argv[])
{
  if (argc != 5) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tri3d_renumber_node  node_file  element_file \n"
       "        node_file_OUT  element_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);

  PerformCommand ();

  WriteNodeFile (argv[3]);
  WriteElementFile (argv[4]);

  return 0;
}


