/* advautomesh_export_ADVENTURE_pcg.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>
#include <automesh.h>



static AutoMesh_Node3D *Node;
static AutoMesh_SurfaceElement3D *Element;
static AutoMesh_SurfaceEdge3D *Edge;
static AutoMesh_SurfaceGrouper3D *Grouper;

static void WriteVertex 
(FILE *fp)
{
  int nNodes = AutoMesh_Node3D_nNodes (Node);
  int nFrameVertexs = AutoMesh_SurfaceGrouper3D_nFrameVertexs (Grouper);
  int iNode;

  fprintf (fp, "#mainVertexInfo\n");

  fprintf (fp, "mainVertexN %d\n", 
	   nFrameVertexs);

  for (iNode = 0; iNode < nNodes; iNode++) {
    int frameVertexId 
      = AutoMesh_SurfaceGrouper3D_frameVertexIdAtNode (Grouper, iNode);
    
    if (0 <= frameVertexId) {
      fprintf (fp, "%d\n", 
	       iNode);
    }
  }
}

static void WriteEdgeGroup 
(FILE *fp)
{
  int nNodes = AutoMesh_Node3D_nNodes (Node);
  int nEdges = AutoMesh_SurfaceEdge3D_nEdges (Edge);
  int nFrameGroups = AutoMesh_SurfaceGrouper3D_nFrameGroups (Grouper);
  int iNode;
  int iEdge;
  int iFrameGroup;
  int flag;
  
  /* NEED TO REDEFINE */
  /* nodes are not ordered */

  flag = AutoMesh_Node3D_makeRenumberedIdColumn (Node);
  assert(flag);

  fprintf (fp, "#edgeGroupInfo\n");
  
  fprintf (fp, "edgeGroupN %d\n", 
	   nFrameGroups);

  for (iFrameGroup = 0;
       iFrameGroup < nFrameGroups;
       iFrameGroup++) {

    AutoMesh_Node3D_initializeRenumbering (Node);
    for (iEdge = 0; iEdge < nEdges; iEdge++) {
      int frameGroupId 
	= AutoMesh_SurfaceGrouper3D_frameGroupIdAtEdge (Grouper, iEdge);

      if (frameGroupId == iFrameGroup) {
	int nodeId0, nodeId1; 

	AutoMesh_SurfaceEdge3D_getVertexNodeIdsAt (Edge, iEdge,
						   &nodeId0, &nodeId1);
	
	AutoMesh_Node3D_markNodeForRenumbering (Node, nodeId0);
	AutoMesh_Node3D_markNodeForRenumbering (Node, nodeId1);
      }
    }
    AutoMesh_Node3D_renumber (Node);

    fprintf (fp, "edgeGroup %d\n", 
	     AutoMesh_Node3D_renumberedCount (Node));

    for (iNode = 0; iNode < nNodes; iNode++) {
      if (AutoMesh_Node3D_isRenumbered (Node, iNode)) {
	fprintf (fp, "%d\n", 
		 iNode);
      }
    }
  }
}

static void WriteFaceGroup 
(FILE *fp)
{
  int nElements = AutoMesh_SurfaceElement3D_nElements (Element);
  int nGroups = AutoMesh_SurfaceGrouper3D_nGroups (Grouper);
  int iElement;
  int iGroup;
  int count;

  fprintf (fp, "#faceGroupInfo\n");
  
  fprintf (fp, "faceGroupN %d\n", 
	   nGroups);
  
  for (iGroup = 0;
       iGroup < nGroups;
       iGroup++) {

    count = 0;
    for (iElement = 0; iElement < nElements; iElement++) {
      int groupId = AutoMesh_SurfaceGrouper3D_groupIdAtElement
	(Grouper, iElement);
      
      if (groupId == iGroup) {
	count++;
      }
    }
    fprintf (fp, "faceGroup %d\n", 
	     count);

    for (iElement = 0; iElement < nElements; iElement++) {
      int groupId = AutoMesh_SurfaceGrouper3D_groupIdAtElement
	(Grouper, iElement);
      
      if (groupId == iGroup) {
	fprintf (fp, "%d\n", 
		 iElement);
      }
    }
  }
}

void WritePcgFile 
(const char *fileName)
{
  FILE *fp;

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  WriteVertex (fp);
  WriteEdgeGroup (fp);
  WriteFaceGroup (fp);

  fclose (fp);
}

int main (int argc, char *argv[])
{
#ifdef BINARY
  int binaryFlag = 1;
  int recordIdFlag = 0;
#else
  int binaryFlag = 0;
  int recordIdFlag = 1;
#endif
  int flag;
  
  if (argc != 8) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_export_ADVENTURE_pcg  node_file  element_file  \n"
       "        edge_vertex_file  \n"
       "        element_group_file \n"
       "        edge_frame_group_file \n"
       "        node_frame_vertex_file \n"
       "        pcg_file_OUT  \n");
    exit (1);
  }
  
  AutoRes_Initialize (0);
  AutoMesh_Initialize ();
  
  AutoRes_TableInOut_SetBinaryFlag (binaryFlag);
  AutoRes_TableInOut_SetRecordIdFlag (recordIdFlag);



  Node = AutoMesh_Node3D_New ();
  assert(Node != NULL);

  flag = AutoMesh_Node3D_makeCoordColumn (Node);
  assert(flag);

  AutoRes_TableInOut_AssociateReal1DColumn (Node->nodes_coord);
  flag = AutoRes_TableInOut_Read (argv[1]);
  assert(flag);



  Element = AutoMesh_SurfaceElement3D_New 
    (Node, AUTOMESH_POLYGON_TRI, 3);
  assert(Element != NULL);

  flag = AutoMesh_SurfaceElement3D_makeNodeIdsColumn (Element);
  assert(flag);

  AutoRes_TableInOut_AssociateInteger1DColumn (Element->elements_nodeIds);
  flag = AutoRes_TableInOut_Read (argv[2]);
  assert(flag);
  


  Edge = AutoMesh_SurfaceEdge3D_New (Element);
  assert(Element != NULL);

  flag = AutoMesh_SurfaceEdge3D_makeVertexNodeIdsColumn (Edge);
  assert(flag);

  AutoRes_TableInOut_AssociateInteger1DColumn (Edge->edges_vertexNodeIds);
  flag = AutoRes_TableInOut_Read (argv[3]);
  assert(flag);
  


  Grouper = AutoMesh_SurfaceGrouper3D_New (Edge);
  assert(Grouper != NULL);



  flag = AutoMesh_SurfaceGrouper3D_makeGroupIdAtElementColumn (Grouper);
  assert(flag);

  AutoRes_TableInOut_AssociateIntegerColumn (Grouper->elements_groupId);
  flag = AutoRes_TableInOut_Read (argv[4]);
  assert(flag);

  AutoMesh_SurfaceGrouper3D_setNGroups (Grouper);  



  flag = AutoMesh_SurfaceGrouper3D_makeFrameGroupIdAtEdgeColumn (Grouper);
  assert(flag);

  AutoRes_TableInOut_AssociateIntegerColumn (Grouper->edges_frameGroupId);
  flag = AutoRes_TableInOut_Read (argv[5]);
  assert(flag);

  AutoMesh_SurfaceGrouper3D_setNFrameGroups (Grouper);



  flag = AutoMesh_SurfaceGrouper3D_makeFrameVertexIdAtNodeColumn (Grouper);
  assert(flag);

  AutoRes_TableInOut_AssociateIntegerColumn (Grouper->nodes_frameVertexId);
  flag = AutoRes_TableInOut_Read (argv[6]);
  assert(flag);

  AutoMesh_SurfaceGrouper3D_setNFrameVertexs (Grouper);



  WritePcgFile (argv[7]);



  AutoMesh_SurfaceGrouper3D_delete (Grouper);
  AutoMesh_SurfaceEdge3D_delete (Edge);
  AutoMesh_SurfaceElement3D_delete (Element);
  AutoMesh_Node3D_delete (Node);

  return 0;
}







