/* automesh_base.c : Base Module */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"



/* Utility */

void AutoMesh_Error (const char *msg)
{
  fprintf (stderr, " AutoMesh ERROR : %s\n", msg);
  abort ();
}

void AutoMesh_Warning (const char *msg)
{
  fprintf (stderr, " AutoMesh WARNING : %s\n", msg);
}



/* General */

void AutoMesh_SortId3 
(int id0, int id1, int id2,
 int *minId_OUT, int *midId_OUT, int *maxId_OUT)
{
  assert(id0 != id1);
  assert(id0 != id2);
  assert(id1 != id2);

  if (id0 < id1 && id0 < id2) {
    *minId_OUT = id0;
  }
  if (id1 < id0 && id1 < id2) {
    *minId_OUT = id1;
  }
  if (id2 < id0 && id2 < id1) {
    *minId_OUT = id2;
  }

  if ((id1 < id0 && id0 < id2)
      || (id2 < id0 && id0 < id1)) {
    *midId_OUT = id0;
  }
  if ((id0 < id1 && id1 < id2)
      || (id2 < id1 && id1 < id0)) {
    *midId_OUT = id1;
  }
  if ((id0 < id2 && id2 < id1)
      || (id1 < id2 && id2 < id0)) {
    *midId_OUT = id2;
  }

  if (id1 < id0 && id2 < id0) {
    *maxId_OUT = id0;
  }
  if (id0 < id1 && id2 < id1) {
    *maxId_OUT = id1;
  }
  if (id0 < id2 && id1 < id2) {
    *maxId_OUT = id2;
  }
}

int AutoMesh_IdIsEqual2 
(int id0_0, int id0_1,
 int id1_0, int id1_1)
{
  assert(id0_0 != id0_1);
  assert(id1_0 != id1_1);

  return (id0_0 == id1_0 && id0_1 == id1_1)
    || (id0_0 == id1_1 && id0_1 == id1_0);
}

int AutoMesh_IdIsEqual3 
(int id0_0, int id0_1, int id0_2,
 int id1_0, int id1_1, int id1_2)
{
  int min0, mid0, max0;
  int min1, mid1, max1;

  assert(id0_0 != id0_1);
  assert(id0_0 != id0_2);
  assert(id0_1 != id0_2);
  assert(id1_0 != id1_1);
  assert(id1_0 != id1_2);
  assert(id1_1 != id1_2);

  AutoMesh_SortId3 (id0_0, id0_1, id0_2,
		    &min0, &mid0, &max0);
  AutoMesh_SortId3 (id1_0, id1_1, id1_2,
		    &min1, &mid1, &max1);
  return min0 == min1 && mid0 == mid1 && max0 == max1;
}



/* 2-D Geometry */

double AutoMesh_Length2D 
(double x, double y) 
{
  return sqrt (x * x + y * y);
}

double AutoMesh_Distance2D 
(double x0, double y0,
 double x1, double y1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0));
}

void AutoMesh_GetNormalizedVector2D 
(double x, double y,
 double *nx_OUT, double *ny_OUT)
{
  double length = sqrt (x * x + y * y);

  *nx_OUT = x / length;
  *ny_OUT = y / length;
}

double AutoMesh_ScalarProduct2D 
(double x0, double y0,
 double x1, double y1)
{
  return x0 * x1 + y0 * y1;
}

double AutoMesh_SignedTriArea2D
(double x0, double y0,
 double x1, double y1,
 double x2, double y2)
{
  return ((x1 - x0) * (y2 - y0) - (y1 - y0) * (x2 - x0)) / 2;
}




/* 3-D Geometry */

double AutoMesh_Length3D 
(double x, double y, double z) 
{
  return sqrt (x * x + y * y + z * z);
}

double AutoMesh_Distance3D 
(double x0, double y0, double z0,
 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}

void AutoMesh_GetNormalizedVector3D 
(double x, double y, double z,
 double *nx_OUT, double *ny_OUT, double *nz_OUT)
{
  double length = sqrt (x * x + y * y + z * z);

  *nx_OUT = x / length;
  *ny_OUT = y / length;
  *nz_OUT = z / length;
}

double AutoMesh_ScalarProduct3D 
(double x0, double y0, double z0,
 double x1, double y1, double z1)
{
  return x0 * x1 + y0 * y1 + z0 * z1;
}

void AutoMesh_GetVectorProduct3D 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  *x_OUT = y0 * z1 - z0 * y1;
  *y_OUT = z0 * x1 - x0 * z1;
  *z_OUT = x0 * y1 - y0 * x1;
}

void AutoMesh_GetAnyPerpendicularDirection3D
(double x, double y, double z,
 double *nx0_OUT, double *ny0_OUT, double *nz0_OUT,
 double *nx1_OUT, double *ny1_OUT, double *nz1_OUT)
{
  double length;

  AutoMesh_GetVectorProduct3D (x, y, z,
			       1, 0, 0,
			       nx0_OUT, ny0_OUT, nz0_OUT);
  length = sqrt (*nx0_OUT * *nx0_OUT
		 + *ny0_OUT * *ny0_OUT
		 + *nz0_OUT * *nz0_OUT);
  if (length == 0.0) {
    AutoMesh_GetVectorProduct3D (x, y, z,
				 0, 1, 0,
				 nx0_OUT, ny0_OUT, nz0_OUT);
  }
  AutoMesh_GetNormalizedVector3D (*nx0_OUT, *ny0_OUT, *nz0_OUT,
				  nx0_OUT, ny0_OUT, nz0_OUT);
  AutoMesh_GetVectorProduct3D (x, y, z,
			       *nx0_OUT, *ny0_OUT, *nz0_OUT,
			       nx1_OUT, ny1_OUT, nz1_OUT);
  AutoMesh_GetNormalizedVector3D (*nx1_OUT, *ny1_OUT, *nz1_OUT,
				  nx1_OUT, ny1_OUT, nz1_OUT);
}

void AutoMesh_GetProjectedPointOnPlane3D
(double a, double b, double c, double d, 
 double x, double y, double z,
 double *px_OUT, double *py_OUT, double *pz_OUT)
{
  double value = a * x + b * y + c * z + d;
  
  *px_OUT = x - a * value;
  *py_OUT = y - b * value;
  *pz_OUT = z - c * value;
}

double AutoMesh_SignedTetraVolume3D
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double x2, double y2, double z2,
 double x3, double y3, double z3)
{
  double x10 = x1 - x0;
  double y10 = y1 - y0;
  double z10 = z1 - z0;

  double x20 = x2 - x0;
  double y20 = y2 - y0;
  double z20 = z2 - z0;
  
  double x30 = x3 - x0;
  double y30 = y3 - y0;
  double z30 = z3 - z0;
  
  double d23 = y20 * z30 - z20 * y30;
  double d31 = y30 * z10 - z30 * y10;
  double d12 = y10 * z20 - z10 * y20;
  
  return (x10 * d23 + x20 * d31 + x30 * d12) / 6;
}


