/* automesh_frame2d.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"
#include "automesh_shape.h"
#include "automesh_frame2d.h"



#define TABLE_NAME_FRAMES "AutoMesh:frames"
#define COLUMN_NAME_END_POINT0 "AutoMesh:endPoint0"
#define COLUMN_NAME_END_POINT1 "AutoMesh:endPoint1"



/* Frame2D */

AUTORES_IMPLEMENT_HANDLE(AutoMesh_Frame2D)

AutoMesh_Frame2D *AutoMesh_Frame2D_New (void)
{
  AutoMesh_Frame2D *frame;

  frame = AutoMesh_Frame2D_NewImplementation ();

  frame->frames = AutoRes_Table_New (TABLE_NAME_FRAMES);
  assert(frame->frames != NULL);
  frame->frames_endPoint0 = NULL;
  frame->frames_endPoint1 = NULL;

  return frame;
}

void AutoMesh_Frame2D_delete 
(AutoMesh_Frame2D *self)
{
  AutoRes_Table_delete (self->frames);
  self->frames = NULL;
  self->frames_endPoint0 = NULL;
  self->frames_endPoint1 = NULL;

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */ 
AutoMesh_Frame2D_allocateFrames 
(AutoMesh_Frame2D *self,
 int nFrames)
{
  return AutoRes_Table_allocateRecords (self->frames, nFrames);
}

int /* boolean */ 
AutoMesh_Frame2D_reserveFrames 
(AutoMesh_Frame2D *self,
 int nFrames)
{
  return AutoRes_Table_reserveRecords (self->frames, nFrames);
}

int AutoMesh_Frame2D_addFrame 
(AutoMesh_Frame2D *self)
{
  return AutoRes_Table_addRecord (self->frames);
}

int /* boolean */
AutoMesh_Frame2D_makeEndPointsColumn 
(AutoMesh_Frame2D *self)
{
  int flag;

  self->frames_endPoint0 = AutoRes_Real1DColumn_New 
    (self->frames, COLUMN_NAME_END_POINT0, 2);
  assert(self->frames_endPoint0 != NULL);
  flag = AutoRes_Table_addReal1DColumn 
    (self->frames, self->frames_endPoint0);
  if (!flag) return 0;

  self->frames_endPoint1 = AutoRes_Real1DColumn_New 
    (self->frames, COLUMN_NAME_END_POINT1, 2);
  assert(self->frames_endPoint1 != NULL);
  flag = AutoRes_Table_addReal1DColumn 
    (self->frames, self->frames_endPoint1);
  if (!flag) return 0;

  return 1;
}

void AutoMesh_Frame2D_setEndPointsAt 
(AutoMesh_Frame2D *self,
 int frameId,
 double x0, double y0,
 double x1, double y1)
{
  AUTORES_R1D_SET(self->frames_endPoint0, frameId, 0, x0);
  AUTORES_R1D_SET(self->frames_endPoint0, frameId, 1, y0);
  
  AUTORES_R1D_SET(self->frames_endPoint1, frameId, 0, x1);
  AUTORES_R1D_SET(self->frames_endPoint1, frameId, 1, y1);
}

int AutoMesh_Frame2D_nFrames 
(AutoMesh_Frame2D *self)
{
  return AutoRes_Table_nRecords (self->frames);
}

void AutoMesh_Frame2D_getEndPointsAt 
(AutoMesh_Frame2D *self,
 int frameId,
 double *x0_OUT, double *y0_OUT,
 double *x1_OUT, double *y1_OUT)
{
  *x0_OUT = AUTORES_R1D(self->frames_endPoint0, frameId, 0);
  *y0_OUT = AUTORES_R1D(self->frames_endPoint0, frameId, 1);

  *x1_OUT = AUTORES_R1D(self->frames_endPoint1, frameId, 0);
  *y1_OUT = AUTORES_R1D(self->frames_endPoint1, frameId, 1);
}

void AutoMesh_Frame2D_getRange 
(AutoMesh_Frame2D *self,
 double *minX_OUT, double *minY_OUT,
 double *maxX_OUT, double *maxY_OUT)
{
  int nFrames = AutoRes_Table_nRecords (self->frames);
  int iFrame;
  double minX, minY;
  double maxX, maxY;
  
  assert(0 < nFrames);

  AutoMesh_Frame2D_getEndPointsAt (self, 0,
				   &minX, &minY,
				   &maxX, &maxY);
  
  for (iFrame = 0; iFrame < nFrames; iFrame++) {
    double x0, y0;
    double x1, y1;

    AutoMesh_Frame2D_getEndPointsAt (self, iFrame,
				     &x0, &y0,
				     &x1, &y1);

    minX = (x0 < minX) ? x0 : minX;
    minY = (y0 < minY) ? y0 : minY;

    minX = (x1 < minX) ? x1 : minX;
    minY = (y1 < minY) ? y1 : minY;

    maxX = (maxX < x0) ? x0 : maxX;
    maxY = (maxY < y0) ? y0 : maxY;

    maxX = (maxX < x1) ? x1 : maxX;
    maxY = (maxY < y1) ? y1 : maxY;
  }

  *minX_OUT = minX;
  *minY_OUT = minY;

  *maxX_OUT = maxX;
  *maxY_OUT = maxY;
}

