/* automesh_frame3d.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"
#include "automesh_shape.h"
#include "automesh_frame3d.h"



#define TABLE_NAME_FRAMES "AutoMesh:frames"
#define COLUMN_NAME_END_POINT0 "AutoMesh:endPoint0"
#define COLUMN_NAME_END_POINT1 "AutoMesh:endPoint1"



/* Frame3D */

AUTORES_IMPLEMENT_HANDLE(AutoMesh_Frame3D)

AutoMesh_Frame3D *AutoMesh_Frame3D_New (void)
{
  AutoMesh_Frame3D *frame;

  frame = AutoMesh_Frame3D_NewImplementation ();

  frame->frames = AutoRes_Table_New (TABLE_NAME_FRAMES);
  assert(frame->frames != NULL);
  frame->frames_endPoint0 = NULL;
  frame->frames_endPoint1 = NULL;

  return frame;
}

void AutoMesh_Frame3D_delete 
(AutoMesh_Frame3D *self)
{
  AutoRes_Table_delete (self->frames);
  self->frames = NULL;
  self->frames_endPoint0 = NULL;
  self->frames_endPoint1 = NULL;

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */ 
AutoMesh_Frame3D_allocateFrames 
(AutoMesh_Frame3D *self,
 int nFrames)
{
  return AutoRes_Table_allocateRecords (self->frames, nFrames);
}

int /* boolean */ 
AutoMesh_Frame3D_reserveFrames 
(AutoMesh_Frame3D *self,
 int nFrames)
{
  return AutoRes_Table_reserveRecords (self->frames, nFrames);
}

int AutoMesh_Frame3D_addFrame 
(AutoMesh_Frame3D *self)
{
  return AutoRes_Table_addRecord (self->frames);
}

int /* boolean */
AutoMesh_Frame3D_makeEndPointsColumn 
(AutoMesh_Frame3D *self)
{
  int flag;

  self->frames_endPoint0 = AutoRes_Real1DColumn_New 
    (self->frames, COLUMN_NAME_END_POINT0, 3);
  assert(self->frames_endPoint0 != NULL);
  flag = AutoRes_Table_addReal1DColumn 
    (self->frames, self->frames_endPoint0);
  if (!flag) return 0;

  self->frames_endPoint1 = AutoRes_Real1DColumn_New 
    (self->frames, COLUMN_NAME_END_POINT1, 3);
  assert(self->frames_endPoint1 != NULL);
  flag = AutoRes_Table_addReal1DColumn 
    (self->frames, self->frames_endPoint1);
  if (!flag) return 0;

  return 1;
}

void AutoMesh_Frame3D_setEndPointsAt 
(AutoMesh_Frame3D *self,
 int frameId,
 double x0, double y0, double z0,
 double x1, double y1, double z1)
{
  AUTORES_R1D_SET(self->frames_endPoint0, frameId, 0, x0);
  AUTORES_R1D_SET(self->frames_endPoint0, frameId, 1, y0);
  AUTORES_R1D_SET(self->frames_endPoint0, frameId, 2, z0);
  
  AUTORES_R1D_SET(self->frames_endPoint1, frameId, 0, x1);
  AUTORES_R1D_SET(self->frames_endPoint1, frameId, 1, y1);
  AUTORES_R1D_SET(self->frames_endPoint1, frameId, 2, z1);
}

int AutoMesh_Frame3D_nFrames 
(AutoMesh_Frame3D *self)
{
  return AutoRes_Table_nRecords (self->frames);
}

void AutoMesh_Frame3D_getEndPointsAt 
(AutoMesh_Frame3D *self,
 int frameId,
 double *x0_OUT, double *y0_OUT, double *z0_OUT,
 double *x1_OUT, double *y1_OUT, double *z1_OUT)
{
  *x0_OUT = AUTORES_R1D(self->frames_endPoint0, frameId, 0);
  *y0_OUT = AUTORES_R1D(self->frames_endPoint0, frameId, 1);
  *z0_OUT = AUTORES_R1D(self->frames_endPoint0, frameId, 2);

  *x1_OUT = AUTORES_R1D(self->frames_endPoint1, frameId, 0);
  *y1_OUT = AUTORES_R1D(self->frames_endPoint1, frameId, 1);
  *z1_OUT = AUTORES_R1D(self->frames_endPoint1, frameId, 2);
}

void AutoMesh_Frame3D_getRange 
(AutoMesh_Frame3D *self,
 double *minX_OUT, double *minY_OUT, double *minZ_OUT,
 double *maxX_OUT, double *maxY_OUT, double *maxZ_OUT)
{
  int nFrames = AutoRes_Table_nRecords (self->frames);
  int iFrame;
  double minX, minY, minZ;
  double maxX, maxY, maxZ;
  
  assert(0 < nFrames);

  AutoMesh_Frame3D_getEndPointsAt (self, 0,
				   &minX, &minY, &minZ,
				   &maxX, &maxY, &maxZ);
  
  for (iFrame = 0; iFrame < nFrames; iFrame++) {
    double x0, y0, z0;
    double x1, y1, z1;

    AutoMesh_Frame3D_getEndPointsAt (self, iFrame,
				     &x0, &y0, &z0,
				     &x1, &y1, &z1);

    minX = (x0 < minX) ? x0 : minX;
    minY = (y0 < minY) ? y0 : minY;
    minZ = (z0 < minZ) ? z0 : minZ;

    minX = (x1 < minX) ? x1 : minX;
    minY = (y1 < minY) ? y1 : minY;
    minZ = (z1 < minZ) ? z1 : minZ;

    maxX = (maxX < x0) ? x0 : maxX;
    maxY = (maxY < y0) ? y0 : maxY;
    maxZ = (maxZ < z0) ? z0 : maxZ;

    maxX = (maxX < x1) ? x1 : maxX;
    maxY = (maxY < y1) ? y1 : maxY;
    maxZ = (maxZ < z1) ? z1 : maxZ;
  }

  *minX_OUT = minX;
  *minY_OUT = minY;
  *minZ_OUT = minZ;

  *maxX_OUT = maxX;
  *maxY_OUT = maxY;
  *maxZ_OUT = maxZ;
}

