/* automesh_node2d.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"
#include "automesh_shape.h"
#include "automesh_node2d.h"



#define TABLE_NAME_NODES "AutoMesh:nodes"
#define COLUMN_NAME_COORD "AutoMesh:coord"
#define COLUMN_NAME_RENUMBERED_ID "AutoMesh:renumberedId"



/* Node2D */

AUTORES_IMPLEMENT_HANDLE(AutoMesh_Node2D)

AutoMesh_Node2D *AutoMesh_Node2D_New (void)
{
  AutoMesh_Node2D *node;

  node = AutoMesh_Node2D_NewImplementation ();

  node->nodes = AutoRes_Table_New (TABLE_NAME_NODES);
  assert(node->nodes != NULL);
  node->nodes_coord = NULL;
  node->nodes_renumberedId = NULL;

  return node;
}

void AutoMesh_Node2D_delete 
(AutoMesh_Node2D *self)
{
  AutoRes_Table_delete (self->nodes);
  self->nodes = NULL;
  self->nodes_coord = NULL;
  self->nodes_renumberedId = NULL;

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */ 
AutoMesh_Node2D_allocateNodes 
(AutoMesh_Node2D *self,
 int nNodes)
{
  return AutoRes_Table_allocateRecords (self->nodes, nNodes);
}

int /* boolean */ 
AutoMesh_Node2D_reserveNodes 
(AutoMesh_Node2D *self,
 int nNodes)
{
  return AutoRes_Table_reserveRecords (self->nodes, nNodes);
}

int AutoMesh_Node2D_addNode 
(AutoMesh_Node2D *self)
{
  return AutoRes_Table_addRecord (self->nodes);
}

int /* boolean */
AutoMesh_Node2D_makeCoordColumn 
(AutoMesh_Node2D *self)
{
  int flag;

  self->nodes_coord = AutoRes_Real1DColumn_New 
    (self->nodes, COLUMN_NAME_COORD, 2);
  assert(self->nodes_coord != NULL);
  flag = AutoRes_Table_addReal1DColumn 
    (self->nodes, self->nodes_coord);
  if (!flag) return 0;

  return 1;
}

void AutoMesh_Node2D_setCoordAt 
(AutoMesh_Node2D *self,
 int nodeId,
 double x, double y)
{
  AUTORES_R1D_SET(self->nodes_coord, nodeId, 0, x);
  AUTORES_R1D_SET(self->nodes_coord, nodeId, 1, y);
}

int /* boolean */
AutoMesh_Node2D_makeRenumberedIdColumn 
(AutoMesh_Node2D *self)
{
  int flag;

  self->nodes_renumberedId = AutoRes_IntegerColumn_New 
    (self->nodes, COLUMN_NAME_RENUMBERED_ID);
  assert(self->nodes_renumberedId != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->nodes, self->nodes_renumberedId);
  if (!flag) return 0;

  return 1;
}

void AutoMesh_Node2D_deleteRenumberedId 
(AutoMesh_Node2D *self)
{
  if (self->nodes_renumberedId != NULL) {
    AutoRes_Table_removeColumn 
      (self->nodes, 
       AutoRes_Table_columnIdOfColumn 
       (self->nodes, 
	self->nodes_renumberedId));
    self->nodes_renumberedId = NULL;
  }

  assert(self->nodes_renumberedId == NULL);
}

void AutoMesh_Node2D_initializeRenumbering 
(AutoMesh_Node2D *self)
{
  int nNodes = AutoRes_Table_nRecords (self->nodes);
  int iNode;

  for (iNode = 0; iNode < nNodes; iNode++) {
    AUTORES_I_SET(self->nodes_renumberedId, iNode, -1);
  }
}

void AutoMesh_Node2D_markNodeForRenumbering 
(AutoMesh_Node2D *self,
 int nodeId)
{
  AUTORES_I_SET(self->nodes_renumberedId, nodeId, 0);
}

void AutoMesh_Node2D_renumber 
(AutoMesh_Node2D *self)
{
  int nNodes = AutoRes_Table_nRecords (self->nodes);
  int iNode;
  int count;

  count = 0;
  for (iNode = 0; iNode < nNodes; iNode++) {
    if (AUTORES_I(self->nodes_renumberedId, iNode) == 0) {
      AUTORES_I_SET(self->nodes_renumberedId, iNode, count);
      count++;
    }
  }
}

int AutoMesh_Node2D_nNodes 
(AutoMesh_Node2D *self)
{
  return AutoRes_Table_nRecords (self->nodes);
}

void AutoMesh_Node2D_getCoordAt 
(AutoMesh_Node2D *self,
 int nodeId,
 double *x_OUT, double *y_OUT)
{
  *x_OUT = AUTORES_R1D(self->nodes_coord, nodeId, 0);
  *y_OUT = AUTORES_R1D(self->nodes_coord, nodeId, 1);
}

void AutoMesh_Node2D_getRange 
(AutoMesh_Node2D *self,
 double *minX_OUT, double *minY_OUT,
 double *maxX_OUT, double *maxY_OUT)
{
  int nNodes = AutoRes_Table_nRecords (self->nodes);
  int iNode;
  double minX, minY;
  double maxX, maxY;
  
  assert(0 < nNodes);

  AutoMesh_Node2D_getCoordAt (self, 0,
			      &minX, &minY);
  AutoMesh_Node2D_getCoordAt (self, 0,
			      &maxX, &maxY);

  for (iNode = 0; iNode < nNodes; iNode++) {
    double x, y;

    AutoMesh_Node2D_getCoordAt (self, iNode,
				&x, &y);

    minX = (x < minX) ? x : minX;
    minY = (y < minY) ? y : minY;

    maxX = (maxX < x) ? x : maxX;
    maxY = (maxY < y) ? y : maxY;
  }

  *minX_OUT = minX;
  *minY_OUT = minY;

  *maxX_OUT = maxX;
  *maxY_OUT = maxY;
}

double AutoMesh_Node2D_distance 
(AutoMesh_Node2D *self,
 int nodeId0, 
 int nodeId1)
{
  double x0, y0;
  double x1, y1;

  AutoMesh_Node2D_getCoordAt (self, nodeId0,
			      &x0, &y0);
  AutoMesh_Node2D_getCoordAt (self, nodeId1,
			      &x1, &y1);
  return AutoMesh_Distance2D 
    (x0, y0, 
     x1, y1);
}

int AutoMesh_Node2D_renumberedCount
(AutoMesh_Node2D *self)
{
  int nNodes = AutoRes_Table_nRecords (self->nodes);
  int iNode;
  int count;

  count = 0;
  for (iNode = 0; iNode < nNodes; iNode++) {
    if (0 <= AUTORES_I(self->nodes_renumberedId, iNode)) {
      count++;
    }
  }
  
  return count;
}

int AutoMesh_Node2D_isRenumbered 
(AutoMesh_Node2D *self,
 int nodeId)
{
  return 0 <= AUTORES_I(self->nodes_renumberedId, nodeId);
}

int AutoMesh_Node2D_renumberedId 
(AutoMesh_Node2D *self,
 int nodeId)
{
  return AUTORES_I(self->nodes_renumberedId, nodeId);
}



