/* automesh_node_vector3d.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"
#include "automesh_shape.h"
#include "automesh_node3d.h"
#include "automesh_node_vector3d.h"



#define COLUMN_NAME_VALUE "AutoMesh:value"



/* NodeVector3D */

AUTORES_IMPLEMENT_HANDLE(AutoMesh_NodeVector3D)

AutoMesh_NodeVector3D *AutoMesh_NodeVector3D_New 
(AutoMesh_Node3D *node)
{
  AutoMesh_NodeVector3D *vector;

  vector = AutoMesh_NodeVector3D_NewImplementation ();

  vector->node = node;
  vector->nodes_value = NULL;

  return vector;
}

void AutoMesh_NodeVector3D_delete 
(AutoMesh_NodeVector3D *self)
{
  if (self->nodes_value != NULL) {
    AutoRes_Table_removeColumn 
      (self->node->nodes, 
       AutoRes_Table_columnIdOfColumn (self->node->nodes, 
				       self->nodes_value));
    self->nodes_value = NULL;
  }

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoMesh_NodeVector3D_makeValueColumn 
(AutoMesh_NodeVector3D *self)
{
  AutoRes_Table *nodes = self->node->nodes;
  int flag;

  self->nodes_value = AutoRes_Real1DColumn_New 
    (nodes, COLUMN_NAME_VALUE, 3);
  assert(self->nodes_value != NULL);
  flag = AutoRes_Table_addReal1DColumn 
    (nodes, self->nodes_value);
  if (!flag) return 0;

  return 1;
}

void AutoMesh_NodeVector3D_setValueAt 
(AutoMesh_NodeVector3D *self,
 int nodeId,
 double valueX, double valueY, double valueZ)
{
  AUTORES_R1D_SET(self->nodes_value, nodeId, 0, valueX);
  AUTORES_R1D_SET(self->nodes_value, nodeId, 1, valueY);
  AUTORES_R1D_SET(self->nodes_value, nodeId, 2, valueZ);
}

void AutoMesh_NodeVector3D_getValueAt 
(AutoMesh_NodeVector3D *self,
 int nodeId,
 double *valueX_OUT, double *valueY_OUT, double *valueZ_OUT)
{
  *valueX_OUT = AUTORES_R1D(self->nodes_value, nodeId, 0);
  *valueY_OUT = AUTORES_R1D(self->nodes_value, nodeId, 1);
  *valueZ_OUT = AUTORES_R1D(self->nodes_value, nodeId, 2);
}

void AutoMesh_NodeVector3D_getRange 
(AutoMesh_NodeVector3D *self,
 double *minValueX_OUT, double *minValueY_OUT, double *minValueZ_OUT,
 double *maxValueX_OUT, double *maxValueY_OUT, double *maxValueZ_OUT)
{
  int nNodes = AutoRes_Table_nRecords (self->node->nodes);
  int iNode;
  double minX, minY, minZ;
  double maxX, maxY, maxZ;
  
  assert(0 < nNodes);
  
  AutoMesh_NodeVector3D_getValueAt (self, 0,
				    &minX, &minY, &minZ);
  AutoMesh_NodeVector3D_getValueAt (self, 0,
				    &maxX, &maxY, &maxZ);

  for (iNode = 0; iNode < nNodes; iNode++) {
    double x, y, z;

    AutoMesh_NodeVector3D_getValueAt (self, iNode,
				      &x, &y, &z);

    minX = (x < minX) ? x : minX;
    minY = (y < minY) ? y : minY;
    minZ = (z < minZ) ? z : minZ;

    maxX = (maxX < x) ? x : maxX;
    maxY = (maxY < y) ? y : maxY;
    maxZ = (maxZ < z) ? z : maxZ;
  }

  *minValueX_OUT = minX;
  *minValueY_OUT = minY;
  *minValueZ_OUT = minZ;

  *maxValueX_OUT = maxX;
  *maxValueY_OUT = maxY;
  *maxValueZ_OUT = maxZ;
}

double AutoMesh_NodeVector3D_maxLength 
(AutoMesh_NodeVector3D *self)
{
  double maxLength;
  int nNodes = AutoRes_Table_nRecords (self->node->nodes);
  int iNode;
  
  assert(0 < nNodes);
  
  maxLength = 0.0;
  for (iNode = 0; iNode < nNodes; iNode++) {
    double x, y, z;
    double length;

    AutoMesh_NodeVector3D_getValueAt (self, iNode,
				      &x, &y, &z);
    length = AutoMesh_Length3D (x, y, z);

    maxLength = (maxLength < length) ? length : maxLength;
  }

  return maxLength;
}
