/* automesh_shape.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"
#include "automesh_shape.h"



/* Triangle */

const int AutoMesh_Tri_EdgeVertexMap[3][2] = {
  { 1, 2 },
  { 2, 0 },
  { 0, 1 }
};



/* Quadrangle */

const int AutoMesh_Quad_EdgeVertexMap[4][2] = {
  { 0, 1 },
  { 1, 2 },
  { 2, 3 },
  { 3, 0 }
};



/* Polygon */

int AutoMesh_Polygon_NVertexs 
(AutoMesh_PolygonType type)
{
  switch (type) {
  case AUTOMESH_POLYGON_TRI:
    return AUTOMESH_TRI_N_VERTEXS;
  case AUTOMESH_POLYGON_QUAD:
    return AUTOMESH_QUAD_N_VERTEXS;
  default:
    assert(0);
    break;
  }
  return 0;
}

int AutoMesh_Polygon_NEdges 
(AutoMesh_PolygonType type)
{
  switch (type) {
  case AUTOMESH_POLYGON_TRI:
    return AUTOMESH_TRI_N_EDGES;
  case AUTOMESH_POLYGON_QUAD:
    return AUTOMESH_QUAD_N_EDGES;
  default:
    assert(0);
    break;
  }
  return 0;
}

int AutoMesh_Polygon_EdgeVertex 
(AutoMesh_PolygonType type, int edgeNum, int vertexNum)
{
  assert(0 <= edgeNum);
  assert(edgeNum < AutoMesh_Polygon_NEdges (type));
  assert(0 <= vertexNum);
  assert(vertexNum < 2);
  
  switch (type) {
  case AUTOMESH_POLYGON_TRI:
    return AutoMesh_Tri_EdgeVertexMap[edgeNum][vertexNum];
  case AUTOMESH_POLYGON_QUAD:
    return AutoMesh_Quad_EdgeVertexMap[edgeNum][vertexNum];
  default:
    assert(0);
    break;
  }
  return 0;
}



/* Tetrahedron */

const int AutoMesh_Tetra_EdgeVertexMap[6][2] = {
  { 0, 1 },
  { 0, 2 },
  { 0, 3 },
  { 1, 2 },
  { 2, 3 },
  { 3, 1 }
};

const int AutoMesh_Tetra_FaceVertexMap[4][3] = {
  { 1, 2, 3 },
  { 0, 3, 2 },
  { 0, 1, 3 },
  { 0, 2, 1 }
};

static const int AutoMesh_Tetra_NIsosurfaceEdges[8] = {
  0, 3, 3, 4, 3, 4, 4, 3
};

static const int AutoMesh_Tetra_IsosurfaceEdgeNums[8][4] = {
  { -1, -1, -1, -1 },
  { 0, 1, 2, -1 },
  { 0, 3, 5, -1 },
  { 3, 1, 2, 5 },
  { 3, 1, 4, -1 },
  { 0, 3, 4, 2 },
  { 0, 5, 4, 1 },
  { 2, 5, 4, -1 }
};

int AutoMesh_Tetra_GetIsosurface 
(double vertexValues[4], 
 double levelValue,
 int *nEdges_OUT,
 int edgeNums_OUT[4], double ratios_OUT[4])
{
  unsigned int mask;
  int iEdge;
  
  mask = 0;
  if (levelValue < vertexValues[0]) mask |= 1;
  if (levelValue < vertexValues[1]) mask |= 2;
  if (levelValue < vertexValues[2]) mask |= 4;
  if (levelValue < vertexValues[3]) mask |= 8;
  if ((mask & 8) != 0) mask = (~mask) & 15;
  assert(0 <= mask);
  assert(mask < 8);
  
  *nEdges_OUT = AutoMesh_Tetra_NIsosurfaceEdges[mask];
  if (*nEdges_OUT == 0) return 0;
  
  assert(*nEdges_OUT == 3 || *nEdges_OUT == 4);

  for (iEdge = 0; iEdge < *nEdges_OUT; iEdge++) {
    int edgeNum = AutoMesh_Tetra_IsosurfaceEdgeNums[mask][iEdge];

    int vertexNum0 = AutoMesh_Tetra_EdgeVertexMap[edgeNum][0];
    int vertexNum1 = AutoMesh_Tetra_EdgeVertexMap[edgeNum][1];
    
    double vertexValue0 = vertexValues[vertexNum0];
    double vertexValue1 = vertexValues[vertexNum1];

    double ratio = (levelValue - vertexValue0) / (vertexValue1 - vertexValue0);
    
    edgeNums_OUT[iEdge] = edgeNum; 
    ratios_OUT[iEdge] = ratio;
  }    

  return 1;
}



/* Hexahedron */

/* NEED TO REDEFINE */
const int AutoMesh_Hexa_EdgeVertexMap[12][2] = {
};

/* NEED TO REDEFINE */
const int AutoMesh_Hexa_FaceVertexMap[6][4] = {
};



/* Prism */

/* NEED TO REDEFINE */
const int AutoMesh_Prism_EdgeVertexMap[9][2] = {
};

/* NEED TO REDEFINE */
const int AutoMesh_Prism_NFaceVertexs[5] = {
};

/* NEED TO REDEFINE */
const int AutoMesh_Prism_FaceVertexMap[5][4] = {
};



/* Polyhedron */

int AutoMesh_Polyhedron_NVertexs 
(AutoMesh_PolyhedronType type)
{
  switch (type) {
  case AUTOMESH_POLYHEDRON_TETRA:
    return AUTOMESH_TETRA_N_VERTEXS;
  case AUTOMESH_POLYHEDRON_HEXA:
    return AUTOMESH_HEXA_N_VERTEXS;
  case AUTOMESH_POLYHEDRON_PRISM:
    return AUTOMESH_PRISM_N_VERTEXS;
  default:
    assert(0);
    break;
  }
  return 0;
}

int AutoMesh_Polyhedron_NEdges 
(AutoMesh_PolyhedronType type)
{
  switch (type) {
  case AUTOMESH_POLYHEDRON_TETRA:
    return AUTOMESH_TETRA_N_EDGES;
  case AUTOMESH_POLYHEDRON_HEXA:
    return AUTOMESH_HEXA_N_EDGES;
  case AUTOMESH_POLYHEDRON_PRISM:
    return AUTOMESH_PRISM_N_EDGES;
  default:
    assert(0);
    break;
  }
  return 0;
}

int AutoMesh_Polyhedron_NFaces 
(AutoMesh_PolyhedronType type)
{
  switch (type) {
  case AUTOMESH_POLYHEDRON_TETRA:
    return AUTOMESH_TETRA_N_FACES;
  case AUTOMESH_POLYHEDRON_HEXA:
    return AUTOMESH_HEXA_N_FACES;
  case AUTOMESH_POLYHEDRON_PRISM:
    return AUTOMESH_PRISM_N_FACES;
  default:
    assert(0);
    break;
  }
  return 0;
}

int AutoMesh_Polyhedron_MaxFaceVertexs 
(AutoMesh_PolyhedronType type)
{
  switch (type) {
  case AUTOMESH_POLYHEDRON_TETRA:
    return AUTOMESH_TETRA_N_FACE_VERTEXS;
  case AUTOMESH_POLYHEDRON_HEXA:
    return AUTOMESH_HEXA_N_FACE_VERTEXS;
  case AUTOMESH_POLYHEDRON_PRISM:
    return AUTOMESH_PRISM_MAX_FACE_VERTEXS;
  default:
    assert(0);
    break;
  }
  return 0;
}

int AutoMesh_Polyhedron_EdgeVertex 
(AutoMesh_PolyhedronType type, int edgeNum, int vertexNum)
{
  assert(0 <= edgeNum);
  assert(edgeNum < AutoMesh_Polyhedron_NEdges (type));
  assert(0 <= vertexNum);
  assert(vertexNum < 2);

  switch (type) {
  case AUTOMESH_POLYHEDRON_TETRA:
    return AutoMesh_Tetra_EdgeVertexMap[edgeNum][vertexNum];
  case AUTOMESH_POLYHEDRON_HEXA:
    return AutoMesh_Hexa_EdgeVertexMap[edgeNum][vertexNum];
  case AUTOMESH_POLYHEDRON_PRISM:
    return AutoMesh_Prism_EdgeVertexMap[edgeNum][vertexNum];
  default:
    assert(0);
    break;
  }
  return 0;
}

int AutoMesh_Polyhedron_NFaceVertexs 
(AutoMesh_PolyhedronType type, int faceNum)
{
  assert(0 <= faceNum);
  assert(faceNum < AutoMesh_Polyhedron_NFaces (type));

  switch (type) {
  case AUTOMESH_POLYHEDRON_TETRA:
    return AUTOMESH_TETRA_N_FACE_VERTEXS;
  case AUTOMESH_POLYHEDRON_HEXA:
    return AUTOMESH_HEXA_N_FACE_VERTEXS;
  case AUTOMESH_POLYHEDRON_PRISM:
    return AutoMesh_Prism_NFaceVertexs[faceNum];
  default:
    assert(0);
    break;
  }
  return 0;
}

int AutoMesh_Polyhedron_FaceVertex 
(AutoMesh_PolyhedronType type, int faceNum, int vertexNum)
{
  assert(0 <= faceNum);
  assert(faceNum < AutoMesh_Polyhedron_NFaces (type));
  assert(0 <= vertexNum);
  assert(vertexNum < AutoMesh_Polyhedron_NFaceVertexs (type, faceNum));

  switch (type) {
  case AUTOMESH_POLYHEDRON_TETRA:
    return AutoMesh_Tetra_FaceVertexMap[faceNum][vertexNum];
  case AUTOMESH_POLYHEDRON_HEXA:
    return AutoMesh_Hexa_FaceVertexMap[faceNum][vertexNum];
  case AUTOMESH_POLYHEDRON_PRISM:
    return AutoMesh_Prism_FaceVertexMap[faceNum][vertexNum];
  default:
    assert(0);
    break;
  }
  return 0;
}




