/* automesh_surface_element3d.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"
#include "automesh_shape.h"
#include "automesh_node3d.h"
#include "automesh_surface_element3d.h"



#define TABLE_NAME_ELEMENTS "AutoMesh:elements"
#define COLUMN_NAME_NODE_IDS "AutoMesh:nodeIds"
#define COLUMN_NAME_NEIGHBOR_IDS "AutoMesh:neighborIds"



/* SurfaceElement3D */

AUTORES_IMPLEMENT_HANDLE(AutoMesh_SurfaceElement3D)

AutoMesh_SurfaceElement3D *AutoMesh_SurfaceElement3D_New 
(AutoMesh_Node3D *node,
 AutoMesh_PolygonType polygonType,
 int nElementNodes)
{
  AutoMesh_SurfaceElement3D *element;

  element = AutoMesh_SurfaceElement3D_NewImplementation ();

  element->polygonType = polygonType;
  element->nElementNodes = nElementNodes;
  element->nElementVertexs = AutoMesh_Polygon_NVertexs (polygonType);
  element->nElementEdges = AutoMesh_Polygon_NEdges (polygonType);

  element->node = node;

  element->elements = AutoRes_Table_New (TABLE_NAME_ELEMENTS);
  assert(element->elements != NULL);
  element->elements_nodeIds = NULL;
  element->elements_neighborIds = NULL;

  return element;
}

void AutoMesh_SurfaceElement3D_delete 
(AutoMesh_SurfaceElement3D *self)
{
  AutoRes_Table_delete (self->elements);
  self->elements = NULL;
  self->elements_nodeIds = NULL;
  self->elements_neighborIds = NULL;

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */ 
AutoMesh_SurfaceElement3D_allocateElements 
(AutoMesh_SurfaceElement3D *self,
 int nElements)
{
  return AutoRes_Table_allocateRecords (self->elements, nElements);
}

int /* boolean */ 
AutoMesh_SurfaceElement3D_reserveElements 
(AutoMesh_SurfaceElement3D *self,
 int nElements)
{
  return AutoRes_Table_reserveRecords (self->elements, nElements);
}

int AutoMesh_SurfaceElement3D_addElement 
(AutoMesh_SurfaceElement3D *self)
{
  return AutoRes_Table_addRecord (self->elements);
}

int /* boolean */
AutoMesh_SurfaceElement3D_makeNodeIdsColumn 
(AutoMesh_SurfaceElement3D *self)
{
  int flag;

  self->elements_nodeIds = AutoRes_Integer1DColumn_New 
    (self->elements, COLUMN_NAME_NODE_IDS, self->nElementNodes);
  assert(self->elements_nodeIds != NULL);
  flag = AutoRes_Table_addInteger1DColumn 
    (self->elements, self->elements_nodeIds);
  if (!flag) return 0;

  return 1;
}

void AutoMesh_SurfaceElement3D_setNodeIdAt 
(AutoMesh_SurfaceElement3D *self,
 int elementId,
 int elementNodeNum,
 int nodeId)
{
  AUTORES_I1D_SET(self->elements_nodeIds, elementId, elementNodeNum, 
		  nodeId);
}

int /* boolean */
AutoMesh_SurfaceElement3D_makeNeighborIdsColumn 
(AutoMesh_SurfaceElement3D *self)
{
  int flag;

  self->elements_neighborIds = AutoRes_Integer1DColumn_New 
    (self->elements, COLUMN_NAME_NEIGHBOR_IDS, self->nElementEdges);
  assert(self->elements_neighborIds != NULL);
  flag = AutoRes_Table_addInteger1DColumn 
    (self->elements, self->elements_neighborIds);
  if (!flag) return 0;

  return 1;
}

void AutoMesh_SurfaceElement3D_setNeighborIdAt 
(AutoMesh_SurfaceElement3D *self,
 int elementId,
 int elementEdgeNum,
 int neighborId)
{
  AUTORES_I1D_SET(self->elements_neighborIds, elementId, elementEdgeNum, 
		  neighborId);
}

void AutoMesh_SurfaceElement3D_markElementForNodeRenumbering 
(AutoMesh_SurfaceElement3D *self,
 int elementId)
{
  int iElementNode;
  
  for (iElementNode = 0; 
       iElementNode < self->nElementNodes; 
       iElementNode++) {
    int nodeId = AutoMesh_SurfaceElement3D_nodeIdAt 
      (self, elementId, iElementNode);
    
    AutoMesh_Node3D_markNodeForRenumbering (self->node, nodeId);
  }
}

int AutoMesh_SurfaceElement3D_nElements 
(AutoMesh_SurfaceElement3D *self)
{
  return AutoRes_Table_nRecords (self->elements);
}

int AutoMesh_SurfaceElement3D_nodeIdAt 
(AutoMesh_SurfaceElement3D *self,
 int elementId,
 int elementNodeNum)
{
  return AUTORES_I1D(self->elements_nodeIds, elementId, elementNodeNum);
}

int AutoMesh_SurfaceElement3D_neighborIdAt 
(AutoMesh_SurfaceElement3D *self,
 int elementId,
 int elementEdgeNum)
{
  return AUTORES_I1D(self->elements_neighborIds, elementId, elementEdgeNum);
}

void AutoMesh_SurfaceElement3D_getRangeAt
(AutoMesh_SurfaceElement3D *self,
 int elementId, 
 double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int nodeId0 = AutoMesh_SurfaceElement3D_nodeIdAt (self, elementId, 0);
  int iElementNode;
  
  AutoMesh_Node3D_getCoordAt (self->node, nodeId0,
			      minRangeX_OUT, minRangeY_OUT, minRangeZ_OUT);
  AutoMesh_Node3D_getCoordAt (self->node, nodeId0,
			      maxRangeX_OUT, maxRangeY_OUT, maxRangeZ_OUT);
  for (iElementNode = 0; 
       iElementNode < self->nElementNodes; 
       iElementNode++) {
    int nodeId = AutoMesh_SurfaceElement3D_nodeIdAt 
      (self, elementId, iElementNode);
    double x, y, z;
    
    AutoMesh_Node3D_getCoordAt (self->node, nodeId,
				&x, &y, &z);
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;
    
    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
    if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
  }
}

double AutoMesh_SurfaceElement3D_averageEdgeLengthAt
(AutoMesh_SurfaceElement3D *self,
 int elementId)
{
  int iElementEdge;
  double average;

  average = 0.0;
  for (iElementEdge = 0; 
       iElementEdge < self->nElementEdges; 
       iElementEdge++) {
    int elementNodeNum0 = AutoMesh_Polygon_EdgeVertex
      (self->polygonType, iElementEdge, 0);
    int elementNodeNum1 = AutoMesh_Polygon_EdgeVertex
      (self->polygonType, iElementEdge, 1);
    
    int nodeId0 = AUTORES_I1D(self->elements_nodeIds, 
			      elementId, elementNodeNum0);
    int nodeId1 = AUTORES_I1D(self->elements_nodeIds, 
			      elementId, elementNodeNum1);

    double x0, y0, z0;
    double x1, y1, z1;
    
    AutoMesh_Node3D_getCoordAt (self->node, nodeId0,
				&x0, &y0, &z0);
    AutoMesh_Node3D_getCoordAt (self->node, nodeId1,
				&x1, &y1, &z1);
    
    average += AutoMesh_Distance3D (x0, y0, z0, 
				    x1, y1, z1);
  }
  average /= self->nElementEdges;
  return average;
}
 
void AutoMesh_SurfaceElement3D_getNormalAt 
(AutoMesh_SurfaceElement3D *self,
 int elementId,
 double *nx_OUT, double *ny_OUT, double *nz_OUT)
{
  switch (self->polygonType) {
  case AUTOMESH_POLYGON_TRI:
    {
      int nodeId0, nodeId1, nodeId2;
      double x0, y0, z0;
      double x1, y1, z1;
      double x2, y2, z2;

      nodeId0 = AutoMesh_SurfaceElement3D_nodeIdAt (self, elementId, 0);
      nodeId1 = AutoMesh_SurfaceElement3D_nodeIdAt (self, elementId, 1);
      nodeId2 = AutoMesh_SurfaceElement3D_nodeIdAt (self, elementId, 2);
      
      AutoMesh_Node3D_getCoordAt (self->node, nodeId0,
				  &x0, &y0, &z0);
      AutoMesh_Node3D_getCoordAt (self->node, nodeId1,
				  &x1, &y1, &z1);
      AutoMesh_Node3D_getCoordAt (self->node, nodeId2,
				  &x2, &y2, &z2);
      
      AutoMesh_GetVectorProduct3D (x0 - x1, y0 - y1, z0 - z1,
				   x0 - x2, y0 - y2, z0 - z2,
				   nx_OUT, ny_OUT, nz_OUT);
      AutoMesh_GetNormalizedVector3D (*nx_OUT, *ny_OUT, *nz_OUT,
				      nx_OUT, ny_OUT, nz_OUT);
    }
    break;
  case AUTOMESH_POLYGON_QUAD:
    assert(0);
    break;
  default:
    assert(0);
    break;
  }
}



