/* automesh_surface_extractor3d.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"
#include "automesh_shape.h"
#include "automesh_node3d.h"
#include "automesh_surface_element3d.h"
#include "automesh_surface_extractor3d.h"



#define COLUMN_NAME_SRC_NODE_ID "AutoMesh:srcNodeId"
#define COLUMN_NAME_SRC_ELEMENT_ID "AutoMesh:srcElementId"
#define COLUMN_NAME_DEST_NODE_ID "AutoMesh:destNodeId"
#define COLUMN_NAME_DEST_ELEMENT_ID "AutoMesh:destElementId"



/* SurfaceExtractor3D */

AUTORES_IMPLEMENT_HANDLE(AutoMesh_SurfaceExtractor3D)

AutoMesh_SurfaceExtractor3D *AutoMesh_SurfaceExtractor3D_New 
(AutoMesh_SurfaceElement3D* element,
 AutoRes_Selection *elementSelection)
{
  AutoMesh_SurfaceExtractor3D *extractor;
  
  extractor = AutoMesh_SurfaceExtractor3D_NewImplementation ();
  
  extractor->elementSelection = elementSelection;

  extractor->srcNode = element->node;
  extractor->srcNodes_destNodeId = NULL;
  
  extractor->srcElement = element;
  extractor->srcElements_destElementId = NULL;

  extractor->destNode = AutoMesh_Node3D_New ();
  assert(extractor->destNode != NULL);
  extractor->destNodes_srcNodeId = NULL;

  extractor->destElement = AutoMesh_SurfaceElement3D_New 
    (extractor->destNode, 
     element->polygonType,
     element->nElementNodes);
  assert(extractor->destElement != NULL);
  extractor->destElements_srcElementId = NULL;

  return extractor;
}

void AutoMesh_SurfaceExtractor3D_delete 
(AutoMesh_SurfaceExtractor3D *self)
{
  if (self->srcNodes_destNodeId != NULL) {
    AutoRes_Table_removeColumn 
      (self->srcNode->nodes, 
       AutoRes_Table_columnIdOfColumn 
       (self->srcNode->nodes, 
	self->srcNodes_destNodeId));
    self->srcNodes_destNodeId = NULL;
  }

  if (self->srcElements_destElementId != NULL) {
    AutoRes_Table_removeColumn 
      (self->srcElement->elements, 
       AutoRes_Table_columnIdOfColumn 
       (self->srcElement->elements, 
	self->srcElements_destElementId));
    self->srcElements_destElementId = NULL;
  }

  if (self->destNodes_srcNodeId != NULL) {
    AutoRes_Table_removeColumn 
      (self->destNode->nodes, 
       AutoRes_Table_columnIdOfColumn 
       (self->destNode->nodes, 
	self->destNodes_srcNodeId));
    self->destNodes_srcNodeId = NULL;
  }

  if (self->destElements_srcElementId != NULL) {
    AutoRes_Table_removeColumn 
      (self->destElement->elements, 
       AutoRes_Table_columnIdOfColumn 
       (self->destElement->elements, 
	self->destElements_srcElementId));
    self->destElements_srcElementId = NULL;
  }

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

static int /* boolean */
AllocateDestElement
(AutoMesh_SurfaceExtractor3D *self)
{
  int nDestElements = AutoRes_Selection_nEntrys (self->elementSelection);
  int flag;
  
  assert(AutoMesh_SurfaceElement3D_nElements (self->destElement) == 0);
  
  flag = AutoMesh_SurfaceElement3D_allocateElements 
    (self->destElement, nDestElements);
  if (!flag) return 0;

  assert(0 < AutoMesh_SurfaceElement3D_nElements (self->destElement));

  return 1;
}

static int /* boolean */
MakeSrcElementId 
(AutoMesh_SurfaceExtractor3D *self)
{
  int nDestElements = AutoRes_Selection_nEntrys (self->elementSelection);
  int iDestElement;
  int flag;

  assert(0 < AutoMesh_SurfaceElement3D_nElements (self->destElement));
  assert(self->destElements_srcElementId == NULL);

  self->destElements_srcElementId = AutoRes_IntegerColumn_New
    (self->destElement->elements, COLUMN_NAME_SRC_ELEMENT_ID);
  assert(self->destElements_srcElementId != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->destElement->elements, 
     self->destElements_srcElementId);
  if (!flag) return 0;

  for (iDestElement = 0; iDestElement < nDestElements; iDestElement++) {
    int srcElementId = AutoRes_Selection_objectId 
      (self->elementSelection, iDestElement);

    AUTORES_I_SET(self->destElements_srcElementId, iDestElement, 
		  srcElementId);
  }

  assert(self->destElements_srcElementId != NULL);

  return 1;
}

static int /* boolean */
MakeDestElementId 
(AutoMesh_SurfaceExtractor3D *self)
{
  int nSrcElements = AutoMesh_SurfaceElement3D_nElements (self->srcElement);
  int nDestElements = AutoMesh_SurfaceElement3D_nElements (self->destElement);
  int iSrcElement;
  int iDestElement;
  int flag;

  assert(0 < nDestElements);
  assert(self->srcElements_destElementId == NULL);

  self->srcElements_destElementId = AutoRes_IntegerColumn_New
    (self->srcElement->elements, COLUMN_NAME_DEST_ELEMENT_ID);
  assert(self->srcElements_destElementId != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->srcElement->elements, 
     self->srcElements_destElementId);
  if (!flag) return 0;

  for (iSrcElement = 0; iSrcElement < nSrcElements; iSrcElement++) {
    AUTORES_I_SET(self->srcElements_destElementId, iSrcElement, -1);
  }

  for (iDestElement = 0; iDestElement < nDestElements; iDestElement++) {
    int srcElementId = AutoRes_Selection_objectId 
      (self->elementSelection, iDestElement);
    
    AUTORES_I_SET(self->srcElements_destElementId, srcElementId, 
		  iDestElement);
  }

  assert(self->srcElements_destElementId != NULL);

  return 1;
}

static int /* boolean */
MakeDestNodeId 
(AutoMesh_SurfaceExtractor3D *self)
{
  int nSrcNodes = AutoMesh_Node3D_nNodes (self->srcNode);
  int nDestElements = AutoMesh_SurfaceElement3D_nElements (self->destElement);
  int nElementNodes = self->srcElement->nElementNodes;
  int iSrcNode;
  int iDestElement;
  int iElementNode;
  int destNodeId;
  int flag;

  assert(0 < nDestElements);
  assert(self->srcNodes_destNodeId == NULL);

  self->srcNodes_destNodeId = AutoRes_IntegerColumn_New
    (self->srcNode->nodes, COLUMN_NAME_DEST_NODE_ID);
  assert(self->srcNodes_destNodeId != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->srcNode->nodes, 
     self->srcNodes_destNodeId);
  if (!flag) return 0;

  for (iSrcNode = 0; iSrcNode < nSrcNodes; iSrcNode++) {
    AUTORES_I_SET(self->srcNodes_destNodeId, iSrcNode, -1);
  }

  for (iDestElement = 0; iDestElement < nDestElements; iDestElement++) {
    int srcElementId = AutoRes_Selection_objectId 
      (self->elementSelection, iDestElement);
    
    for (iElementNode = 0; 
	 iElementNode < nElementNodes; 
	 iElementNode++) {
      int srcNodeId = AutoMesh_SurfaceElement3D_nodeIdAt
	(self->srcElement, srcElementId, iElementNode);

      AUTORES_I_SET(self->srcNodes_destNodeId, srcNodeId, 0);
    }
  }

  destNodeId = 0;
  for (iSrcNode = 0; iSrcNode < nSrcNodes; iSrcNode++) {
    if (AUTORES_I(self->srcNodes_destNodeId, iSrcNode) == 0) {
      AUTORES_I_SET(self->srcNodes_destNodeId, iSrcNode, 
		    destNodeId);
      destNodeId++;
    }
  }

  assert(self->srcNodes_destNodeId != NULL);

  return 1;
}

static int /* boolean */
AllocateDestNode
(AutoMesh_SurfaceExtractor3D *self)
{
  int nSrcNodes = AutoMesh_Node3D_nNodes (self->srcNode);
  int nDestNodes;
  int iSrcNode;
  int flag;

  assert(self->srcNodes_destNodeId != NULL);
  assert(AutoMesh_Node3D_nNodes (self->destNode) == 0);

  nDestNodes = 0;
  for (iSrcNode = 0; iSrcNode < nSrcNodes; iSrcNode++) {
    if (AUTORES_I(self->srcNodes_destNodeId, iSrcNode) != -1) {
      nDestNodes++;
    }
  }

  flag = AutoMesh_Node3D_allocateNodes 
    (self->destNode, nDestNodes);
  if (!flag) return 0;

  assert(0 < AutoMesh_Node3D_nNodes (self->destNode));

  return 1;
}

static int /* boolean */
MakeSrcNodeId 
(AutoMesh_SurfaceExtractor3D *self)
{
  int nSrcNodes = AutoMesh_Node3D_nNodes (self->srcNode);
  int iSrcNode;
  int flag;

  assert(0 < AutoMesh_Node3D_nNodes (self->destNode));
  assert(self->destNodes_srcNodeId == NULL);

  self->destNodes_srcNodeId = AutoRes_IntegerColumn_New
    (self->destNode->nodes, COLUMN_NAME_SRC_NODE_ID);
  assert(self->destNodes_srcNodeId != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->destNode->nodes, 
     self->destNodes_srcNodeId);
  if (!flag) return 0;

  for (iSrcNode = 0; iSrcNode < nSrcNodes; iSrcNode++) {
    int destNodeId = AUTORES_I(self->srcNodes_destNodeId, iSrcNode);

    if (destNodeId != -1) {
      AUTORES_I_SET(self->destNodes_srcNodeId, destNodeId, 
		    iSrcNode);
    }
  }

  assert(self->destNodes_srcNodeId != NULL);

  return 1;
}

static int /* boolean */
MakeNode 
(AutoMesh_SurfaceExtractor3D *self)
{
  int nDestNodes = AutoMesh_Node3D_nNodes (self->destNode);
  int iDestNode;
  int flag;

  assert(0 < nDestNodes);
  assert(self->destNodes_srcNodeId != NULL);
  assert(self->destNode->nodes_coord == NULL);

  flag = AutoMesh_Node3D_makeCoordColumn (self->destNode);
  if (!flag) return 0;

  for (iDestNode = 0; iDestNode < nDestNodes; iDestNode++) {
    int srcNodeId = AUTORES_I(self->destNodes_srcNodeId, iDestNode);
    double x, y, z;

    AutoMesh_Node3D_getCoordAt (self->srcNode, srcNodeId,
				&x, &y, &z);
    AutoMesh_Node3D_setCoordAt (self->destNode, iDestNode,
				x, y, z);
  }

  assert(self->destNode->nodes_coord != NULL);

  return 1;
}

static int /* boolean */
MakeElement 
(AutoMesh_SurfaceExtractor3D *self)
{
  int nDestElements = AutoMesh_SurfaceElement3D_nElements (self->destElement);
  int nElementNodes = self->srcElement->nElementNodes;
  int iDestElement;
  int iElementNode;
  int flag;

  assert(0 < nDestElements);
  assert(self->srcNodes_destNodeId != NULL);
  assert(self->destElement->elements_nodeIds == NULL);

  flag = AutoMesh_SurfaceElement3D_makeNodeIdsColumn (self->destElement);
  if (!flag) return 0;

  for (iDestElement = 0; iDestElement < nDestElements; iDestElement++) {
    int srcElementId = AutoRes_Selection_objectId 
      (self->elementSelection, iDestElement);
    
    for (iElementNode = 0; 
	 iElementNode < nElementNodes; 
	 iElementNode++) {
      int srcNodeId = AutoMesh_SurfaceElement3D_nodeIdAt
	(self->srcElement, srcElementId, iElementNode);
      int destNodeId = AUTORES_I(self->srcNodes_destNodeId, srcNodeId);
      
      AutoMesh_SurfaceElement3D_setNodeIdAt
	(self->destElement, iDestElement, iElementNode, destNodeId);
    }
  }

  assert(self->destElement->elements_nodeIds != NULL);

  return 1;
}

int /* boolean */
AutoMesh_SurfaceExtractor3D_do 
(AutoMesh_SurfaceExtractor3D *self,
 int nodeFlag,
 int elementFlag,
 int srcElementIdFlag,
 int destElementIdFlag,
 int srcNodeIdFlag)
{
  int flag;

  flag = AllocateDestElement (self);
  if (!flag) return 0;

  if (srcElementIdFlag) {
    flag = MakeSrcElementId (self);
    if (!flag) return 0;
  }

  if (destElementIdFlag) {
    flag = MakeDestElementId (self);
    if (!flag) return 0;
  }

  flag = MakeDestNodeId (self);
  if (!flag) return 0;

  if (srcNodeIdFlag) {
    flag = AllocateDestNode (self);
    if (!flag) return 0;

    flag = MakeSrcNodeId (self);
    if (!flag) return 0;
  }

  if (srcNodeIdFlag && nodeFlag) {
    flag = MakeNode (self);
    if (!flag) return 0;
  }

  if (elementFlag) {
    flag = MakeElement (self);
    if (!flag) return 0;
  }

  return 1;
}

int AutoMesh_SurfaceExtractor3D_destNodeIdAtSrcNode
(AutoMesh_SurfaceExtractor3D *self,
 int srcNodeId)
{
  return AUTORES_I(self->srcNodes_destNodeId, srcNodeId);
}

int AutoMesh_SurfaceExtractor3D_srcNodeIdAtDestNode
(AutoMesh_SurfaceExtractor3D *self,
 int destNodeId)
{
  return AUTORES_I(self->destNodes_srcNodeId, destNodeId);
}

int AutoMesh_SurfaceExtractor3D_destElementIdAtSrcElement
(AutoMesh_SurfaceExtractor3D *self,
 int srcElementId)
{
  return AUTORES_I(self->srcElements_destElementId, srcElementId);
}

int AutoMesh_SurfaceExtractor3D_srcElementIdAtDestElement
(AutoMesh_SurfaceExtractor3D *self,
 int destElementId)
{
  return AUTORES_I(self->destElements_srcElementId, destElementId);
}







