/* automesh_surface_grouper3d.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"
#include "automesh_shape.h"
#include "automesh_node3d.h"
#include "automesh_frame3d.h"
#include "automesh_surface_element3d.h"
#include "automesh_surface_edge3d.h"
#include "automesh_surface_grouper3d.h"



#define TABLE_NAME_FRAMES "AutoMesh:frames"
#define COLUMN_NAME_GROUP_ID "AutoMesh:groupId"
#define COLUMN_NAME_ELEMENT_ID "AutoMesh:elementId"
#define COLUMN_NAME_ELEMENT_EDGE_NUM "AutoMesh:elementEdgeNum"
#define COLUMN_NAME_FRAME_FLAG "AutoMesh:frameFlag"
#define COLUMN_NAME_MARK_FLAG "AutoMesh:markFlag"
#define COLUMN_NAME_FRAME_GROUP_ID "AutoMesh:frameGroupId"
#define COLUMN_NAME_FRAME_VERTEX_ID "AutoMesh:frameVertexId"



/* SurfaceGrouper3D */

AUTORES_IMPLEMENT_HANDLE(AutoMesh_SurfaceGrouper3D)

AutoMesh_SurfaceGrouper3D *AutoMesh_SurfaceGrouper3D_New 
(AutoMesh_SurfaceEdge3D *edge)
{
  AutoMesh_SurfaceGrouper3D *grouper;

  grouper = AutoMesh_SurfaceGrouper3D_NewImplementation ();

  grouper->nGroups = 0;
  grouper->nFrameGroups = 0;
  grouper->nFrameVertexs = 0;

  grouper->node = edge->element->node;
  grouper->nodes_frameVertexId = NULL;
  grouper->nodes_markFlag = NULL;

  grouper->element = edge->element;
  grouper->elements_groupId = NULL;
  grouper->elements_elementId = NULL;
  grouper->elements_elementEdgeNum = NULL;

  grouper->edge = edge;
  grouper->edges_frameFlag = NULL;
  grouper->edges_frameGroupId = NULL;

  grouper->frame = AutoMesh_Frame3D_New ();
  assert(grouper->frame != NULL);
  
  return grouper;
}

AutoMesh_SurfaceGrouper3D *AutoMesh_SurfaceGrouper3D_NewWithoutEdge 
(AutoMesh_SurfaceElement3D *element)
{
  AutoMesh_SurfaceGrouper3D *grouper;

  grouper = AutoMesh_SurfaceGrouper3D_NewImplementation ();

  grouper->nGroups = 0;
  grouper->nFrameGroups = 0;
  grouper->nFrameVertexs = 0;

  grouper->node = element->node;
  grouper->nodes_frameVertexId = NULL;
  grouper->nodes_markFlag = NULL;

  grouper->element = element;
  grouper->elements_groupId = NULL;
  grouper->elements_elementId = NULL;
  grouper->elements_elementEdgeNum = NULL;

  grouper->edge = NULL;
  grouper->edges_frameFlag = NULL;
  grouper->edges_frameGroupId = NULL;

  grouper->frame = NULL;
  
  return grouper;
}

void AutoMesh_SurfaceGrouper3D_delete 
(AutoMesh_SurfaceGrouper3D *self)
{
  if (self->nodes_frameVertexId != NULL) {
    AutoRes_Table_removeColumn 
      (self->node->nodes, 
       AutoRes_Table_columnIdOfColumn (self->node->nodes, 
				       self->nodes_frameVertexId));
    self->nodes_frameVertexId = NULL;
  }

  if (self->nodes_markFlag != NULL) {
    AutoRes_Table_removeColumn 
      (self->node->nodes, 
       AutoRes_Table_columnIdOfColumn (self->node->nodes, 
				       self->nodes_markFlag));
    self->nodes_markFlag = NULL;
  }

  if (self->elements_groupId != NULL) {
    AutoRes_Table_removeColumn 
      (self->element->elements, 
       AutoRes_Table_columnIdOfColumn (self->element->elements, 
				       self->elements_groupId));
    self->elements_groupId = NULL;
  }

  if (self->elements_elementId != NULL) {
    AutoRes_Table_removeColumn 
      (self->element->elements, 
       AutoRes_Table_columnIdOfColumn (self->element->elements, 
				       self->elements_elementId));
    self->elements_elementId = NULL;
  }

  if (self->elements_elementEdgeNum != NULL) {
    AutoRes_Table_removeColumn 
      (self->element->elements, 
       AutoRes_Table_columnIdOfColumn (self->element->elements, 
				       self->elements_elementEdgeNum));
    self->elements_elementEdgeNum = NULL;
  }

  if (self->edges_frameFlag != NULL) {
    AutoRes_Table_removeColumn 
      (self->edge->edges, 
       AutoRes_Table_columnIdOfColumn (self->edge->edges, 
				       self->edges_frameFlag));
    self->edges_frameFlag = NULL;
  }

  if (self->edges_frameGroupId != NULL) {
    AutoRes_Table_removeColumn 
      (self->edge->edges, 
       AutoRes_Table_columnIdOfColumn (self->edge->edges, 
				       self->edges_frameGroupId));
    self->edges_frameGroupId = NULL;
  }

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoMesh_SurfaceGrouper3D_makeGroupIdAtElementColumn 
(AutoMesh_SurfaceGrouper3D *self)
{
  int flag;

  self->elements_groupId = AutoRes_IntegerColumn_New 
    (self->element->elements, COLUMN_NAME_GROUP_ID);
  assert(self->elements_groupId != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->element->elements, self->elements_groupId);
  if (!flag) return 0;

  return 1;
}

void AutoMesh_SurfaceGrouper3D_setNGroups 
(AutoMesh_SurfaceGrouper3D *self)
{
  int nElements = AutoMesh_SurfaceElement3D_nElements (self->element);
  int iElement;
  int maxGroupId;
  
  assert(self->elements_groupId != NULL);
  assert(self->nGroups == 0);

  maxGroupId = -1;
  for (iElement = 0; iElement < nElements; iElement++) {
    int groupId = AutoMesh_SurfaceGrouper3D_groupIdAtElement
      (self, iElement);
    
    if (maxGroupId < groupId) {
      maxGroupId = groupId;
    }
  }
  assert(0 <= maxGroupId);
  self->nGroups = maxGroupId + 1;
}

int /* boolean */
AutoMesh_SurfaceGrouper3D_makeFrameFlagAtEdgeColumn 
(AutoMesh_SurfaceGrouper3D *self)
{
  int flag;

  self->edges_frameFlag = AutoRes_IntegerColumn_New 
    (self->edge->edges, COLUMN_NAME_FRAME_FLAG);
  assert(self->edges_frameFlag != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->edge->edges, self->edges_frameFlag);
  if (!flag) return 0;

  return 1;
}

int /* boolean */
AutoMesh_SurfaceGrouper3D_makeFrameGroupIdAtEdgeColumn 
(AutoMesh_SurfaceGrouper3D *self)
{
  int flag;

  self->edges_frameGroupId = AutoRes_IntegerColumn_New 
    (self->edge->edges, COLUMN_NAME_FRAME_GROUP_ID);
  assert(self->edges_frameGroupId != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->edge->edges, self->edges_frameGroupId);
  if (!flag) return 0;

  return 1;
}

void AutoMesh_SurfaceGrouper3D_setNFrameGroups 
(AutoMesh_SurfaceGrouper3D *self)
{
  int nEdges = AutoMesh_SurfaceEdge3D_nEdges (self->edge);
  int iEdge;
  int maxFrameGroupId;
  
  assert(self->edge != NULL);
  assert(self->edges_frameGroupId != NULL);
  assert(self->nFrameGroups == 0);
  
  maxFrameGroupId = -1;
  for (iEdge = 0; iEdge < nEdges; iEdge++) {
    int groupId = AutoMesh_SurfaceGrouper3D_frameGroupIdAtEdge
      (self, iEdge);
    
    if (maxFrameGroupId < groupId) {
      maxFrameGroupId = groupId;
    }
  }
  assert(-1 <= maxFrameGroupId);
  self->nFrameGroups = maxFrameGroupId + 1;
}

int /* boolean */
AutoMesh_SurfaceGrouper3D_makeFrameVertexIdAtNodeColumn 
(AutoMesh_SurfaceGrouper3D *self)
{
  int flag;

  self->nodes_frameVertexId = AutoRes_IntegerColumn_New 
    (self->node->nodes, COLUMN_NAME_FRAME_VERTEX_ID);
  assert(self->nodes_frameVertexId != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->node->nodes, self->nodes_frameVertexId);
  if (!flag) return 0;

  return 1;
}

void AutoMesh_SurfaceGrouper3D_setNFrameVertexs 
(AutoMesh_SurfaceGrouper3D *self)
{
  int nNodes = AutoMesh_Node3D_nNodes (self->node);
  int iNode;
  int maxFrameVertexId;
  
  assert(self->nodes_frameVertexId != NULL);
  assert(self->nFrameVertexs == 0);
  
  maxFrameVertexId = -1;
  for (iNode = 0; iNode < nNodes; iNode++) {
    int vertexId = AutoMesh_SurfaceGrouper3D_frameVertexIdAtNode
      (self, iNode);
    
    if (maxFrameVertexId < vertexId) {
      maxFrameVertexId = vertexId;
    }
  }
  assert(-1 <= maxFrameVertexId);
  self->nFrameVertexs = maxFrameVertexId + 1;
}

static void SetFrameFlag 
(AutoMesh_SurfaceGrouper3D *self,
 double angle)
{
  double radian = M_PI * angle / 180;
  int nEdges = AutoMesh_SurfaceEdge3D_nEdges (self->edge);
  int iEdge;

  assert(self->edge->edges_elementIds != NULL);
  assert(self->edge->edges_elementEdgeNums != NULL);

  for (iEdge = 0; iEdge < nEdges; iEdge++) {
    int elementIds[2];
    int elementEdgeNums[2];
    int elementId0, elementId1;

    AutoMesh_SurfaceEdge3D_getEdgeAt 
      (self->edge, iEdge, 
       elementIds, elementEdgeNums);
    elementId0 = elementIds[0];
    elementId1 = elementIds[1];
    
    AUTORES_I_SET(self->edges_frameFlag, iEdge, 0);

    if (elementId1 == -1) {
      AUTORES_I_SET(self->edges_frameFlag, iEdge, 1);
    } else {
      double nx0, ny0, nz0;
      double nx1, ny1, nz1;
      double sp;
      
      assert(elementId0 != -1);
      assert(elementId1 != -1);
      
      AutoMesh_SurfaceElement3D_getNormalAt (self->element, elementId0,
					     &nx0, &ny0, &nz0);
      AutoMesh_SurfaceElement3D_getNormalAt (self->element, elementId1,
					     &nx1, &ny1, &nz1);
      sp = AutoMesh_ScalarProduct3D (nx0, ny0, nz0,
				     nx1, ny1, nz1);
      if (sp <= cos (radian)) {
	AUTORES_I_SET(self->edges_frameFlag, iEdge, 1);
      }
    }
  }
}

static int /* boolean */
MakeWorkArea 
(AutoMesh_SurfaceGrouper3D *self)
{
  int flag;
  
  self->elements_elementId = AutoRes_IntegerColumn_New 
    (self->element->elements, COLUMN_NAME_ELEMENT_ID);
  assert(self->elements_elementId != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->element->elements, self->elements_elementId);
  if (!flag) return 0;
  
  self->elements_elementEdgeNum = AutoRes_IntegerColumn_New 
    (self->element->elements, COLUMN_NAME_ELEMENT_EDGE_NUM);
  assert(self->elements_elementEdgeNum != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->element->elements, self->elements_elementEdgeNum);
  if (!flag) return 0;

  return 1;
}

static void Mark 
(AutoMesh_SurfaceGrouper3D *self,
 int firstElementId,
 int groupId)
{
  int nElements = AutoMesh_SurfaceElement3D_nElements (self->element);
  int nElementEdges 
    = AutoMesh_Polygon_NEdges (self->element->polygonType);
  int stackId;

  assert(0 <= groupId);

  AUTORES_I_SET(self->elements_groupId, firstElementId, groupId);

  stackId = 0;
  AUTORES_I_SET(self->elements_elementId, stackId, 
		firstElementId);
  AUTORES_I_SET(self->elements_elementEdgeNum, stackId, 
		0);
  
  for (;;) {
    if (0 <= stackId) {
      int elementId 
	= AUTORES_I(self->elements_elementId, stackId);
      int elementEdgeNum 
	= AUTORES_I(self->elements_elementEdgeNum, stackId);
      
      if (elementEdgeNum == nElementEdges) {
	stackId--;
      } else {
	int neighborId = AutoMesh_SurfaceElement3D_neighborIdAt
	  (self->element, elementId, elementEdgeNum);
	int edgeId = AutoMesh_SurfaceEdge3D_edgeIdAtElement
	  (self->edge, elementId, elementEdgeNum);
	
	if (!AUTORES_I(self->edges_frameFlag, edgeId)
	    && AUTORES_I(self->elements_groupId, neighborId) == -1) {
	  AUTORES_I_SET(self->elements_groupId, neighborId, groupId);

	  stackId++;
	  assert(stackId < nElements);
	  AUTORES_I_SET(self->elements_elementId, stackId, 
			neighborId);
	  AUTORES_I_SET(self->elements_elementEdgeNum, stackId, 
			0);

	} else {
	  AUTORES_I_SET(self->elements_elementEdgeNum, stackId, 
			elementEdgeNum + 1);
	}
      }
    } else {
      break;
    }
  }
}

static void MakeGroup 
(AutoMesh_SurfaceGrouper3D *self)
{
  int nElements = AutoMesh_SurfaceElement3D_nElements (self->element);
  int iElement;
  int currentGroupId;

  assert(self->edge->elements_edgeIds != NULL);
  assert(self->element->elements_neighborIds != NULL);

  for (iElement = 0; iElement < nElements; iElement++) {
    AUTORES_I_SET(self->elements_groupId, iElement, -1);
  }

  currentGroupId = 0;
  for (;;) {
    int elementIdOfNewGroup = -1;

    for (iElement = 0; iElement < nElements; iElement++) {
      int groupId = AUTORES_I(self->elements_groupId, iElement);
      
      if (groupId == -1) {
	elementIdOfNewGroup = iElement;
	break;
      }
    }
    if (elementIdOfNewGroup == -1) {
      break;
    }
    Mark (self, elementIdOfNewGroup, currentGroupId);
    currentGroupId++;
  }
  self->nGroups = currentGroupId;

  for (iElement = 0; iElement < nElements; iElement++) {
    int groupId = AUTORES_I(self->elements_groupId, iElement);
    
    assert(0 <= groupId);
  }
}

static void ResetFrameFlag 
(AutoMesh_SurfaceGrouper3D *self)
{
  int nEdges = AutoMesh_SurfaceEdge3D_nEdges (self->edge);
  int iEdge;

  assert(self->edges_frameFlag != NULL);
  assert(self->elements_groupId != NULL);

  for (iEdge = 0; iEdge < nEdges; iEdge++) {
    if (AUTORES_I(self->edges_frameFlag, iEdge)) {
      int elementIds[2];
      int elementEdgeNums[2];
      int elementId0, elementId1;
      
      AutoMesh_SurfaceEdge3D_getEdgeAt 
	(self->edge, iEdge, 
	 elementIds, elementEdgeNums);
      elementId0 = elementIds[0];
      elementId1 = elementIds[1];

      if (elementId1 != -1) {
	int groupId0 = AUTORES_I(self->elements_groupId, elementId0);
	int groupId1 = AUTORES_I(self->elements_groupId, elementId1);
	
	if (groupId0 == groupId1) {
	  AUTORES_I_SET(self->edges_frameFlag, iEdge, 0);
	}
      }
    }
  }
}

static int /* boolean */
MakeFrame 
(AutoMesh_SurfaceGrouper3D *self)
{
  int nEdges = AutoMesh_SurfaceEdge3D_nEdges (self->edge);
  int iEdge;
  int nFrames;
  int frameId;
  int flag;

  assert(self->edges_frameFlag != NULL);

  nFrames = 0;
  for (iEdge = 0; iEdge < nEdges; iEdge++) {
    if (AUTORES_I(self->edges_frameFlag, iEdge)) {
      nFrames++;
    }
  }

  flag = AutoMesh_Frame3D_allocateFrames (self->frame, nFrames);
  if (!flag) return 0;

  flag = AutoMesh_Frame3D_makeEndPointsColumn (self->frame);
  if (!flag) return 0;

  frameId = 0;
  for (iEdge = 0; iEdge < nEdges; iEdge++) {
    if (AUTORES_I(self->edges_frameFlag, iEdge)) {
      int nodeId0, nodeId1;
      double x0, y0, z0;
      double x1, y1, z1;

      AutoMesh_SurfaceEdge3D_getVertexNodeIdsAt (self->edge, iEdge,
						 &nodeId0, &nodeId1);
      AutoMesh_Node3D_getCoordAt (self->element->node, nodeId0,
				  &x0, &y0, &z0);
      AutoMesh_Node3D_getCoordAt (self->element->node, nodeId1,
				  &x1, &y1, &z1);

      AutoMesh_Frame3D_setEndPointsAt 
	(self->frame, frameId, 
	 x0, y0, z0, 
	 x1, y1, z1);

      frameId++;
    }
  }

  return 1;
}

static void GroupFrame 
(AutoMesh_SurfaceGrouper3D *self)
{
  int nEdges = AutoMesh_SurfaceEdge3D_nEdges (self->edge);
  int iEdge;
  int currentFrameGroupId;

  assert(self->edges_frameFlag != NULL);
  assert(self->edges_frameGroupId != NULL);

  for (iEdge = 0; iEdge < nEdges; iEdge++) {
    AUTORES_I_SET(self->edges_frameGroupId, iEdge, -1);
  }

  currentFrameGroupId = 0;
  for (;;) {
    int currentGroupId0, currentGroupId1;

    currentGroupId0 = -1;
    currentGroupId1 = -1;
    for (iEdge = 0; iEdge < nEdges; iEdge++) {
      if (AUTORES_I(self->edges_frameFlag, iEdge)) {
	int elementIds[2];
	int elementEdgeNums[2];
	int elementId0, elementId1;
	int groupId0, groupId1;
	
	AutoMesh_SurfaceEdge3D_getEdgeAt 
	  (self->edge, iEdge, 
	   elementIds, elementEdgeNums);
	elementId0 = elementIds[0];
	elementId1 = elementIds[1];
	
	groupId0 = AUTORES_I(self->elements_groupId, elementId0);
	assert(0 <= groupId0);

	if (elementId1 != -1) {
	  groupId1 = AUTORES_I(self->elements_groupId, elementId1);
	  assert(0 <= groupId1);
	} else {
	  groupId1 = -1;
	}
	assert(groupId0 != groupId1);
	
	if (AUTORES_I(self->edges_frameGroupId, iEdge) == -1) {
	  currentGroupId0 = groupId0;
	  currentGroupId1 = groupId1;
	  break;
	}
      }
    }

    if (currentGroupId0 == -1) {
      break;
    }
    assert(0 <= currentGroupId0);
    assert(0 <= currentGroupId1
	   || currentGroupId1 == -1);
    assert(currentGroupId0 != currentGroupId1);

    for (iEdge = 0; iEdge < nEdges; iEdge++) {
      if (AUTORES_I(self->edges_frameFlag, iEdge)) {
	int elementIds[2];
	int elementEdgeNums[2];
	int elementId0, elementId1;
	int groupId0, groupId1;
	
	AutoMesh_SurfaceEdge3D_getEdgeAt 
	  (self->edge, iEdge, 
	   elementIds, elementEdgeNums);
	elementId0 = elementIds[0];
	elementId1 = elementIds[1];
	
	groupId0 = AUTORES_I(self->elements_groupId, elementId0);
	assert(0 <= groupId0);

	if (elementId1 != -1) {
	  groupId1 = AUTORES_I(self->elements_groupId, elementId1);
	  assert(0 <= groupId1);
	} else {
	  groupId1 = -1;
	}
	assert(groupId0 != groupId1);
	
	if ((currentGroupId0 == groupId0 
	     && currentGroupId1 == groupId1)
	    || (currentGroupId0 == groupId1 
		&& currentGroupId1 == groupId0)) {
	  assert(AUTORES_I(self->edges_frameGroupId, iEdge) == -1);
	  
	  AUTORES_I_SET(self->edges_frameGroupId, iEdge, 
			currentFrameGroupId);
	}
      }
    }

    currentFrameGroupId++;
  }
  self->nFrameGroups = currentFrameGroupId;

  for (iEdge = 0; iEdge < nEdges; iEdge++) {
    int frameGroupId = AUTORES_I(self->edges_frameGroupId, iEdge);

    if (AUTORES_I(self->edges_frameFlag, iEdge)) {
      assert(0 <= frameGroupId);
    } else {
      assert(frameGroupId == -1);
    }
  }

}

static void IdentifyFrameVertex 
(AutoMesh_SurfaceGrouper3D *self)
{
  int nNodes = AutoMesh_Node3D_nNodes (self->node);
  int iNode;
  int nEdges = AutoMesh_SurfaceEdge3D_nEdges (self->edge);
  int iEdge;
  int count;

  /* 
     temporary state
       node on frame vertex : <nFrameGroups>
       node on frame : frame group ID, 0 <= ID < <nFrameGroups>
       other node : -1
   */

  if (self->nFrameGroups == 0) {
    return;
  }

  assert(self->edges_frameFlag != NULL);
  assert(self->edges_frameGroupId != NULL);
  assert(0 < self->nFrameGroups);
  assert(self->nodes_frameVertexId != NULL);

  for (iNode = 0; iNode < nNodes; iNode++) {
    AUTORES_I_SET(self->nodes_frameVertexId, iNode, -1);
  }

  for (iEdge = 0; iEdge < nEdges; iEdge++) {
    if (AUTORES_I(self->edges_frameFlag, iEdge)) {
      int frameGroupId = AUTORES_I(self->edges_frameGroupId, iEdge);      
      int nodeId0, nodeId1;
      int frameVertexId0, frameVertexId1;
      
      assert(0 <= frameGroupId);
      assert(frameGroupId < self->nFrameGroups);
      
      AutoMesh_SurfaceEdge3D_getVertexNodeIdsAt (self->edge, iEdge,
						 &nodeId0, &nodeId1);
      
      frameVertexId0 = AUTORES_I(self->nodes_frameVertexId, nodeId0);
      if (frameVertexId0 == self->nFrameGroups) {
	/* already on frame vertex */

	/* do nothing */

      } else if (0 <= frameVertexId0
		 && frameVertexId0 < self->nFrameGroups) {
	/* on frame, or on frame vertex */

	if (frameVertexId0 != frameGroupId) {
	  /* on frame vertex */

	  AUTORES_I_SET(self->nodes_frameVertexId, nodeId0, 
			self->nFrameGroups);
	} else {
	  /* still on frame */

	  /* do nothing */
	}

      } else {
	/* other */
	assert(frameVertexId0 == -1);

	AUTORES_I_SET(self->nodes_frameVertexId, nodeId0, 
		      frameGroupId);
      }
      
      frameVertexId1 = AUTORES_I(self->nodes_frameVertexId, nodeId1);
      if (frameVertexId1 == self->nFrameGroups) {
	/* already on frame vertex */

	/* do nothing */

      } else if (0 <= frameVertexId1
		 && frameVertexId1 < self->nFrameGroups) {
	/* on frame, or on frame vertex */

	if (frameVertexId1 != frameGroupId) {
	  /* on frame vertex */

	  AUTORES_I_SET(self->nodes_frameVertexId, nodeId1, 
			self->nFrameGroups);
	} else {
	  /* still on frame */

	  /* do nothing */
	}

      } else {
	/* other */
	assert(frameVertexId1 == -1);

	AUTORES_I_SET(self->nodes_frameVertexId, nodeId1, 
		      frameGroupId);
      }
    }
  }

  for (iNode = 0; iNode < nNodes; iNode++) {
    if (AUTORES_I(self->nodes_frameVertexId, iNode) < self->nFrameGroups) {
      AUTORES_I_SET(self->nodes_frameVertexId, iNode, -1);
    }
  }

  count = 0;
  for (iNode = 0; iNode < nNodes; iNode++) {
    if (AUTORES_I(self->nodes_frameVertexId, iNode) == self->nFrameGroups) {
      AUTORES_I_SET(self->nodes_frameVertexId, iNode, count);
      count++;
    }
  }
  self->nFrameVertexs = count;
}

int /* boolean */
AutoMesh_SurfaceGrouper3D_do 
(AutoMesh_SurfaceGrouper3D *self,
 double angle,
 int groupIdFlag,
 int frameFlag,
 int frameGroupIdFlag,
 int frameVertexIdFlag)
{
  int flag;
  
  assert(self->elements_groupId == NULL);
  assert(self->edges_frameFlag == NULL);
  assert(self->edges_frameGroupId == NULL);
  assert(self->nodes_frameVertexId == NULL);
  assert(self->nGroups == 0);
  assert(self->nFrameGroups == 0);
  assert(self->nFrameVertexs == 0);
  assert(self->frame->frames_endPoint0 == NULL);
  assert(self->frame->frames_endPoint1 == NULL);

  flag = AutoMesh_SurfaceGrouper3D_makeFrameFlagAtEdgeColumn (self);
  if (!flag) return 0;

  SetFrameFlag (self, angle);

  if (groupIdFlag) {
    flag = AutoMesh_SurfaceGrouper3D_makeGroupIdAtElementColumn (self);
    if (!flag) return 0;

    flag = MakeWorkArea (self);
    if (!flag) return 0;
    
    MakeGroup (self);
    ResetFrameFlag (self);
  }

  if (frameFlag) {
    flag = MakeFrame (self);
    if (!flag) return 0;
  }

  if (frameGroupIdFlag) {
    flag = AutoMesh_SurfaceGrouper3D_makeFrameGroupIdAtEdgeColumn (self);
    if (!flag) return 0;
    
    GroupFrame (self);
  }
  
  if (frameVertexIdFlag && frameGroupIdFlag) {
    flag = AutoMesh_SurfaceGrouper3D_makeFrameVertexIdAtNodeColumn (self);
    if (!flag) return 0;

    IdentifyFrameVertex (self);
  }

  assert(self->edges_frameFlag != NULL);
  if (groupIdFlag) {
    assert(self->elements_groupId != NULL);
    assert(0 < self->nGroups);
  }
  if (frameFlag) {
    assert(self->frame->frames_endPoint0 != NULL);
    assert(self->frame->frames_endPoint1 != NULL);
  }
  if (frameGroupIdFlag) {
    assert(self->edges_frameGroupId != NULL);
    assert(0 <= self->nFrameGroups);
  }
  if (frameVertexIdFlag) {
    assert(self->nodes_frameVertexId != NULL);
    assert(0 <= self->nFrameVertexs);
  }

  return 1;
}

int /* boolean */
AutoMesh_SurfaceGrouper3D_makeMarkFlagAtNodeColumn 
(AutoMesh_SurfaceGrouper3D *self)
{
  int flag;

  self->nodes_markFlag = AutoRes_IntegerColumn_New 
    (self->node->nodes, COLUMN_NAME_MARK_FLAG);
  assert(self->nodes_markFlag != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->node->nodes, self->nodes_markFlag);
  if (!flag) return 0;

  return 1;
}

AutoRes_Selection *   /* CREATE */ 
AutoMesh_SurfaceGrouper3D_getGroupNode 
(AutoMesh_SurfaceGrouper3D *self,
 AutoRes_Selection *groupSelection)
{
  int nNodes = AutoMesh_Node3D_nNodes (self->node);
  int nElements = AutoMesh_SurfaceElement3D_nElements (self->element);
  int nEntrys = AutoRes_Selection_nEntrys (groupSelection);
  int iNode;
  int iElement;
  int iElementNode;
  int iEntry;
  AutoRes_Selection *nodeSelection;
  int flag;

  assert(self->nodes_markFlag != NULL);

  for (iNode = 0; iNode < nNodes; iNode++) {
    AUTORES_I_SET(self->nodes_markFlag, iNode, 0);
  }

  for (iElement = 0; iElement < nElements; iElement++) {
    int groupId = AUTORES_I(self->elements_groupId, iElement);
    int found;
    
    found = 0;
    for (iEntry = 0; iEntry < nEntrys; iEntry++) {
      int selectedGroupId 
	= AutoRes_Selection_objectId (groupSelection, iEntry);

      if (selectedGroupId == groupId) {
	found = 1;
	break;
      }
    }
    if (!found) continue;
    
    for (iElementNode = 0; 
	 iElementNode < self->element->nElementNodes; 
	 iElementNode++) {
      int nodeId = AutoMesh_SurfaceElement3D_nodeIdAt 
	(self->element, iElement, iElementNode);

      AUTORES_I_SET(self->nodes_markFlag, nodeId, 1);
    }
  }

  nodeSelection = AutoRes_Selection_New ();  
  assert(nodeSelection != NULL);

  flag = AutoRes_Selection_import (nodeSelection, self->nodes_markFlag);
  if (!flag) return NULL;

  return nodeSelection;
}

AutoRes_Selection *   /* CREATE */ 
AutoMesh_SurfaceGrouper3D_getGroupElement 
(AutoMesh_SurfaceGrouper3D *self,
 AutoRes_Selection *groupSelection)
{
  int nElements = AutoMesh_SurfaceElement3D_nElements (self->element);
  int nEntrys = AutoRes_Selection_nEntrys (groupSelection);
  int iElement;
  int iEntry;
  int nSelectedElements;
  AutoRes_Selection *elementSelection;
  int flag;

  nSelectedElements = 0;
  for (iElement = 0; iElement < nElements; iElement++) {
    int groupId = AUTORES_I(self->elements_groupId, iElement);
    int found;
    
    found = 0;
    for (iEntry = 0; iEntry < nEntrys; iEntry++) {
      int selectedGroupId 
	= AutoRes_Selection_objectId (groupSelection, iEntry);

      if (selectedGroupId == groupId) {
	found = 1;
	break;
      }
    }

    if (found) {
      nSelectedElements++;
    }
  }

  elementSelection = AutoRes_Selection_New ();  
  assert(elementSelection != NULL);

  flag = AutoRes_Selection_reserve (elementSelection, nSelectedElements);
  if (!flag) return NULL;

  for (iElement = 0; iElement < nElements; iElement++) {
    int groupId = AUTORES_I(self->elements_groupId, iElement);
    int found;
    
    found = 0;
    for (iEntry = 0; iEntry < nEntrys; iEntry++) {
      int selectedGroupId 
	= AutoRes_Selection_objectId (groupSelection, iEntry);

      if (selectedGroupId == groupId) {
	found = 1;
	break;
      }
    }

    if (found) {
      AutoRes_Selection_addObject (elementSelection, iElement);
    }
  }

  return elementSelection;
}

void AutoMesh_SurfaceGrouper3D_purgeWorkArea 
(AutoMesh_SurfaceGrouper3D *self)
{
  if (self->nodes_markFlag != NULL) {
    AutoRes_Table_removeColumn 
      (self->node->nodes, 
       AutoRes_Table_columnIdOfColumn (self->node->nodes, 
				       self->nodes_markFlag));
    self->nodes_markFlag = NULL;
  }

  if (self->elements_elementId != NULL) {
    AutoRes_Table_removeColumn 
      (self->element->elements, 
       AutoRes_Table_columnIdOfColumn 
       (self->element->elements, 
	self->elements_elementId));
    self->elements_elementId = NULL;
  }

  if (self->elements_elementEdgeNum != NULL) {
    AutoRes_Table_removeColumn 
      (self->element->elements, 
       AutoRes_Table_columnIdOfColumn 
       (self->element->elements, 
	self->elements_elementEdgeNum));
    self->elements_elementEdgeNum = NULL;
  }

  assert(self->elements_elementId == NULL);
  assert(self->elements_elementEdgeNum == NULL);
}

int AutoMesh_SurfaceGrouper3D_nGroups 
(AutoMesh_SurfaceGrouper3D *self)
{
  return self->nGroups;
}

int AutoMesh_SurfaceGrouper3D_groupIdAtElement 
(AutoMesh_SurfaceGrouper3D *self,
 int elementId)
{
  return AUTORES_I(self->elements_groupId, elementId);
}

int /* boolean */
AutoMesh_SurfaceGrouper3D_frameFlagAtEdge 
(AutoMesh_SurfaceGrouper3D *self,
 int edgeId)
{
  return AUTORES_I(self->edges_frameFlag, edgeId);
}

int AutoMesh_SurfaceGrouper3D_nFrameGroups 
(AutoMesh_SurfaceGrouper3D *self)
{
  return self->nFrameGroups;
}

int AutoMesh_SurfaceGrouper3D_frameGroupIdAtEdge 
(AutoMesh_SurfaceGrouper3D *self,
 int edgeId)
{
  return AUTORES_I(self->edges_frameGroupId, edgeId);
}

int AutoMesh_SurfaceGrouper3D_nFrameVertexs 
(AutoMesh_SurfaceGrouper3D *self)
{
  return self->nFrameVertexs;
}

int AutoMesh_SurfaceGrouper3D_frameVertexIdAtNode 
(AutoMesh_SurfaceGrouper3D *self,
 int nodeId)
{
  return AUTORES_I(self->nodes_frameVertexId, nodeId);
}


