/* automesh_volume_boundary_edge2d.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"
#include "automesh_shape.h"
#include "automesh_node2d.h"
#include "automesh_volume_element2d.h"
#include "automesh_volume_boundary_edge2d.h"



#define TABLE_NAME_EDGES "AutoMesh:edges"
#define COLUMN_NAME_VERTEX_NODE_IDS "AutoMesh:vertexNodeIds"
#define COLUMN_NAME_ELEMENT_ID "AutoMesh:elementId"
#define COLUMN_NAME_ELEMENT_EDGE_NUM "AutoMesh:elementEdgeNum"



/* VolumeBoundaryEdge2D */

AUTORES_IMPLEMENT_HANDLE(AutoMesh_VolumeBoundaryEdge2D)

AutoMesh_VolumeBoundaryEdge2D *AutoMesh_VolumeBoundaryEdge2D_New 
(AutoMesh_VolumeElement2D *element)
{
  AutoMesh_VolumeBoundaryEdge2D *edge;

  edge = AutoMesh_VolumeBoundaryEdge2D_NewImplementation ();

  edge->polygonType = element->polygonType;
  edge->nElementEdges 
    = AutoMesh_Polygon_NEdges (element->polygonType);

  edge->element = element;

  edge->edges = AutoRes_Table_New (TABLE_NAME_EDGES);
  assert(edge->edges != NULL);
  edge->edgeCollection = AutoRes_FreeListCollection_New (edge->edges);
  assert(edge->edgeCollection != NULL);
  edge->edges_vertexNodeIds = NULL;
  edge->edges_elementId = NULL;
  edge->edges_elementEdgeNum = NULL;

  return edge;
}

void AutoMesh_VolumeBoundaryEdge2D_delete 
(AutoMesh_VolumeBoundaryEdge2D *self)
{
  if (self->edgeCollection != NULL) {
    AutoRes_FreeListCollection_delete (self->edgeCollection);
    self->edgeCollection = NULL;
  }

  AutoRes_Table_delete (self->edges);
  self->edges = NULL;
  self->edges_vertexNodeIds = NULL;
  self->edges_elementId = NULL;
  self->edges_elementEdgeNum = NULL;

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoMesh_VolumeBoundaryEdge2D_makeVertexNodeIdsColumn 
(AutoMesh_VolumeBoundaryEdge2D *self)
{
  int flag;
  
  self->edges_vertexNodeIds = AutoRes_Integer1DColumn_New 
    (self->edges, COLUMN_NAME_VERTEX_NODE_IDS, 2);
  assert(self->edges_vertexNodeIds != NULL);
  flag = AutoRes_Table_addInteger1DColumn 
    (self->edges, self->edges_vertexNodeIds);
  if (!flag) return 0;

  return 1;
}

int /* boolean */
AutoMesh_VolumeBoundaryEdge2D_makeElementIdColumn 
(AutoMesh_VolumeBoundaryEdge2D *self)
{
  int flag;
  
  self->edges_elementId = AutoRes_IntegerColumn_New 
    (self->edges, COLUMN_NAME_ELEMENT_ID);
  assert(self->edges_elementId != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->edges, self->edges_elementId);
  if (!flag) return 0;
  
  self->edges_elementEdgeNum = AutoRes_IntegerColumn_New 
    (self->edges, COLUMN_NAME_ELEMENT_EDGE_NUM);
  assert(self->edges_elementEdgeNum != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->edges, self->edges_elementEdgeNum);
  if (!flag) return 0;

  return 1;
}

static int FindEdge 
(AutoMesh_VolumeBoundaryEdge2D *self,
 int nodeId0, int nodeId1,
 AutoRes_ListTable *edgeTable)
{
  int hashTableSize;
  int hashValue;
  int linkId;

  hashTableSize = AutoRes_ListTable_nLists (edgeTable);
  hashValue = (nodeId0 + nodeId1) % hashTableSize;
  for (linkId = AutoRes_ListTable_startLinkId (edgeTable, hashValue);
       linkId != -1;
       linkId = AutoRes_ListTable_nextLinkId (edgeTable, linkId)) {
    int edgeId = AutoRes_ListTable_objectId (edgeTable, linkId);
    
    int eNodeId0 = AUTORES_I1D(self->edges_vertexNodeIds, edgeId, 0);
    int eNodeId1 = AUTORES_I1D(self->edges_vertexNodeIds, edgeId, 1);

    if (AutoMesh_IdIsEqual2 (nodeId0, nodeId1,
			     eNodeId0, eNodeId1)) {
      return edgeId;
    }
  }

  return -1;
}

static int AddEdge 
(AutoMesh_VolumeBoundaryEdge2D *self,
 int elementId, int elementEdgeNum,
 int nodeId0, int nodeId1,
 AutoRes_ListTable *edgeTable)
{
  int edgeId;
  int hashTableSize;
  int hashValue;
  int linkId;

  edgeId = AutoRes_FreeListCollection_addRecord (self->edgeCollection);
  if (edgeId == -1) return -1;

  AUTORES_I_SET(self->edges_elementId, edgeId, elementId);
  AUTORES_I_SET(self->edges_elementEdgeNum, edgeId, elementEdgeNum);
  
  AUTORES_I1D_SET(self->edges_vertexNodeIds, edgeId, 0, nodeId0);
  AUTORES_I1D_SET(self->edges_vertexNodeIds, edgeId, 1, nodeId1);

  hashTableSize = AutoRes_ListTable_nLists (edgeTable);
  hashValue = (nodeId0 + nodeId1) % hashTableSize;
  linkId = AutoRes_ListTable_addObject (edgeTable, hashValue, edgeId);
  if (linkId == -1) return -1;

  return edgeId;
}

static void RemoveEdge 
(AutoMesh_VolumeBoundaryEdge2D *self,
 int edgeId,
 int nodeId0, int nodeId1,
 AutoRes_ListTable *edgeTable)
{
  int hashTableSize;
  int hashValue;

  hashTableSize = AutoRes_ListTable_nLists (edgeTable);
  hashValue = (nodeId0 + nodeId1) % hashTableSize;
  AutoRes_ListTable_removeObject (edgeTable, hashValue, edgeId);

  AutoRes_FreeListCollection_removeRecord (self->edgeCollection, edgeId);
}

int /* boolean */
AutoMesh_VolumeBoundaryEdge2D_extract 
(AutoMesh_VolumeBoundaryEdge2D *self)
{
  int nElements = AutoMesh_VolumeElement2D_nElements (self->element);
  AutoRes_ListTable *edgeTable;
  int hashTableSize;
  int nAverageLinksPerList;
  int iElement;
  int iElementEdge;
  int flag;

  assert(AutoRes_Table_nRecords (self->edges) == 0);
  assert(self->edges_vertexNodeIds == NULL);
  assert(self->edges_elementId == NULL);
  assert(self->edges_elementEdgeNum == NULL);

  /* NEED TO REDEFINE */
  /* experience-based value */
  flag = AutoRes_Table_reserveRecords (self->edges, nElements * 1);
  if (!flag) return 0;

  flag = AutoRes_FreeListCollection_makeColumn (self->edgeCollection);
  if (!flag) return 0;

  flag = AutoMesh_VolumeBoundaryEdge2D_makeVertexNodeIdsColumn (self);
  if (!flag) return 0;

  flag = AutoMesh_VolumeBoundaryEdge2D_makeElementIdColumn (self);
  if (!flag) return 0;

  edgeTable = AutoRes_ListTable_New ();
  assert(edgeTable != NULL);

  /* NEED TO REDEFINE */
  /* experience-based value */
  hashTableSize = nElements * 1;
  nAverageLinksPerList = 2;
  flag = AutoRes_ListTable_reserve 
    (edgeTable, hashTableSize, nAverageLinksPerList);
  if (!flag) return 0;

  for (iElement = 0; iElement < nElements; iElement++) {
    for (iElementEdge = 0; 
	 iElementEdge < self->nElementEdges; 
	 iElementEdge++) {
      int elementNodeNum0 = AutoMesh_Polygon_EdgeVertex
	(self->polygonType, iElementEdge, 0);
      int elementNodeNum1 = AutoMesh_Polygon_EdgeVertex
	(self->polygonType, iElementEdge, 1);

      int nodeId0 = AutoMesh_VolumeElement2D_nodeIdAt
	(self->element, iElement, elementNodeNum0);
      int nodeId1 = AutoMesh_VolumeElement2D_nodeIdAt
	(self->element, iElement, elementNodeNum1);

      int edgeId = FindEdge (self, nodeId0, nodeId1, edgeTable);

      if (edgeId == -1) {
	edgeId = AddEdge (self, iElement, iElementEdge,
			  nodeId0, nodeId1, edgeTable);
	if (edgeId == -1) return 0;
      } else {
	RemoveEdge (self, edgeId, nodeId0, nodeId1, edgeTable);
      }
    }
  }

  AutoRes_ListTable_delete (edgeTable);

  AutoRes_FreeListCollection_pack (self->edgeCollection);

  assert(0 < AutoRes_Table_nRecords (self->edges));
  assert(self->edges_vertexNodeIds != NULL);
  assert(self->edges_elementId != NULL);
  assert(self->edges_elementEdgeNum != NULL);

  return 1;
}

void AutoMesh_VolumeBoundaryEdge2D_purgeWorkArea 
(AutoMesh_VolumeBoundaryEdge2D *self)
{
  if (self->edgeCollection != NULL) {
    AutoRes_FreeListCollection_delete (self->edgeCollection);
    self->edgeCollection = NULL;
  }

  if (self->edges_vertexNodeIds != NULL) {
    AutoRes_Table_removeColumn 
      (self->edges, 
       AutoRes_Table_columnIdOfColumn 
       (self->edges, 
	self->edges_vertexNodeIds));
    self->edges_vertexNodeIds = NULL;
  }

  assert(self->edgeCollection == NULL);
  assert(self->edges_vertexNodeIds == NULL);
}

int AutoMesh_VolumeBoundaryEdge2D_nEdges 
(AutoMesh_VolumeBoundaryEdge2D *self)
{
  return AutoRes_Table_nRecords (self->edges);
}

void AutoMesh_VolumeBoundaryEdge2D_getEdgeAt 
(AutoMesh_VolumeBoundaryEdge2D *self,
 int edgeId,
 int *elementId_OUT,
 int *elementEdgeNum_OUT)
{
  *elementId_OUT = AUTORES_I(self->edges_elementId, edgeId);
  *elementEdgeNum_OUT = AUTORES_I(self->edges_elementEdgeNum, edgeId);
}

void AutoMesh_VolumeBoundaryEdge2D_getVertexNodeIdsAt 
(AutoMesh_VolumeBoundaryEdge2D *self,
 int edgeId,
 int *nodeId0_OUT, int *nodeId1_OUT)
{
  assert(self->edges_vertexNodeIds != NULL);

  *nodeId0_OUT = AUTORES_I1D(self->edges_vertexNodeIds, edgeId, 0);
  *nodeId1_OUT = AUTORES_I1D(self->edges_vertexNodeIds, edgeId, 1);
}

void AutoMesh_VolumeBoundaryEdge2D_getMidPointAt 
(AutoMesh_VolumeBoundaryEdge2D *self,
 int edgeId,
 double *x_OUT, double *y_OUT)
{
  AutoMesh_Node2D *node = self->element->node;

  assert(self->edges_vertexNodeIds != NULL);

  /* NEED TO REDEFINE */
  /* assuming straight line */
	 
  {
    int nodeId0 = AUTORES_I1D(self->edges_vertexNodeIds, edgeId, 0);
    int nodeId1 = AUTORES_I1D(self->edges_vertexNodeIds, edgeId, 1);
    double x0, y0;
    double x1, y1;

    AutoMesh_Node2D_getCoordAt (node, nodeId0,
				&x0, &y0);
    AutoMesh_Node2D_getCoordAt (node, nodeId1,
				&x1, &y1);
    *x_OUT = (x0 + x1) / 2;
    *y_OUT = (y0 + y1) / 2;
  }
}


