/* automesh_volume_boundary_face3d.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"
#include "automesh_shape.h"
#include "automesh_node3d.h"
#include "automesh_volume_element3d.h"
#include "automesh_volume_boundary_face3d.h"



#define TABLE_NAME_FACES "AutoMesh:faces"
#define COLUMN_NAME_VERTEX_NODE_IDS "AutoMesh:vertexNodeIds"
#define COLUMN_NAME_ELEMENT_ID "AutoMesh:elementId"
#define COLUMN_NAME_ELEMENT_FACE_NUM "AutoMesh:elementFaceNum"



/* VolumeBoundaryFace3D */

AUTORES_IMPLEMENT_HANDLE(AutoMesh_VolumeBoundaryFace3D)

AutoMesh_VolumeBoundaryFace3D *AutoMesh_VolumeBoundaryFace3D_New 
(AutoMesh_VolumeElement3D *element)
{
  AutoMesh_VolumeBoundaryFace3D *face;

  face = AutoMesh_VolumeBoundaryFace3D_NewImplementation ();

  face->polyhedronType = element->polyhedronType;
  face->nElementFaces 
    = AutoMesh_Polyhedron_NFaces (element->polyhedronType);

  face->element = element;

  face->faces = AutoRes_Table_New (TABLE_NAME_FACES);
  assert(face->faces != NULL);
  face->faceCollection = AutoRes_FreeListCollection_New (face->faces);
  assert(face->faceCollection != NULL);
  face->faces_vertexNodeIds = NULL;
  face->faces_elementId = NULL;
  face->faces_elementFaceNum = NULL;

  return face;
}

void AutoMesh_VolumeBoundaryFace3D_delete 
(AutoMesh_VolumeBoundaryFace3D *self)
{
  if (self->faceCollection != NULL) {
    AutoRes_FreeListCollection_delete (self->faceCollection);
    self->faceCollection = NULL;
  }

  AutoRes_Table_delete (self->faces);
  self->faces = NULL;
  self->faces_vertexNodeIds = NULL;
  self->faces_elementId = NULL;
  self->faces_elementFaceNum = NULL;

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoMesh_VolumeBoundaryFace3D_makeVertexNodeIdsColumn 
(AutoMesh_VolumeBoundaryFace3D *self,
 int nFaceVertexs)
{
  int flag;

  self->faces_vertexNodeIds = AutoRes_Integer1DColumn_New 
    (self->faces, COLUMN_NAME_VERTEX_NODE_IDS, nFaceVertexs);
  assert(self->faces_vertexNodeIds != NULL);
  flag = AutoRes_Table_addInteger1DColumn 
    (self->faces, self->faces_vertexNodeIds);
  if (!flag) return 0;

  return 1;
}

int /* boolean */
AutoMesh_VolumeBoundaryFace3D_makeElementIdColumn 
(AutoMesh_VolumeBoundaryFace3D *self)
{
  int flag;

  self->faces_elementId = AutoRes_IntegerColumn_New 
    (self->faces, COLUMN_NAME_ELEMENT_ID);
  assert(self->faces_elementId != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->faces, self->faces_elementId);
  if (!flag) return 0;
  
  self->faces_elementFaceNum = AutoRes_IntegerColumn_New 
    (self->faces, COLUMN_NAME_ELEMENT_FACE_NUM);
  assert(self->faces_elementFaceNum != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->faces, self->faces_elementFaceNum);
  if (!flag) return 0;

  return 1;
}

static int FindTriangleFace 
(AutoMesh_VolumeBoundaryFace3D *self,
 int nodeId0, int nodeId1, int nodeId2,
 AutoRes_ListTable *faceTable)
{
  int minNodeId, midNodeId, maxNodeId;
  int hashTableSize;
  int hashValue;
  int linkId;

  AutoMesh_SortId3 (nodeId0, nodeId1, nodeId2,
		    &minNodeId, &midNodeId, &maxNodeId);

  hashTableSize = AutoRes_ListTable_nLists (faceTable);
  hashValue = (nodeId0 + nodeId1 + nodeId2) % hashTableSize;
  for (linkId = AutoRes_ListTable_startLinkId (faceTable, hashValue);
       linkId != -1;
       linkId = AutoRes_ListTable_nextLinkId (faceTable, linkId)) {
    int faceId = AutoRes_ListTable_objectId (faceTable, linkId);
    
    int fNodeId0 = AUTORES_I1D(self->faces_vertexNodeIds, faceId, 0);
    int fNodeId1 = AUTORES_I1D(self->faces_vertexNodeIds, faceId, 1);
    int fNodeId2 = AUTORES_I1D(self->faces_vertexNodeIds, faceId, 2);

    int minFNodeId, midFNodeId, maxFNodeId; 

    AutoMesh_SortId3 (fNodeId0, fNodeId1, fNodeId2,
		      &minFNodeId, &midFNodeId, &maxFNodeId);
    
    if (minFNodeId == minNodeId
	&& midFNodeId == midNodeId
	&& maxFNodeId == maxNodeId) {
      return faceId;
    }
  }

  return -1;
}

static int AddTriangleFace 
(AutoMesh_VolumeBoundaryFace3D *self,
 int elementId, int elementFaceNum,
 int nodeId0, int nodeId1, int nodeId2,
 AutoRes_ListTable *faceTable)
{
  int faceId;
  int hashTableSize;
  int hashValue;
  int linkId;

  faceId = AutoRes_FreeListCollection_addRecord (self->faceCollection);
  if (faceId == -1) return -1;

  AUTORES_I_SET(self->faces_elementId, faceId, elementId);
  AUTORES_I_SET(self->faces_elementFaceNum, faceId, elementFaceNum);
  
  AUTORES_I1D_SET(self->faces_vertexNodeIds, faceId, 0, nodeId0);
  AUTORES_I1D_SET(self->faces_vertexNodeIds, faceId, 1, nodeId1);
  AUTORES_I1D_SET(self->faces_vertexNodeIds, faceId, 2, nodeId2);

  hashTableSize = AutoRes_ListTable_nLists (faceTable);
  hashValue = (nodeId0 + nodeId1 + nodeId2) % hashTableSize;
  linkId = AutoRes_ListTable_addObject (faceTable, hashValue, faceId);
  if (linkId == -1) return -1;

  return faceId;
}

static void RemoveTriangleFace 
(AutoMesh_VolumeBoundaryFace3D *self,
 int faceId,
 int nodeId0, int nodeId1, int nodeId2,
 AutoRes_ListTable *faceTable)
{
  int hashTableSize;
  int hashValue;

  hashTableSize = AutoRes_ListTable_nLists (faceTable);
  hashValue = (nodeId0 + nodeId1 + nodeId2) % hashTableSize;
  AutoRes_ListTable_removeObject (faceTable, hashValue, faceId);

  AutoRes_FreeListCollection_removeRecord (self->faceCollection, faceId);
}

static int /* boolean */
ExtractTetraFace
(AutoMesh_VolumeBoundaryFace3D *self)
{
  int nElements = AutoMesh_VolumeElement3D_nElements (self->element);
  AutoRes_ListTable *faceTable;
  int hashTableSize;
  int nAverageLinksPerList;
  int iElement;
  int iElementFace;
  int flag;

  assert(self->polyhedronType == AUTOMESH_POLYHEDRON_TETRA);

  /* NEED TO REDEFINE */
  /* experience-based value */
  flag = AutoRes_Table_reserveRecords (self->faces, nElements * 1);
  if (!flag) return 0;

  flag = AutoRes_FreeListCollection_makeColumn (self->faceCollection);
  if (!flag) return 0;
  
  flag = AutoMesh_VolumeBoundaryFace3D_makeVertexNodeIdsColumn (self, 3);
  if (!flag) return 0;
  
  flag = AutoMesh_VolumeBoundaryFace3D_makeElementIdColumn (self);
  if (!flag) return 0;

  faceTable = AutoRes_ListTable_New ();
  assert(faceTable != NULL);

  /* NEED TO REDEFINE */
  /* experience-based value */
  hashTableSize = nElements * 1;
  nAverageLinksPerList = 2;
  flag = AutoRes_ListTable_reserve 
    (faceTable, hashTableSize, nAverageLinksPerList);
  if (!flag) return 0;

  for (iElement = 0; iElement < nElements; iElement++) {
    for (iElementFace = 0; 
	 iElementFace < self->nElementFaces; 
	 iElementFace++) {
      int elementNodeNum0 = AutoMesh_Tetra_FaceVertexMap[iElementFace][0];
      int elementNodeNum1 = AutoMesh_Tetra_FaceVertexMap[iElementFace][1];
      int elementNodeNum2 = AutoMesh_Tetra_FaceVertexMap[iElementFace][2];

      int nodeId0 = AutoMesh_VolumeElement3D_nodeIdAt 
	(self->element, iElement, elementNodeNum0);
      int nodeId1 = AutoMesh_VolumeElement3D_nodeIdAt 
	(self->element, iElement, elementNodeNum1);
      int nodeId2 = AutoMesh_VolumeElement3D_nodeIdAt 
	(self->element, iElement, elementNodeNum2);

      int faceId = FindTriangleFace 
	(self, nodeId0, nodeId1, nodeId2, faceTable);

      if (faceId == -1) {
	faceId = AddTriangleFace 
	  (self, iElement, iElementFace,
	   nodeId0, nodeId1, nodeId2, faceTable);
	if (faceId == -1) return 0;
      } else {
	RemoveTriangleFace 
	  (self, faceId, nodeId0, nodeId1, nodeId2, faceTable);
      }
    }
  }

  AutoRes_ListTable_delete (faceTable);

  AutoRes_FreeListCollection_pack (self->faceCollection);

  return 1;
}

int /* boolean */
AutoMesh_VolumeBoundaryFace3D_extract 
(AutoMesh_VolumeBoundaryFace3D *self)
{
  int flag = 0;

  assert(AutoRes_Table_nRecords (self->faces) == 0);
  assert(self->faces_vertexNodeIds == NULL);
  assert(self->faces_elementId == NULL);
  assert(self->faces_elementFaceNum == NULL);

  switch (self->polyhedronType) {
  case AUTOMESH_POLYHEDRON_TETRA:
    flag = ExtractTetraFace (self);
    break;
  case AUTOMESH_POLYHEDRON_HEXA:
    /* NEED TO REDEFINE */
    assert(0);
    break;
  case AUTOMESH_POLYHEDRON_PRISM:
    /* NEED TO REDEFINE */
    assert(0);
    break;
  default:
    assert(0);
    break;
  }

  if (flag) {
    assert(0 < AutoRes_Table_nRecords (self->faces));
    assert(self->faces_vertexNodeIds != NULL);
    assert(self->faces_elementId != NULL);
    assert(self->faces_elementFaceNum != NULL);
  }

  return flag;
}

void AutoMesh_VolumeBoundaryFace3D_purgeWorkArea 
(AutoMesh_VolumeBoundaryFace3D *self)
{
  if (self->faceCollection != NULL) {
    AutoRes_FreeListCollection_delete (self->faceCollection);
    self->faceCollection = NULL;
  }

  if (self->faces_vertexNodeIds != NULL) {
    AutoRes_Table_removeColumn 
      (self->faces, 
       AutoRes_Table_columnIdOfColumn 
       (self->faces, 
	self->faces_vertexNodeIds));
    self->faces_vertexNodeIds = NULL;
  }

  assert(self->faceCollection == NULL);
  assert(self->faces_vertexNodeIds == NULL);
}

int AutoMesh_VolumeBoundaryFace3D_nFaces 
(AutoMesh_VolumeBoundaryFace3D *self)
{
  return AutoRes_Table_nRecords (self->faces);
}

void AutoMesh_VolumeBoundaryFace3D_getFaceAt 
(AutoMesh_VolumeBoundaryFace3D *self,
 int faceId,
 int *elementId_OUT,
 int *elementFaceNum_OUT)
{
  *elementId_OUT = AUTORES_I(self->faces_elementId, faceId);
  *elementFaceNum_OUT = AUTORES_I(self->faces_elementFaceNum, faceId);
}

