/* automesh_volume_boundary_maker3d.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"
#include "automesh_shape.h"
#include "automesh_node3d.h"
#include "automesh_surface_element3d.h"
#include "automesh_volume_element3d.h"
#include "automesh_volume_face3d.h"
#include "automesh_volume_boundary_maker3d.h"



#define COLUMN_NAME_SURFACE_NODE_ID "AutoMesh:surfaceNodeId"
#define COLUMN_NAME_VOLUME_NODE_ID "AutoMesh:volumNodeId"
#define COLUMN_NAME_VOLUME_FACE_ID "AutoMesh:volumFaceId"
#define COLUMN_NAME_VOLUME_ELEMENT_ID "AutoMesh:volumElementId"
#define COLUMN_NAME_VOLUME_ELEMENT_FACE_ID "AutoMesh:volumeElementFaceId"



/* VolumeBoundaryMaker3D */

/* NEED TO REDEFINE */
/* 1st order node only on surface, even if volume is high order */

AUTORES_IMPLEMENT_HANDLE(AutoMesh_VolumeBoundaryMaker3D)

AutoMesh_VolumeBoundaryMaker3D *AutoMesh_VolumeBoundaryMaker3D_New 
(AutoMesh_VolumeFace3D *face)
{
  AutoMesh_VolumeBoundaryMaker3D *maker;
  AutoMesh_PolygonType polygonType;
  int nElementNodes;
  
  maker = AutoMesh_VolumeBoundaryMaker3D_NewImplementation ();
  
  maker->face = face;

  maker->volumeNode = face->element->node;
  maker->volumeNodes_surfaceNodeId = NULL;
  
  maker->volumeElement = face->element;

  maker->surfaceNode = AutoMesh_Node3D_New ();
  assert(maker->surfaceNode != NULL);
  maker->surfaceNodes_volumeNodeId = NULL;

  /* NEED TO REDEFINE */
  /* vertex node only */
  switch (maker->volumeElement->polyhedronType) {
  case AUTOMESH_POLYHEDRON_TETRA:
    polygonType = AUTOMESH_POLYGON_TRI;
    nElementNodes = 3;
    break;
  case AUTOMESH_POLYHEDRON_HEXA:
    polygonType = AUTOMESH_POLYGON_QUAD;
    nElementNodes = 4;
    break;
  case AUTOMESH_POLYHEDRON_PRISM:
    /* NEED TO REDEFINE */
    assert(0);
    break;
  default:
    assert(0);
    break;
  }

  maker->surfaceElement = AutoMesh_SurfaceElement3D_New 
    (maker->surfaceNode, polygonType, nElementNodes);
  assert(maker->surfaceElement != NULL);
  maker->surfaceElements_volumeFaceId = NULL;
  maker->surfaceElements_volumeElementId = NULL;
  maker->surfaceElements_volumeElementFaceId = NULL;

  return maker;
}

void AutoMesh_VolumeBoundaryMaker3D_delete 
(AutoMesh_VolumeBoundaryMaker3D *self)
{
  if (self->volumeNodes_surfaceNodeId != NULL) {
    AutoRes_Table_removeColumn 
      (self->volumeNode->nodes, 
       AutoRes_Table_columnIdOfColumn 
       (self->volumeNode->nodes, 
	self->volumeNodes_surfaceNodeId));
    self->volumeNodes_surfaceNodeId = NULL;
  }

  if (self->surfaceNodes_volumeNodeId != NULL) {
    AutoRes_Table_removeColumn 
      (self->surfaceNode->nodes, 
       AutoRes_Table_columnIdOfColumn 
       (self->surfaceNode->nodes, 
	self->surfaceNodes_volumeNodeId));
    self->surfaceNodes_volumeNodeId = NULL;
  }

  if (self->surfaceElements_volumeFaceId != NULL) {
    AutoRes_Table_removeColumn 
      (self->surfaceElement->elements, 
       AutoRes_Table_columnIdOfColumn 
       (self->surfaceElement->elements, 
	self->surfaceElements_volumeFaceId));
    self->surfaceElements_volumeFaceId = NULL;
  }

  if (self->surfaceElements_volumeElementId != NULL) {
    AutoRes_Table_removeColumn 
      (self->surfaceElement->elements, 
       AutoRes_Table_columnIdOfColumn 
       (self->surfaceElement->elements, 
	self->surfaceElements_volumeElementId));
    self->surfaceElements_volumeElementId = NULL;
  }

  if (self->surfaceElements_volumeElementFaceId != NULL) {
    AutoRes_Table_removeColumn 
      (self->surfaceElement->elements, 
       AutoRes_Table_columnIdOfColumn 
       (self->surfaceElement->elements, 
	self->surfaceElements_volumeElementFaceId));
    self->surfaceElements_volumeElementFaceId = NULL;
  }

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

static int /* boolean */
AllocateElement
(AutoMesh_VolumeBoundaryMaker3D *self)
{
  int nFaces = AutoMesh_VolumeFace3D_nFaces (self->face);
  int nSurfaceElements;
  int iFace;
  int flag;

  assert(AutoMesh_SurfaceElement3D_nElements (self->surfaceElement) == 0);

  nSurfaceElements = 0;
  for (iFace = 0; iFace < nFaces; iFace++) {
    if (AutoMesh_VolumeFace3D_isOnBoundaryAt (self->face, iFace)) {
      nSurfaceElements++;
    }
  }

  flag = AutoMesh_SurfaceElement3D_allocateElements 
    (self->surfaceElement, nSurfaceElements);
  if (!flag) return 0;

  assert(0 < AutoMesh_SurfaceElement3D_nElements (self->surfaceElement));

  return 1;
}

static int /* boolean */
MakeVolumeFaceId 
(AutoMesh_VolumeBoundaryMaker3D *self)
{
  int nFaces = AutoMesh_VolumeFace3D_nFaces (self->face);
  int surfaceElementId;
  int iFace;
  int flag;

  assert(0 < AutoMesh_SurfaceElement3D_nElements (self->surfaceElement));
  assert(self->surfaceElements_volumeFaceId == NULL);

  self->surfaceElements_volumeFaceId = AutoRes_IntegerColumn_New
    (self->surfaceElement->elements, COLUMN_NAME_VOLUME_FACE_ID);
  assert(self->surfaceElements_volumeFaceId != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->surfaceElement->elements, 
     self->surfaceElements_volumeFaceId);
  if (!flag) return 0;

  surfaceElementId = 0;
  for (iFace = 0; iFace < nFaces; iFace++) {
    if (AutoMesh_VolumeFace3D_isOnBoundaryAt (self->face, iFace)) {
      AUTORES_I_SET(self->surfaceElements_volumeFaceId, surfaceElementId, 
		    iFace);
      surfaceElementId++;
    }
  }

  assert(self->surfaceElements_volumeFaceId != NULL);

  return 1;
}

static int /* boolean */
MakeVolumeElementId 
(AutoMesh_VolumeBoundaryMaker3D *self)
{
  int nSurfaceElements 
    = AutoMesh_SurfaceElement3D_nElements (self->surfaceElement);
  int iSurfaceElement;
  int flag;

  assert(0 < nSurfaceElements);
  assert(self->surfaceElements_volumeFaceId != NULL);
  assert(self->surfaceElements_volumeElementId == NULL);
  assert(self->surfaceElements_volumeElementFaceId == NULL);

  self->surfaceElements_volumeElementId = AutoRes_IntegerColumn_New
    (self->surfaceElement->elements, COLUMN_NAME_VOLUME_ELEMENT_ID);
  assert(self->surfaceElements_volumeElementId != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->surfaceElement->elements, 
     self->surfaceElements_volumeElementId);
  if (!flag) return 0;

  self->surfaceElements_volumeElementFaceId = AutoRes_IntegerColumn_New
    (self->surfaceElement->elements, COLUMN_NAME_VOLUME_ELEMENT_FACE_ID);
  assert(self->surfaceElements_volumeElementFaceId != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->surfaceElement->elements, 
     self->surfaceElements_volumeElementFaceId);
  if (!flag) return 0;

  for (iSurfaceElement = 0; 
       iSurfaceElement < nSurfaceElements; 
       iSurfaceElement++) {
    int faceId = AUTORES_I
      (self->surfaceElements_volumeFaceId, iSurfaceElement);
    int volumeElementId;
    int volumeElementFaceId;
    
    if (AutoMesh_VolumeFace3D_getBoundaryAt 
	(self->face, faceId,
	 &volumeElementId, &volumeElementFaceId)) {
      AUTORES_I_SET
	(self->surfaceElements_volumeElementId, iSurfaceElement, 
	 volumeElementId);
      AUTORES_I_SET
	(self->surfaceElements_volumeElementFaceId, iSurfaceElement, 
	 volumeElementFaceId);
    }
  }

  assert(self->surfaceElements_volumeElementId != NULL);
  assert(self->surfaceElements_volumeElementFaceId != NULL);

  return 1;
}

static int /* boolean */
MakeSurfaceNodeId 
(AutoMesh_VolumeBoundaryMaker3D *self)
{
  AutoMesh_PolyhedronType polyhedronType
    = self->volumeElement->polyhedronType;
  int nVolumeNodes = AutoMesh_Node3D_nNodes (self->volumeNode);
  int nSurfaceElements 
    = AutoMesh_SurfaceElement3D_nElements (self->surfaceElement);
  int iVolumeNode;
  int iSurfaceElement;
  int iFaceNode;
  int surfaceNodeId;
  int flag;

  assert(0 < nSurfaceElements);
  assert(self->surfaceElements_volumeFaceId != NULL);
  assert(self->volumeNodes_surfaceNodeId == NULL);

  self->volumeNodes_surfaceNodeId = AutoRes_IntegerColumn_New
    (self->volumeNode->nodes, COLUMN_NAME_SURFACE_NODE_ID);
  assert(self->volumeNodes_surfaceNodeId != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->volumeNode->nodes, 
     self->volumeNodes_surfaceNodeId);
  if (!flag) return 0;

  for (iVolumeNode = 0; iVolumeNode < nVolumeNodes; iVolumeNode++) {
    AUTORES_I_SET(self->volumeNodes_surfaceNodeId, iVolumeNode, -1);
  }

  for (iSurfaceElement = 0; 
       iSurfaceElement < nSurfaceElements; 
       iSurfaceElement++) {
    int faceId = AUTORES_I
      (self->surfaceElements_volumeFaceId, iSurfaceElement);
    int volumeElementId;
    int volumeElementFaceId;
    int nFaceNodes;
    
    flag = AutoMesh_VolumeFace3D_getBoundaryAt 
      (self->face, faceId,
       &volumeElementId, &volumeElementFaceId);
    assert(flag);
    
    /* NEED TO REDEFINE */
    /* vertex node only */
    nFaceNodes = AutoMesh_Polyhedron_NFaceVertexs 
      (polyhedronType, volumeElementFaceId);
    
    for (iFaceNode = 0; iFaceNode < nFaceNodes; iFaceNode++) {
      int elementNodeNum = AutoMesh_Polyhedron_FaceVertex 
	(polyhedronType, volumeElementFaceId, iFaceNode);
      int volumeNodeId = AUTORES_I1D(self->volumeElement->elements_nodeIds, 
				     volumeElementId, elementNodeNum);

      AUTORES_I_SET(self->volumeNodes_surfaceNodeId, volumeNodeId, 0);
    }
  }

  surfaceNodeId = 0;
  for (iVolumeNode = 0; iVolumeNode < nVolumeNodes; iVolumeNode++) {
    if (AUTORES_I(self->volumeNodes_surfaceNodeId, iVolumeNode) == 0) {
      AUTORES_I_SET(self->volumeNodes_surfaceNodeId, iVolumeNode, 
		    surfaceNodeId);
      surfaceNodeId++;
    }
  }

  assert(self->volumeNodes_surfaceNodeId != NULL);

  return 1;
}

static int /* boolean */
AllocateNode
(AutoMesh_VolumeBoundaryMaker3D *self)
{
  int nVolumeNodes = AutoMesh_Node3D_nNodes (self->volumeNode);
  int nSurfaceNodes;
  int iVolumeNode;
  int flag;

  assert(self->volumeNodes_surfaceNodeId != NULL);
  assert(AutoMesh_Node3D_nNodes (self->surfaceNode) == 0);

  nSurfaceNodes = 0;
  for (iVolumeNode = 0; iVolumeNode < nVolumeNodes; iVolumeNode++) {
    if (AUTORES_I(self->volumeNodes_surfaceNodeId, iVolumeNode) != -1) {
      nSurfaceNodes++;
    }
  }

  flag = AutoMesh_Node3D_allocateNodes 
    (self->surfaceNode, nSurfaceNodes);
  if (!flag) return 0;

  assert(0 < AutoMesh_Node3D_nNodes (self->surfaceNode));

  return 1;
}

static int /* boolean */
MakeVolumeNodeId 
(AutoMesh_VolumeBoundaryMaker3D *self)
{
  int nVolumeNodes = AutoMesh_Node3D_nNodes (self->volumeNode);
  int iVolumeNode;
  int flag;

  assert(0 < AutoMesh_Node3D_nNodes (self->surfaceNode));
  assert(self->surfaceNodes_volumeNodeId == NULL);

  self->surfaceNodes_volumeNodeId = AutoRes_IntegerColumn_New
    (self->surfaceNode->nodes, COLUMN_NAME_VOLUME_NODE_ID);
  assert(self->surfaceNodes_volumeNodeId != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->surfaceNode->nodes, 
     self->surfaceNodes_volumeNodeId);
  if (!flag) return 0;

  for (iVolumeNode = 0; iVolumeNode < nVolumeNodes; iVolumeNode++) {
    int surfaceNodeId = AUTORES_I
      (self->volumeNodes_surfaceNodeId, iVolumeNode);

    if (surfaceNodeId != -1) {
      AUTORES_I_SET(self->surfaceNodes_volumeNodeId, surfaceNodeId, 
		    iVolumeNode);
    }
  }

  assert(self->surfaceNodes_volumeNodeId != NULL);

  return 1;
}

static int /* boolean */
MakeNode 
(AutoMesh_VolumeBoundaryMaker3D *self)
{
  int nSurfaceNodes = AutoMesh_Node3D_nNodes (self->surfaceNode);
  int iSurfaceNode;
  int flag;

  assert(0 < nSurfaceNodes);
  assert(self->surfaceNodes_volumeNodeId != NULL);
  assert(self->surfaceNode->nodes_coord == NULL);

  flag = AutoMesh_Node3D_makeCoordColumn (self->surfaceNode);
  if (!flag) return 0;

  for (iSurfaceNode = 0; iSurfaceNode < nSurfaceNodes; iSurfaceNode++) {
    int iVolumeNodeId = AUTORES_I
      (self->surfaceNodes_volumeNodeId, iSurfaceNode);
    double x, y, z;

    AutoMesh_Node3D_getCoordAt (self->volumeNode, iVolumeNodeId,
				&x, &y, &z);
    AutoMesh_Node3D_setCoordAt (self->surfaceNode, iSurfaceNode,
				x, y, z);
  }

  assert(self->surfaceNode->nodes_coord != NULL);

  return 1;
}

static int /* boolean */
MakeElement 
(AutoMesh_VolumeBoundaryMaker3D *self)
{
  AutoMesh_PolyhedronType polyhedronType
    = self->volumeElement->polyhedronType;
  int nSurfaceElements 
    = AutoMesh_SurfaceElement3D_nElements (self->surfaceElement);
  int iSurfaceElement;
  int iFaceNode;
  int flag;

  assert(0 < nSurfaceElements);
  assert(self->surfaceElements_volumeFaceId != NULL);
  assert(self->volumeNodes_surfaceNodeId != NULL);
  assert(self->surfaceElement->elements_nodeIds == NULL);

  flag = AutoMesh_SurfaceElement3D_makeNodeIdsColumn (self->surfaceElement);
  if (!flag) return 0;

  for (iSurfaceElement = 0; 
       iSurfaceElement < nSurfaceElements; 
       iSurfaceElement++) {
    int faceId = AUTORES_I
      (self->surfaceElements_volumeFaceId, iSurfaceElement);
    int volumeElementId;
    int volumeElementFaceId;
    int nFaceNodes;
    
    flag = AutoMesh_VolumeFace3D_getBoundaryAt 
      (self->face, faceId,
       &volumeElementId, &volumeElementFaceId);
    assert(flag);
    
    /* NEED TO REDEFINE */
    /* vertex node only */
    nFaceNodes = AutoMesh_Polyhedron_NFaceVertexs 
      (polyhedronType, volumeElementFaceId);
    
    for (iFaceNode = 0; iFaceNode < nFaceNodes; iFaceNode++) {
      int elementNodeNum = AutoMesh_Polyhedron_FaceVertex 
	(polyhedronType, volumeElementFaceId, iFaceNode);
      int volumeNodeId = AutoMesh_VolumeElement3D_nodeIdAt
	(self->volumeElement, volumeElementId, elementNodeNum);
      int surfaceNodeId = AUTORES_I
	(self->volumeNodes_surfaceNodeId, volumeNodeId);
      
      AutoMesh_SurfaceElement3D_setNodeIdAt
	(self->surfaceElement, iSurfaceElement, iFaceNode, 
	 surfaceNodeId);

      /* NEED TO REDEFINE */
      /* what if prism element ? */
    }
  }

  assert(self->surfaceElement->elements_nodeIds != NULL);

  return 1;
}

int /* boolean */
AutoMesh_VolumeBoundaryMaker3D_do 
(AutoMesh_VolumeBoundaryMaker3D *self,
 int nodeFlag,
 int elementFlag,
 int volumeNodeIdFlag,
 int volumeElementIdFlag)
{
  int flag;

  flag = AllocateElement (self);
  if (!flag) return 0;

  flag = MakeVolumeFaceId (self);
  if (!flag) return 0;

  if (volumeElementIdFlag) {
    flag = MakeVolumeElementId (self);
    if (!flag) return 0;
  }

  flag = MakeSurfaceNodeId (self);
  if (!flag) return 0;
  
  flag = AllocateNode (self);
  if (!flag) return 0;

  if (volumeNodeIdFlag) {
    flag = MakeVolumeNodeId (self);
    if (!flag) return 0;

    if (nodeFlag) {
      flag = MakeNode (self);
      if (!flag) return 0;
    }
  }

  if (elementFlag) {
    flag = MakeElement (self);
    if (!flag) return 0;
  }

  return 1;
}

int AutoMesh_VolumeBoundaryMaker3D_surfaceNodeIdAtVolumeNode
(AutoMesh_VolumeBoundaryMaker3D *self,
 int volumeNodeId)
{
  return AUTORES_I(self->volumeNodes_surfaceNodeId, volumeNodeId);
}

int AutoMesh_VolumeBoundaryMaker3D_volumeNodeIdAtSurfaceNode
(AutoMesh_VolumeBoundaryMaker3D *self,
 int surfaceNodeId)
{
  return AUTORES_I(self->surfaceNodes_volumeNodeId, surfaceNodeId);
}

int AutoMesh_VolumeBoundaryMaker3D_volumeFaceIdAtSurfaceElement
(AutoMesh_VolumeBoundaryMaker3D *self,
 int surfaceElementId)
{
  return AUTORES_I(self->surfaceElements_volumeFaceId, surfaceElementId);
}

void AutoMesh_VolumeBoundaryMaker3D_getVolumeElementAtSurfaceElement
(AutoMesh_VolumeBoundaryMaker3D *self,
 int surfaceElementId,
 int *volumeElementId_OUT,
 int *volumeElementFaceId_OUT)
{
  *volumeElementId_OUT 
    = AUTORES_I(self->surfaceElements_volumeElementId, surfaceElementId);
  *volumeElementFaceId_OUT 
    = AUTORES_I(self->surfaceElements_volumeElementFaceId, surfaceElementId);
}


