/* automesh_volume_divider3d.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"
#include "automesh_shape.h"
#include "automesh_node3d.h"
#include "automesh_volume_element3d.h"
#include "automesh_volume_edge3d.h"
#include "automesh_volume_divider3d.h"



#define COLUMN_NAME_SRC_NODE_IDS "AutoMesh:srcNodeIds"
#define COLUMN_NAME_SRC_ELEMENT_ID "AutoMesh:srcElementId"
#define COLUMN_NAME_SRC_ELEMENT_PART_ID "AutoMesh:srcElementPartId"
#define COLUMN_NAME_DEST_NODE_ID "AutoMesh:destNodeId"
#define COLUMN_NAME_DEST_ELEMENT_IDS "AutoMesh:destElementIds"



/* VolumeDivider3D */

AUTORES_IMPLEMENT_HANDLE(AutoMesh_VolumeDivider3D)

AutoMesh_VolumeDivider3D *AutoMesh_VolumeDivider3D_New 
(AutoMesh_VolumeEdge3D *edge)
{
  AutoMesh_VolumeDivider3D *divider;
  
  divider = AutoMesh_VolumeDivider3D_NewImplementation ();
  
  divider->edge = edge;

  divider->srcNode = edge->element->node;
  divider->srcNodes_destNodeId = NULL;

  divider->srcElement = edge->element;
  divider->srcElements_destElementIds = NULL;

  divider->destNode = AutoMesh_Node3D_New ();
  assert(divider->destNode != NULL);
  divider->destNodes_srcNodeIds = NULL;

  divider->destElement = AutoMesh_VolumeElement3D_New 
    (divider->destNode, 
     divider->srcElement->polyhedronType,
     divider->srcElement->nElementNodes);
  assert(divider->destElement != NULL);
  divider->destElements_srcElementId = NULL;
  divider->destElements_srcElementPartId = NULL;

  return divider;
}

void AutoMesh_VolumeDivider3D_delete 
(AutoMesh_VolumeDivider3D *self)
{
  if (self->srcNodes_destNodeId != NULL) {
    AutoRes_Table_removeColumn 
      (self->srcNode->nodes, 
       AutoRes_Table_columnIdOfColumn 
       (self->srcNode->nodes, 
	self->srcNodes_destNodeId));
    self->srcNodes_destNodeId = NULL;
  }

  if (self->srcElements_destElementIds != NULL) {
    AutoRes_Table_removeColumn 
      (self->srcElement->elements, 
       AutoRes_Table_columnIdOfColumn 
       (self->srcElement->elements, 
	self->srcElements_destElementIds));
    self->srcElements_destElementIds = NULL;
  }

  if (self->destNodes_srcNodeIds != NULL) {
    AutoRes_Table_removeColumn 
      (self->destNode->nodes, 
       AutoRes_Table_columnIdOfColumn 
       (self->destNode->nodes, 
	self->destNodes_srcNodeIds));
    self->destNodes_srcNodeIds = NULL;
  }

  if (self->destElements_srcElementId != NULL) {
    AutoRes_Table_removeColumn 
      (self->destElement->elements, 
       AutoRes_Table_columnIdOfColumn 
       (self->destElement->elements, 
	self->destElements_srcElementId));
    self->destElements_srcElementId = NULL;
  }

  if (self->destElements_srcElementPartId != NULL) {
    AutoRes_Table_removeColumn 
      (self->destElement->elements, 
       AutoRes_Table_columnIdOfColumn 
       (self->destElement->elements, 
	self->destElements_srcElementPartId));
    self->destElements_srcElementPartId = NULL;
  }

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

static int /* boolean */
Initialize
(AutoMesh_VolumeDivider3D *self,
 int srcNodeIdsFlag,
 int srcElementIdFlag,
 int destNodeIdFlag,
 int destElementIdsFlag)
{
  int nSrcNodes = AutoRes_Table_nRecords (self->srcNode->nodes);
  int nSrcElements = AutoRes_Table_nRecords (self->srcElement->elements);
  int flag;

  flag = AutoMesh_Node3D_reserveNodes 
    (self->destNode, nSrcNodes * 8);
  if (!flag) return 0;

  flag = AutoMesh_Node3D_makeCoordColumn (self->destNode);
  if (!flag) return 0;

  flag = AutoMesh_VolumeElement3D_reserveElements 
    (self->destElement, nSrcElements * 8);
  if (!flag) return 0;

  flag = AutoMesh_VolumeElement3D_makeNodeIdsColumn (self->destElement);
  if (!flag) return 0;

  if (srcNodeIdsFlag) {
    self->destNodes_srcNodeIds = AutoRes_Integer1DColumn_New
      (self->destNode->nodes, COLUMN_NAME_SRC_NODE_IDS, 2);
    assert(self->destNodes_srcNodeIds != NULL);
    flag = AutoRes_Table_addInteger1DColumn 
      (self->destNode->nodes, 
       self->destNodes_srcNodeIds);
    if (!flag) return 0;
  }

  if (srcElementIdFlag) {
    self->destElements_srcElementId = AutoRes_IntegerColumn_New
      (self->destElement->elements, COLUMN_NAME_SRC_ELEMENT_ID);
    assert(self->destElements_srcElementId != NULL);
    flag = AutoRes_Table_addIntegerColumn 
      (self->destElement->elements, 
       self->destElements_srcElementId);
    if (!flag) return 0;
    
    self->destElements_srcElementPartId = AutoRes_IntegerColumn_New
      (self->destElement->elements, COLUMN_NAME_SRC_ELEMENT_PART_ID);
    assert(self->destElements_srcElementPartId != NULL);
    flag = AutoRes_Table_addIntegerColumn 
      (self->destElement->elements, 
     self->destElements_srcElementPartId);
    if (!flag) return 0;
  }

  if (destNodeIdFlag) {
    self->srcNodes_destNodeId = AutoRes_IntegerColumn_New
      (self->srcNode->nodes, COLUMN_NAME_DEST_NODE_ID);
    assert(self->srcNodes_destNodeId != NULL);
    flag = AutoRes_Table_addIntegerColumn 
      (self->srcNode->nodes, 
       self->srcNodes_destNodeId);
    if (!flag) return 0;
  }

  if (destElementIdsFlag) {
    self->srcElements_destElementIds = AutoRes_Integer1DColumn_New
      (self->srcElement->elements, COLUMN_NAME_DEST_ELEMENT_IDS, 8);
    assert(self->srcElements_destElementIds != NULL);
    flag = AutoRes_Table_addInteger1DColumn 
      (self->srcElement->elements, 
       self->srcElements_destElementIds);
    if (!flag) return 0;
  }

  return 1;
}

static void MakeNode
(AutoMesh_VolumeDivider3D *self)
{
  int nEdges = AutoMesh_VolumeEdge3D_nEdges (self->edge);
  int nSrcNodes = AutoMesh_Node3D_nNodes (self->srcNode);
  int iEdge;
  int iSrcNode;

  for (iSrcNode = 0; iSrcNode < nSrcNodes; iSrcNode++) {
    int destNodeId;
    double x, y, z;
    
    destNodeId = AutoMesh_Node3D_addNode (self->destNode);
    assert(destNodeId != -1);
    AutoMesh_Node3D_getCoordAt (self->srcNode, iSrcNode, 
				&x, &y, &z);
    AutoMesh_Node3D_setCoordAt (self->destNode, destNodeId, x, y, z);
  }

  if (self->srcNodes_destNodeId != NULL) {
    for (iSrcNode = 0; iSrcNode < nSrcNodes; iSrcNode++) {
      AUTORES_I_SET(self->srcNodes_destNodeId, iSrcNode, iSrcNode);
    }
  }

  if (self->destNodes_srcNodeIds != NULL) {
    /* 1st node : [0] = nodeId, [1] = -1 */
    for (iSrcNode = 0; iSrcNode < nSrcNodes; iSrcNode++) {
      AUTORES_I1D_SET(self->destNodes_srcNodeIds, iSrcNode, 0, 
		      iSrcNode);
      AUTORES_I1D_SET(self->destNodes_srcNodeIds, iSrcNode, 1, 
		      -1);
    }
  }

  for (iEdge = 0; iEdge < nEdges; iEdge++) {
    int destNodeId;
    double x, y, z;
    
    destNodeId = AutoMesh_Node3D_addNode (self->destNode);
    assert(destNodeId != -1);
    AutoMesh_VolumeEdge3D_getMidPointAt (self->edge, iEdge,
					 &x, &y, &z);
    AutoMesh_Node3D_setCoordAt (self->destNode, destNodeId, x, y, z);
  }

  if (self->destNodes_srcNodeIds != NULL) {
    /* 2nd node : [0] = nodeId0, [1] = nodeId1 */
    for (iEdge = 0; iEdge < nEdges; iEdge++) {
      int destNodeId = iEdge + nSrcNodes;
      int srcNodeId0, srcNodeId1;
      
      AutoMesh_VolumeEdge3D_getVertexNodeIdsAt 
	(self->edge, iEdge,
	 &srcNodeId0, &srcNodeId1);
      AUTORES_I1D_SET(self->destNodes_srcNodeIds, destNodeId, 0, 
		      srcNodeId0);
      AUTORES_I1D_SET(self->destNodes_srcNodeIds, destNodeId, 1, 
		      srcNodeId1);
    }
  }
}

/* NEED TO REDEFINE */
/* tetra, 1st order only */
static void AddElement 
(AutoMesh_VolumeDivider3D *self,
 int srcElementId,
 int srcElementPartId,
 int destNodeId0, int destNodeId1, int destNodeId2, int destNodeId3)
{
  int destElementId;

  destElementId = AutoMesh_VolumeElement3D_addElement (self->destElement);
  assert(destElementId != -1);
  
  AutoMesh_VolumeElement3D_setNodeIdAt
    (self->destElement, destElementId, 0, destNodeId0);
  AutoMesh_VolumeElement3D_setNodeIdAt
    (self->destElement, destElementId, 1, destNodeId1);
  AutoMesh_VolumeElement3D_setNodeIdAt
    (self->destElement, destElementId, 2, destNodeId2);
  AutoMesh_VolumeElement3D_setNodeIdAt
    (self->destElement, destElementId, 3, destNodeId3);
  
  if (self->srcElements_destElementIds != NULL) {
    AUTORES_I1D_SET(self->srcElements_destElementIds, 
		    srcElementId, srcElementPartId,
		    destElementId);
  }
  
  if (self->destElements_srcElementId != NULL) {
    AUTORES_I_SET(self->destElements_srcElementId, destElementId, 
		  srcElementId);
    AUTORES_I_SET(self->destElements_srcElementPartId, destElementId, 
		  srcElementPartId);
  }
}

static void MakeElement
(AutoMesh_VolumeDivider3D *self)
{
  int nSrcNodes = AutoMesh_Node3D_nNodes (self->srcNode);
  int nSrcElements 
    = AutoMesh_VolumeElement3D_nElements (self->srcElement);
  int iSrcElement;

  for (iSrcElement = 0; 
       iSrcElement < nSrcElements; 
       iSrcElement++) {
    int nodeId0 = AutoMesh_VolumeElement3D_nodeIdAt
      (self->srcElement, iSrcElement, 0);
    int nodeId1 = AutoMesh_VolumeElement3D_nodeIdAt
      (self->srcElement, iSrcElement, 1);
    int nodeId2 = AutoMesh_VolumeElement3D_nodeIdAt
      (self->srcElement, iSrcElement, 2);
    int nodeId3 = AutoMesh_VolumeElement3D_nodeIdAt
      (self->srcElement, iSrcElement, 3);

    int edgeId0 = AutoMesh_VolumeEdge3D_edgeIdAtElement
      (self->edge, iSrcElement, 0);
    int edgeId1 = AutoMesh_VolumeEdge3D_edgeIdAtElement
      (self->edge, iSrcElement, 1);
    int edgeId2 = AutoMesh_VolumeEdge3D_edgeIdAtElement
      (self->edge, iSrcElement, 2);
    int edgeId3 = AutoMesh_VolumeEdge3D_edgeIdAtElement
      (self->edge, iSrcElement, 3);
    int edgeId4 = AutoMesh_VolumeEdge3D_edgeIdAtElement
      (self->edge, iSrcElement, 4);
    int edgeId5 = AutoMesh_VolumeEdge3D_edgeIdAtElement
      (self->edge, iSrcElement, 5);

    int edgeNodeId0 = edgeId0 + nSrcNodes;
    int edgeNodeId1 = edgeId1 + nSrcNodes;
    int edgeNodeId2 = edgeId2 + nSrcNodes;
    int edgeNodeId3 = edgeId3 + nSrcNodes;
    int edgeNodeId4 = edgeId4 + nSrcNodes;
    int edgeNodeId5 = edgeId5 + nSrcNodes;

    double dist04 = AutoMesh_Node3D_distance 
      (self->destNode, edgeNodeId0, edgeNodeId4);
    double dist15 = AutoMesh_Node3D_distance  
      (self->destNode, edgeNodeId1, edgeNodeId5);
    double dist23 = AutoMesh_Node3D_distance   
      (self->destNode, edgeNodeId2, edgeNodeId3);
    
    AddElement (self, iSrcElement, 0,
		edgeNodeId2, edgeNodeId1, edgeNodeId0, nodeId0);
    AddElement (self, iSrcElement, 1,
		edgeNodeId0, edgeNodeId3, edgeNodeId5, nodeId1);
    AddElement (self, iSrcElement, 2,
		edgeNodeId1, edgeNodeId4, edgeNodeId3, nodeId2);
    AddElement (self, iSrcElement, 3,
		edgeNodeId2, edgeNodeId5, edgeNodeId4, nodeId3);

    if (dist04 < dist15 && dist04 < dist23) {
      /* e0 - e4 */
      AddElement (self, iSrcElement, 4,
		  edgeNodeId0, edgeNodeId4, edgeNodeId1, edgeNodeId2);
      AddElement (self, iSrcElement, 5,
		  edgeNodeId0, edgeNodeId4, edgeNodeId2, edgeNodeId5);
      AddElement (self, iSrcElement, 6,
		  edgeNodeId0, edgeNodeId4, edgeNodeId5, edgeNodeId3);
      AddElement (self, iSrcElement, 7,
		  edgeNodeId0, edgeNodeId4, edgeNodeId3, edgeNodeId1);
    } else if (dist15 < dist04 && dist15 < dist23) {
      /* e1 - e5 */
      AddElement (self, iSrcElement, 4,
		  edgeNodeId1, edgeNodeId5, edgeNodeId4, edgeNodeId2);
      AddElement (self, iSrcElement, 5, 
		  edgeNodeId1, edgeNodeId5, edgeNodeId2, edgeNodeId0);
      AddElement (self, iSrcElement, 6,
		  edgeNodeId1, edgeNodeId5, edgeNodeId0, edgeNodeId3);
      AddElement (self, iSrcElement, 7,
		  edgeNodeId1, edgeNodeId5, edgeNodeId3, edgeNodeId4);
    } else {
      /* e2 - e3 */
      AddElement (self, iSrcElement, 4,
		  edgeNodeId2, edgeNodeId3, edgeNodeId0, edgeNodeId1);
      AddElement (self, iSrcElement, 5,
		  edgeNodeId2, edgeNodeId3, edgeNodeId1, edgeNodeId4);
      AddElement (self, iSrcElement, 6,
		  edgeNodeId2, edgeNodeId3, edgeNodeId4, edgeNodeId5);
      AddElement (self, iSrcElement, 7,
		  edgeNodeId2, edgeNodeId3, edgeNodeId5, edgeNodeId0);
    }
  }
}

int /* boolean */
AutoMesh_VolumeDivider3D_do 
(AutoMesh_VolumeDivider3D *self,
 int srcNodeIdsFlag,
 int srcElementIdFlag,
 int destNodeIdFlag,
 int destElementIdsFlag)
{
  int flag;

  /* NEED TO REDEFINE */
  /* tetra, 1st order only */
  assert(self->srcElement->polyhedronType == AUTOMESH_POLYHEDRON_TETRA);
  assert(self->srcElement->nElementNodes == 4);

  flag = Initialize (self, 
		     srcNodeIdsFlag, srcElementIdFlag,
		     destNodeIdFlag, destElementIdsFlag);
  if (!flag) return 0;

  MakeNode (self);
  MakeElement (self);

  return 1;
}

int AutoMesh_VolumeDivider3D_destNodeIdAtSrcNode
(AutoMesh_VolumeDivider3D *self,
 int srcNodeId)
{
  return AUTORES_I(self->srcNodes_destNodeId, srcNodeId);
}

void AutoMesh_VolumeDivider3D_getDestElementIdsAtSrcElement
(AutoMesh_VolumeDivider3D *self,
 int srcElementId,
 int destElementIds_OUT[8])
{
  int iPart;

  for (iPart = 0; iPart < 8; iPart++) {
    destElementIds_OUT[iPart] 
      = AUTORES_I1D(self->srcElements_destElementIds, srcElementId, iPart);
  }
}

void AutoMesh_VolumeDivider3D_getSrcNodeIdsAtDestNode
(AutoMesh_VolumeDivider3D *self,
 int destNodeId,
 int srcNodeIds_OUT[2])
{
  srcNodeIds_OUT[0] = AUTORES_I1D(self->destNodes_srcNodeIds, destNodeId, 0);
  srcNodeIds_OUT[1] = AUTORES_I1D(self->destNodes_srcNodeIds, destNodeId, 1);
}

void AutoMesh_VolumeDivider3D_getSrcElementAtDestElement
(AutoMesh_VolumeDivider3D *self,
 int destElementId,
 int *srcElementId_OUT,
 int *srcElementPartId_OUT)
{
  *srcElementId_OUT 
    = AUTORES_I(self->destElements_srcElementId, destElementId);
  *srcElementPartId_OUT 
    = AUTORES_I(self->destElements_srcElementPartId, destElementId);
}




