/* automesh_volume_edge2d.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"
#include "automesh_shape.h"
#include "automesh_node2d.h"
#include "automesh_volume_element2d.h"
#include "automesh_volume_edge2d.h"



#define TABLE_NAME_EDGES "AutoMesh:edges"
#define COLUMN_NAME_EDGE_IDS "AutoMesh:edgeIds"
#define COLUMN_NAME_VERTEX_NODE_IDS "AutoMesh:vertexNodeIds"
#define COLUMN_NAME_ELEMENT_IDS "AutoMesh:elementIds"
#define COLUMN_NAME_ELEMENT_EDGE_NUMS "AutoMesh:elementEdgeNums"



/* VolumeEdge2D */

AUTORES_IMPLEMENT_HANDLE(AutoMesh_VolumeEdge2D)

AutoMesh_VolumeEdge2D *AutoMesh_VolumeEdge2D_New 
(AutoMesh_VolumeElement2D *element)
{
  AutoMesh_VolumeEdge2D *edge;

  edge = AutoMesh_VolumeEdge2D_NewImplementation ();

  edge->polygonType = element->polygonType;
  edge->nElementEdges 
    = AutoMesh_Polygon_NEdges (element->polygonType);

  edge->element = element;
  edge->elements_edgeIds = NULL;

  edge->edges = AutoRes_Table_New (TABLE_NAME_EDGES);
  assert(edge->edges != NULL);
  edge->edges_vertexNodeIds = NULL;
  edge->edges_elementIds = NULL;
  edge->edges_elementEdgeNums = NULL;

  return edge;
}

void AutoMesh_VolumeEdge2D_delete 
(AutoMesh_VolumeEdge2D *self)
{
  AutoRes_Table_delete (self->edges);
  self->edges = NULL;
  self->edges_vertexNodeIds = NULL;
  self->edges_elementIds = NULL;
  self->edges_elementEdgeNums = NULL;

  if (self->elements_edgeIds != NULL) {
    AutoRes_Table_removeColumn 
      (self->element->elements, 
       AutoRes_Table_columnIdOfColumn (self->element->elements, 
				       self->elements_edgeIds));
    self->elements_edgeIds = NULL;
  }

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoMesh_VolumeEdge2D_makeEdgeIdsAtElementColumn 
(AutoMesh_VolumeEdge2D *self)
{
  int flag;

  self->elements_edgeIds = AutoRes_Integer1DColumn_New 
    (self->element->elements, COLUMN_NAME_EDGE_IDS, self->nElementEdges);
  assert(self->elements_edgeIds != NULL);
  flag = AutoRes_Table_addInteger1DColumn 
    (self->element->elements, self->elements_edgeIds);
  if (!flag) return 0;

  return 1;
}

int /* boolean */
AutoMesh_VolumeEdge2D_makeVertexNodeIdsColumn 
(AutoMesh_VolumeEdge2D *self)
{
  int flag;

  self->edges_vertexNodeIds = AutoRes_Integer1DColumn_New 
    (self->edges, COLUMN_NAME_VERTEX_NODE_IDS, 2);
  assert(self->edges_vertexNodeIds != NULL);
  flag = AutoRes_Table_addInteger1DColumn 
    (self->edges, self->edges_vertexNodeIds);
  if (!flag) return 0;

  return 1;
}

int /* boolean */
AutoMesh_VolumeEdge2D_makeElementIdsColumn 
(AutoMesh_VolumeEdge2D *self)
{
  int flag;

  self->edges_elementIds = AutoRes_Integer1DColumn_New 
    (self->edges, COLUMN_NAME_ELEMENT_IDS, 2);
  assert(self->edges_elementIds != NULL);
  flag = AutoRes_Table_addInteger1DColumn 
    (self->edges, self->edges_elementIds);
  if (!flag) return 0;
  
  self->edges_elementEdgeNums = AutoRes_Integer1DColumn_New 
    (self->edges, COLUMN_NAME_ELEMENT_EDGE_NUMS, 2);
  assert(self->edges_elementEdgeNums != NULL);
  flag = AutoRes_Table_addInteger1DColumn 
    (self->edges, self->edges_elementEdgeNums);
  if (!flag) return 0;

  return 1;
}

static int FindEdge 
(AutoMesh_VolumeEdge2D *self,
 int nodeId0, int nodeId1,
 AutoRes_ListTable *edgeTable)
{
  int hashTableSize;
  int hashValue;
  int linkId;

  hashTableSize = AutoRes_ListTable_nLists (edgeTable);
  hashValue = (nodeId0 + nodeId1) % hashTableSize;
  for (linkId = AutoRes_ListTable_startLinkId (edgeTable, hashValue);
       linkId != -1;
       linkId = AutoRes_ListTable_nextLinkId (edgeTable, linkId)) {
    int edgeId = AutoRes_ListTable_objectId (edgeTable, linkId);

    int eNodeId0 = AUTORES_I1D(self->edges_vertexNodeIds, edgeId, 0);
    int eNodeId1 = AUTORES_I1D(self->edges_vertexNodeIds, edgeId, 1);
    
    if (AutoMesh_IdIsEqual2 (nodeId0, nodeId1,
			     eNodeId0, eNodeId1)) {
      return edgeId;
    }
  }

  return -1;
}

static int AddEdge 
(AutoMesh_VolumeEdge2D *self,
 int nodeId0, int nodeId1,
 AutoRes_ListTable *edgeTable,
 int elementIdsFlag) 
{
  int edgeId;
  int hashTableSize;
  int hashValue;
  int linkId;

  edgeId = AutoRes_Table_addRecord (self->edges);
  if (edgeId == -1) return -1;

  if (elementIdsFlag) {
    AUTORES_I1D_SET(self->edges_elementIds, edgeId, 0, -1);
    AUTORES_I1D_SET(self->edges_elementIds, edgeId, 1, -1);
    
    AUTORES_I1D_SET(self->edges_elementEdgeNums, edgeId, 0, -1);
    AUTORES_I1D_SET(self->edges_elementEdgeNums, edgeId, 1, -1);
  }

  AUTORES_I1D_SET(self->edges_vertexNodeIds, edgeId, 0, nodeId0);
  AUTORES_I1D_SET(self->edges_vertexNodeIds, edgeId, 1, nodeId1);
  
  hashTableSize = AutoRes_ListTable_nLists (edgeTable);
  hashValue = (nodeId0 + nodeId1) % hashTableSize;
  linkId = AutoRes_ListTable_addObject (edgeTable, hashValue, edgeId);
  if (linkId == -1) return -1;

  return edgeId;
}

static void SetElementEdge 
(AutoMesh_VolumeEdge2D *self,
 int elementId, 
 int elementEdgeNum, 
 int edgeId,
 int elementIdsFlag)
{
  AUTORES_I1D_SET(self->elements_edgeIds, elementId, elementEdgeNum, 
		  edgeId);
  
  if (elementIdsFlag) {
    assert(AUTORES_I1D(self->edges_elementIds, edgeId, 1) == -1);
    if (AUTORES_I1D(self->edges_elementIds, edgeId, 0) == -1) {
      AUTORES_I1D_SET(self->edges_elementIds, edgeId, 0, elementId);
      AUTORES_I1D_SET(self->edges_elementEdgeNums, edgeId, 0, elementEdgeNum);
    } else {
      AUTORES_I1D_SET(self->edges_elementIds, edgeId, 1, elementId);
      AUTORES_I1D_SET(self->edges_elementEdgeNums, edgeId, 1, elementEdgeNum);
    }
  }
}

int /* boolean */
AutoMesh_VolumeEdge2D_extract 
(AutoMesh_VolumeEdge2D *self,
 int elementIdsFlag)
{
  int nElements = AutoMesh_VolumeElement2D_nElements (self->element);
  AutoRes_ListTable *edgeTable;
  int hashTableSize;
  int nAverageLinksPerList;
  int iElement;
  int iElementEdge;
  int flag;

  assert(AutoRes_Table_nRecords (self->edges) == 0);
  assert(self->edges_vertexNodeIds == NULL);
  assert(self->edges_elementIds == NULL);
  assert(self->edges_elementEdgeNums == NULL);
  assert(self->elements_edgeIds == NULL);

  /* NEED TO REDEFINE */
  /* experience-based value */
  flag = AutoRes_Table_reserveRecords (self->edges, nElements * 2);
  if (!flag) return 0;

  flag = AutoMesh_VolumeEdge2D_makeVertexNodeIdsColumn (self);
  if (!flag) return 0;

  if (elementIdsFlag) {
    flag = AutoMesh_VolumeEdge2D_makeElementIdsColumn (self);
    if (!flag) return 0;
  }

  flag = AutoMesh_VolumeEdge2D_makeEdgeIdsAtElementColumn (self);
  if (!flag) return 0;

  edgeTable = AutoRes_ListTable_New ();
  assert(edgeTable != NULL);

  /* NEED TO REDEFINE */
  /* experience-based value */
  hashTableSize = nElements * 2;
  nAverageLinksPerList = 2;
  flag = AutoRes_ListTable_reserve 
    (edgeTable, hashTableSize, nAverageLinksPerList);
  if (!flag) return 0;

  for (iElement = 0; iElement < nElements; iElement++) {
    for (iElementEdge = 0; 
	 iElementEdge < self->nElementEdges; 
	 iElementEdge++) {
      int elementNodeNum0 = AutoMesh_Polygon_EdgeVertex
	(self->polygonType, iElementEdge, 0);
      int elementNodeNum1 = AutoMesh_Polygon_EdgeVertex
	(self->polygonType, iElementEdge, 1);

      int nodeId0 = AutoMesh_VolumeElement2D_nodeIdAt
	(self->element, iElement, elementNodeNum0);
      int nodeId1 = AutoMesh_VolumeElement2D_nodeIdAt
	(self->element, iElement, elementNodeNum1);

      int edgeId = FindEdge (self, nodeId0, nodeId1, edgeTable);

      if (edgeId == -1) {
	edgeId = AddEdge (self, nodeId0, nodeId1, edgeTable,
			  elementIdsFlag);
	if (edgeId == -1) return 0;
      }
      SetElementEdge (self, iElement, iElementEdge, edgeId,
		      elementIdsFlag);
    }
  }

  AutoRes_ListTable_delete (edgeTable);

  assert(0 < AutoRes_Table_nRecords (self->edges));
  assert(self->edges_vertexNodeIds != NULL);
  if (elementIdsFlag) {
    assert(self->edges_elementIds != NULL);
    assert(self->edges_elementEdgeNums != NULL);
  }
  assert(self->elements_edgeIds != NULL);

  return 1;
}

void AutoMesh_VolumeEdge2D_purgeWorkArea 
(AutoMesh_VolumeEdge2D *self)
{
  if (self->edges_vertexNodeIds != NULL) {
    AutoRes_Table_removeColumn 
      (self->edges, 
       AutoRes_Table_columnIdOfColumn 
       (self->edges, 
	self->edges_vertexNodeIds));
    self->edges_vertexNodeIds = NULL;
  }

  assert(self->edges_vertexNodeIds == NULL);
}

int AutoMesh_VolumeEdge2D_collectNeighbor 
(AutoMesh_VolumeEdge2D *self)
{
  int nElements = AutoMesh_VolumeElement2D_nElements (self->element);
  int iElement;
  int iElementEdge;
  int flag;
  
  assert(self->edges_elementIds != NULL);
  assert(self->elements_edgeIds != NULL);
  assert(self->element->elements_neighborIds == NULL);

  flag = AutoMesh_VolumeElement2D_makeNeighborIdsColumn (self->element);
  if (!flag) return 0;

  for (iElement = 0; iElement < nElements; iElement++) {
    for (iElementEdge = 0; 
	 iElementEdge < self->nElementEdges; 
	 iElementEdge++) {
      int edgeId = AUTORES_I1D(self->elements_edgeIds, iElement, iElementEdge);
      int neighborId;
      
      if (AUTORES_I1D(self->edges_elementIds, edgeId, 0) == iElement) {
	neighborId = AUTORES_I1D(self->edges_elementIds, edgeId, 1);
      } else {
	assert(AUTORES_I1D(self->edges_elementIds, edgeId, 1) == iElement);
	neighborId = AUTORES_I1D(self->edges_elementIds, edgeId, 0);
      }
      AutoMesh_VolumeElement2D_setNeighborIdAt
	(self->element, iElement, iElementEdge, 
	 neighborId);
    }
  }

  assert(self->element->elements_neighborIds != NULL);

  return 1;
}

int AutoMesh_VolumeEdge2D_nEdges 
(AutoMesh_VolumeEdge2D *self)
{
  return AutoRes_Table_nRecords (self->edges);
}

void AutoMesh_VolumeEdge2D_getEdgeAt 
(AutoMesh_VolumeEdge2D *self,
 int edgeId,
 int elementIds_OUT[2],
 int elementEdgeNums_OUT[2])
{
  elementIds_OUT[0] = AUTORES_I1D(self->edges_elementIds, edgeId, 0);
  elementIds_OUT[1] = AUTORES_I1D(self->edges_elementIds, edgeId, 1);

  elementEdgeNums_OUT[0] = AUTORES_I1D(self->edges_elementEdgeNums, edgeId, 0);
  elementEdgeNums_OUT[1] = AUTORES_I1D(self->edges_elementEdgeNums, edgeId, 1);
}

int /* boolean */
AutoMesh_VolumeEdge2D_isOnBoundaryAt 
(AutoMesh_VolumeEdge2D *self,
 int edgeId)
{
  int elementId1 = AUTORES_I1D(self->edges_elementIds, edgeId, 1);
  
  return (elementId1 == -1);
}

int /* boolean */
AutoMesh_VolumeEdge2D_getBoundaryAt 
(AutoMesh_VolumeEdge2D *self,
 int edgeId,
 int *elementId_OUT,
 int *elementEdgeNum_OUT)
{
  int elementId1 = AUTORES_I1D(self->edges_elementIds, edgeId, 1);

  if (elementId1 == -1) {
    *elementId_OUT = AUTORES_I1D(self->edges_elementIds, edgeId, 0);
    *elementEdgeNum_OUT = AUTORES_I1D(self->edges_elementEdgeNums, edgeId, 0);
    return 1;
  }

  return 0;
}

void AutoMesh_VolumeEdge2D_getVertexNodeIdsAt 
(AutoMesh_VolumeEdge2D *self,
 int edgeId,
 int *nodeId0_OUT, int *nodeId1_OUT)
{
  assert(self->edges_vertexNodeIds != NULL);

  *nodeId0_OUT = AUTORES_I1D(self->edges_vertexNodeIds, edgeId, 0);
  *nodeId1_OUT = AUTORES_I1D(self->edges_vertexNodeIds, edgeId, 1);
}

void AutoMesh_VolumeEdge2D_getMidPointAt 
(AutoMesh_VolumeEdge2D *self,
 int edgeId,
 double *x_OUT, double *y_OUT)
{
  AutoMesh_Node2D *node = self->element->node;

  assert(self->edges_vertexNodeIds != NULL);

  /* NEED TO REDEFINE */
  /* assuming straight line */
	 
  {
    int nodeId0 = AUTORES_I1D(self->edges_vertexNodeIds, edgeId, 0);
    int nodeId1 = AUTORES_I1D(self->edges_vertexNodeIds, edgeId, 1);
    double x0, y0;
    double x1, y1;

    AutoMesh_Node2D_getCoordAt (node, nodeId0,
				&x0, &y0);
    AutoMesh_Node2D_getCoordAt (node, nodeId1,
				&x1, &y1);
    *x_OUT = (x0 + x1) / 2;
    *y_OUT = (y0 + y1) / 2;
  }
}

int AutoMesh_VolumeEdge2D_edgeIdAtElement 
(AutoMesh_VolumeEdge2D *self,
 int elementId,
 int elementEdgeNum)
{
  return AUTORES_I1D(self->elements_edgeIds, elementId, elementEdgeNum);
}










