/* automesh_volume_element2d.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"
#include "automesh_shape.h"
#include "automesh_node2d.h"
#include "automesh_node_scalar2d.h"
#include "automesh_node_vector2d.h"
#include "automesh_volume_element2d.h"



#define TABLE_NAME_ELEMENTS "AutoMesh:elements"
#define COLUMN_NAME_NODE_IDS "AutoMesh:nodeIds"
#define COLUMN_NAME_NEIGHBOR_IDS "AutoMesh:neighborIds"



/* VolumeElement2D */

AUTORES_IMPLEMENT_HANDLE(AutoMesh_VolumeElement2D)

AutoMesh_VolumeElement2D *AutoMesh_VolumeElement2D_New 
(AutoMesh_Node2D *node,
 AutoMesh_PolygonType polygonType,
 int nElementNodes)
{
  AutoMesh_VolumeElement2D *element;

  element = AutoMesh_VolumeElement2D_NewImplementation ();

  element->polygonType = polygonType;
  element->nElementNodes = nElementNodes;
  element->nElementVertexs = AutoMesh_Polygon_NVertexs (polygonType);
  element->nElementEdges = AutoMesh_Polygon_NEdges (polygonType);

  element->node = node;

  element->elements = AutoRes_Table_New (TABLE_NAME_ELEMENTS);
  assert(element->elements != NULL);
  element->elements_nodeIds = NULL;
  element->elements_neighborIds = NULL;

  element->bucket = NULL;

  return element;
}

void AutoMesh_VolumeElement2D_delete 
(AutoMesh_VolumeElement2D *self)
{
  AutoRes_Table_delete (self->elements);
  self->elements = NULL;
  self->elements_nodeIds = NULL;
  self->elements_neighborIds = NULL;

  if (self->bucket != NULL) {
    AutoRes_Bucket2D_delete (self->bucket);
    self->bucket = NULL;
  }

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */ 
AutoMesh_VolumeElement2D_allocateElements 
(AutoMesh_VolumeElement2D *self,
 int nElements)
{
  return AutoRes_Table_allocateRecords (self->elements, nElements);
}

int /* boolean */ 
AutoMesh_VolumeElement2D_reserveElements 
(AutoMesh_VolumeElement2D *self,
 int nElements)
{
  return AutoRes_Table_reserveRecords (self->elements, nElements);
}

int AutoMesh_VolumeElement2D_addElement 
(AutoMesh_VolumeElement2D *self)
{
  return AutoRes_Table_addRecord (self->elements);
}

int /* boolean */
AutoMesh_VolumeElement2D_makeNodeIdsColumn 
(AutoMesh_VolumeElement2D *self)
{
  int flag;

  self->elements_nodeIds = AutoRes_Integer1DColumn_New 
    (self->elements, COLUMN_NAME_NODE_IDS, self->nElementNodes);
  assert(self->elements_nodeIds != NULL);
  flag = AutoRes_Table_addInteger1DColumn 
    (self->elements, self->elements_nodeIds);
  if (!flag) return 0;

  return 1;
}

void AutoMesh_VolumeElement2D_setNodeIdAt 
(AutoMesh_VolumeElement2D *self,
 int elementId,
 int elementNodeNum,
 int nodeId)
{
  AUTORES_I1D_SET(self->elements_nodeIds, elementId, elementNodeNum, 
		  nodeId);
}

int /* boolean */
AutoMesh_VolumeElement2D_makeNeighborIdsColumn 
(AutoMesh_VolumeElement2D *self)
{
  int flag;

  self->elements_neighborIds = AutoRes_Integer1DColumn_New 
    (self->elements, COLUMN_NAME_NEIGHBOR_IDS, self->nElementEdges);
  assert(self->elements_neighborIds != NULL);
  flag = AutoRes_Table_addInteger1DColumn 
    (self->elements, self->elements_neighborIds);
  if (!flag) return 0;

  return 1;
}

void AutoMesh_VolumeElement2D_setNeighborIdAt 
(AutoMesh_VolumeElement2D *self,
 int elementId,
 int elementEdgeNum,
 int neighborId)
{
  AUTORES_I1D_SET(self->elements_neighborIds, elementId, elementEdgeNum, 
		  neighborId);
}

int /* boolean */
AutoMesh_VolumeElement2D_makeBucket 
(AutoMesh_VolumeElement2D *self)
{
  double minRangeX, minRangeY;
  double maxRangeX, maxRangeY;
  double elementSize;
  int nElements = AutoRes_Table_nRecords (self->elements);
  int iElement;
  int xDivisions, yDivisions;
  int nOffsetDivisions = 22;
  double offsetX, offsetY;
  int nAverageLinksPerCell = 50;
  int flag;

  assert(self->elements_nodeIds != NULL);
  assert(self->bucket == NULL);

  AutoMesh_Node2D_getRange 
    (self->node,
     &minRangeX, &minRangeY,
     &maxRangeX, &maxRangeY);

  elementSize = 0.0;
  for (iElement = 0; iElement < nElements; iElement++) {
    elementSize += AutoMesh_VolumeElement2D_averageEdgeLengthAt 
      (self, iElement);
  }
  elementSize /= nElements;
#if 0
  elementSize *= 0.7;
#else
  elementSize *= 1.5;
#endif

  xDivisions = (maxRangeX - minRangeX) / elementSize;
  if (xDivisions < 1) xDivisions = 1;
  yDivisions = (maxRangeY - minRangeY) / elementSize;
  if (yDivisions < 1) yDivisions = 1;

  offsetX = (maxRangeX - minRangeX) / nOffsetDivisions;
  offsetY = (maxRangeY - minRangeY) / nOffsetDivisions;

  self->bucket = AutoRes_Bucket2D_New 
    (minRangeX - offsetX, minRangeY - offsetY,
     maxRangeX + offsetX, maxRangeY + offsetY);
  assert(self->bucket != NULL);
  
  flag = AutoRes_Bucket2D_reserve 
    (self->bucket, 
     xDivisions, yDivisions, 
     nAverageLinksPerCell);
  if (!flag) return 0;

  for (iElement = 0; iElement < nElements; iElement++) {
    double minX, minY;
    double maxX, maxY;
    
    AutoMesh_VolumeElement2D_getRangeAt 
      (self, iElement,
       &minX, &minY,
       &maxX, &maxY);
    
    flag = AutoRes_Bucket2D_addObjectWithin 
      (self->bucket,
       minX, minY,
       maxX, maxY,
       iElement);
    if (!flag) return 0;
  }

  assert(self->bucket != NULL);

  return 1;
}

void AutoMesh_VolumeElement2D_markElementForNodeRenumbering 
(AutoMesh_VolumeElement2D *self,
 int elementId)
{
  int iElementNode;
  
  for (iElementNode = 0; 
       iElementNode < self->nElementNodes; 
       iElementNode++) {
    int nodeId = AutoMesh_VolumeElement2D_nodeIdAt 
      (self, elementId, iElementNode);
    
    AutoMesh_Node2D_markNodeForRenumbering (self->node, nodeId);
  }
}

int AutoMesh_VolumeElement2D_nElements 
(AutoMesh_VolumeElement2D *self)
{
  return AutoRes_Table_nRecords (self->elements);
}

int AutoMesh_VolumeElement2D_nodeIdAt 
(AutoMesh_VolumeElement2D *self,
 int elementId,
 int elementNodeNum)
{
  return AUTORES_I1D(self->elements_nodeIds, elementId, elementNodeNum);
}

int AutoMesh_VolumeElement2D_neighborIdAt 
(AutoMesh_VolumeElement2D *self,
 int elementId,
 int elementEdgeNum)
{
  return AUTORES_I1D(self->elements_neighborIds, elementId, elementEdgeNum);
}

void AutoMesh_VolumeElement2D_getRangeAt
(AutoMesh_VolumeElement2D *self,
 int elementId, 
 double *minRangeX_OUT, double *minRangeY_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT)
{
  int nodeId0 = AutoMesh_VolumeElement2D_nodeIdAt (self, elementId, 0);
  int iElementNode;
  
  AutoMesh_Node2D_getCoordAt (self->node, nodeId0,
			      minRangeX_OUT, minRangeY_OUT);
  AutoMesh_Node2D_getCoordAt (self->node, nodeId0,
			      maxRangeX_OUT, maxRangeY_OUT);
  for (iElementNode = 0; 
       iElementNode < self->nElementNodes; 
       iElementNode++) {
    int nodeId = AutoMesh_VolumeElement2D_nodeIdAt 
      (self, elementId, iElementNode);
    double x, y;
    
    AutoMesh_Node2D_getCoordAt (self->node, nodeId,
				&x, &y);
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    
    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
  }
}

double AutoMesh_VolumeElement2D_averageEdgeLengthAt
(AutoMesh_VolumeElement2D *self,
 int elementId)
{
  int iElementEdge;
  double average;

  average = 0.0;
  for (iElementEdge = 0; 
       iElementEdge < self->nElementEdges; 
       iElementEdge++) {
    int elementNodeNum0 = AutoMesh_Polygon_EdgeVertex
      (self->polygonType, iElementEdge, 0);
    int elementNodeNum1 = AutoMesh_Polygon_EdgeVertex
      (self->polygonType, iElementEdge, 1);
    
    int nodeId0 = AUTORES_I1D(self->elements_nodeIds, 
			      elementId, elementNodeNum0);
    int nodeId1 = AUTORES_I1D(self->elements_nodeIds, 
			      elementId, elementNodeNum1);

    double x0, y0;
    double x1, y1;
    
    AutoMesh_Node2D_getCoordAt (self->node, nodeId0,
				&x0, &y0);
    AutoMesh_Node2D_getCoordAt (self->node, nodeId1,
				&x1, &y1);
    
    average += AutoMesh_Distance2D (x0, y0, 
				    x1, y1);
  }
  average /= self->nElementEdges;
  return average;
}
 
/* 
   mapping between area coordinate (L0, L1, L2) 
   and local coordinate (lx, ly) of triangle element
   
   L0 = 1.0 - lx - ly 
   L1 = lx
   L2 = ly
*/

#define AREA_COORD_EPS 0.00001

static int /* boolean */
ContainsTriPointAt 
(AutoMesh_VolumeElement2D *self,
 int elementId, 
 double x, double y,
 double *lx_OUT, double *ly_OUT)
{
  int nodeId0, nodeId1, nodeId2;
  double x0, y0;
  double x1, y1;
  double x2, y2;
  double area;
  double area0, area1, area2;
  double L0, L1, L2;

  assert(self->polygonType == AUTOMESH_POLYGON_TRI);

  nodeId0 = AUTORES_I1D(self->elements_nodeIds, elementId, 0);
  nodeId1 = AUTORES_I1D(self->elements_nodeIds, elementId, 1);
  nodeId2 = AUTORES_I1D(self->elements_nodeIds, elementId, 2);

  AutoMesh_Node2D_getCoordAt (self->node, nodeId0,
			      &x0, &y0);
  AutoMesh_Node2D_getCoordAt (self->node, nodeId1,
			      &x1, &y1);
  AutoMesh_Node2D_getCoordAt (self->node, nodeId2,
			      &x2, &y2);
  
  area = AutoMesh_SignedTriArea2D 
    (x0, y0, 
     x1, y1, 
     x2, y2);
  area0 = AutoMesh_SignedTriArea2D 
    (x, y, 
     x1, y1, 
     x2, y2);
  area1 = AutoMesh_SignedTriArea2D 
    (x0, y0, 
     x, y, 
     x2, y2);
  area2 = AutoMesh_SignedTriArea2D 
    (x0, y0, 
     x1, y1, 
     x, y);

  L0 = area0 / area;
  L1 = area1 / area;
  L2 = area2 / area;

  if (0.0 - AREA_COORD_EPS < L0 && L0 < 1.0 + AREA_COORD_EPS
      && 0.0 - AREA_COORD_EPS < L1 && L1 < 1.0 + AREA_COORD_EPS
      && 0.0 - AREA_COORD_EPS < L2 && L2 < 1.0 + AREA_COORD_EPS) {
    *lx_OUT = L1;
    *ly_OUT = L2;
    
    return 1;
  }
  return 0;
}

static double TriScalarAt
(AutoMesh_VolumeElement2D *self,
 int elementId,
 AutoMesh_NodeScalar2D *scalar,
 double lx, double ly)
{
  double L0 = 1.0 - lx - ly;
  double L1 = lx;
  double L2 = ly;
  int nodeId0, nodeId1, nodeId2;
  double s0, s1, s2;

  /* NEED TO REDEFINE */
  /* ignore high order nodes */

  assert(self->polygonType == AUTOMESH_POLYGON_TRI);

  nodeId0 = AUTORES_I1D(self->elements_nodeIds, elementId, 0);
  nodeId1 = AUTORES_I1D(self->elements_nodeIds, elementId, 1);
  nodeId2 = AUTORES_I1D(self->elements_nodeIds, elementId, 2);

  s0 = AutoMesh_NodeScalar2D_valueAt (scalar, nodeId0);
  s1 = AutoMesh_NodeScalar2D_valueAt (scalar, nodeId1);
  s2 = AutoMesh_NodeScalar2D_valueAt (scalar, nodeId2);

  return s0 * L0 + s1 * L1 + s2 * L2;
}

static void GetTriVectorAt
(AutoMesh_VolumeElement2D *self,
 int elementId,
 AutoMesh_NodeVector2D *vector,
 double lx, double ly,
 double *vectorX_OUT, double *vectorY_OUT)
{
  double L0 = 1.0 - lx - ly;
  double L1 = lx;
  double L2 = ly;
  int nodeId0, nodeId1, nodeId2;
  double vx0, vy0;
  double vx1, vy1;
  double vx2, vy2;

  /* NEED TO REDEFINE */
  /* ignore high order nodes */

  assert(self->polygonType == AUTOMESH_POLYGON_TRI);

  nodeId0 = AUTORES_I1D(self->elements_nodeIds, elementId, 0);
  nodeId1 = AUTORES_I1D(self->elements_nodeIds, elementId, 1);
  nodeId2 = AUTORES_I1D(self->elements_nodeIds, elementId, 2);

  AutoMesh_NodeVector2D_getValueAt (vector, nodeId0,
				    &vx0, &vy0);
  AutoMesh_NodeVector2D_getValueAt (vector, nodeId1,
				    &vx1, &vy1);
  AutoMesh_NodeVector2D_getValueAt (vector, nodeId2,
				    &vx2, &vy2);

  *vectorX_OUT = vx0 * L0 + vx1 * L1 + vx2 * L2;
  *vectorY_OUT = vy0 * L0 + vy1 * L1 + vy2 * L2;
}

int /* boolean */
AutoMesh_VolumeElement2D_containsPointAt 
(AutoMesh_VolumeElement2D *self,
 int elementId, 
 double x, double y,
 double *lx_OUT, double *ly_OUT)
{
  int flag = 0;

  switch (self->polygonType) {
  case AUTOMESH_POLYGON_TRI:
    flag = ContainsTriPointAt (self, elementId, x, y,
			       lx_OUT, ly_OUT);
    break;
  case AUTOMESH_POLYGON_QUAD:
    /* NEED TO REDEFINE */
    assert(0);
    break;
  default:
    assert(0);
    break;
  }

  return flag;
}

double AutoMesh_VolumeElement2D_scalarAt
(AutoMesh_VolumeElement2D *self,
 int elementId,
 AutoMesh_NodeScalar2D *scalar,
 double lx, double ly)
{
  switch (self->polygonType) {
  case AUTOMESH_POLYGON_TRI:
    return TriScalarAt (self, elementId, scalar, lx, ly);
  case AUTOMESH_POLYGON_QUAD:
    /* NEED TO REDEFINE */
    assert(0);
    break;
  default:
    assert(0);
    break;
  }
  return 0.0;
}

void AutoMesh_VolumeElement2D_getVectorAt
(AutoMesh_VolumeElement2D *self,
 int elementId,
 AutoMesh_NodeVector2D *vector,
 double lx, double ly,
 double *vectorX_OUT, double *vectorY_OUT)
{
  switch (self->polygonType) {
  case AUTOMESH_POLYGON_TRI:
    GetTriVectorAt (self, elementId, vector, lx, ly,
		    vectorX_OUT, vectorY_OUT);
    break;
  case AUTOMESH_POLYGON_QUAD:
    /* NEED TO REDEFINE */
    assert(0);
    break;
  default:
    assert(0);
    break;
  }
}

int AutoMesh_VolumeElement2D_findElementUsingBucket 
(AutoMesh_VolumeElement2D *self,
 double x, double y,
 double *lx_OUT, double *ly_OUT)
{
  int cx, cy;
  
  assert(self->bucket != NULL);

  AutoRes_Bucket2D_getCellIdAt (self->bucket, x, y,
				&cx, &cy);
  
  if (AutoRes_Bucket2D_isOutOfRange (self->bucket, cx, cy)) {
    return -1;
  } else {
    int linkId;
    
    for (linkId = AutoRes_Bucket2D_startLinkIdIn (self->bucket, cx, cy);
	 linkId != -1;
	 linkId = AutoRes_Bucket2D_nextLinkId (self->bucket, linkId)) {
      int elementId = AutoRes_Bucket2D_objectId (self->bucket, linkId);
      
      if (AutoMesh_VolumeElement2D_containsPointAt 
	  (self, elementId, x, y,
	   lx_OUT, ly_OUT)) {
	return elementId;
      }
    }
  }

  return -1;
}

int AutoMesh_VolumeElement2D_findElementNear 
(AutoMesh_VolumeElement2D *self,
 int elementId, 
 double x, double y,
 double *lx_OUT, double *ly_OUT)
{
  int iElementEdge;
  int jElementEdge;
  int kElementEdge;
  
  assert(elementId != -1);

  if (AutoMesh_VolumeElement2D_containsPointAt 
      (self, elementId, x, y,
       lx_OUT, ly_OUT)) {
    return elementId;
  }

  for (iElementEdge = 0; 
       iElementEdge < self->nElementEdges; 
       iElementEdge++) {
    int neighborId = AUTORES_I1D
      (self->elements_neighborIds, elementId, iElementEdge);
    
    if (neighborId == -1) continue;

    if (AutoMesh_VolumeElement2D_containsPointAt 
	(self, neighborId, x, y,
	 lx_OUT, ly_OUT)) {
      return neighborId;
    }
  }

  for (iElementEdge = 0; 
       iElementEdge < self->nElementEdges; 
       iElementEdge++) {
    int neighborIdI = AUTORES_I1D
      (self->elements_neighborIds, elementId, iElementEdge);
    
    if (neighborIdI == -1) continue;

    for (jElementEdge = 0; 
	 jElementEdge < self->nElementEdges; 
	 jElementEdge++) {
      int neighborIdJ = AUTORES_I1D
	(self->elements_neighborIds, neighborIdI, jElementEdge);
      
      if (neighborIdJ == -1) continue;
      if (neighborIdJ == elementId) continue;

      if (AutoMesh_VolumeElement2D_containsPointAt 
	  (self, neighborIdJ, x, y,
	   lx_OUT, ly_OUT)) {
	return neighborIdJ;
      }
    }
  }

  for (iElementEdge = 0; 
       iElementEdge < self->nElementEdges; 
       iElementEdge++) {
    int neighborIdI = AUTORES_I1D
      (self->elements_neighborIds, elementId, iElementEdge);
    
    if (neighborIdI == -1) continue;

    for (jElementEdge = 0; 
	 jElementEdge < self->nElementEdges; 
	 jElementEdge++) {
      int neighborIdJ = AUTORES_I1D
	(self->elements_neighborIds, neighborIdI, jElementEdge);
      
      if (neighborIdJ == -1) continue;
      if (neighborIdJ == elementId) continue;

      for (kElementEdge = 0; 
	   kElementEdge < self->nElementEdges; 
	   kElementEdge++) {
	int neighborIdK = AUTORES_I1D
	  (self->elements_neighborIds, neighborIdJ, kElementEdge);
	
	if (neighborIdK == -1) continue;
	if (neighborIdK == elementId) continue;
	if (neighborIdK == neighborIdI) continue;
	
	if (AutoMesh_VolumeElement2D_containsPointAt 
	    (self, neighborIdK, x, y,
	     lx_OUT, ly_OUT)) {
	  return neighborIdK;
	}
      }
    }
  }

  return -1;
}






