/* automesh_volume_element3d.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"
#include "automesh_shape.h"
#include "automesh_node3d.h"
#include "automesh_node_scalar3d.h"
#include "automesh_node_vector3d.h"
#include "automesh_volume_element3d.h"



#define TABLE_NAME_ELEMENTS "AutoMesh:elements"
#define COLUMN_NAME_NODE_IDS "AutoMesh:nodeIds"
#define COLUMN_NAME_NEIGHBOR_IDS "AutoMesh:neighborIds"



/* VolumeElement3D */

AUTORES_IMPLEMENT_HANDLE(AutoMesh_VolumeElement3D)

AutoMesh_VolumeElement3D *AutoMesh_VolumeElement3D_New 
(AutoMesh_Node3D *node,
 AutoMesh_PolyhedronType polyhedronType,
 int nElementNodes)
{
  AutoMesh_VolumeElement3D *element;

  element = AutoMesh_VolumeElement3D_NewImplementation ();

  element->polyhedronType = polyhedronType;
  element->nElementNodes = nElementNodes;
  element->nElementVertexs = AutoMesh_Polyhedron_NVertexs (polyhedronType);
  element->nElementEdges = AutoMesh_Polyhedron_NEdges (polyhedronType);
  element->nElementFaces = AutoMesh_Polyhedron_NFaces (polyhedronType);

  element->node = node;

  element->elements = AutoRes_Table_New (TABLE_NAME_ELEMENTS);
  assert(element->elements != NULL);
  element->elements_nodeIds = NULL;
  element->elements_neighborIds = NULL;

  element->bucket = NULL;

  return element;
}

void AutoMesh_VolumeElement3D_delete 
(AutoMesh_VolumeElement3D *self)
{
  AutoRes_Table_delete (self->elements);
  self->elements = NULL;
  self->elements_nodeIds = NULL;
  self->elements_neighborIds = NULL;

  if (self->bucket != NULL) {
    AutoRes_Bucket3D_delete (self->bucket);
    self->bucket = NULL;
  }

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */ 
AutoMesh_VolumeElement3D_allocateElements 
(AutoMesh_VolumeElement3D *self,
 int nElements)
{
  return AutoRes_Table_allocateRecords (self->elements, nElements);
}

int /* boolean */ 
AutoMesh_VolumeElement3D_reserveElements 
(AutoMesh_VolumeElement3D *self,
 int nElements)
{
  return AutoRes_Table_reserveRecords (self->elements, nElements);
}

int AutoMesh_VolumeElement3D_addElement 
(AutoMesh_VolumeElement3D *self)
{
  return AutoRes_Table_addRecord (self->elements);
}

int /* boolean */
AutoMesh_VolumeElement3D_makeNodeIdsColumn 
(AutoMesh_VolumeElement3D *self)
{
  int flag;

  self->elements_nodeIds = AutoRes_Integer1DColumn_New 
    (self->elements, COLUMN_NAME_NODE_IDS, self->nElementNodes);
  assert(self->elements_nodeIds != NULL);
  flag = AutoRes_Table_addInteger1DColumn 
    (self->elements, self->elements_nodeIds);
  if (!flag) return 0;

  return 1;
}

void AutoMesh_VolumeElement3D_setNodeIdAt 
(AutoMesh_VolumeElement3D *self,
 int elementId,
 int elementNodeNum,
 int nodeId)
{
  AUTORES_I1D_SET(self->elements_nodeIds, elementId, elementNodeNum, 
		  nodeId);
}

int /* boolean */
AutoMesh_VolumeElement3D_makeNeighborIdsColumn 
(AutoMesh_VolumeElement3D *self)
{
  int flag;

  self->elements_neighborIds = AutoRes_Integer1DColumn_New 
    (self->elements, COLUMN_NAME_NEIGHBOR_IDS, self->nElementFaces);
  assert(self->elements_neighborIds != NULL);
  flag = AutoRes_Table_addInteger1DColumn 
    (self->elements, self->elements_neighborIds);
  if (!flag) return 0;

  return 1;
}

void AutoMesh_VolumeElement3D_setNeighborIdAt 
(AutoMesh_VolumeElement3D *self,
 int elementId,
 int elementFaceNum,
 int neighborId)
{
  AUTORES_I1D_SET(self->elements_neighborIds, elementId, elementFaceNum, 
		  neighborId);
}

int /* boolean */
AutoMesh_VolumeElement3D_makeBucket 
(AutoMesh_VolumeElement3D *self)
{
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  double elementSize;
  int nElements = AutoRes_Table_nRecords (self->elements);
  int iElement;
  int xDivisions, yDivisions, zDivisions;
  int nOffsetDivisions = 23;
  double offsetX, offsetY, offsetZ;
  int nAverageLinksPerCell = 50;
  int flag;

  assert(self->elements_nodeIds != NULL);
  assert(self->bucket == NULL);

  AutoMesh_Node3D_getRange 
    (self->node,
     &minRangeX, &minRangeY, &minRangeZ,
     &maxRangeX, &maxRangeY, &maxRangeZ);

  elementSize = 0.0;
  for (iElement = 0; iElement < nElements; iElement++) {
    elementSize += AutoMesh_VolumeElement3D_averageEdgeLengthAt 
      (self, iElement);
  }
  elementSize /= nElements;
#if 0
  elementSize *= 0.7;
#else
  elementSize *= 1.5;
#endif

  xDivisions = (maxRangeX - minRangeX) / elementSize;
  if (xDivisions < 1) xDivisions = 1;
  yDivisions = (maxRangeY - minRangeY) / elementSize;
  if (yDivisions < 1) yDivisions = 1;
  zDivisions = (maxRangeZ - minRangeZ) / elementSize;
  if (zDivisions < 1) zDivisions = 1;

  offsetX = (maxRangeX - minRangeX) / nOffsetDivisions;
  offsetY = (maxRangeY - minRangeY) / nOffsetDivisions;
  offsetZ = (maxRangeZ - minRangeZ) / nOffsetDivisions;

  self->bucket = AutoRes_Bucket3D_New 
    (minRangeX - offsetX, minRangeY - offsetY, minRangeZ - offsetZ,
     maxRangeX + offsetX, maxRangeY + offsetY, maxRangeZ + offsetZ);
  assert(self->bucket != NULL);
  
  flag = AutoRes_Bucket3D_reserve 
    (self->bucket, 
     xDivisions, yDivisions, zDivisions, 
     nAverageLinksPerCell);
  if (!flag) return 0;

  for (iElement = 0; iElement < nElements; iElement++) {
    double minX, minY, minZ;
    double maxX, maxY, maxZ;
    
    AutoMesh_VolumeElement3D_getRangeAt 
      (self, iElement,
       &minX, &minY, &minZ,
       &maxX, &maxY, &maxZ);
    
    flag = AutoRes_Bucket3D_addObjectWithin 
      (self->bucket,
       minX, minY, minZ,
       maxX, maxY, maxZ,
       iElement);
    if (!flag) return 0;
  }

  assert(self->bucket != NULL);

  return 1;
}

void AutoMesh_VolumeElement3D_markElementForNodeRenumbering 
(AutoMesh_VolumeElement3D *self,
 int elementId)
{
  int iElementNode;
  
  for (iElementNode = 0; 
       iElementNode < self->nElementNodes; 
       iElementNode++) {
    int nodeId = AutoMesh_VolumeElement3D_nodeIdAt 
      (self, elementId, iElementNode);
    
    AutoMesh_Node3D_markNodeForRenumbering (self->node, nodeId);
  }
}

int AutoMesh_VolumeElement3D_nElements 
(AutoMesh_VolumeElement3D *self)
{
  return AutoRes_Table_nRecords (self->elements);
}

int AutoMesh_VolumeElement3D_nodeIdAt 
(AutoMesh_VolumeElement3D *self,
 int elementId,
 int elementNodeNum)
{
  return AUTORES_I1D(self->elements_nodeIds, elementId, elementNodeNum);
}

int AutoMesh_VolumeElement3D_neighborIdAt 
(AutoMesh_VolumeElement3D *self,
 int elementId,
 int elementFaceNum)
{
  return AUTORES_I1D(self->elements_neighborIds, elementId, elementFaceNum);
}

void AutoMesh_VolumeElement3D_getRangeAt
(AutoMesh_VolumeElement3D *self,
 int elementId, 
 double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int nodeId0 = AutoMesh_VolumeElement3D_nodeIdAt (self, elementId, 0);
  int iElementNode;
  
  AutoMesh_Node3D_getCoordAt (self->node, nodeId0,
			      minRangeX_OUT, minRangeY_OUT, minRangeZ_OUT);
  AutoMesh_Node3D_getCoordAt (self->node, nodeId0,
			      maxRangeX_OUT, maxRangeY_OUT, maxRangeZ_OUT);
  for (iElementNode = 0; 
       iElementNode < self->nElementNodes; 
       iElementNode++) {
    int nodeId = AutoMesh_VolumeElement3D_nodeIdAt 
      (self, elementId, iElementNode);
    double x, y, z;
    
    AutoMesh_Node3D_getCoordAt (self->node, nodeId,
				&x, &y, &z);
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;
    if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;
    
    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
    if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
  }
}

double AutoMesh_VolumeElement3D_averageEdgeLengthAt
(AutoMesh_VolumeElement3D *self,
 int elementId)
{
  int iElementEdge;
  double average;

  average = 0.0;
  for (iElementEdge = 0; 
       iElementEdge < self->nElementEdges; 
       iElementEdge++) {
    int elementNodeNum0 = AutoMesh_Polyhedron_EdgeVertex
      (self->polyhedronType, iElementEdge, 0);
    int elementNodeNum1 = AutoMesh_Polyhedron_EdgeVertex
      (self->polyhedronType, iElementEdge, 1);
    
    int nodeId0 = AUTORES_I1D(self->elements_nodeIds, 
			      elementId, elementNodeNum0);
    int nodeId1 = AUTORES_I1D(self->elements_nodeIds, 
			      elementId, elementNodeNum1);

    double x0, y0, z0;
    double x1, y1, z1;
    
    AutoMesh_Node3D_getCoordAt (self->node, nodeId0,
				&x0, &y0, &z0);
    AutoMesh_Node3D_getCoordAt (self->node, nodeId1,
				&x1, &y1, &z1);
    
    average += AutoMesh_Distance3D (x0, y0, z0, 
				    x1, y1, z1);
  }
  average /= self->nElementEdges;
  return average;
}
 
/* 
   mapping between volume coordinate (L0, L1, L2, L3) 
   and local coordinate (lx, ly, lz) of tetrahedron element
   
   L0 = 1.0 - lx - ly - lz 
   L1 = lx
   L2 = ly
   L3 = lz
*/

#define VOLUME_COORD_EPS 0.00001

static int /* boolean */
ContainsTetraPointAt 
(AutoMesh_VolumeElement3D *self,
 int elementId, 
 double x, double y, double z,
 double *lx_OUT, double *ly_OUT, double *lz_OUT)
{
  int nodeId0, nodeId1, nodeId2, nodeId3;
  double x0, y0, z0;
  double x1, y1, z1;
  double x2, y2, z2;
  double x3, y3, z3;
  double volume;
  double volume0, volume1, volume2, volume3;
  double L0, L1, L2, L3;

  assert(self->polyhedronType == AUTOMESH_POLYHEDRON_TETRA);

  nodeId0 = AUTORES_I1D(self->elements_nodeIds, elementId, 0);
  nodeId1 = AUTORES_I1D(self->elements_nodeIds, elementId, 1);
  nodeId2 = AUTORES_I1D(self->elements_nodeIds, elementId, 2);
  nodeId3 = AUTORES_I1D(self->elements_nodeIds, elementId, 3);

  AutoMesh_Node3D_getCoordAt (self->node, nodeId0,
			      &x0, &y0, &z0);
  AutoMesh_Node3D_getCoordAt (self->node, nodeId1,
			      &x1, &y1, &z1);
  AutoMesh_Node3D_getCoordAt (self->node, nodeId2,
			      &x2, &y2, &z2);
  AutoMesh_Node3D_getCoordAt (self->node, nodeId3,
			      &x3, &y3, &z3);
  
  volume = AutoMesh_SignedTetraVolume3D 
    (x0, y0, z0, 
     x1, y1, z1, 
     x2, y2, z2, 
     x3, y3, z3);
  volume0 = AutoMesh_SignedTetraVolume3D 
    (x, y, z, 
     x1, y1, z1, 
     x2, y2, z2, 
     x3, y3, z3);
  volume1 = AutoMesh_SignedTetraVolume3D 
    (x0, y0, z0, 
     x, y, z, 
     x2, y2, z2, 
     x3, y3, z3);
  volume2 = AutoMesh_SignedTetraVolume3D 
    (x0, y0, z0, 
     x1, y1, z1, 
     x, y, z, 
     x3, y3, z3);
  volume3 = AutoMesh_SignedTetraVolume3D 
    (x0, y0, z0, 
     x1, y1, z1, 
     x2, y2, z2, 
     x, y, z);

  L0 = volume0 / volume;
  L1 = volume1 / volume;
  L2 = volume2 / volume;
  L3 = volume3 / volume;

  if (0.0 - VOLUME_COORD_EPS < L0 && L0 < 1.0 + VOLUME_COORD_EPS
      && 0.0 - VOLUME_COORD_EPS < L1 && L1 < 1.0 + VOLUME_COORD_EPS
      && 0.0 - VOLUME_COORD_EPS < L2 && L2 < 1.0 + VOLUME_COORD_EPS
      && 0.0 - VOLUME_COORD_EPS < L3 && L3 < 1.0 + VOLUME_COORD_EPS) {
    *lx_OUT = L1;
    *ly_OUT = L2;
    *lz_OUT = L3;
    
    return 1;
  }
  return 0;
}

static double TetraScalarAt
(AutoMesh_VolumeElement3D *self,
 int elementId,
 AutoMesh_NodeScalar3D *scalar,
 double lx, double ly, double lz)
{
  double L0 = 1.0 - lx - ly - lz;
  double L1 = lx;
  double L2 = ly;
  double L3 = lz;
  int nodeId0, nodeId1, nodeId2, nodeId3;
  double s0, s1, s2, s3;

  /* NEED TO REDEFINE */
  /* ignore high order nodes */

  assert(self->polyhedronType == AUTOMESH_POLYHEDRON_TETRA);

  nodeId0 = AUTORES_I1D(self->elements_nodeIds, elementId, 0);
  nodeId1 = AUTORES_I1D(self->elements_nodeIds, elementId, 1);
  nodeId2 = AUTORES_I1D(self->elements_nodeIds, elementId, 2);
  nodeId3 = AUTORES_I1D(self->elements_nodeIds, elementId, 3);

  s0 = AutoMesh_NodeScalar3D_valueAt (scalar, nodeId0);
  s1 = AutoMesh_NodeScalar3D_valueAt (scalar, nodeId1);
  s2 = AutoMesh_NodeScalar3D_valueAt (scalar, nodeId2);
  s3 = AutoMesh_NodeScalar3D_valueAt (scalar, nodeId3);

  return s0 * L0 + s1 * L1 + s2 * L2 + s3 * L3;
}

static void GetTetraVectorAt
(AutoMesh_VolumeElement3D *self,
 int elementId,
 AutoMesh_NodeVector3D *vector,
 double lx, double ly, double lz,
 double *vectorX_OUT, double *vectorY_OUT, double *vectorZ_OUT)
{
  double L0 = 1.0 - lx - ly - lz;
  double L1 = lx;
  double L2 = ly;
  double L3 = lz;
  int nodeId0, nodeId1, nodeId2, nodeId3;
  double vx0, vy0, vz0;
  double vx1, vy1, vz1;
  double vx2, vy2, vz2;
  double vx3, vy3, vz3;

  /* NEED TO REDEFINE */
  /* ignore high order nodes */

  assert(self->polyhedronType == AUTOMESH_POLYHEDRON_TETRA);

  nodeId0 = AUTORES_I1D(self->elements_nodeIds, elementId, 0);
  nodeId1 = AUTORES_I1D(self->elements_nodeIds, elementId, 1);
  nodeId2 = AUTORES_I1D(self->elements_nodeIds, elementId, 2);
  nodeId3 = AUTORES_I1D(self->elements_nodeIds, elementId, 3);

  AutoMesh_NodeVector3D_getValueAt (vector, nodeId0,
				    &vx0, &vy0, &vz0);
  AutoMesh_NodeVector3D_getValueAt (vector, nodeId1,
				    &vx1, &vy1, &vz1);
  AutoMesh_NodeVector3D_getValueAt (vector, nodeId2,
				    &vx2, &vy2, &vz2);
  AutoMesh_NodeVector3D_getValueAt (vector, nodeId3,
				    &vx3, &vy3, &vz3);

  *vectorX_OUT = vx0 * L0 + vx1 * L1 + vx2 * L2 + vx3 * L3;
  *vectorY_OUT = vy0 * L0 + vy1 * L1 + vy2 * L2 + vy3 * L3;
  *vectorZ_OUT = vz0 * L0 + vz1 * L1 + vz2 * L2 + vz3 * L3;
}

int /* boolean */
AutoMesh_VolumeElement3D_containsPointAt 
(AutoMesh_VolumeElement3D *self,
 int elementId, 
 double x, double y, double z,
 double *lx_OUT, double *ly_OUT, double *lz_OUT)
{
  int flag = 0;

  switch (self->polyhedronType) {
  case AUTOMESH_POLYHEDRON_TETRA:
    flag = ContainsTetraPointAt (self, elementId, x, y, z,
				 lx_OUT, ly_OUT, lz_OUT);
    break;
  case AUTOMESH_POLYHEDRON_HEXA:
    /* NEED TO REDEFINE */
    assert(0);
    break;
  case AUTOMESH_POLYHEDRON_PRISM:
    /* NEED TO REDEFINE */
    assert(0);
    break;
  default:
    assert(0);
    break;
  }

  return flag;
}

double AutoMesh_VolumeElement3D_scalarAt
(AutoMesh_VolumeElement3D *self,
 int elementId,
 AutoMesh_NodeScalar3D *scalar,
 double lx, double ly, double lz)
{
  switch (self->polyhedronType) {
  case AUTOMESH_POLYHEDRON_TETRA:
    return TetraScalarAt (self, elementId, scalar, lx, ly, lz);
  case AUTOMESH_POLYHEDRON_HEXA:
    /* NEED TO REDEFINE */
    assert(0);
    break;
  case AUTOMESH_POLYHEDRON_PRISM:
    /* NEED TO REDEFINE */
    assert(0);
    break;
  default:
    assert(0);
    break;
  }
  return 0.0;
}

void AutoMesh_VolumeElement3D_getVectorAt
(AutoMesh_VolumeElement3D *self,
 int elementId,
 AutoMesh_NodeVector3D *vector,
 double lx, double ly, double lz,
 double *vectorX_OUT, double *vectorY_OUT, double *vectorZ_OUT)
{
  switch (self->polyhedronType) {
  case AUTOMESH_POLYHEDRON_TETRA:
    GetTetraVectorAt (self, elementId, vector, lx, ly, lz,
		      vectorX_OUT, vectorY_OUT, vectorZ_OUT);
    break;
  case AUTOMESH_POLYHEDRON_HEXA:
    /* NEED TO REDEFINE */
    assert(0);
    break;
  case AUTOMESH_POLYHEDRON_PRISM:
    /* NEED TO REDEFINE */
    assert(0);
    break;
  default:
    assert(0);
    break;
  }
}

int AutoMesh_VolumeElement3D_findElementUsingBucket 
(AutoMesh_VolumeElement3D *self,
 double x, double y, double z,
 double *lx_OUT, double *ly_OUT, double *lz_OUT)
{
  int cx, cy, cz;
  
  assert(self->bucket != NULL);

  AutoRes_Bucket3D_getCellIdAt (self->bucket, x, y, z,
				&cx, &cy, &cz);
  
  if (AutoRes_Bucket3D_isOutOfRange (self->bucket, cx, cy, cz)) {
    return -1;
  } else {
    int linkId;
    
    for (linkId = AutoRes_Bucket3D_startLinkIdIn (self->bucket, cx, cy, cz);
	 linkId != -1;
	 linkId = AutoRes_Bucket3D_nextLinkId (self->bucket, linkId)) {
      int elementId = AutoRes_Bucket3D_objectId (self->bucket, linkId);
      
      if (AutoMesh_VolumeElement3D_containsPointAt 
	  (self, elementId, x, y, z,
	   lx_OUT, ly_OUT, lz_OUT)) {
	return elementId;
      }
    }
  }

  return -1;
}

int AutoMesh_VolumeElement3D_findElementNear 
(AutoMesh_VolumeElement3D *self,
 int elementId, 
 double x, double y, double z,
 double *lx_OUT, double *ly_OUT, double *lz_OUT)
{
  int iElementFace;
  int jElementFace;
  int kElementFace;
  
  assert(elementId != -1);

  if (AutoMesh_VolumeElement3D_containsPointAt 
      (self, elementId, x, y, z,
       lx_OUT, ly_OUT, lz_OUT)) {
    return elementId;
  }

  for (iElementFace = 0; 
       iElementFace < self->nElementFaces; 
       iElementFace++) {
    int neighborId = AUTORES_I1D
      (self->elements_neighborIds, elementId, iElementFace);
    
    if (neighborId == -1) continue;

    if (AutoMesh_VolumeElement3D_containsPointAt 
	(self, neighborId, x, y, z,
	 lx_OUT, ly_OUT, lz_OUT)) {
      return neighborId;
    }
  }

  for (iElementFace = 0; 
       iElementFace < self->nElementFaces; 
       iElementFace++) {
    int neighborIdI = AUTORES_I1D
      (self->elements_neighborIds, elementId, iElementFace);
    
    if (neighborIdI == -1) continue;

    for (jElementFace = 0; 
	 jElementFace < self->nElementFaces; 
	 jElementFace++) {
      int neighborIdJ = AUTORES_I1D
	(self->elements_neighborIds, neighborIdI, jElementFace);
      
      if (neighborIdJ == -1) continue;
      if (neighborIdJ == elementId) continue;

      if (AutoMesh_VolumeElement3D_containsPointAt 
	  (self, neighborIdJ, x, y, z,
	   lx_OUT, ly_OUT, lz_OUT)) {
	return neighborIdJ;
      }
    }
  }

  for (iElementFace = 0; 
       iElementFace < self->nElementFaces; 
       iElementFace++) {
    int neighborIdI = AUTORES_I1D
      (self->elements_neighborIds, elementId, iElementFace);
    
    if (neighborIdI == -1) continue;

    for (jElementFace = 0; 
	 jElementFace < self->nElementFaces; 
	 jElementFace++) {
      int neighborIdJ = AUTORES_I1D
	(self->elements_neighborIds, neighborIdI, jElementFace);
      
      if (neighborIdJ == -1) continue;
      if (neighborIdJ == elementId) continue;

      for (kElementFace = 0; 
	   kElementFace < self->nElementFaces; 
	   kElementFace++) {
	int neighborIdK = AUTORES_I1D
	  (self->elements_neighborIds, neighborIdJ, kElementFace);
	
	if (neighborIdK == -1) continue;
	if (neighborIdK == elementId) continue;
	if (neighborIdK == neighborIdI) continue;
	
	if (AutoMesh_VolumeElement3D_containsPointAt 
	    (self, neighborIdK, x, y, z,
	     lx_OUT, ly_OUT, lz_OUT)) {
	  return neighborIdK;
	}
      }
    }
  }

  return -1;
}






