/* automesh_volume_face3d.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"
#include "automesh_shape.h"
#include "automesh_node3d.h"
#include "automesh_volume_element3d.h"
#include "automesh_volume_face3d.h"



#define TABLE_NAME_FACES "AutoMesh:faces"
#define COLUMN_NAME_FACE_IDS "AutoMesh:faceIds"
#define COLUMN_NAME_VERTEX_NODE_IDS "AutoMesh:vertexNodeIds"
#define COLUMN_NAME_ELEMENT_IDS "AutoMesh:elementIds"
#define COLUMN_NAME_ELEMENT_FACE_NUMS "AutoMesh:elementFaceNums"



/* VolumeFace3D */

AUTORES_IMPLEMENT_HANDLE(AutoMesh_VolumeFace3D)

AutoMesh_VolumeFace3D *AutoMesh_VolumeFace3D_New 
(AutoMesh_VolumeElement3D *element)
{
  AutoMesh_VolumeFace3D *face;

  face = AutoMesh_VolumeFace3D_NewImplementation ();

  face->polyhedronType = element->polyhedronType;
  face->nElementFaces 
    = AutoMesh_Polyhedron_NFaces (element->polyhedronType);

  face->element = element;
  face->elements_faceIds = NULL;

  face->faces = AutoRes_Table_New (TABLE_NAME_FACES);
  assert(face->faces != NULL);
  face->faces_vertexNodeIds = NULL;
  face->faces_elementIds = NULL;
  face->faces_elementFaceNums = NULL;

  return face;
}

void AutoMesh_VolumeFace3D_delete 
(AutoMesh_VolumeFace3D *self)
{
  AutoRes_Table_delete (self->faces);
  self->faces = NULL;
  self->faces_vertexNodeIds = NULL;
  self->faces_elementIds = NULL;
  self->faces_elementFaceNums = NULL;

  if (self->elements_faceIds != NULL) {
    AutoRes_Table_removeColumn 
      (self->element->elements, 
       AutoRes_Table_columnIdOfColumn (self->element->elements, 
				       self->elements_faceIds));
    self->elements_faceIds = NULL;
  }

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoMesh_VolumeFace3D_makeFaceIdsAtElementColumn 
(AutoMesh_VolumeFace3D *self)
{
  int flag;

  self->elements_faceIds = AutoRes_Integer1DColumn_New 
    (self->element->elements, COLUMN_NAME_FACE_IDS, self->nElementFaces);
  assert(self->elements_faceIds != NULL);
  flag = AutoRes_Table_addInteger1DColumn 
    (self->element->elements, self->elements_faceIds);
  if (!flag) return 0;

  return 1;
}

int /* boolean */
AutoMesh_VolumeFace3D_makeVertexNodeIdsColumn 
(AutoMesh_VolumeFace3D *self,
 int nFaceVertexs)
{
  int flag;

  self->faces_vertexNodeIds = AutoRes_Integer1DColumn_New 
    (self->faces, COLUMN_NAME_VERTEX_NODE_IDS, nFaceVertexs);
  assert(self->faces_vertexNodeIds != NULL);
  flag = AutoRes_Table_addInteger1DColumn 
    (self->faces, self->faces_vertexNodeIds);
  if (!flag) return 0;

  return 1;
}

int /* boolean */
AutoMesh_VolumeFace3D_makeElementIdsColumn 
(AutoMesh_VolumeFace3D *self)
{
  int flag;

  self->faces_elementIds = AutoRes_Integer1DColumn_New 
    (self->faces, COLUMN_NAME_ELEMENT_IDS, 2);
  assert(self->faces_elementIds != NULL);
  flag = AutoRes_Table_addInteger1DColumn 
    (self->faces, self->faces_elementIds);
  if (!flag) return 0;
  
  self->faces_elementFaceNums = AutoRes_Integer1DColumn_New 
    (self->faces, COLUMN_NAME_ELEMENT_FACE_NUMS, 2);
  assert(self->faces_elementFaceNums != NULL);
  flag = AutoRes_Table_addInteger1DColumn 
    (self->faces, self->faces_elementFaceNums);
  if (!flag) return 0;

  return 1;
}

static void SetElementFace 
(AutoMesh_VolumeFace3D *self,
 int elementId, 
 int elementFaceNum, 
 int faceId,
 int elementIdsFlag)
{
  AUTORES_I1D_SET(self->elements_faceIds, elementId, elementFaceNum, 
		  faceId);
  
  if (elementIdsFlag) {
    assert(AUTORES_I1D(self->faces_elementIds, faceId, 1) == -1);
    if (AUTORES_I1D(self->faces_elementIds, faceId, 0) == -1) {
      AUTORES_I1D_SET(self->faces_elementIds, faceId, 0, elementId);
      AUTORES_I1D_SET(self->faces_elementFaceNums, faceId, 0, elementFaceNum);
    } else {
      AUTORES_I1D_SET(self->faces_elementIds, faceId, 1, elementId);
      AUTORES_I1D_SET(self->faces_elementFaceNums, faceId, 1, elementFaceNum);
    }
  }
}

static int FindTriangleFace 
(AutoMesh_VolumeFace3D *self,
 int nodeId0, int nodeId1, int nodeId2,
 AutoRes_ListTable *faceTable)
{
  int minNodeId, midNodeId, maxNodeId;
  int hashTableSize;
  int hashValue;
  int linkId;

  AutoMesh_SortId3 (nodeId0, nodeId1, nodeId2,
		    &minNodeId, &midNodeId, &maxNodeId);

  hashTableSize = AutoRes_ListTable_nLists (faceTable);
  hashValue = (nodeId0 + nodeId1 + nodeId2) % hashTableSize;
  for (linkId = AutoRes_ListTable_startLinkId (faceTable, hashValue);
       linkId != -1;
       linkId = AutoRes_ListTable_nextLinkId (faceTable, linkId)) {
    int faceId = AutoRes_ListTable_objectId (faceTable, linkId);
    
    int fNodeId0 = AUTORES_I1D(self->faces_vertexNodeIds, faceId, 0);
    int fNodeId1 = AUTORES_I1D(self->faces_vertexNodeIds, faceId, 1);
    int fNodeId2 = AUTORES_I1D(self->faces_vertexNodeIds, faceId, 2);
    
    int minFNodeId, midFNodeId, maxFNodeId; 
    
    AutoMesh_SortId3 (fNodeId0, fNodeId1, fNodeId2,
		      &minFNodeId, &midFNodeId, &maxFNodeId);
    
    if (minFNodeId == minNodeId
	&& midFNodeId == midNodeId
	&& maxFNodeId == maxNodeId) {
      return faceId;
    }
  }

  return -1;
}

static int AddTriangleFace 
(AutoMesh_VolumeFace3D *self,
 int nodeId0, int nodeId1, int nodeId2,
 AutoRes_ListTable *faceTable,
 int elementIdsFlag)
{
  int faceId;
  int hashTableSize;
  int hashValue;
  int linkId;

  faceId = AutoRes_Table_addRecord (self->faces);
  if (faceId == -1) return -1;

  if (elementIdsFlag) {
    AUTORES_I1D_SET(self->faces_elementIds, faceId, 0, -1);
    AUTORES_I1D_SET(self->faces_elementIds, faceId, 1, -1);
    
    AUTORES_I1D_SET(self->faces_elementFaceNums, faceId, 0, -1);
    AUTORES_I1D_SET(self->faces_elementFaceNums, faceId, 1, -1);
  }

  AUTORES_I1D_SET(self->faces_vertexNodeIds, faceId, 0, nodeId0);
  AUTORES_I1D_SET(self->faces_vertexNodeIds, faceId, 1, nodeId1);
  AUTORES_I1D_SET(self->faces_vertexNodeIds, faceId, 2, nodeId2);

  hashTableSize = AutoRes_ListTable_nLists (faceTable);
  hashValue = (nodeId0 + nodeId1 + nodeId2) % hashTableSize;
  linkId = AutoRes_ListTable_addObject (faceTable, hashValue, faceId);
  if (linkId == -1) return -1;

  return faceId;
}

static int /* boolean */
ExtractTetraFace
(AutoMesh_VolumeFace3D *self,
 int elementIdsFlag)
{
  int nElements = AutoMesh_VolumeElement3D_nElements (self->element);
  AutoRes_ListTable *faceTable;
  int hashTableSize;
  int nAverageLinksPerList;
  int iElement;
  int iElementFace;
  int flag;

  assert(self->polyhedronType == AUTOMESH_POLYHEDRON_TETRA);

  /* NEED TO REDEFINE */
  /* experience-based value */
  flag = AutoRes_Table_reserveRecords (self->faces, nElements * 3);
  if (!flag) return 0;
  
  flag = AutoMesh_VolumeFace3D_makeVertexNodeIdsColumn (self, 3);
  if (!flag) return 0;

  if (elementIdsFlag) {
    flag = AutoMesh_VolumeFace3D_makeElementIdsColumn (self);
    if (!flag) return 0;
  }

  flag = AutoMesh_VolumeFace3D_makeFaceIdsAtElementColumn (self);
  if (!flag) return 0;

  faceTable = AutoRes_ListTable_New ();
  assert(faceTable != NULL);

  /* NEED TO REDEFINE */
  /* experience-based value */
  hashTableSize = nElements * 2;
  nAverageLinksPerList = 2;
  flag = AutoRes_ListTable_reserve 
    (faceTable, hashTableSize, nAverageLinksPerList);
  if (!flag) return 0;

  for (iElement = 0; iElement < nElements; iElement++) {
    for (iElementFace = 0; 
	 iElementFace < self->nElementFaces; 
	 iElementFace++) {
      int elementNodeNum0 = AutoMesh_Tetra_FaceVertexMap[iElementFace][0];
      int elementNodeNum1 = AutoMesh_Tetra_FaceVertexMap[iElementFace][1];
      int elementNodeNum2 = AutoMesh_Tetra_FaceVertexMap[iElementFace][2];

      int nodeId0 = AutoMesh_VolumeElement3D_nodeIdAt 
	(self->element, iElement, elementNodeNum0);
      int nodeId1 = AutoMesh_VolumeElement3D_nodeIdAt 
	(self->element, iElement, elementNodeNum1);
      int nodeId2 = AutoMesh_VolumeElement3D_nodeIdAt 
	(self->element, iElement, elementNodeNum2);

      int faceId = FindTriangleFace 
	(self, nodeId0, nodeId1, nodeId2, faceTable);

      if (faceId == -1) {
	faceId = AddTriangleFace 
	  (self, nodeId0, nodeId1, nodeId2, faceTable,
	   elementIdsFlag);
	if (faceId == -1) return 0;
      }
      SetElementFace (self, iElement, iElementFace, faceId,
		      elementIdsFlag);
    }
  }

  AutoRes_ListTable_delete (faceTable);

  return 1;
}

int /* boolean */
AutoMesh_VolumeFace3D_extract 
(AutoMesh_VolumeFace3D *self,
 int elementIdsFlag)
{
  int flag = 0;

  assert(AutoRes_Table_nRecords (self->faces) == 0);
  assert(self->faces_vertexNodeIds == NULL);
  assert(self->faces_elementIds == NULL);
  assert(self->faces_elementFaceNums == NULL);
  assert(self->elements_faceIds == NULL);

  switch (self->polyhedronType) {
  case AUTOMESH_POLYHEDRON_TETRA:
    flag = ExtractTetraFace (self, elementIdsFlag);
    break;
  case AUTOMESH_POLYHEDRON_HEXA:
    /* NEED TO REDEFINE */
    assert(0);
    break;
  case AUTOMESH_POLYHEDRON_PRISM:
    /* NEED TO REDEFINE */
    assert(0);
    break;
  default:
    assert(0);
    break;
  }

  if (flag) {
    assert(0 < AutoRes_Table_nRecords (self->faces));
    assert(self->faces_vertexNodeIds != NULL);
    if (elementIdsFlag) {
      assert(self->faces_elementIds != NULL);
      assert(self->faces_elementFaceNums != NULL);
    }
    assert(self->elements_faceIds != NULL);
  }

  return flag;
}

void AutoMesh_VolumeFace3D_purgeWorkArea 
(AutoMesh_VolumeFace3D *self)
{
  if (self->faces_vertexNodeIds != NULL) {
    AutoRes_Table_removeColumn 
      (self->faces, 
       AutoRes_Table_columnIdOfColumn 
       (self->faces, 
	self->faces_vertexNodeIds));
    self->faces_vertexNodeIds = NULL;
  }

  assert(self->faces_vertexNodeIds == NULL);
}

int AutoMesh_VolumeFace3D_collectNeighbor 
(AutoMesh_VolumeFace3D *self)
{
  int nElements = AutoMesh_VolumeElement3D_nElements (self->element);
  int iElement;
  int iElementFace;
  int flag;
  
  assert(0 < AutoRes_Table_nRecords (self->faces));
  assert(self->faces_elementIds != NULL);
  assert(self->elements_faceIds != NULL);
  assert(self->element->elements_neighborIds == NULL);

  flag = AutoMesh_VolumeElement3D_makeNeighborIdsColumn (self->element);
  if (!flag) return 0;

  for (iElement = 0; iElement < nElements; iElement++) {
    for (iElementFace = 0; 
	 iElementFace < self->nElementFaces; 
	 iElementFace++) {
      int faceId = AUTORES_I1D(self->elements_faceIds, iElement, iElementFace);
      int neighborId;
      
      if (AUTORES_I1D(self->faces_elementIds, faceId, 0) == iElement) {
	neighborId = AUTORES_I1D(self->faces_elementIds, faceId, 1);
      } else {
	assert(AUTORES_I1D(self->faces_elementIds, faceId, 1) == iElement);
	neighborId = AUTORES_I1D(self->faces_elementIds, faceId, 0);
      }
      AutoMesh_VolumeElement3D_setNeighborIdAt 
	(self->element, iElement, iElementFace, 
	 neighborId);
    }
  }

  assert(self->element->elements_neighborIds != NULL);

  return 1;
}

int AutoMesh_VolumeFace3D_nFaces 
(AutoMesh_VolumeFace3D *self)
{
  return AutoRes_Table_nRecords (self->faces);
}

void AutoMesh_VolumeFace3D_getFaceAt 
(AutoMesh_VolumeFace3D *self,
 int faceId,
 int elementIds_OUT[2],
 int elementFaceNums_OUT[2])
{
  elementIds_OUT[0] = AUTORES_I1D(self->faces_elementIds, faceId, 0);
  elementIds_OUT[1] = AUTORES_I1D(self->faces_elementIds, faceId, 1);

  elementFaceNums_OUT[0] = AUTORES_I1D(self->faces_elementFaceNums, faceId, 0);
  elementFaceNums_OUT[1] = AUTORES_I1D(self->faces_elementFaceNums, faceId, 1);
}

int /* boolean */
AutoMesh_VolumeFace3D_isOnBoundaryAt 
(AutoMesh_VolumeFace3D *self,
 int faceId)
{
  int elementId1 = AUTORES_I1D(self->faces_elementIds, faceId, 1);
  
  return (elementId1 == -1);
}

int /* boolean */
AutoMesh_VolumeFace3D_getBoundaryAt 
(AutoMesh_VolumeFace3D *self,
 int faceId,
 int *elementId_OUT,
 int *elementFaceNum_OUT)
{
  int elementId1 = AUTORES_I1D(self->faces_elementIds, faceId, 1);

  if (elementId1 == -1) {
    *elementId_OUT = AUTORES_I1D(self->faces_elementIds, faceId, 0);
    *elementFaceNum_OUT = AUTORES_I1D(self->faces_elementFaceNums, faceId, 0);
    return 1;
  }

  return 0;
}

int AutoMesh_VolumeFace3D_faceIdAtElement 
(AutoMesh_VolumeFace3D *self,
 int elementId,
 int elementFaceNum)
{
  return AUTORES_I1D(self->elements_faceIds, elementId, elementFaceNum);
}





