/* automesh_volume_isosurface3d.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"
#include "automesh_shape.h"
#include "automesh_node3d.h"
#include "automesh_node_scalar3d.h"
#include "automesh_volume_element3d.h"
#include "automesh_volume_isosurface3d.h"



#define TABLE_NAME_ISOSURFACES "AutoMesh:isosurfaces"
#define COLUMN_NAME_ELEMENT_ID "AutoMesh:elementId"
#define COLUMN_NAME_ELEMENT_EDGE_NUMS "AutoMesh:elementEdgeNums"
#define COLUMN_NAME_RATIOS "AutoMesh:ratios"



/* VolumeIsosurface3D */

AUTORES_IMPLEMENT_HANDLE(AutoMesh_VolumeIsosurface3D)

AutoMesh_VolumeIsosurface3D *AutoMesh_VolumeIsosurface3D_New 
(AutoMesh_VolumeElement3D *element,
 AutoMesh_NodeScalar3D *scalar)
{
  AutoMesh_VolumeIsosurface3D *isosurface;

  isosurface = AutoMesh_VolumeIsosurface3D_NewImplementation ();

  isosurface->node = element->node;
  isosurface->element = element;
  isosurface->scalar = scalar;

  isosurface->isosurfaces = AutoRes_Table_New (TABLE_NAME_ISOSURFACES);
  assert(isosurface->isosurfaces != NULL);
  isosurface->isosurfaces_elementId = NULL;
  isosurface->isosurfaces_elementEdgeNums = NULL;
  isosurface->isosurfaces_ratios = NULL;

  return isosurface;
}

void AutoMesh_VolumeIsosurface3D_delete 
(AutoMesh_VolumeIsosurface3D *self)
{
  AutoRes_Table_delete (self->isosurfaces);
  self->isosurfaces = NULL;
  self->isosurfaces_elementId = NULL;
  self->isosurfaces_elementEdgeNums = NULL;
  self->isosurfaces_ratios = NULL;

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoMesh_VolumeIsosurface3D_makeColumn 
(AutoMesh_VolumeIsosurface3D *self)
{
  int flag;

  self->isosurfaces_elementId = AutoRes_IntegerColumn_New 
    (self->isosurfaces, COLUMN_NAME_ELEMENT_ID);
  assert(self->isosurfaces_elementId != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->isosurfaces, self->isosurfaces_elementId);
  if (!flag) return 0;
  
  self->isosurfaces_elementEdgeNums = AutoRes_Integer1DColumn_New 
    (self->isosurfaces, COLUMN_NAME_ELEMENT_EDGE_NUMS, 3);
  assert(self->isosurfaces_elementEdgeNums != NULL);
  flag = AutoRes_Table_addInteger1DColumn 
    (self->isosurfaces, self->isosurfaces_elementEdgeNums);
  if (!flag) return 0;

  self->isosurfaces_ratios = AutoRes_Real1DColumn_New 
    (self->isosurfaces, COLUMN_NAME_RATIOS, 3);
  assert(self->isosurfaces_ratios != NULL);
  flag = AutoRes_Table_addReal1DColumn 
    (self->isosurfaces, self->isosurfaces_ratios);
  if (!flag) return 0;

  return 1;
}

static int AddIsosurface 
(AutoMesh_VolumeIsosurface3D *self,
 int elementId, 
 int elementEdgeNum0, double ratio0,
 int elementEdgeNum1, double ratio1,
 int elementEdgeNum2, double ratio2)
{
  int isosurfaceId;

  isosurfaceId = AutoRes_Table_addRecord (self->isosurfaces);
  if (isosurfaceId == -1) return -1;

  AUTORES_I_SET(self->isosurfaces_elementId, isosurfaceId, elementId);

  AUTORES_I1D_SET(self->isosurfaces_elementEdgeNums, 
		  isosurfaceId, 0, elementEdgeNum0);
  AUTORES_I1D_SET(self->isosurfaces_elementEdgeNums, 
		  isosurfaceId, 1, elementEdgeNum1);
  AUTORES_I1D_SET(self->isosurfaces_elementEdgeNums, 
		  isosurfaceId, 2, elementEdgeNum2);

  AUTORES_R1D_SET(self->isosurfaces_ratios, isosurfaceId, 0, ratio0);
  AUTORES_R1D_SET(self->isosurfaces_ratios, isosurfaceId, 1, ratio1);
  AUTORES_R1D_SET(self->isosurfaces_ratios, isosurfaceId, 2, ratio2);

  return isosurfaceId;
}

static int GetTetraIsosurface 
(AutoMesh_VolumeIsosurface3D *self,
 int elementId, 
 double levelValue,
 int *nEdges_OUT,
 int elementEdgeNums_OUT[4], double ratios_OUT[4])
{
  int nodeId0 = AutoMesh_VolumeElement3D_nodeIdAt (self->element, elementId, 0);
  int nodeId1 = AutoMesh_VolumeElement3D_nodeIdAt (self->element, elementId, 1);
  int nodeId2 = AutoMesh_VolumeElement3D_nodeIdAt (self->element, elementId, 2);
  int nodeId3 = AutoMesh_VolumeElement3D_nodeIdAt (self->element, elementId, 3);
  
  double vertexValues[4];

  vertexValues[0] = AUTORES_R(self->scalar->nodes_value, nodeId0);
  vertexValues[1] = AUTORES_R(self->scalar->nodes_value, nodeId1);
  vertexValues[2] = AUTORES_R(self->scalar->nodes_value, nodeId2);
  vertexValues[3] = AUTORES_R(self->scalar->nodes_value, nodeId3);

  return AutoMesh_Tetra_GetIsosurface 
    (vertexValues, levelValue,
     nEdges_OUT, elementEdgeNums_OUT, ratios_OUT);
}

static int /* boolean */
CheckTetraIsosurface 
(AutoMesh_VolumeIsosurface3D *self,
 int elementId, 
 double levelValue)
{
  int nEdges;
  int elementEdgeNums[4];
  double ratios[4];
  int isosurfaceId;
  
  if (GetTetraIsosurface
      (self, 
       elementId, levelValue,
       &nEdges, elementEdgeNums, ratios)) {
    assert(nEdges == 3 || nEdges == 4);

    isosurfaceId = AddIsosurface 
      (self, 
       elementId, 
       elementEdgeNums[0], ratios[0],
       elementEdgeNums[1], ratios[1],
       elementEdgeNums[2], ratios[2]);
    if (isosurfaceId == -1) return 0;

    if (nEdges == 4) {
      isosurfaceId = AddIsosurface 
	(self, 
	 elementId, 
	 elementEdgeNums[2], ratios[2],
	 elementEdgeNums[3], ratios[3],
	 elementEdgeNums[0], ratios[0]);
      if (isosurfaceId == -1) return 0;
    }
  }

  return 1;
}

static int /* boolean */
ExtractTetraIsosurface
(AutoMesh_VolumeIsosurface3D *self,
 double levelValue)
{
  int nElements = AutoMesh_VolumeElement3D_nElements (self->element);
  int iElement;
  int flag;

  {
    /* NEED TO REDEFINE */
    /* experience-based value */
    int nIsosurfaces = nElements / 4;
    
    if (nIsosurfaces < 10) nIsosurfaces = 10;
    flag = AutoRes_Table_reserveRecords (self->isosurfaces, nIsosurfaces);
    if (!flag) return 0;
  }

  for (iElement = 0; iElement < nElements; iElement++) {
    flag = CheckTetraIsosurface (self, iElement, levelValue);
    if (!flag) return 0;
  }

  return 1;
}

int /* boolean */
AutoMesh_VolumeIsosurface3D_extract 
(AutoMesh_VolumeIsosurface3D *self,
 double levelValue)
{
  int flag = 0;
  
  assert(self->isosurfaces_elementId != NULL);
  assert(self->isosurfaces_elementEdgeNums != NULL);
  assert(self->isosurfaces_ratios != NULL);

  switch (self->element->polyhedronType) {
  case AUTOMESH_POLYHEDRON_TETRA:
    flag = ExtractTetraIsosurface (self, levelValue);
    break;
  case AUTOMESH_POLYHEDRON_HEXA:
    /* NEED TO REDEFINE */
    assert(0);
    break;
  case AUTOMESH_POLYHEDRON_PRISM:
    /* NEED TO REDEFINE */
    assert(0);
    break;
  default:
    assert(0);
    break;
  }

  return flag;
}

int AutoMesh_VolumeIsosurface3D_nIsosurfaces 
(AutoMesh_VolumeIsosurface3D *self)
{
  return AutoRes_Table_nRecords (self->isosurfaces);
}

void AutoMesh_VolumeIsosurface3D_getIsosurfaceAt 
(AutoMesh_VolumeIsosurface3D *self,
 int isosurfaceId,
 int *elementId_OUT,
 int *elementEdgeNum0_OUT, double *ratio0_OUT, 
 int *elementEdgeNum1_OUT, double *ratio1_OUT, 
 int *elementEdgeNum2_OUT, double *ratio2_OUT)
{
  *elementId_OUT = AUTORES_I(self->isosurfaces_elementId, isosurfaceId);

  *elementEdgeNum0_OUT = AUTORES_I1D
    (self->isosurfaces_elementEdgeNums, isosurfaceId, 0);
  *elementEdgeNum1_OUT = AUTORES_I1D
    (self->isosurfaces_elementEdgeNums, isosurfaceId, 1);
  *elementEdgeNum2_OUT = AUTORES_I1D
    (self->isosurfaces_elementEdgeNums, isosurfaceId, 2);

  *ratio0_OUT = AUTORES_R1D(self->isosurfaces_ratios, isosurfaceId, 0);
  *ratio1_OUT = AUTORES_R1D(self->isosurfaces_ratios, isosurfaceId, 1);
  *ratio2_OUT = AUTORES_R1D(self->isosurfaces_ratios, isosurfaceId, 2);
}

static void GetEdgePoint 
(AutoMesh_VolumeIsosurface3D *self,
 int elementId, 
 int elementEdgeNum, double ratio,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  int elementNodeNum0 = AutoMesh_Polyhedron_EdgeVertex
    (self->element->polyhedronType, elementEdgeNum, 0);
  int elementNodeNum1 = AutoMesh_Polyhedron_EdgeVertex
    (self->element->polyhedronType, elementEdgeNum, 1);
  
  int nodeId0 = AutoMesh_VolumeElement3D_nodeIdAt 
    (self->element, elementId, elementNodeNum0);
  int nodeId1 = AutoMesh_VolumeElement3D_nodeIdAt 
    (self->element, elementId, elementNodeNum1);

  double x0, y0, z0;
  double x1, y1, z1;

  AutoMesh_Node3D_getCoordAt (self->node, nodeId0,
			      &x0, &y0, &z0);
  AutoMesh_Node3D_getCoordAt (self->node, nodeId1,
			      &x1, &y1, &z1);
  
  *x_OUT = x0 + (x1 - x0) * ratio;
  *y_OUT = y0 + (y1 - y0) * ratio;
  *z_OUT = z0 + (z1 - z0) * ratio;
}

void AutoMesh_VolumeIsosurface3D_getIsosurfacePointAt 
(AutoMesh_VolumeIsosurface3D *self,
 int isosurfaceId,
 double *x0_OUT, double *y0_OUT, double *z0_OUT, 
 double *x1_OUT, double *y1_OUT, double *z1_OUT, 
 double *x2_OUT, double *y2_OUT, double *z2_OUT)
{
  int elementId;
  int elementEdgeNum0, elementEdgeNum1, elementEdgeNum2;
  double ratio0, ratio1, ratio2;
  
  AutoMesh_VolumeIsosurface3D_getIsosurfaceAt 
    (self, isosurfaceId,
     &elementId,
     &elementEdgeNum0, &ratio0,
     &elementEdgeNum1, &ratio1,
     &elementEdgeNum2, &ratio2);
  
  GetEdgePoint (self, elementId, elementEdgeNum0, ratio0,
		x0_OUT, y0_OUT, z0_OUT);
  GetEdgePoint (self, elementId, elementEdgeNum1, ratio1,
		x1_OUT, y1_OUT, z1_OUT);
  GetEdgePoint (self, elementId, elementEdgeNum2, ratio2,
		x2_OUT, y2_OUT, z2_OUT);
}

