/* automesh_volume_particle3d.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"
#include "automesh_shape.h"
#include "automesh_node3d.h"
#include "automesh_node_vector3d.h"
#include "automesh_volume_element3d.h"
#include "automesh_volume_particle3d.h"



#define TABLE_NAME_PARTICLES "AutoMesh:particles"
#define COLUMN_NAME_START_COORD "AutoMesh:startCoord"
#define TABLE_NAME_TRACES "AutoMesh:traces"
#define COLUMN_NAME_COORD "AutoMesh:coord"
#define COLUMN_NAME_ELEMENT_ID "AutoMesh:elementId"
#define COLUMN_NAME_LOCAL_COORD "AutoMesh:localCoord"



/* VolumeParticle3D */

AUTORES_IMPLEMENT_HANDLE(AutoMesh_VolumeParticle3D)

AutoMesh_VolumeParticle3D *AutoMesh_VolumeParticle3D_New 
(AutoMesh_VolumeElement3D *element)
{
  AutoMesh_VolumeParticle3D *particle;

  particle = AutoMesh_VolumeParticle3D_NewImplementation ();

  particle->node = element->node;
  particle->element = element;
  particle->nParticles = 0;
  particle->nSteps = 0;
  particle->deltaT = 1.0;

  particle->particles = AutoRes_Table_New (TABLE_NAME_PARTICLES);
  assert(particle->particles != NULL);
  particle->particles_startCoord = NULL;

  particle->traces = AutoRes_Table_New (TABLE_NAME_TRACES);
  assert(particle->traces != NULL);
  particle->traces_coord = NULL;
  particle->traces_elementId = NULL;
  particle->traces_localCoord = NULL;

  return particle;
}

void AutoMesh_VolumeParticle3D_delete 
(AutoMesh_VolumeParticle3D *self)
{
  AutoRes_Table_delete (self->particles);
  self->particles = NULL;
  self->particles_startCoord = NULL;

  AutoRes_Table_delete (self->traces);
  self->traces = NULL;
  self->traces_coord = NULL;
  self->traces_elementId = NULL;
  self->traces_localCoord = NULL;

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoMesh_VolumeParticle3D_makeColumn 
(AutoMesh_VolumeParticle3D *self)
{
  int flag;

  self->particles_startCoord = AutoRes_Real1DColumn_New 
    (self->particles, COLUMN_NAME_START_COORD, 3);
  assert(self->particles_startCoord != NULL);
  flag = AutoRes_Table_addReal1DColumn 
    (self->particles, self->particles_startCoord);
  if (!flag) return 0;

  self->traces_coord = AutoRes_Real1DColumn_New 
    (self->traces, COLUMN_NAME_COORD, 3);
  assert(self->traces_coord != NULL);
  flag = AutoRes_Table_addReal1DColumn 
    (self->traces, self->traces_coord);
  if (!flag) return 0;

  self->traces_elementId = AutoRes_IntegerColumn_New 
    (self->traces, COLUMN_NAME_ELEMENT_ID);
  assert(self->traces_elementId != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->traces, self->traces_elementId);
  if (!flag) return 0;
  
  self->traces_localCoord = AutoRes_Real1DColumn_New 
    (self->traces, COLUMN_NAME_LOCAL_COORD, 3);
  assert(self->traces_localCoord != NULL);
  flag = AutoRes_Table_addReal1DColumn 
    (self->traces, self->traces_localCoord);
  if (!flag) return 0;

  return 1;
}

void AutoMesh_VolumeParticle3D_allocateParticle 
(AutoMesh_VolumeParticle3D *self,
 int nParticles,
 int nSteps)
{
  assert(0 < nParticles);
  assert(0 < nSteps);

  self->nParticles = nParticles;
  self->nSteps = nSteps;
  AutoRes_Table_allocateRecords (self->particles, nParticles);
  AutoRes_Table_allocateRecords (self->traces, nParticles * nSteps);
}

void AutoMesh_VolumeParticle3D_setParticleStartPointAt 
(AutoMesh_VolumeParticle3D *self,
 int particleId,
 double x, double y, double z)
{
  AUTORES_R1D_SET(self->particles_startCoord, particleId, 0, x);
  AUTORES_R1D_SET(self->particles_startCoord, particleId, 1, y);
  AUTORES_R1D_SET(self->particles_startCoord, particleId, 2, z);
}

static void SetTraceAt 
(AutoMesh_VolumeParticle3D *self,
 int particleId,
 int step,
 double x, double y, double z,
 int elementId,
 double lx, double ly, double lz)
{
  int traceId = particleId * self->nSteps + step;

  assert(0 <= particleId);
  assert(particleId < self->nParticles);
  assert(0 <= step);
  assert(step < self->nSteps);

  AUTORES_R1D_SET(self->traces_coord, traceId, 0, x);
  AUTORES_R1D_SET(self->traces_coord, traceId, 1, y);
  AUTORES_R1D_SET(self->traces_coord, traceId, 2, z);

  AUTORES_I_SET(self->traces_elementId, traceId, elementId);

  AUTORES_R1D_SET(self->traces_localCoord, traceId, 0, lx);
  AUTORES_R1D_SET(self->traces_localCoord, traceId, 1, ly);
  AUTORES_R1D_SET(self->traces_localCoord, traceId, 2, lz);
}

void AutoMesh_VolumeParticle3D_traceParticle  
(AutoMesh_VolumeParticle3D *self,
 AutoMesh_NodeVector3D *vector,
 double deltaT)
{
  int iParticle;
  int iStep;

  self->deltaT = deltaT;

  iStep = 0;
  for (iParticle = 0; iParticle < self->nParticles; iParticle++) {
    double x = AUTORES_R1D(self->particles_startCoord, iParticle, 0);
    double y = AUTORES_R1D(self->particles_startCoord, iParticle, 1);
    double z = AUTORES_R1D(self->particles_startCoord, iParticle, 2);
    int elementId;
    double lx, ly, lz;

    elementId = AutoMesh_VolumeElement3D_findElementUsingBucket 
      (self->element, x, y, z,
       &lx, &ly, &lz);
    if (elementId != -1) {
      SetTraceAt (self, iParticle, iStep,
		  x, y, z, 
		  elementId, lx, ly, lz);
    } else {
      SetTraceAt (self, iParticle, iStep,
		  x, y, z, 
		  -1, 0.0, 0.0, 0.0);
    }
  }

  for (iStep = 1; iStep < self->nSteps; iStep++) {
    for (iParticle = 0; iParticle < self->nParticles; iParticle++) {
      double x, y, z;
      int elementId;
      double lx, ly, lz;

      AutoMesh_VolumeParticle3D_getTraceAt 
	(self, iParticle, iStep - 1,
	 &x, &y, &z,
	 &elementId,
	 &lx, &ly, &lz);
      
      if (elementId != -1) {
	double vectorX, vectorY, vectorZ;

	AutoMesh_VolumeElement3D_getVectorAt
	  (self->element, elementId, vector, lx, ly, lz,
	   &vectorX, &vectorY, &vectorZ);
	x += vectorX * self->deltaT;
	y += vectorY * self->deltaT;
	z += vectorZ * self->deltaT;

	elementId = AutoMesh_VolumeElement3D_findElementNear 
	  (self->element, elementId, x, y, z, 
	   &lx, &ly, &lz);
	if (elementId == -1) {
	  elementId = AutoMesh_VolumeElement3D_findElementUsingBucket 
	    (self->element, x, y, z,
	     &lx, &ly, &lz);
	}
      }

      if (elementId != -1) {
	SetTraceAt (self, iParticle, iStep,
		    x, y, z, 
		    elementId, lx, ly, lz);
      } else {
	SetTraceAt (self, iParticle, iStep,
		    x, y, z, 
		    -1, 0.0, 0.0, 0.0);
      }
    }
  }
}

int AutoMesh_VolumeParticle3D_nParticles
(AutoMesh_VolumeParticle3D *self)
{
  return self->nParticles;
}

int AutoMesh_VolumeParticle3D_nSteps
(AutoMesh_VolumeParticle3D *self)
{
  return self->nSteps;
}

void AutoMesh_VolumeParticle3D_getParticleStartPointAt  
(AutoMesh_VolumeParticle3D *self,
 int particleId,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  *x_OUT = AUTORES_R1D(self->particles_startCoord, particleId, 0);
  *y_OUT = AUTORES_R1D(self->particles_startCoord, particleId, 1);
  *z_OUT = AUTORES_R1D(self->particles_startCoord, particleId, 2);
}

void AutoMesh_VolumeParticle3D_getTraceAt 
(AutoMesh_VolumeParticle3D *self,
 int particleId,
 int step,
 double *x_OUT, double *y_OUT, double *z_OUT,
 int *elementId_OUT,
 double *lx_OUT, double *ly_OUT, double *lz_OUT)
{
  int traceId = particleId * self->nSteps + step;
  
  assert(0 <= particleId);
  assert(particleId < self->nParticles);
  assert(0 <= step);
  assert(step < self->nSteps);
  
  *x_OUT = AUTORES_R1D(self->traces_coord, traceId, 0);
  *y_OUT = AUTORES_R1D(self->traces_coord, traceId, 1);
  *z_OUT = AUTORES_R1D(self->traces_coord, traceId, 2);
  
  *elementId_OUT = AUTORES_I(self->traces_elementId, traceId);

  *lx_OUT = AUTORES_R1D(self->traces_localCoord, traceId, 0);
  *ly_OUT = AUTORES_R1D(self->traces_localCoord, traceId, 1);
  *lz_OUT = AUTORES_R1D(self->traces_localCoord, traceId, 2);
}

