/* automesh_volume_sample2d.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"
#include "automesh_shape.h"
#include "automesh_node2d.h"
#include "automesh_node_scalar2d.h"
#include "automesh_node_vector2d.h"
#include "automesh_volume_element2d.h"
#include "automesh_volume_sample2d.h"



#define TABLE_NAME_SAMPLES "AutoMesh:samples"
#define COLUMN_NAME_COORD "AutoMesh:coord"
#define COLUMN_NAME_ELEMENT_ID "AutoMesh:elementId"
#define COLUMN_NAME_LOCAL_COORD "AutoMesh:localCoord"



/* VolumeSample2D */

AUTORES_IMPLEMENT_HANDLE(AutoMesh_VolumeSample2D)

AutoMesh_VolumeSample2D *AutoMesh_VolumeSample2D_New 
(AutoMesh_VolumeElement2D *element)
{
  AutoMesh_VolumeSample2D *sample;

  sample = AutoMesh_VolumeSample2D_NewImplementation ();

  sample->node = element->node;
  sample->element = element;

  sample->samples = AutoRes_Table_New (TABLE_NAME_SAMPLES);
  assert(sample->samples != NULL);
  sample->samples_coord = NULL;
  sample->samples_elementId = NULL;
  sample->samples_localCoord = NULL;

  return sample;
}

void AutoMesh_VolumeSample2D_delete 
(AutoMesh_VolumeSample2D *self)
{
  AutoRes_Table_delete (self->samples);
  self->samples = NULL;
  self->samples_coord = NULL;
  self->samples_elementId = NULL;
  self->samples_localCoord = NULL;

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoMesh_VolumeSample2D_makeColumn 
(AutoMesh_VolumeSample2D *self)
{
  int flag;

  self->samples_coord = AutoRes_Real1DColumn_New 
    (self->samples, COLUMN_NAME_COORD, 2);
  assert(self->samples_coord != NULL);
  flag = AutoRes_Table_addReal1DColumn 
    (self->samples, self->samples_coord);
  if (!flag) return 0;

  self->samples_elementId = AutoRes_IntegerColumn_New 
    (self->samples, COLUMN_NAME_ELEMENT_ID);
  assert(self->samples_elementId != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->samples, self->samples_elementId);
  if (!flag) return 0;
  
  self->samples_localCoord = AutoRes_Real1DColumn_New 
    (self->samples, COLUMN_NAME_LOCAL_COORD, 2);
  assert(self->samples_localCoord != NULL);
  flag = AutoRes_Table_addReal1DColumn 
    (self->samples, self->samples_localCoord);
  if (!flag) return 0;

  return 1;
}

int /* boolean */ 
AutoMesh_VolumeSample2D_reserveSamples 
(AutoMesh_VolumeSample2D *self,
 int nSamples)
{
  return AutoRes_Table_reserveRecords (self->samples, nSamples);
}

int AutoMesh_VolumeSample2D_addSample 
(AutoMesh_VolumeSample2D *self,
 double x, double y)
{
  int sampleId;

  sampleId = AutoRes_Table_addRecord (self->samples);
  if (sampleId == -1) return -1;

  AUTORES_R1D_SET(self->samples_coord, sampleId, 0, x);
  AUTORES_R1D_SET(self->samples_coord, sampleId, 1, y);

  AUTORES_I_SET(self->samples_elementId, sampleId, -1);

  AUTORES_R1D_SET(self->samples_localCoord, sampleId, 0, 0.0);
  AUTORES_R1D_SET(self->samples_localCoord, sampleId, 1, 0.0);

  return sampleId;
}

void AutoMesh_VolumeSample2D_setElementOnSample 
(AutoMesh_VolumeSample2D *self)
{
  int nSamples = AutoRes_Table_nRecords (self->samples);
  int iSample;

  assert(self->samples_coord != NULL);
  assert(self->samples_elementId != NULL);
  assert(self->samples_localCoord != NULL);
  assert(self->element->bucket != NULL);

  for (iSample = 0; iSample < nSamples; iSample++) {
    double x = AUTORES_R1D(self->samples_coord, iSample, 0);
    double y = AUTORES_R1D(self->samples_coord, iSample, 1);
    
    int elementId;
    double lx, ly;

    elementId = AutoMesh_VolumeElement2D_findElementUsingBucket 
      (self->element, x, y,
       &lx, &ly);
    if (elementId != -1) {
      AUTORES_I_SET(self->samples_elementId, iSample, elementId);
      
      AUTORES_R1D_SET(self->samples_localCoord, iSample, 0, lx);
      AUTORES_R1D_SET(self->samples_localCoord, iSample, 1, ly);
    }
  }
}

int AutoMesh_VolumeSample2D_nSamples 
(AutoMesh_VolumeSample2D *self)
{
  return AutoRes_Table_nRecords (self->samples);
}

void AutoMesh_VolumeSample2D_getSamplePointAt 
(AutoMesh_VolumeSample2D *self,
 int sampleId,
 double *x_OUT, double *y_OUT)
{
  *x_OUT = AUTORES_R1D(self->samples_coord, sampleId, 0);
  *y_OUT = AUTORES_R1D(self->samples_coord, sampleId, 1);
}

void AutoMesh_VolumeSample2D_getSampleAt 
(AutoMesh_VolumeSample2D *self,
 int sampleId,
 int *elementId_OUT,
 double *lx_OUT, double *ly_OUT)
{
  *elementId_OUT = AUTORES_I(self->samples_elementId, sampleId);

  *lx_OUT = AUTORES_R1D(self->samples_localCoord, sampleId, 0);
  *ly_OUT = AUTORES_R1D(self->samples_localCoord, sampleId, 1);
}

double AutoMesh_VolumeSample2D_scalarAt 
(AutoMesh_VolumeSample2D *self,
 int sampleId,
 AutoMesh_NodeScalar2D *scalar)
{
  int elementId;
  double lx, ly;

  AutoMesh_VolumeSample2D_getSampleAt (self, sampleId,
				       &elementId, 
				       &lx, &ly);
  return AutoMesh_VolumeElement2D_scalarAt 
    (self->element, elementId, scalar, lx, ly);
}

void AutoMesh_VolumeSample2D_getVectorAt 
(AutoMesh_VolumeSample2D *self,
 int sampleId,
 AutoMesh_NodeVector2D *vector,
 double *valueX_OUT, double *valueY_OUT)
{
  int elementId;
  double lx, ly;

  AutoMesh_VolumeSample2D_getSampleAt (self, sampleId,
				       &elementId, 
				       &lx, &ly);
  AutoMesh_VolumeElement2D_getVectorAt 
    (self->element, elementId, vector, lx, ly,
     valueX_OUT, valueY_OUT);
}
