/* automesh_volume_section3d.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"
#include "automesh_shape.h"
#include "automesh_node3d.h"
#include "automesh_node_scalar3d.h"
#include "automesh_volume_element3d.h"
#include "automesh_volume_section3d.h"



#define TABLE_NAME_SECTIONS "AutoMesh:sections"
#define COLUMN_NAME_ELEMENT_ID "AutoMesh:elementId"
#define COLUMN_NAME_ELEMENT_EDGE_NUMS "AutoMesh:elementEdgeNums"
#define COLUMN_NAME_RATIOS "AutoMesh:ratios"
#define COLUMN_NAME_SECTION_VALUE "AutoMesh:sectionValue"



/* VolumeSection3D */

AUTORES_IMPLEMENT_HANDLE(AutoMesh_VolumeSection3D)

AutoMesh_VolumeSection3D *AutoMesh_VolumeSection3D_New 
(AutoMesh_VolumeElement3D *element)
{
  AutoMesh_VolumeSection3D *section;

  section = AutoMesh_VolumeSection3D_NewImplementation ();

  section->node = element->node;
  section->nodes_sectionValue = NULL;

  section->element = element;

  section->a = 0.0;
  section->b = 0.0;
  section->c = 1.0;
  section->d = 0.0;

  section->sections = AutoRes_Table_New (TABLE_NAME_SECTIONS);
  assert(section->sections != NULL);
  section->sections_elementId = NULL;
  section->sections_elementEdgeNums = NULL;
  section->sections_ratios = NULL;

  return section;
}

void AutoMesh_VolumeSection3D_delete 
(AutoMesh_VolumeSection3D *self)
{
  AutoRes_Table_delete (self->sections);
  self->sections = NULL;
  self->sections_elementId = NULL;
  self->sections_elementEdgeNums = NULL;
  self->sections_ratios = NULL;

  if (self->nodes_sectionValue != NULL) {
    AutoRes_Table_removeColumn 
      (self->node->nodes, 
       AutoRes_Table_columnIdOfColumn 
       (self->node->nodes, 
	self->nodes_sectionValue));
    self->nodes_sectionValue = NULL;
  }

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoMesh_VolumeSection3D_makeColumn 
(AutoMesh_VolumeSection3D *self)
{
  int flag;

  self->sections_elementId = AutoRes_IntegerColumn_New 
    (self->sections, COLUMN_NAME_ELEMENT_ID);
  assert(self->sections_elementId != NULL);
  flag = AutoRes_Table_addIntegerColumn 
    (self->sections, self->sections_elementId);
  if (!flag) return 0;
  
  self->sections_elementEdgeNums = AutoRes_Integer1DColumn_New 
    (self->sections, COLUMN_NAME_ELEMENT_EDGE_NUMS, 3);
  assert(self->sections_elementEdgeNums != NULL);
  flag = AutoRes_Table_addInteger1DColumn 
    (self->sections, self->sections_elementEdgeNums);
  if (!flag) return 0;

  self->sections_ratios = AutoRes_Real1DColumn_New 
    (self->sections, COLUMN_NAME_RATIOS, 3);
  assert(self->sections_ratios != NULL);
  flag = AutoRes_Table_addReal1DColumn 
    (self->sections, self->sections_ratios);
  if (!flag) return 0;

  self->nodes_sectionValue = AutoRes_RealColumn_New 
    (self->node->nodes, COLUMN_NAME_SECTION_VALUE);
  assert(self->nodes_sectionValue != NULL);
  flag = AutoRes_Table_addRealColumn 
    (self->node->nodes, self->nodes_sectionValue);
  if (!flag) return 0;

  return 1;
}

void AutoMesh_VolumeSection3D_setEquation 
(AutoMesh_VolumeSection3D *self,
 double a, double b, double c, double d)
{
  int nNodes = AutoMesh_Node3D_nNodes (self->node);
  int iNode;

  self->a = a;
  self->b = b;
  self->c = c;
  self->d = d;

  assert(self->nodes_sectionValue != NULL);

  for (iNode = 0; iNode < nNodes; iNode++) {
    double x, y, z;
    double sectionValue;
    
    AutoMesh_Node3D_getCoordAt (self->node, iNode,
				&x, &y, &z);
    sectionValue = a * x + b * y + c * z + d;
    AUTORES_R_SET(self->nodes_sectionValue, iNode, sectionValue);
  }
}

static int AddSection 
(AutoMesh_VolumeSection3D *self,
 int elementId, 
 int elementEdgeNum0, double ratio0,
 int elementEdgeNum1, double ratio1,
 int elementEdgeNum2, double ratio2)
{
  int sectionId;

  sectionId = AutoRes_Table_addRecord (self->sections);
  if (sectionId == -1) return -1;

  AUTORES_I_SET(self->sections_elementId, sectionId, elementId);

  AUTORES_I1D_SET(self->sections_elementEdgeNums, 
		  sectionId, 0, elementEdgeNum0);
  AUTORES_I1D_SET(self->sections_elementEdgeNums, 
		  sectionId, 1, elementEdgeNum1);
  AUTORES_I1D_SET(self->sections_elementEdgeNums, 
		  sectionId, 2, elementEdgeNum2);

  AUTORES_R1D_SET(self->sections_ratios, sectionId, 0, ratio0);
  AUTORES_R1D_SET(self->sections_ratios, sectionId, 1, ratio1);
  AUTORES_R1D_SET(self->sections_ratios, sectionId, 2, ratio2);

  return sectionId;
}

static int GetTetraSection 
(AutoMesh_VolumeSection3D *self,
 int elementId, 
 double offset,
 int *nEdges_OUT,
 int elementEdgeNums_OUT[4], double ratios_OUT[4])
{
  int nodeId0 = AutoMesh_VolumeElement3D_nodeIdAt (self->element, elementId, 0);
  int nodeId1 = AutoMesh_VolumeElement3D_nodeIdAt (self->element, elementId, 1);
  int nodeId2 = AutoMesh_VolumeElement3D_nodeIdAt (self->element, elementId, 2);
  int nodeId3 = AutoMesh_VolumeElement3D_nodeIdAt (self->element, elementId, 3);
  
  double vertexValues[4];

  vertexValues[0] = AUTORES_R(self->nodes_sectionValue, nodeId0);
  vertexValues[1] = AUTORES_R(self->nodes_sectionValue, nodeId1);
  vertexValues[2] = AUTORES_R(self->nodes_sectionValue, nodeId2);
  vertexValues[3] = AUTORES_R(self->nodes_sectionValue, nodeId3);

  return AutoMesh_Tetra_GetIsosurface 
    (vertexValues, offset,
     nEdges_OUT, elementEdgeNums_OUT, ratios_OUT);
}

static int /* boolean */
CheckTetraSection 
(AutoMesh_VolumeSection3D *self,
 int elementId, 
 double offset)
{
  int nEdges;
  int elementEdgeNums[4];
  double ratios[4];
  int sectionId;
  
  if (GetTetraSection
      (self, 
       elementId, offset,
       &nEdges, elementEdgeNums, ratios)) {
    assert(nEdges == 3 || nEdges == 4);

    sectionId = AddSection 
      (self, 
       elementId, 
       elementEdgeNums[0], ratios[0],
       elementEdgeNums[1], ratios[1],
       elementEdgeNums[2], ratios[2]);
    if (sectionId == -1) return 0;

    if (nEdges == 4) {
      sectionId = AddSection 
	(self, 
	 elementId, 
	 elementEdgeNums[2], ratios[2],
	 elementEdgeNums[3], ratios[3],
	 elementEdgeNums[0], ratios[0]);
      if (sectionId == -1) return 0;
    }
  }

  return 1;
}

static int /* boolean */
ExtractTetraSection
(AutoMesh_VolumeSection3D *self,
 double offset)
{
  int nElements = AutoMesh_VolumeElement3D_nElements (self->element);
  int iElement;
  int flag;

  {
    /* NEED TO REDEFINE */
    /* experience-based value */
    int nSections = nElements / 4;
    
    if (nSections < 10) nSections = 10;
    flag = AutoRes_Table_reserveRecords (self->sections, nSections);
    if (!flag) return 0;
  }

  for (iElement = 0; iElement < nElements; iElement++) {
    flag = CheckTetraSection (self, iElement, offset);
    if (!flag) return 0;
  }

  return 1;
}

int /* boolean */
AutoMesh_VolumeSection3D_extract 
(AutoMesh_VolumeSection3D *self,
 double offset)
{
  int flag = 0;
  
  assert(self->sections_elementId != NULL);
  assert(self->sections_elementEdgeNums != NULL);
  assert(self->sections_ratios != NULL);
  assert(self->nodes_sectionValue != NULL);

  switch (self->element->polyhedronType) {
  case AUTOMESH_POLYHEDRON_TETRA:
    flag = ExtractTetraSection (self, offset);
    break;
  case AUTOMESH_POLYHEDRON_HEXA:
    /* NEED TO REDEFINE */
    assert(0);
    break;
  case AUTOMESH_POLYHEDRON_PRISM:
    /* NEED TO REDEFINE */
    assert(0);
    break;
  default:
    assert(0);
    break;
  }

  return flag;
}

int AutoMesh_VolumeSection3D_nSections 
(AutoMesh_VolumeSection3D *self)
{
  return AutoRes_Table_nRecords (self->sections);
}

void AutoMesh_VolumeSection3D_getSectionAt 
(AutoMesh_VolumeSection3D *self,
 int sectionId,
 int *elementId_OUT,
 int *elementEdgeNum0_OUT, double *ratio0_OUT, 
 int *elementEdgeNum1_OUT, double *ratio1_OUT, 
 int *elementEdgeNum2_OUT, double *ratio2_OUT)
{
  *elementId_OUT = AUTORES_I(self->sections_elementId, sectionId);

  *elementEdgeNum0_OUT = AUTORES_I1D
    (self->sections_elementEdgeNums, sectionId, 0);
  *elementEdgeNum1_OUT = AUTORES_I1D
    (self->sections_elementEdgeNums, sectionId, 1);
  *elementEdgeNum2_OUT = AUTORES_I1D
    (self->sections_elementEdgeNums, sectionId, 2);

  *ratio0_OUT = AUTORES_R1D(self->sections_ratios, sectionId, 0);
  *ratio1_OUT = AUTORES_R1D(self->sections_ratios, sectionId, 1);
  *ratio2_OUT = AUTORES_R1D(self->sections_ratios, sectionId, 2);
}

static void GetEdgePoint 
(AutoMesh_VolumeSection3D *self,
 int elementId, 
 int elementEdgeNum, double ratio,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  int elementNodeNum0 = AutoMesh_Polyhedron_EdgeVertex
    (self->element->polyhedronType, elementEdgeNum, 0);
  int elementNodeNum1 = AutoMesh_Polyhedron_EdgeVertex
    (self->element->polyhedronType, elementEdgeNum, 1);
  
  int nodeId0 = AutoMesh_VolumeElement3D_nodeIdAt 
    (self->element, elementId, elementNodeNum0);
  int nodeId1 = AutoMesh_VolumeElement3D_nodeIdAt 
    (self->element, elementId, elementNodeNum1);

  double x0, y0, z0;
  double x1, y1, z1;

  AutoMesh_Node3D_getCoordAt (self->node, nodeId0,
			      &x0, &y0, &z0);
  AutoMesh_Node3D_getCoordAt (self->node, nodeId1,
			      &x1, &y1, &z1);
  
  *x_OUT = x0 + (x1 - x0) * ratio;
  *y_OUT = y0 + (y1 - y0) * ratio;
  *z_OUT = z0 + (z1 - z0) * ratio;
}

void AutoMesh_VolumeSection3D_getSectionPointAt 
(AutoMesh_VolumeSection3D *self,
 int sectionId,
 double *x0_OUT, double *y0_OUT, double *z0_OUT, 
 double *x1_OUT, double *y1_OUT, double *z1_OUT, 
 double *x2_OUT, double *y2_OUT, double *z2_OUT)
{
  int elementId;
  int elementEdgeNum0, elementEdgeNum1, elementEdgeNum2;
  double ratio0, ratio1, ratio2;
  
  AutoMesh_VolumeSection3D_getSectionAt 
    (self, sectionId,
     &elementId,
     &elementEdgeNum0, &ratio0,
     &elementEdgeNum1, &ratio1,
     &elementEdgeNum2, &ratio2);
  
  GetEdgePoint (self, elementId, elementEdgeNum0, ratio0,
		x0_OUT, y0_OUT, z0_OUT);
  GetEdgePoint (self, elementId, elementEdgeNum1, ratio1,
		x1_OUT, y1_OUT, z1_OUT);
  GetEdgePoint (self, elementId, elementEdgeNum2, ratio2,
		x2_OUT, y2_OUT, z2_OUT);
}

static double EdgeScalar 
(AutoMesh_VolumeSection3D *self,
 int elementId, 
 int elementEdgeNum, double ratio,
 AutoMesh_NodeScalar3D *scalar)
{
  int elementNodeNum0 = AutoMesh_Polyhedron_EdgeVertex
    (self->element->polyhedronType, elementEdgeNum, 0);
  int elementNodeNum1 = AutoMesh_Polyhedron_EdgeVertex
    (self->element->polyhedronType, elementEdgeNum, 1);
  
  int nodeId0 = AutoMesh_VolumeElement3D_nodeIdAt 
    (self->element, elementId, elementNodeNum0);
  int nodeId1 = AutoMesh_VolumeElement3D_nodeIdAt 
    (self->element, elementId, elementNodeNum1);
  
  double scalar0 = AutoMesh_NodeScalar3D_valueAt (scalar, nodeId0);
  double scalar1 = AutoMesh_NodeScalar3D_valueAt (scalar, nodeId1);
  
  return scalar0 + (scalar1 - scalar0) * ratio;
}

void AutoMesh_VolumeSection3D_getSectionScalarAt 
(AutoMesh_VolumeSection3D *self,
 int sectionId,
 AutoMesh_NodeScalar3D *scalar,
 double *value0_OUT, double *value1_OUT, double *value2_OUT)
{
  int elementId;
  int elementEdgeNum0, elementEdgeNum1, elementEdgeNum2;
  double ratio0, ratio1, ratio2;
  
  AutoMesh_VolumeSection3D_getSectionAt 
    (self, sectionId,
     &elementId,
     &elementEdgeNum0, &ratio0,
     &elementEdgeNum1, &ratio1,
     &elementEdgeNum2, &ratio2);

  *value0_OUT = EdgeScalar (self, elementId, elementEdgeNum0, ratio0, scalar);
  *value1_OUT = EdgeScalar (self, elementId, elementEdgeNum1, ratio1, scalar);
  *value2_OUT = EdgeScalar (self, elementId, elementEdgeNum2, ratio2, scalar);
}

