/* test_volume_edge3d.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>

#include "automesh_base.h"
#include "automesh_shape.h"
#include "automesh_node3d.h"
#include "automesh_volume_element3d.h"
#include "automesh_volume_edge3d.h"
#include "automesh_initialize.h"



#define EQ(a,b)  fabs ((a) - (b)) < 0.00001

static int EdgeIsFound 
(AutoMesh_VolumeEdge3D *edge,
 int nodeId0, int nodeId1)
{
  int nEdges = AutoMesh_VolumeEdge3D_nEdges (edge);
  int iEdge;

  for (iEdge = 0; iEdge < nEdges; iEdge++) {
    int eNodeId0, eNodeId1;

    AutoMesh_VolumeEdge3D_getVertexNodeIdsAt (edge, iEdge,
					      &eNodeId0, &eNodeId1);
    if (AutoMesh_IdIsEqual2 (nodeId0, nodeId1,
			     eNodeId0, eNodeId1)) {
      return 1;
    }
  }
  return 0;
}

static void CheckElementEdge 
(AutoMesh_VolumeEdge3D *edge)
{
  int nElements = AutoMesh_VolumeElement3D_nElements (edge->element);
  int iElement;
  int iElementEdge;
  
  for (iElement = 0; iElement < nElements; iElement++) {
    for (iElementEdge = 0; 
	 iElementEdge < edge->nElementEdges; 
	 iElementEdge++) {
      int elementNodeNum0 = AutoMesh_Polyhedron_EdgeVertex
	(edge->polyhedronType, iElementEdge, 0);
      int elementNodeNum1 = AutoMesh_Polyhedron_EdgeVertex
	(edge->polyhedronType, iElementEdge, 1);
      
      int nodeId0 = AutoMesh_VolumeElement3D_nodeIdAt 
	(edge->element, iElement, elementNodeNum0);
      int nodeId1 = AutoMesh_VolumeElement3D_nodeIdAt 
	(edge->element, iElement, elementNodeNum1);
      
      int edgeId = AutoMesh_VolumeEdge3D_edgeIdAtElement 
	(edge, iElement, iElementEdge);
      int eNodeId0, eNodeId1;
      
      AutoMesh_VolumeEdge3D_getVertexNodeIdsAt (edge, edgeId,
						&eNodeId0, &eNodeId1);
      assert(AutoMesh_IdIsEqual2 (nodeId0, nodeId1,
				  eNodeId0, eNodeId1));
    }
  }
}

/* VolumeElement3D, 1st order 4 node tetrahedron */
static void Test0 (void)
{
  AutoMesh_Node3D *node;
  AutoMesh_VolumeElement3D *element;
  AutoMesh_VolumeEdge3D *edge;
  int nNodes;
  int nElements;
  int flag;



  printf (" Test0 \n"); 



  /* node */

  printf (" node \n"); 

  node = AutoMesh_Node3D_New ();

  nNodes = 6;

  flag = AutoMesh_Node3D_allocateNodes (node, nNodes);
  assert(flag);

  flag = AutoMesh_Node3D_makeCoordColumn (node);
  assert(flag);

  AutoMesh_Node3D_setCoordAt (node, 0,  0.0, 0.0, 0.0);
  AutoMesh_Node3D_setCoordAt (node, 1,  1.0, 0.0, 0.0);
  AutoMesh_Node3D_setCoordAt (node, 2,  0.0, 2.0, 0.0);
  AutoMesh_Node3D_setCoordAt (node, 3,  0.0, 0.0, 3.0);
  AutoMesh_Node3D_setCoordAt (node, 4,  1.0, 0.0, 3.0);
  AutoMesh_Node3D_setCoordAt (node, 5,  0.0, 2.0, 3.0);



  /* element */

  printf (" element \n"); 

  element = AutoMesh_VolumeElement3D_New 
    (node, AUTOMESH_POLYHEDRON_TETRA, 4);

  nElements = 3;

  flag = AutoMesh_VolumeElement3D_allocateElements (element, nElements);
  assert(flag);

  flag = AutoMesh_VolumeElement3D_makeNodeIdsColumn (element);
  assert(flag);

  AutoMesh_VolumeElement3D_setNodeIdAt (element, 0, 0, 0);
  AutoMesh_VolumeElement3D_setNodeIdAt (element, 0, 1, 1);
  AutoMesh_VolumeElement3D_setNodeIdAt (element, 0, 2, 2);
  AutoMesh_VolumeElement3D_setNodeIdAt (element, 0, 3, 3);

  AutoMesh_VolumeElement3D_setNodeIdAt (element, 1, 0, 1);
  AutoMesh_VolumeElement3D_setNodeIdAt (element, 1, 1, 2);
  AutoMesh_VolumeElement3D_setNodeIdAt (element, 1, 2, 3);
  AutoMesh_VolumeElement3D_setNodeIdAt (element, 1, 3, 4);

  AutoMesh_VolumeElement3D_setNodeIdAt (element, 2, 0, 2);
  AutoMesh_VolumeElement3D_setNodeIdAt (element, 2, 1, 3);
  AutoMesh_VolumeElement3D_setNodeIdAt (element, 2, 2, 4);
  AutoMesh_VolumeElement3D_setNodeIdAt (element, 2, 3, 5);



  /* edge */

  printf (" edge \n"); 

  edge = AutoMesh_VolumeEdge3D_New (element);

  flag = AutoMesh_VolumeEdge3D_extract (edge);
  assert(flag);

  assert(AutoMesh_VolumeEdge3D_nEdges (edge) == 12);
  assert(EdgeIsFound (edge, 0, 1));
  assert(EdgeIsFound (edge, 1, 2));
  assert(EdgeIsFound (edge, 2, 0));
  assert(EdgeIsFound (edge, 0, 3));
  assert(EdgeIsFound (edge, 1, 4));
  assert(EdgeIsFound (edge, 2, 5));
  assert(EdgeIsFound (edge, 3, 4));
  assert(EdgeIsFound (edge, 4, 5));
  assert(EdgeIsFound (edge, 5, 3));
  assert(EdgeIsFound (edge, 2, 3));
  assert(EdgeIsFound (edge, 3, 1));
  assert(EdgeIsFound (edge, 2, 4));
  
  CheckElementEdge (edge);

  AutoMesh_VolumeEdge3D_purgeWorkArea (edge);



  printf (" delete edge \n"); 

  AutoMesh_VolumeEdge3D_delete (edge);

  printf (" delete element \n"); 

  AutoMesh_VolumeElement3D_delete (element);

  printf (" delete node \n"); 

  AutoMesh_Node3D_delete (node);
}

int main (int argc, char *argv[])
{
  AutoRes_Initialize (0);
  AutoMesh_Initialize ();

  Test0 ();

  return 0;
}

