/* advautomesh_h_tetra3d_sc_crack.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001
#define AREA_COORD_EPS 0.00001

double Eps = DEFAULT_EPS;



double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}

void GetNormalizedVector 
(double x, double y, double z,
 double *nx_OUT, double *ny_OUT, double *nz_OUT)
{
  double length = sqrt (x * x + y * y + z * z);

  *nx_OUT = x / length;
  *ny_OUT = y / length;
  *nz_OUT = z / length;
}

void GetVectorProduct 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  *x_OUT = y0 * z1 - z0 * y1;
  *y_OUT = z0 * x1 - x0 * z1;
  *z_OUT = x0 * y1 - y0 * x1;
}

void GetPlaneEquationToward 
(double nx, double ny, double nz,
 double x, double y, double z,
 double *a_OUT, double *b_OUT, double *c_OUT, double *d_OUT)
{
  double nx2, ny2, nz2;

  GetNormalizedVector (nx, ny, nz,
		       &nx2, &ny2, &nz2);
  *a_OUT = nx2;
  *b_OUT = ny2;
  *c_OUT = nz2;
  *d_OUT = -(nx2 * x + ny2 * y + nz2 * z);
}

double TriArea
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double x2, double y2, double z2)
{
  double vx, vy, vz;
  
  GetVectorProduct (x1 - x0, y1 - y0, z1 - z0,
		    x2 - x0, y2 - y0, z2 - z0,
		    &vx, &vy, &vz);
  return sqrt (vx * vx + vy * vy + vz * vz) / 2;
}

int TriContainsPoint 
(double tx0, double ty0, double tz0, 
 double tx1, double ty1, double tz1, 
 double tx2, double ty2, double tz2, 
 double x, double y, double z)
{
  double area = TriArea (tx0, ty0, tz0, 
			 tx1, ty1, tz1, 
			 tx2, ty2, tz2);
  double area0 = TriArea (tx1, ty1, tz1, 
			  tx2, ty2, tz2, 
			  x, y, z);
  double area1 = TriArea (tx2, ty2, tz2, 
			  tx0, ty0, tz0, 
			  x, y, z);
  double area2 = TriArea (tx0, ty0, tz0, 
			  tx1, ty1, tz1, 
			  x, y, z);
  
  if (fabs ((area0 + area1 + area2) / area - 1.0) < AREA_COORD_EPS) {
    return 1;
  }
  return 0;
}



/************************
	 Mesh
*************************/

/* Node */

typedef enum {
  NODE_CRACK_OTHER,
  NODE_CRACK_ON_CRACK,
  NODE_CRACK_ON_CRACK_TIP
} NodeCrackType;

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][3];
NodeCrackType Node_crackType[MAX_NODES];



/* Element */

#define N_ELEMENT_NODES 4
#define MAX_ELEMENTS 1000000

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];
int Element_isOnTheOtherSide[MAX_ELEMENTS];



void Element_getCenter 
(int elementId, 
 double *centerX_OUT, double *centerY_OUT, double *centerZ_OUT)
{
  int iElementNode;

  assert(0 <= elementId);
  assert(elementId < NElements);

  *centerX_OUT = 0.0;
  *centerY_OUT = 0.0;
  *centerZ_OUT = 0.0;
  for (iElementNode = 0; 
       iElementNode < N_ELEMENT_NODES; 
       iElementNode++) {
    int nodeId = Element_nodeIds[elementId][iElementNode];
    double x = Node_coord[nodeId][0];
    double y = Node_coord[nodeId][1];
    double z = Node_coord[nodeId][2];
    
    *centerX_OUT += x;
    *centerY_OUT += y;
    *centerZ_OUT += z;
  }
  *centerX_OUT /= N_ELEMENT_NODES;
  *centerY_OUT /= N_ELEMENT_NODES;
  *centerZ_OUT /= N_ELEMENT_NODES;
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void WriteNodeCrackFile (const char *fileName)
{
  FILE *fp;
  int iNode;
  
  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    fprintf (fp, "%d   %d\n", 
	     iNode, Node_crackType[iNode]);
  }

  fclose (fp);
}

void WriteCrackElementListFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int count;
  
  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  count = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isOnTheOtherSide[iElement]) {
      count++;
    }
  }

  fprintf (fp, "%d\n", 
	   count);
  
  count = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isOnTheOtherSide[iElement]) {
      fprintf (fp, "%d   %d\n", 
	       count, iElement);
      count++;
    }
  }

  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Crack */

double Center[3];
double XDir[3], YDir[3], ZDir[3];
double XRadius, YRadius;

#define N_DIVISIONS 100

double CrackPlaneA, CrackPlaneB, CrackPlaneC, CrackPlaneD;



void ReadSurfaceCrackFile (const char *fileName)
{
  FILE *fp;
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%lf %lf %lf", 
	  &Center[0], &Center[1], &Center[2]);
  fscanf (fp, "%lf %lf %lf", 
	  &ZDir[0], &ZDir[1], &ZDir[2]);
  fscanf (fp, "%lf %lf %lf", 
	  &XDir[0], &XDir[1], &XDir[2]);
  fscanf (fp, "%lf %lf", 
	  &XRadius, &YRadius);

  GetNormalizedVector
    (XDir[0], XDir[1], XDir[2],
     &XDir[0], &XDir[1], &XDir[2]);
  GetNormalizedVector
    (ZDir[0], ZDir[1], ZDir[2],
     &ZDir[0], &ZDir[1], &ZDir[2]);
  GetVectorProduct 
    (ZDir[0], ZDir[1], ZDir[2],
     XDir[0], XDir[1], XDir[2],
     &YDir[0], &YDir[1], &YDir[2]);

  fclose (fp);

  GetPlaneEquationToward 
    (ZDir[0], ZDir[1], ZDir[2],
     Center[0], Center[1], Center[2],
     &CrackPlaneA, &CrackPlaneB, &CrackPlaneC, &CrackPlaneD);
}



/************************
	 Command
*************************/

void IdentifyNodeCrack (void)
{
  int iNode;
  int i;

  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    double z = Node_coord[iNode][2];
    
    double crackPlaneValue 
      = CrackPlaneA * x 
      + CrackPlaneB * y 
      + CrackPlaneC * z 
      + CrackPlaneD;

    int onCrack;
    int onCrackTip;

    if (Eps < fabs (crackPlaneValue)) {
      Node_crackType[iNode] = NODE_CRACK_OTHER;
      continue;
    }

    onCrack = 0;
    onCrackTip = 0;
    for (i = 0; i < N_DIVISIONS * 4; i++) {
      double angle0 = M_PI / 2 / N_DIVISIONS * i;
      double s0 = sin (angle0);
      double c0 = cos (angle0);
      
      double x0 = Center[0] 
	+ XDir[0] * XRadius * c0 + YDir[0] * YRadius * s0;
      double y0 = Center[1] 
	+ XDir[1] * XRadius * c0 + YDir[1] * YRadius * s0;
      double z0 = Center[2] 
	+ XDir[2] * XRadius * c0 + YDir[2] * YRadius * s0;
      
      double angle1 = M_PI / 2 / N_DIVISIONS * (i + 1);
      double s1 = sin (angle1);
      double c1 = cos (angle1);
      
      double x1 = Center[0] 
	+ XDir[0] * XRadius * c1 + YDir[0] * YRadius * s1;
      double y1 = Center[1] 
	+ XDir[1] * XRadius * c1 + YDir[1] * YRadius * s1;
      double z1 = Center[2] 
	+ XDir[2] * XRadius * c1 + YDir[2] * YRadius * s1;

      if (TriContainsPoint 
	  (x0, y0, z0,
	   x1, y1, z1,
	   Center[0], Center[1], Center[2],
	   x, y, z)) {
	double area = TriArea (x0, y0, z0,
			       x1, y1, z1,
			       Center[0], Center[1], Center[2]);
	double area0 = TriArea (x0, y0, z0, 
				x1, y1, z1, 
				x, y, z);

	onCrack = 1;
	if (area0 / area < AREA_COORD_EPS) {
	  onCrackTip = 1;
	}
      }
    }

    if (onCrack) {
      if (onCrackTip) {
	Node_crackType[iNode] = NODE_CRACK_ON_CRACK_TIP;
      } else {
	Node_crackType[iNode] = NODE_CRACK_ON_CRACK;
      }
    } else {
      Node_crackType[iNode] = NODE_CRACK_OTHER;
    }
  }
}

void IdentifyElementCrack (void)
{
  int iElement;
  int iElementNode;

  for (iElement = 0; iElement < NElements; iElement++) {
    Element_isOnTheOtherSide[iElement] = 0;
  }

  for (iElement = 0; iElement < NElements; iElement++) {
    int hasNodeOnCrack;
    
    hasNodeOnCrack = 0;
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId = Element_nodeIds[iElement][iElementNode];
      
      if (Node_crackType[nodeId] == NODE_CRACK_ON_CRACK) {
	hasNodeOnCrack = 1;
	break;
      }
    }
    
    if (hasNodeOnCrack) {
      double centerX, centerY, centerZ;
      double crackPlaneValue;
      
      Element_getCenter (iElement, 
			 &centerX, &centerY, &centerZ);
      crackPlaneValue 
	= CrackPlaneA * centerX 
	+ CrackPlaneB * centerY 
	+ CrackPlaneC * centerZ 
	+ CrackPlaneD;
      if (crackPlaneValue < -Eps) {
	Element_isOnTheOtherSide[iElement] = 1;
      }
    }
  }
}

void PerformCommand (void)
{
  IdentifyNodeCrack ();
  IdentifyElementCrack ();
}



int main (int argc, char *argv[])
{
  if (argc != 6) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tetra3d_sc_crack  node_file  element_file \n"
       "        surface_crack_file \n"
       "        node_crack_file_OUT  crack_element_list_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadSurfaceCrackFile (argv[3]);

  PerformCommand ();

  WriteNodeCrackFile (argv[4]);
  WriteCrackElementListFile (argv[5]);

  return 0;
}
