/* advautomesh_h_tetra3d_twc_crack.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001

double Eps = DEFAULT_EPS;



double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}

void GetNormalizedVector 
(double x, double y, double z,
 double *nx_OUT, double *ny_OUT, double *nz_OUT)
{
  double length = sqrt (x * x + y * y + z * z);

  *nx_OUT = x / length;
  *ny_OUT = y / length;
  *nz_OUT = z / length;
}

void GetVectorProduct 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  *x_OUT = y0 * z1 - z0 * y1;
  *y_OUT = z0 * x1 - x0 * z1;
  *z_OUT = x0 * y1 - y0 * x1;
}

void GetPlaneEquation 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double x2, double y2, double z2,
 double *a_OUT, double *b_OUT, double *c_OUT, double *d_OUT)
{
  double nx, ny, nz;

  GetVectorProduct (x1 - x0, y1 - y0, z1 - z0,
		    x2 - x0, y2 - y0, z2 - z0,
		    &nx, &ny, &nz);
  GetNormalizedVector (nx, ny, nz,
		       &nx, &ny, &nz);
  *a_OUT = nx;
  *b_OUT = ny;
  *c_OUT = nz;
  *d_OUT = -(nx * x0 + ny * y0 + nz * z0);
}

void GetPlaneEquationToward 
(double nx, double ny, double nz,
 double x, double y, double z,
 double *a_OUT, double *b_OUT, double *c_OUT, double *d_OUT)
{
  double nx2, ny2, nz2;

  GetNormalizedVector (nx, ny, nz,
		       &nx2, &ny2, &nz2);
  *a_OUT = nx2;
  *b_OUT = ny2;
  *c_OUT = nz2;
  *d_OUT = -(nx2 * x + ny2 * y + nz2 * z);
}

double DistanceFromLine 
(double originX, double originY, double originZ,
 double dirX, double dirY, double dirZ,
 double x, double y, double z)
{
  double t = (x - originX) * dirX
    + (y - originY) * dirY
    + (z - originZ) * dirZ;
  
  return Distance (x, y, z,
		   originX + dirX * t, 
		   originY + dirY * t, 
		   originZ + dirZ * t);
}



/************************
	 Mesh
*************************/

/* Node */

typedef enum {
  NODE_CRACK_OTHER,
  NODE_CRACK_ON_CRACK,
  NODE_CRACK_ON_CRACK_TIP
} NodeCrackType;

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][3];
NodeCrackType Node_crackType[MAX_NODES];



/* Element */

#define N_ELEMENT_NODES 4
#define MAX_ELEMENTS 1000000

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];
int Element_isOnTheOtherSide[MAX_ELEMENTS];



void Element_getCenter 
(int elementId, 
 double *centerX_OUT, double *centerY_OUT, double *centerZ_OUT)
{
  int iElementNode;

  assert(0 <= elementId);
  assert(elementId < NElements);

  *centerX_OUT = 0.0;
  *centerY_OUT = 0.0;
  *centerZ_OUT = 0.0;
  for (iElementNode = 0; 
       iElementNode < N_ELEMENT_NODES; 
       iElementNode++) {
    int nodeId = Element_nodeIds[elementId][iElementNode];
    double x = Node_coord[nodeId][0];
    double y = Node_coord[nodeId][1];
    double z = Node_coord[nodeId][2];
    
    *centerX_OUT += x;
    *centerY_OUT += y;
    *centerZ_OUT += z;
  }
  *centerX_OUT /= N_ELEMENT_NODES;
  *centerY_OUT /= N_ELEMENT_NODES;
  *centerZ_OUT /= N_ELEMENT_NODES;
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &nodeId, 
	    &x, &y, &z);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
    Node_coord[iNode][2] = z; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void WriteNodeCrackFile (const char *fileName)
{
  FILE *fp;
  int iNode;
  
  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    fprintf (fp, "%d   %d\n", 
	     iNode, Node_crackType[iNode]);
  }

  fclose (fp);
}

void WriteCrackElementListFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int count;
  
  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  count = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isOnTheOtherSide[iElement]) {
      count++;
    }
  }

  fprintf (fp, "%d\n", 
	   count);
  
  count = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isOnTheOtherSide[iElement]) {
      fprintf (fp, "%d   %d\n", 
	       count, iElement);
      count++;
    }
  }

  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Crack */

double Vertex0_coord[3];
double Vertex1_coord[3];
double Vertex2_coord[3];
double Vertex3_coord[3];

double CrackLineOriginX, CrackLineOriginY, CrackLineOriginZ;
double CrackLineDirX, CrackLineDirY, CrackLineDirZ;

double CrackPlaneA, CrackPlaneB, CrackPlaneC, CrackPlaneD;

double SidePlaneA, SidePlaneB, SidePlaneC, SidePlaneD;



void SetCrack (void)
{
  double lx, ly, lz;
  double sx, sy, sz;

  /* crack tip line */
  CrackLineOriginX = Vertex0_coord[0];
  CrackLineOriginY = Vertex0_coord[1];
  CrackLineOriginZ = Vertex0_coord[2];
  lx = Vertex1_coord[0] - Vertex0_coord[0];
  ly = Vertex1_coord[1] - Vertex0_coord[1];
  lz = Vertex1_coord[2] - Vertex0_coord[2];
  GetNormalizedVector 
    (lx, ly, lz,
     &CrackLineDirX, &CrackLineDirY, &CrackLineDirZ);
  
  /* crack plane */
  GetPlaneEquation 
    (Vertex0_coord[0], Vertex0_coord[1], Vertex0_coord[2], 
     Vertex1_coord[0], Vertex1_coord[1], Vertex1_coord[2], 
     Vertex2_coord[0], Vertex2_coord[1], Vertex2_coord[2], 
     &CrackPlaneA, &CrackPlaneB, &CrackPlaneC, &CrackPlaneD);
  
  /* side plane */
  GetVectorProduct 
    (CrackPlaneA, CrackPlaneB, CrackPlaneC,
     CrackLineDirX, CrackLineDirY, CrackLineDirZ,
     &sx, &sy, &sz);
  GetPlaneEquationToward 
    (sx, sy, sz,
     Vertex0_coord[0], Vertex0_coord[1], Vertex0_coord[2], 
     &SidePlaneA, &SidePlaneB, &SidePlaneC, &SidePlaneD);
}

void ReadThroughWallCrackFile (const char *fileName)
{
  FILE *fp;
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%lf %lf %lf", 
	  &Vertex0_coord[0], &Vertex0_coord[1], &Vertex0_coord[2]);
  fscanf (fp, "%lf %lf %lf", 
	  &Vertex1_coord[0], &Vertex1_coord[1], &Vertex1_coord[2]);
  fscanf (fp, "%lf %lf %lf", 
	  &Vertex2_coord[0], &Vertex2_coord[1], &Vertex2_coord[2]);
  fscanf (fp, "%lf %lf %lf", 
	  &Vertex3_coord[0], &Vertex3_coord[1], &Vertex3_coord[2]);

  fclose (fp);

  SetCrack ();
}



/************************
	 Command
*************************/

void IdentifyNodeCrack (void)
{
  int iNode;

  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    double z = Node_coord[iNode][2];
    
    double crackPlaneValue 
      = CrackPlaneA * x 
      + CrackPlaneB * y 
      + CrackPlaneC * z 
      + CrackPlaneD;
    double sidePlaneValue 
      = SidePlaneA * x 
      + SidePlaneB * y 
      + SidePlaneC * z 
      + SidePlaneD;

    if (DistanceFromLine 
	(CrackLineOriginX, CrackLineOriginY, CrackLineOriginZ,
	 CrackLineDirX, CrackLineDirY, CrackLineDirZ,
	 x, y, z) < Eps) {
      Node_crackType[iNode] = NODE_CRACK_ON_CRACK_TIP;
    } else if (fabs (crackPlaneValue) < Eps && Eps < sidePlaneValue) {
      Node_crackType[iNode] = NODE_CRACK_ON_CRACK;
    } else {
      Node_crackType[iNode] = NODE_CRACK_OTHER;
    }
  }
}

void IdentifyElementCrack (void)
{
  int iElement;
  int iElementNode;

  for (iElement = 0; iElement < NElements; iElement++) {
    Element_isOnTheOtherSide[iElement] = 0;
  }

  for (iElement = 0; iElement < NElements; iElement++) {
    int hasNodeOnCrack;
    
    hasNodeOnCrack = 0;
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId = Element_nodeIds[iElement][iElementNode];
      
      if (Node_crackType[nodeId] == NODE_CRACK_ON_CRACK) {
	hasNodeOnCrack = 1;
	break;
      }
    }
    
    if (hasNodeOnCrack) {
      double centerX, centerY, centerZ;
      double crackPlaneValue;
      
      Element_getCenter (iElement, 
			 &centerX, &centerY, &centerZ);
      crackPlaneValue 
	= CrackPlaneA * centerX 
	+ CrackPlaneB * centerY 
	+ CrackPlaneC * centerZ 
	+ CrackPlaneD;
      if (crackPlaneValue < -Eps) {
	Element_isOnTheOtherSide[iElement] = 1;
      }
    }
  }
}

void PerformCommand (void)
{
  IdentifyNodeCrack ();
  IdentifyElementCrack ();
}



int main (int argc, char *argv[])
{
  if (argc != 6) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tetra3d_twc_crack  node_file  element_file \n"
       "        through_wall_crack_file \n"
       "        node_crack_file_OUT  crack_element_list_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadThroughWallCrackFile (argv[3]);

  PerformCommand ();

  WriteNodeCrackFile (argv[4]);
  WriteCrackElementListFile (argv[5]);

  return 0;
}
