/* advautomesh_h_tri2d_node_crack_view.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autogl.h>



/************************
	 Basic
*************************/

double Distance (double x0, double y0,
		 double x1, double y1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0));
}



/************************
	 Mesh
*************************/

/* Node */

typedef enum {
  NODE_CRACK_OTHER,
  NODE_CRACK_ON_CRACK,
  NODE_CRACK_ON_CRACK_TIP
} NodeCrackType;

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][2];
NodeCrackType Node_crackType[MAX_NODES];



void Mesh_getRange
(double *minRangeX_OUT, double *minRangeY_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT)
{
  int iNode;

  assert(0 < NNodes);

  *minRangeX_OUT = *maxRangeX_OUT = Node_coord[0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Node_coord[0][1];
  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    
    if (x < *minRangeX_OUT) *minRangeX_OUT = x;
    if (y < *minRangeY_OUT) *minRangeY_OUT = y;

    if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
    if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
  }
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y;

    fscanf (fp, "%d %lf %lf", 
	    &nodeId, 
	    &x, &y);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
  }

  fclose (fp);
}

void ReadNodeCrackFile (const char *fileName)
{
  FILE *fp;
  int count;
  int iNode;
  
  assert(0 < NNodes);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &count);
  assert(count == NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    NodeCrackType crackType;

    fscanf (fp, "%d %d", 
	    &nodeId,
	    (int *)&crackType);
    assert(nodeId == iNode);
    switch (crackType) {
    case NODE_CRACK_OTHER:
    case NODE_CRACK_ON_CRACK:
    case NODE_CRACK_ON_CRACK_TIP:
      Node_crackType[iNode] = crackType;
      break;
    default:
      assert(0);
      break;
    }
  }

  fclose (fp);
}



/********************************
	 Visualization
********************************/

int NodeId = 0;
int CrackFlag = 0;

void SetView (void)
{
  double minRangeX, minRangeY;
  double maxRangeX, maxRangeY;
  double centerX, centerY;
  double size;
  
  Mesh_getRange 
    (&minRangeX, &minRangeY,
     &maxRangeX, &maxRangeY);
  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;
  size = Distance (minRangeX, minRangeY,
		   maxRangeX, maxRangeY) / 2;

  AutoGL_SetViewSize (size);
  AutoGL_SetViewCenter (centerX, centerY, 0.0); 
}

void RedrawView (void) 
{
  int iNode;
  char buf[256];

  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1]; 

    if (CrackFlag) {
      switch (Node_crackType[iNode]) {
      case NODE_CRACK_OTHER:
	AutoGL_SetColor (1, 1, 0);
	break;
      case NODE_CRACK_ON_CRACK:
	AutoGL_SetColor (0, 1, 1);
	break;
      case NODE_CRACK_ON_CRACK_TIP:
	AutoGL_SetColor (1, 0, 0);
	break;
      default:
	assert(0);
	break;
      }
    } else {
      AutoGL_SetColor (1, 1, 1);
    }
    AutoGL_DrawCrossMark3D (x, y, 0.0,
			    3);

    if (NodeId) {
      sprintf(buf, "%d", iNode);
      AutoGL_DrawString (x, y, 0.0,
			 buf);
    }
  }
}

void UpdateModel (void) 
{
  AutoGL_DrawView ();
}

void Quit (void) 
{
  AutoGL_Terminate ();
  exit (0);
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
  if (argc != 3) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tri2d_node_crack_view  node_file \n"
       "     node_crack_file \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadNodeCrackFile (argv[2]);

  SetView ();
  
  AutoGL_SetViewRedrawCallback (RedrawView);

  AutoGL_SetDefaultCallbackInMode3D (NULL);
  AutoGL_EnableDragInMode3D ();



  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode3D ();

  AutoGL_AddCallback (Quit, "Quit");       

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");

  AutoGL_AddBoolean (&NodeId, "NodeId");
  AutoGL_AddBoolean (&CrackFlag, "CrackFlag");

  AutoGL_AddCallback (UpdateModel, "UpdateModel");      
}
