/* advautomesh_h_tri2d_ec_crack.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001

double Eps = DEFAULT_EPS;



double Distance (double x0, double y0, 
		 double x1, double y1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) + (y1 - y0) * (y1 - y0));
}

int Between (double value, double range0, double range1) 
{
  return (range0 - Eps < value && value < range1 + Eps)
    || (range1 - Eps < value && value < range0 + Eps);
}

void GetLineEquation 
(double x0, double y0, 
 double x1, double y1, 
 double *a_OUT, double *b_OUT, double *c_OUT) 
{
  double dx = x1 - x0;
  double dy = y1 - y0;

  double a0 = -dy;
  double b0 = dx;
  double c0 = dy * x0 - y0 * dx;

  double len = sqrt (a0 * a0 + b0 * b0);

  *a_OUT = a0 / len;  
  *b_OUT = b0 / len;  
  *c_OUT = c0 / len;
}

int SegmentContainsPoint 
(double sx0, double sy0, double sx1, double sy1,
 double x, double y)
{
  if (Between (x, sx0, sx1) || Between (y, sy0, sy1)) {
    double a, b, c;
    
    GetLineEquation (sx0, sy0, sx1, sy1,
		     &a, &b, &c);
    if (fabs (a * x + b * y + c) < Eps) {
      return 1;
    }
  }

  return 0;
}
     


/************************
	 Mesh
*************************/

/* Node */

typedef enum {
  NODE_CRACK_OTHER,
  NODE_CRACK_ON_CRACK,
  NODE_CRACK_ON_CRACK_TIP
} NodeCrackType;

#define MAX_NODES 1000000

int NNodes;
double Node_coord[MAX_NODES][2];
NodeCrackType Node_crackType[MAX_NODES];



/* Element */

#define N_ELEMENT_NODES 3
#define MAX_ELEMENTS 1000000

int NElements;
int Element_nodeIds[MAX_ELEMENTS][N_ELEMENT_NODES];
int Element_isOnTheOtherSide[MAX_ELEMENTS];



void Element_getCenter 
(int elementId, 
 double *centerX_OUT, double *centerY_OUT)
{
  int iElementNode;

  assert(0 <= elementId);
  assert(elementId < NElements);

  *centerX_OUT = 0.0;
  *centerY_OUT = 0.0;
  for (iElementNode = 0; 
       iElementNode < N_ELEMENT_NODES; 
       iElementNode++) {
    int nodeId = Element_nodeIds[elementId][iElementNode];
    double x = Node_coord[nodeId][0];
    double y = Node_coord[nodeId][1];
    
    *centerX_OUT += x;
    *centerY_OUT += y;
  }
  *centerX_OUT /= N_ELEMENT_NODES;
  *centerY_OUT /= N_ELEMENT_NODES;
}

void ReadNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;

  assert(NNodes == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NNodes);
  assert(0 < NNodes);
  assert(NNodes <= MAX_NODES);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    int nodeId;
    double x, y;

    fscanf (fp, "%d %lf %lf", 
	    &nodeId, 
	    &x, &y);
    assert(nodeId == iNode);
    Node_coord[iNode][0] = x;
    Node_coord[iNode][1] = y; 
  }

  fclose (fp);
}

void ReadElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int iElementNode;

  assert(0 < NNodes);
  assert(NElements == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d", 
	  &NElements);
  assert(0 < NElements);
  assert(NElements <= MAX_ELEMENTS);
  
  for (iElement = 0; iElement < NElements; iElement++) {
    int elementId;

    fscanf (fp, "%d", 
	    &elementId); 
    assert(elementId == iElement);

    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId;

      fscanf (fp, "%d", 
	      &nodeId);
      assert(0 <= nodeId);
      assert(nodeId < NNodes);
      Element_nodeIds[iElement][iElementNode] = nodeId; 
    }
  }
  
  fclose (fp);
}

void WriteNodeCrackFile (const char *fileName)
{
  FILE *fp;
  int iNode;
  
  assert(0 < NNodes);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  fprintf (fp, "%d\n", 
	   NNodes);
  
  for (iNode = 0; iNode < NNodes; iNode++) {
    fprintf (fp, "%d   %d\n", 
	     iNode, Node_crackType[iNode]);
  }

  fclose (fp);
}

void WriteCrackElementListFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  int count;
  
  assert(0 < NElements);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  count = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isOnTheOtherSide[iElement]) {
      count++;
    }
  }

  fprintf (fp, "%d\n", 
	   count);
  
  count = 0;
  for (iElement = 0; iElement < NElements; iElement++) {
    if (Element_isOnTheOtherSide[iElement]) {
      fprintf (fp, "%d   %d\n", 
	       count, iElement);
      count++;
    }
  }

  fclose (fp);
}



/************************
	 Other Data
*************************/

/* Crack */

double CrackTipX, CrackTipY;
double CrackEndX, CrackEndY;



void ReadEdgeCrackFile (const char *fileName)
{
  FILE *fp;
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%lf %lf", 
	  &CrackTipX, &CrackTipY);
  fscanf (fp, "%lf %lf", 
	  &CrackEndX, &CrackEndY);

  fclose (fp);
}



/************************
	 Command
*************************/

void IdentifyNodeCrack (void)
{
  int iNode;

  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = Node_coord[iNode][0];
    double y = Node_coord[iNode][1];
    
    if (Distance (x, y, 
		  CrackTipX, CrackTipY) < Eps) {
      Node_crackType[iNode] = NODE_CRACK_ON_CRACK_TIP;
    } else if (SegmentContainsPoint 
	       (CrackTipX, CrackTipY, CrackEndX, CrackEndY,
		x, y)) {
      Node_crackType[iNode] = NODE_CRACK_ON_CRACK;
    } else {
      Node_crackType[iNode] = NODE_CRACK_OTHER;
    }
  }
}

void IdentifyElementCrack (void)
{
  int iElement;
  int iElementNode;

  for (iElement = 0; iElement < NElements; iElement++) {
    Element_isOnTheOtherSide[iElement] = 0;
  }

  for (iElement = 0; iElement < NElements; iElement++) {
    int hasNodeOnCrack;
    
    hasNodeOnCrack = 0;
    for (iElementNode = 0; 
	 iElementNode < N_ELEMENT_NODES; 
	 iElementNode++) {
      int nodeId = Element_nodeIds[iElement][iElementNode];
      
      if (Node_crackType[nodeId] == NODE_CRACK_ON_CRACK) {
	hasNodeOnCrack = 1;
	break;
      }
    }
    
    if (hasNodeOnCrack) {
      double centerX, centerY;
      double a, b, c;
      
      Element_getCenter (iElement, 
			 &centerX, &centerY);
      GetLineEquation (CrackTipX, CrackTipY, CrackEndX, CrackEndY,
		       &a, &b, &c);
      if (Eps < a * centerX + b * centerY + c) {
	Element_isOnTheOtherSide[iElement] = 1;
      }
    }
  }
}

void PerformCommand (void)
{
  IdentifyNodeCrack ();
  IdentifyElementCrack ();
}



int main (int argc, char *argv[])
{
  if (argc != 6) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautomesh_h_tri2d_ec_crack  node_file  element_file \n"
       "        edge_crack_file \n"
       "        node_crack_file_OUT  crack_element_list_file_OUT \n");
    exit (1);
  }

  ReadNodeFile (argv[1]);
  ReadElementFile (argv[2]);
  ReadEdgeCrackFile (argv[3]);

  PerformCommand ();

  WriteNodeCrackFile (argv[4]);
  WriteCrackElementListFile (argv[5]);

  return 0;
}
